"""
Autogenerated Enums for Q-Chem LevelOfTheory, TaskType, and CalcType
Do not edit this by hand. Edit generate.py or types.py instead
"""
from emmet.core.utils import ValueEnum


class LevelOfTheory(ValueEnum):
    """ Levels of theory for calculations in Q-Chem """

    PBE_6_31g_d_VACUUM = "PBE/6-31g*/VACUUM"
    PBE_6_31g_d_PCM_WATER_ = "PBE/6-31g*/PCM(WATER)"
    PBE_6_31g_d_PCM_THF_ = "PBE/6-31g*/PCM(THF)"
    PBE_6_31g_d_PCM_G2_ = "PBE/6-31g*/PCM(G2)"
    PBE_6_31g_d_PCM_DME_ = "PBE/6-31g*/PCM(DME)"
    PBE_6_31g_d_PCM_EC_ = "PBE/6-31g*/PCM(EC)"
    PBE_6_31g_d_PCM_EC_EMC_ = "PBE/6-31g*/PCM(EC/EMC)"
    PBE_6_31g_d_PCM_PC_ = "PBE/6-31g*/PCM(PC)"
    PBE_6_31g_d_ISOSVP_WATER_ = "PBE/6-31g*/ISOSVP(WATER)"
    PBE_6_31g_d_ISOSVP_THF_ = "PBE/6-31g*/ISOSVP(THF)"
    PBE_6_31g_d_ISOSVP_G2_ = "PBE/6-31g*/ISOSVP(G2)"
    PBE_6_31g_d_ISOSVP_DME_ = "PBE/6-31g*/ISOSVP(DME)"
    PBE_6_31g_d_ISOSVP_EC_ = "PBE/6-31g*/ISOSVP(EC)"
    PBE_6_31g_d_ISOSVP_EC_EMC_ = "PBE/6-31g*/ISOSVP(EC/EMC)"
    PBE_6_31g_d_ISOSVP_PC_ = "PBE/6-31g*/ISOSVP(PC)"
    PBE_6_31g_d_SMD_WATER_ = "PBE/6-31g*/SMD(WATER)"
    PBE_6_31g_d_SMD_THF_ = "PBE/6-31g*/SMD(THF)"
    PBE_6_31g_d_SMD_G2_ = "PBE/6-31g*/SMD(G2)"
    PBE_6_31g_d_SMD_DME_ = "PBE/6-31g*/SMD(DME)"
    PBE_6_31g_d_SMD_EC_ = "PBE/6-31g*/SMD(EC)"
    PBE_6_31g_d_SMD_EC_EMC_ = "PBE/6-31g*/SMD(EC/EMC)"
    PBE_6_31g_d_SMD_PC_ = "PBE/6-31g*/SMD(PC)"
    PBE_6_31g_d_CMIRS_WATER_ = "PBE/6-31g*/CMIRS(WATER)"
    PBE_6_31g_d_CMIRS_THF_ = "PBE/6-31g*/CMIRS(THF)"
    PBE_6_31g_d_CMIRS_G2_ = "PBE/6-31g*/CMIRS(G2)"
    PBE_6_31g_d_CMIRS_DME_ = "PBE/6-31g*/CMIRS(DME)"
    PBE_6_31g_d_CMIRS_EC_ = "PBE/6-31g*/CMIRS(EC)"
    PBE_6_31g_d_CMIRS_EC_EMC_ = "PBE/6-31g*/CMIRS(EC/EMC)"
    PBE_6_31g_d_CMIRS_PC_ = "PBE/6-31g*/CMIRS(PC)"
    PBE_def2_SVPD_VACUUM = "PBE/def2-SVPD/VACUUM"
    PBE_def2_SVPD_PCM_WATER_ = "PBE/def2-SVPD/PCM(WATER)"
    PBE_def2_SVPD_PCM_THF_ = "PBE/def2-SVPD/PCM(THF)"
    PBE_def2_SVPD_PCM_G2_ = "PBE/def2-SVPD/PCM(G2)"
    PBE_def2_SVPD_PCM_DME_ = "PBE/def2-SVPD/PCM(DME)"
    PBE_def2_SVPD_PCM_EC_ = "PBE/def2-SVPD/PCM(EC)"
    PBE_def2_SVPD_PCM_EC_EMC_ = "PBE/def2-SVPD/PCM(EC/EMC)"
    PBE_def2_SVPD_PCM_PC_ = "PBE/def2-SVPD/PCM(PC)"
    PBE_def2_SVPD_ISOSVP_WATER_ = "PBE/def2-SVPD/ISOSVP(WATER)"
    PBE_def2_SVPD_ISOSVP_THF_ = "PBE/def2-SVPD/ISOSVP(THF)"
    PBE_def2_SVPD_ISOSVP_G2_ = "PBE/def2-SVPD/ISOSVP(G2)"
    PBE_def2_SVPD_ISOSVP_DME_ = "PBE/def2-SVPD/ISOSVP(DME)"
    PBE_def2_SVPD_ISOSVP_EC_ = "PBE/def2-SVPD/ISOSVP(EC)"
    PBE_def2_SVPD_ISOSVP_EC_EMC_ = "PBE/def2-SVPD/ISOSVP(EC/EMC)"
    PBE_def2_SVPD_ISOSVP_PC_ = "PBE/def2-SVPD/ISOSVP(PC)"
    PBE_def2_SVPD_SMD_WATER_ = "PBE/def2-SVPD/SMD(WATER)"
    PBE_def2_SVPD_SMD_THF_ = "PBE/def2-SVPD/SMD(THF)"
    PBE_def2_SVPD_SMD_G2_ = "PBE/def2-SVPD/SMD(G2)"
    PBE_def2_SVPD_SMD_DME_ = "PBE/def2-SVPD/SMD(DME)"
    PBE_def2_SVPD_SMD_EC_ = "PBE/def2-SVPD/SMD(EC)"
    PBE_def2_SVPD_SMD_EC_EMC_ = "PBE/def2-SVPD/SMD(EC/EMC)"
    PBE_def2_SVPD_SMD_PC_ = "PBE/def2-SVPD/SMD(PC)"
    PBE_def2_SVPD_CMIRS_WATER_ = "PBE/def2-SVPD/CMIRS(WATER)"
    PBE_def2_SVPD_CMIRS_THF_ = "PBE/def2-SVPD/CMIRS(THF)"
    PBE_def2_SVPD_CMIRS_G2_ = "PBE/def2-SVPD/CMIRS(G2)"
    PBE_def2_SVPD_CMIRS_DME_ = "PBE/def2-SVPD/CMIRS(DME)"
    PBE_def2_SVPD_CMIRS_EC_ = "PBE/def2-SVPD/CMIRS(EC)"
    PBE_def2_SVPD_CMIRS_EC_EMC_ = "PBE/def2-SVPD/CMIRS(EC/EMC)"
    PBE_def2_SVPD_CMIRS_PC_ = "PBE/def2-SVPD/CMIRS(PC)"
    PBE_def2_TZVP_VACUUM = "PBE/def2-TZVP/VACUUM"
    PBE_def2_TZVP_PCM_WATER_ = "PBE/def2-TZVP/PCM(WATER)"
    PBE_def2_TZVP_PCM_THF_ = "PBE/def2-TZVP/PCM(THF)"
    PBE_def2_TZVP_PCM_G2_ = "PBE/def2-TZVP/PCM(G2)"
    PBE_def2_TZVP_PCM_DME_ = "PBE/def2-TZVP/PCM(DME)"
    PBE_def2_TZVP_PCM_EC_ = "PBE/def2-TZVP/PCM(EC)"
    PBE_def2_TZVP_PCM_EC_EMC_ = "PBE/def2-TZVP/PCM(EC/EMC)"
    PBE_def2_TZVP_PCM_PC_ = "PBE/def2-TZVP/PCM(PC)"
    PBE_def2_TZVP_ISOSVP_WATER_ = "PBE/def2-TZVP/ISOSVP(WATER)"
    PBE_def2_TZVP_ISOSVP_THF_ = "PBE/def2-TZVP/ISOSVP(THF)"
    PBE_def2_TZVP_ISOSVP_G2_ = "PBE/def2-TZVP/ISOSVP(G2)"
    PBE_def2_TZVP_ISOSVP_DME_ = "PBE/def2-TZVP/ISOSVP(DME)"
    PBE_def2_TZVP_ISOSVP_EC_ = "PBE/def2-TZVP/ISOSVP(EC)"
    PBE_def2_TZVP_ISOSVP_EC_EMC_ = "PBE/def2-TZVP/ISOSVP(EC/EMC)"
    PBE_def2_TZVP_ISOSVP_PC_ = "PBE/def2-TZVP/ISOSVP(PC)"
    PBE_def2_TZVP_SMD_WATER_ = "PBE/def2-TZVP/SMD(WATER)"
    PBE_def2_TZVP_SMD_THF_ = "PBE/def2-TZVP/SMD(THF)"
    PBE_def2_TZVP_SMD_G2_ = "PBE/def2-TZVP/SMD(G2)"
    PBE_def2_TZVP_SMD_DME_ = "PBE/def2-TZVP/SMD(DME)"
    PBE_def2_TZVP_SMD_EC_ = "PBE/def2-TZVP/SMD(EC)"
    PBE_def2_TZVP_SMD_EC_EMC_ = "PBE/def2-TZVP/SMD(EC/EMC)"
    PBE_def2_TZVP_SMD_PC_ = "PBE/def2-TZVP/SMD(PC)"
    PBE_def2_TZVP_CMIRS_WATER_ = "PBE/def2-TZVP/CMIRS(WATER)"
    PBE_def2_TZVP_CMIRS_THF_ = "PBE/def2-TZVP/CMIRS(THF)"
    PBE_def2_TZVP_CMIRS_G2_ = "PBE/def2-TZVP/CMIRS(G2)"
    PBE_def2_TZVP_CMIRS_DME_ = "PBE/def2-TZVP/CMIRS(DME)"
    PBE_def2_TZVP_CMIRS_EC_ = "PBE/def2-TZVP/CMIRS(EC)"
    PBE_def2_TZVP_CMIRS_EC_EMC_ = "PBE/def2-TZVP/CMIRS(EC/EMC)"
    PBE_def2_TZVP_CMIRS_PC_ = "PBE/def2-TZVP/CMIRS(PC)"
    PBE_def2_TZVPD_VACUUM = "PBE/def2-TZVPD/VACUUM"
    PBE_def2_TZVPD_PCM_WATER_ = "PBE/def2-TZVPD/PCM(WATER)"
    PBE_def2_TZVPD_PCM_THF_ = "PBE/def2-TZVPD/PCM(THF)"
    PBE_def2_TZVPD_PCM_G2_ = "PBE/def2-TZVPD/PCM(G2)"
    PBE_def2_TZVPD_PCM_DME_ = "PBE/def2-TZVPD/PCM(DME)"
    PBE_def2_TZVPD_PCM_EC_ = "PBE/def2-TZVPD/PCM(EC)"
    PBE_def2_TZVPD_PCM_EC_EMC_ = "PBE/def2-TZVPD/PCM(EC/EMC)"
    PBE_def2_TZVPD_PCM_PC_ = "PBE/def2-TZVPD/PCM(PC)"
    PBE_def2_TZVPD_ISOSVP_WATER_ = "PBE/def2-TZVPD/ISOSVP(WATER)"
    PBE_def2_TZVPD_ISOSVP_THF_ = "PBE/def2-TZVPD/ISOSVP(THF)"
    PBE_def2_TZVPD_ISOSVP_G2_ = "PBE/def2-TZVPD/ISOSVP(G2)"
    PBE_def2_TZVPD_ISOSVP_DME_ = "PBE/def2-TZVPD/ISOSVP(DME)"
    PBE_def2_TZVPD_ISOSVP_EC_ = "PBE/def2-TZVPD/ISOSVP(EC)"
    PBE_def2_TZVPD_ISOSVP_EC_EMC_ = "PBE/def2-TZVPD/ISOSVP(EC/EMC)"
    PBE_def2_TZVPD_ISOSVP_PC_ = "PBE/def2-TZVPD/ISOSVP(PC)"
    PBE_def2_TZVPD_SMD_WATER_ = "PBE/def2-TZVPD/SMD(WATER)"
    PBE_def2_TZVPD_SMD_THF_ = "PBE/def2-TZVPD/SMD(THF)"
    PBE_def2_TZVPD_SMD_G2_ = "PBE/def2-TZVPD/SMD(G2)"
    PBE_def2_TZVPD_SMD_DME_ = "PBE/def2-TZVPD/SMD(DME)"
    PBE_def2_TZVPD_SMD_EC_ = "PBE/def2-TZVPD/SMD(EC)"
    PBE_def2_TZVPD_SMD_EC_EMC_ = "PBE/def2-TZVPD/SMD(EC/EMC)"
    PBE_def2_TZVPD_SMD_PC_ = "PBE/def2-TZVPD/SMD(PC)"
    PBE_def2_TZVPD_CMIRS_WATER_ = "PBE/def2-TZVPD/CMIRS(WATER)"
    PBE_def2_TZVPD_CMIRS_THF_ = "PBE/def2-TZVPD/CMIRS(THF)"
    PBE_def2_TZVPD_CMIRS_G2_ = "PBE/def2-TZVPD/CMIRS(G2)"
    PBE_def2_TZVPD_CMIRS_DME_ = "PBE/def2-TZVPD/CMIRS(DME)"
    PBE_def2_TZVPD_CMIRS_EC_ = "PBE/def2-TZVPD/CMIRS(EC)"
    PBE_def2_TZVPD_CMIRS_EC_EMC_ = "PBE/def2-TZVPD/CMIRS(EC/EMC)"
    PBE_def2_TZVPD_CMIRS_PC_ = "PBE/def2-TZVPD/CMIRS(PC)"
    PBE_def2_TZVPP_VACUUM = "PBE/def2-TZVPP/VACUUM"
    PBE_def2_TZVPP_PCM_WATER_ = "PBE/def2-TZVPP/PCM(WATER)"
    PBE_def2_TZVPP_PCM_THF_ = "PBE/def2-TZVPP/PCM(THF)"
    PBE_def2_TZVPP_PCM_G2_ = "PBE/def2-TZVPP/PCM(G2)"
    PBE_def2_TZVPP_PCM_DME_ = "PBE/def2-TZVPP/PCM(DME)"
    PBE_def2_TZVPP_PCM_EC_ = "PBE/def2-TZVPP/PCM(EC)"
    PBE_def2_TZVPP_PCM_EC_EMC_ = "PBE/def2-TZVPP/PCM(EC/EMC)"
    PBE_def2_TZVPP_PCM_PC_ = "PBE/def2-TZVPP/PCM(PC)"
    PBE_def2_TZVPP_ISOSVP_WATER_ = "PBE/def2-TZVPP/ISOSVP(WATER)"
    PBE_def2_TZVPP_ISOSVP_THF_ = "PBE/def2-TZVPP/ISOSVP(THF)"
    PBE_def2_TZVPP_ISOSVP_G2_ = "PBE/def2-TZVPP/ISOSVP(G2)"
    PBE_def2_TZVPP_ISOSVP_DME_ = "PBE/def2-TZVPP/ISOSVP(DME)"
    PBE_def2_TZVPP_ISOSVP_EC_ = "PBE/def2-TZVPP/ISOSVP(EC)"
    PBE_def2_TZVPP_ISOSVP_EC_EMC_ = "PBE/def2-TZVPP/ISOSVP(EC/EMC)"
    PBE_def2_TZVPP_ISOSVP_PC_ = "PBE/def2-TZVPP/ISOSVP(PC)"
    PBE_def2_TZVPP_SMD_WATER_ = "PBE/def2-TZVPP/SMD(WATER)"
    PBE_def2_TZVPP_SMD_THF_ = "PBE/def2-TZVPP/SMD(THF)"
    PBE_def2_TZVPP_SMD_G2_ = "PBE/def2-TZVPP/SMD(G2)"
    PBE_def2_TZVPP_SMD_DME_ = "PBE/def2-TZVPP/SMD(DME)"
    PBE_def2_TZVPP_SMD_EC_ = "PBE/def2-TZVPP/SMD(EC)"
    PBE_def2_TZVPP_SMD_EC_EMC_ = "PBE/def2-TZVPP/SMD(EC/EMC)"
    PBE_def2_TZVPP_SMD_PC_ = "PBE/def2-TZVPP/SMD(PC)"
    PBE_def2_TZVPP_CMIRS_WATER_ = "PBE/def2-TZVPP/CMIRS(WATER)"
    PBE_def2_TZVPP_CMIRS_THF_ = "PBE/def2-TZVPP/CMIRS(THF)"
    PBE_def2_TZVPP_CMIRS_G2_ = "PBE/def2-TZVPP/CMIRS(G2)"
    PBE_def2_TZVPP_CMIRS_DME_ = "PBE/def2-TZVPP/CMIRS(DME)"
    PBE_def2_TZVPP_CMIRS_EC_ = "PBE/def2-TZVPP/CMIRS(EC)"
    PBE_def2_TZVPP_CMIRS_EC_EMC_ = "PBE/def2-TZVPP/CMIRS(EC/EMC)"
    PBE_def2_TZVPP_CMIRS_PC_ = "PBE/def2-TZVPP/CMIRS(PC)"
    PBE_def2_TZVPPD_VACUUM = "PBE/def2-TZVPPD/VACUUM"
    PBE_def2_TZVPPD_PCM_WATER_ = "PBE/def2-TZVPPD/PCM(WATER)"
    PBE_def2_TZVPPD_PCM_THF_ = "PBE/def2-TZVPPD/PCM(THF)"
    PBE_def2_TZVPPD_PCM_G2_ = "PBE/def2-TZVPPD/PCM(G2)"
    PBE_def2_TZVPPD_PCM_DME_ = "PBE/def2-TZVPPD/PCM(DME)"
    PBE_def2_TZVPPD_PCM_EC_ = "PBE/def2-TZVPPD/PCM(EC)"
    PBE_def2_TZVPPD_PCM_EC_EMC_ = "PBE/def2-TZVPPD/PCM(EC/EMC)"
    PBE_def2_TZVPPD_PCM_PC_ = "PBE/def2-TZVPPD/PCM(PC)"
    PBE_def2_TZVPPD_ISOSVP_WATER_ = "PBE/def2-TZVPPD/ISOSVP(WATER)"
    PBE_def2_TZVPPD_ISOSVP_THF_ = "PBE/def2-TZVPPD/ISOSVP(THF)"
    PBE_def2_TZVPPD_ISOSVP_G2_ = "PBE/def2-TZVPPD/ISOSVP(G2)"
    PBE_def2_TZVPPD_ISOSVP_DME_ = "PBE/def2-TZVPPD/ISOSVP(DME)"
    PBE_def2_TZVPPD_ISOSVP_EC_ = "PBE/def2-TZVPPD/ISOSVP(EC)"
    PBE_def2_TZVPPD_ISOSVP_EC_EMC_ = "PBE/def2-TZVPPD/ISOSVP(EC/EMC)"
    PBE_def2_TZVPPD_ISOSVP_PC_ = "PBE/def2-TZVPPD/ISOSVP(PC)"
    PBE_def2_TZVPPD_SMD_WATER_ = "PBE/def2-TZVPPD/SMD(WATER)"
    PBE_def2_TZVPPD_SMD_THF_ = "PBE/def2-TZVPPD/SMD(THF)"
    PBE_def2_TZVPPD_SMD_G2_ = "PBE/def2-TZVPPD/SMD(G2)"
    PBE_def2_TZVPPD_SMD_DME_ = "PBE/def2-TZVPPD/SMD(DME)"
    PBE_def2_TZVPPD_SMD_EC_ = "PBE/def2-TZVPPD/SMD(EC)"
    PBE_def2_TZVPPD_SMD_EC_EMC_ = "PBE/def2-TZVPPD/SMD(EC/EMC)"
    PBE_def2_TZVPPD_SMD_PC_ = "PBE/def2-TZVPPD/SMD(PC)"
    PBE_def2_TZVPPD_CMIRS_WATER_ = "PBE/def2-TZVPPD/CMIRS(WATER)"
    PBE_def2_TZVPPD_CMIRS_THF_ = "PBE/def2-TZVPPD/CMIRS(THF)"
    PBE_def2_TZVPPD_CMIRS_G2_ = "PBE/def2-TZVPPD/CMIRS(G2)"
    PBE_def2_TZVPPD_CMIRS_DME_ = "PBE/def2-TZVPPD/CMIRS(DME)"
    PBE_def2_TZVPPD_CMIRS_EC_ = "PBE/def2-TZVPPD/CMIRS(EC)"
    PBE_def2_TZVPPD_CMIRS_EC_EMC_ = "PBE/def2-TZVPPD/CMIRS(EC/EMC)"
    PBE_def2_TZVPPD_CMIRS_PC_ = "PBE/def2-TZVPPD/CMIRS(PC)"
    PBE_def2_QZVPPD_VACUUM = "PBE/def2-QZVPPD/VACUUM"
    PBE_def2_QZVPPD_PCM_WATER_ = "PBE/def2-QZVPPD/PCM(WATER)"
    PBE_def2_QZVPPD_PCM_THF_ = "PBE/def2-QZVPPD/PCM(THF)"
    PBE_def2_QZVPPD_PCM_G2_ = "PBE/def2-QZVPPD/PCM(G2)"
    PBE_def2_QZVPPD_PCM_DME_ = "PBE/def2-QZVPPD/PCM(DME)"
    PBE_def2_QZVPPD_PCM_EC_ = "PBE/def2-QZVPPD/PCM(EC)"
    PBE_def2_QZVPPD_PCM_EC_EMC_ = "PBE/def2-QZVPPD/PCM(EC/EMC)"
    PBE_def2_QZVPPD_PCM_PC_ = "PBE/def2-QZVPPD/PCM(PC)"
    PBE_def2_QZVPPD_ISOSVP_WATER_ = "PBE/def2-QZVPPD/ISOSVP(WATER)"
    PBE_def2_QZVPPD_ISOSVP_THF_ = "PBE/def2-QZVPPD/ISOSVP(THF)"
    PBE_def2_QZVPPD_ISOSVP_G2_ = "PBE/def2-QZVPPD/ISOSVP(G2)"
    PBE_def2_QZVPPD_ISOSVP_DME_ = "PBE/def2-QZVPPD/ISOSVP(DME)"
    PBE_def2_QZVPPD_ISOSVP_EC_ = "PBE/def2-QZVPPD/ISOSVP(EC)"
    PBE_def2_QZVPPD_ISOSVP_EC_EMC_ = "PBE/def2-QZVPPD/ISOSVP(EC/EMC)"
    PBE_def2_QZVPPD_ISOSVP_PC_ = "PBE/def2-QZVPPD/ISOSVP(PC)"
    PBE_def2_QZVPPD_SMD_WATER_ = "PBE/def2-QZVPPD/SMD(WATER)"
    PBE_def2_QZVPPD_SMD_THF_ = "PBE/def2-QZVPPD/SMD(THF)"
    PBE_def2_QZVPPD_SMD_G2_ = "PBE/def2-QZVPPD/SMD(G2)"
    PBE_def2_QZVPPD_SMD_DME_ = "PBE/def2-QZVPPD/SMD(DME)"
    PBE_def2_QZVPPD_SMD_EC_ = "PBE/def2-QZVPPD/SMD(EC)"
    PBE_def2_QZVPPD_SMD_EC_EMC_ = "PBE/def2-QZVPPD/SMD(EC/EMC)"
    PBE_def2_QZVPPD_SMD_PC_ = "PBE/def2-QZVPPD/SMD(PC)"
    PBE_def2_QZVPPD_CMIRS_WATER_ = "PBE/def2-QZVPPD/CMIRS(WATER)"
    PBE_def2_QZVPPD_CMIRS_THF_ = "PBE/def2-QZVPPD/CMIRS(THF)"
    PBE_def2_QZVPPD_CMIRS_G2_ = "PBE/def2-QZVPPD/CMIRS(G2)"
    PBE_def2_QZVPPD_CMIRS_DME_ = "PBE/def2-QZVPPD/CMIRS(DME)"
    PBE_def2_QZVPPD_CMIRS_EC_ = "PBE/def2-QZVPPD/CMIRS(EC)"
    PBE_def2_QZVPPD_CMIRS_EC_EMC_ = "PBE/def2-QZVPPD/CMIRS(EC/EMC)"
    PBE_def2_QZVPPD_CMIRS_PC_ = "PBE/def2-QZVPPD/CMIRS(PC)"
    B97_D_6_31g_d_VACUUM = "B97-D/6-31g*/VACUUM"
    B97_D_6_31g_d_PCM_WATER_ = "B97-D/6-31g*/PCM(WATER)"
    B97_D_6_31g_d_PCM_THF_ = "B97-D/6-31g*/PCM(THF)"
    B97_D_6_31g_d_PCM_G2_ = "B97-D/6-31g*/PCM(G2)"
    B97_D_6_31g_d_PCM_DME_ = "B97-D/6-31g*/PCM(DME)"
    B97_D_6_31g_d_PCM_EC_ = "B97-D/6-31g*/PCM(EC)"
    B97_D_6_31g_d_PCM_EC_EMC_ = "B97-D/6-31g*/PCM(EC/EMC)"
    B97_D_6_31g_d_PCM_PC_ = "B97-D/6-31g*/PCM(PC)"
    B97_D_6_31g_d_ISOSVP_WATER_ = "B97-D/6-31g*/ISOSVP(WATER)"
    B97_D_6_31g_d_ISOSVP_THF_ = "B97-D/6-31g*/ISOSVP(THF)"
    B97_D_6_31g_d_ISOSVP_G2_ = "B97-D/6-31g*/ISOSVP(G2)"
    B97_D_6_31g_d_ISOSVP_DME_ = "B97-D/6-31g*/ISOSVP(DME)"
    B97_D_6_31g_d_ISOSVP_EC_ = "B97-D/6-31g*/ISOSVP(EC)"
    B97_D_6_31g_d_ISOSVP_EC_EMC_ = "B97-D/6-31g*/ISOSVP(EC/EMC)"
    B97_D_6_31g_d_ISOSVP_PC_ = "B97-D/6-31g*/ISOSVP(PC)"
    B97_D_6_31g_d_SMD_WATER_ = "B97-D/6-31g*/SMD(WATER)"
    B97_D_6_31g_d_SMD_THF_ = "B97-D/6-31g*/SMD(THF)"
    B97_D_6_31g_d_SMD_G2_ = "B97-D/6-31g*/SMD(G2)"
    B97_D_6_31g_d_SMD_DME_ = "B97-D/6-31g*/SMD(DME)"
    B97_D_6_31g_d_SMD_EC_ = "B97-D/6-31g*/SMD(EC)"
    B97_D_6_31g_d_SMD_EC_EMC_ = "B97-D/6-31g*/SMD(EC/EMC)"
    B97_D_6_31g_d_SMD_PC_ = "B97-D/6-31g*/SMD(PC)"
    B97_D_6_31g_d_CMIRS_WATER_ = "B97-D/6-31g*/CMIRS(WATER)"
    B97_D_6_31g_d_CMIRS_THF_ = "B97-D/6-31g*/CMIRS(THF)"
    B97_D_6_31g_d_CMIRS_G2_ = "B97-D/6-31g*/CMIRS(G2)"
    B97_D_6_31g_d_CMIRS_DME_ = "B97-D/6-31g*/CMIRS(DME)"
    B97_D_6_31g_d_CMIRS_EC_ = "B97-D/6-31g*/CMIRS(EC)"
    B97_D_6_31g_d_CMIRS_EC_EMC_ = "B97-D/6-31g*/CMIRS(EC/EMC)"
    B97_D_6_31g_d_CMIRS_PC_ = "B97-D/6-31g*/CMIRS(PC)"
    B97_D_def2_SVPD_VACUUM = "B97-D/def2-SVPD/VACUUM"
    B97_D_def2_SVPD_PCM_WATER_ = "B97-D/def2-SVPD/PCM(WATER)"
    B97_D_def2_SVPD_PCM_THF_ = "B97-D/def2-SVPD/PCM(THF)"
    B97_D_def2_SVPD_PCM_G2_ = "B97-D/def2-SVPD/PCM(G2)"
    B97_D_def2_SVPD_PCM_DME_ = "B97-D/def2-SVPD/PCM(DME)"
    B97_D_def2_SVPD_PCM_EC_ = "B97-D/def2-SVPD/PCM(EC)"
    B97_D_def2_SVPD_PCM_EC_EMC_ = "B97-D/def2-SVPD/PCM(EC/EMC)"
    B97_D_def2_SVPD_PCM_PC_ = "B97-D/def2-SVPD/PCM(PC)"
    B97_D_def2_SVPD_ISOSVP_WATER_ = "B97-D/def2-SVPD/ISOSVP(WATER)"
    B97_D_def2_SVPD_ISOSVP_THF_ = "B97-D/def2-SVPD/ISOSVP(THF)"
    B97_D_def2_SVPD_ISOSVP_G2_ = "B97-D/def2-SVPD/ISOSVP(G2)"
    B97_D_def2_SVPD_ISOSVP_DME_ = "B97-D/def2-SVPD/ISOSVP(DME)"
    B97_D_def2_SVPD_ISOSVP_EC_ = "B97-D/def2-SVPD/ISOSVP(EC)"
    B97_D_def2_SVPD_ISOSVP_EC_EMC_ = "B97-D/def2-SVPD/ISOSVP(EC/EMC)"
    B97_D_def2_SVPD_ISOSVP_PC_ = "B97-D/def2-SVPD/ISOSVP(PC)"
    B97_D_def2_SVPD_SMD_WATER_ = "B97-D/def2-SVPD/SMD(WATER)"
    B97_D_def2_SVPD_SMD_THF_ = "B97-D/def2-SVPD/SMD(THF)"
    B97_D_def2_SVPD_SMD_G2_ = "B97-D/def2-SVPD/SMD(G2)"
    B97_D_def2_SVPD_SMD_DME_ = "B97-D/def2-SVPD/SMD(DME)"
    B97_D_def2_SVPD_SMD_EC_ = "B97-D/def2-SVPD/SMD(EC)"
    B97_D_def2_SVPD_SMD_EC_EMC_ = "B97-D/def2-SVPD/SMD(EC/EMC)"
    B97_D_def2_SVPD_SMD_PC_ = "B97-D/def2-SVPD/SMD(PC)"
    B97_D_def2_SVPD_CMIRS_WATER_ = "B97-D/def2-SVPD/CMIRS(WATER)"
    B97_D_def2_SVPD_CMIRS_THF_ = "B97-D/def2-SVPD/CMIRS(THF)"
    B97_D_def2_SVPD_CMIRS_G2_ = "B97-D/def2-SVPD/CMIRS(G2)"
    B97_D_def2_SVPD_CMIRS_DME_ = "B97-D/def2-SVPD/CMIRS(DME)"
    B97_D_def2_SVPD_CMIRS_EC_ = "B97-D/def2-SVPD/CMIRS(EC)"
    B97_D_def2_SVPD_CMIRS_EC_EMC_ = "B97-D/def2-SVPD/CMIRS(EC/EMC)"
    B97_D_def2_SVPD_CMIRS_PC_ = "B97-D/def2-SVPD/CMIRS(PC)"
    B97_D_def2_TZVP_VACUUM = "B97-D/def2-TZVP/VACUUM"
    B97_D_def2_TZVP_PCM_WATER_ = "B97-D/def2-TZVP/PCM(WATER)"
    B97_D_def2_TZVP_PCM_THF_ = "B97-D/def2-TZVP/PCM(THF)"
    B97_D_def2_TZVP_PCM_G2_ = "B97-D/def2-TZVP/PCM(G2)"
    B97_D_def2_TZVP_PCM_DME_ = "B97-D/def2-TZVP/PCM(DME)"
    B97_D_def2_TZVP_PCM_EC_ = "B97-D/def2-TZVP/PCM(EC)"
    B97_D_def2_TZVP_PCM_EC_EMC_ = "B97-D/def2-TZVP/PCM(EC/EMC)"
    B97_D_def2_TZVP_PCM_PC_ = "B97-D/def2-TZVP/PCM(PC)"
    B97_D_def2_TZVP_ISOSVP_WATER_ = "B97-D/def2-TZVP/ISOSVP(WATER)"
    B97_D_def2_TZVP_ISOSVP_THF_ = "B97-D/def2-TZVP/ISOSVP(THF)"
    B97_D_def2_TZVP_ISOSVP_G2_ = "B97-D/def2-TZVP/ISOSVP(G2)"
    B97_D_def2_TZVP_ISOSVP_DME_ = "B97-D/def2-TZVP/ISOSVP(DME)"
    B97_D_def2_TZVP_ISOSVP_EC_ = "B97-D/def2-TZVP/ISOSVP(EC)"
    B97_D_def2_TZVP_ISOSVP_EC_EMC_ = "B97-D/def2-TZVP/ISOSVP(EC/EMC)"
    B97_D_def2_TZVP_ISOSVP_PC_ = "B97-D/def2-TZVP/ISOSVP(PC)"
    B97_D_def2_TZVP_SMD_WATER_ = "B97-D/def2-TZVP/SMD(WATER)"
    B97_D_def2_TZVP_SMD_THF_ = "B97-D/def2-TZVP/SMD(THF)"
    B97_D_def2_TZVP_SMD_G2_ = "B97-D/def2-TZVP/SMD(G2)"
    B97_D_def2_TZVP_SMD_DME_ = "B97-D/def2-TZVP/SMD(DME)"
    B97_D_def2_TZVP_SMD_EC_ = "B97-D/def2-TZVP/SMD(EC)"
    B97_D_def2_TZVP_SMD_EC_EMC_ = "B97-D/def2-TZVP/SMD(EC/EMC)"
    B97_D_def2_TZVP_SMD_PC_ = "B97-D/def2-TZVP/SMD(PC)"
    B97_D_def2_TZVP_CMIRS_WATER_ = "B97-D/def2-TZVP/CMIRS(WATER)"
    B97_D_def2_TZVP_CMIRS_THF_ = "B97-D/def2-TZVP/CMIRS(THF)"
    B97_D_def2_TZVP_CMIRS_G2_ = "B97-D/def2-TZVP/CMIRS(G2)"
    B97_D_def2_TZVP_CMIRS_DME_ = "B97-D/def2-TZVP/CMIRS(DME)"
    B97_D_def2_TZVP_CMIRS_EC_ = "B97-D/def2-TZVP/CMIRS(EC)"
    B97_D_def2_TZVP_CMIRS_EC_EMC_ = "B97-D/def2-TZVP/CMIRS(EC/EMC)"
    B97_D_def2_TZVP_CMIRS_PC_ = "B97-D/def2-TZVP/CMIRS(PC)"
    B97_D_def2_TZVPD_VACUUM = "B97-D/def2-TZVPD/VACUUM"
    B97_D_def2_TZVPD_PCM_WATER_ = "B97-D/def2-TZVPD/PCM(WATER)"
    B97_D_def2_TZVPD_PCM_THF_ = "B97-D/def2-TZVPD/PCM(THF)"
    B97_D_def2_TZVPD_PCM_G2_ = "B97-D/def2-TZVPD/PCM(G2)"
    B97_D_def2_TZVPD_PCM_DME_ = "B97-D/def2-TZVPD/PCM(DME)"
    B97_D_def2_TZVPD_PCM_EC_ = "B97-D/def2-TZVPD/PCM(EC)"
    B97_D_def2_TZVPD_PCM_EC_EMC_ = "B97-D/def2-TZVPD/PCM(EC/EMC)"
    B97_D_def2_TZVPD_PCM_PC_ = "B97-D/def2-TZVPD/PCM(PC)"
    B97_D_def2_TZVPD_ISOSVP_WATER_ = "B97-D/def2-TZVPD/ISOSVP(WATER)"
    B97_D_def2_TZVPD_ISOSVP_THF_ = "B97-D/def2-TZVPD/ISOSVP(THF)"
    B97_D_def2_TZVPD_ISOSVP_G2_ = "B97-D/def2-TZVPD/ISOSVP(G2)"
    B97_D_def2_TZVPD_ISOSVP_DME_ = "B97-D/def2-TZVPD/ISOSVP(DME)"
    B97_D_def2_TZVPD_ISOSVP_EC_ = "B97-D/def2-TZVPD/ISOSVP(EC)"
    B97_D_def2_TZVPD_ISOSVP_EC_EMC_ = "B97-D/def2-TZVPD/ISOSVP(EC/EMC)"
    B97_D_def2_TZVPD_ISOSVP_PC_ = "B97-D/def2-TZVPD/ISOSVP(PC)"
    B97_D_def2_TZVPD_SMD_WATER_ = "B97-D/def2-TZVPD/SMD(WATER)"
    B97_D_def2_TZVPD_SMD_THF_ = "B97-D/def2-TZVPD/SMD(THF)"
    B97_D_def2_TZVPD_SMD_G2_ = "B97-D/def2-TZVPD/SMD(G2)"
    B97_D_def2_TZVPD_SMD_DME_ = "B97-D/def2-TZVPD/SMD(DME)"
    B97_D_def2_TZVPD_SMD_EC_ = "B97-D/def2-TZVPD/SMD(EC)"
    B97_D_def2_TZVPD_SMD_EC_EMC_ = "B97-D/def2-TZVPD/SMD(EC/EMC)"
    B97_D_def2_TZVPD_SMD_PC_ = "B97-D/def2-TZVPD/SMD(PC)"
    B97_D_def2_TZVPD_CMIRS_WATER_ = "B97-D/def2-TZVPD/CMIRS(WATER)"
    B97_D_def2_TZVPD_CMIRS_THF_ = "B97-D/def2-TZVPD/CMIRS(THF)"
    B97_D_def2_TZVPD_CMIRS_G2_ = "B97-D/def2-TZVPD/CMIRS(G2)"
    B97_D_def2_TZVPD_CMIRS_DME_ = "B97-D/def2-TZVPD/CMIRS(DME)"
    B97_D_def2_TZVPD_CMIRS_EC_ = "B97-D/def2-TZVPD/CMIRS(EC)"
    B97_D_def2_TZVPD_CMIRS_EC_EMC_ = "B97-D/def2-TZVPD/CMIRS(EC/EMC)"
    B97_D_def2_TZVPD_CMIRS_PC_ = "B97-D/def2-TZVPD/CMIRS(PC)"
    B97_D_def2_TZVPP_VACUUM = "B97-D/def2-TZVPP/VACUUM"
    B97_D_def2_TZVPP_PCM_WATER_ = "B97-D/def2-TZVPP/PCM(WATER)"
    B97_D_def2_TZVPP_PCM_THF_ = "B97-D/def2-TZVPP/PCM(THF)"
    B97_D_def2_TZVPP_PCM_G2_ = "B97-D/def2-TZVPP/PCM(G2)"
    B97_D_def2_TZVPP_PCM_DME_ = "B97-D/def2-TZVPP/PCM(DME)"
    B97_D_def2_TZVPP_PCM_EC_ = "B97-D/def2-TZVPP/PCM(EC)"
    B97_D_def2_TZVPP_PCM_EC_EMC_ = "B97-D/def2-TZVPP/PCM(EC/EMC)"
    B97_D_def2_TZVPP_PCM_PC_ = "B97-D/def2-TZVPP/PCM(PC)"
    B97_D_def2_TZVPP_ISOSVP_WATER_ = "B97-D/def2-TZVPP/ISOSVP(WATER)"
    B97_D_def2_TZVPP_ISOSVP_THF_ = "B97-D/def2-TZVPP/ISOSVP(THF)"
    B97_D_def2_TZVPP_ISOSVP_G2_ = "B97-D/def2-TZVPP/ISOSVP(G2)"
    B97_D_def2_TZVPP_ISOSVP_DME_ = "B97-D/def2-TZVPP/ISOSVP(DME)"
    B97_D_def2_TZVPP_ISOSVP_EC_ = "B97-D/def2-TZVPP/ISOSVP(EC)"
    B97_D_def2_TZVPP_ISOSVP_EC_EMC_ = "B97-D/def2-TZVPP/ISOSVP(EC/EMC)"
    B97_D_def2_TZVPP_ISOSVP_PC_ = "B97-D/def2-TZVPP/ISOSVP(PC)"
    B97_D_def2_TZVPP_SMD_WATER_ = "B97-D/def2-TZVPP/SMD(WATER)"
    B97_D_def2_TZVPP_SMD_THF_ = "B97-D/def2-TZVPP/SMD(THF)"
    B97_D_def2_TZVPP_SMD_G2_ = "B97-D/def2-TZVPP/SMD(G2)"
    B97_D_def2_TZVPP_SMD_DME_ = "B97-D/def2-TZVPP/SMD(DME)"
    B97_D_def2_TZVPP_SMD_EC_ = "B97-D/def2-TZVPP/SMD(EC)"
    B97_D_def2_TZVPP_SMD_EC_EMC_ = "B97-D/def2-TZVPP/SMD(EC/EMC)"
    B97_D_def2_TZVPP_SMD_PC_ = "B97-D/def2-TZVPP/SMD(PC)"
    B97_D_def2_TZVPP_CMIRS_WATER_ = "B97-D/def2-TZVPP/CMIRS(WATER)"
    B97_D_def2_TZVPP_CMIRS_THF_ = "B97-D/def2-TZVPP/CMIRS(THF)"
    B97_D_def2_TZVPP_CMIRS_G2_ = "B97-D/def2-TZVPP/CMIRS(G2)"
    B97_D_def2_TZVPP_CMIRS_DME_ = "B97-D/def2-TZVPP/CMIRS(DME)"
    B97_D_def2_TZVPP_CMIRS_EC_ = "B97-D/def2-TZVPP/CMIRS(EC)"
    B97_D_def2_TZVPP_CMIRS_EC_EMC_ = "B97-D/def2-TZVPP/CMIRS(EC/EMC)"
    B97_D_def2_TZVPP_CMIRS_PC_ = "B97-D/def2-TZVPP/CMIRS(PC)"
    B97_D_def2_TZVPPD_VACUUM = "B97-D/def2-TZVPPD/VACUUM"
    B97_D_def2_TZVPPD_PCM_WATER_ = "B97-D/def2-TZVPPD/PCM(WATER)"
    B97_D_def2_TZVPPD_PCM_THF_ = "B97-D/def2-TZVPPD/PCM(THF)"
    B97_D_def2_TZVPPD_PCM_G2_ = "B97-D/def2-TZVPPD/PCM(G2)"
    B97_D_def2_TZVPPD_PCM_DME_ = "B97-D/def2-TZVPPD/PCM(DME)"
    B97_D_def2_TZVPPD_PCM_EC_ = "B97-D/def2-TZVPPD/PCM(EC)"
    B97_D_def2_TZVPPD_PCM_EC_EMC_ = "B97-D/def2-TZVPPD/PCM(EC/EMC)"
    B97_D_def2_TZVPPD_PCM_PC_ = "B97-D/def2-TZVPPD/PCM(PC)"
    B97_D_def2_TZVPPD_ISOSVP_WATER_ = "B97-D/def2-TZVPPD/ISOSVP(WATER)"
    B97_D_def2_TZVPPD_ISOSVP_THF_ = "B97-D/def2-TZVPPD/ISOSVP(THF)"
    B97_D_def2_TZVPPD_ISOSVP_G2_ = "B97-D/def2-TZVPPD/ISOSVP(G2)"
    B97_D_def2_TZVPPD_ISOSVP_DME_ = "B97-D/def2-TZVPPD/ISOSVP(DME)"
    B97_D_def2_TZVPPD_ISOSVP_EC_ = "B97-D/def2-TZVPPD/ISOSVP(EC)"
    B97_D_def2_TZVPPD_ISOSVP_EC_EMC_ = "B97-D/def2-TZVPPD/ISOSVP(EC/EMC)"
    B97_D_def2_TZVPPD_ISOSVP_PC_ = "B97-D/def2-TZVPPD/ISOSVP(PC)"
    B97_D_def2_TZVPPD_SMD_WATER_ = "B97-D/def2-TZVPPD/SMD(WATER)"
    B97_D_def2_TZVPPD_SMD_THF_ = "B97-D/def2-TZVPPD/SMD(THF)"
    B97_D_def2_TZVPPD_SMD_G2_ = "B97-D/def2-TZVPPD/SMD(G2)"
    B97_D_def2_TZVPPD_SMD_DME_ = "B97-D/def2-TZVPPD/SMD(DME)"
    B97_D_def2_TZVPPD_SMD_EC_ = "B97-D/def2-TZVPPD/SMD(EC)"
    B97_D_def2_TZVPPD_SMD_EC_EMC_ = "B97-D/def2-TZVPPD/SMD(EC/EMC)"
    B97_D_def2_TZVPPD_SMD_PC_ = "B97-D/def2-TZVPPD/SMD(PC)"
    B97_D_def2_TZVPPD_CMIRS_WATER_ = "B97-D/def2-TZVPPD/CMIRS(WATER)"
    B97_D_def2_TZVPPD_CMIRS_THF_ = "B97-D/def2-TZVPPD/CMIRS(THF)"
    B97_D_def2_TZVPPD_CMIRS_G2_ = "B97-D/def2-TZVPPD/CMIRS(G2)"
    B97_D_def2_TZVPPD_CMIRS_DME_ = "B97-D/def2-TZVPPD/CMIRS(DME)"
    B97_D_def2_TZVPPD_CMIRS_EC_ = "B97-D/def2-TZVPPD/CMIRS(EC)"
    B97_D_def2_TZVPPD_CMIRS_EC_EMC_ = "B97-D/def2-TZVPPD/CMIRS(EC/EMC)"
    B97_D_def2_TZVPPD_CMIRS_PC_ = "B97-D/def2-TZVPPD/CMIRS(PC)"
    B97_D_def2_QZVPPD_VACUUM = "B97-D/def2-QZVPPD/VACUUM"
    B97_D_def2_QZVPPD_PCM_WATER_ = "B97-D/def2-QZVPPD/PCM(WATER)"
    B97_D_def2_QZVPPD_PCM_THF_ = "B97-D/def2-QZVPPD/PCM(THF)"
    B97_D_def2_QZVPPD_PCM_G2_ = "B97-D/def2-QZVPPD/PCM(G2)"
    B97_D_def2_QZVPPD_PCM_DME_ = "B97-D/def2-QZVPPD/PCM(DME)"
    B97_D_def2_QZVPPD_PCM_EC_ = "B97-D/def2-QZVPPD/PCM(EC)"
    B97_D_def2_QZVPPD_PCM_EC_EMC_ = "B97-D/def2-QZVPPD/PCM(EC/EMC)"
    B97_D_def2_QZVPPD_PCM_PC_ = "B97-D/def2-QZVPPD/PCM(PC)"
    B97_D_def2_QZVPPD_ISOSVP_WATER_ = "B97-D/def2-QZVPPD/ISOSVP(WATER)"
    B97_D_def2_QZVPPD_ISOSVP_THF_ = "B97-D/def2-QZVPPD/ISOSVP(THF)"
    B97_D_def2_QZVPPD_ISOSVP_G2_ = "B97-D/def2-QZVPPD/ISOSVP(G2)"
    B97_D_def2_QZVPPD_ISOSVP_DME_ = "B97-D/def2-QZVPPD/ISOSVP(DME)"
    B97_D_def2_QZVPPD_ISOSVP_EC_ = "B97-D/def2-QZVPPD/ISOSVP(EC)"
    B97_D_def2_QZVPPD_ISOSVP_EC_EMC_ = "B97-D/def2-QZVPPD/ISOSVP(EC/EMC)"
    B97_D_def2_QZVPPD_ISOSVP_PC_ = "B97-D/def2-QZVPPD/ISOSVP(PC)"
    B97_D_def2_QZVPPD_SMD_WATER_ = "B97-D/def2-QZVPPD/SMD(WATER)"
    B97_D_def2_QZVPPD_SMD_THF_ = "B97-D/def2-QZVPPD/SMD(THF)"
    B97_D_def2_QZVPPD_SMD_G2_ = "B97-D/def2-QZVPPD/SMD(G2)"
    B97_D_def2_QZVPPD_SMD_DME_ = "B97-D/def2-QZVPPD/SMD(DME)"
    B97_D_def2_QZVPPD_SMD_EC_ = "B97-D/def2-QZVPPD/SMD(EC)"
    B97_D_def2_QZVPPD_SMD_EC_EMC_ = "B97-D/def2-QZVPPD/SMD(EC/EMC)"
    B97_D_def2_QZVPPD_SMD_PC_ = "B97-D/def2-QZVPPD/SMD(PC)"
    B97_D_def2_QZVPPD_CMIRS_WATER_ = "B97-D/def2-QZVPPD/CMIRS(WATER)"
    B97_D_def2_QZVPPD_CMIRS_THF_ = "B97-D/def2-QZVPPD/CMIRS(THF)"
    B97_D_def2_QZVPPD_CMIRS_G2_ = "B97-D/def2-QZVPPD/CMIRS(G2)"
    B97_D_def2_QZVPPD_CMIRS_DME_ = "B97-D/def2-QZVPPD/CMIRS(DME)"
    B97_D_def2_QZVPPD_CMIRS_EC_ = "B97-D/def2-QZVPPD/CMIRS(EC)"
    B97_D_def2_QZVPPD_CMIRS_EC_EMC_ = "B97-D/def2-QZVPPD/CMIRS(EC/EMC)"
    B97_D_def2_QZVPPD_CMIRS_PC_ = "B97-D/def2-QZVPPD/CMIRS(PC)"
    B97_D3_6_31g_d_VACUUM = "B97-D3/6-31g*/VACUUM"
    B97_D3_6_31g_d_PCM_WATER_ = "B97-D3/6-31g*/PCM(WATER)"
    B97_D3_6_31g_d_PCM_THF_ = "B97-D3/6-31g*/PCM(THF)"
    B97_D3_6_31g_d_PCM_G2_ = "B97-D3/6-31g*/PCM(G2)"
    B97_D3_6_31g_d_PCM_DME_ = "B97-D3/6-31g*/PCM(DME)"
    B97_D3_6_31g_d_PCM_EC_ = "B97-D3/6-31g*/PCM(EC)"
    B97_D3_6_31g_d_PCM_EC_EMC_ = "B97-D3/6-31g*/PCM(EC/EMC)"
    B97_D3_6_31g_d_PCM_PC_ = "B97-D3/6-31g*/PCM(PC)"
    B97_D3_6_31g_d_ISOSVP_WATER_ = "B97-D3/6-31g*/ISOSVP(WATER)"
    B97_D3_6_31g_d_ISOSVP_THF_ = "B97-D3/6-31g*/ISOSVP(THF)"
    B97_D3_6_31g_d_ISOSVP_G2_ = "B97-D3/6-31g*/ISOSVP(G2)"
    B97_D3_6_31g_d_ISOSVP_DME_ = "B97-D3/6-31g*/ISOSVP(DME)"
    B97_D3_6_31g_d_ISOSVP_EC_ = "B97-D3/6-31g*/ISOSVP(EC)"
    B97_D3_6_31g_d_ISOSVP_EC_EMC_ = "B97-D3/6-31g*/ISOSVP(EC/EMC)"
    B97_D3_6_31g_d_ISOSVP_PC_ = "B97-D3/6-31g*/ISOSVP(PC)"
    B97_D3_6_31g_d_SMD_WATER_ = "B97-D3/6-31g*/SMD(WATER)"
    B97_D3_6_31g_d_SMD_THF_ = "B97-D3/6-31g*/SMD(THF)"
    B97_D3_6_31g_d_SMD_G2_ = "B97-D3/6-31g*/SMD(G2)"
    B97_D3_6_31g_d_SMD_DME_ = "B97-D3/6-31g*/SMD(DME)"
    B97_D3_6_31g_d_SMD_EC_ = "B97-D3/6-31g*/SMD(EC)"
    B97_D3_6_31g_d_SMD_EC_EMC_ = "B97-D3/6-31g*/SMD(EC/EMC)"
    B97_D3_6_31g_d_SMD_PC_ = "B97-D3/6-31g*/SMD(PC)"
    B97_D3_6_31g_d_CMIRS_WATER_ = "B97-D3/6-31g*/CMIRS(WATER)"
    B97_D3_6_31g_d_CMIRS_THF_ = "B97-D3/6-31g*/CMIRS(THF)"
    B97_D3_6_31g_d_CMIRS_G2_ = "B97-D3/6-31g*/CMIRS(G2)"
    B97_D3_6_31g_d_CMIRS_DME_ = "B97-D3/6-31g*/CMIRS(DME)"
    B97_D3_6_31g_d_CMIRS_EC_ = "B97-D3/6-31g*/CMIRS(EC)"
    B97_D3_6_31g_d_CMIRS_EC_EMC_ = "B97-D3/6-31g*/CMIRS(EC/EMC)"
    B97_D3_6_31g_d_CMIRS_PC_ = "B97-D3/6-31g*/CMIRS(PC)"
    B97_D3_def2_SVPD_VACUUM = "B97-D3/def2-SVPD/VACUUM"
    B97_D3_def2_SVPD_PCM_WATER_ = "B97-D3/def2-SVPD/PCM(WATER)"
    B97_D3_def2_SVPD_PCM_THF_ = "B97-D3/def2-SVPD/PCM(THF)"
    B97_D3_def2_SVPD_PCM_G2_ = "B97-D3/def2-SVPD/PCM(G2)"
    B97_D3_def2_SVPD_PCM_DME_ = "B97-D3/def2-SVPD/PCM(DME)"
    B97_D3_def2_SVPD_PCM_EC_ = "B97-D3/def2-SVPD/PCM(EC)"
    B97_D3_def2_SVPD_PCM_EC_EMC_ = "B97-D3/def2-SVPD/PCM(EC/EMC)"
    B97_D3_def2_SVPD_PCM_PC_ = "B97-D3/def2-SVPD/PCM(PC)"
    B97_D3_def2_SVPD_ISOSVP_WATER_ = "B97-D3/def2-SVPD/ISOSVP(WATER)"
    B97_D3_def2_SVPD_ISOSVP_THF_ = "B97-D3/def2-SVPD/ISOSVP(THF)"
    B97_D3_def2_SVPD_ISOSVP_G2_ = "B97-D3/def2-SVPD/ISOSVP(G2)"
    B97_D3_def2_SVPD_ISOSVP_DME_ = "B97-D3/def2-SVPD/ISOSVP(DME)"
    B97_D3_def2_SVPD_ISOSVP_EC_ = "B97-D3/def2-SVPD/ISOSVP(EC)"
    B97_D3_def2_SVPD_ISOSVP_EC_EMC_ = "B97-D3/def2-SVPD/ISOSVP(EC/EMC)"
    B97_D3_def2_SVPD_ISOSVP_PC_ = "B97-D3/def2-SVPD/ISOSVP(PC)"
    B97_D3_def2_SVPD_SMD_WATER_ = "B97-D3/def2-SVPD/SMD(WATER)"
    B97_D3_def2_SVPD_SMD_THF_ = "B97-D3/def2-SVPD/SMD(THF)"
    B97_D3_def2_SVPD_SMD_G2_ = "B97-D3/def2-SVPD/SMD(G2)"
    B97_D3_def2_SVPD_SMD_DME_ = "B97-D3/def2-SVPD/SMD(DME)"
    B97_D3_def2_SVPD_SMD_EC_ = "B97-D3/def2-SVPD/SMD(EC)"
    B97_D3_def2_SVPD_SMD_EC_EMC_ = "B97-D3/def2-SVPD/SMD(EC/EMC)"
    B97_D3_def2_SVPD_SMD_PC_ = "B97-D3/def2-SVPD/SMD(PC)"
    B97_D3_def2_SVPD_CMIRS_WATER_ = "B97-D3/def2-SVPD/CMIRS(WATER)"
    B97_D3_def2_SVPD_CMIRS_THF_ = "B97-D3/def2-SVPD/CMIRS(THF)"
    B97_D3_def2_SVPD_CMIRS_G2_ = "B97-D3/def2-SVPD/CMIRS(G2)"
    B97_D3_def2_SVPD_CMIRS_DME_ = "B97-D3/def2-SVPD/CMIRS(DME)"
    B97_D3_def2_SVPD_CMIRS_EC_ = "B97-D3/def2-SVPD/CMIRS(EC)"
    B97_D3_def2_SVPD_CMIRS_EC_EMC_ = "B97-D3/def2-SVPD/CMIRS(EC/EMC)"
    B97_D3_def2_SVPD_CMIRS_PC_ = "B97-D3/def2-SVPD/CMIRS(PC)"
    B97_D3_def2_TZVP_VACUUM = "B97-D3/def2-TZVP/VACUUM"
    B97_D3_def2_TZVP_PCM_WATER_ = "B97-D3/def2-TZVP/PCM(WATER)"
    B97_D3_def2_TZVP_PCM_THF_ = "B97-D3/def2-TZVP/PCM(THF)"
    B97_D3_def2_TZVP_PCM_G2_ = "B97-D3/def2-TZVP/PCM(G2)"
    B97_D3_def2_TZVP_PCM_DME_ = "B97-D3/def2-TZVP/PCM(DME)"
    B97_D3_def2_TZVP_PCM_EC_ = "B97-D3/def2-TZVP/PCM(EC)"
    B97_D3_def2_TZVP_PCM_EC_EMC_ = "B97-D3/def2-TZVP/PCM(EC/EMC)"
    B97_D3_def2_TZVP_PCM_PC_ = "B97-D3/def2-TZVP/PCM(PC)"
    B97_D3_def2_TZVP_ISOSVP_WATER_ = "B97-D3/def2-TZVP/ISOSVP(WATER)"
    B97_D3_def2_TZVP_ISOSVP_THF_ = "B97-D3/def2-TZVP/ISOSVP(THF)"
    B97_D3_def2_TZVP_ISOSVP_G2_ = "B97-D3/def2-TZVP/ISOSVP(G2)"
    B97_D3_def2_TZVP_ISOSVP_DME_ = "B97-D3/def2-TZVP/ISOSVP(DME)"
    B97_D3_def2_TZVP_ISOSVP_EC_ = "B97-D3/def2-TZVP/ISOSVP(EC)"
    B97_D3_def2_TZVP_ISOSVP_EC_EMC_ = "B97-D3/def2-TZVP/ISOSVP(EC/EMC)"
    B97_D3_def2_TZVP_ISOSVP_PC_ = "B97-D3/def2-TZVP/ISOSVP(PC)"
    B97_D3_def2_TZVP_SMD_WATER_ = "B97-D3/def2-TZVP/SMD(WATER)"
    B97_D3_def2_TZVP_SMD_THF_ = "B97-D3/def2-TZVP/SMD(THF)"
    B97_D3_def2_TZVP_SMD_G2_ = "B97-D3/def2-TZVP/SMD(G2)"
    B97_D3_def2_TZVP_SMD_DME_ = "B97-D3/def2-TZVP/SMD(DME)"
    B97_D3_def2_TZVP_SMD_EC_ = "B97-D3/def2-TZVP/SMD(EC)"
    B97_D3_def2_TZVP_SMD_EC_EMC_ = "B97-D3/def2-TZVP/SMD(EC/EMC)"
    B97_D3_def2_TZVP_SMD_PC_ = "B97-D3/def2-TZVP/SMD(PC)"
    B97_D3_def2_TZVP_CMIRS_WATER_ = "B97-D3/def2-TZVP/CMIRS(WATER)"
    B97_D3_def2_TZVP_CMIRS_THF_ = "B97-D3/def2-TZVP/CMIRS(THF)"
    B97_D3_def2_TZVP_CMIRS_G2_ = "B97-D3/def2-TZVP/CMIRS(G2)"
    B97_D3_def2_TZVP_CMIRS_DME_ = "B97-D3/def2-TZVP/CMIRS(DME)"
    B97_D3_def2_TZVP_CMIRS_EC_ = "B97-D3/def2-TZVP/CMIRS(EC)"
    B97_D3_def2_TZVP_CMIRS_EC_EMC_ = "B97-D3/def2-TZVP/CMIRS(EC/EMC)"
    B97_D3_def2_TZVP_CMIRS_PC_ = "B97-D3/def2-TZVP/CMIRS(PC)"
    B97_D3_def2_TZVPD_VACUUM = "B97-D3/def2-TZVPD/VACUUM"
    B97_D3_def2_TZVPD_PCM_WATER_ = "B97-D3/def2-TZVPD/PCM(WATER)"
    B97_D3_def2_TZVPD_PCM_THF_ = "B97-D3/def2-TZVPD/PCM(THF)"
    B97_D3_def2_TZVPD_PCM_G2_ = "B97-D3/def2-TZVPD/PCM(G2)"
    B97_D3_def2_TZVPD_PCM_DME_ = "B97-D3/def2-TZVPD/PCM(DME)"
    B97_D3_def2_TZVPD_PCM_EC_ = "B97-D3/def2-TZVPD/PCM(EC)"
    B97_D3_def2_TZVPD_PCM_EC_EMC_ = "B97-D3/def2-TZVPD/PCM(EC/EMC)"
    B97_D3_def2_TZVPD_PCM_PC_ = "B97-D3/def2-TZVPD/PCM(PC)"
    B97_D3_def2_TZVPD_ISOSVP_WATER_ = "B97-D3/def2-TZVPD/ISOSVP(WATER)"
    B97_D3_def2_TZVPD_ISOSVP_THF_ = "B97-D3/def2-TZVPD/ISOSVP(THF)"
    B97_D3_def2_TZVPD_ISOSVP_G2_ = "B97-D3/def2-TZVPD/ISOSVP(G2)"
    B97_D3_def2_TZVPD_ISOSVP_DME_ = "B97-D3/def2-TZVPD/ISOSVP(DME)"
    B97_D3_def2_TZVPD_ISOSVP_EC_ = "B97-D3/def2-TZVPD/ISOSVP(EC)"
    B97_D3_def2_TZVPD_ISOSVP_EC_EMC_ = "B97-D3/def2-TZVPD/ISOSVP(EC/EMC)"
    B97_D3_def2_TZVPD_ISOSVP_PC_ = "B97-D3/def2-TZVPD/ISOSVP(PC)"
    B97_D3_def2_TZVPD_SMD_WATER_ = "B97-D3/def2-TZVPD/SMD(WATER)"
    B97_D3_def2_TZVPD_SMD_THF_ = "B97-D3/def2-TZVPD/SMD(THF)"
    B97_D3_def2_TZVPD_SMD_G2_ = "B97-D3/def2-TZVPD/SMD(G2)"
    B97_D3_def2_TZVPD_SMD_DME_ = "B97-D3/def2-TZVPD/SMD(DME)"
    B97_D3_def2_TZVPD_SMD_EC_ = "B97-D3/def2-TZVPD/SMD(EC)"
    B97_D3_def2_TZVPD_SMD_EC_EMC_ = "B97-D3/def2-TZVPD/SMD(EC/EMC)"
    B97_D3_def2_TZVPD_SMD_PC_ = "B97-D3/def2-TZVPD/SMD(PC)"
    B97_D3_def2_TZVPD_CMIRS_WATER_ = "B97-D3/def2-TZVPD/CMIRS(WATER)"
    B97_D3_def2_TZVPD_CMIRS_THF_ = "B97-D3/def2-TZVPD/CMIRS(THF)"
    B97_D3_def2_TZVPD_CMIRS_G2_ = "B97-D3/def2-TZVPD/CMIRS(G2)"
    B97_D3_def2_TZVPD_CMIRS_DME_ = "B97-D3/def2-TZVPD/CMIRS(DME)"
    B97_D3_def2_TZVPD_CMIRS_EC_ = "B97-D3/def2-TZVPD/CMIRS(EC)"
    B97_D3_def2_TZVPD_CMIRS_EC_EMC_ = "B97-D3/def2-TZVPD/CMIRS(EC/EMC)"
    B97_D3_def2_TZVPD_CMIRS_PC_ = "B97-D3/def2-TZVPD/CMIRS(PC)"
    B97_D3_def2_TZVPP_VACUUM = "B97-D3/def2-TZVPP/VACUUM"
    B97_D3_def2_TZVPP_PCM_WATER_ = "B97-D3/def2-TZVPP/PCM(WATER)"
    B97_D3_def2_TZVPP_PCM_THF_ = "B97-D3/def2-TZVPP/PCM(THF)"
    B97_D3_def2_TZVPP_PCM_G2_ = "B97-D3/def2-TZVPP/PCM(G2)"
    B97_D3_def2_TZVPP_PCM_DME_ = "B97-D3/def2-TZVPP/PCM(DME)"
    B97_D3_def2_TZVPP_PCM_EC_ = "B97-D3/def2-TZVPP/PCM(EC)"
    B97_D3_def2_TZVPP_PCM_EC_EMC_ = "B97-D3/def2-TZVPP/PCM(EC/EMC)"
    B97_D3_def2_TZVPP_PCM_PC_ = "B97-D3/def2-TZVPP/PCM(PC)"
    B97_D3_def2_TZVPP_ISOSVP_WATER_ = "B97-D3/def2-TZVPP/ISOSVP(WATER)"
    B97_D3_def2_TZVPP_ISOSVP_THF_ = "B97-D3/def2-TZVPP/ISOSVP(THF)"
    B97_D3_def2_TZVPP_ISOSVP_G2_ = "B97-D3/def2-TZVPP/ISOSVP(G2)"
    B97_D3_def2_TZVPP_ISOSVP_DME_ = "B97-D3/def2-TZVPP/ISOSVP(DME)"
    B97_D3_def2_TZVPP_ISOSVP_EC_ = "B97-D3/def2-TZVPP/ISOSVP(EC)"
    B97_D3_def2_TZVPP_ISOSVP_EC_EMC_ = "B97-D3/def2-TZVPP/ISOSVP(EC/EMC)"
    B97_D3_def2_TZVPP_ISOSVP_PC_ = "B97-D3/def2-TZVPP/ISOSVP(PC)"
    B97_D3_def2_TZVPP_SMD_WATER_ = "B97-D3/def2-TZVPP/SMD(WATER)"
    B97_D3_def2_TZVPP_SMD_THF_ = "B97-D3/def2-TZVPP/SMD(THF)"
    B97_D3_def2_TZVPP_SMD_G2_ = "B97-D3/def2-TZVPP/SMD(G2)"
    B97_D3_def2_TZVPP_SMD_DME_ = "B97-D3/def2-TZVPP/SMD(DME)"
    B97_D3_def2_TZVPP_SMD_EC_ = "B97-D3/def2-TZVPP/SMD(EC)"
    B97_D3_def2_TZVPP_SMD_EC_EMC_ = "B97-D3/def2-TZVPP/SMD(EC/EMC)"
    B97_D3_def2_TZVPP_SMD_PC_ = "B97-D3/def2-TZVPP/SMD(PC)"
    B97_D3_def2_TZVPP_CMIRS_WATER_ = "B97-D3/def2-TZVPP/CMIRS(WATER)"
    B97_D3_def2_TZVPP_CMIRS_THF_ = "B97-D3/def2-TZVPP/CMIRS(THF)"
    B97_D3_def2_TZVPP_CMIRS_G2_ = "B97-D3/def2-TZVPP/CMIRS(G2)"
    B97_D3_def2_TZVPP_CMIRS_DME_ = "B97-D3/def2-TZVPP/CMIRS(DME)"
    B97_D3_def2_TZVPP_CMIRS_EC_ = "B97-D3/def2-TZVPP/CMIRS(EC)"
    B97_D3_def2_TZVPP_CMIRS_EC_EMC_ = "B97-D3/def2-TZVPP/CMIRS(EC/EMC)"
    B97_D3_def2_TZVPP_CMIRS_PC_ = "B97-D3/def2-TZVPP/CMIRS(PC)"
    B97_D3_def2_TZVPPD_VACUUM = "B97-D3/def2-TZVPPD/VACUUM"
    B97_D3_def2_TZVPPD_PCM_WATER_ = "B97-D3/def2-TZVPPD/PCM(WATER)"
    B97_D3_def2_TZVPPD_PCM_THF_ = "B97-D3/def2-TZVPPD/PCM(THF)"
    B97_D3_def2_TZVPPD_PCM_G2_ = "B97-D3/def2-TZVPPD/PCM(G2)"
    B97_D3_def2_TZVPPD_PCM_DME_ = "B97-D3/def2-TZVPPD/PCM(DME)"
    B97_D3_def2_TZVPPD_PCM_EC_ = "B97-D3/def2-TZVPPD/PCM(EC)"
    B97_D3_def2_TZVPPD_PCM_EC_EMC_ = "B97-D3/def2-TZVPPD/PCM(EC/EMC)"
    B97_D3_def2_TZVPPD_PCM_PC_ = "B97-D3/def2-TZVPPD/PCM(PC)"
    B97_D3_def2_TZVPPD_ISOSVP_WATER_ = "B97-D3/def2-TZVPPD/ISOSVP(WATER)"
    B97_D3_def2_TZVPPD_ISOSVP_THF_ = "B97-D3/def2-TZVPPD/ISOSVP(THF)"
    B97_D3_def2_TZVPPD_ISOSVP_G2_ = "B97-D3/def2-TZVPPD/ISOSVP(G2)"
    B97_D3_def2_TZVPPD_ISOSVP_DME_ = "B97-D3/def2-TZVPPD/ISOSVP(DME)"
    B97_D3_def2_TZVPPD_ISOSVP_EC_ = "B97-D3/def2-TZVPPD/ISOSVP(EC)"
    B97_D3_def2_TZVPPD_ISOSVP_EC_EMC_ = "B97-D3/def2-TZVPPD/ISOSVP(EC/EMC)"
    B97_D3_def2_TZVPPD_ISOSVP_PC_ = "B97-D3/def2-TZVPPD/ISOSVP(PC)"
    B97_D3_def2_TZVPPD_SMD_WATER_ = "B97-D3/def2-TZVPPD/SMD(WATER)"
    B97_D3_def2_TZVPPD_SMD_THF_ = "B97-D3/def2-TZVPPD/SMD(THF)"
    B97_D3_def2_TZVPPD_SMD_G2_ = "B97-D3/def2-TZVPPD/SMD(G2)"
    B97_D3_def2_TZVPPD_SMD_DME_ = "B97-D3/def2-TZVPPD/SMD(DME)"
    B97_D3_def2_TZVPPD_SMD_EC_ = "B97-D3/def2-TZVPPD/SMD(EC)"
    B97_D3_def2_TZVPPD_SMD_EC_EMC_ = "B97-D3/def2-TZVPPD/SMD(EC/EMC)"
    B97_D3_def2_TZVPPD_SMD_PC_ = "B97-D3/def2-TZVPPD/SMD(PC)"
    B97_D3_def2_TZVPPD_CMIRS_WATER_ = "B97-D3/def2-TZVPPD/CMIRS(WATER)"
    B97_D3_def2_TZVPPD_CMIRS_THF_ = "B97-D3/def2-TZVPPD/CMIRS(THF)"
    B97_D3_def2_TZVPPD_CMIRS_G2_ = "B97-D3/def2-TZVPPD/CMIRS(G2)"
    B97_D3_def2_TZVPPD_CMIRS_DME_ = "B97-D3/def2-TZVPPD/CMIRS(DME)"
    B97_D3_def2_TZVPPD_CMIRS_EC_ = "B97-D3/def2-TZVPPD/CMIRS(EC)"
    B97_D3_def2_TZVPPD_CMIRS_EC_EMC_ = "B97-D3/def2-TZVPPD/CMIRS(EC/EMC)"
    B97_D3_def2_TZVPPD_CMIRS_PC_ = "B97-D3/def2-TZVPPD/CMIRS(PC)"
    B97_D3_def2_QZVPPD_VACUUM = "B97-D3/def2-QZVPPD/VACUUM"
    B97_D3_def2_QZVPPD_PCM_WATER_ = "B97-D3/def2-QZVPPD/PCM(WATER)"
    B97_D3_def2_QZVPPD_PCM_THF_ = "B97-D3/def2-QZVPPD/PCM(THF)"
    B97_D3_def2_QZVPPD_PCM_G2_ = "B97-D3/def2-QZVPPD/PCM(G2)"
    B97_D3_def2_QZVPPD_PCM_DME_ = "B97-D3/def2-QZVPPD/PCM(DME)"
    B97_D3_def2_QZVPPD_PCM_EC_ = "B97-D3/def2-QZVPPD/PCM(EC)"
    B97_D3_def2_QZVPPD_PCM_EC_EMC_ = "B97-D3/def2-QZVPPD/PCM(EC/EMC)"
    B97_D3_def2_QZVPPD_PCM_PC_ = "B97-D3/def2-QZVPPD/PCM(PC)"
    B97_D3_def2_QZVPPD_ISOSVP_WATER_ = "B97-D3/def2-QZVPPD/ISOSVP(WATER)"
    B97_D3_def2_QZVPPD_ISOSVP_THF_ = "B97-D3/def2-QZVPPD/ISOSVP(THF)"
    B97_D3_def2_QZVPPD_ISOSVP_G2_ = "B97-D3/def2-QZVPPD/ISOSVP(G2)"
    B97_D3_def2_QZVPPD_ISOSVP_DME_ = "B97-D3/def2-QZVPPD/ISOSVP(DME)"
    B97_D3_def2_QZVPPD_ISOSVP_EC_ = "B97-D3/def2-QZVPPD/ISOSVP(EC)"
    B97_D3_def2_QZVPPD_ISOSVP_EC_EMC_ = "B97-D3/def2-QZVPPD/ISOSVP(EC/EMC)"
    B97_D3_def2_QZVPPD_ISOSVP_PC_ = "B97-D3/def2-QZVPPD/ISOSVP(PC)"
    B97_D3_def2_QZVPPD_SMD_WATER_ = "B97-D3/def2-QZVPPD/SMD(WATER)"
    B97_D3_def2_QZVPPD_SMD_THF_ = "B97-D3/def2-QZVPPD/SMD(THF)"
    B97_D3_def2_QZVPPD_SMD_G2_ = "B97-D3/def2-QZVPPD/SMD(G2)"
    B97_D3_def2_QZVPPD_SMD_DME_ = "B97-D3/def2-QZVPPD/SMD(DME)"
    B97_D3_def2_QZVPPD_SMD_EC_ = "B97-D3/def2-QZVPPD/SMD(EC)"
    B97_D3_def2_QZVPPD_SMD_EC_EMC_ = "B97-D3/def2-QZVPPD/SMD(EC/EMC)"
    B97_D3_def2_QZVPPD_SMD_PC_ = "B97-D3/def2-QZVPPD/SMD(PC)"
    B97_D3_def2_QZVPPD_CMIRS_WATER_ = "B97-D3/def2-QZVPPD/CMIRS(WATER)"
    B97_D3_def2_QZVPPD_CMIRS_THF_ = "B97-D3/def2-QZVPPD/CMIRS(THF)"
    B97_D3_def2_QZVPPD_CMIRS_G2_ = "B97-D3/def2-QZVPPD/CMIRS(G2)"
    B97_D3_def2_QZVPPD_CMIRS_DME_ = "B97-D3/def2-QZVPPD/CMIRS(DME)"
    B97_D3_def2_QZVPPD_CMIRS_EC_ = "B97-D3/def2-QZVPPD/CMIRS(EC)"
    B97_D3_def2_QZVPPD_CMIRS_EC_EMC_ = "B97-D3/def2-QZVPPD/CMIRS(EC/EMC)"
    B97_D3_def2_QZVPPD_CMIRS_PC_ = "B97-D3/def2-QZVPPD/CMIRS(PC)"
    B97M_rV_6_31g_d_VACUUM = "B97M-rV/6-31g*/VACUUM"
    B97M_rV_6_31g_d_PCM_WATER_ = "B97M-rV/6-31g*/PCM(WATER)"
    B97M_rV_6_31g_d_PCM_THF_ = "B97M-rV/6-31g*/PCM(THF)"
    B97M_rV_6_31g_d_PCM_G2_ = "B97M-rV/6-31g*/PCM(G2)"
    B97M_rV_6_31g_d_PCM_DME_ = "B97M-rV/6-31g*/PCM(DME)"
    B97M_rV_6_31g_d_PCM_EC_ = "B97M-rV/6-31g*/PCM(EC)"
    B97M_rV_6_31g_d_PCM_EC_EMC_ = "B97M-rV/6-31g*/PCM(EC/EMC)"
    B97M_rV_6_31g_d_PCM_PC_ = "B97M-rV/6-31g*/PCM(PC)"
    B97M_rV_6_31g_d_ISOSVP_WATER_ = "B97M-rV/6-31g*/ISOSVP(WATER)"
    B97M_rV_6_31g_d_ISOSVP_THF_ = "B97M-rV/6-31g*/ISOSVP(THF)"
    B97M_rV_6_31g_d_ISOSVP_G2_ = "B97M-rV/6-31g*/ISOSVP(G2)"
    B97M_rV_6_31g_d_ISOSVP_DME_ = "B97M-rV/6-31g*/ISOSVP(DME)"
    B97M_rV_6_31g_d_ISOSVP_EC_ = "B97M-rV/6-31g*/ISOSVP(EC)"
    B97M_rV_6_31g_d_ISOSVP_EC_EMC_ = "B97M-rV/6-31g*/ISOSVP(EC/EMC)"
    B97M_rV_6_31g_d_ISOSVP_PC_ = "B97M-rV/6-31g*/ISOSVP(PC)"
    B97M_rV_6_31g_d_SMD_WATER_ = "B97M-rV/6-31g*/SMD(WATER)"
    B97M_rV_6_31g_d_SMD_THF_ = "B97M-rV/6-31g*/SMD(THF)"
    B97M_rV_6_31g_d_SMD_G2_ = "B97M-rV/6-31g*/SMD(G2)"
    B97M_rV_6_31g_d_SMD_DME_ = "B97M-rV/6-31g*/SMD(DME)"
    B97M_rV_6_31g_d_SMD_EC_ = "B97M-rV/6-31g*/SMD(EC)"
    B97M_rV_6_31g_d_SMD_EC_EMC_ = "B97M-rV/6-31g*/SMD(EC/EMC)"
    B97M_rV_6_31g_d_SMD_PC_ = "B97M-rV/6-31g*/SMD(PC)"
    B97M_rV_6_31g_d_CMIRS_WATER_ = "B97M-rV/6-31g*/CMIRS(WATER)"
    B97M_rV_6_31g_d_CMIRS_THF_ = "B97M-rV/6-31g*/CMIRS(THF)"
    B97M_rV_6_31g_d_CMIRS_G2_ = "B97M-rV/6-31g*/CMIRS(G2)"
    B97M_rV_6_31g_d_CMIRS_DME_ = "B97M-rV/6-31g*/CMIRS(DME)"
    B97M_rV_6_31g_d_CMIRS_EC_ = "B97M-rV/6-31g*/CMIRS(EC)"
    B97M_rV_6_31g_d_CMIRS_EC_EMC_ = "B97M-rV/6-31g*/CMIRS(EC/EMC)"
    B97M_rV_6_31g_d_CMIRS_PC_ = "B97M-rV/6-31g*/CMIRS(PC)"
    B97M_rV_def2_SVPD_VACUUM = "B97M-rV/def2-SVPD/VACUUM"
    B97M_rV_def2_SVPD_PCM_WATER_ = "B97M-rV/def2-SVPD/PCM(WATER)"
    B97M_rV_def2_SVPD_PCM_THF_ = "B97M-rV/def2-SVPD/PCM(THF)"
    B97M_rV_def2_SVPD_PCM_G2_ = "B97M-rV/def2-SVPD/PCM(G2)"
    B97M_rV_def2_SVPD_PCM_DME_ = "B97M-rV/def2-SVPD/PCM(DME)"
    B97M_rV_def2_SVPD_PCM_EC_ = "B97M-rV/def2-SVPD/PCM(EC)"
    B97M_rV_def2_SVPD_PCM_EC_EMC_ = "B97M-rV/def2-SVPD/PCM(EC/EMC)"
    B97M_rV_def2_SVPD_PCM_PC_ = "B97M-rV/def2-SVPD/PCM(PC)"
    B97M_rV_def2_SVPD_ISOSVP_WATER_ = "B97M-rV/def2-SVPD/ISOSVP(WATER)"
    B97M_rV_def2_SVPD_ISOSVP_THF_ = "B97M-rV/def2-SVPD/ISOSVP(THF)"
    B97M_rV_def2_SVPD_ISOSVP_G2_ = "B97M-rV/def2-SVPD/ISOSVP(G2)"
    B97M_rV_def2_SVPD_ISOSVP_DME_ = "B97M-rV/def2-SVPD/ISOSVP(DME)"
    B97M_rV_def2_SVPD_ISOSVP_EC_ = "B97M-rV/def2-SVPD/ISOSVP(EC)"
    B97M_rV_def2_SVPD_ISOSVP_EC_EMC_ = "B97M-rV/def2-SVPD/ISOSVP(EC/EMC)"
    B97M_rV_def2_SVPD_ISOSVP_PC_ = "B97M-rV/def2-SVPD/ISOSVP(PC)"
    B97M_rV_def2_SVPD_SMD_WATER_ = "B97M-rV/def2-SVPD/SMD(WATER)"
    B97M_rV_def2_SVPD_SMD_THF_ = "B97M-rV/def2-SVPD/SMD(THF)"
    B97M_rV_def2_SVPD_SMD_G2_ = "B97M-rV/def2-SVPD/SMD(G2)"
    B97M_rV_def2_SVPD_SMD_DME_ = "B97M-rV/def2-SVPD/SMD(DME)"
    B97M_rV_def2_SVPD_SMD_EC_ = "B97M-rV/def2-SVPD/SMD(EC)"
    B97M_rV_def2_SVPD_SMD_EC_EMC_ = "B97M-rV/def2-SVPD/SMD(EC/EMC)"
    B97M_rV_def2_SVPD_SMD_PC_ = "B97M-rV/def2-SVPD/SMD(PC)"
    B97M_rV_def2_SVPD_CMIRS_WATER_ = "B97M-rV/def2-SVPD/CMIRS(WATER)"
    B97M_rV_def2_SVPD_CMIRS_THF_ = "B97M-rV/def2-SVPD/CMIRS(THF)"
    B97M_rV_def2_SVPD_CMIRS_G2_ = "B97M-rV/def2-SVPD/CMIRS(G2)"
    B97M_rV_def2_SVPD_CMIRS_DME_ = "B97M-rV/def2-SVPD/CMIRS(DME)"
    B97M_rV_def2_SVPD_CMIRS_EC_ = "B97M-rV/def2-SVPD/CMIRS(EC)"
    B97M_rV_def2_SVPD_CMIRS_EC_EMC_ = "B97M-rV/def2-SVPD/CMIRS(EC/EMC)"
    B97M_rV_def2_SVPD_CMIRS_PC_ = "B97M-rV/def2-SVPD/CMIRS(PC)"
    B97M_rV_def2_TZVP_VACUUM = "B97M-rV/def2-TZVP/VACUUM"
    B97M_rV_def2_TZVP_PCM_WATER_ = "B97M-rV/def2-TZVP/PCM(WATER)"
    B97M_rV_def2_TZVP_PCM_THF_ = "B97M-rV/def2-TZVP/PCM(THF)"
    B97M_rV_def2_TZVP_PCM_G2_ = "B97M-rV/def2-TZVP/PCM(G2)"
    B97M_rV_def2_TZVP_PCM_DME_ = "B97M-rV/def2-TZVP/PCM(DME)"
    B97M_rV_def2_TZVP_PCM_EC_ = "B97M-rV/def2-TZVP/PCM(EC)"
    B97M_rV_def2_TZVP_PCM_EC_EMC_ = "B97M-rV/def2-TZVP/PCM(EC/EMC)"
    B97M_rV_def2_TZVP_PCM_PC_ = "B97M-rV/def2-TZVP/PCM(PC)"
    B97M_rV_def2_TZVP_ISOSVP_WATER_ = "B97M-rV/def2-TZVP/ISOSVP(WATER)"
    B97M_rV_def2_TZVP_ISOSVP_THF_ = "B97M-rV/def2-TZVP/ISOSVP(THF)"
    B97M_rV_def2_TZVP_ISOSVP_G2_ = "B97M-rV/def2-TZVP/ISOSVP(G2)"
    B97M_rV_def2_TZVP_ISOSVP_DME_ = "B97M-rV/def2-TZVP/ISOSVP(DME)"
    B97M_rV_def2_TZVP_ISOSVP_EC_ = "B97M-rV/def2-TZVP/ISOSVP(EC)"
    B97M_rV_def2_TZVP_ISOSVP_EC_EMC_ = "B97M-rV/def2-TZVP/ISOSVP(EC/EMC)"
    B97M_rV_def2_TZVP_ISOSVP_PC_ = "B97M-rV/def2-TZVP/ISOSVP(PC)"
    B97M_rV_def2_TZVP_SMD_WATER_ = "B97M-rV/def2-TZVP/SMD(WATER)"
    B97M_rV_def2_TZVP_SMD_THF_ = "B97M-rV/def2-TZVP/SMD(THF)"
    B97M_rV_def2_TZVP_SMD_G2_ = "B97M-rV/def2-TZVP/SMD(G2)"
    B97M_rV_def2_TZVP_SMD_DME_ = "B97M-rV/def2-TZVP/SMD(DME)"
    B97M_rV_def2_TZVP_SMD_EC_ = "B97M-rV/def2-TZVP/SMD(EC)"
    B97M_rV_def2_TZVP_SMD_EC_EMC_ = "B97M-rV/def2-TZVP/SMD(EC/EMC)"
    B97M_rV_def2_TZVP_SMD_PC_ = "B97M-rV/def2-TZVP/SMD(PC)"
    B97M_rV_def2_TZVP_CMIRS_WATER_ = "B97M-rV/def2-TZVP/CMIRS(WATER)"
    B97M_rV_def2_TZVP_CMIRS_THF_ = "B97M-rV/def2-TZVP/CMIRS(THF)"
    B97M_rV_def2_TZVP_CMIRS_G2_ = "B97M-rV/def2-TZVP/CMIRS(G2)"
    B97M_rV_def2_TZVP_CMIRS_DME_ = "B97M-rV/def2-TZVP/CMIRS(DME)"
    B97M_rV_def2_TZVP_CMIRS_EC_ = "B97M-rV/def2-TZVP/CMIRS(EC)"
    B97M_rV_def2_TZVP_CMIRS_EC_EMC_ = "B97M-rV/def2-TZVP/CMIRS(EC/EMC)"
    B97M_rV_def2_TZVP_CMIRS_PC_ = "B97M-rV/def2-TZVP/CMIRS(PC)"
    B97M_rV_def2_TZVPD_VACUUM = "B97M-rV/def2-TZVPD/VACUUM"
    B97M_rV_def2_TZVPD_PCM_WATER_ = "B97M-rV/def2-TZVPD/PCM(WATER)"
    B97M_rV_def2_TZVPD_PCM_THF_ = "B97M-rV/def2-TZVPD/PCM(THF)"
    B97M_rV_def2_TZVPD_PCM_G2_ = "B97M-rV/def2-TZVPD/PCM(G2)"
    B97M_rV_def2_TZVPD_PCM_DME_ = "B97M-rV/def2-TZVPD/PCM(DME)"
    B97M_rV_def2_TZVPD_PCM_EC_ = "B97M-rV/def2-TZVPD/PCM(EC)"
    B97M_rV_def2_TZVPD_PCM_EC_EMC_ = "B97M-rV/def2-TZVPD/PCM(EC/EMC)"
    B97M_rV_def2_TZVPD_PCM_PC_ = "B97M-rV/def2-TZVPD/PCM(PC)"
    B97M_rV_def2_TZVPD_ISOSVP_WATER_ = "B97M-rV/def2-TZVPD/ISOSVP(WATER)"
    B97M_rV_def2_TZVPD_ISOSVP_THF_ = "B97M-rV/def2-TZVPD/ISOSVP(THF)"
    B97M_rV_def2_TZVPD_ISOSVP_G2_ = "B97M-rV/def2-TZVPD/ISOSVP(G2)"
    B97M_rV_def2_TZVPD_ISOSVP_DME_ = "B97M-rV/def2-TZVPD/ISOSVP(DME)"
    B97M_rV_def2_TZVPD_ISOSVP_EC_ = "B97M-rV/def2-TZVPD/ISOSVP(EC)"
    B97M_rV_def2_TZVPD_ISOSVP_EC_EMC_ = "B97M-rV/def2-TZVPD/ISOSVP(EC/EMC)"
    B97M_rV_def2_TZVPD_ISOSVP_PC_ = "B97M-rV/def2-TZVPD/ISOSVP(PC)"
    B97M_rV_def2_TZVPD_SMD_WATER_ = "B97M-rV/def2-TZVPD/SMD(WATER)"
    B97M_rV_def2_TZVPD_SMD_THF_ = "B97M-rV/def2-TZVPD/SMD(THF)"
    B97M_rV_def2_TZVPD_SMD_G2_ = "B97M-rV/def2-TZVPD/SMD(G2)"
    B97M_rV_def2_TZVPD_SMD_DME_ = "B97M-rV/def2-TZVPD/SMD(DME)"
    B97M_rV_def2_TZVPD_SMD_EC_ = "B97M-rV/def2-TZVPD/SMD(EC)"
    B97M_rV_def2_TZVPD_SMD_EC_EMC_ = "B97M-rV/def2-TZVPD/SMD(EC/EMC)"
    B97M_rV_def2_TZVPD_SMD_PC_ = "B97M-rV/def2-TZVPD/SMD(PC)"
    B97M_rV_def2_TZVPD_CMIRS_WATER_ = "B97M-rV/def2-TZVPD/CMIRS(WATER)"
    B97M_rV_def2_TZVPD_CMIRS_THF_ = "B97M-rV/def2-TZVPD/CMIRS(THF)"
    B97M_rV_def2_TZVPD_CMIRS_G2_ = "B97M-rV/def2-TZVPD/CMIRS(G2)"
    B97M_rV_def2_TZVPD_CMIRS_DME_ = "B97M-rV/def2-TZVPD/CMIRS(DME)"
    B97M_rV_def2_TZVPD_CMIRS_EC_ = "B97M-rV/def2-TZVPD/CMIRS(EC)"
    B97M_rV_def2_TZVPD_CMIRS_EC_EMC_ = "B97M-rV/def2-TZVPD/CMIRS(EC/EMC)"
    B97M_rV_def2_TZVPD_CMIRS_PC_ = "B97M-rV/def2-TZVPD/CMIRS(PC)"
    B97M_rV_def2_TZVPP_VACUUM = "B97M-rV/def2-TZVPP/VACUUM"
    B97M_rV_def2_TZVPP_PCM_WATER_ = "B97M-rV/def2-TZVPP/PCM(WATER)"
    B97M_rV_def2_TZVPP_PCM_THF_ = "B97M-rV/def2-TZVPP/PCM(THF)"
    B97M_rV_def2_TZVPP_PCM_G2_ = "B97M-rV/def2-TZVPP/PCM(G2)"
    B97M_rV_def2_TZVPP_PCM_DME_ = "B97M-rV/def2-TZVPP/PCM(DME)"
    B97M_rV_def2_TZVPP_PCM_EC_ = "B97M-rV/def2-TZVPP/PCM(EC)"
    B97M_rV_def2_TZVPP_PCM_EC_EMC_ = "B97M-rV/def2-TZVPP/PCM(EC/EMC)"
    B97M_rV_def2_TZVPP_PCM_PC_ = "B97M-rV/def2-TZVPP/PCM(PC)"
    B97M_rV_def2_TZVPP_ISOSVP_WATER_ = "B97M-rV/def2-TZVPP/ISOSVP(WATER)"
    B97M_rV_def2_TZVPP_ISOSVP_THF_ = "B97M-rV/def2-TZVPP/ISOSVP(THF)"
    B97M_rV_def2_TZVPP_ISOSVP_G2_ = "B97M-rV/def2-TZVPP/ISOSVP(G2)"
    B97M_rV_def2_TZVPP_ISOSVP_DME_ = "B97M-rV/def2-TZVPP/ISOSVP(DME)"
    B97M_rV_def2_TZVPP_ISOSVP_EC_ = "B97M-rV/def2-TZVPP/ISOSVP(EC)"
    B97M_rV_def2_TZVPP_ISOSVP_EC_EMC_ = "B97M-rV/def2-TZVPP/ISOSVP(EC/EMC)"
    B97M_rV_def2_TZVPP_ISOSVP_PC_ = "B97M-rV/def2-TZVPP/ISOSVP(PC)"
    B97M_rV_def2_TZVPP_SMD_WATER_ = "B97M-rV/def2-TZVPP/SMD(WATER)"
    B97M_rV_def2_TZVPP_SMD_THF_ = "B97M-rV/def2-TZVPP/SMD(THF)"
    B97M_rV_def2_TZVPP_SMD_G2_ = "B97M-rV/def2-TZVPP/SMD(G2)"
    B97M_rV_def2_TZVPP_SMD_DME_ = "B97M-rV/def2-TZVPP/SMD(DME)"
    B97M_rV_def2_TZVPP_SMD_EC_ = "B97M-rV/def2-TZVPP/SMD(EC)"
    B97M_rV_def2_TZVPP_SMD_EC_EMC_ = "B97M-rV/def2-TZVPP/SMD(EC/EMC)"
    B97M_rV_def2_TZVPP_SMD_PC_ = "B97M-rV/def2-TZVPP/SMD(PC)"
    B97M_rV_def2_TZVPP_CMIRS_WATER_ = "B97M-rV/def2-TZVPP/CMIRS(WATER)"
    B97M_rV_def2_TZVPP_CMIRS_THF_ = "B97M-rV/def2-TZVPP/CMIRS(THF)"
    B97M_rV_def2_TZVPP_CMIRS_G2_ = "B97M-rV/def2-TZVPP/CMIRS(G2)"
    B97M_rV_def2_TZVPP_CMIRS_DME_ = "B97M-rV/def2-TZVPP/CMIRS(DME)"
    B97M_rV_def2_TZVPP_CMIRS_EC_ = "B97M-rV/def2-TZVPP/CMIRS(EC)"
    B97M_rV_def2_TZVPP_CMIRS_EC_EMC_ = "B97M-rV/def2-TZVPP/CMIRS(EC/EMC)"
    B97M_rV_def2_TZVPP_CMIRS_PC_ = "B97M-rV/def2-TZVPP/CMIRS(PC)"
    B97M_rV_def2_TZVPPD_VACUUM = "B97M-rV/def2-TZVPPD/VACUUM"
    B97M_rV_def2_TZVPPD_PCM_WATER_ = "B97M-rV/def2-TZVPPD/PCM(WATER)"
    B97M_rV_def2_TZVPPD_PCM_THF_ = "B97M-rV/def2-TZVPPD/PCM(THF)"
    B97M_rV_def2_TZVPPD_PCM_G2_ = "B97M-rV/def2-TZVPPD/PCM(G2)"
    B97M_rV_def2_TZVPPD_PCM_DME_ = "B97M-rV/def2-TZVPPD/PCM(DME)"
    B97M_rV_def2_TZVPPD_PCM_EC_ = "B97M-rV/def2-TZVPPD/PCM(EC)"
    B97M_rV_def2_TZVPPD_PCM_EC_EMC_ = "B97M-rV/def2-TZVPPD/PCM(EC/EMC)"
    B97M_rV_def2_TZVPPD_PCM_PC_ = "B97M-rV/def2-TZVPPD/PCM(PC)"
    B97M_rV_def2_TZVPPD_ISOSVP_WATER_ = "B97M-rV/def2-TZVPPD/ISOSVP(WATER)"
    B97M_rV_def2_TZVPPD_ISOSVP_THF_ = "B97M-rV/def2-TZVPPD/ISOSVP(THF)"
    B97M_rV_def2_TZVPPD_ISOSVP_G2_ = "B97M-rV/def2-TZVPPD/ISOSVP(G2)"
    B97M_rV_def2_TZVPPD_ISOSVP_DME_ = "B97M-rV/def2-TZVPPD/ISOSVP(DME)"
    B97M_rV_def2_TZVPPD_ISOSVP_EC_ = "B97M-rV/def2-TZVPPD/ISOSVP(EC)"
    B97M_rV_def2_TZVPPD_ISOSVP_EC_EMC_ = "B97M-rV/def2-TZVPPD/ISOSVP(EC/EMC)"
    B97M_rV_def2_TZVPPD_ISOSVP_PC_ = "B97M-rV/def2-TZVPPD/ISOSVP(PC)"
    B97M_rV_def2_TZVPPD_SMD_WATER_ = "B97M-rV/def2-TZVPPD/SMD(WATER)"
    B97M_rV_def2_TZVPPD_SMD_THF_ = "B97M-rV/def2-TZVPPD/SMD(THF)"
    B97M_rV_def2_TZVPPD_SMD_G2_ = "B97M-rV/def2-TZVPPD/SMD(G2)"
    B97M_rV_def2_TZVPPD_SMD_DME_ = "B97M-rV/def2-TZVPPD/SMD(DME)"
    B97M_rV_def2_TZVPPD_SMD_EC_ = "B97M-rV/def2-TZVPPD/SMD(EC)"
    B97M_rV_def2_TZVPPD_SMD_EC_EMC_ = "B97M-rV/def2-TZVPPD/SMD(EC/EMC)"
    B97M_rV_def2_TZVPPD_SMD_PC_ = "B97M-rV/def2-TZVPPD/SMD(PC)"
    B97M_rV_def2_TZVPPD_CMIRS_WATER_ = "B97M-rV/def2-TZVPPD/CMIRS(WATER)"
    B97M_rV_def2_TZVPPD_CMIRS_THF_ = "B97M-rV/def2-TZVPPD/CMIRS(THF)"
    B97M_rV_def2_TZVPPD_CMIRS_G2_ = "B97M-rV/def2-TZVPPD/CMIRS(G2)"
    B97M_rV_def2_TZVPPD_CMIRS_DME_ = "B97M-rV/def2-TZVPPD/CMIRS(DME)"
    B97M_rV_def2_TZVPPD_CMIRS_EC_ = "B97M-rV/def2-TZVPPD/CMIRS(EC)"
    B97M_rV_def2_TZVPPD_CMIRS_EC_EMC_ = "B97M-rV/def2-TZVPPD/CMIRS(EC/EMC)"
    B97M_rV_def2_TZVPPD_CMIRS_PC_ = "B97M-rV/def2-TZVPPD/CMIRS(PC)"
    B97M_rV_def2_QZVPPD_VACUUM = "B97M-rV/def2-QZVPPD/VACUUM"
    B97M_rV_def2_QZVPPD_PCM_WATER_ = "B97M-rV/def2-QZVPPD/PCM(WATER)"
    B97M_rV_def2_QZVPPD_PCM_THF_ = "B97M-rV/def2-QZVPPD/PCM(THF)"
    B97M_rV_def2_QZVPPD_PCM_G2_ = "B97M-rV/def2-QZVPPD/PCM(G2)"
    B97M_rV_def2_QZVPPD_PCM_DME_ = "B97M-rV/def2-QZVPPD/PCM(DME)"
    B97M_rV_def2_QZVPPD_PCM_EC_ = "B97M-rV/def2-QZVPPD/PCM(EC)"
    B97M_rV_def2_QZVPPD_PCM_EC_EMC_ = "B97M-rV/def2-QZVPPD/PCM(EC/EMC)"
    B97M_rV_def2_QZVPPD_PCM_PC_ = "B97M-rV/def2-QZVPPD/PCM(PC)"
    B97M_rV_def2_QZVPPD_ISOSVP_WATER_ = "B97M-rV/def2-QZVPPD/ISOSVP(WATER)"
    B97M_rV_def2_QZVPPD_ISOSVP_THF_ = "B97M-rV/def2-QZVPPD/ISOSVP(THF)"
    B97M_rV_def2_QZVPPD_ISOSVP_G2_ = "B97M-rV/def2-QZVPPD/ISOSVP(G2)"
    B97M_rV_def2_QZVPPD_ISOSVP_DME_ = "B97M-rV/def2-QZVPPD/ISOSVP(DME)"
    B97M_rV_def2_QZVPPD_ISOSVP_EC_ = "B97M-rV/def2-QZVPPD/ISOSVP(EC)"
    B97M_rV_def2_QZVPPD_ISOSVP_EC_EMC_ = "B97M-rV/def2-QZVPPD/ISOSVP(EC/EMC)"
    B97M_rV_def2_QZVPPD_ISOSVP_PC_ = "B97M-rV/def2-QZVPPD/ISOSVP(PC)"
    B97M_rV_def2_QZVPPD_SMD_WATER_ = "B97M-rV/def2-QZVPPD/SMD(WATER)"
    B97M_rV_def2_QZVPPD_SMD_THF_ = "B97M-rV/def2-QZVPPD/SMD(THF)"
    B97M_rV_def2_QZVPPD_SMD_G2_ = "B97M-rV/def2-QZVPPD/SMD(G2)"
    B97M_rV_def2_QZVPPD_SMD_DME_ = "B97M-rV/def2-QZVPPD/SMD(DME)"
    B97M_rV_def2_QZVPPD_SMD_EC_ = "B97M-rV/def2-QZVPPD/SMD(EC)"
    B97M_rV_def2_QZVPPD_SMD_EC_EMC_ = "B97M-rV/def2-QZVPPD/SMD(EC/EMC)"
    B97M_rV_def2_QZVPPD_SMD_PC_ = "B97M-rV/def2-QZVPPD/SMD(PC)"
    B97M_rV_def2_QZVPPD_CMIRS_WATER_ = "B97M-rV/def2-QZVPPD/CMIRS(WATER)"
    B97M_rV_def2_QZVPPD_CMIRS_THF_ = "B97M-rV/def2-QZVPPD/CMIRS(THF)"
    B97M_rV_def2_QZVPPD_CMIRS_G2_ = "B97M-rV/def2-QZVPPD/CMIRS(G2)"
    B97M_rV_def2_QZVPPD_CMIRS_DME_ = "B97M-rV/def2-QZVPPD/CMIRS(DME)"
    B97M_rV_def2_QZVPPD_CMIRS_EC_ = "B97M-rV/def2-QZVPPD/CMIRS(EC)"
    B97M_rV_def2_QZVPPD_CMIRS_EC_EMC_ = "B97M-rV/def2-QZVPPD/CMIRS(EC/EMC)"
    B97M_rV_def2_QZVPPD_CMIRS_PC_ = "B97M-rV/def2-QZVPPD/CMIRS(PC)"
    B3LYP_6_31g_d_VACUUM = "B3LYP/6-31g*/VACUUM"
    B3LYP_6_31g_d_PCM_WATER_ = "B3LYP/6-31g*/PCM(WATER)"
    B3LYP_6_31g_d_PCM_THF_ = "B3LYP/6-31g*/PCM(THF)"
    B3LYP_6_31g_d_PCM_G2_ = "B3LYP/6-31g*/PCM(G2)"
    B3LYP_6_31g_d_PCM_DME_ = "B3LYP/6-31g*/PCM(DME)"
    B3LYP_6_31g_d_PCM_EC_ = "B3LYP/6-31g*/PCM(EC)"
    B3LYP_6_31g_d_PCM_EC_EMC_ = "B3LYP/6-31g*/PCM(EC/EMC)"
    B3LYP_6_31g_d_PCM_PC_ = "B3LYP/6-31g*/PCM(PC)"
    B3LYP_6_31g_d_ISOSVP_WATER_ = "B3LYP/6-31g*/ISOSVP(WATER)"
    B3LYP_6_31g_d_ISOSVP_THF_ = "B3LYP/6-31g*/ISOSVP(THF)"
    B3LYP_6_31g_d_ISOSVP_G2_ = "B3LYP/6-31g*/ISOSVP(G2)"
    B3LYP_6_31g_d_ISOSVP_DME_ = "B3LYP/6-31g*/ISOSVP(DME)"
    B3LYP_6_31g_d_ISOSVP_EC_ = "B3LYP/6-31g*/ISOSVP(EC)"
    B3LYP_6_31g_d_ISOSVP_EC_EMC_ = "B3LYP/6-31g*/ISOSVP(EC/EMC)"
    B3LYP_6_31g_d_ISOSVP_PC_ = "B3LYP/6-31g*/ISOSVP(PC)"
    B3LYP_6_31g_d_SMD_WATER_ = "B3LYP/6-31g*/SMD(WATER)"
    B3LYP_6_31g_d_SMD_THF_ = "B3LYP/6-31g*/SMD(THF)"
    B3LYP_6_31g_d_SMD_G2_ = "B3LYP/6-31g*/SMD(G2)"
    B3LYP_6_31g_d_SMD_DME_ = "B3LYP/6-31g*/SMD(DME)"
    B3LYP_6_31g_d_SMD_EC_ = "B3LYP/6-31g*/SMD(EC)"
    B3LYP_6_31g_d_SMD_EC_EMC_ = "B3LYP/6-31g*/SMD(EC/EMC)"
    B3LYP_6_31g_d_SMD_PC_ = "B3LYP/6-31g*/SMD(PC)"
    B3LYP_6_31g_d_CMIRS_WATER_ = "B3LYP/6-31g*/CMIRS(WATER)"
    B3LYP_6_31g_d_CMIRS_THF_ = "B3LYP/6-31g*/CMIRS(THF)"
    B3LYP_6_31g_d_CMIRS_G2_ = "B3LYP/6-31g*/CMIRS(G2)"
    B3LYP_6_31g_d_CMIRS_DME_ = "B3LYP/6-31g*/CMIRS(DME)"
    B3LYP_6_31g_d_CMIRS_EC_ = "B3LYP/6-31g*/CMIRS(EC)"
    B3LYP_6_31g_d_CMIRS_EC_EMC_ = "B3LYP/6-31g*/CMIRS(EC/EMC)"
    B3LYP_6_31g_d_CMIRS_PC_ = "B3LYP/6-31g*/CMIRS(PC)"
    B3LYP_def2_SVPD_VACUUM = "B3LYP/def2-SVPD/VACUUM"
    B3LYP_def2_SVPD_PCM_WATER_ = "B3LYP/def2-SVPD/PCM(WATER)"
    B3LYP_def2_SVPD_PCM_THF_ = "B3LYP/def2-SVPD/PCM(THF)"
    B3LYP_def2_SVPD_PCM_G2_ = "B3LYP/def2-SVPD/PCM(G2)"
    B3LYP_def2_SVPD_PCM_DME_ = "B3LYP/def2-SVPD/PCM(DME)"
    B3LYP_def2_SVPD_PCM_EC_ = "B3LYP/def2-SVPD/PCM(EC)"
    B3LYP_def2_SVPD_PCM_EC_EMC_ = "B3LYP/def2-SVPD/PCM(EC/EMC)"
    B3LYP_def2_SVPD_PCM_PC_ = "B3LYP/def2-SVPD/PCM(PC)"
    B3LYP_def2_SVPD_ISOSVP_WATER_ = "B3LYP/def2-SVPD/ISOSVP(WATER)"
    B3LYP_def2_SVPD_ISOSVP_THF_ = "B3LYP/def2-SVPD/ISOSVP(THF)"
    B3LYP_def2_SVPD_ISOSVP_G2_ = "B3LYP/def2-SVPD/ISOSVP(G2)"
    B3LYP_def2_SVPD_ISOSVP_DME_ = "B3LYP/def2-SVPD/ISOSVP(DME)"
    B3LYP_def2_SVPD_ISOSVP_EC_ = "B3LYP/def2-SVPD/ISOSVP(EC)"
    B3LYP_def2_SVPD_ISOSVP_EC_EMC_ = "B3LYP/def2-SVPD/ISOSVP(EC/EMC)"
    B3LYP_def2_SVPD_ISOSVP_PC_ = "B3LYP/def2-SVPD/ISOSVP(PC)"
    B3LYP_def2_SVPD_SMD_WATER_ = "B3LYP/def2-SVPD/SMD(WATER)"
    B3LYP_def2_SVPD_SMD_THF_ = "B3LYP/def2-SVPD/SMD(THF)"
    B3LYP_def2_SVPD_SMD_G2_ = "B3LYP/def2-SVPD/SMD(G2)"
    B3LYP_def2_SVPD_SMD_DME_ = "B3LYP/def2-SVPD/SMD(DME)"
    B3LYP_def2_SVPD_SMD_EC_ = "B3LYP/def2-SVPD/SMD(EC)"
    B3LYP_def2_SVPD_SMD_EC_EMC_ = "B3LYP/def2-SVPD/SMD(EC/EMC)"
    B3LYP_def2_SVPD_SMD_PC_ = "B3LYP/def2-SVPD/SMD(PC)"
    B3LYP_def2_SVPD_CMIRS_WATER_ = "B3LYP/def2-SVPD/CMIRS(WATER)"
    B3LYP_def2_SVPD_CMIRS_THF_ = "B3LYP/def2-SVPD/CMIRS(THF)"
    B3LYP_def2_SVPD_CMIRS_G2_ = "B3LYP/def2-SVPD/CMIRS(G2)"
    B3LYP_def2_SVPD_CMIRS_DME_ = "B3LYP/def2-SVPD/CMIRS(DME)"
    B3LYP_def2_SVPD_CMIRS_EC_ = "B3LYP/def2-SVPD/CMIRS(EC)"
    B3LYP_def2_SVPD_CMIRS_EC_EMC_ = "B3LYP/def2-SVPD/CMIRS(EC/EMC)"
    B3LYP_def2_SVPD_CMIRS_PC_ = "B3LYP/def2-SVPD/CMIRS(PC)"
    B3LYP_def2_TZVP_VACUUM = "B3LYP/def2-TZVP/VACUUM"
    B3LYP_def2_TZVP_PCM_WATER_ = "B3LYP/def2-TZVP/PCM(WATER)"
    B3LYP_def2_TZVP_PCM_THF_ = "B3LYP/def2-TZVP/PCM(THF)"
    B3LYP_def2_TZVP_PCM_G2_ = "B3LYP/def2-TZVP/PCM(G2)"
    B3LYP_def2_TZVP_PCM_DME_ = "B3LYP/def2-TZVP/PCM(DME)"
    B3LYP_def2_TZVP_PCM_EC_ = "B3LYP/def2-TZVP/PCM(EC)"
    B3LYP_def2_TZVP_PCM_EC_EMC_ = "B3LYP/def2-TZVP/PCM(EC/EMC)"
    B3LYP_def2_TZVP_PCM_PC_ = "B3LYP/def2-TZVP/PCM(PC)"
    B3LYP_def2_TZVP_ISOSVP_WATER_ = "B3LYP/def2-TZVP/ISOSVP(WATER)"
    B3LYP_def2_TZVP_ISOSVP_THF_ = "B3LYP/def2-TZVP/ISOSVP(THF)"
    B3LYP_def2_TZVP_ISOSVP_G2_ = "B3LYP/def2-TZVP/ISOSVP(G2)"
    B3LYP_def2_TZVP_ISOSVP_DME_ = "B3LYP/def2-TZVP/ISOSVP(DME)"
    B3LYP_def2_TZVP_ISOSVP_EC_ = "B3LYP/def2-TZVP/ISOSVP(EC)"
    B3LYP_def2_TZVP_ISOSVP_EC_EMC_ = "B3LYP/def2-TZVP/ISOSVP(EC/EMC)"
    B3LYP_def2_TZVP_ISOSVP_PC_ = "B3LYP/def2-TZVP/ISOSVP(PC)"
    B3LYP_def2_TZVP_SMD_WATER_ = "B3LYP/def2-TZVP/SMD(WATER)"
    B3LYP_def2_TZVP_SMD_THF_ = "B3LYP/def2-TZVP/SMD(THF)"
    B3LYP_def2_TZVP_SMD_G2_ = "B3LYP/def2-TZVP/SMD(G2)"
    B3LYP_def2_TZVP_SMD_DME_ = "B3LYP/def2-TZVP/SMD(DME)"
    B3LYP_def2_TZVP_SMD_EC_ = "B3LYP/def2-TZVP/SMD(EC)"
    B3LYP_def2_TZVP_SMD_EC_EMC_ = "B3LYP/def2-TZVP/SMD(EC/EMC)"
    B3LYP_def2_TZVP_SMD_PC_ = "B3LYP/def2-TZVP/SMD(PC)"
    B3LYP_def2_TZVP_CMIRS_WATER_ = "B3LYP/def2-TZVP/CMIRS(WATER)"
    B3LYP_def2_TZVP_CMIRS_THF_ = "B3LYP/def2-TZVP/CMIRS(THF)"
    B3LYP_def2_TZVP_CMIRS_G2_ = "B3LYP/def2-TZVP/CMIRS(G2)"
    B3LYP_def2_TZVP_CMIRS_DME_ = "B3LYP/def2-TZVP/CMIRS(DME)"
    B3LYP_def2_TZVP_CMIRS_EC_ = "B3LYP/def2-TZVP/CMIRS(EC)"
    B3LYP_def2_TZVP_CMIRS_EC_EMC_ = "B3LYP/def2-TZVP/CMIRS(EC/EMC)"
    B3LYP_def2_TZVP_CMIRS_PC_ = "B3LYP/def2-TZVP/CMIRS(PC)"
    B3LYP_def2_TZVPD_VACUUM = "B3LYP/def2-TZVPD/VACUUM"
    B3LYP_def2_TZVPD_PCM_WATER_ = "B3LYP/def2-TZVPD/PCM(WATER)"
    B3LYP_def2_TZVPD_PCM_THF_ = "B3LYP/def2-TZVPD/PCM(THF)"
    B3LYP_def2_TZVPD_PCM_G2_ = "B3LYP/def2-TZVPD/PCM(G2)"
    B3LYP_def2_TZVPD_PCM_DME_ = "B3LYP/def2-TZVPD/PCM(DME)"
    B3LYP_def2_TZVPD_PCM_EC_ = "B3LYP/def2-TZVPD/PCM(EC)"
    B3LYP_def2_TZVPD_PCM_EC_EMC_ = "B3LYP/def2-TZVPD/PCM(EC/EMC)"
    B3LYP_def2_TZVPD_PCM_PC_ = "B3LYP/def2-TZVPD/PCM(PC)"
    B3LYP_def2_TZVPD_ISOSVP_WATER_ = "B3LYP/def2-TZVPD/ISOSVP(WATER)"
    B3LYP_def2_TZVPD_ISOSVP_THF_ = "B3LYP/def2-TZVPD/ISOSVP(THF)"
    B3LYP_def2_TZVPD_ISOSVP_G2_ = "B3LYP/def2-TZVPD/ISOSVP(G2)"
    B3LYP_def2_TZVPD_ISOSVP_DME_ = "B3LYP/def2-TZVPD/ISOSVP(DME)"
    B3LYP_def2_TZVPD_ISOSVP_EC_ = "B3LYP/def2-TZVPD/ISOSVP(EC)"
    B3LYP_def2_TZVPD_ISOSVP_EC_EMC_ = "B3LYP/def2-TZVPD/ISOSVP(EC/EMC)"
    B3LYP_def2_TZVPD_ISOSVP_PC_ = "B3LYP/def2-TZVPD/ISOSVP(PC)"
    B3LYP_def2_TZVPD_SMD_WATER_ = "B3LYP/def2-TZVPD/SMD(WATER)"
    B3LYP_def2_TZVPD_SMD_THF_ = "B3LYP/def2-TZVPD/SMD(THF)"
    B3LYP_def2_TZVPD_SMD_G2_ = "B3LYP/def2-TZVPD/SMD(G2)"
    B3LYP_def2_TZVPD_SMD_DME_ = "B3LYP/def2-TZVPD/SMD(DME)"
    B3LYP_def2_TZVPD_SMD_EC_ = "B3LYP/def2-TZVPD/SMD(EC)"
    B3LYP_def2_TZVPD_SMD_EC_EMC_ = "B3LYP/def2-TZVPD/SMD(EC/EMC)"
    B3LYP_def2_TZVPD_SMD_PC_ = "B3LYP/def2-TZVPD/SMD(PC)"
    B3LYP_def2_TZVPD_CMIRS_WATER_ = "B3LYP/def2-TZVPD/CMIRS(WATER)"
    B3LYP_def2_TZVPD_CMIRS_THF_ = "B3LYP/def2-TZVPD/CMIRS(THF)"
    B3LYP_def2_TZVPD_CMIRS_G2_ = "B3LYP/def2-TZVPD/CMIRS(G2)"
    B3LYP_def2_TZVPD_CMIRS_DME_ = "B3LYP/def2-TZVPD/CMIRS(DME)"
    B3LYP_def2_TZVPD_CMIRS_EC_ = "B3LYP/def2-TZVPD/CMIRS(EC)"
    B3LYP_def2_TZVPD_CMIRS_EC_EMC_ = "B3LYP/def2-TZVPD/CMIRS(EC/EMC)"
    B3LYP_def2_TZVPD_CMIRS_PC_ = "B3LYP/def2-TZVPD/CMIRS(PC)"
    B3LYP_def2_TZVPP_VACUUM = "B3LYP/def2-TZVPP/VACUUM"
    B3LYP_def2_TZVPP_PCM_WATER_ = "B3LYP/def2-TZVPP/PCM(WATER)"
    B3LYP_def2_TZVPP_PCM_THF_ = "B3LYP/def2-TZVPP/PCM(THF)"
    B3LYP_def2_TZVPP_PCM_G2_ = "B3LYP/def2-TZVPP/PCM(G2)"
    B3LYP_def2_TZVPP_PCM_DME_ = "B3LYP/def2-TZVPP/PCM(DME)"
    B3LYP_def2_TZVPP_PCM_EC_ = "B3LYP/def2-TZVPP/PCM(EC)"
    B3LYP_def2_TZVPP_PCM_EC_EMC_ = "B3LYP/def2-TZVPP/PCM(EC/EMC)"
    B3LYP_def2_TZVPP_PCM_PC_ = "B3LYP/def2-TZVPP/PCM(PC)"
    B3LYP_def2_TZVPP_ISOSVP_WATER_ = "B3LYP/def2-TZVPP/ISOSVP(WATER)"
    B3LYP_def2_TZVPP_ISOSVP_THF_ = "B3LYP/def2-TZVPP/ISOSVP(THF)"
    B3LYP_def2_TZVPP_ISOSVP_G2_ = "B3LYP/def2-TZVPP/ISOSVP(G2)"
    B3LYP_def2_TZVPP_ISOSVP_DME_ = "B3LYP/def2-TZVPP/ISOSVP(DME)"
    B3LYP_def2_TZVPP_ISOSVP_EC_ = "B3LYP/def2-TZVPP/ISOSVP(EC)"
    B3LYP_def2_TZVPP_ISOSVP_EC_EMC_ = "B3LYP/def2-TZVPP/ISOSVP(EC/EMC)"
    B3LYP_def2_TZVPP_ISOSVP_PC_ = "B3LYP/def2-TZVPP/ISOSVP(PC)"
    B3LYP_def2_TZVPP_SMD_WATER_ = "B3LYP/def2-TZVPP/SMD(WATER)"
    B3LYP_def2_TZVPP_SMD_THF_ = "B3LYP/def2-TZVPP/SMD(THF)"
    B3LYP_def2_TZVPP_SMD_G2_ = "B3LYP/def2-TZVPP/SMD(G2)"
    B3LYP_def2_TZVPP_SMD_DME_ = "B3LYP/def2-TZVPP/SMD(DME)"
    B3LYP_def2_TZVPP_SMD_EC_ = "B3LYP/def2-TZVPP/SMD(EC)"
    B3LYP_def2_TZVPP_SMD_EC_EMC_ = "B3LYP/def2-TZVPP/SMD(EC/EMC)"
    B3LYP_def2_TZVPP_SMD_PC_ = "B3LYP/def2-TZVPP/SMD(PC)"
    B3LYP_def2_TZVPP_CMIRS_WATER_ = "B3LYP/def2-TZVPP/CMIRS(WATER)"
    B3LYP_def2_TZVPP_CMIRS_THF_ = "B3LYP/def2-TZVPP/CMIRS(THF)"
    B3LYP_def2_TZVPP_CMIRS_G2_ = "B3LYP/def2-TZVPP/CMIRS(G2)"
    B3LYP_def2_TZVPP_CMIRS_DME_ = "B3LYP/def2-TZVPP/CMIRS(DME)"
    B3LYP_def2_TZVPP_CMIRS_EC_ = "B3LYP/def2-TZVPP/CMIRS(EC)"
    B3LYP_def2_TZVPP_CMIRS_EC_EMC_ = "B3LYP/def2-TZVPP/CMIRS(EC/EMC)"
    B3LYP_def2_TZVPP_CMIRS_PC_ = "B3LYP/def2-TZVPP/CMIRS(PC)"
    B3LYP_def2_TZVPPD_VACUUM = "B3LYP/def2-TZVPPD/VACUUM"
    B3LYP_def2_TZVPPD_PCM_WATER_ = "B3LYP/def2-TZVPPD/PCM(WATER)"
    B3LYP_def2_TZVPPD_PCM_THF_ = "B3LYP/def2-TZVPPD/PCM(THF)"
    B3LYP_def2_TZVPPD_PCM_G2_ = "B3LYP/def2-TZVPPD/PCM(G2)"
    B3LYP_def2_TZVPPD_PCM_DME_ = "B3LYP/def2-TZVPPD/PCM(DME)"
    B3LYP_def2_TZVPPD_PCM_EC_ = "B3LYP/def2-TZVPPD/PCM(EC)"
    B3LYP_def2_TZVPPD_PCM_EC_EMC_ = "B3LYP/def2-TZVPPD/PCM(EC/EMC)"
    B3LYP_def2_TZVPPD_PCM_PC_ = "B3LYP/def2-TZVPPD/PCM(PC)"
    B3LYP_def2_TZVPPD_ISOSVP_WATER_ = "B3LYP/def2-TZVPPD/ISOSVP(WATER)"
    B3LYP_def2_TZVPPD_ISOSVP_THF_ = "B3LYP/def2-TZVPPD/ISOSVP(THF)"
    B3LYP_def2_TZVPPD_ISOSVP_G2_ = "B3LYP/def2-TZVPPD/ISOSVP(G2)"
    B3LYP_def2_TZVPPD_ISOSVP_DME_ = "B3LYP/def2-TZVPPD/ISOSVP(DME)"
    B3LYP_def2_TZVPPD_ISOSVP_EC_ = "B3LYP/def2-TZVPPD/ISOSVP(EC)"
    B3LYP_def2_TZVPPD_ISOSVP_EC_EMC_ = "B3LYP/def2-TZVPPD/ISOSVP(EC/EMC)"
    B3LYP_def2_TZVPPD_ISOSVP_PC_ = "B3LYP/def2-TZVPPD/ISOSVP(PC)"
    B3LYP_def2_TZVPPD_SMD_WATER_ = "B3LYP/def2-TZVPPD/SMD(WATER)"
    B3LYP_def2_TZVPPD_SMD_THF_ = "B3LYP/def2-TZVPPD/SMD(THF)"
    B3LYP_def2_TZVPPD_SMD_G2_ = "B3LYP/def2-TZVPPD/SMD(G2)"
    B3LYP_def2_TZVPPD_SMD_DME_ = "B3LYP/def2-TZVPPD/SMD(DME)"
    B3LYP_def2_TZVPPD_SMD_EC_ = "B3LYP/def2-TZVPPD/SMD(EC)"
    B3LYP_def2_TZVPPD_SMD_EC_EMC_ = "B3LYP/def2-TZVPPD/SMD(EC/EMC)"
    B3LYP_def2_TZVPPD_SMD_PC_ = "B3LYP/def2-TZVPPD/SMD(PC)"
    B3LYP_def2_TZVPPD_CMIRS_WATER_ = "B3LYP/def2-TZVPPD/CMIRS(WATER)"
    B3LYP_def2_TZVPPD_CMIRS_THF_ = "B3LYP/def2-TZVPPD/CMIRS(THF)"
    B3LYP_def2_TZVPPD_CMIRS_G2_ = "B3LYP/def2-TZVPPD/CMIRS(G2)"
    B3LYP_def2_TZVPPD_CMIRS_DME_ = "B3LYP/def2-TZVPPD/CMIRS(DME)"
    B3LYP_def2_TZVPPD_CMIRS_EC_ = "B3LYP/def2-TZVPPD/CMIRS(EC)"
    B3LYP_def2_TZVPPD_CMIRS_EC_EMC_ = "B3LYP/def2-TZVPPD/CMIRS(EC/EMC)"
    B3LYP_def2_TZVPPD_CMIRS_PC_ = "B3LYP/def2-TZVPPD/CMIRS(PC)"
    B3LYP_def2_QZVPPD_VACUUM = "B3LYP/def2-QZVPPD/VACUUM"
    B3LYP_def2_QZVPPD_PCM_WATER_ = "B3LYP/def2-QZVPPD/PCM(WATER)"
    B3LYP_def2_QZVPPD_PCM_THF_ = "B3LYP/def2-QZVPPD/PCM(THF)"
    B3LYP_def2_QZVPPD_PCM_G2_ = "B3LYP/def2-QZVPPD/PCM(G2)"
    B3LYP_def2_QZVPPD_PCM_DME_ = "B3LYP/def2-QZVPPD/PCM(DME)"
    B3LYP_def2_QZVPPD_PCM_EC_ = "B3LYP/def2-QZVPPD/PCM(EC)"
    B3LYP_def2_QZVPPD_PCM_EC_EMC_ = "B3LYP/def2-QZVPPD/PCM(EC/EMC)"
    B3LYP_def2_QZVPPD_PCM_PC_ = "B3LYP/def2-QZVPPD/PCM(PC)"
    B3LYP_def2_QZVPPD_ISOSVP_WATER_ = "B3LYP/def2-QZVPPD/ISOSVP(WATER)"
    B3LYP_def2_QZVPPD_ISOSVP_THF_ = "B3LYP/def2-QZVPPD/ISOSVP(THF)"
    B3LYP_def2_QZVPPD_ISOSVP_G2_ = "B3LYP/def2-QZVPPD/ISOSVP(G2)"
    B3LYP_def2_QZVPPD_ISOSVP_DME_ = "B3LYP/def2-QZVPPD/ISOSVP(DME)"
    B3LYP_def2_QZVPPD_ISOSVP_EC_ = "B3LYP/def2-QZVPPD/ISOSVP(EC)"
    B3LYP_def2_QZVPPD_ISOSVP_EC_EMC_ = "B3LYP/def2-QZVPPD/ISOSVP(EC/EMC)"
    B3LYP_def2_QZVPPD_ISOSVP_PC_ = "B3LYP/def2-QZVPPD/ISOSVP(PC)"
    B3LYP_def2_QZVPPD_SMD_WATER_ = "B3LYP/def2-QZVPPD/SMD(WATER)"
    B3LYP_def2_QZVPPD_SMD_THF_ = "B3LYP/def2-QZVPPD/SMD(THF)"
    B3LYP_def2_QZVPPD_SMD_G2_ = "B3LYP/def2-QZVPPD/SMD(G2)"
    B3LYP_def2_QZVPPD_SMD_DME_ = "B3LYP/def2-QZVPPD/SMD(DME)"
    B3LYP_def2_QZVPPD_SMD_EC_ = "B3LYP/def2-QZVPPD/SMD(EC)"
    B3LYP_def2_QZVPPD_SMD_EC_EMC_ = "B3LYP/def2-QZVPPD/SMD(EC/EMC)"
    B3LYP_def2_QZVPPD_SMD_PC_ = "B3LYP/def2-QZVPPD/SMD(PC)"
    B3LYP_def2_QZVPPD_CMIRS_WATER_ = "B3LYP/def2-QZVPPD/CMIRS(WATER)"
    B3LYP_def2_QZVPPD_CMIRS_THF_ = "B3LYP/def2-QZVPPD/CMIRS(THF)"
    B3LYP_def2_QZVPPD_CMIRS_G2_ = "B3LYP/def2-QZVPPD/CMIRS(G2)"
    B3LYP_def2_QZVPPD_CMIRS_DME_ = "B3LYP/def2-QZVPPD/CMIRS(DME)"
    B3LYP_def2_QZVPPD_CMIRS_EC_ = "B3LYP/def2-QZVPPD/CMIRS(EC)"
    B3LYP_def2_QZVPPD_CMIRS_EC_EMC_ = "B3LYP/def2-QZVPPD/CMIRS(EC/EMC)"
    B3LYP_def2_QZVPPD_CMIRS_PC_ = "B3LYP/def2-QZVPPD/CMIRS(PC)"
    wB97X_D_6_31g_d_VACUUM = "wB97X-D/6-31g*/VACUUM"
    wB97X_D_6_31g_d_PCM_WATER_ = "wB97X-D/6-31g*/PCM(WATER)"
    wB97X_D_6_31g_d_PCM_THF_ = "wB97X-D/6-31g*/PCM(THF)"
    wB97X_D_6_31g_d_PCM_G2_ = "wB97X-D/6-31g*/PCM(G2)"
    wB97X_D_6_31g_d_PCM_DME_ = "wB97X-D/6-31g*/PCM(DME)"
    wB97X_D_6_31g_d_PCM_EC_ = "wB97X-D/6-31g*/PCM(EC)"
    wB97X_D_6_31g_d_PCM_EC_EMC_ = "wB97X-D/6-31g*/PCM(EC/EMC)"
    wB97X_D_6_31g_d_PCM_PC_ = "wB97X-D/6-31g*/PCM(PC)"
    wB97X_D_6_31g_d_ISOSVP_WATER_ = "wB97X-D/6-31g*/ISOSVP(WATER)"
    wB97X_D_6_31g_d_ISOSVP_THF_ = "wB97X-D/6-31g*/ISOSVP(THF)"
    wB97X_D_6_31g_d_ISOSVP_G2_ = "wB97X-D/6-31g*/ISOSVP(G2)"
    wB97X_D_6_31g_d_ISOSVP_DME_ = "wB97X-D/6-31g*/ISOSVP(DME)"
    wB97X_D_6_31g_d_ISOSVP_EC_ = "wB97X-D/6-31g*/ISOSVP(EC)"
    wB97X_D_6_31g_d_ISOSVP_EC_EMC_ = "wB97X-D/6-31g*/ISOSVP(EC/EMC)"
    wB97X_D_6_31g_d_ISOSVP_PC_ = "wB97X-D/6-31g*/ISOSVP(PC)"
    wB97X_D_6_31g_d_SMD_WATER_ = "wB97X-D/6-31g*/SMD(WATER)"
    wB97X_D_6_31g_d_SMD_THF_ = "wB97X-D/6-31g*/SMD(THF)"
    wB97X_D_6_31g_d_SMD_G2_ = "wB97X-D/6-31g*/SMD(G2)"
    wB97X_D_6_31g_d_SMD_DME_ = "wB97X-D/6-31g*/SMD(DME)"
    wB97X_D_6_31g_d_SMD_EC_ = "wB97X-D/6-31g*/SMD(EC)"
    wB97X_D_6_31g_d_SMD_EC_EMC_ = "wB97X-D/6-31g*/SMD(EC/EMC)"
    wB97X_D_6_31g_d_SMD_PC_ = "wB97X-D/6-31g*/SMD(PC)"
    wB97X_D_6_31g_d_CMIRS_WATER_ = "wB97X-D/6-31g*/CMIRS(WATER)"
    wB97X_D_6_31g_d_CMIRS_THF_ = "wB97X-D/6-31g*/CMIRS(THF)"
    wB97X_D_6_31g_d_CMIRS_G2_ = "wB97X-D/6-31g*/CMIRS(G2)"
    wB97X_D_6_31g_d_CMIRS_DME_ = "wB97X-D/6-31g*/CMIRS(DME)"
    wB97X_D_6_31g_d_CMIRS_EC_ = "wB97X-D/6-31g*/CMIRS(EC)"
    wB97X_D_6_31g_d_CMIRS_EC_EMC_ = "wB97X-D/6-31g*/CMIRS(EC/EMC)"
    wB97X_D_6_31g_d_CMIRS_PC_ = "wB97X-D/6-31g*/CMIRS(PC)"
    wB97X_D_def2_SVPD_VACUUM = "wB97X-D/def2-SVPD/VACUUM"
    wB97X_D_def2_SVPD_PCM_WATER_ = "wB97X-D/def2-SVPD/PCM(WATER)"
    wB97X_D_def2_SVPD_PCM_THF_ = "wB97X-D/def2-SVPD/PCM(THF)"
    wB97X_D_def2_SVPD_PCM_G2_ = "wB97X-D/def2-SVPD/PCM(G2)"
    wB97X_D_def2_SVPD_PCM_DME_ = "wB97X-D/def2-SVPD/PCM(DME)"
    wB97X_D_def2_SVPD_PCM_EC_ = "wB97X-D/def2-SVPD/PCM(EC)"
    wB97X_D_def2_SVPD_PCM_EC_EMC_ = "wB97X-D/def2-SVPD/PCM(EC/EMC)"
    wB97X_D_def2_SVPD_PCM_PC_ = "wB97X-D/def2-SVPD/PCM(PC)"
    wB97X_D_def2_SVPD_ISOSVP_WATER_ = "wB97X-D/def2-SVPD/ISOSVP(WATER)"
    wB97X_D_def2_SVPD_ISOSVP_THF_ = "wB97X-D/def2-SVPD/ISOSVP(THF)"
    wB97X_D_def2_SVPD_ISOSVP_G2_ = "wB97X-D/def2-SVPD/ISOSVP(G2)"
    wB97X_D_def2_SVPD_ISOSVP_DME_ = "wB97X-D/def2-SVPD/ISOSVP(DME)"
    wB97X_D_def2_SVPD_ISOSVP_EC_ = "wB97X-D/def2-SVPD/ISOSVP(EC)"
    wB97X_D_def2_SVPD_ISOSVP_EC_EMC_ = "wB97X-D/def2-SVPD/ISOSVP(EC/EMC)"
    wB97X_D_def2_SVPD_ISOSVP_PC_ = "wB97X-D/def2-SVPD/ISOSVP(PC)"
    wB97X_D_def2_SVPD_SMD_WATER_ = "wB97X-D/def2-SVPD/SMD(WATER)"
    wB97X_D_def2_SVPD_SMD_THF_ = "wB97X-D/def2-SVPD/SMD(THF)"
    wB97X_D_def2_SVPD_SMD_G2_ = "wB97X-D/def2-SVPD/SMD(G2)"
    wB97X_D_def2_SVPD_SMD_DME_ = "wB97X-D/def2-SVPD/SMD(DME)"
    wB97X_D_def2_SVPD_SMD_EC_ = "wB97X-D/def2-SVPD/SMD(EC)"
    wB97X_D_def2_SVPD_SMD_EC_EMC_ = "wB97X-D/def2-SVPD/SMD(EC/EMC)"
    wB97X_D_def2_SVPD_SMD_PC_ = "wB97X-D/def2-SVPD/SMD(PC)"
    wB97X_D_def2_SVPD_CMIRS_WATER_ = "wB97X-D/def2-SVPD/CMIRS(WATER)"
    wB97X_D_def2_SVPD_CMIRS_THF_ = "wB97X-D/def2-SVPD/CMIRS(THF)"
    wB97X_D_def2_SVPD_CMIRS_G2_ = "wB97X-D/def2-SVPD/CMIRS(G2)"
    wB97X_D_def2_SVPD_CMIRS_DME_ = "wB97X-D/def2-SVPD/CMIRS(DME)"
    wB97X_D_def2_SVPD_CMIRS_EC_ = "wB97X-D/def2-SVPD/CMIRS(EC)"
    wB97X_D_def2_SVPD_CMIRS_EC_EMC_ = "wB97X-D/def2-SVPD/CMIRS(EC/EMC)"
    wB97X_D_def2_SVPD_CMIRS_PC_ = "wB97X-D/def2-SVPD/CMIRS(PC)"
    wB97X_D_def2_TZVP_VACUUM = "wB97X-D/def2-TZVP/VACUUM"
    wB97X_D_def2_TZVP_PCM_WATER_ = "wB97X-D/def2-TZVP/PCM(WATER)"
    wB97X_D_def2_TZVP_PCM_THF_ = "wB97X-D/def2-TZVP/PCM(THF)"
    wB97X_D_def2_TZVP_PCM_G2_ = "wB97X-D/def2-TZVP/PCM(G2)"
    wB97X_D_def2_TZVP_PCM_DME_ = "wB97X-D/def2-TZVP/PCM(DME)"
    wB97X_D_def2_TZVP_PCM_EC_ = "wB97X-D/def2-TZVP/PCM(EC)"
    wB97X_D_def2_TZVP_PCM_EC_EMC_ = "wB97X-D/def2-TZVP/PCM(EC/EMC)"
    wB97X_D_def2_TZVP_PCM_PC_ = "wB97X-D/def2-TZVP/PCM(PC)"
    wB97X_D_def2_TZVP_ISOSVP_WATER_ = "wB97X-D/def2-TZVP/ISOSVP(WATER)"
    wB97X_D_def2_TZVP_ISOSVP_THF_ = "wB97X-D/def2-TZVP/ISOSVP(THF)"
    wB97X_D_def2_TZVP_ISOSVP_G2_ = "wB97X-D/def2-TZVP/ISOSVP(G2)"
    wB97X_D_def2_TZVP_ISOSVP_DME_ = "wB97X-D/def2-TZVP/ISOSVP(DME)"
    wB97X_D_def2_TZVP_ISOSVP_EC_ = "wB97X-D/def2-TZVP/ISOSVP(EC)"
    wB97X_D_def2_TZVP_ISOSVP_EC_EMC_ = "wB97X-D/def2-TZVP/ISOSVP(EC/EMC)"
    wB97X_D_def2_TZVP_ISOSVP_PC_ = "wB97X-D/def2-TZVP/ISOSVP(PC)"
    wB97X_D_def2_TZVP_SMD_WATER_ = "wB97X-D/def2-TZVP/SMD(WATER)"
    wB97X_D_def2_TZVP_SMD_THF_ = "wB97X-D/def2-TZVP/SMD(THF)"
    wB97X_D_def2_TZVP_SMD_G2_ = "wB97X-D/def2-TZVP/SMD(G2)"
    wB97X_D_def2_TZVP_SMD_DME_ = "wB97X-D/def2-TZVP/SMD(DME)"
    wB97X_D_def2_TZVP_SMD_EC_ = "wB97X-D/def2-TZVP/SMD(EC)"
    wB97X_D_def2_TZVP_SMD_EC_EMC_ = "wB97X-D/def2-TZVP/SMD(EC/EMC)"
    wB97X_D_def2_TZVP_SMD_PC_ = "wB97X-D/def2-TZVP/SMD(PC)"
    wB97X_D_def2_TZVP_CMIRS_WATER_ = "wB97X-D/def2-TZVP/CMIRS(WATER)"
    wB97X_D_def2_TZVP_CMIRS_THF_ = "wB97X-D/def2-TZVP/CMIRS(THF)"
    wB97X_D_def2_TZVP_CMIRS_G2_ = "wB97X-D/def2-TZVP/CMIRS(G2)"
    wB97X_D_def2_TZVP_CMIRS_DME_ = "wB97X-D/def2-TZVP/CMIRS(DME)"
    wB97X_D_def2_TZVP_CMIRS_EC_ = "wB97X-D/def2-TZVP/CMIRS(EC)"
    wB97X_D_def2_TZVP_CMIRS_EC_EMC_ = "wB97X-D/def2-TZVP/CMIRS(EC/EMC)"
    wB97X_D_def2_TZVP_CMIRS_PC_ = "wB97X-D/def2-TZVP/CMIRS(PC)"
    wB97X_D_def2_TZVPD_VACUUM = "wB97X-D/def2-TZVPD/VACUUM"
    wB97X_D_def2_TZVPD_PCM_WATER_ = "wB97X-D/def2-TZVPD/PCM(WATER)"
    wB97X_D_def2_TZVPD_PCM_THF_ = "wB97X-D/def2-TZVPD/PCM(THF)"
    wB97X_D_def2_TZVPD_PCM_G2_ = "wB97X-D/def2-TZVPD/PCM(G2)"
    wB97X_D_def2_TZVPD_PCM_DME_ = "wB97X-D/def2-TZVPD/PCM(DME)"
    wB97X_D_def2_TZVPD_PCM_EC_ = "wB97X-D/def2-TZVPD/PCM(EC)"
    wB97X_D_def2_TZVPD_PCM_EC_EMC_ = "wB97X-D/def2-TZVPD/PCM(EC/EMC)"
    wB97X_D_def2_TZVPD_PCM_PC_ = "wB97X-D/def2-TZVPD/PCM(PC)"
    wB97X_D_def2_TZVPD_ISOSVP_WATER_ = "wB97X-D/def2-TZVPD/ISOSVP(WATER)"
    wB97X_D_def2_TZVPD_ISOSVP_THF_ = "wB97X-D/def2-TZVPD/ISOSVP(THF)"
    wB97X_D_def2_TZVPD_ISOSVP_G2_ = "wB97X-D/def2-TZVPD/ISOSVP(G2)"
    wB97X_D_def2_TZVPD_ISOSVP_DME_ = "wB97X-D/def2-TZVPD/ISOSVP(DME)"
    wB97X_D_def2_TZVPD_ISOSVP_EC_ = "wB97X-D/def2-TZVPD/ISOSVP(EC)"
    wB97X_D_def2_TZVPD_ISOSVP_EC_EMC_ = "wB97X-D/def2-TZVPD/ISOSVP(EC/EMC)"
    wB97X_D_def2_TZVPD_ISOSVP_PC_ = "wB97X-D/def2-TZVPD/ISOSVP(PC)"
    wB97X_D_def2_TZVPD_SMD_WATER_ = "wB97X-D/def2-TZVPD/SMD(WATER)"
    wB97X_D_def2_TZVPD_SMD_THF_ = "wB97X-D/def2-TZVPD/SMD(THF)"
    wB97X_D_def2_TZVPD_SMD_G2_ = "wB97X-D/def2-TZVPD/SMD(G2)"
    wB97X_D_def2_TZVPD_SMD_DME_ = "wB97X-D/def2-TZVPD/SMD(DME)"
    wB97X_D_def2_TZVPD_SMD_EC_ = "wB97X-D/def2-TZVPD/SMD(EC)"
    wB97X_D_def2_TZVPD_SMD_EC_EMC_ = "wB97X-D/def2-TZVPD/SMD(EC/EMC)"
    wB97X_D_def2_TZVPD_SMD_PC_ = "wB97X-D/def2-TZVPD/SMD(PC)"
    wB97X_D_def2_TZVPD_CMIRS_WATER_ = "wB97X-D/def2-TZVPD/CMIRS(WATER)"
    wB97X_D_def2_TZVPD_CMIRS_THF_ = "wB97X-D/def2-TZVPD/CMIRS(THF)"
    wB97X_D_def2_TZVPD_CMIRS_G2_ = "wB97X-D/def2-TZVPD/CMIRS(G2)"
    wB97X_D_def2_TZVPD_CMIRS_DME_ = "wB97X-D/def2-TZVPD/CMIRS(DME)"
    wB97X_D_def2_TZVPD_CMIRS_EC_ = "wB97X-D/def2-TZVPD/CMIRS(EC)"
    wB97X_D_def2_TZVPD_CMIRS_EC_EMC_ = "wB97X-D/def2-TZVPD/CMIRS(EC/EMC)"
    wB97X_D_def2_TZVPD_CMIRS_PC_ = "wB97X-D/def2-TZVPD/CMIRS(PC)"
    wB97X_D_def2_TZVPP_VACUUM = "wB97X-D/def2-TZVPP/VACUUM"
    wB97X_D_def2_TZVPP_PCM_WATER_ = "wB97X-D/def2-TZVPP/PCM(WATER)"
    wB97X_D_def2_TZVPP_PCM_THF_ = "wB97X-D/def2-TZVPP/PCM(THF)"
    wB97X_D_def2_TZVPP_PCM_G2_ = "wB97X-D/def2-TZVPP/PCM(G2)"
    wB97X_D_def2_TZVPP_PCM_DME_ = "wB97X-D/def2-TZVPP/PCM(DME)"
    wB97X_D_def2_TZVPP_PCM_EC_ = "wB97X-D/def2-TZVPP/PCM(EC)"
    wB97X_D_def2_TZVPP_PCM_EC_EMC_ = "wB97X-D/def2-TZVPP/PCM(EC/EMC)"
    wB97X_D_def2_TZVPP_PCM_PC_ = "wB97X-D/def2-TZVPP/PCM(PC)"
    wB97X_D_def2_TZVPP_ISOSVP_WATER_ = "wB97X-D/def2-TZVPP/ISOSVP(WATER)"
    wB97X_D_def2_TZVPP_ISOSVP_THF_ = "wB97X-D/def2-TZVPP/ISOSVP(THF)"
    wB97X_D_def2_TZVPP_ISOSVP_G2_ = "wB97X-D/def2-TZVPP/ISOSVP(G2)"
    wB97X_D_def2_TZVPP_ISOSVP_DME_ = "wB97X-D/def2-TZVPP/ISOSVP(DME)"
    wB97X_D_def2_TZVPP_ISOSVP_EC_ = "wB97X-D/def2-TZVPP/ISOSVP(EC)"
    wB97X_D_def2_TZVPP_ISOSVP_EC_EMC_ = "wB97X-D/def2-TZVPP/ISOSVP(EC/EMC)"
    wB97X_D_def2_TZVPP_ISOSVP_PC_ = "wB97X-D/def2-TZVPP/ISOSVP(PC)"
    wB97X_D_def2_TZVPP_SMD_WATER_ = "wB97X-D/def2-TZVPP/SMD(WATER)"
    wB97X_D_def2_TZVPP_SMD_THF_ = "wB97X-D/def2-TZVPP/SMD(THF)"
    wB97X_D_def2_TZVPP_SMD_G2_ = "wB97X-D/def2-TZVPP/SMD(G2)"
    wB97X_D_def2_TZVPP_SMD_DME_ = "wB97X-D/def2-TZVPP/SMD(DME)"
    wB97X_D_def2_TZVPP_SMD_EC_ = "wB97X-D/def2-TZVPP/SMD(EC)"
    wB97X_D_def2_TZVPP_SMD_EC_EMC_ = "wB97X-D/def2-TZVPP/SMD(EC/EMC)"
    wB97X_D_def2_TZVPP_SMD_PC_ = "wB97X-D/def2-TZVPP/SMD(PC)"
    wB97X_D_def2_TZVPP_CMIRS_WATER_ = "wB97X-D/def2-TZVPP/CMIRS(WATER)"
    wB97X_D_def2_TZVPP_CMIRS_THF_ = "wB97X-D/def2-TZVPP/CMIRS(THF)"
    wB97X_D_def2_TZVPP_CMIRS_G2_ = "wB97X-D/def2-TZVPP/CMIRS(G2)"
    wB97X_D_def2_TZVPP_CMIRS_DME_ = "wB97X-D/def2-TZVPP/CMIRS(DME)"
    wB97X_D_def2_TZVPP_CMIRS_EC_ = "wB97X-D/def2-TZVPP/CMIRS(EC)"
    wB97X_D_def2_TZVPP_CMIRS_EC_EMC_ = "wB97X-D/def2-TZVPP/CMIRS(EC/EMC)"
    wB97X_D_def2_TZVPP_CMIRS_PC_ = "wB97X-D/def2-TZVPP/CMIRS(PC)"
    wB97X_D_def2_TZVPPD_VACUUM = "wB97X-D/def2-TZVPPD/VACUUM"
    wB97X_D_def2_TZVPPD_PCM_WATER_ = "wB97X-D/def2-TZVPPD/PCM(WATER)"
    wB97X_D_def2_TZVPPD_PCM_THF_ = "wB97X-D/def2-TZVPPD/PCM(THF)"
    wB97X_D_def2_TZVPPD_PCM_G2_ = "wB97X-D/def2-TZVPPD/PCM(G2)"
    wB97X_D_def2_TZVPPD_PCM_DME_ = "wB97X-D/def2-TZVPPD/PCM(DME)"
    wB97X_D_def2_TZVPPD_PCM_EC_ = "wB97X-D/def2-TZVPPD/PCM(EC)"
    wB97X_D_def2_TZVPPD_PCM_EC_EMC_ = "wB97X-D/def2-TZVPPD/PCM(EC/EMC)"
    wB97X_D_def2_TZVPPD_PCM_PC_ = "wB97X-D/def2-TZVPPD/PCM(PC)"
    wB97X_D_def2_TZVPPD_ISOSVP_WATER_ = "wB97X-D/def2-TZVPPD/ISOSVP(WATER)"
    wB97X_D_def2_TZVPPD_ISOSVP_THF_ = "wB97X-D/def2-TZVPPD/ISOSVP(THF)"
    wB97X_D_def2_TZVPPD_ISOSVP_G2_ = "wB97X-D/def2-TZVPPD/ISOSVP(G2)"
    wB97X_D_def2_TZVPPD_ISOSVP_DME_ = "wB97X-D/def2-TZVPPD/ISOSVP(DME)"
    wB97X_D_def2_TZVPPD_ISOSVP_EC_ = "wB97X-D/def2-TZVPPD/ISOSVP(EC)"
    wB97X_D_def2_TZVPPD_ISOSVP_EC_EMC_ = "wB97X-D/def2-TZVPPD/ISOSVP(EC/EMC)"
    wB97X_D_def2_TZVPPD_ISOSVP_PC_ = "wB97X-D/def2-TZVPPD/ISOSVP(PC)"
    wB97X_D_def2_TZVPPD_SMD_WATER_ = "wB97X-D/def2-TZVPPD/SMD(WATER)"
    wB97X_D_def2_TZVPPD_SMD_THF_ = "wB97X-D/def2-TZVPPD/SMD(THF)"
    wB97X_D_def2_TZVPPD_SMD_G2_ = "wB97X-D/def2-TZVPPD/SMD(G2)"
    wB97X_D_def2_TZVPPD_SMD_DME_ = "wB97X-D/def2-TZVPPD/SMD(DME)"
    wB97X_D_def2_TZVPPD_SMD_EC_ = "wB97X-D/def2-TZVPPD/SMD(EC)"
    wB97X_D_def2_TZVPPD_SMD_EC_EMC_ = "wB97X-D/def2-TZVPPD/SMD(EC/EMC)"
    wB97X_D_def2_TZVPPD_SMD_PC_ = "wB97X-D/def2-TZVPPD/SMD(PC)"
    wB97X_D_def2_TZVPPD_CMIRS_WATER_ = "wB97X-D/def2-TZVPPD/CMIRS(WATER)"
    wB97X_D_def2_TZVPPD_CMIRS_THF_ = "wB97X-D/def2-TZVPPD/CMIRS(THF)"
    wB97X_D_def2_TZVPPD_CMIRS_G2_ = "wB97X-D/def2-TZVPPD/CMIRS(G2)"
    wB97X_D_def2_TZVPPD_CMIRS_DME_ = "wB97X-D/def2-TZVPPD/CMIRS(DME)"
    wB97X_D_def2_TZVPPD_CMIRS_EC_ = "wB97X-D/def2-TZVPPD/CMIRS(EC)"
    wB97X_D_def2_TZVPPD_CMIRS_EC_EMC_ = "wB97X-D/def2-TZVPPD/CMIRS(EC/EMC)"
    wB97X_D_def2_TZVPPD_CMIRS_PC_ = "wB97X-D/def2-TZVPPD/CMIRS(PC)"
    wB97X_D_def2_QZVPPD_VACUUM = "wB97X-D/def2-QZVPPD/VACUUM"
    wB97X_D_def2_QZVPPD_PCM_WATER_ = "wB97X-D/def2-QZVPPD/PCM(WATER)"
    wB97X_D_def2_QZVPPD_PCM_THF_ = "wB97X-D/def2-QZVPPD/PCM(THF)"
    wB97X_D_def2_QZVPPD_PCM_G2_ = "wB97X-D/def2-QZVPPD/PCM(G2)"
    wB97X_D_def2_QZVPPD_PCM_DME_ = "wB97X-D/def2-QZVPPD/PCM(DME)"
    wB97X_D_def2_QZVPPD_PCM_EC_ = "wB97X-D/def2-QZVPPD/PCM(EC)"
    wB97X_D_def2_QZVPPD_PCM_EC_EMC_ = "wB97X-D/def2-QZVPPD/PCM(EC/EMC)"
    wB97X_D_def2_QZVPPD_PCM_PC_ = "wB97X-D/def2-QZVPPD/PCM(PC)"
    wB97X_D_def2_QZVPPD_ISOSVP_WATER_ = "wB97X-D/def2-QZVPPD/ISOSVP(WATER)"
    wB97X_D_def2_QZVPPD_ISOSVP_THF_ = "wB97X-D/def2-QZVPPD/ISOSVP(THF)"
    wB97X_D_def2_QZVPPD_ISOSVP_G2_ = "wB97X-D/def2-QZVPPD/ISOSVP(G2)"
    wB97X_D_def2_QZVPPD_ISOSVP_DME_ = "wB97X-D/def2-QZVPPD/ISOSVP(DME)"
    wB97X_D_def2_QZVPPD_ISOSVP_EC_ = "wB97X-D/def2-QZVPPD/ISOSVP(EC)"
    wB97X_D_def2_QZVPPD_ISOSVP_EC_EMC_ = "wB97X-D/def2-QZVPPD/ISOSVP(EC/EMC)"
    wB97X_D_def2_QZVPPD_ISOSVP_PC_ = "wB97X-D/def2-QZVPPD/ISOSVP(PC)"
    wB97X_D_def2_QZVPPD_SMD_WATER_ = "wB97X-D/def2-QZVPPD/SMD(WATER)"
    wB97X_D_def2_QZVPPD_SMD_THF_ = "wB97X-D/def2-QZVPPD/SMD(THF)"
    wB97X_D_def2_QZVPPD_SMD_G2_ = "wB97X-D/def2-QZVPPD/SMD(G2)"
    wB97X_D_def2_QZVPPD_SMD_DME_ = "wB97X-D/def2-QZVPPD/SMD(DME)"
    wB97X_D_def2_QZVPPD_SMD_EC_ = "wB97X-D/def2-QZVPPD/SMD(EC)"
    wB97X_D_def2_QZVPPD_SMD_EC_EMC_ = "wB97X-D/def2-QZVPPD/SMD(EC/EMC)"
    wB97X_D_def2_QZVPPD_SMD_PC_ = "wB97X-D/def2-QZVPPD/SMD(PC)"
    wB97X_D_def2_QZVPPD_CMIRS_WATER_ = "wB97X-D/def2-QZVPPD/CMIRS(WATER)"
    wB97X_D_def2_QZVPPD_CMIRS_THF_ = "wB97X-D/def2-QZVPPD/CMIRS(THF)"
    wB97X_D_def2_QZVPPD_CMIRS_G2_ = "wB97X-D/def2-QZVPPD/CMIRS(G2)"
    wB97X_D_def2_QZVPPD_CMIRS_DME_ = "wB97X-D/def2-QZVPPD/CMIRS(DME)"
    wB97X_D_def2_QZVPPD_CMIRS_EC_ = "wB97X-D/def2-QZVPPD/CMIRS(EC)"
    wB97X_D_def2_QZVPPD_CMIRS_EC_EMC_ = "wB97X-D/def2-QZVPPD/CMIRS(EC/EMC)"
    wB97X_D_def2_QZVPPD_CMIRS_PC_ = "wB97X-D/def2-QZVPPD/CMIRS(PC)"
    wB97X_D3_6_31g_d_VACUUM = "wB97X-D3/6-31g*/VACUUM"
    wB97X_D3_6_31g_d_PCM_WATER_ = "wB97X-D3/6-31g*/PCM(WATER)"
    wB97X_D3_6_31g_d_PCM_THF_ = "wB97X-D3/6-31g*/PCM(THF)"
    wB97X_D3_6_31g_d_PCM_G2_ = "wB97X-D3/6-31g*/PCM(G2)"
    wB97X_D3_6_31g_d_PCM_DME_ = "wB97X-D3/6-31g*/PCM(DME)"
    wB97X_D3_6_31g_d_PCM_EC_ = "wB97X-D3/6-31g*/PCM(EC)"
    wB97X_D3_6_31g_d_PCM_EC_EMC_ = "wB97X-D3/6-31g*/PCM(EC/EMC)"
    wB97X_D3_6_31g_d_PCM_PC_ = "wB97X-D3/6-31g*/PCM(PC)"
    wB97X_D3_6_31g_d_ISOSVP_WATER_ = "wB97X-D3/6-31g*/ISOSVP(WATER)"
    wB97X_D3_6_31g_d_ISOSVP_THF_ = "wB97X-D3/6-31g*/ISOSVP(THF)"
    wB97X_D3_6_31g_d_ISOSVP_G2_ = "wB97X-D3/6-31g*/ISOSVP(G2)"
    wB97X_D3_6_31g_d_ISOSVP_DME_ = "wB97X-D3/6-31g*/ISOSVP(DME)"
    wB97X_D3_6_31g_d_ISOSVP_EC_ = "wB97X-D3/6-31g*/ISOSVP(EC)"
    wB97X_D3_6_31g_d_ISOSVP_EC_EMC_ = "wB97X-D3/6-31g*/ISOSVP(EC/EMC)"
    wB97X_D3_6_31g_d_ISOSVP_PC_ = "wB97X-D3/6-31g*/ISOSVP(PC)"
    wB97X_D3_6_31g_d_SMD_WATER_ = "wB97X-D3/6-31g*/SMD(WATER)"
    wB97X_D3_6_31g_d_SMD_THF_ = "wB97X-D3/6-31g*/SMD(THF)"
    wB97X_D3_6_31g_d_SMD_G2_ = "wB97X-D3/6-31g*/SMD(G2)"
    wB97X_D3_6_31g_d_SMD_DME_ = "wB97X-D3/6-31g*/SMD(DME)"
    wB97X_D3_6_31g_d_SMD_EC_ = "wB97X-D3/6-31g*/SMD(EC)"
    wB97X_D3_6_31g_d_SMD_EC_EMC_ = "wB97X-D3/6-31g*/SMD(EC/EMC)"
    wB97X_D3_6_31g_d_SMD_PC_ = "wB97X-D3/6-31g*/SMD(PC)"
    wB97X_D3_6_31g_d_CMIRS_WATER_ = "wB97X-D3/6-31g*/CMIRS(WATER)"
    wB97X_D3_6_31g_d_CMIRS_THF_ = "wB97X-D3/6-31g*/CMIRS(THF)"
    wB97X_D3_6_31g_d_CMIRS_G2_ = "wB97X-D3/6-31g*/CMIRS(G2)"
    wB97X_D3_6_31g_d_CMIRS_DME_ = "wB97X-D3/6-31g*/CMIRS(DME)"
    wB97X_D3_6_31g_d_CMIRS_EC_ = "wB97X-D3/6-31g*/CMIRS(EC)"
    wB97X_D3_6_31g_d_CMIRS_EC_EMC_ = "wB97X-D3/6-31g*/CMIRS(EC/EMC)"
    wB97X_D3_6_31g_d_CMIRS_PC_ = "wB97X-D3/6-31g*/CMIRS(PC)"
    wB97X_D3_def2_SVPD_VACUUM = "wB97X-D3/def2-SVPD/VACUUM"
    wB97X_D3_def2_SVPD_PCM_WATER_ = "wB97X-D3/def2-SVPD/PCM(WATER)"
    wB97X_D3_def2_SVPD_PCM_THF_ = "wB97X-D3/def2-SVPD/PCM(THF)"
    wB97X_D3_def2_SVPD_PCM_G2_ = "wB97X-D3/def2-SVPD/PCM(G2)"
    wB97X_D3_def2_SVPD_PCM_DME_ = "wB97X-D3/def2-SVPD/PCM(DME)"
    wB97X_D3_def2_SVPD_PCM_EC_ = "wB97X-D3/def2-SVPD/PCM(EC)"
    wB97X_D3_def2_SVPD_PCM_EC_EMC_ = "wB97X-D3/def2-SVPD/PCM(EC/EMC)"
    wB97X_D3_def2_SVPD_PCM_PC_ = "wB97X-D3/def2-SVPD/PCM(PC)"
    wB97X_D3_def2_SVPD_ISOSVP_WATER_ = "wB97X-D3/def2-SVPD/ISOSVP(WATER)"
    wB97X_D3_def2_SVPD_ISOSVP_THF_ = "wB97X-D3/def2-SVPD/ISOSVP(THF)"
    wB97X_D3_def2_SVPD_ISOSVP_G2_ = "wB97X-D3/def2-SVPD/ISOSVP(G2)"
    wB97X_D3_def2_SVPD_ISOSVP_DME_ = "wB97X-D3/def2-SVPD/ISOSVP(DME)"
    wB97X_D3_def2_SVPD_ISOSVP_EC_ = "wB97X-D3/def2-SVPD/ISOSVP(EC)"
    wB97X_D3_def2_SVPD_ISOSVP_EC_EMC_ = "wB97X-D3/def2-SVPD/ISOSVP(EC/EMC)"
    wB97X_D3_def2_SVPD_ISOSVP_PC_ = "wB97X-D3/def2-SVPD/ISOSVP(PC)"
    wB97X_D3_def2_SVPD_SMD_WATER_ = "wB97X-D3/def2-SVPD/SMD(WATER)"
    wB97X_D3_def2_SVPD_SMD_THF_ = "wB97X-D3/def2-SVPD/SMD(THF)"
    wB97X_D3_def2_SVPD_SMD_G2_ = "wB97X-D3/def2-SVPD/SMD(G2)"
    wB97X_D3_def2_SVPD_SMD_DME_ = "wB97X-D3/def2-SVPD/SMD(DME)"
    wB97X_D3_def2_SVPD_SMD_EC_ = "wB97X-D3/def2-SVPD/SMD(EC)"
    wB97X_D3_def2_SVPD_SMD_EC_EMC_ = "wB97X-D3/def2-SVPD/SMD(EC/EMC)"
    wB97X_D3_def2_SVPD_SMD_PC_ = "wB97X-D3/def2-SVPD/SMD(PC)"
    wB97X_D3_def2_SVPD_CMIRS_WATER_ = "wB97X-D3/def2-SVPD/CMIRS(WATER)"
    wB97X_D3_def2_SVPD_CMIRS_THF_ = "wB97X-D3/def2-SVPD/CMIRS(THF)"
    wB97X_D3_def2_SVPD_CMIRS_G2_ = "wB97X-D3/def2-SVPD/CMIRS(G2)"
    wB97X_D3_def2_SVPD_CMIRS_DME_ = "wB97X-D3/def2-SVPD/CMIRS(DME)"
    wB97X_D3_def2_SVPD_CMIRS_EC_ = "wB97X-D3/def2-SVPD/CMIRS(EC)"
    wB97X_D3_def2_SVPD_CMIRS_EC_EMC_ = "wB97X-D3/def2-SVPD/CMIRS(EC/EMC)"
    wB97X_D3_def2_SVPD_CMIRS_PC_ = "wB97X-D3/def2-SVPD/CMIRS(PC)"
    wB97X_D3_def2_TZVP_VACUUM = "wB97X-D3/def2-TZVP/VACUUM"
    wB97X_D3_def2_TZVP_PCM_WATER_ = "wB97X-D3/def2-TZVP/PCM(WATER)"
    wB97X_D3_def2_TZVP_PCM_THF_ = "wB97X-D3/def2-TZVP/PCM(THF)"
    wB97X_D3_def2_TZVP_PCM_G2_ = "wB97X-D3/def2-TZVP/PCM(G2)"
    wB97X_D3_def2_TZVP_PCM_DME_ = "wB97X-D3/def2-TZVP/PCM(DME)"
    wB97X_D3_def2_TZVP_PCM_EC_ = "wB97X-D3/def2-TZVP/PCM(EC)"
    wB97X_D3_def2_TZVP_PCM_EC_EMC_ = "wB97X-D3/def2-TZVP/PCM(EC/EMC)"
    wB97X_D3_def2_TZVP_PCM_PC_ = "wB97X-D3/def2-TZVP/PCM(PC)"
    wB97X_D3_def2_TZVP_ISOSVP_WATER_ = "wB97X-D3/def2-TZVP/ISOSVP(WATER)"
    wB97X_D3_def2_TZVP_ISOSVP_THF_ = "wB97X-D3/def2-TZVP/ISOSVP(THF)"
    wB97X_D3_def2_TZVP_ISOSVP_G2_ = "wB97X-D3/def2-TZVP/ISOSVP(G2)"
    wB97X_D3_def2_TZVP_ISOSVP_DME_ = "wB97X-D3/def2-TZVP/ISOSVP(DME)"
    wB97X_D3_def2_TZVP_ISOSVP_EC_ = "wB97X-D3/def2-TZVP/ISOSVP(EC)"
    wB97X_D3_def2_TZVP_ISOSVP_EC_EMC_ = "wB97X-D3/def2-TZVP/ISOSVP(EC/EMC)"
    wB97X_D3_def2_TZVP_ISOSVP_PC_ = "wB97X-D3/def2-TZVP/ISOSVP(PC)"
    wB97X_D3_def2_TZVP_SMD_WATER_ = "wB97X-D3/def2-TZVP/SMD(WATER)"
    wB97X_D3_def2_TZVP_SMD_THF_ = "wB97X-D3/def2-TZVP/SMD(THF)"
    wB97X_D3_def2_TZVP_SMD_G2_ = "wB97X-D3/def2-TZVP/SMD(G2)"
    wB97X_D3_def2_TZVP_SMD_DME_ = "wB97X-D3/def2-TZVP/SMD(DME)"
    wB97X_D3_def2_TZVP_SMD_EC_ = "wB97X-D3/def2-TZVP/SMD(EC)"
    wB97X_D3_def2_TZVP_SMD_EC_EMC_ = "wB97X-D3/def2-TZVP/SMD(EC/EMC)"
    wB97X_D3_def2_TZVP_SMD_PC_ = "wB97X-D3/def2-TZVP/SMD(PC)"
    wB97X_D3_def2_TZVP_CMIRS_WATER_ = "wB97X-D3/def2-TZVP/CMIRS(WATER)"
    wB97X_D3_def2_TZVP_CMIRS_THF_ = "wB97X-D3/def2-TZVP/CMIRS(THF)"
    wB97X_D3_def2_TZVP_CMIRS_G2_ = "wB97X-D3/def2-TZVP/CMIRS(G2)"
    wB97X_D3_def2_TZVP_CMIRS_DME_ = "wB97X-D3/def2-TZVP/CMIRS(DME)"
    wB97X_D3_def2_TZVP_CMIRS_EC_ = "wB97X-D3/def2-TZVP/CMIRS(EC)"
    wB97X_D3_def2_TZVP_CMIRS_EC_EMC_ = "wB97X-D3/def2-TZVP/CMIRS(EC/EMC)"
    wB97X_D3_def2_TZVP_CMIRS_PC_ = "wB97X-D3/def2-TZVP/CMIRS(PC)"
    wB97X_D3_def2_TZVPD_VACUUM = "wB97X-D3/def2-TZVPD/VACUUM"
    wB97X_D3_def2_TZVPD_PCM_WATER_ = "wB97X-D3/def2-TZVPD/PCM(WATER)"
    wB97X_D3_def2_TZVPD_PCM_THF_ = "wB97X-D3/def2-TZVPD/PCM(THF)"
    wB97X_D3_def2_TZVPD_PCM_G2_ = "wB97X-D3/def2-TZVPD/PCM(G2)"
    wB97X_D3_def2_TZVPD_PCM_DME_ = "wB97X-D3/def2-TZVPD/PCM(DME)"
    wB97X_D3_def2_TZVPD_PCM_EC_ = "wB97X-D3/def2-TZVPD/PCM(EC)"
    wB97X_D3_def2_TZVPD_PCM_EC_EMC_ = "wB97X-D3/def2-TZVPD/PCM(EC/EMC)"
    wB97X_D3_def2_TZVPD_PCM_PC_ = "wB97X-D3/def2-TZVPD/PCM(PC)"
    wB97X_D3_def2_TZVPD_ISOSVP_WATER_ = "wB97X-D3/def2-TZVPD/ISOSVP(WATER)"
    wB97X_D3_def2_TZVPD_ISOSVP_THF_ = "wB97X-D3/def2-TZVPD/ISOSVP(THF)"
    wB97X_D3_def2_TZVPD_ISOSVP_G2_ = "wB97X-D3/def2-TZVPD/ISOSVP(G2)"
    wB97X_D3_def2_TZVPD_ISOSVP_DME_ = "wB97X-D3/def2-TZVPD/ISOSVP(DME)"
    wB97X_D3_def2_TZVPD_ISOSVP_EC_ = "wB97X-D3/def2-TZVPD/ISOSVP(EC)"
    wB97X_D3_def2_TZVPD_ISOSVP_EC_EMC_ = "wB97X-D3/def2-TZVPD/ISOSVP(EC/EMC)"
    wB97X_D3_def2_TZVPD_ISOSVP_PC_ = "wB97X-D3/def2-TZVPD/ISOSVP(PC)"
    wB97X_D3_def2_TZVPD_SMD_WATER_ = "wB97X-D3/def2-TZVPD/SMD(WATER)"
    wB97X_D3_def2_TZVPD_SMD_THF_ = "wB97X-D3/def2-TZVPD/SMD(THF)"
    wB97X_D3_def2_TZVPD_SMD_G2_ = "wB97X-D3/def2-TZVPD/SMD(G2)"
    wB97X_D3_def2_TZVPD_SMD_DME_ = "wB97X-D3/def2-TZVPD/SMD(DME)"
    wB97X_D3_def2_TZVPD_SMD_EC_ = "wB97X-D3/def2-TZVPD/SMD(EC)"
    wB97X_D3_def2_TZVPD_SMD_EC_EMC_ = "wB97X-D3/def2-TZVPD/SMD(EC/EMC)"
    wB97X_D3_def2_TZVPD_SMD_PC_ = "wB97X-D3/def2-TZVPD/SMD(PC)"
    wB97X_D3_def2_TZVPD_CMIRS_WATER_ = "wB97X-D3/def2-TZVPD/CMIRS(WATER)"
    wB97X_D3_def2_TZVPD_CMIRS_THF_ = "wB97X-D3/def2-TZVPD/CMIRS(THF)"
    wB97X_D3_def2_TZVPD_CMIRS_G2_ = "wB97X-D3/def2-TZVPD/CMIRS(G2)"
    wB97X_D3_def2_TZVPD_CMIRS_DME_ = "wB97X-D3/def2-TZVPD/CMIRS(DME)"
    wB97X_D3_def2_TZVPD_CMIRS_EC_ = "wB97X-D3/def2-TZVPD/CMIRS(EC)"
    wB97X_D3_def2_TZVPD_CMIRS_EC_EMC_ = "wB97X-D3/def2-TZVPD/CMIRS(EC/EMC)"
    wB97X_D3_def2_TZVPD_CMIRS_PC_ = "wB97X-D3/def2-TZVPD/CMIRS(PC)"
    wB97X_D3_def2_TZVPP_VACUUM = "wB97X-D3/def2-TZVPP/VACUUM"
    wB97X_D3_def2_TZVPP_PCM_WATER_ = "wB97X-D3/def2-TZVPP/PCM(WATER)"
    wB97X_D3_def2_TZVPP_PCM_THF_ = "wB97X-D3/def2-TZVPP/PCM(THF)"
    wB97X_D3_def2_TZVPP_PCM_G2_ = "wB97X-D3/def2-TZVPP/PCM(G2)"
    wB97X_D3_def2_TZVPP_PCM_DME_ = "wB97X-D3/def2-TZVPP/PCM(DME)"
    wB97X_D3_def2_TZVPP_PCM_EC_ = "wB97X-D3/def2-TZVPP/PCM(EC)"
    wB97X_D3_def2_TZVPP_PCM_EC_EMC_ = "wB97X-D3/def2-TZVPP/PCM(EC/EMC)"
    wB97X_D3_def2_TZVPP_PCM_PC_ = "wB97X-D3/def2-TZVPP/PCM(PC)"
    wB97X_D3_def2_TZVPP_ISOSVP_WATER_ = "wB97X-D3/def2-TZVPP/ISOSVP(WATER)"
    wB97X_D3_def2_TZVPP_ISOSVP_THF_ = "wB97X-D3/def2-TZVPP/ISOSVP(THF)"
    wB97X_D3_def2_TZVPP_ISOSVP_G2_ = "wB97X-D3/def2-TZVPP/ISOSVP(G2)"
    wB97X_D3_def2_TZVPP_ISOSVP_DME_ = "wB97X-D3/def2-TZVPP/ISOSVP(DME)"
    wB97X_D3_def2_TZVPP_ISOSVP_EC_ = "wB97X-D3/def2-TZVPP/ISOSVP(EC)"
    wB97X_D3_def2_TZVPP_ISOSVP_EC_EMC_ = "wB97X-D3/def2-TZVPP/ISOSVP(EC/EMC)"
    wB97X_D3_def2_TZVPP_ISOSVP_PC_ = "wB97X-D3/def2-TZVPP/ISOSVP(PC)"
    wB97X_D3_def2_TZVPP_SMD_WATER_ = "wB97X-D3/def2-TZVPP/SMD(WATER)"
    wB97X_D3_def2_TZVPP_SMD_THF_ = "wB97X-D3/def2-TZVPP/SMD(THF)"
    wB97X_D3_def2_TZVPP_SMD_G2_ = "wB97X-D3/def2-TZVPP/SMD(G2)"
    wB97X_D3_def2_TZVPP_SMD_DME_ = "wB97X-D3/def2-TZVPP/SMD(DME)"
    wB97X_D3_def2_TZVPP_SMD_EC_ = "wB97X-D3/def2-TZVPP/SMD(EC)"
    wB97X_D3_def2_TZVPP_SMD_EC_EMC_ = "wB97X-D3/def2-TZVPP/SMD(EC/EMC)"
    wB97X_D3_def2_TZVPP_SMD_PC_ = "wB97X-D3/def2-TZVPP/SMD(PC)"
    wB97X_D3_def2_TZVPP_CMIRS_WATER_ = "wB97X-D3/def2-TZVPP/CMIRS(WATER)"
    wB97X_D3_def2_TZVPP_CMIRS_THF_ = "wB97X-D3/def2-TZVPP/CMIRS(THF)"
    wB97X_D3_def2_TZVPP_CMIRS_G2_ = "wB97X-D3/def2-TZVPP/CMIRS(G2)"
    wB97X_D3_def2_TZVPP_CMIRS_DME_ = "wB97X-D3/def2-TZVPP/CMIRS(DME)"
    wB97X_D3_def2_TZVPP_CMIRS_EC_ = "wB97X-D3/def2-TZVPP/CMIRS(EC)"
    wB97X_D3_def2_TZVPP_CMIRS_EC_EMC_ = "wB97X-D3/def2-TZVPP/CMIRS(EC/EMC)"
    wB97X_D3_def2_TZVPP_CMIRS_PC_ = "wB97X-D3/def2-TZVPP/CMIRS(PC)"
    wB97X_D3_def2_TZVPPD_VACUUM = "wB97X-D3/def2-TZVPPD/VACUUM"
    wB97X_D3_def2_TZVPPD_PCM_WATER_ = "wB97X-D3/def2-TZVPPD/PCM(WATER)"
    wB97X_D3_def2_TZVPPD_PCM_THF_ = "wB97X-D3/def2-TZVPPD/PCM(THF)"
    wB97X_D3_def2_TZVPPD_PCM_G2_ = "wB97X-D3/def2-TZVPPD/PCM(G2)"
    wB97X_D3_def2_TZVPPD_PCM_DME_ = "wB97X-D3/def2-TZVPPD/PCM(DME)"
    wB97X_D3_def2_TZVPPD_PCM_EC_ = "wB97X-D3/def2-TZVPPD/PCM(EC)"
    wB97X_D3_def2_TZVPPD_PCM_EC_EMC_ = "wB97X-D3/def2-TZVPPD/PCM(EC/EMC)"
    wB97X_D3_def2_TZVPPD_PCM_PC_ = "wB97X-D3/def2-TZVPPD/PCM(PC)"
    wB97X_D3_def2_TZVPPD_ISOSVP_WATER_ = "wB97X-D3/def2-TZVPPD/ISOSVP(WATER)"
    wB97X_D3_def2_TZVPPD_ISOSVP_THF_ = "wB97X-D3/def2-TZVPPD/ISOSVP(THF)"
    wB97X_D3_def2_TZVPPD_ISOSVP_G2_ = "wB97X-D3/def2-TZVPPD/ISOSVP(G2)"
    wB97X_D3_def2_TZVPPD_ISOSVP_DME_ = "wB97X-D3/def2-TZVPPD/ISOSVP(DME)"
    wB97X_D3_def2_TZVPPD_ISOSVP_EC_ = "wB97X-D3/def2-TZVPPD/ISOSVP(EC)"
    wB97X_D3_def2_TZVPPD_ISOSVP_EC_EMC_ = "wB97X-D3/def2-TZVPPD/ISOSVP(EC/EMC)"
    wB97X_D3_def2_TZVPPD_ISOSVP_PC_ = "wB97X-D3/def2-TZVPPD/ISOSVP(PC)"
    wB97X_D3_def2_TZVPPD_SMD_WATER_ = "wB97X-D3/def2-TZVPPD/SMD(WATER)"
    wB97X_D3_def2_TZVPPD_SMD_THF_ = "wB97X-D3/def2-TZVPPD/SMD(THF)"
    wB97X_D3_def2_TZVPPD_SMD_G2_ = "wB97X-D3/def2-TZVPPD/SMD(G2)"
    wB97X_D3_def2_TZVPPD_SMD_DME_ = "wB97X-D3/def2-TZVPPD/SMD(DME)"
    wB97X_D3_def2_TZVPPD_SMD_EC_ = "wB97X-D3/def2-TZVPPD/SMD(EC)"
    wB97X_D3_def2_TZVPPD_SMD_EC_EMC_ = "wB97X-D3/def2-TZVPPD/SMD(EC/EMC)"
    wB97X_D3_def2_TZVPPD_SMD_PC_ = "wB97X-D3/def2-TZVPPD/SMD(PC)"
    wB97X_D3_def2_TZVPPD_CMIRS_WATER_ = "wB97X-D3/def2-TZVPPD/CMIRS(WATER)"
    wB97X_D3_def2_TZVPPD_CMIRS_THF_ = "wB97X-D3/def2-TZVPPD/CMIRS(THF)"
    wB97X_D3_def2_TZVPPD_CMIRS_G2_ = "wB97X-D3/def2-TZVPPD/CMIRS(G2)"
    wB97X_D3_def2_TZVPPD_CMIRS_DME_ = "wB97X-D3/def2-TZVPPD/CMIRS(DME)"
    wB97X_D3_def2_TZVPPD_CMIRS_EC_ = "wB97X-D3/def2-TZVPPD/CMIRS(EC)"
    wB97X_D3_def2_TZVPPD_CMIRS_EC_EMC_ = "wB97X-D3/def2-TZVPPD/CMIRS(EC/EMC)"
    wB97X_D3_def2_TZVPPD_CMIRS_PC_ = "wB97X-D3/def2-TZVPPD/CMIRS(PC)"
    wB97X_D3_def2_QZVPPD_VACUUM = "wB97X-D3/def2-QZVPPD/VACUUM"
    wB97X_D3_def2_QZVPPD_PCM_WATER_ = "wB97X-D3/def2-QZVPPD/PCM(WATER)"
    wB97X_D3_def2_QZVPPD_PCM_THF_ = "wB97X-D3/def2-QZVPPD/PCM(THF)"
    wB97X_D3_def2_QZVPPD_PCM_G2_ = "wB97X-D3/def2-QZVPPD/PCM(G2)"
    wB97X_D3_def2_QZVPPD_PCM_DME_ = "wB97X-D3/def2-QZVPPD/PCM(DME)"
    wB97X_D3_def2_QZVPPD_PCM_EC_ = "wB97X-D3/def2-QZVPPD/PCM(EC)"
    wB97X_D3_def2_QZVPPD_PCM_EC_EMC_ = "wB97X-D3/def2-QZVPPD/PCM(EC/EMC)"
    wB97X_D3_def2_QZVPPD_PCM_PC_ = "wB97X-D3/def2-QZVPPD/PCM(PC)"
    wB97X_D3_def2_QZVPPD_ISOSVP_WATER_ = "wB97X-D3/def2-QZVPPD/ISOSVP(WATER)"
    wB97X_D3_def2_QZVPPD_ISOSVP_THF_ = "wB97X-D3/def2-QZVPPD/ISOSVP(THF)"
    wB97X_D3_def2_QZVPPD_ISOSVP_G2_ = "wB97X-D3/def2-QZVPPD/ISOSVP(G2)"
    wB97X_D3_def2_QZVPPD_ISOSVP_DME_ = "wB97X-D3/def2-QZVPPD/ISOSVP(DME)"
    wB97X_D3_def2_QZVPPD_ISOSVP_EC_ = "wB97X-D3/def2-QZVPPD/ISOSVP(EC)"
    wB97X_D3_def2_QZVPPD_ISOSVP_EC_EMC_ = "wB97X-D3/def2-QZVPPD/ISOSVP(EC/EMC)"
    wB97X_D3_def2_QZVPPD_ISOSVP_PC_ = "wB97X-D3/def2-QZVPPD/ISOSVP(PC)"
    wB97X_D3_def2_QZVPPD_SMD_WATER_ = "wB97X-D3/def2-QZVPPD/SMD(WATER)"
    wB97X_D3_def2_QZVPPD_SMD_THF_ = "wB97X-D3/def2-QZVPPD/SMD(THF)"
    wB97X_D3_def2_QZVPPD_SMD_G2_ = "wB97X-D3/def2-QZVPPD/SMD(G2)"
    wB97X_D3_def2_QZVPPD_SMD_DME_ = "wB97X-D3/def2-QZVPPD/SMD(DME)"
    wB97X_D3_def2_QZVPPD_SMD_EC_ = "wB97X-D3/def2-QZVPPD/SMD(EC)"
    wB97X_D3_def2_QZVPPD_SMD_EC_EMC_ = "wB97X-D3/def2-QZVPPD/SMD(EC/EMC)"
    wB97X_D3_def2_QZVPPD_SMD_PC_ = "wB97X-D3/def2-QZVPPD/SMD(PC)"
    wB97X_D3_def2_QZVPPD_CMIRS_WATER_ = "wB97X-D3/def2-QZVPPD/CMIRS(WATER)"
    wB97X_D3_def2_QZVPPD_CMIRS_THF_ = "wB97X-D3/def2-QZVPPD/CMIRS(THF)"
    wB97X_D3_def2_QZVPPD_CMIRS_G2_ = "wB97X-D3/def2-QZVPPD/CMIRS(G2)"
    wB97X_D3_def2_QZVPPD_CMIRS_DME_ = "wB97X-D3/def2-QZVPPD/CMIRS(DME)"
    wB97X_D3_def2_QZVPPD_CMIRS_EC_ = "wB97X-D3/def2-QZVPPD/CMIRS(EC)"
    wB97X_D3_def2_QZVPPD_CMIRS_EC_EMC_ = "wB97X-D3/def2-QZVPPD/CMIRS(EC/EMC)"
    wB97X_D3_def2_QZVPPD_CMIRS_PC_ = "wB97X-D3/def2-QZVPPD/CMIRS(PC)"
    wB97X_V_6_31g_d_VACUUM = "wB97X-V/6-31g*/VACUUM"
    wB97X_V_6_31g_d_PCM_WATER_ = "wB97X-V/6-31g*/PCM(WATER)"
    wB97X_V_6_31g_d_PCM_THF_ = "wB97X-V/6-31g*/PCM(THF)"
    wB97X_V_6_31g_d_PCM_G2_ = "wB97X-V/6-31g*/PCM(G2)"
    wB97X_V_6_31g_d_PCM_DME_ = "wB97X-V/6-31g*/PCM(DME)"
    wB97X_V_6_31g_d_PCM_EC_ = "wB97X-V/6-31g*/PCM(EC)"
    wB97X_V_6_31g_d_PCM_EC_EMC_ = "wB97X-V/6-31g*/PCM(EC/EMC)"
    wB97X_V_6_31g_d_PCM_PC_ = "wB97X-V/6-31g*/PCM(PC)"
    wB97X_V_6_31g_d_ISOSVP_WATER_ = "wB97X-V/6-31g*/ISOSVP(WATER)"
    wB97X_V_6_31g_d_ISOSVP_THF_ = "wB97X-V/6-31g*/ISOSVP(THF)"
    wB97X_V_6_31g_d_ISOSVP_G2_ = "wB97X-V/6-31g*/ISOSVP(G2)"
    wB97X_V_6_31g_d_ISOSVP_DME_ = "wB97X-V/6-31g*/ISOSVP(DME)"
    wB97X_V_6_31g_d_ISOSVP_EC_ = "wB97X-V/6-31g*/ISOSVP(EC)"
    wB97X_V_6_31g_d_ISOSVP_EC_EMC_ = "wB97X-V/6-31g*/ISOSVP(EC/EMC)"
    wB97X_V_6_31g_d_ISOSVP_PC_ = "wB97X-V/6-31g*/ISOSVP(PC)"
    wB97X_V_6_31g_d_SMD_WATER_ = "wB97X-V/6-31g*/SMD(WATER)"
    wB97X_V_6_31g_d_SMD_THF_ = "wB97X-V/6-31g*/SMD(THF)"
    wB97X_V_6_31g_d_SMD_G2_ = "wB97X-V/6-31g*/SMD(G2)"
    wB97X_V_6_31g_d_SMD_DME_ = "wB97X-V/6-31g*/SMD(DME)"
    wB97X_V_6_31g_d_SMD_EC_ = "wB97X-V/6-31g*/SMD(EC)"
    wB97X_V_6_31g_d_SMD_EC_EMC_ = "wB97X-V/6-31g*/SMD(EC/EMC)"
    wB97X_V_6_31g_d_SMD_PC_ = "wB97X-V/6-31g*/SMD(PC)"
    wB97X_V_6_31g_d_CMIRS_WATER_ = "wB97X-V/6-31g*/CMIRS(WATER)"
    wB97X_V_6_31g_d_CMIRS_THF_ = "wB97X-V/6-31g*/CMIRS(THF)"
    wB97X_V_6_31g_d_CMIRS_G2_ = "wB97X-V/6-31g*/CMIRS(G2)"
    wB97X_V_6_31g_d_CMIRS_DME_ = "wB97X-V/6-31g*/CMIRS(DME)"
    wB97X_V_6_31g_d_CMIRS_EC_ = "wB97X-V/6-31g*/CMIRS(EC)"
    wB97X_V_6_31g_d_CMIRS_EC_EMC_ = "wB97X-V/6-31g*/CMIRS(EC/EMC)"
    wB97X_V_6_31g_d_CMIRS_PC_ = "wB97X-V/6-31g*/CMIRS(PC)"
    wB97X_V_def2_SVPD_VACUUM = "wB97X-V/def2-SVPD/VACUUM"
    wB97X_V_def2_SVPD_PCM_WATER_ = "wB97X-V/def2-SVPD/PCM(WATER)"
    wB97X_V_def2_SVPD_PCM_THF_ = "wB97X-V/def2-SVPD/PCM(THF)"
    wB97X_V_def2_SVPD_PCM_G2_ = "wB97X-V/def2-SVPD/PCM(G2)"
    wB97X_V_def2_SVPD_PCM_DME_ = "wB97X-V/def2-SVPD/PCM(DME)"
    wB97X_V_def2_SVPD_PCM_EC_ = "wB97X-V/def2-SVPD/PCM(EC)"
    wB97X_V_def2_SVPD_PCM_EC_EMC_ = "wB97X-V/def2-SVPD/PCM(EC/EMC)"
    wB97X_V_def2_SVPD_PCM_PC_ = "wB97X-V/def2-SVPD/PCM(PC)"
    wB97X_V_def2_SVPD_ISOSVP_WATER_ = "wB97X-V/def2-SVPD/ISOSVP(WATER)"
    wB97X_V_def2_SVPD_ISOSVP_THF_ = "wB97X-V/def2-SVPD/ISOSVP(THF)"
    wB97X_V_def2_SVPD_ISOSVP_G2_ = "wB97X-V/def2-SVPD/ISOSVP(G2)"
    wB97X_V_def2_SVPD_ISOSVP_DME_ = "wB97X-V/def2-SVPD/ISOSVP(DME)"
    wB97X_V_def2_SVPD_ISOSVP_EC_ = "wB97X-V/def2-SVPD/ISOSVP(EC)"
    wB97X_V_def2_SVPD_ISOSVP_EC_EMC_ = "wB97X-V/def2-SVPD/ISOSVP(EC/EMC)"
    wB97X_V_def2_SVPD_ISOSVP_PC_ = "wB97X-V/def2-SVPD/ISOSVP(PC)"
    wB97X_V_def2_SVPD_SMD_WATER_ = "wB97X-V/def2-SVPD/SMD(WATER)"
    wB97X_V_def2_SVPD_SMD_THF_ = "wB97X-V/def2-SVPD/SMD(THF)"
    wB97X_V_def2_SVPD_SMD_G2_ = "wB97X-V/def2-SVPD/SMD(G2)"
    wB97X_V_def2_SVPD_SMD_DME_ = "wB97X-V/def2-SVPD/SMD(DME)"
    wB97X_V_def2_SVPD_SMD_EC_ = "wB97X-V/def2-SVPD/SMD(EC)"
    wB97X_V_def2_SVPD_SMD_EC_EMC_ = "wB97X-V/def2-SVPD/SMD(EC/EMC)"
    wB97X_V_def2_SVPD_SMD_PC_ = "wB97X-V/def2-SVPD/SMD(PC)"
    wB97X_V_def2_SVPD_CMIRS_WATER_ = "wB97X-V/def2-SVPD/CMIRS(WATER)"
    wB97X_V_def2_SVPD_CMIRS_THF_ = "wB97X-V/def2-SVPD/CMIRS(THF)"
    wB97X_V_def2_SVPD_CMIRS_G2_ = "wB97X-V/def2-SVPD/CMIRS(G2)"
    wB97X_V_def2_SVPD_CMIRS_DME_ = "wB97X-V/def2-SVPD/CMIRS(DME)"
    wB97X_V_def2_SVPD_CMIRS_EC_ = "wB97X-V/def2-SVPD/CMIRS(EC)"
    wB97X_V_def2_SVPD_CMIRS_EC_EMC_ = "wB97X-V/def2-SVPD/CMIRS(EC/EMC)"
    wB97X_V_def2_SVPD_CMIRS_PC_ = "wB97X-V/def2-SVPD/CMIRS(PC)"
    wB97X_V_def2_TZVP_VACUUM = "wB97X-V/def2-TZVP/VACUUM"
    wB97X_V_def2_TZVP_PCM_WATER_ = "wB97X-V/def2-TZVP/PCM(WATER)"
    wB97X_V_def2_TZVP_PCM_THF_ = "wB97X-V/def2-TZVP/PCM(THF)"
    wB97X_V_def2_TZVP_PCM_G2_ = "wB97X-V/def2-TZVP/PCM(G2)"
    wB97X_V_def2_TZVP_PCM_DME_ = "wB97X-V/def2-TZVP/PCM(DME)"
    wB97X_V_def2_TZVP_PCM_EC_ = "wB97X-V/def2-TZVP/PCM(EC)"
    wB97X_V_def2_TZVP_PCM_EC_EMC_ = "wB97X-V/def2-TZVP/PCM(EC/EMC)"
    wB97X_V_def2_TZVP_PCM_PC_ = "wB97X-V/def2-TZVP/PCM(PC)"
    wB97X_V_def2_TZVP_ISOSVP_WATER_ = "wB97X-V/def2-TZVP/ISOSVP(WATER)"
    wB97X_V_def2_TZVP_ISOSVP_THF_ = "wB97X-V/def2-TZVP/ISOSVP(THF)"
    wB97X_V_def2_TZVP_ISOSVP_G2_ = "wB97X-V/def2-TZVP/ISOSVP(G2)"
    wB97X_V_def2_TZVP_ISOSVP_DME_ = "wB97X-V/def2-TZVP/ISOSVP(DME)"
    wB97X_V_def2_TZVP_ISOSVP_EC_ = "wB97X-V/def2-TZVP/ISOSVP(EC)"
    wB97X_V_def2_TZVP_ISOSVP_EC_EMC_ = "wB97X-V/def2-TZVP/ISOSVP(EC/EMC)"
    wB97X_V_def2_TZVP_ISOSVP_PC_ = "wB97X-V/def2-TZVP/ISOSVP(PC)"
    wB97X_V_def2_TZVP_SMD_WATER_ = "wB97X-V/def2-TZVP/SMD(WATER)"
    wB97X_V_def2_TZVP_SMD_THF_ = "wB97X-V/def2-TZVP/SMD(THF)"
    wB97X_V_def2_TZVP_SMD_G2_ = "wB97X-V/def2-TZVP/SMD(G2)"
    wB97X_V_def2_TZVP_SMD_DME_ = "wB97X-V/def2-TZVP/SMD(DME)"
    wB97X_V_def2_TZVP_SMD_EC_ = "wB97X-V/def2-TZVP/SMD(EC)"
    wB97X_V_def2_TZVP_SMD_EC_EMC_ = "wB97X-V/def2-TZVP/SMD(EC/EMC)"
    wB97X_V_def2_TZVP_SMD_PC_ = "wB97X-V/def2-TZVP/SMD(PC)"
    wB97X_V_def2_TZVP_CMIRS_WATER_ = "wB97X-V/def2-TZVP/CMIRS(WATER)"
    wB97X_V_def2_TZVP_CMIRS_THF_ = "wB97X-V/def2-TZVP/CMIRS(THF)"
    wB97X_V_def2_TZVP_CMIRS_G2_ = "wB97X-V/def2-TZVP/CMIRS(G2)"
    wB97X_V_def2_TZVP_CMIRS_DME_ = "wB97X-V/def2-TZVP/CMIRS(DME)"
    wB97X_V_def2_TZVP_CMIRS_EC_ = "wB97X-V/def2-TZVP/CMIRS(EC)"
    wB97X_V_def2_TZVP_CMIRS_EC_EMC_ = "wB97X-V/def2-TZVP/CMIRS(EC/EMC)"
    wB97X_V_def2_TZVP_CMIRS_PC_ = "wB97X-V/def2-TZVP/CMIRS(PC)"
    wB97X_V_def2_TZVPD_VACUUM = "wB97X-V/def2-TZVPD/VACUUM"
    wB97X_V_def2_TZVPD_PCM_WATER_ = "wB97X-V/def2-TZVPD/PCM(WATER)"
    wB97X_V_def2_TZVPD_PCM_THF_ = "wB97X-V/def2-TZVPD/PCM(THF)"
    wB97X_V_def2_TZVPD_PCM_G2_ = "wB97X-V/def2-TZVPD/PCM(G2)"
    wB97X_V_def2_TZVPD_PCM_DME_ = "wB97X-V/def2-TZVPD/PCM(DME)"
    wB97X_V_def2_TZVPD_PCM_EC_ = "wB97X-V/def2-TZVPD/PCM(EC)"
    wB97X_V_def2_TZVPD_PCM_EC_EMC_ = "wB97X-V/def2-TZVPD/PCM(EC/EMC)"
    wB97X_V_def2_TZVPD_PCM_PC_ = "wB97X-V/def2-TZVPD/PCM(PC)"
    wB97X_V_def2_TZVPD_ISOSVP_WATER_ = "wB97X-V/def2-TZVPD/ISOSVP(WATER)"
    wB97X_V_def2_TZVPD_ISOSVP_THF_ = "wB97X-V/def2-TZVPD/ISOSVP(THF)"
    wB97X_V_def2_TZVPD_ISOSVP_G2_ = "wB97X-V/def2-TZVPD/ISOSVP(G2)"
    wB97X_V_def2_TZVPD_ISOSVP_DME_ = "wB97X-V/def2-TZVPD/ISOSVP(DME)"
    wB97X_V_def2_TZVPD_ISOSVP_EC_ = "wB97X-V/def2-TZVPD/ISOSVP(EC)"
    wB97X_V_def2_TZVPD_ISOSVP_EC_EMC_ = "wB97X-V/def2-TZVPD/ISOSVP(EC/EMC)"
    wB97X_V_def2_TZVPD_ISOSVP_PC_ = "wB97X-V/def2-TZVPD/ISOSVP(PC)"
    wB97X_V_def2_TZVPD_SMD_WATER_ = "wB97X-V/def2-TZVPD/SMD(WATER)"
    wB97X_V_def2_TZVPD_SMD_THF_ = "wB97X-V/def2-TZVPD/SMD(THF)"
    wB97X_V_def2_TZVPD_SMD_G2_ = "wB97X-V/def2-TZVPD/SMD(G2)"
    wB97X_V_def2_TZVPD_SMD_DME_ = "wB97X-V/def2-TZVPD/SMD(DME)"
    wB97X_V_def2_TZVPD_SMD_EC_ = "wB97X-V/def2-TZVPD/SMD(EC)"
    wB97X_V_def2_TZVPD_SMD_EC_EMC_ = "wB97X-V/def2-TZVPD/SMD(EC/EMC)"
    wB97X_V_def2_TZVPD_SMD_PC_ = "wB97X-V/def2-TZVPD/SMD(PC)"
    wB97X_V_def2_TZVPD_CMIRS_WATER_ = "wB97X-V/def2-TZVPD/CMIRS(WATER)"
    wB97X_V_def2_TZVPD_CMIRS_THF_ = "wB97X-V/def2-TZVPD/CMIRS(THF)"
    wB97X_V_def2_TZVPD_CMIRS_G2_ = "wB97X-V/def2-TZVPD/CMIRS(G2)"
    wB97X_V_def2_TZVPD_CMIRS_DME_ = "wB97X-V/def2-TZVPD/CMIRS(DME)"
    wB97X_V_def2_TZVPD_CMIRS_EC_ = "wB97X-V/def2-TZVPD/CMIRS(EC)"
    wB97X_V_def2_TZVPD_CMIRS_EC_EMC_ = "wB97X-V/def2-TZVPD/CMIRS(EC/EMC)"
    wB97X_V_def2_TZVPD_CMIRS_PC_ = "wB97X-V/def2-TZVPD/CMIRS(PC)"
    wB97X_V_def2_TZVPP_VACUUM = "wB97X-V/def2-TZVPP/VACUUM"
    wB97X_V_def2_TZVPP_PCM_WATER_ = "wB97X-V/def2-TZVPP/PCM(WATER)"
    wB97X_V_def2_TZVPP_PCM_THF_ = "wB97X-V/def2-TZVPP/PCM(THF)"
    wB97X_V_def2_TZVPP_PCM_G2_ = "wB97X-V/def2-TZVPP/PCM(G2)"
    wB97X_V_def2_TZVPP_PCM_DME_ = "wB97X-V/def2-TZVPP/PCM(DME)"
    wB97X_V_def2_TZVPP_PCM_EC_ = "wB97X-V/def2-TZVPP/PCM(EC)"
    wB97X_V_def2_TZVPP_PCM_EC_EMC_ = "wB97X-V/def2-TZVPP/PCM(EC/EMC)"
    wB97X_V_def2_TZVPP_PCM_PC_ = "wB97X-V/def2-TZVPP/PCM(PC)"
    wB97X_V_def2_TZVPP_ISOSVP_WATER_ = "wB97X-V/def2-TZVPP/ISOSVP(WATER)"
    wB97X_V_def2_TZVPP_ISOSVP_THF_ = "wB97X-V/def2-TZVPP/ISOSVP(THF)"
    wB97X_V_def2_TZVPP_ISOSVP_G2_ = "wB97X-V/def2-TZVPP/ISOSVP(G2)"
    wB97X_V_def2_TZVPP_ISOSVP_DME_ = "wB97X-V/def2-TZVPP/ISOSVP(DME)"
    wB97X_V_def2_TZVPP_ISOSVP_EC_ = "wB97X-V/def2-TZVPP/ISOSVP(EC)"
    wB97X_V_def2_TZVPP_ISOSVP_EC_EMC_ = "wB97X-V/def2-TZVPP/ISOSVP(EC/EMC)"
    wB97X_V_def2_TZVPP_ISOSVP_PC_ = "wB97X-V/def2-TZVPP/ISOSVP(PC)"
    wB97X_V_def2_TZVPP_SMD_WATER_ = "wB97X-V/def2-TZVPP/SMD(WATER)"
    wB97X_V_def2_TZVPP_SMD_THF_ = "wB97X-V/def2-TZVPP/SMD(THF)"
    wB97X_V_def2_TZVPP_SMD_G2_ = "wB97X-V/def2-TZVPP/SMD(G2)"
    wB97X_V_def2_TZVPP_SMD_DME_ = "wB97X-V/def2-TZVPP/SMD(DME)"
    wB97X_V_def2_TZVPP_SMD_EC_ = "wB97X-V/def2-TZVPP/SMD(EC)"
    wB97X_V_def2_TZVPP_SMD_EC_EMC_ = "wB97X-V/def2-TZVPP/SMD(EC/EMC)"
    wB97X_V_def2_TZVPP_SMD_PC_ = "wB97X-V/def2-TZVPP/SMD(PC)"
    wB97X_V_def2_TZVPP_CMIRS_WATER_ = "wB97X-V/def2-TZVPP/CMIRS(WATER)"
    wB97X_V_def2_TZVPP_CMIRS_THF_ = "wB97X-V/def2-TZVPP/CMIRS(THF)"
    wB97X_V_def2_TZVPP_CMIRS_G2_ = "wB97X-V/def2-TZVPP/CMIRS(G2)"
    wB97X_V_def2_TZVPP_CMIRS_DME_ = "wB97X-V/def2-TZVPP/CMIRS(DME)"
    wB97X_V_def2_TZVPP_CMIRS_EC_ = "wB97X-V/def2-TZVPP/CMIRS(EC)"
    wB97X_V_def2_TZVPP_CMIRS_EC_EMC_ = "wB97X-V/def2-TZVPP/CMIRS(EC/EMC)"
    wB97X_V_def2_TZVPP_CMIRS_PC_ = "wB97X-V/def2-TZVPP/CMIRS(PC)"
    wB97X_V_def2_TZVPPD_VACUUM = "wB97X-V/def2-TZVPPD/VACUUM"
    wB97X_V_def2_TZVPPD_PCM_WATER_ = "wB97X-V/def2-TZVPPD/PCM(WATER)"
    wB97X_V_def2_TZVPPD_PCM_THF_ = "wB97X-V/def2-TZVPPD/PCM(THF)"
    wB97X_V_def2_TZVPPD_PCM_G2_ = "wB97X-V/def2-TZVPPD/PCM(G2)"
    wB97X_V_def2_TZVPPD_PCM_DME_ = "wB97X-V/def2-TZVPPD/PCM(DME)"
    wB97X_V_def2_TZVPPD_PCM_EC_ = "wB97X-V/def2-TZVPPD/PCM(EC)"
    wB97X_V_def2_TZVPPD_PCM_EC_EMC_ = "wB97X-V/def2-TZVPPD/PCM(EC/EMC)"
    wB97X_V_def2_TZVPPD_PCM_PC_ = "wB97X-V/def2-TZVPPD/PCM(PC)"
    wB97X_V_def2_TZVPPD_ISOSVP_WATER_ = "wB97X-V/def2-TZVPPD/ISOSVP(WATER)"
    wB97X_V_def2_TZVPPD_ISOSVP_THF_ = "wB97X-V/def2-TZVPPD/ISOSVP(THF)"
    wB97X_V_def2_TZVPPD_ISOSVP_G2_ = "wB97X-V/def2-TZVPPD/ISOSVP(G2)"
    wB97X_V_def2_TZVPPD_ISOSVP_DME_ = "wB97X-V/def2-TZVPPD/ISOSVP(DME)"
    wB97X_V_def2_TZVPPD_ISOSVP_EC_ = "wB97X-V/def2-TZVPPD/ISOSVP(EC)"
    wB97X_V_def2_TZVPPD_ISOSVP_EC_EMC_ = "wB97X-V/def2-TZVPPD/ISOSVP(EC/EMC)"
    wB97X_V_def2_TZVPPD_ISOSVP_PC_ = "wB97X-V/def2-TZVPPD/ISOSVP(PC)"
    wB97X_V_def2_TZVPPD_SMD_WATER_ = "wB97X-V/def2-TZVPPD/SMD(WATER)"
    wB97X_V_def2_TZVPPD_SMD_THF_ = "wB97X-V/def2-TZVPPD/SMD(THF)"
    wB97X_V_def2_TZVPPD_SMD_G2_ = "wB97X-V/def2-TZVPPD/SMD(G2)"
    wB97X_V_def2_TZVPPD_SMD_DME_ = "wB97X-V/def2-TZVPPD/SMD(DME)"
    wB97X_V_def2_TZVPPD_SMD_EC_ = "wB97X-V/def2-TZVPPD/SMD(EC)"
    wB97X_V_def2_TZVPPD_SMD_EC_EMC_ = "wB97X-V/def2-TZVPPD/SMD(EC/EMC)"
    wB97X_V_def2_TZVPPD_SMD_PC_ = "wB97X-V/def2-TZVPPD/SMD(PC)"
    wB97X_V_def2_TZVPPD_CMIRS_WATER_ = "wB97X-V/def2-TZVPPD/CMIRS(WATER)"
    wB97X_V_def2_TZVPPD_CMIRS_THF_ = "wB97X-V/def2-TZVPPD/CMIRS(THF)"
    wB97X_V_def2_TZVPPD_CMIRS_G2_ = "wB97X-V/def2-TZVPPD/CMIRS(G2)"
    wB97X_V_def2_TZVPPD_CMIRS_DME_ = "wB97X-V/def2-TZVPPD/CMIRS(DME)"
    wB97X_V_def2_TZVPPD_CMIRS_EC_ = "wB97X-V/def2-TZVPPD/CMIRS(EC)"
    wB97X_V_def2_TZVPPD_CMIRS_EC_EMC_ = "wB97X-V/def2-TZVPPD/CMIRS(EC/EMC)"
    wB97X_V_def2_TZVPPD_CMIRS_PC_ = "wB97X-V/def2-TZVPPD/CMIRS(PC)"
    wB97X_V_def2_QZVPPD_VACUUM = "wB97X-V/def2-QZVPPD/VACUUM"
    wB97X_V_def2_QZVPPD_PCM_WATER_ = "wB97X-V/def2-QZVPPD/PCM(WATER)"
    wB97X_V_def2_QZVPPD_PCM_THF_ = "wB97X-V/def2-QZVPPD/PCM(THF)"
    wB97X_V_def2_QZVPPD_PCM_G2_ = "wB97X-V/def2-QZVPPD/PCM(G2)"
    wB97X_V_def2_QZVPPD_PCM_DME_ = "wB97X-V/def2-QZVPPD/PCM(DME)"
    wB97X_V_def2_QZVPPD_PCM_EC_ = "wB97X-V/def2-QZVPPD/PCM(EC)"
    wB97X_V_def2_QZVPPD_PCM_EC_EMC_ = "wB97X-V/def2-QZVPPD/PCM(EC/EMC)"
    wB97X_V_def2_QZVPPD_PCM_PC_ = "wB97X-V/def2-QZVPPD/PCM(PC)"
    wB97X_V_def2_QZVPPD_ISOSVP_WATER_ = "wB97X-V/def2-QZVPPD/ISOSVP(WATER)"
    wB97X_V_def2_QZVPPD_ISOSVP_THF_ = "wB97X-V/def2-QZVPPD/ISOSVP(THF)"
    wB97X_V_def2_QZVPPD_ISOSVP_G2_ = "wB97X-V/def2-QZVPPD/ISOSVP(G2)"
    wB97X_V_def2_QZVPPD_ISOSVP_DME_ = "wB97X-V/def2-QZVPPD/ISOSVP(DME)"
    wB97X_V_def2_QZVPPD_ISOSVP_EC_ = "wB97X-V/def2-QZVPPD/ISOSVP(EC)"
    wB97X_V_def2_QZVPPD_ISOSVP_EC_EMC_ = "wB97X-V/def2-QZVPPD/ISOSVP(EC/EMC)"
    wB97X_V_def2_QZVPPD_ISOSVP_PC_ = "wB97X-V/def2-QZVPPD/ISOSVP(PC)"
    wB97X_V_def2_QZVPPD_SMD_WATER_ = "wB97X-V/def2-QZVPPD/SMD(WATER)"
    wB97X_V_def2_QZVPPD_SMD_THF_ = "wB97X-V/def2-QZVPPD/SMD(THF)"
    wB97X_V_def2_QZVPPD_SMD_G2_ = "wB97X-V/def2-QZVPPD/SMD(G2)"
    wB97X_V_def2_QZVPPD_SMD_DME_ = "wB97X-V/def2-QZVPPD/SMD(DME)"
    wB97X_V_def2_QZVPPD_SMD_EC_ = "wB97X-V/def2-QZVPPD/SMD(EC)"
    wB97X_V_def2_QZVPPD_SMD_EC_EMC_ = "wB97X-V/def2-QZVPPD/SMD(EC/EMC)"
    wB97X_V_def2_QZVPPD_SMD_PC_ = "wB97X-V/def2-QZVPPD/SMD(PC)"
    wB97X_V_def2_QZVPPD_CMIRS_WATER_ = "wB97X-V/def2-QZVPPD/CMIRS(WATER)"
    wB97X_V_def2_QZVPPD_CMIRS_THF_ = "wB97X-V/def2-QZVPPD/CMIRS(THF)"
    wB97X_V_def2_QZVPPD_CMIRS_G2_ = "wB97X-V/def2-QZVPPD/CMIRS(G2)"
    wB97X_V_def2_QZVPPD_CMIRS_DME_ = "wB97X-V/def2-QZVPPD/CMIRS(DME)"
    wB97X_V_def2_QZVPPD_CMIRS_EC_ = "wB97X-V/def2-QZVPPD/CMIRS(EC)"
    wB97X_V_def2_QZVPPD_CMIRS_EC_EMC_ = "wB97X-V/def2-QZVPPD/CMIRS(EC/EMC)"
    wB97X_V_def2_QZVPPD_CMIRS_PC_ = "wB97X-V/def2-QZVPPD/CMIRS(PC)"
    wB97M_V_6_31g_d_VACUUM = "wB97M-V/6-31g*/VACUUM"
    wB97M_V_6_31g_d_PCM_WATER_ = "wB97M-V/6-31g*/PCM(WATER)"
    wB97M_V_6_31g_d_PCM_THF_ = "wB97M-V/6-31g*/PCM(THF)"
    wB97M_V_6_31g_d_PCM_G2_ = "wB97M-V/6-31g*/PCM(G2)"
    wB97M_V_6_31g_d_PCM_DME_ = "wB97M-V/6-31g*/PCM(DME)"
    wB97M_V_6_31g_d_PCM_EC_ = "wB97M-V/6-31g*/PCM(EC)"
    wB97M_V_6_31g_d_PCM_EC_EMC_ = "wB97M-V/6-31g*/PCM(EC/EMC)"
    wB97M_V_6_31g_d_PCM_PC_ = "wB97M-V/6-31g*/PCM(PC)"
    wB97M_V_6_31g_d_ISOSVP_WATER_ = "wB97M-V/6-31g*/ISOSVP(WATER)"
    wB97M_V_6_31g_d_ISOSVP_THF_ = "wB97M-V/6-31g*/ISOSVP(THF)"
    wB97M_V_6_31g_d_ISOSVP_G2_ = "wB97M-V/6-31g*/ISOSVP(G2)"
    wB97M_V_6_31g_d_ISOSVP_DME_ = "wB97M-V/6-31g*/ISOSVP(DME)"
    wB97M_V_6_31g_d_ISOSVP_EC_ = "wB97M-V/6-31g*/ISOSVP(EC)"
    wB97M_V_6_31g_d_ISOSVP_EC_EMC_ = "wB97M-V/6-31g*/ISOSVP(EC/EMC)"
    wB97M_V_6_31g_d_ISOSVP_PC_ = "wB97M-V/6-31g*/ISOSVP(PC)"
    wB97M_V_6_31g_d_SMD_WATER_ = "wB97M-V/6-31g*/SMD(WATER)"
    wB97M_V_6_31g_d_SMD_THF_ = "wB97M-V/6-31g*/SMD(THF)"
    wB97M_V_6_31g_d_SMD_G2_ = "wB97M-V/6-31g*/SMD(G2)"
    wB97M_V_6_31g_d_SMD_DME_ = "wB97M-V/6-31g*/SMD(DME)"
    wB97M_V_6_31g_d_SMD_EC_ = "wB97M-V/6-31g*/SMD(EC)"
    wB97M_V_6_31g_d_SMD_EC_EMC_ = "wB97M-V/6-31g*/SMD(EC/EMC)"
    wB97M_V_6_31g_d_SMD_PC_ = "wB97M-V/6-31g*/SMD(PC)"
    wB97M_V_6_31g_d_CMIRS_WATER_ = "wB97M-V/6-31g*/CMIRS(WATER)"
    wB97M_V_6_31g_d_CMIRS_THF_ = "wB97M-V/6-31g*/CMIRS(THF)"
    wB97M_V_6_31g_d_CMIRS_G2_ = "wB97M-V/6-31g*/CMIRS(G2)"
    wB97M_V_6_31g_d_CMIRS_DME_ = "wB97M-V/6-31g*/CMIRS(DME)"
    wB97M_V_6_31g_d_CMIRS_EC_ = "wB97M-V/6-31g*/CMIRS(EC)"
    wB97M_V_6_31g_d_CMIRS_EC_EMC_ = "wB97M-V/6-31g*/CMIRS(EC/EMC)"
    wB97M_V_6_31g_d_CMIRS_PC_ = "wB97M-V/6-31g*/CMIRS(PC)"
    wB97M_V_def2_SVPD_VACUUM = "wB97M-V/def2-SVPD/VACUUM"
    wB97M_V_def2_SVPD_PCM_WATER_ = "wB97M-V/def2-SVPD/PCM(WATER)"
    wB97M_V_def2_SVPD_PCM_THF_ = "wB97M-V/def2-SVPD/PCM(THF)"
    wB97M_V_def2_SVPD_PCM_G2_ = "wB97M-V/def2-SVPD/PCM(G2)"
    wB97M_V_def2_SVPD_PCM_DME_ = "wB97M-V/def2-SVPD/PCM(DME)"
    wB97M_V_def2_SVPD_PCM_EC_ = "wB97M-V/def2-SVPD/PCM(EC)"
    wB97M_V_def2_SVPD_PCM_EC_EMC_ = "wB97M-V/def2-SVPD/PCM(EC/EMC)"
    wB97M_V_def2_SVPD_PCM_PC_ = "wB97M-V/def2-SVPD/PCM(PC)"
    wB97M_V_def2_SVPD_ISOSVP_WATER_ = "wB97M-V/def2-SVPD/ISOSVP(WATER)"
    wB97M_V_def2_SVPD_ISOSVP_THF_ = "wB97M-V/def2-SVPD/ISOSVP(THF)"
    wB97M_V_def2_SVPD_ISOSVP_G2_ = "wB97M-V/def2-SVPD/ISOSVP(G2)"
    wB97M_V_def2_SVPD_ISOSVP_DME_ = "wB97M-V/def2-SVPD/ISOSVP(DME)"
    wB97M_V_def2_SVPD_ISOSVP_EC_ = "wB97M-V/def2-SVPD/ISOSVP(EC)"
    wB97M_V_def2_SVPD_ISOSVP_EC_EMC_ = "wB97M-V/def2-SVPD/ISOSVP(EC/EMC)"
    wB97M_V_def2_SVPD_ISOSVP_PC_ = "wB97M-V/def2-SVPD/ISOSVP(PC)"
    wB97M_V_def2_SVPD_SMD_WATER_ = "wB97M-V/def2-SVPD/SMD(WATER)"
    wB97M_V_def2_SVPD_SMD_THF_ = "wB97M-V/def2-SVPD/SMD(THF)"
    wB97M_V_def2_SVPD_SMD_G2_ = "wB97M-V/def2-SVPD/SMD(G2)"
    wB97M_V_def2_SVPD_SMD_DME_ = "wB97M-V/def2-SVPD/SMD(DME)"
    wB97M_V_def2_SVPD_SMD_EC_ = "wB97M-V/def2-SVPD/SMD(EC)"
    wB97M_V_def2_SVPD_SMD_EC_EMC_ = "wB97M-V/def2-SVPD/SMD(EC/EMC)"
    wB97M_V_def2_SVPD_SMD_PC_ = "wB97M-V/def2-SVPD/SMD(PC)"
    wB97M_V_def2_SVPD_CMIRS_WATER_ = "wB97M-V/def2-SVPD/CMIRS(WATER)"
    wB97M_V_def2_SVPD_CMIRS_THF_ = "wB97M-V/def2-SVPD/CMIRS(THF)"
    wB97M_V_def2_SVPD_CMIRS_G2_ = "wB97M-V/def2-SVPD/CMIRS(G2)"
    wB97M_V_def2_SVPD_CMIRS_DME_ = "wB97M-V/def2-SVPD/CMIRS(DME)"
    wB97M_V_def2_SVPD_CMIRS_EC_ = "wB97M-V/def2-SVPD/CMIRS(EC)"
    wB97M_V_def2_SVPD_CMIRS_EC_EMC_ = "wB97M-V/def2-SVPD/CMIRS(EC/EMC)"
    wB97M_V_def2_SVPD_CMIRS_PC_ = "wB97M-V/def2-SVPD/CMIRS(PC)"
    wB97M_V_def2_TZVP_VACUUM = "wB97M-V/def2-TZVP/VACUUM"
    wB97M_V_def2_TZVP_PCM_WATER_ = "wB97M-V/def2-TZVP/PCM(WATER)"
    wB97M_V_def2_TZVP_PCM_THF_ = "wB97M-V/def2-TZVP/PCM(THF)"
    wB97M_V_def2_TZVP_PCM_G2_ = "wB97M-V/def2-TZVP/PCM(G2)"
    wB97M_V_def2_TZVP_PCM_DME_ = "wB97M-V/def2-TZVP/PCM(DME)"
    wB97M_V_def2_TZVP_PCM_EC_ = "wB97M-V/def2-TZVP/PCM(EC)"
    wB97M_V_def2_TZVP_PCM_EC_EMC_ = "wB97M-V/def2-TZVP/PCM(EC/EMC)"
    wB97M_V_def2_TZVP_PCM_PC_ = "wB97M-V/def2-TZVP/PCM(PC)"
    wB97M_V_def2_TZVP_ISOSVP_WATER_ = "wB97M-V/def2-TZVP/ISOSVP(WATER)"
    wB97M_V_def2_TZVP_ISOSVP_THF_ = "wB97M-V/def2-TZVP/ISOSVP(THF)"
    wB97M_V_def2_TZVP_ISOSVP_G2_ = "wB97M-V/def2-TZVP/ISOSVP(G2)"
    wB97M_V_def2_TZVP_ISOSVP_DME_ = "wB97M-V/def2-TZVP/ISOSVP(DME)"
    wB97M_V_def2_TZVP_ISOSVP_EC_ = "wB97M-V/def2-TZVP/ISOSVP(EC)"
    wB97M_V_def2_TZVP_ISOSVP_EC_EMC_ = "wB97M-V/def2-TZVP/ISOSVP(EC/EMC)"
    wB97M_V_def2_TZVP_ISOSVP_PC_ = "wB97M-V/def2-TZVP/ISOSVP(PC)"
    wB97M_V_def2_TZVP_SMD_WATER_ = "wB97M-V/def2-TZVP/SMD(WATER)"
    wB97M_V_def2_TZVP_SMD_THF_ = "wB97M-V/def2-TZVP/SMD(THF)"
    wB97M_V_def2_TZVP_SMD_G2_ = "wB97M-V/def2-TZVP/SMD(G2)"
    wB97M_V_def2_TZVP_SMD_DME_ = "wB97M-V/def2-TZVP/SMD(DME)"
    wB97M_V_def2_TZVP_SMD_EC_ = "wB97M-V/def2-TZVP/SMD(EC)"
    wB97M_V_def2_TZVP_SMD_EC_EMC_ = "wB97M-V/def2-TZVP/SMD(EC/EMC)"
    wB97M_V_def2_TZVP_SMD_PC_ = "wB97M-V/def2-TZVP/SMD(PC)"
    wB97M_V_def2_TZVP_CMIRS_WATER_ = "wB97M-V/def2-TZVP/CMIRS(WATER)"
    wB97M_V_def2_TZVP_CMIRS_THF_ = "wB97M-V/def2-TZVP/CMIRS(THF)"
    wB97M_V_def2_TZVP_CMIRS_G2_ = "wB97M-V/def2-TZVP/CMIRS(G2)"
    wB97M_V_def2_TZVP_CMIRS_DME_ = "wB97M-V/def2-TZVP/CMIRS(DME)"
    wB97M_V_def2_TZVP_CMIRS_EC_ = "wB97M-V/def2-TZVP/CMIRS(EC)"
    wB97M_V_def2_TZVP_CMIRS_EC_EMC_ = "wB97M-V/def2-TZVP/CMIRS(EC/EMC)"
    wB97M_V_def2_TZVP_CMIRS_PC_ = "wB97M-V/def2-TZVP/CMIRS(PC)"
    wB97M_V_def2_TZVPD_VACUUM = "wB97M-V/def2-TZVPD/VACUUM"
    wB97M_V_def2_TZVPD_PCM_WATER_ = "wB97M-V/def2-TZVPD/PCM(WATER)"
    wB97M_V_def2_TZVPD_PCM_THF_ = "wB97M-V/def2-TZVPD/PCM(THF)"
    wB97M_V_def2_TZVPD_PCM_G2_ = "wB97M-V/def2-TZVPD/PCM(G2)"
    wB97M_V_def2_TZVPD_PCM_DME_ = "wB97M-V/def2-TZVPD/PCM(DME)"
    wB97M_V_def2_TZVPD_PCM_EC_ = "wB97M-V/def2-TZVPD/PCM(EC)"
    wB97M_V_def2_TZVPD_PCM_EC_EMC_ = "wB97M-V/def2-TZVPD/PCM(EC/EMC)"
    wB97M_V_def2_TZVPD_PCM_PC_ = "wB97M-V/def2-TZVPD/PCM(PC)"
    wB97M_V_def2_TZVPD_ISOSVP_WATER_ = "wB97M-V/def2-TZVPD/ISOSVP(WATER)"
    wB97M_V_def2_TZVPD_ISOSVP_THF_ = "wB97M-V/def2-TZVPD/ISOSVP(THF)"
    wB97M_V_def2_TZVPD_ISOSVP_G2_ = "wB97M-V/def2-TZVPD/ISOSVP(G2)"
    wB97M_V_def2_TZVPD_ISOSVP_DME_ = "wB97M-V/def2-TZVPD/ISOSVP(DME)"
    wB97M_V_def2_TZVPD_ISOSVP_EC_ = "wB97M-V/def2-TZVPD/ISOSVP(EC)"
    wB97M_V_def2_TZVPD_ISOSVP_EC_EMC_ = "wB97M-V/def2-TZVPD/ISOSVP(EC/EMC)"
    wB97M_V_def2_TZVPD_ISOSVP_PC_ = "wB97M-V/def2-TZVPD/ISOSVP(PC)"
    wB97M_V_def2_TZVPD_SMD_WATER_ = "wB97M-V/def2-TZVPD/SMD(WATER)"
    wB97M_V_def2_TZVPD_SMD_THF_ = "wB97M-V/def2-TZVPD/SMD(THF)"
    wB97M_V_def2_TZVPD_SMD_G2_ = "wB97M-V/def2-TZVPD/SMD(G2)"
    wB97M_V_def2_TZVPD_SMD_DME_ = "wB97M-V/def2-TZVPD/SMD(DME)"
    wB97M_V_def2_TZVPD_SMD_EC_ = "wB97M-V/def2-TZVPD/SMD(EC)"
    wB97M_V_def2_TZVPD_SMD_EC_EMC_ = "wB97M-V/def2-TZVPD/SMD(EC/EMC)"
    wB97M_V_def2_TZVPD_SMD_PC_ = "wB97M-V/def2-TZVPD/SMD(PC)"
    wB97M_V_def2_TZVPD_CMIRS_WATER_ = "wB97M-V/def2-TZVPD/CMIRS(WATER)"
    wB97M_V_def2_TZVPD_CMIRS_THF_ = "wB97M-V/def2-TZVPD/CMIRS(THF)"
    wB97M_V_def2_TZVPD_CMIRS_G2_ = "wB97M-V/def2-TZVPD/CMIRS(G2)"
    wB97M_V_def2_TZVPD_CMIRS_DME_ = "wB97M-V/def2-TZVPD/CMIRS(DME)"
    wB97M_V_def2_TZVPD_CMIRS_EC_ = "wB97M-V/def2-TZVPD/CMIRS(EC)"
    wB97M_V_def2_TZVPD_CMIRS_EC_EMC_ = "wB97M-V/def2-TZVPD/CMIRS(EC/EMC)"
    wB97M_V_def2_TZVPD_CMIRS_PC_ = "wB97M-V/def2-TZVPD/CMIRS(PC)"
    wB97M_V_def2_TZVPP_VACUUM = "wB97M-V/def2-TZVPP/VACUUM"
    wB97M_V_def2_TZVPP_PCM_WATER_ = "wB97M-V/def2-TZVPP/PCM(WATER)"
    wB97M_V_def2_TZVPP_PCM_THF_ = "wB97M-V/def2-TZVPP/PCM(THF)"
    wB97M_V_def2_TZVPP_PCM_G2_ = "wB97M-V/def2-TZVPP/PCM(G2)"
    wB97M_V_def2_TZVPP_PCM_DME_ = "wB97M-V/def2-TZVPP/PCM(DME)"
    wB97M_V_def2_TZVPP_PCM_EC_ = "wB97M-V/def2-TZVPP/PCM(EC)"
    wB97M_V_def2_TZVPP_PCM_EC_EMC_ = "wB97M-V/def2-TZVPP/PCM(EC/EMC)"
    wB97M_V_def2_TZVPP_PCM_PC_ = "wB97M-V/def2-TZVPP/PCM(PC)"
    wB97M_V_def2_TZVPP_ISOSVP_WATER_ = "wB97M-V/def2-TZVPP/ISOSVP(WATER)"
    wB97M_V_def2_TZVPP_ISOSVP_THF_ = "wB97M-V/def2-TZVPP/ISOSVP(THF)"
    wB97M_V_def2_TZVPP_ISOSVP_G2_ = "wB97M-V/def2-TZVPP/ISOSVP(G2)"
    wB97M_V_def2_TZVPP_ISOSVP_DME_ = "wB97M-V/def2-TZVPP/ISOSVP(DME)"
    wB97M_V_def2_TZVPP_ISOSVP_EC_ = "wB97M-V/def2-TZVPP/ISOSVP(EC)"
    wB97M_V_def2_TZVPP_ISOSVP_EC_EMC_ = "wB97M-V/def2-TZVPP/ISOSVP(EC/EMC)"
    wB97M_V_def2_TZVPP_ISOSVP_PC_ = "wB97M-V/def2-TZVPP/ISOSVP(PC)"
    wB97M_V_def2_TZVPP_SMD_WATER_ = "wB97M-V/def2-TZVPP/SMD(WATER)"
    wB97M_V_def2_TZVPP_SMD_THF_ = "wB97M-V/def2-TZVPP/SMD(THF)"
    wB97M_V_def2_TZVPP_SMD_G2_ = "wB97M-V/def2-TZVPP/SMD(G2)"
    wB97M_V_def2_TZVPP_SMD_DME_ = "wB97M-V/def2-TZVPP/SMD(DME)"
    wB97M_V_def2_TZVPP_SMD_EC_ = "wB97M-V/def2-TZVPP/SMD(EC)"
    wB97M_V_def2_TZVPP_SMD_EC_EMC_ = "wB97M-V/def2-TZVPP/SMD(EC/EMC)"
    wB97M_V_def2_TZVPP_SMD_PC_ = "wB97M-V/def2-TZVPP/SMD(PC)"
    wB97M_V_def2_TZVPP_CMIRS_WATER_ = "wB97M-V/def2-TZVPP/CMIRS(WATER)"
    wB97M_V_def2_TZVPP_CMIRS_THF_ = "wB97M-V/def2-TZVPP/CMIRS(THF)"
    wB97M_V_def2_TZVPP_CMIRS_G2_ = "wB97M-V/def2-TZVPP/CMIRS(G2)"
    wB97M_V_def2_TZVPP_CMIRS_DME_ = "wB97M-V/def2-TZVPP/CMIRS(DME)"
    wB97M_V_def2_TZVPP_CMIRS_EC_ = "wB97M-V/def2-TZVPP/CMIRS(EC)"
    wB97M_V_def2_TZVPP_CMIRS_EC_EMC_ = "wB97M-V/def2-TZVPP/CMIRS(EC/EMC)"
    wB97M_V_def2_TZVPP_CMIRS_PC_ = "wB97M-V/def2-TZVPP/CMIRS(PC)"
    wB97M_V_def2_TZVPPD_VACUUM = "wB97M-V/def2-TZVPPD/VACUUM"
    wB97M_V_def2_TZVPPD_PCM_WATER_ = "wB97M-V/def2-TZVPPD/PCM(WATER)"
    wB97M_V_def2_TZVPPD_PCM_THF_ = "wB97M-V/def2-TZVPPD/PCM(THF)"
    wB97M_V_def2_TZVPPD_PCM_G2_ = "wB97M-V/def2-TZVPPD/PCM(G2)"
    wB97M_V_def2_TZVPPD_PCM_DME_ = "wB97M-V/def2-TZVPPD/PCM(DME)"
    wB97M_V_def2_TZVPPD_PCM_EC_ = "wB97M-V/def2-TZVPPD/PCM(EC)"
    wB97M_V_def2_TZVPPD_PCM_EC_EMC_ = "wB97M-V/def2-TZVPPD/PCM(EC/EMC)"
    wB97M_V_def2_TZVPPD_PCM_PC_ = "wB97M-V/def2-TZVPPD/PCM(PC)"
    wB97M_V_def2_TZVPPD_ISOSVP_WATER_ = "wB97M-V/def2-TZVPPD/ISOSVP(WATER)"
    wB97M_V_def2_TZVPPD_ISOSVP_THF_ = "wB97M-V/def2-TZVPPD/ISOSVP(THF)"
    wB97M_V_def2_TZVPPD_ISOSVP_G2_ = "wB97M-V/def2-TZVPPD/ISOSVP(G2)"
    wB97M_V_def2_TZVPPD_ISOSVP_DME_ = "wB97M-V/def2-TZVPPD/ISOSVP(DME)"
    wB97M_V_def2_TZVPPD_ISOSVP_EC_ = "wB97M-V/def2-TZVPPD/ISOSVP(EC)"
    wB97M_V_def2_TZVPPD_ISOSVP_EC_EMC_ = "wB97M-V/def2-TZVPPD/ISOSVP(EC/EMC)"
    wB97M_V_def2_TZVPPD_ISOSVP_PC_ = "wB97M-V/def2-TZVPPD/ISOSVP(PC)"
    wB97M_V_def2_TZVPPD_SMD_WATER_ = "wB97M-V/def2-TZVPPD/SMD(WATER)"
    wB97M_V_def2_TZVPPD_SMD_THF_ = "wB97M-V/def2-TZVPPD/SMD(THF)"
    wB97M_V_def2_TZVPPD_SMD_G2_ = "wB97M-V/def2-TZVPPD/SMD(G2)"
    wB97M_V_def2_TZVPPD_SMD_DME_ = "wB97M-V/def2-TZVPPD/SMD(DME)"
    wB97M_V_def2_TZVPPD_SMD_EC_ = "wB97M-V/def2-TZVPPD/SMD(EC)"
    wB97M_V_def2_TZVPPD_SMD_EC_EMC_ = "wB97M-V/def2-TZVPPD/SMD(EC/EMC)"
    wB97M_V_def2_TZVPPD_SMD_PC_ = "wB97M-V/def2-TZVPPD/SMD(PC)"
    wB97M_V_def2_TZVPPD_CMIRS_WATER_ = "wB97M-V/def2-TZVPPD/CMIRS(WATER)"
    wB97M_V_def2_TZVPPD_CMIRS_THF_ = "wB97M-V/def2-TZVPPD/CMIRS(THF)"
    wB97M_V_def2_TZVPPD_CMIRS_G2_ = "wB97M-V/def2-TZVPPD/CMIRS(G2)"
    wB97M_V_def2_TZVPPD_CMIRS_DME_ = "wB97M-V/def2-TZVPPD/CMIRS(DME)"
    wB97M_V_def2_TZVPPD_CMIRS_EC_ = "wB97M-V/def2-TZVPPD/CMIRS(EC)"
    wB97M_V_def2_TZVPPD_CMIRS_EC_EMC_ = "wB97M-V/def2-TZVPPD/CMIRS(EC/EMC)"
    wB97M_V_def2_TZVPPD_CMIRS_PC_ = "wB97M-V/def2-TZVPPD/CMIRS(PC)"
    wB97M_V_def2_QZVPPD_VACUUM = "wB97M-V/def2-QZVPPD/VACUUM"
    wB97M_V_def2_QZVPPD_PCM_WATER_ = "wB97M-V/def2-QZVPPD/PCM(WATER)"
    wB97M_V_def2_QZVPPD_PCM_THF_ = "wB97M-V/def2-QZVPPD/PCM(THF)"
    wB97M_V_def2_QZVPPD_PCM_G2_ = "wB97M-V/def2-QZVPPD/PCM(G2)"
    wB97M_V_def2_QZVPPD_PCM_DME_ = "wB97M-V/def2-QZVPPD/PCM(DME)"
    wB97M_V_def2_QZVPPD_PCM_EC_ = "wB97M-V/def2-QZVPPD/PCM(EC)"
    wB97M_V_def2_QZVPPD_PCM_EC_EMC_ = "wB97M-V/def2-QZVPPD/PCM(EC/EMC)"
    wB97M_V_def2_QZVPPD_PCM_PC_ = "wB97M-V/def2-QZVPPD/PCM(PC)"
    wB97M_V_def2_QZVPPD_ISOSVP_WATER_ = "wB97M-V/def2-QZVPPD/ISOSVP(WATER)"
    wB97M_V_def2_QZVPPD_ISOSVP_THF_ = "wB97M-V/def2-QZVPPD/ISOSVP(THF)"
    wB97M_V_def2_QZVPPD_ISOSVP_G2_ = "wB97M-V/def2-QZVPPD/ISOSVP(G2)"
    wB97M_V_def2_QZVPPD_ISOSVP_DME_ = "wB97M-V/def2-QZVPPD/ISOSVP(DME)"
    wB97M_V_def2_QZVPPD_ISOSVP_EC_ = "wB97M-V/def2-QZVPPD/ISOSVP(EC)"
    wB97M_V_def2_QZVPPD_ISOSVP_EC_EMC_ = "wB97M-V/def2-QZVPPD/ISOSVP(EC/EMC)"
    wB97M_V_def2_QZVPPD_ISOSVP_PC_ = "wB97M-V/def2-QZVPPD/ISOSVP(PC)"
    wB97M_V_def2_QZVPPD_SMD_WATER_ = "wB97M-V/def2-QZVPPD/SMD(WATER)"
    wB97M_V_def2_QZVPPD_SMD_THF_ = "wB97M-V/def2-QZVPPD/SMD(THF)"
    wB97M_V_def2_QZVPPD_SMD_G2_ = "wB97M-V/def2-QZVPPD/SMD(G2)"
    wB97M_V_def2_QZVPPD_SMD_DME_ = "wB97M-V/def2-QZVPPD/SMD(DME)"
    wB97M_V_def2_QZVPPD_SMD_EC_ = "wB97M-V/def2-QZVPPD/SMD(EC)"
    wB97M_V_def2_QZVPPD_SMD_EC_EMC_ = "wB97M-V/def2-QZVPPD/SMD(EC/EMC)"
    wB97M_V_def2_QZVPPD_SMD_PC_ = "wB97M-V/def2-QZVPPD/SMD(PC)"
    wB97M_V_def2_QZVPPD_CMIRS_WATER_ = "wB97M-V/def2-QZVPPD/CMIRS(WATER)"
    wB97M_V_def2_QZVPPD_CMIRS_THF_ = "wB97M-V/def2-QZVPPD/CMIRS(THF)"
    wB97M_V_def2_QZVPPD_CMIRS_G2_ = "wB97M-V/def2-QZVPPD/CMIRS(G2)"
    wB97M_V_def2_QZVPPD_CMIRS_DME_ = "wB97M-V/def2-QZVPPD/CMIRS(DME)"
    wB97M_V_def2_QZVPPD_CMIRS_EC_ = "wB97M-V/def2-QZVPPD/CMIRS(EC)"
    wB97M_V_def2_QZVPPD_CMIRS_EC_EMC_ = "wB97M-V/def2-QZVPPD/CMIRS(EC/EMC)"
    wB97M_V_def2_QZVPPD_CMIRS_PC_ = "wB97M-V/def2-QZVPPD/CMIRS(PC)"


class TaskType(ValueEnum):
    """ Calculation task types for Q-Chem """

    Single_Point = "Single Point"
    Force = "Force"
    Geometry_Optimization = "Geometry Optimization"
    Frequency_Analysis = "Frequency Analysis"
    Frequency_Flattening_Geometry_Optimization = "Frequency Flattening Geometry Optimization"
    Transition_State_Geometry_Optimization = "Transition State Geometry Optimization"
    Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "Frequency Flattening Transition State Geometry Optimization"
    )
    Unknown = "Unknown"


class CalcType(ValueEnum):
    """ Calculation types (LOT + task type) for Q-Chem """

    PBE_6_31g_d_VACUUM_Single_Point = "PBE/6-31g*/VACUUM Single Point"
    PBE_6_31g_d_VACUUM_Force = "PBE/6-31g*/VACUUM Force"
    PBE_6_31g_d_VACUUM_Geometry_Optimization = "PBE/6-31g*/VACUUM Geometry Optimization"
    PBE_6_31g_d_VACUUM_Frequency_Analysis = "PBE/6-31g*/VACUUM Frequency Analysis"
    PBE_6_31g_d_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/VACUUM Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_VACUUM_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/VACUUM Transition State Geometry Optimization"
    )
    PBE_6_31g_d_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_VACUUM_Unknown = "PBE/6-31g*/VACUUM Unknown"
    PBE_6_31g_d_PCM_WATER__Single_Point = "PBE/6-31g*/PCM(WATER) Single Point"
    PBE_6_31g_d_PCM_WATER__Force = "PBE/6-31g*/PCM(WATER) Force"
    PBE_6_31g_d_PCM_WATER__Geometry_Optimization = "PBE/6-31g*/PCM(WATER) Geometry Optimization"
    PBE_6_31g_d_PCM_WATER__Frequency_Analysis = "PBE/6-31g*/PCM(WATER) Frequency Analysis"
    PBE_6_31g_d_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_PCM_WATER__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(WATER) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_WATER__Unknown = "PBE/6-31g*/PCM(WATER) Unknown"
    PBE_6_31g_d_PCM_THF__Single_Point = "PBE/6-31g*/PCM(THF) Single Point"
    PBE_6_31g_d_PCM_THF__Force = "PBE/6-31g*/PCM(THF) Force"
    PBE_6_31g_d_PCM_THF__Geometry_Optimization = "PBE/6-31g*/PCM(THF) Geometry Optimization"
    PBE_6_31g_d_PCM_THF__Frequency_Analysis = "PBE/6-31g*/PCM(THF) Frequency Analysis"
    PBE_6_31g_d_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_PCM_THF__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(THF) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_THF__Unknown = "PBE/6-31g*/PCM(THF) Unknown"
    PBE_6_31g_d_PCM_G2__Single_Point = "PBE/6-31g*/PCM(G2) Single Point"
    PBE_6_31g_d_PCM_G2__Force = "PBE/6-31g*/PCM(G2) Force"
    PBE_6_31g_d_PCM_G2__Geometry_Optimization = "PBE/6-31g*/PCM(G2) Geometry Optimization"
    PBE_6_31g_d_PCM_G2__Frequency_Analysis = "PBE/6-31g*/PCM(G2) Frequency Analysis"
    PBE_6_31g_d_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_PCM_G2__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(G2) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_G2__Unknown = "PBE/6-31g*/PCM(G2) Unknown"
    PBE_6_31g_d_PCM_DME__Single_Point = "PBE/6-31g*/PCM(DME) Single Point"
    PBE_6_31g_d_PCM_DME__Force = "PBE/6-31g*/PCM(DME) Force"
    PBE_6_31g_d_PCM_DME__Geometry_Optimization = "PBE/6-31g*/PCM(DME) Geometry Optimization"
    PBE_6_31g_d_PCM_DME__Frequency_Analysis = "PBE/6-31g*/PCM(DME) Frequency Analysis"
    PBE_6_31g_d_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_PCM_DME__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(DME) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_DME__Unknown = "PBE/6-31g*/PCM(DME) Unknown"
    PBE_6_31g_d_PCM_EC__Single_Point = "PBE/6-31g*/PCM(EC) Single Point"
    PBE_6_31g_d_PCM_EC__Force = "PBE/6-31g*/PCM(EC) Force"
    PBE_6_31g_d_PCM_EC__Geometry_Optimization = "PBE/6-31g*/PCM(EC) Geometry Optimization"
    PBE_6_31g_d_PCM_EC__Frequency_Analysis = "PBE/6-31g*/PCM(EC) Frequency Analysis"
    PBE_6_31g_d_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_PCM_EC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(EC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_EC__Unknown = "PBE/6-31g*/PCM(EC) Unknown"
    PBE_6_31g_d_PCM_EC_EMC__Single_Point = "PBE/6-31g*/PCM(EC/EMC) Single Point"
    PBE_6_31g_d_PCM_EC_EMC__Force = "PBE/6-31g*/PCM(EC/EMC) Force"
    PBE_6_31g_d_PCM_EC_EMC__Geometry_Optimization = "PBE/6-31g*/PCM(EC/EMC) Geometry Optimization"
    PBE_6_31g_d_PCM_EC_EMC__Frequency_Analysis = "PBE/6-31g*/PCM(EC/EMC) Frequency Analysis"
    PBE_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_EC_EMC__Unknown = "PBE/6-31g*/PCM(EC/EMC) Unknown"
    PBE_6_31g_d_PCM_PC__Single_Point = "PBE/6-31g*/PCM(PC) Single Point"
    PBE_6_31g_d_PCM_PC__Force = "PBE/6-31g*/PCM(PC) Force"
    PBE_6_31g_d_PCM_PC__Geometry_Optimization = "PBE/6-31g*/PCM(PC) Geometry Optimization"
    PBE_6_31g_d_PCM_PC__Frequency_Analysis = "PBE/6-31g*/PCM(PC) Frequency Analysis"
    PBE_6_31g_d_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_PCM_PC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(PC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_PCM_PC__Unknown = "PBE/6-31g*/PCM(PC) Unknown"
    PBE_6_31g_d_ISOSVP_WATER__Single_Point = "PBE/6-31g*/ISOSVP(WATER) Single Point"
    PBE_6_31g_d_ISOSVP_WATER__Force = "PBE/6-31g*/ISOSVP(WATER) Force"
    PBE_6_31g_d_ISOSVP_WATER__Geometry_Optimization = "PBE/6-31g*/ISOSVP(WATER) Geometry Optimization"
    PBE_6_31g_d_ISOSVP_WATER__Frequency_Analysis = "PBE/6-31g*/ISOSVP(WATER) Frequency Analysis"
    PBE_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_WATER__Unknown = "PBE/6-31g*/ISOSVP(WATER) Unknown"
    PBE_6_31g_d_ISOSVP_THF__Single_Point = "PBE/6-31g*/ISOSVP(THF) Single Point"
    PBE_6_31g_d_ISOSVP_THF__Force = "PBE/6-31g*/ISOSVP(THF) Force"
    PBE_6_31g_d_ISOSVP_THF__Geometry_Optimization = "PBE/6-31g*/ISOSVP(THF) Geometry Optimization"
    PBE_6_31g_d_ISOSVP_THF__Frequency_Analysis = "PBE/6-31g*/ISOSVP(THF) Frequency Analysis"
    PBE_6_31g_d_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(THF) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_THF__Unknown = "PBE/6-31g*/ISOSVP(THF) Unknown"
    PBE_6_31g_d_ISOSVP_G2__Single_Point = "PBE/6-31g*/ISOSVP(G2) Single Point"
    PBE_6_31g_d_ISOSVP_G2__Force = "PBE/6-31g*/ISOSVP(G2) Force"
    PBE_6_31g_d_ISOSVP_G2__Geometry_Optimization = "PBE/6-31g*/ISOSVP(G2) Geometry Optimization"
    PBE_6_31g_d_ISOSVP_G2__Frequency_Analysis = "PBE/6-31g*/ISOSVP(G2) Frequency Analysis"
    PBE_6_31g_d_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(G2) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_G2__Unknown = "PBE/6-31g*/ISOSVP(G2) Unknown"
    PBE_6_31g_d_ISOSVP_DME__Single_Point = "PBE/6-31g*/ISOSVP(DME) Single Point"
    PBE_6_31g_d_ISOSVP_DME__Force = "PBE/6-31g*/ISOSVP(DME) Force"
    PBE_6_31g_d_ISOSVP_DME__Geometry_Optimization = "PBE/6-31g*/ISOSVP(DME) Geometry Optimization"
    PBE_6_31g_d_ISOSVP_DME__Frequency_Analysis = "PBE/6-31g*/ISOSVP(DME) Frequency Analysis"
    PBE_6_31g_d_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(DME) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_DME__Unknown = "PBE/6-31g*/ISOSVP(DME) Unknown"
    PBE_6_31g_d_ISOSVP_EC__Single_Point = "PBE/6-31g*/ISOSVP(EC) Single Point"
    PBE_6_31g_d_ISOSVP_EC__Force = "PBE/6-31g*/ISOSVP(EC) Force"
    PBE_6_31g_d_ISOSVP_EC__Geometry_Optimization = "PBE/6-31g*/ISOSVP(EC) Geometry Optimization"
    PBE_6_31g_d_ISOSVP_EC__Frequency_Analysis = "PBE/6-31g*/ISOSVP(EC) Frequency Analysis"
    PBE_6_31g_d_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(EC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_EC__Unknown = "PBE/6-31g*/ISOSVP(EC) Unknown"
    PBE_6_31g_d_ISOSVP_EC_EMC__Single_Point = "PBE/6-31g*/ISOSVP(EC/EMC) Single Point"
    PBE_6_31g_d_ISOSVP_EC_EMC__Force = "PBE/6-31g*/ISOSVP(EC/EMC) Force"
    PBE_6_31g_d_ISOSVP_EC_EMC__Geometry_Optimization = "PBE/6-31g*/ISOSVP(EC/EMC) Geometry Optimization"
    PBE_6_31g_d_ISOSVP_EC_EMC__Frequency_Analysis = "PBE/6-31g*/ISOSVP(EC/EMC) Frequency Analysis"
    PBE_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_EC_EMC__Unknown = "PBE/6-31g*/ISOSVP(EC/EMC) Unknown"
    PBE_6_31g_d_ISOSVP_PC__Single_Point = "PBE/6-31g*/ISOSVP(PC) Single Point"
    PBE_6_31g_d_ISOSVP_PC__Force = "PBE/6-31g*/ISOSVP(PC) Force"
    PBE_6_31g_d_ISOSVP_PC__Geometry_Optimization = "PBE/6-31g*/ISOSVP(PC) Geometry Optimization"
    PBE_6_31g_d_ISOSVP_PC__Frequency_Analysis = "PBE/6-31g*/ISOSVP(PC) Frequency Analysis"
    PBE_6_31g_d_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(PC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_ISOSVP_PC__Unknown = "PBE/6-31g*/ISOSVP(PC) Unknown"
    PBE_6_31g_d_SMD_WATER__Single_Point = "PBE/6-31g*/SMD(WATER) Single Point"
    PBE_6_31g_d_SMD_WATER__Force = "PBE/6-31g*/SMD(WATER) Force"
    PBE_6_31g_d_SMD_WATER__Geometry_Optimization = "PBE/6-31g*/SMD(WATER) Geometry Optimization"
    PBE_6_31g_d_SMD_WATER__Frequency_Analysis = "PBE/6-31g*/SMD(WATER) Frequency Analysis"
    PBE_6_31g_d_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_SMD_WATER__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(WATER) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_WATER__Unknown = "PBE/6-31g*/SMD(WATER) Unknown"
    PBE_6_31g_d_SMD_THF__Single_Point = "PBE/6-31g*/SMD(THF) Single Point"
    PBE_6_31g_d_SMD_THF__Force = "PBE/6-31g*/SMD(THF) Force"
    PBE_6_31g_d_SMD_THF__Geometry_Optimization = "PBE/6-31g*/SMD(THF) Geometry Optimization"
    PBE_6_31g_d_SMD_THF__Frequency_Analysis = "PBE/6-31g*/SMD(THF) Frequency Analysis"
    PBE_6_31g_d_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_SMD_THF__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(THF) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_THF__Unknown = "PBE/6-31g*/SMD(THF) Unknown"
    PBE_6_31g_d_SMD_G2__Single_Point = "PBE/6-31g*/SMD(G2) Single Point"
    PBE_6_31g_d_SMD_G2__Force = "PBE/6-31g*/SMD(G2) Force"
    PBE_6_31g_d_SMD_G2__Geometry_Optimization = "PBE/6-31g*/SMD(G2) Geometry Optimization"
    PBE_6_31g_d_SMD_G2__Frequency_Analysis = "PBE/6-31g*/SMD(G2) Frequency Analysis"
    PBE_6_31g_d_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_SMD_G2__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(G2) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_G2__Unknown = "PBE/6-31g*/SMD(G2) Unknown"
    PBE_6_31g_d_SMD_DME__Single_Point = "PBE/6-31g*/SMD(DME) Single Point"
    PBE_6_31g_d_SMD_DME__Force = "PBE/6-31g*/SMD(DME) Force"
    PBE_6_31g_d_SMD_DME__Geometry_Optimization = "PBE/6-31g*/SMD(DME) Geometry Optimization"
    PBE_6_31g_d_SMD_DME__Frequency_Analysis = "PBE/6-31g*/SMD(DME) Frequency Analysis"
    PBE_6_31g_d_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_SMD_DME__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(DME) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_DME__Unknown = "PBE/6-31g*/SMD(DME) Unknown"
    PBE_6_31g_d_SMD_EC__Single_Point = "PBE/6-31g*/SMD(EC) Single Point"
    PBE_6_31g_d_SMD_EC__Force = "PBE/6-31g*/SMD(EC) Force"
    PBE_6_31g_d_SMD_EC__Geometry_Optimization = "PBE/6-31g*/SMD(EC) Geometry Optimization"
    PBE_6_31g_d_SMD_EC__Frequency_Analysis = "PBE/6-31g*/SMD(EC) Frequency Analysis"
    PBE_6_31g_d_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_SMD_EC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(EC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_EC__Unknown = "PBE/6-31g*/SMD(EC) Unknown"
    PBE_6_31g_d_SMD_EC_EMC__Single_Point = "PBE/6-31g*/SMD(EC/EMC) Single Point"
    PBE_6_31g_d_SMD_EC_EMC__Force = "PBE/6-31g*/SMD(EC/EMC) Force"
    PBE_6_31g_d_SMD_EC_EMC__Geometry_Optimization = "PBE/6-31g*/SMD(EC/EMC) Geometry Optimization"
    PBE_6_31g_d_SMD_EC_EMC__Frequency_Analysis = "PBE/6-31g*/SMD(EC/EMC) Frequency Analysis"
    PBE_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_EC_EMC__Unknown = "PBE/6-31g*/SMD(EC/EMC) Unknown"
    PBE_6_31g_d_SMD_PC__Single_Point = "PBE/6-31g*/SMD(PC) Single Point"
    PBE_6_31g_d_SMD_PC__Force = "PBE/6-31g*/SMD(PC) Force"
    PBE_6_31g_d_SMD_PC__Geometry_Optimization = "PBE/6-31g*/SMD(PC) Geometry Optimization"
    PBE_6_31g_d_SMD_PC__Frequency_Analysis = "PBE/6-31g*/SMD(PC) Frequency Analysis"
    PBE_6_31g_d_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_SMD_PC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(PC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_SMD_PC__Unknown = "PBE/6-31g*/SMD(PC) Unknown"
    PBE_6_31g_d_CMIRS_WATER__Single_Point = "PBE/6-31g*/CMIRS(WATER) Single Point"
    PBE_6_31g_d_CMIRS_WATER__Force = "PBE/6-31g*/CMIRS(WATER) Force"
    PBE_6_31g_d_CMIRS_WATER__Geometry_Optimization = "PBE/6-31g*/CMIRS(WATER) Geometry Optimization"
    PBE_6_31g_d_CMIRS_WATER__Frequency_Analysis = "PBE/6-31g*/CMIRS(WATER) Frequency Analysis"
    PBE_6_31g_d_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(WATER) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_WATER__Unknown = "PBE/6-31g*/CMIRS(WATER) Unknown"
    PBE_6_31g_d_CMIRS_THF__Single_Point = "PBE/6-31g*/CMIRS(THF) Single Point"
    PBE_6_31g_d_CMIRS_THF__Force = "PBE/6-31g*/CMIRS(THF) Force"
    PBE_6_31g_d_CMIRS_THF__Geometry_Optimization = "PBE/6-31g*/CMIRS(THF) Geometry Optimization"
    PBE_6_31g_d_CMIRS_THF__Frequency_Analysis = "PBE/6-31g*/CMIRS(THF) Frequency Analysis"
    PBE_6_31g_d_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(THF) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_THF__Unknown = "PBE/6-31g*/CMIRS(THF) Unknown"
    PBE_6_31g_d_CMIRS_G2__Single_Point = "PBE/6-31g*/CMIRS(G2) Single Point"
    PBE_6_31g_d_CMIRS_G2__Force = "PBE/6-31g*/CMIRS(G2) Force"
    PBE_6_31g_d_CMIRS_G2__Geometry_Optimization = "PBE/6-31g*/CMIRS(G2) Geometry Optimization"
    PBE_6_31g_d_CMIRS_G2__Frequency_Analysis = "PBE/6-31g*/CMIRS(G2) Frequency Analysis"
    PBE_6_31g_d_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(G2) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_G2__Unknown = "PBE/6-31g*/CMIRS(G2) Unknown"
    PBE_6_31g_d_CMIRS_DME__Single_Point = "PBE/6-31g*/CMIRS(DME) Single Point"
    PBE_6_31g_d_CMIRS_DME__Force = "PBE/6-31g*/CMIRS(DME) Force"
    PBE_6_31g_d_CMIRS_DME__Geometry_Optimization = "PBE/6-31g*/CMIRS(DME) Geometry Optimization"
    PBE_6_31g_d_CMIRS_DME__Frequency_Analysis = "PBE/6-31g*/CMIRS(DME) Frequency Analysis"
    PBE_6_31g_d_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(DME) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_DME__Unknown = "PBE/6-31g*/CMIRS(DME) Unknown"
    PBE_6_31g_d_CMIRS_EC__Single_Point = "PBE/6-31g*/CMIRS(EC) Single Point"
    PBE_6_31g_d_CMIRS_EC__Force = "PBE/6-31g*/CMIRS(EC) Force"
    PBE_6_31g_d_CMIRS_EC__Geometry_Optimization = "PBE/6-31g*/CMIRS(EC) Geometry Optimization"
    PBE_6_31g_d_CMIRS_EC__Frequency_Analysis = "PBE/6-31g*/CMIRS(EC) Frequency Analysis"
    PBE_6_31g_d_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(EC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_EC__Unknown = "PBE/6-31g*/CMIRS(EC) Unknown"
    PBE_6_31g_d_CMIRS_EC_EMC__Single_Point = "PBE/6-31g*/CMIRS(EC/EMC) Single Point"
    PBE_6_31g_d_CMIRS_EC_EMC__Force = "PBE/6-31g*/CMIRS(EC/EMC) Force"
    PBE_6_31g_d_CMIRS_EC_EMC__Geometry_Optimization = "PBE/6-31g*/CMIRS(EC/EMC) Geometry Optimization"
    PBE_6_31g_d_CMIRS_EC_EMC__Frequency_Analysis = "PBE/6-31g*/CMIRS(EC/EMC) Frequency Analysis"
    PBE_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_EC_EMC__Unknown = "PBE/6-31g*/CMIRS(EC/EMC) Unknown"
    PBE_6_31g_d_CMIRS_PC__Single_Point = "PBE/6-31g*/CMIRS(PC) Single Point"
    PBE_6_31g_d_CMIRS_PC__Force = "PBE/6-31g*/CMIRS(PC) Force"
    PBE_6_31g_d_CMIRS_PC__Geometry_Optimization = "PBE/6-31g*/CMIRS(PC) Geometry Optimization"
    PBE_6_31g_d_CMIRS_PC__Frequency_Analysis = "PBE/6-31g*/CMIRS(PC) Frequency Analysis"
    PBE_6_31g_d_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(PC) Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/6-31g*/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_6_31g_d_CMIRS_PC__Unknown = "PBE/6-31g*/CMIRS(PC) Unknown"
    PBE_def2_SVPD_VACUUM_Single_Point = "PBE/def2-SVPD/VACUUM Single Point"
    PBE_def2_SVPD_VACUUM_Force = "PBE/def2-SVPD/VACUUM Force"
    PBE_def2_SVPD_VACUUM_Geometry_Optimization = "PBE/def2-SVPD/VACUUM Geometry Optimization"
    PBE_def2_SVPD_VACUUM_Frequency_Analysis = "PBE/def2-SVPD/VACUUM Frequency Analysis"
    PBE_def2_SVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/VACUUM Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_VACUUM_Unknown = "PBE/def2-SVPD/VACUUM Unknown"
    PBE_def2_SVPD_PCM_WATER__Single_Point = "PBE/def2-SVPD/PCM(WATER) Single Point"
    PBE_def2_SVPD_PCM_WATER__Force = "PBE/def2-SVPD/PCM(WATER) Force"
    PBE_def2_SVPD_PCM_WATER__Geometry_Optimization = "PBE/def2-SVPD/PCM(WATER) Geometry Optimization"
    PBE_def2_SVPD_PCM_WATER__Frequency_Analysis = "PBE/def2-SVPD/PCM(WATER) Frequency Analysis"
    PBE_def2_SVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_WATER__Unknown = "PBE/def2-SVPD/PCM(WATER) Unknown"
    PBE_def2_SVPD_PCM_THF__Single_Point = "PBE/def2-SVPD/PCM(THF) Single Point"
    PBE_def2_SVPD_PCM_THF__Force = "PBE/def2-SVPD/PCM(THF) Force"
    PBE_def2_SVPD_PCM_THF__Geometry_Optimization = "PBE/def2-SVPD/PCM(THF) Geometry Optimization"
    PBE_def2_SVPD_PCM_THF__Frequency_Analysis = "PBE/def2-SVPD/PCM(THF) Frequency Analysis"
    PBE_def2_SVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(THF) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_THF__Unknown = "PBE/def2-SVPD/PCM(THF) Unknown"
    PBE_def2_SVPD_PCM_G2__Single_Point = "PBE/def2-SVPD/PCM(G2) Single Point"
    PBE_def2_SVPD_PCM_G2__Force = "PBE/def2-SVPD/PCM(G2) Force"
    PBE_def2_SVPD_PCM_G2__Geometry_Optimization = "PBE/def2-SVPD/PCM(G2) Geometry Optimization"
    PBE_def2_SVPD_PCM_G2__Frequency_Analysis = "PBE/def2-SVPD/PCM(G2) Frequency Analysis"
    PBE_def2_SVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(G2) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_G2__Unknown = "PBE/def2-SVPD/PCM(G2) Unknown"
    PBE_def2_SVPD_PCM_DME__Single_Point = "PBE/def2-SVPD/PCM(DME) Single Point"
    PBE_def2_SVPD_PCM_DME__Force = "PBE/def2-SVPD/PCM(DME) Force"
    PBE_def2_SVPD_PCM_DME__Geometry_Optimization = "PBE/def2-SVPD/PCM(DME) Geometry Optimization"
    PBE_def2_SVPD_PCM_DME__Frequency_Analysis = "PBE/def2-SVPD/PCM(DME) Frequency Analysis"
    PBE_def2_SVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(DME) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_DME__Unknown = "PBE/def2-SVPD/PCM(DME) Unknown"
    PBE_def2_SVPD_PCM_EC__Single_Point = "PBE/def2-SVPD/PCM(EC) Single Point"
    PBE_def2_SVPD_PCM_EC__Force = "PBE/def2-SVPD/PCM(EC) Force"
    PBE_def2_SVPD_PCM_EC__Geometry_Optimization = "PBE/def2-SVPD/PCM(EC) Geometry Optimization"
    PBE_def2_SVPD_PCM_EC__Frequency_Analysis = "PBE/def2-SVPD/PCM(EC) Frequency Analysis"
    PBE_def2_SVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(EC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_EC__Unknown = "PBE/def2-SVPD/PCM(EC) Unknown"
    PBE_def2_SVPD_PCM_EC_EMC__Single_Point = "PBE/def2-SVPD/PCM(EC/EMC) Single Point"
    PBE_def2_SVPD_PCM_EC_EMC__Force = "PBE/def2-SVPD/PCM(EC/EMC) Force"
    PBE_def2_SVPD_PCM_EC_EMC__Geometry_Optimization = "PBE/def2-SVPD/PCM(EC/EMC) Geometry Optimization"
    PBE_def2_SVPD_PCM_EC_EMC__Frequency_Analysis = "PBE/def2-SVPD/PCM(EC/EMC) Frequency Analysis"
    PBE_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_EC_EMC__Unknown = "PBE/def2-SVPD/PCM(EC/EMC) Unknown"
    PBE_def2_SVPD_PCM_PC__Single_Point = "PBE/def2-SVPD/PCM(PC) Single Point"
    PBE_def2_SVPD_PCM_PC__Force = "PBE/def2-SVPD/PCM(PC) Force"
    PBE_def2_SVPD_PCM_PC__Geometry_Optimization = "PBE/def2-SVPD/PCM(PC) Geometry Optimization"
    PBE_def2_SVPD_PCM_PC__Frequency_Analysis = "PBE/def2-SVPD/PCM(PC) Frequency Analysis"
    PBE_def2_SVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(PC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_PCM_PC__Unknown = "PBE/def2-SVPD/PCM(PC) Unknown"
    PBE_def2_SVPD_ISOSVP_WATER__Single_Point = "PBE/def2-SVPD/ISOSVP(WATER) Single Point"
    PBE_def2_SVPD_ISOSVP_WATER__Force = "PBE/def2-SVPD/ISOSVP(WATER) Force"
    PBE_def2_SVPD_ISOSVP_WATER__Geometry_Optimization = "PBE/def2-SVPD/ISOSVP(WATER) Geometry Optimization"
    PBE_def2_SVPD_ISOSVP_WATER__Frequency_Analysis = "PBE/def2-SVPD/ISOSVP(WATER) Frequency Analysis"
    PBE_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_WATER__Unknown = "PBE/def2-SVPD/ISOSVP(WATER) Unknown"
    PBE_def2_SVPD_ISOSVP_THF__Single_Point = "PBE/def2-SVPD/ISOSVP(THF) Single Point"
    PBE_def2_SVPD_ISOSVP_THF__Force = "PBE/def2-SVPD/ISOSVP(THF) Force"
    PBE_def2_SVPD_ISOSVP_THF__Geometry_Optimization = "PBE/def2-SVPD/ISOSVP(THF) Geometry Optimization"
    PBE_def2_SVPD_ISOSVP_THF__Frequency_Analysis = "PBE/def2-SVPD/ISOSVP(THF) Frequency Analysis"
    PBE_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_THF__Unknown = "PBE/def2-SVPD/ISOSVP(THF) Unknown"
    PBE_def2_SVPD_ISOSVP_G2__Single_Point = "PBE/def2-SVPD/ISOSVP(G2) Single Point"
    PBE_def2_SVPD_ISOSVP_G2__Force = "PBE/def2-SVPD/ISOSVP(G2) Force"
    PBE_def2_SVPD_ISOSVP_G2__Geometry_Optimization = "PBE/def2-SVPD/ISOSVP(G2) Geometry Optimization"
    PBE_def2_SVPD_ISOSVP_G2__Frequency_Analysis = "PBE/def2-SVPD/ISOSVP(G2) Frequency Analysis"
    PBE_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_G2__Unknown = "PBE/def2-SVPD/ISOSVP(G2) Unknown"
    PBE_def2_SVPD_ISOSVP_DME__Single_Point = "PBE/def2-SVPD/ISOSVP(DME) Single Point"
    PBE_def2_SVPD_ISOSVP_DME__Force = "PBE/def2-SVPD/ISOSVP(DME) Force"
    PBE_def2_SVPD_ISOSVP_DME__Geometry_Optimization = "PBE/def2-SVPD/ISOSVP(DME) Geometry Optimization"
    PBE_def2_SVPD_ISOSVP_DME__Frequency_Analysis = "PBE/def2-SVPD/ISOSVP(DME) Frequency Analysis"
    PBE_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_DME__Unknown = "PBE/def2-SVPD/ISOSVP(DME) Unknown"
    PBE_def2_SVPD_ISOSVP_EC__Single_Point = "PBE/def2-SVPD/ISOSVP(EC) Single Point"
    PBE_def2_SVPD_ISOSVP_EC__Force = "PBE/def2-SVPD/ISOSVP(EC) Force"
    PBE_def2_SVPD_ISOSVP_EC__Geometry_Optimization = "PBE/def2-SVPD/ISOSVP(EC) Geometry Optimization"
    PBE_def2_SVPD_ISOSVP_EC__Frequency_Analysis = "PBE/def2-SVPD/ISOSVP(EC) Frequency Analysis"
    PBE_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_EC__Unknown = "PBE/def2-SVPD/ISOSVP(EC) Unknown"
    PBE_def2_SVPD_ISOSVP_EC_EMC__Single_Point = "PBE/def2-SVPD/ISOSVP(EC/EMC) Single Point"
    PBE_def2_SVPD_ISOSVP_EC_EMC__Force = "PBE/def2-SVPD/ISOSVP(EC/EMC) Force"
    PBE_def2_SVPD_ISOSVP_EC_EMC__Geometry_Optimization = "PBE/def2-SVPD/ISOSVP(EC/EMC) Geometry Optimization"
    PBE_def2_SVPD_ISOSVP_EC_EMC__Frequency_Analysis = "PBE/def2-SVPD/ISOSVP(EC/EMC) Frequency Analysis"
    PBE_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_EC_EMC__Unknown = "PBE/def2-SVPD/ISOSVP(EC/EMC) Unknown"
    PBE_def2_SVPD_ISOSVP_PC__Single_Point = "PBE/def2-SVPD/ISOSVP(PC) Single Point"
    PBE_def2_SVPD_ISOSVP_PC__Force = "PBE/def2-SVPD/ISOSVP(PC) Force"
    PBE_def2_SVPD_ISOSVP_PC__Geometry_Optimization = "PBE/def2-SVPD/ISOSVP(PC) Geometry Optimization"
    PBE_def2_SVPD_ISOSVP_PC__Frequency_Analysis = "PBE/def2-SVPD/ISOSVP(PC) Frequency Analysis"
    PBE_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_ISOSVP_PC__Unknown = "PBE/def2-SVPD/ISOSVP(PC) Unknown"
    PBE_def2_SVPD_SMD_WATER__Single_Point = "PBE/def2-SVPD/SMD(WATER) Single Point"
    PBE_def2_SVPD_SMD_WATER__Force = "PBE/def2-SVPD/SMD(WATER) Force"
    PBE_def2_SVPD_SMD_WATER__Geometry_Optimization = "PBE/def2-SVPD/SMD(WATER) Geometry Optimization"
    PBE_def2_SVPD_SMD_WATER__Frequency_Analysis = "PBE/def2-SVPD/SMD(WATER) Frequency Analysis"
    PBE_def2_SVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_WATER__Unknown = "PBE/def2-SVPD/SMD(WATER) Unknown"
    PBE_def2_SVPD_SMD_THF__Single_Point = "PBE/def2-SVPD/SMD(THF) Single Point"
    PBE_def2_SVPD_SMD_THF__Force = "PBE/def2-SVPD/SMD(THF) Force"
    PBE_def2_SVPD_SMD_THF__Geometry_Optimization = "PBE/def2-SVPD/SMD(THF) Geometry Optimization"
    PBE_def2_SVPD_SMD_THF__Frequency_Analysis = "PBE/def2-SVPD/SMD(THF) Frequency Analysis"
    PBE_def2_SVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(THF) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_THF__Unknown = "PBE/def2-SVPD/SMD(THF) Unknown"
    PBE_def2_SVPD_SMD_G2__Single_Point = "PBE/def2-SVPD/SMD(G2) Single Point"
    PBE_def2_SVPD_SMD_G2__Force = "PBE/def2-SVPD/SMD(G2) Force"
    PBE_def2_SVPD_SMD_G2__Geometry_Optimization = "PBE/def2-SVPD/SMD(G2) Geometry Optimization"
    PBE_def2_SVPD_SMD_G2__Frequency_Analysis = "PBE/def2-SVPD/SMD(G2) Frequency Analysis"
    PBE_def2_SVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(G2) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_G2__Unknown = "PBE/def2-SVPD/SMD(G2) Unknown"
    PBE_def2_SVPD_SMD_DME__Single_Point = "PBE/def2-SVPD/SMD(DME) Single Point"
    PBE_def2_SVPD_SMD_DME__Force = "PBE/def2-SVPD/SMD(DME) Force"
    PBE_def2_SVPD_SMD_DME__Geometry_Optimization = "PBE/def2-SVPD/SMD(DME) Geometry Optimization"
    PBE_def2_SVPD_SMD_DME__Frequency_Analysis = "PBE/def2-SVPD/SMD(DME) Frequency Analysis"
    PBE_def2_SVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(DME) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_DME__Unknown = "PBE/def2-SVPD/SMD(DME) Unknown"
    PBE_def2_SVPD_SMD_EC__Single_Point = "PBE/def2-SVPD/SMD(EC) Single Point"
    PBE_def2_SVPD_SMD_EC__Force = "PBE/def2-SVPD/SMD(EC) Force"
    PBE_def2_SVPD_SMD_EC__Geometry_Optimization = "PBE/def2-SVPD/SMD(EC) Geometry Optimization"
    PBE_def2_SVPD_SMD_EC__Frequency_Analysis = "PBE/def2-SVPD/SMD(EC) Frequency Analysis"
    PBE_def2_SVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(EC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_EC__Unknown = "PBE/def2-SVPD/SMD(EC) Unknown"
    PBE_def2_SVPD_SMD_EC_EMC__Single_Point = "PBE/def2-SVPD/SMD(EC/EMC) Single Point"
    PBE_def2_SVPD_SMD_EC_EMC__Force = "PBE/def2-SVPD/SMD(EC/EMC) Force"
    PBE_def2_SVPD_SMD_EC_EMC__Geometry_Optimization = "PBE/def2-SVPD/SMD(EC/EMC) Geometry Optimization"
    PBE_def2_SVPD_SMD_EC_EMC__Frequency_Analysis = "PBE/def2-SVPD/SMD(EC/EMC) Frequency Analysis"
    PBE_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_EC_EMC__Unknown = "PBE/def2-SVPD/SMD(EC/EMC) Unknown"
    PBE_def2_SVPD_SMD_PC__Single_Point = "PBE/def2-SVPD/SMD(PC) Single Point"
    PBE_def2_SVPD_SMD_PC__Force = "PBE/def2-SVPD/SMD(PC) Force"
    PBE_def2_SVPD_SMD_PC__Geometry_Optimization = "PBE/def2-SVPD/SMD(PC) Geometry Optimization"
    PBE_def2_SVPD_SMD_PC__Frequency_Analysis = "PBE/def2-SVPD/SMD(PC) Frequency Analysis"
    PBE_def2_SVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(PC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_SMD_PC__Unknown = "PBE/def2-SVPD/SMD(PC) Unknown"
    PBE_def2_SVPD_CMIRS_WATER__Single_Point = "PBE/def2-SVPD/CMIRS(WATER) Single Point"
    PBE_def2_SVPD_CMIRS_WATER__Force = "PBE/def2-SVPD/CMIRS(WATER) Force"
    PBE_def2_SVPD_CMIRS_WATER__Geometry_Optimization = "PBE/def2-SVPD/CMIRS(WATER) Geometry Optimization"
    PBE_def2_SVPD_CMIRS_WATER__Frequency_Analysis = "PBE/def2-SVPD/CMIRS(WATER) Frequency Analysis"
    PBE_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_WATER__Unknown = "PBE/def2-SVPD/CMIRS(WATER) Unknown"
    PBE_def2_SVPD_CMIRS_THF__Single_Point = "PBE/def2-SVPD/CMIRS(THF) Single Point"
    PBE_def2_SVPD_CMIRS_THF__Force = "PBE/def2-SVPD/CMIRS(THF) Force"
    PBE_def2_SVPD_CMIRS_THF__Geometry_Optimization = "PBE/def2-SVPD/CMIRS(THF) Geometry Optimization"
    PBE_def2_SVPD_CMIRS_THF__Frequency_Analysis = "PBE/def2-SVPD/CMIRS(THF) Frequency Analysis"
    PBE_def2_SVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_THF__Unknown = "PBE/def2-SVPD/CMIRS(THF) Unknown"
    PBE_def2_SVPD_CMIRS_G2__Single_Point = "PBE/def2-SVPD/CMIRS(G2) Single Point"
    PBE_def2_SVPD_CMIRS_G2__Force = "PBE/def2-SVPD/CMIRS(G2) Force"
    PBE_def2_SVPD_CMIRS_G2__Geometry_Optimization = "PBE/def2-SVPD/CMIRS(G2) Geometry Optimization"
    PBE_def2_SVPD_CMIRS_G2__Frequency_Analysis = "PBE/def2-SVPD/CMIRS(G2) Frequency Analysis"
    PBE_def2_SVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_G2__Unknown = "PBE/def2-SVPD/CMIRS(G2) Unknown"
    PBE_def2_SVPD_CMIRS_DME__Single_Point = "PBE/def2-SVPD/CMIRS(DME) Single Point"
    PBE_def2_SVPD_CMIRS_DME__Force = "PBE/def2-SVPD/CMIRS(DME) Force"
    PBE_def2_SVPD_CMIRS_DME__Geometry_Optimization = "PBE/def2-SVPD/CMIRS(DME) Geometry Optimization"
    PBE_def2_SVPD_CMIRS_DME__Frequency_Analysis = "PBE/def2-SVPD/CMIRS(DME) Frequency Analysis"
    PBE_def2_SVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_DME__Unknown = "PBE/def2-SVPD/CMIRS(DME) Unknown"
    PBE_def2_SVPD_CMIRS_EC__Single_Point = "PBE/def2-SVPD/CMIRS(EC) Single Point"
    PBE_def2_SVPD_CMIRS_EC__Force = "PBE/def2-SVPD/CMIRS(EC) Force"
    PBE_def2_SVPD_CMIRS_EC__Geometry_Optimization = "PBE/def2-SVPD/CMIRS(EC) Geometry Optimization"
    PBE_def2_SVPD_CMIRS_EC__Frequency_Analysis = "PBE/def2-SVPD/CMIRS(EC) Frequency Analysis"
    PBE_def2_SVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_EC__Unknown = "PBE/def2-SVPD/CMIRS(EC) Unknown"
    PBE_def2_SVPD_CMIRS_EC_EMC__Single_Point = "PBE/def2-SVPD/CMIRS(EC/EMC) Single Point"
    PBE_def2_SVPD_CMIRS_EC_EMC__Force = "PBE/def2-SVPD/CMIRS(EC/EMC) Force"
    PBE_def2_SVPD_CMIRS_EC_EMC__Geometry_Optimization = "PBE/def2-SVPD/CMIRS(EC/EMC) Geometry Optimization"
    PBE_def2_SVPD_CMIRS_EC_EMC__Frequency_Analysis = "PBE/def2-SVPD/CMIRS(EC/EMC) Frequency Analysis"
    PBE_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_EC_EMC__Unknown = "PBE/def2-SVPD/CMIRS(EC/EMC) Unknown"
    PBE_def2_SVPD_CMIRS_PC__Single_Point = "PBE/def2-SVPD/CMIRS(PC) Single Point"
    PBE_def2_SVPD_CMIRS_PC__Force = "PBE/def2-SVPD/CMIRS(PC) Force"
    PBE_def2_SVPD_CMIRS_PC__Geometry_Optimization = "PBE/def2-SVPD/CMIRS(PC) Geometry Optimization"
    PBE_def2_SVPD_CMIRS_PC__Frequency_Analysis = "PBE/def2-SVPD/CMIRS(PC) Frequency Analysis"
    PBE_def2_SVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-SVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_SVPD_CMIRS_PC__Unknown = "PBE/def2-SVPD/CMIRS(PC) Unknown"
    PBE_def2_TZVP_VACUUM_Single_Point = "PBE/def2-TZVP/VACUUM Single Point"
    PBE_def2_TZVP_VACUUM_Force = "PBE/def2-TZVP/VACUUM Force"
    PBE_def2_TZVP_VACUUM_Geometry_Optimization = "PBE/def2-TZVP/VACUUM Geometry Optimization"
    PBE_def2_TZVP_VACUUM_Frequency_Analysis = "PBE/def2-TZVP/VACUUM Frequency Analysis"
    PBE_def2_TZVP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/VACUUM Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_VACUUM_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/VACUUM Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_VACUUM_Unknown = "PBE/def2-TZVP/VACUUM Unknown"
    PBE_def2_TZVP_PCM_WATER__Single_Point = "PBE/def2-TZVP/PCM(WATER) Single Point"
    PBE_def2_TZVP_PCM_WATER__Force = "PBE/def2-TZVP/PCM(WATER) Force"
    PBE_def2_TZVP_PCM_WATER__Geometry_Optimization = "PBE/def2-TZVP/PCM(WATER) Geometry Optimization"
    PBE_def2_TZVP_PCM_WATER__Frequency_Analysis = "PBE/def2-TZVP/PCM(WATER) Frequency Analysis"
    PBE_def2_TZVP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_WATER__Unknown = "PBE/def2-TZVP/PCM(WATER) Unknown"
    PBE_def2_TZVP_PCM_THF__Single_Point = "PBE/def2-TZVP/PCM(THF) Single Point"
    PBE_def2_TZVP_PCM_THF__Force = "PBE/def2-TZVP/PCM(THF) Force"
    PBE_def2_TZVP_PCM_THF__Geometry_Optimization = "PBE/def2-TZVP/PCM(THF) Geometry Optimization"
    PBE_def2_TZVP_PCM_THF__Frequency_Analysis = "PBE/def2-TZVP/PCM(THF) Frequency Analysis"
    PBE_def2_TZVP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_THF__Unknown = "PBE/def2-TZVP/PCM(THF) Unknown"
    PBE_def2_TZVP_PCM_G2__Single_Point = "PBE/def2-TZVP/PCM(G2) Single Point"
    PBE_def2_TZVP_PCM_G2__Force = "PBE/def2-TZVP/PCM(G2) Force"
    PBE_def2_TZVP_PCM_G2__Geometry_Optimization = "PBE/def2-TZVP/PCM(G2) Geometry Optimization"
    PBE_def2_TZVP_PCM_G2__Frequency_Analysis = "PBE/def2-TZVP/PCM(G2) Frequency Analysis"
    PBE_def2_TZVP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_G2__Unknown = "PBE/def2-TZVP/PCM(G2) Unknown"
    PBE_def2_TZVP_PCM_DME__Single_Point = "PBE/def2-TZVP/PCM(DME) Single Point"
    PBE_def2_TZVP_PCM_DME__Force = "PBE/def2-TZVP/PCM(DME) Force"
    PBE_def2_TZVP_PCM_DME__Geometry_Optimization = "PBE/def2-TZVP/PCM(DME) Geometry Optimization"
    PBE_def2_TZVP_PCM_DME__Frequency_Analysis = "PBE/def2-TZVP/PCM(DME) Frequency Analysis"
    PBE_def2_TZVP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_DME__Unknown = "PBE/def2-TZVP/PCM(DME) Unknown"
    PBE_def2_TZVP_PCM_EC__Single_Point = "PBE/def2-TZVP/PCM(EC) Single Point"
    PBE_def2_TZVP_PCM_EC__Force = "PBE/def2-TZVP/PCM(EC) Force"
    PBE_def2_TZVP_PCM_EC__Geometry_Optimization = "PBE/def2-TZVP/PCM(EC) Geometry Optimization"
    PBE_def2_TZVP_PCM_EC__Frequency_Analysis = "PBE/def2-TZVP/PCM(EC) Frequency Analysis"
    PBE_def2_TZVP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_EC__Unknown = "PBE/def2-TZVP/PCM(EC) Unknown"
    PBE_def2_TZVP_PCM_EC_EMC__Single_Point = "PBE/def2-TZVP/PCM(EC/EMC) Single Point"
    PBE_def2_TZVP_PCM_EC_EMC__Force = "PBE/def2-TZVP/PCM(EC/EMC) Force"
    PBE_def2_TZVP_PCM_EC_EMC__Geometry_Optimization = "PBE/def2-TZVP/PCM(EC/EMC) Geometry Optimization"
    PBE_def2_TZVP_PCM_EC_EMC__Frequency_Analysis = "PBE/def2-TZVP/PCM(EC/EMC) Frequency Analysis"
    PBE_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_EC_EMC__Unknown = "PBE/def2-TZVP/PCM(EC/EMC) Unknown"
    PBE_def2_TZVP_PCM_PC__Single_Point = "PBE/def2-TZVP/PCM(PC) Single Point"
    PBE_def2_TZVP_PCM_PC__Force = "PBE/def2-TZVP/PCM(PC) Force"
    PBE_def2_TZVP_PCM_PC__Geometry_Optimization = "PBE/def2-TZVP/PCM(PC) Geometry Optimization"
    PBE_def2_TZVP_PCM_PC__Frequency_Analysis = "PBE/def2-TZVP/PCM(PC) Frequency Analysis"
    PBE_def2_TZVP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_PCM_PC__Unknown = "PBE/def2-TZVP/PCM(PC) Unknown"
    PBE_def2_TZVP_ISOSVP_WATER__Single_Point = "PBE/def2-TZVP/ISOSVP(WATER) Single Point"
    PBE_def2_TZVP_ISOSVP_WATER__Force = "PBE/def2-TZVP/ISOSVP(WATER) Force"
    PBE_def2_TZVP_ISOSVP_WATER__Geometry_Optimization = "PBE/def2-TZVP/ISOSVP(WATER) Geometry Optimization"
    PBE_def2_TZVP_ISOSVP_WATER__Frequency_Analysis = "PBE/def2-TZVP/ISOSVP(WATER) Frequency Analysis"
    PBE_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_WATER__Unknown = "PBE/def2-TZVP/ISOSVP(WATER) Unknown"
    PBE_def2_TZVP_ISOSVP_THF__Single_Point = "PBE/def2-TZVP/ISOSVP(THF) Single Point"
    PBE_def2_TZVP_ISOSVP_THF__Force = "PBE/def2-TZVP/ISOSVP(THF) Force"
    PBE_def2_TZVP_ISOSVP_THF__Geometry_Optimization = "PBE/def2-TZVP/ISOSVP(THF) Geometry Optimization"
    PBE_def2_TZVP_ISOSVP_THF__Frequency_Analysis = "PBE/def2-TZVP/ISOSVP(THF) Frequency Analysis"
    PBE_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_THF__Unknown = "PBE/def2-TZVP/ISOSVP(THF) Unknown"
    PBE_def2_TZVP_ISOSVP_G2__Single_Point = "PBE/def2-TZVP/ISOSVP(G2) Single Point"
    PBE_def2_TZVP_ISOSVP_G2__Force = "PBE/def2-TZVP/ISOSVP(G2) Force"
    PBE_def2_TZVP_ISOSVP_G2__Geometry_Optimization = "PBE/def2-TZVP/ISOSVP(G2) Geometry Optimization"
    PBE_def2_TZVP_ISOSVP_G2__Frequency_Analysis = "PBE/def2-TZVP/ISOSVP(G2) Frequency Analysis"
    PBE_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_G2__Unknown = "PBE/def2-TZVP/ISOSVP(G2) Unknown"
    PBE_def2_TZVP_ISOSVP_DME__Single_Point = "PBE/def2-TZVP/ISOSVP(DME) Single Point"
    PBE_def2_TZVP_ISOSVP_DME__Force = "PBE/def2-TZVP/ISOSVP(DME) Force"
    PBE_def2_TZVP_ISOSVP_DME__Geometry_Optimization = "PBE/def2-TZVP/ISOSVP(DME) Geometry Optimization"
    PBE_def2_TZVP_ISOSVP_DME__Frequency_Analysis = "PBE/def2-TZVP/ISOSVP(DME) Frequency Analysis"
    PBE_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_DME__Unknown = "PBE/def2-TZVP/ISOSVP(DME) Unknown"
    PBE_def2_TZVP_ISOSVP_EC__Single_Point = "PBE/def2-TZVP/ISOSVP(EC) Single Point"
    PBE_def2_TZVP_ISOSVP_EC__Force = "PBE/def2-TZVP/ISOSVP(EC) Force"
    PBE_def2_TZVP_ISOSVP_EC__Geometry_Optimization = "PBE/def2-TZVP/ISOSVP(EC) Geometry Optimization"
    PBE_def2_TZVP_ISOSVP_EC__Frequency_Analysis = "PBE/def2-TZVP/ISOSVP(EC) Frequency Analysis"
    PBE_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_EC__Unknown = "PBE/def2-TZVP/ISOSVP(EC) Unknown"
    PBE_def2_TZVP_ISOSVP_EC_EMC__Single_Point = "PBE/def2-TZVP/ISOSVP(EC/EMC) Single Point"
    PBE_def2_TZVP_ISOSVP_EC_EMC__Force = "PBE/def2-TZVP/ISOSVP(EC/EMC) Force"
    PBE_def2_TZVP_ISOSVP_EC_EMC__Geometry_Optimization = "PBE/def2-TZVP/ISOSVP(EC/EMC) Geometry Optimization"
    PBE_def2_TZVP_ISOSVP_EC_EMC__Frequency_Analysis = "PBE/def2-TZVP/ISOSVP(EC/EMC) Frequency Analysis"
    PBE_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_EC_EMC__Unknown = "PBE/def2-TZVP/ISOSVP(EC/EMC) Unknown"
    PBE_def2_TZVP_ISOSVP_PC__Single_Point = "PBE/def2-TZVP/ISOSVP(PC) Single Point"
    PBE_def2_TZVP_ISOSVP_PC__Force = "PBE/def2-TZVP/ISOSVP(PC) Force"
    PBE_def2_TZVP_ISOSVP_PC__Geometry_Optimization = "PBE/def2-TZVP/ISOSVP(PC) Geometry Optimization"
    PBE_def2_TZVP_ISOSVP_PC__Frequency_Analysis = "PBE/def2-TZVP/ISOSVP(PC) Frequency Analysis"
    PBE_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_ISOSVP_PC__Unknown = "PBE/def2-TZVP/ISOSVP(PC) Unknown"
    PBE_def2_TZVP_SMD_WATER__Single_Point = "PBE/def2-TZVP/SMD(WATER) Single Point"
    PBE_def2_TZVP_SMD_WATER__Force = "PBE/def2-TZVP/SMD(WATER) Force"
    PBE_def2_TZVP_SMD_WATER__Geometry_Optimization = "PBE/def2-TZVP/SMD(WATER) Geometry Optimization"
    PBE_def2_TZVP_SMD_WATER__Frequency_Analysis = "PBE/def2-TZVP/SMD(WATER) Frequency Analysis"
    PBE_def2_TZVP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_WATER__Unknown = "PBE/def2-TZVP/SMD(WATER) Unknown"
    PBE_def2_TZVP_SMD_THF__Single_Point = "PBE/def2-TZVP/SMD(THF) Single Point"
    PBE_def2_TZVP_SMD_THF__Force = "PBE/def2-TZVP/SMD(THF) Force"
    PBE_def2_TZVP_SMD_THF__Geometry_Optimization = "PBE/def2-TZVP/SMD(THF) Geometry Optimization"
    PBE_def2_TZVP_SMD_THF__Frequency_Analysis = "PBE/def2-TZVP/SMD(THF) Frequency Analysis"
    PBE_def2_TZVP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_THF__Unknown = "PBE/def2-TZVP/SMD(THF) Unknown"
    PBE_def2_TZVP_SMD_G2__Single_Point = "PBE/def2-TZVP/SMD(G2) Single Point"
    PBE_def2_TZVP_SMD_G2__Force = "PBE/def2-TZVP/SMD(G2) Force"
    PBE_def2_TZVP_SMD_G2__Geometry_Optimization = "PBE/def2-TZVP/SMD(G2) Geometry Optimization"
    PBE_def2_TZVP_SMD_G2__Frequency_Analysis = "PBE/def2-TZVP/SMD(G2) Frequency Analysis"
    PBE_def2_TZVP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_G2__Unknown = "PBE/def2-TZVP/SMD(G2) Unknown"
    PBE_def2_TZVP_SMD_DME__Single_Point = "PBE/def2-TZVP/SMD(DME) Single Point"
    PBE_def2_TZVP_SMD_DME__Force = "PBE/def2-TZVP/SMD(DME) Force"
    PBE_def2_TZVP_SMD_DME__Geometry_Optimization = "PBE/def2-TZVP/SMD(DME) Geometry Optimization"
    PBE_def2_TZVP_SMD_DME__Frequency_Analysis = "PBE/def2-TZVP/SMD(DME) Frequency Analysis"
    PBE_def2_TZVP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_DME__Unknown = "PBE/def2-TZVP/SMD(DME) Unknown"
    PBE_def2_TZVP_SMD_EC__Single_Point = "PBE/def2-TZVP/SMD(EC) Single Point"
    PBE_def2_TZVP_SMD_EC__Force = "PBE/def2-TZVP/SMD(EC) Force"
    PBE_def2_TZVP_SMD_EC__Geometry_Optimization = "PBE/def2-TZVP/SMD(EC) Geometry Optimization"
    PBE_def2_TZVP_SMD_EC__Frequency_Analysis = "PBE/def2-TZVP/SMD(EC) Frequency Analysis"
    PBE_def2_TZVP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_EC__Unknown = "PBE/def2-TZVP/SMD(EC) Unknown"
    PBE_def2_TZVP_SMD_EC_EMC__Single_Point = "PBE/def2-TZVP/SMD(EC/EMC) Single Point"
    PBE_def2_TZVP_SMD_EC_EMC__Force = "PBE/def2-TZVP/SMD(EC/EMC) Force"
    PBE_def2_TZVP_SMD_EC_EMC__Geometry_Optimization = "PBE/def2-TZVP/SMD(EC/EMC) Geometry Optimization"
    PBE_def2_TZVP_SMD_EC_EMC__Frequency_Analysis = "PBE/def2-TZVP/SMD(EC/EMC) Frequency Analysis"
    PBE_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_EC_EMC__Unknown = "PBE/def2-TZVP/SMD(EC/EMC) Unknown"
    PBE_def2_TZVP_SMD_PC__Single_Point = "PBE/def2-TZVP/SMD(PC) Single Point"
    PBE_def2_TZVP_SMD_PC__Force = "PBE/def2-TZVP/SMD(PC) Force"
    PBE_def2_TZVP_SMD_PC__Geometry_Optimization = "PBE/def2-TZVP/SMD(PC) Geometry Optimization"
    PBE_def2_TZVP_SMD_PC__Frequency_Analysis = "PBE/def2-TZVP/SMD(PC) Frequency Analysis"
    PBE_def2_TZVP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_SMD_PC__Unknown = "PBE/def2-TZVP/SMD(PC) Unknown"
    PBE_def2_TZVP_CMIRS_WATER__Single_Point = "PBE/def2-TZVP/CMIRS(WATER) Single Point"
    PBE_def2_TZVP_CMIRS_WATER__Force = "PBE/def2-TZVP/CMIRS(WATER) Force"
    PBE_def2_TZVP_CMIRS_WATER__Geometry_Optimization = "PBE/def2-TZVP/CMIRS(WATER) Geometry Optimization"
    PBE_def2_TZVP_CMIRS_WATER__Frequency_Analysis = "PBE/def2-TZVP/CMIRS(WATER) Frequency Analysis"
    PBE_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_WATER__Unknown = "PBE/def2-TZVP/CMIRS(WATER) Unknown"
    PBE_def2_TZVP_CMIRS_THF__Single_Point = "PBE/def2-TZVP/CMIRS(THF) Single Point"
    PBE_def2_TZVP_CMIRS_THF__Force = "PBE/def2-TZVP/CMIRS(THF) Force"
    PBE_def2_TZVP_CMIRS_THF__Geometry_Optimization = "PBE/def2-TZVP/CMIRS(THF) Geometry Optimization"
    PBE_def2_TZVP_CMIRS_THF__Frequency_Analysis = "PBE/def2-TZVP/CMIRS(THF) Frequency Analysis"
    PBE_def2_TZVP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_THF__Unknown = "PBE/def2-TZVP/CMIRS(THF) Unknown"
    PBE_def2_TZVP_CMIRS_G2__Single_Point = "PBE/def2-TZVP/CMIRS(G2) Single Point"
    PBE_def2_TZVP_CMIRS_G2__Force = "PBE/def2-TZVP/CMIRS(G2) Force"
    PBE_def2_TZVP_CMIRS_G2__Geometry_Optimization = "PBE/def2-TZVP/CMIRS(G2) Geometry Optimization"
    PBE_def2_TZVP_CMIRS_G2__Frequency_Analysis = "PBE/def2-TZVP/CMIRS(G2) Frequency Analysis"
    PBE_def2_TZVP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_G2__Unknown = "PBE/def2-TZVP/CMIRS(G2) Unknown"
    PBE_def2_TZVP_CMIRS_DME__Single_Point = "PBE/def2-TZVP/CMIRS(DME) Single Point"
    PBE_def2_TZVP_CMIRS_DME__Force = "PBE/def2-TZVP/CMIRS(DME) Force"
    PBE_def2_TZVP_CMIRS_DME__Geometry_Optimization = "PBE/def2-TZVP/CMIRS(DME) Geometry Optimization"
    PBE_def2_TZVP_CMIRS_DME__Frequency_Analysis = "PBE/def2-TZVP/CMIRS(DME) Frequency Analysis"
    PBE_def2_TZVP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_DME__Unknown = "PBE/def2-TZVP/CMIRS(DME) Unknown"
    PBE_def2_TZVP_CMIRS_EC__Single_Point = "PBE/def2-TZVP/CMIRS(EC) Single Point"
    PBE_def2_TZVP_CMIRS_EC__Force = "PBE/def2-TZVP/CMIRS(EC) Force"
    PBE_def2_TZVP_CMIRS_EC__Geometry_Optimization = "PBE/def2-TZVP/CMIRS(EC) Geometry Optimization"
    PBE_def2_TZVP_CMIRS_EC__Frequency_Analysis = "PBE/def2-TZVP/CMIRS(EC) Frequency Analysis"
    PBE_def2_TZVP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_EC__Unknown = "PBE/def2-TZVP/CMIRS(EC) Unknown"
    PBE_def2_TZVP_CMIRS_EC_EMC__Single_Point = "PBE/def2-TZVP/CMIRS(EC/EMC) Single Point"
    PBE_def2_TZVP_CMIRS_EC_EMC__Force = "PBE/def2-TZVP/CMIRS(EC/EMC) Force"
    PBE_def2_TZVP_CMIRS_EC_EMC__Geometry_Optimization = "PBE/def2-TZVP/CMIRS(EC/EMC) Geometry Optimization"
    PBE_def2_TZVP_CMIRS_EC_EMC__Frequency_Analysis = "PBE/def2-TZVP/CMIRS(EC/EMC) Frequency Analysis"
    PBE_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_EC_EMC__Unknown = "PBE/def2-TZVP/CMIRS(EC/EMC) Unknown"
    PBE_def2_TZVP_CMIRS_PC__Single_Point = "PBE/def2-TZVP/CMIRS(PC) Single Point"
    PBE_def2_TZVP_CMIRS_PC__Force = "PBE/def2-TZVP/CMIRS(PC) Force"
    PBE_def2_TZVP_CMIRS_PC__Geometry_Optimization = "PBE/def2-TZVP/CMIRS(PC) Geometry Optimization"
    PBE_def2_TZVP_CMIRS_PC__Frequency_Analysis = "PBE/def2-TZVP/CMIRS(PC) Frequency Analysis"
    PBE_def2_TZVP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVP_CMIRS_PC__Unknown = "PBE/def2-TZVP/CMIRS(PC) Unknown"
    PBE_def2_TZVPD_VACUUM_Single_Point = "PBE/def2-TZVPD/VACUUM Single Point"
    PBE_def2_TZVPD_VACUUM_Force = "PBE/def2-TZVPD/VACUUM Force"
    PBE_def2_TZVPD_VACUUM_Geometry_Optimization = "PBE/def2-TZVPD/VACUUM Geometry Optimization"
    PBE_def2_TZVPD_VACUUM_Frequency_Analysis = "PBE/def2-TZVPD/VACUUM Frequency Analysis"
    PBE_def2_TZVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/VACUUM Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_VACUUM_Unknown = "PBE/def2-TZVPD/VACUUM Unknown"
    PBE_def2_TZVPD_PCM_WATER__Single_Point = "PBE/def2-TZVPD/PCM(WATER) Single Point"
    PBE_def2_TZVPD_PCM_WATER__Force = "PBE/def2-TZVPD/PCM(WATER) Force"
    PBE_def2_TZVPD_PCM_WATER__Geometry_Optimization = "PBE/def2-TZVPD/PCM(WATER) Geometry Optimization"
    PBE_def2_TZVPD_PCM_WATER__Frequency_Analysis = "PBE/def2-TZVPD/PCM(WATER) Frequency Analysis"
    PBE_def2_TZVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_WATER__Unknown = "PBE/def2-TZVPD/PCM(WATER) Unknown"
    PBE_def2_TZVPD_PCM_THF__Single_Point = "PBE/def2-TZVPD/PCM(THF) Single Point"
    PBE_def2_TZVPD_PCM_THF__Force = "PBE/def2-TZVPD/PCM(THF) Force"
    PBE_def2_TZVPD_PCM_THF__Geometry_Optimization = "PBE/def2-TZVPD/PCM(THF) Geometry Optimization"
    PBE_def2_TZVPD_PCM_THF__Frequency_Analysis = "PBE/def2-TZVPD/PCM(THF) Frequency Analysis"
    PBE_def2_TZVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_THF__Unknown = "PBE/def2-TZVPD/PCM(THF) Unknown"
    PBE_def2_TZVPD_PCM_G2__Single_Point = "PBE/def2-TZVPD/PCM(G2) Single Point"
    PBE_def2_TZVPD_PCM_G2__Force = "PBE/def2-TZVPD/PCM(G2) Force"
    PBE_def2_TZVPD_PCM_G2__Geometry_Optimization = "PBE/def2-TZVPD/PCM(G2) Geometry Optimization"
    PBE_def2_TZVPD_PCM_G2__Frequency_Analysis = "PBE/def2-TZVPD/PCM(G2) Frequency Analysis"
    PBE_def2_TZVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_G2__Unknown = "PBE/def2-TZVPD/PCM(G2) Unknown"
    PBE_def2_TZVPD_PCM_DME__Single_Point = "PBE/def2-TZVPD/PCM(DME) Single Point"
    PBE_def2_TZVPD_PCM_DME__Force = "PBE/def2-TZVPD/PCM(DME) Force"
    PBE_def2_TZVPD_PCM_DME__Geometry_Optimization = "PBE/def2-TZVPD/PCM(DME) Geometry Optimization"
    PBE_def2_TZVPD_PCM_DME__Frequency_Analysis = "PBE/def2-TZVPD/PCM(DME) Frequency Analysis"
    PBE_def2_TZVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_DME__Unknown = "PBE/def2-TZVPD/PCM(DME) Unknown"
    PBE_def2_TZVPD_PCM_EC__Single_Point = "PBE/def2-TZVPD/PCM(EC) Single Point"
    PBE_def2_TZVPD_PCM_EC__Force = "PBE/def2-TZVPD/PCM(EC) Force"
    PBE_def2_TZVPD_PCM_EC__Geometry_Optimization = "PBE/def2-TZVPD/PCM(EC) Geometry Optimization"
    PBE_def2_TZVPD_PCM_EC__Frequency_Analysis = "PBE/def2-TZVPD/PCM(EC) Frequency Analysis"
    PBE_def2_TZVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_EC__Unknown = "PBE/def2-TZVPD/PCM(EC) Unknown"
    PBE_def2_TZVPD_PCM_EC_EMC__Single_Point = "PBE/def2-TZVPD/PCM(EC/EMC) Single Point"
    PBE_def2_TZVPD_PCM_EC_EMC__Force = "PBE/def2-TZVPD/PCM(EC/EMC) Force"
    PBE_def2_TZVPD_PCM_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPD/PCM(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPD_PCM_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPD/PCM(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_EC_EMC__Unknown = "PBE/def2-TZVPD/PCM(EC/EMC) Unknown"
    PBE_def2_TZVPD_PCM_PC__Single_Point = "PBE/def2-TZVPD/PCM(PC) Single Point"
    PBE_def2_TZVPD_PCM_PC__Force = "PBE/def2-TZVPD/PCM(PC) Force"
    PBE_def2_TZVPD_PCM_PC__Geometry_Optimization = "PBE/def2-TZVPD/PCM(PC) Geometry Optimization"
    PBE_def2_TZVPD_PCM_PC__Frequency_Analysis = "PBE/def2-TZVPD/PCM(PC) Frequency Analysis"
    PBE_def2_TZVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_PCM_PC__Unknown = "PBE/def2-TZVPD/PCM(PC) Unknown"
    PBE_def2_TZVPD_ISOSVP_WATER__Single_Point = "PBE/def2-TZVPD/ISOSVP(WATER) Single Point"
    PBE_def2_TZVPD_ISOSVP_WATER__Force = "PBE/def2-TZVPD/ISOSVP(WATER) Force"
    PBE_def2_TZVPD_ISOSVP_WATER__Geometry_Optimization = "PBE/def2-TZVPD/ISOSVP(WATER) Geometry Optimization"
    PBE_def2_TZVPD_ISOSVP_WATER__Frequency_Analysis = "PBE/def2-TZVPD/ISOSVP(WATER) Frequency Analysis"
    PBE_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_WATER__Unknown = "PBE/def2-TZVPD/ISOSVP(WATER) Unknown"
    PBE_def2_TZVPD_ISOSVP_THF__Single_Point = "PBE/def2-TZVPD/ISOSVP(THF) Single Point"
    PBE_def2_TZVPD_ISOSVP_THF__Force = "PBE/def2-TZVPD/ISOSVP(THF) Force"
    PBE_def2_TZVPD_ISOSVP_THF__Geometry_Optimization = "PBE/def2-TZVPD/ISOSVP(THF) Geometry Optimization"
    PBE_def2_TZVPD_ISOSVP_THF__Frequency_Analysis = "PBE/def2-TZVPD/ISOSVP(THF) Frequency Analysis"
    PBE_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_THF__Unknown = "PBE/def2-TZVPD/ISOSVP(THF) Unknown"
    PBE_def2_TZVPD_ISOSVP_G2__Single_Point = "PBE/def2-TZVPD/ISOSVP(G2) Single Point"
    PBE_def2_TZVPD_ISOSVP_G2__Force = "PBE/def2-TZVPD/ISOSVP(G2) Force"
    PBE_def2_TZVPD_ISOSVP_G2__Geometry_Optimization = "PBE/def2-TZVPD/ISOSVP(G2) Geometry Optimization"
    PBE_def2_TZVPD_ISOSVP_G2__Frequency_Analysis = "PBE/def2-TZVPD/ISOSVP(G2) Frequency Analysis"
    PBE_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_G2__Unknown = "PBE/def2-TZVPD/ISOSVP(G2) Unknown"
    PBE_def2_TZVPD_ISOSVP_DME__Single_Point = "PBE/def2-TZVPD/ISOSVP(DME) Single Point"
    PBE_def2_TZVPD_ISOSVP_DME__Force = "PBE/def2-TZVPD/ISOSVP(DME) Force"
    PBE_def2_TZVPD_ISOSVP_DME__Geometry_Optimization = "PBE/def2-TZVPD/ISOSVP(DME) Geometry Optimization"
    PBE_def2_TZVPD_ISOSVP_DME__Frequency_Analysis = "PBE/def2-TZVPD/ISOSVP(DME) Frequency Analysis"
    PBE_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_DME__Unknown = "PBE/def2-TZVPD/ISOSVP(DME) Unknown"
    PBE_def2_TZVPD_ISOSVP_EC__Single_Point = "PBE/def2-TZVPD/ISOSVP(EC) Single Point"
    PBE_def2_TZVPD_ISOSVP_EC__Force = "PBE/def2-TZVPD/ISOSVP(EC) Force"
    PBE_def2_TZVPD_ISOSVP_EC__Geometry_Optimization = "PBE/def2-TZVPD/ISOSVP(EC) Geometry Optimization"
    PBE_def2_TZVPD_ISOSVP_EC__Frequency_Analysis = "PBE/def2-TZVPD/ISOSVP(EC) Frequency Analysis"
    PBE_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_EC__Unknown = "PBE/def2-TZVPD/ISOSVP(EC) Unknown"
    PBE_def2_TZVPD_ISOSVP_EC_EMC__Single_Point = "PBE/def2-TZVPD/ISOSVP(EC/EMC) Single Point"
    PBE_def2_TZVPD_ISOSVP_EC_EMC__Force = "PBE/def2-TZVPD/ISOSVP(EC/EMC) Force"
    PBE_def2_TZVPD_ISOSVP_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPD/ISOSVP(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPD/ISOSVP(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_EC_EMC__Unknown = "PBE/def2-TZVPD/ISOSVP(EC/EMC) Unknown"
    PBE_def2_TZVPD_ISOSVP_PC__Single_Point = "PBE/def2-TZVPD/ISOSVP(PC) Single Point"
    PBE_def2_TZVPD_ISOSVP_PC__Force = "PBE/def2-TZVPD/ISOSVP(PC) Force"
    PBE_def2_TZVPD_ISOSVP_PC__Geometry_Optimization = "PBE/def2-TZVPD/ISOSVP(PC) Geometry Optimization"
    PBE_def2_TZVPD_ISOSVP_PC__Frequency_Analysis = "PBE/def2-TZVPD/ISOSVP(PC) Frequency Analysis"
    PBE_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_ISOSVP_PC__Unknown = "PBE/def2-TZVPD/ISOSVP(PC) Unknown"
    PBE_def2_TZVPD_SMD_WATER__Single_Point = "PBE/def2-TZVPD/SMD(WATER) Single Point"
    PBE_def2_TZVPD_SMD_WATER__Force = "PBE/def2-TZVPD/SMD(WATER) Force"
    PBE_def2_TZVPD_SMD_WATER__Geometry_Optimization = "PBE/def2-TZVPD/SMD(WATER) Geometry Optimization"
    PBE_def2_TZVPD_SMD_WATER__Frequency_Analysis = "PBE/def2-TZVPD/SMD(WATER) Frequency Analysis"
    PBE_def2_TZVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_WATER__Unknown = "PBE/def2-TZVPD/SMD(WATER) Unknown"
    PBE_def2_TZVPD_SMD_THF__Single_Point = "PBE/def2-TZVPD/SMD(THF) Single Point"
    PBE_def2_TZVPD_SMD_THF__Force = "PBE/def2-TZVPD/SMD(THF) Force"
    PBE_def2_TZVPD_SMD_THF__Geometry_Optimization = "PBE/def2-TZVPD/SMD(THF) Geometry Optimization"
    PBE_def2_TZVPD_SMD_THF__Frequency_Analysis = "PBE/def2-TZVPD/SMD(THF) Frequency Analysis"
    PBE_def2_TZVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_THF__Unknown = "PBE/def2-TZVPD/SMD(THF) Unknown"
    PBE_def2_TZVPD_SMD_G2__Single_Point = "PBE/def2-TZVPD/SMD(G2) Single Point"
    PBE_def2_TZVPD_SMD_G2__Force = "PBE/def2-TZVPD/SMD(G2) Force"
    PBE_def2_TZVPD_SMD_G2__Geometry_Optimization = "PBE/def2-TZVPD/SMD(G2) Geometry Optimization"
    PBE_def2_TZVPD_SMD_G2__Frequency_Analysis = "PBE/def2-TZVPD/SMD(G2) Frequency Analysis"
    PBE_def2_TZVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_G2__Unknown = "PBE/def2-TZVPD/SMD(G2) Unknown"
    PBE_def2_TZVPD_SMD_DME__Single_Point = "PBE/def2-TZVPD/SMD(DME) Single Point"
    PBE_def2_TZVPD_SMD_DME__Force = "PBE/def2-TZVPD/SMD(DME) Force"
    PBE_def2_TZVPD_SMD_DME__Geometry_Optimization = "PBE/def2-TZVPD/SMD(DME) Geometry Optimization"
    PBE_def2_TZVPD_SMD_DME__Frequency_Analysis = "PBE/def2-TZVPD/SMD(DME) Frequency Analysis"
    PBE_def2_TZVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_DME__Unknown = "PBE/def2-TZVPD/SMD(DME) Unknown"
    PBE_def2_TZVPD_SMD_EC__Single_Point = "PBE/def2-TZVPD/SMD(EC) Single Point"
    PBE_def2_TZVPD_SMD_EC__Force = "PBE/def2-TZVPD/SMD(EC) Force"
    PBE_def2_TZVPD_SMD_EC__Geometry_Optimization = "PBE/def2-TZVPD/SMD(EC) Geometry Optimization"
    PBE_def2_TZVPD_SMD_EC__Frequency_Analysis = "PBE/def2-TZVPD/SMD(EC) Frequency Analysis"
    PBE_def2_TZVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_EC__Unknown = "PBE/def2-TZVPD/SMD(EC) Unknown"
    PBE_def2_TZVPD_SMD_EC_EMC__Single_Point = "PBE/def2-TZVPD/SMD(EC/EMC) Single Point"
    PBE_def2_TZVPD_SMD_EC_EMC__Force = "PBE/def2-TZVPD/SMD(EC/EMC) Force"
    PBE_def2_TZVPD_SMD_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPD/SMD(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPD_SMD_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPD/SMD(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_EC_EMC__Unknown = "PBE/def2-TZVPD/SMD(EC/EMC) Unknown"
    PBE_def2_TZVPD_SMD_PC__Single_Point = "PBE/def2-TZVPD/SMD(PC) Single Point"
    PBE_def2_TZVPD_SMD_PC__Force = "PBE/def2-TZVPD/SMD(PC) Force"
    PBE_def2_TZVPD_SMD_PC__Geometry_Optimization = "PBE/def2-TZVPD/SMD(PC) Geometry Optimization"
    PBE_def2_TZVPD_SMD_PC__Frequency_Analysis = "PBE/def2-TZVPD/SMD(PC) Frequency Analysis"
    PBE_def2_TZVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_SMD_PC__Unknown = "PBE/def2-TZVPD/SMD(PC) Unknown"
    PBE_def2_TZVPD_CMIRS_WATER__Single_Point = "PBE/def2-TZVPD/CMIRS(WATER) Single Point"
    PBE_def2_TZVPD_CMIRS_WATER__Force = "PBE/def2-TZVPD/CMIRS(WATER) Force"
    PBE_def2_TZVPD_CMIRS_WATER__Geometry_Optimization = "PBE/def2-TZVPD/CMIRS(WATER) Geometry Optimization"
    PBE_def2_TZVPD_CMIRS_WATER__Frequency_Analysis = "PBE/def2-TZVPD/CMIRS(WATER) Frequency Analysis"
    PBE_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_WATER__Unknown = "PBE/def2-TZVPD/CMIRS(WATER) Unknown"
    PBE_def2_TZVPD_CMIRS_THF__Single_Point = "PBE/def2-TZVPD/CMIRS(THF) Single Point"
    PBE_def2_TZVPD_CMIRS_THF__Force = "PBE/def2-TZVPD/CMIRS(THF) Force"
    PBE_def2_TZVPD_CMIRS_THF__Geometry_Optimization = "PBE/def2-TZVPD/CMIRS(THF) Geometry Optimization"
    PBE_def2_TZVPD_CMIRS_THF__Frequency_Analysis = "PBE/def2-TZVPD/CMIRS(THF) Frequency Analysis"
    PBE_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_THF__Unknown = "PBE/def2-TZVPD/CMIRS(THF) Unknown"
    PBE_def2_TZVPD_CMIRS_G2__Single_Point = "PBE/def2-TZVPD/CMIRS(G2) Single Point"
    PBE_def2_TZVPD_CMIRS_G2__Force = "PBE/def2-TZVPD/CMIRS(G2) Force"
    PBE_def2_TZVPD_CMIRS_G2__Geometry_Optimization = "PBE/def2-TZVPD/CMIRS(G2) Geometry Optimization"
    PBE_def2_TZVPD_CMIRS_G2__Frequency_Analysis = "PBE/def2-TZVPD/CMIRS(G2) Frequency Analysis"
    PBE_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_G2__Unknown = "PBE/def2-TZVPD/CMIRS(G2) Unknown"
    PBE_def2_TZVPD_CMIRS_DME__Single_Point = "PBE/def2-TZVPD/CMIRS(DME) Single Point"
    PBE_def2_TZVPD_CMIRS_DME__Force = "PBE/def2-TZVPD/CMIRS(DME) Force"
    PBE_def2_TZVPD_CMIRS_DME__Geometry_Optimization = "PBE/def2-TZVPD/CMIRS(DME) Geometry Optimization"
    PBE_def2_TZVPD_CMIRS_DME__Frequency_Analysis = "PBE/def2-TZVPD/CMIRS(DME) Frequency Analysis"
    PBE_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_DME__Unknown = "PBE/def2-TZVPD/CMIRS(DME) Unknown"
    PBE_def2_TZVPD_CMIRS_EC__Single_Point = "PBE/def2-TZVPD/CMIRS(EC) Single Point"
    PBE_def2_TZVPD_CMIRS_EC__Force = "PBE/def2-TZVPD/CMIRS(EC) Force"
    PBE_def2_TZVPD_CMIRS_EC__Geometry_Optimization = "PBE/def2-TZVPD/CMIRS(EC) Geometry Optimization"
    PBE_def2_TZVPD_CMIRS_EC__Frequency_Analysis = "PBE/def2-TZVPD/CMIRS(EC) Frequency Analysis"
    PBE_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_EC__Unknown = "PBE/def2-TZVPD/CMIRS(EC) Unknown"
    PBE_def2_TZVPD_CMIRS_EC_EMC__Single_Point = "PBE/def2-TZVPD/CMIRS(EC/EMC) Single Point"
    PBE_def2_TZVPD_CMIRS_EC_EMC__Force = "PBE/def2-TZVPD/CMIRS(EC/EMC) Force"
    PBE_def2_TZVPD_CMIRS_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPD/CMIRS(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPD_CMIRS_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPD/CMIRS(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_EC_EMC__Unknown = "PBE/def2-TZVPD/CMIRS(EC/EMC) Unknown"
    PBE_def2_TZVPD_CMIRS_PC__Single_Point = "PBE/def2-TZVPD/CMIRS(PC) Single Point"
    PBE_def2_TZVPD_CMIRS_PC__Force = "PBE/def2-TZVPD/CMIRS(PC) Force"
    PBE_def2_TZVPD_CMIRS_PC__Geometry_Optimization = "PBE/def2-TZVPD/CMIRS(PC) Geometry Optimization"
    PBE_def2_TZVPD_CMIRS_PC__Frequency_Analysis = "PBE/def2-TZVPD/CMIRS(PC) Frequency Analysis"
    PBE_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPD_CMIRS_PC__Unknown = "PBE/def2-TZVPD/CMIRS(PC) Unknown"
    PBE_def2_TZVPP_VACUUM_Single_Point = "PBE/def2-TZVPP/VACUUM Single Point"
    PBE_def2_TZVPP_VACUUM_Force = "PBE/def2-TZVPP/VACUUM Force"
    PBE_def2_TZVPP_VACUUM_Geometry_Optimization = "PBE/def2-TZVPP/VACUUM Geometry Optimization"
    PBE_def2_TZVPP_VACUUM_Frequency_Analysis = "PBE/def2-TZVPP/VACUUM Frequency Analysis"
    PBE_def2_TZVPP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/VACUUM Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_VACUUM_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/VACUUM Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_VACUUM_Unknown = "PBE/def2-TZVPP/VACUUM Unknown"
    PBE_def2_TZVPP_PCM_WATER__Single_Point = "PBE/def2-TZVPP/PCM(WATER) Single Point"
    PBE_def2_TZVPP_PCM_WATER__Force = "PBE/def2-TZVPP/PCM(WATER) Force"
    PBE_def2_TZVPP_PCM_WATER__Geometry_Optimization = "PBE/def2-TZVPP/PCM(WATER) Geometry Optimization"
    PBE_def2_TZVPP_PCM_WATER__Frequency_Analysis = "PBE/def2-TZVPP/PCM(WATER) Frequency Analysis"
    PBE_def2_TZVPP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_WATER__Unknown = "PBE/def2-TZVPP/PCM(WATER) Unknown"
    PBE_def2_TZVPP_PCM_THF__Single_Point = "PBE/def2-TZVPP/PCM(THF) Single Point"
    PBE_def2_TZVPP_PCM_THF__Force = "PBE/def2-TZVPP/PCM(THF) Force"
    PBE_def2_TZVPP_PCM_THF__Geometry_Optimization = "PBE/def2-TZVPP/PCM(THF) Geometry Optimization"
    PBE_def2_TZVPP_PCM_THF__Frequency_Analysis = "PBE/def2-TZVPP/PCM(THF) Frequency Analysis"
    PBE_def2_TZVPP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_THF__Unknown = "PBE/def2-TZVPP/PCM(THF) Unknown"
    PBE_def2_TZVPP_PCM_G2__Single_Point = "PBE/def2-TZVPP/PCM(G2) Single Point"
    PBE_def2_TZVPP_PCM_G2__Force = "PBE/def2-TZVPP/PCM(G2) Force"
    PBE_def2_TZVPP_PCM_G2__Geometry_Optimization = "PBE/def2-TZVPP/PCM(G2) Geometry Optimization"
    PBE_def2_TZVPP_PCM_G2__Frequency_Analysis = "PBE/def2-TZVPP/PCM(G2) Frequency Analysis"
    PBE_def2_TZVPP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_G2__Unknown = "PBE/def2-TZVPP/PCM(G2) Unknown"
    PBE_def2_TZVPP_PCM_DME__Single_Point = "PBE/def2-TZVPP/PCM(DME) Single Point"
    PBE_def2_TZVPP_PCM_DME__Force = "PBE/def2-TZVPP/PCM(DME) Force"
    PBE_def2_TZVPP_PCM_DME__Geometry_Optimization = "PBE/def2-TZVPP/PCM(DME) Geometry Optimization"
    PBE_def2_TZVPP_PCM_DME__Frequency_Analysis = "PBE/def2-TZVPP/PCM(DME) Frequency Analysis"
    PBE_def2_TZVPP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_DME__Unknown = "PBE/def2-TZVPP/PCM(DME) Unknown"
    PBE_def2_TZVPP_PCM_EC__Single_Point = "PBE/def2-TZVPP/PCM(EC) Single Point"
    PBE_def2_TZVPP_PCM_EC__Force = "PBE/def2-TZVPP/PCM(EC) Force"
    PBE_def2_TZVPP_PCM_EC__Geometry_Optimization = "PBE/def2-TZVPP/PCM(EC) Geometry Optimization"
    PBE_def2_TZVPP_PCM_EC__Frequency_Analysis = "PBE/def2-TZVPP/PCM(EC) Frequency Analysis"
    PBE_def2_TZVPP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_EC__Unknown = "PBE/def2-TZVPP/PCM(EC) Unknown"
    PBE_def2_TZVPP_PCM_EC_EMC__Single_Point = "PBE/def2-TZVPP/PCM(EC/EMC) Single Point"
    PBE_def2_TZVPP_PCM_EC_EMC__Force = "PBE/def2-TZVPP/PCM(EC/EMC) Force"
    PBE_def2_TZVPP_PCM_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPP/PCM(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPP_PCM_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPP/PCM(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_EC_EMC__Unknown = "PBE/def2-TZVPP/PCM(EC/EMC) Unknown"
    PBE_def2_TZVPP_PCM_PC__Single_Point = "PBE/def2-TZVPP/PCM(PC) Single Point"
    PBE_def2_TZVPP_PCM_PC__Force = "PBE/def2-TZVPP/PCM(PC) Force"
    PBE_def2_TZVPP_PCM_PC__Geometry_Optimization = "PBE/def2-TZVPP/PCM(PC) Geometry Optimization"
    PBE_def2_TZVPP_PCM_PC__Frequency_Analysis = "PBE/def2-TZVPP/PCM(PC) Frequency Analysis"
    PBE_def2_TZVPP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_PCM_PC__Unknown = "PBE/def2-TZVPP/PCM(PC) Unknown"
    PBE_def2_TZVPP_ISOSVP_WATER__Single_Point = "PBE/def2-TZVPP/ISOSVP(WATER) Single Point"
    PBE_def2_TZVPP_ISOSVP_WATER__Force = "PBE/def2-TZVPP/ISOSVP(WATER) Force"
    PBE_def2_TZVPP_ISOSVP_WATER__Geometry_Optimization = "PBE/def2-TZVPP/ISOSVP(WATER) Geometry Optimization"
    PBE_def2_TZVPP_ISOSVP_WATER__Frequency_Analysis = "PBE/def2-TZVPP/ISOSVP(WATER) Frequency Analysis"
    PBE_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_WATER__Unknown = "PBE/def2-TZVPP/ISOSVP(WATER) Unknown"
    PBE_def2_TZVPP_ISOSVP_THF__Single_Point = "PBE/def2-TZVPP/ISOSVP(THF) Single Point"
    PBE_def2_TZVPP_ISOSVP_THF__Force = "PBE/def2-TZVPP/ISOSVP(THF) Force"
    PBE_def2_TZVPP_ISOSVP_THF__Geometry_Optimization = "PBE/def2-TZVPP/ISOSVP(THF) Geometry Optimization"
    PBE_def2_TZVPP_ISOSVP_THF__Frequency_Analysis = "PBE/def2-TZVPP/ISOSVP(THF) Frequency Analysis"
    PBE_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_THF__Unknown = "PBE/def2-TZVPP/ISOSVP(THF) Unknown"
    PBE_def2_TZVPP_ISOSVP_G2__Single_Point = "PBE/def2-TZVPP/ISOSVP(G2) Single Point"
    PBE_def2_TZVPP_ISOSVP_G2__Force = "PBE/def2-TZVPP/ISOSVP(G2) Force"
    PBE_def2_TZVPP_ISOSVP_G2__Geometry_Optimization = "PBE/def2-TZVPP/ISOSVP(G2) Geometry Optimization"
    PBE_def2_TZVPP_ISOSVP_G2__Frequency_Analysis = "PBE/def2-TZVPP/ISOSVP(G2) Frequency Analysis"
    PBE_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_G2__Unknown = "PBE/def2-TZVPP/ISOSVP(G2) Unknown"
    PBE_def2_TZVPP_ISOSVP_DME__Single_Point = "PBE/def2-TZVPP/ISOSVP(DME) Single Point"
    PBE_def2_TZVPP_ISOSVP_DME__Force = "PBE/def2-TZVPP/ISOSVP(DME) Force"
    PBE_def2_TZVPP_ISOSVP_DME__Geometry_Optimization = "PBE/def2-TZVPP/ISOSVP(DME) Geometry Optimization"
    PBE_def2_TZVPP_ISOSVP_DME__Frequency_Analysis = "PBE/def2-TZVPP/ISOSVP(DME) Frequency Analysis"
    PBE_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_DME__Unknown = "PBE/def2-TZVPP/ISOSVP(DME) Unknown"
    PBE_def2_TZVPP_ISOSVP_EC__Single_Point = "PBE/def2-TZVPP/ISOSVP(EC) Single Point"
    PBE_def2_TZVPP_ISOSVP_EC__Force = "PBE/def2-TZVPP/ISOSVP(EC) Force"
    PBE_def2_TZVPP_ISOSVP_EC__Geometry_Optimization = "PBE/def2-TZVPP/ISOSVP(EC) Geometry Optimization"
    PBE_def2_TZVPP_ISOSVP_EC__Frequency_Analysis = "PBE/def2-TZVPP/ISOSVP(EC) Frequency Analysis"
    PBE_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_EC__Unknown = "PBE/def2-TZVPP/ISOSVP(EC) Unknown"
    PBE_def2_TZVPP_ISOSVP_EC_EMC__Single_Point = "PBE/def2-TZVPP/ISOSVP(EC/EMC) Single Point"
    PBE_def2_TZVPP_ISOSVP_EC_EMC__Force = "PBE/def2-TZVPP/ISOSVP(EC/EMC) Force"
    PBE_def2_TZVPP_ISOSVP_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPP/ISOSVP(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPP/ISOSVP(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_EC_EMC__Unknown = "PBE/def2-TZVPP/ISOSVP(EC/EMC) Unknown"
    PBE_def2_TZVPP_ISOSVP_PC__Single_Point = "PBE/def2-TZVPP/ISOSVP(PC) Single Point"
    PBE_def2_TZVPP_ISOSVP_PC__Force = "PBE/def2-TZVPP/ISOSVP(PC) Force"
    PBE_def2_TZVPP_ISOSVP_PC__Geometry_Optimization = "PBE/def2-TZVPP/ISOSVP(PC) Geometry Optimization"
    PBE_def2_TZVPP_ISOSVP_PC__Frequency_Analysis = "PBE/def2-TZVPP/ISOSVP(PC) Frequency Analysis"
    PBE_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_ISOSVP_PC__Unknown = "PBE/def2-TZVPP/ISOSVP(PC) Unknown"
    PBE_def2_TZVPP_SMD_WATER__Single_Point = "PBE/def2-TZVPP/SMD(WATER) Single Point"
    PBE_def2_TZVPP_SMD_WATER__Force = "PBE/def2-TZVPP/SMD(WATER) Force"
    PBE_def2_TZVPP_SMD_WATER__Geometry_Optimization = "PBE/def2-TZVPP/SMD(WATER) Geometry Optimization"
    PBE_def2_TZVPP_SMD_WATER__Frequency_Analysis = "PBE/def2-TZVPP/SMD(WATER) Frequency Analysis"
    PBE_def2_TZVPP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_WATER__Unknown = "PBE/def2-TZVPP/SMD(WATER) Unknown"
    PBE_def2_TZVPP_SMD_THF__Single_Point = "PBE/def2-TZVPP/SMD(THF) Single Point"
    PBE_def2_TZVPP_SMD_THF__Force = "PBE/def2-TZVPP/SMD(THF) Force"
    PBE_def2_TZVPP_SMD_THF__Geometry_Optimization = "PBE/def2-TZVPP/SMD(THF) Geometry Optimization"
    PBE_def2_TZVPP_SMD_THF__Frequency_Analysis = "PBE/def2-TZVPP/SMD(THF) Frequency Analysis"
    PBE_def2_TZVPP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_THF__Unknown = "PBE/def2-TZVPP/SMD(THF) Unknown"
    PBE_def2_TZVPP_SMD_G2__Single_Point = "PBE/def2-TZVPP/SMD(G2) Single Point"
    PBE_def2_TZVPP_SMD_G2__Force = "PBE/def2-TZVPP/SMD(G2) Force"
    PBE_def2_TZVPP_SMD_G2__Geometry_Optimization = "PBE/def2-TZVPP/SMD(G2) Geometry Optimization"
    PBE_def2_TZVPP_SMD_G2__Frequency_Analysis = "PBE/def2-TZVPP/SMD(G2) Frequency Analysis"
    PBE_def2_TZVPP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_G2__Unknown = "PBE/def2-TZVPP/SMD(G2) Unknown"
    PBE_def2_TZVPP_SMD_DME__Single_Point = "PBE/def2-TZVPP/SMD(DME) Single Point"
    PBE_def2_TZVPP_SMD_DME__Force = "PBE/def2-TZVPP/SMD(DME) Force"
    PBE_def2_TZVPP_SMD_DME__Geometry_Optimization = "PBE/def2-TZVPP/SMD(DME) Geometry Optimization"
    PBE_def2_TZVPP_SMD_DME__Frequency_Analysis = "PBE/def2-TZVPP/SMD(DME) Frequency Analysis"
    PBE_def2_TZVPP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_DME__Unknown = "PBE/def2-TZVPP/SMD(DME) Unknown"
    PBE_def2_TZVPP_SMD_EC__Single_Point = "PBE/def2-TZVPP/SMD(EC) Single Point"
    PBE_def2_TZVPP_SMD_EC__Force = "PBE/def2-TZVPP/SMD(EC) Force"
    PBE_def2_TZVPP_SMD_EC__Geometry_Optimization = "PBE/def2-TZVPP/SMD(EC) Geometry Optimization"
    PBE_def2_TZVPP_SMD_EC__Frequency_Analysis = "PBE/def2-TZVPP/SMD(EC) Frequency Analysis"
    PBE_def2_TZVPP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_EC__Unknown = "PBE/def2-TZVPP/SMD(EC) Unknown"
    PBE_def2_TZVPP_SMD_EC_EMC__Single_Point = "PBE/def2-TZVPP/SMD(EC/EMC) Single Point"
    PBE_def2_TZVPP_SMD_EC_EMC__Force = "PBE/def2-TZVPP/SMD(EC/EMC) Force"
    PBE_def2_TZVPP_SMD_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPP/SMD(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPP_SMD_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPP/SMD(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_EC_EMC__Unknown = "PBE/def2-TZVPP/SMD(EC/EMC) Unknown"
    PBE_def2_TZVPP_SMD_PC__Single_Point = "PBE/def2-TZVPP/SMD(PC) Single Point"
    PBE_def2_TZVPP_SMD_PC__Force = "PBE/def2-TZVPP/SMD(PC) Force"
    PBE_def2_TZVPP_SMD_PC__Geometry_Optimization = "PBE/def2-TZVPP/SMD(PC) Geometry Optimization"
    PBE_def2_TZVPP_SMD_PC__Frequency_Analysis = "PBE/def2-TZVPP/SMD(PC) Frequency Analysis"
    PBE_def2_TZVPP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_SMD_PC__Unknown = "PBE/def2-TZVPP/SMD(PC) Unknown"
    PBE_def2_TZVPP_CMIRS_WATER__Single_Point = "PBE/def2-TZVPP/CMIRS(WATER) Single Point"
    PBE_def2_TZVPP_CMIRS_WATER__Force = "PBE/def2-TZVPP/CMIRS(WATER) Force"
    PBE_def2_TZVPP_CMIRS_WATER__Geometry_Optimization = "PBE/def2-TZVPP/CMIRS(WATER) Geometry Optimization"
    PBE_def2_TZVPP_CMIRS_WATER__Frequency_Analysis = "PBE/def2-TZVPP/CMIRS(WATER) Frequency Analysis"
    PBE_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_WATER__Unknown = "PBE/def2-TZVPP/CMIRS(WATER) Unknown"
    PBE_def2_TZVPP_CMIRS_THF__Single_Point = "PBE/def2-TZVPP/CMIRS(THF) Single Point"
    PBE_def2_TZVPP_CMIRS_THF__Force = "PBE/def2-TZVPP/CMIRS(THF) Force"
    PBE_def2_TZVPP_CMIRS_THF__Geometry_Optimization = "PBE/def2-TZVPP/CMIRS(THF) Geometry Optimization"
    PBE_def2_TZVPP_CMIRS_THF__Frequency_Analysis = "PBE/def2-TZVPP/CMIRS(THF) Frequency Analysis"
    PBE_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_THF__Unknown = "PBE/def2-TZVPP/CMIRS(THF) Unknown"
    PBE_def2_TZVPP_CMIRS_G2__Single_Point = "PBE/def2-TZVPP/CMIRS(G2) Single Point"
    PBE_def2_TZVPP_CMIRS_G2__Force = "PBE/def2-TZVPP/CMIRS(G2) Force"
    PBE_def2_TZVPP_CMIRS_G2__Geometry_Optimization = "PBE/def2-TZVPP/CMIRS(G2) Geometry Optimization"
    PBE_def2_TZVPP_CMIRS_G2__Frequency_Analysis = "PBE/def2-TZVPP/CMIRS(G2) Frequency Analysis"
    PBE_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_G2__Unknown = "PBE/def2-TZVPP/CMIRS(G2) Unknown"
    PBE_def2_TZVPP_CMIRS_DME__Single_Point = "PBE/def2-TZVPP/CMIRS(DME) Single Point"
    PBE_def2_TZVPP_CMIRS_DME__Force = "PBE/def2-TZVPP/CMIRS(DME) Force"
    PBE_def2_TZVPP_CMIRS_DME__Geometry_Optimization = "PBE/def2-TZVPP/CMIRS(DME) Geometry Optimization"
    PBE_def2_TZVPP_CMIRS_DME__Frequency_Analysis = "PBE/def2-TZVPP/CMIRS(DME) Frequency Analysis"
    PBE_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_DME__Unknown = "PBE/def2-TZVPP/CMIRS(DME) Unknown"
    PBE_def2_TZVPP_CMIRS_EC__Single_Point = "PBE/def2-TZVPP/CMIRS(EC) Single Point"
    PBE_def2_TZVPP_CMIRS_EC__Force = "PBE/def2-TZVPP/CMIRS(EC) Force"
    PBE_def2_TZVPP_CMIRS_EC__Geometry_Optimization = "PBE/def2-TZVPP/CMIRS(EC) Geometry Optimization"
    PBE_def2_TZVPP_CMIRS_EC__Frequency_Analysis = "PBE/def2-TZVPP/CMIRS(EC) Frequency Analysis"
    PBE_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_EC__Unknown = "PBE/def2-TZVPP/CMIRS(EC) Unknown"
    PBE_def2_TZVPP_CMIRS_EC_EMC__Single_Point = "PBE/def2-TZVPP/CMIRS(EC/EMC) Single Point"
    PBE_def2_TZVPP_CMIRS_EC_EMC__Force = "PBE/def2-TZVPP/CMIRS(EC/EMC) Force"
    PBE_def2_TZVPP_CMIRS_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPP/CMIRS(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPP_CMIRS_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPP/CMIRS(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_EC_EMC__Unknown = "PBE/def2-TZVPP/CMIRS(EC/EMC) Unknown"
    PBE_def2_TZVPP_CMIRS_PC__Single_Point = "PBE/def2-TZVPP/CMIRS(PC) Single Point"
    PBE_def2_TZVPP_CMIRS_PC__Force = "PBE/def2-TZVPP/CMIRS(PC) Force"
    PBE_def2_TZVPP_CMIRS_PC__Geometry_Optimization = "PBE/def2-TZVPP/CMIRS(PC) Geometry Optimization"
    PBE_def2_TZVPP_CMIRS_PC__Frequency_Analysis = "PBE/def2-TZVPP/CMIRS(PC) Frequency Analysis"
    PBE_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPP_CMIRS_PC__Unknown = "PBE/def2-TZVPP/CMIRS(PC) Unknown"
    PBE_def2_TZVPPD_VACUUM_Single_Point = "PBE/def2-TZVPPD/VACUUM Single Point"
    PBE_def2_TZVPPD_VACUUM_Force = "PBE/def2-TZVPPD/VACUUM Force"
    PBE_def2_TZVPPD_VACUUM_Geometry_Optimization = "PBE/def2-TZVPPD/VACUUM Geometry Optimization"
    PBE_def2_TZVPPD_VACUUM_Frequency_Analysis = "PBE/def2-TZVPPD/VACUUM Frequency Analysis"
    PBE_def2_TZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/VACUUM Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_VACUUM_Unknown = "PBE/def2-TZVPPD/VACUUM Unknown"
    PBE_def2_TZVPPD_PCM_WATER__Single_Point = "PBE/def2-TZVPPD/PCM(WATER) Single Point"
    PBE_def2_TZVPPD_PCM_WATER__Force = "PBE/def2-TZVPPD/PCM(WATER) Force"
    PBE_def2_TZVPPD_PCM_WATER__Geometry_Optimization = "PBE/def2-TZVPPD/PCM(WATER) Geometry Optimization"
    PBE_def2_TZVPPD_PCM_WATER__Frequency_Analysis = "PBE/def2-TZVPPD/PCM(WATER) Frequency Analysis"
    PBE_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_WATER__Unknown = "PBE/def2-TZVPPD/PCM(WATER) Unknown"
    PBE_def2_TZVPPD_PCM_THF__Single_Point = "PBE/def2-TZVPPD/PCM(THF) Single Point"
    PBE_def2_TZVPPD_PCM_THF__Force = "PBE/def2-TZVPPD/PCM(THF) Force"
    PBE_def2_TZVPPD_PCM_THF__Geometry_Optimization = "PBE/def2-TZVPPD/PCM(THF) Geometry Optimization"
    PBE_def2_TZVPPD_PCM_THF__Frequency_Analysis = "PBE/def2-TZVPPD/PCM(THF) Frequency Analysis"
    PBE_def2_TZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_THF__Unknown = "PBE/def2-TZVPPD/PCM(THF) Unknown"
    PBE_def2_TZVPPD_PCM_G2__Single_Point = "PBE/def2-TZVPPD/PCM(G2) Single Point"
    PBE_def2_TZVPPD_PCM_G2__Force = "PBE/def2-TZVPPD/PCM(G2) Force"
    PBE_def2_TZVPPD_PCM_G2__Geometry_Optimization = "PBE/def2-TZVPPD/PCM(G2) Geometry Optimization"
    PBE_def2_TZVPPD_PCM_G2__Frequency_Analysis = "PBE/def2-TZVPPD/PCM(G2) Frequency Analysis"
    PBE_def2_TZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_G2__Unknown = "PBE/def2-TZVPPD/PCM(G2) Unknown"
    PBE_def2_TZVPPD_PCM_DME__Single_Point = "PBE/def2-TZVPPD/PCM(DME) Single Point"
    PBE_def2_TZVPPD_PCM_DME__Force = "PBE/def2-TZVPPD/PCM(DME) Force"
    PBE_def2_TZVPPD_PCM_DME__Geometry_Optimization = "PBE/def2-TZVPPD/PCM(DME) Geometry Optimization"
    PBE_def2_TZVPPD_PCM_DME__Frequency_Analysis = "PBE/def2-TZVPPD/PCM(DME) Frequency Analysis"
    PBE_def2_TZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_DME__Unknown = "PBE/def2-TZVPPD/PCM(DME) Unknown"
    PBE_def2_TZVPPD_PCM_EC__Single_Point = "PBE/def2-TZVPPD/PCM(EC) Single Point"
    PBE_def2_TZVPPD_PCM_EC__Force = "PBE/def2-TZVPPD/PCM(EC) Force"
    PBE_def2_TZVPPD_PCM_EC__Geometry_Optimization = "PBE/def2-TZVPPD/PCM(EC) Geometry Optimization"
    PBE_def2_TZVPPD_PCM_EC__Frequency_Analysis = "PBE/def2-TZVPPD/PCM(EC) Frequency Analysis"
    PBE_def2_TZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_EC__Unknown = "PBE/def2-TZVPPD/PCM(EC) Unknown"
    PBE_def2_TZVPPD_PCM_EC_EMC__Single_Point = "PBE/def2-TZVPPD/PCM(EC/EMC) Single Point"
    PBE_def2_TZVPPD_PCM_EC_EMC__Force = "PBE/def2-TZVPPD/PCM(EC/EMC) Force"
    PBE_def2_TZVPPD_PCM_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPPD/PCM(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPPD_PCM_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPPD/PCM(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_EC_EMC__Unknown = "PBE/def2-TZVPPD/PCM(EC/EMC) Unknown"
    PBE_def2_TZVPPD_PCM_PC__Single_Point = "PBE/def2-TZVPPD/PCM(PC) Single Point"
    PBE_def2_TZVPPD_PCM_PC__Force = "PBE/def2-TZVPPD/PCM(PC) Force"
    PBE_def2_TZVPPD_PCM_PC__Geometry_Optimization = "PBE/def2-TZVPPD/PCM(PC) Geometry Optimization"
    PBE_def2_TZVPPD_PCM_PC__Frequency_Analysis = "PBE/def2-TZVPPD/PCM(PC) Frequency Analysis"
    PBE_def2_TZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_PCM_PC__Unknown = "PBE/def2-TZVPPD/PCM(PC) Unknown"
    PBE_def2_TZVPPD_ISOSVP_WATER__Single_Point = "PBE/def2-TZVPPD/ISOSVP(WATER) Single Point"
    PBE_def2_TZVPPD_ISOSVP_WATER__Force = "PBE/def2-TZVPPD/ISOSVP(WATER) Force"
    PBE_def2_TZVPPD_ISOSVP_WATER__Geometry_Optimization = "PBE/def2-TZVPPD/ISOSVP(WATER) Geometry Optimization"
    PBE_def2_TZVPPD_ISOSVP_WATER__Frequency_Analysis = "PBE/def2-TZVPPD/ISOSVP(WATER) Frequency Analysis"
    PBE_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_WATER__Unknown = "PBE/def2-TZVPPD/ISOSVP(WATER) Unknown"
    PBE_def2_TZVPPD_ISOSVP_THF__Single_Point = "PBE/def2-TZVPPD/ISOSVP(THF) Single Point"
    PBE_def2_TZVPPD_ISOSVP_THF__Force = "PBE/def2-TZVPPD/ISOSVP(THF) Force"
    PBE_def2_TZVPPD_ISOSVP_THF__Geometry_Optimization = "PBE/def2-TZVPPD/ISOSVP(THF) Geometry Optimization"
    PBE_def2_TZVPPD_ISOSVP_THF__Frequency_Analysis = "PBE/def2-TZVPPD/ISOSVP(THF) Frequency Analysis"
    PBE_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_THF__Unknown = "PBE/def2-TZVPPD/ISOSVP(THF) Unknown"
    PBE_def2_TZVPPD_ISOSVP_G2__Single_Point = "PBE/def2-TZVPPD/ISOSVP(G2) Single Point"
    PBE_def2_TZVPPD_ISOSVP_G2__Force = "PBE/def2-TZVPPD/ISOSVP(G2) Force"
    PBE_def2_TZVPPD_ISOSVP_G2__Geometry_Optimization = "PBE/def2-TZVPPD/ISOSVP(G2) Geometry Optimization"
    PBE_def2_TZVPPD_ISOSVP_G2__Frequency_Analysis = "PBE/def2-TZVPPD/ISOSVP(G2) Frequency Analysis"
    PBE_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_G2__Unknown = "PBE/def2-TZVPPD/ISOSVP(G2) Unknown"
    PBE_def2_TZVPPD_ISOSVP_DME__Single_Point = "PBE/def2-TZVPPD/ISOSVP(DME) Single Point"
    PBE_def2_TZVPPD_ISOSVP_DME__Force = "PBE/def2-TZVPPD/ISOSVP(DME) Force"
    PBE_def2_TZVPPD_ISOSVP_DME__Geometry_Optimization = "PBE/def2-TZVPPD/ISOSVP(DME) Geometry Optimization"
    PBE_def2_TZVPPD_ISOSVP_DME__Frequency_Analysis = "PBE/def2-TZVPPD/ISOSVP(DME) Frequency Analysis"
    PBE_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_DME__Unknown = "PBE/def2-TZVPPD/ISOSVP(DME) Unknown"
    PBE_def2_TZVPPD_ISOSVP_EC__Single_Point = "PBE/def2-TZVPPD/ISOSVP(EC) Single Point"
    PBE_def2_TZVPPD_ISOSVP_EC__Force = "PBE/def2-TZVPPD/ISOSVP(EC) Force"
    PBE_def2_TZVPPD_ISOSVP_EC__Geometry_Optimization = "PBE/def2-TZVPPD/ISOSVP(EC) Geometry Optimization"
    PBE_def2_TZVPPD_ISOSVP_EC__Frequency_Analysis = "PBE/def2-TZVPPD/ISOSVP(EC) Frequency Analysis"
    PBE_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_EC__Unknown = "PBE/def2-TZVPPD/ISOSVP(EC) Unknown"
    PBE_def2_TZVPPD_ISOSVP_EC_EMC__Single_Point = "PBE/def2-TZVPPD/ISOSVP(EC/EMC) Single Point"
    PBE_def2_TZVPPD_ISOSVP_EC_EMC__Force = "PBE/def2-TZVPPD/ISOSVP(EC/EMC) Force"
    PBE_def2_TZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_EC_EMC__Unknown = "PBE/def2-TZVPPD/ISOSVP(EC/EMC) Unknown"
    PBE_def2_TZVPPD_ISOSVP_PC__Single_Point = "PBE/def2-TZVPPD/ISOSVP(PC) Single Point"
    PBE_def2_TZVPPD_ISOSVP_PC__Force = "PBE/def2-TZVPPD/ISOSVP(PC) Force"
    PBE_def2_TZVPPD_ISOSVP_PC__Geometry_Optimization = "PBE/def2-TZVPPD/ISOSVP(PC) Geometry Optimization"
    PBE_def2_TZVPPD_ISOSVP_PC__Frequency_Analysis = "PBE/def2-TZVPPD/ISOSVP(PC) Frequency Analysis"
    PBE_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_ISOSVP_PC__Unknown = "PBE/def2-TZVPPD/ISOSVP(PC) Unknown"
    PBE_def2_TZVPPD_SMD_WATER__Single_Point = "PBE/def2-TZVPPD/SMD(WATER) Single Point"
    PBE_def2_TZVPPD_SMD_WATER__Force = "PBE/def2-TZVPPD/SMD(WATER) Force"
    PBE_def2_TZVPPD_SMD_WATER__Geometry_Optimization = "PBE/def2-TZVPPD/SMD(WATER) Geometry Optimization"
    PBE_def2_TZVPPD_SMD_WATER__Frequency_Analysis = "PBE/def2-TZVPPD/SMD(WATER) Frequency Analysis"
    PBE_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_WATER__Unknown = "PBE/def2-TZVPPD/SMD(WATER) Unknown"
    PBE_def2_TZVPPD_SMD_THF__Single_Point = "PBE/def2-TZVPPD/SMD(THF) Single Point"
    PBE_def2_TZVPPD_SMD_THF__Force = "PBE/def2-TZVPPD/SMD(THF) Force"
    PBE_def2_TZVPPD_SMD_THF__Geometry_Optimization = "PBE/def2-TZVPPD/SMD(THF) Geometry Optimization"
    PBE_def2_TZVPPD_SMD_THF__Frequency_Analysis = "PBE/def2-TZVPPD/SMD(THF) Frequency Analysis"
    PBE_def2_TZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_THF__Unknown = "PBE/def2-TZVPPD/SMD(THF) Unknown"
    PBE_def2_TZVPPD_SMD_G2__Single_Point = "PBE/def2-TZVPPD/SMD(G2) Single Point"
    PBE_def2_TZVPPD_SMD_G2__Force = "PBE/def2-TZVPPD/SMD(G2) Force"
    PBE_def2_TZVPPD_SMD_G2__Geometry_Optimization = "PBE/def2-TZVPPD/SMD(G2) Geometry Optimization"
    PBE_def2_TZVPPD_SMD_G2__Frequency_Analysis = "PBE/def2-TZVPPD/SMD(G2) Frequency Analysis"
    PBE_def2_TZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_G2__Unknown = "PBE/def2-TZVPPD/SMD(G2) Unknown"
    PBE_def2_TZVPPD_SMD_DME__Single_Point = "PBE/def2-TZVPPD/SMD(DME) Single Point"
    PBE_def2_TZVPPD_SMD_DME__Force = "PBE/def2-TZVPPD/SMD(DME) Force"
    PBE_def2_TZVPPD_SMD_DME__Geometry_Optimization = "PBE/def2-TZVPPD/SMD(DME) Geometry Optimization"
    PBE_def2_TZVPPD_SMD_DME__Frequency_Analysis = "PBE/def2-TZVPPD/SMD(DME) Frequency Analysis"
    PBE_def2_TZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_DME__Unknown = "PBE/def2-TZVPPD/SMD(DME) Unknown"
    PBE_def2_TZVPPD_SMD_EC__Single_Point = "PBE/def2-TZVPPD/SMD(EC) Single Point"
    PBE_def2_TZVPPD_SMD_EC__Force = "PBE/def2-TZVPPD/SMD(EC) Force"
    PBE_def2_TZVPPD_SMD_EC__Geometry_Optimization = "PBE/def2-TZVPPD/SMD(EC) Geometry Optimization"
    PBE_def2_TZVPPD_SMD_EC__Frequency_Analysis = "PBE/def2-TZVPPD/SMD(EC) Frequency Analysis"
    PBE_def2_TZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_EC__Unknown = "PBE/def2-TZVPPD/SMD(EC) Unknown"
    PBE_def2_TZVPPD_SMD_EC_EMC__Single_Point = "PBE/def2-TZVPPD/SMD(EC/EMC) Single Point"
    PBE_def2_TZVPPD_SMD_EC_EMC__Force = "PBE/def2-TZVPPD/SMD(EC/EMC) Force"
    PBE_def2_TZVPPD_SMD_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPPD/SMD(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPPD_SMD_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPPD/SMD(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_EC_EMC__Unknown = "PBE/def2-TZVPPD/SMD(EC/EMC) Unknown"
    PBE_def2_TZVPPD_SMD_PC__Single_Point = "PBE/def2-TZVPPD/SMD(PC) Single Point"
    PBE_def2_TZVPPD_SMD_PC__Force = "PBE/def2-TZVPPD/SMD(PC) Force"
    PBE_def2_TZVPPD_SMD_PC__Geometry_Optimization = "PBE/def2-TZVPPD/SMD(PC) Geometry Optimization"
    PBE_def2_TZVPPD_SMD_PC__Frequency_Analysis = "PBE/def2-TZVPPD/SMD(PC) Frequency Analysis"
    PBE_def2_TZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_SMD_PC__Unknown = "PBE/def2-TZVPPD/SMD(PC) Unknown"
    PBE_def2_TZVPPD_CMIRS_WATER__Single_Point = "PBE/def2-TZVPPD/CMIRS(WATER) Single Point"
    PBE_def2_TZVPPD_CMIRS_WATER__Force = "PBE/def2-TZVPPD/CMIRS(WATER) Force"
    PBE_def2_TZVPPD_CMIRS_WATER__Geometry_Optimization = "PBE/def2-TZVPPD/CMIRS(WATER) Geometry Optimization"
    PBE_def2_TZVPPD_CMIRS_WATER__Frequency_Analysis = "PBE/def2-TZVPPD/CMIRS(WATER) Frequency Analysis"
    PBE_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_WATER__Unknown = "PBE/def2-TZVPPD/CMIRS(WATER) Unknown"
    PBE_def2_TZVPPD_CMIRS_THF__Single_Point = "PBE/def2-TZVPPD/CMIRS(THF) Single Point"
    PBE_def2_TZVPPD_CMIRS_THF__Force = "PBE/def2-TZVPPD/CMIRS(THF) Force"
    PBE_def2_TZVPPD_CMIRS_THF__Geometry_Optimization = "PBE/def2-TZVPPD/CMIRS(THF) Geometry Optimization"
    PBE_def2_TZVPPD_CMIRS_THF__Frequency_Analysis = "PBE/def2-TZVPPD/CMIRS(THF) Frequency Analysis"
    PBE_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_THF__Unknown = "PBE/def2-TZVPPD/CMIRS(THF) Unknown"
    PBE_def2_TZVPPD_CMIRS_G2__Single_Point = "PBE/def2-TZVPPD/CMIRS(G2) Single Point"
    PBE_def2_TZVPPD_CMIRS_G2__Force = "PBE/def2-TZVPPD/CMIRS(G2) Force"
    PBE_def2_TZVPPD_CMIRS_G2__Geometry_Optimization = "PBE/def2-TZVPPD/CMIRS(G2) Geometry Optimization"
    PBE_def2_TZVPPD_CMIRS_G2__Frequency_Analysis = "PBE/def2-TZVPPD/CMIRS(G2) Frequency Analysis"
    PBE_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_G2__Unknown = "PBE/def2-TZVPPD/CMIRS(G2) Unknown"
    PBE_def2_TZVPPD_CMIRS_DME__Single_Point = "PBE/def2-TZVPPD/CMIRS(DME) Single Point"
    PBE_def2_TZVPPD_CMIRS_DME__Force = "PBE/def2-TZVPPD/CMIRS(DME) Force"
    PBE_def2_TZVPPD_CMIRS_DME__Geometry_Optimization = "PBE/def2-TZVPPD/CMIRS(DME) Geometry Optimization"
    PBE_def2_TZVPPD_CMIRS_DME__Frequency_Analysis = "PBE/def2-TZVPPD/CMIRS(DME) Frequency Analysis"
    PBE_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_DME__Unknown = "PBE/def2-TZVPPD/CMIRS(DME) Unknown"
    PBE_def2_TZVPPD_CMIRS_EC__Single_Point = "PBE/def2-TZVPPD/CMIRS(EC) Single Point"
    PBE_def2_TZVPPD_CMIRS_EC__Force = "PBE/def2-TZVPPD/CMIRS(EC) Force"
    PBE_def2_TZVPPD_CMIRS_EC__Geometry_Optimization = "PBE/def2-TZVPPD/CMIRS(EC) Geometry Optimization"
    PBE_def2_TZVPPD_CMIRS_EC__Frequency_Analysis = "PBE/def2-TZVPPD/CMIRS(EC) Frequency Analysis"
    PBE_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_EC__Unknown = "PBE/def2-TZVPPD/CMIRS(EC) Unknown"
    PBE_def2_TZVPPD_CMIRS_EC_EMC__Single_Point = "PBE/def2-TZVPPD/CMIRS(EC/EMC) Single Point"
    PBE_def2_TZVPPD_CMIRS_EC_EMC__Force = "PBE/def2-TZVPPD/CMIRS(EC/EMC) Force"
    PBE_def2_TZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "PBE/def2-TZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    PBE_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "PBE/def2-TZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    PBE_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_EC_EMC__Unknown = "PBE/def2-TZVPPD/CMIRS(EC/EMC) Unknown"
    PBE_def2_TZVPPD_CMIRS_PC__Single_Point = "PBE/def2-TZVPPD/CMIRS(PC) Single Point"
    PBE_def2_TZVPPD_CMIRS_PC__Force = "PBE/def2-TZVPPD/CMIRS(PC) Force"
    PBE_def2_TZVPPD_CMIRS_PC__Geometry_Optimization = "PBE/def2-TZVPPD/CMIRS(PC) Geometry Optimization"
    PBE_def2_TZVPPD_CMIRS_PC__Frequency_Analysis = "PBE/def2-TZVPPD/CMIRS(PC) Frequency Analysis"
    PBE_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-TZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_TZVPPD_CMIRS_PC__Unknown = "PBE/def2-TZVPPD/CMIRS(PC) Unknown"
    PBE_def2_QZVPPD_VACUUM_Single_Point = "PBE/def2-QZVPPD/VACUUM Single Point"
    PBE_def2_QZVPPD_VACUUM_Force = "PBE/def2-QZVPPD/VACUUM Force"
    PBE_def2_QZVPPD_VACUUM_Geometry_Optimization = "PBE/def2-QZVPPD/VACUUM Geometry Optimization"
    PBE_def2_QZVPPD_VACUUM_Frequency_Analysis = "PBE/def2-QZVPPD/VACUUM Frequency Analysis"
    PBE_def2_QZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/VACUUM Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_VACUUM_Unknown = "PBE/def2-QZVPPD/VACUUM Unknown"
    PBE_def2_QZVPPD_PCM_WATER__Single_Point = "PBE/def2-QZVPPD/PCM(WATER) Single Point"
    PBE_def2_QZVPPD_PCM_WATER__Force = "PBE/def2-QZVPPD/PCM(WATER) Force"
    PBE_def2_QZVPPD_PCM_WATER__Geometry_Optimization = "PBE/def2-QZVPPD/PCM(WATER) Geometry Optimization"
    PBE_def2_QZVPPD_PCM_WATER__Frequency_Analysis = "PBE/def2-QZVPPD/PCM(WATER) Frequency Analysis"
    PBE_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_WATER__Unknown = "PBE/def2-QZVPPD/PCM(WATER) Unknown"
    PBE_def2_QZVPPD_PCM_THF__Single_Point = "PBE/def2-QZVPPD/PCM(THF) Single Point"
    PBE_def2_QZVPPD_PCM_THF__Force = "PBE/def2-QZVPPD/PCM(THF) Force"
    PBE_def2_QZVPPD_PCM_THF__Geometry_Optimization = "PBE/def2-QZVPPD/PCM(THF) Geometry Optimization"
    PBE_def2_QZVPPD_PCM_THF__Frequency_Analysis = "PBE/def2-QZVPPD/PCM(THF) Frequency Analysis"
    PBE_def2_QZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_THF__Unknown = "PBE/def2-QZVPPD/PCM(THF) Unknown"
    PBE_def2_QZVPPD_PCM_G2__Single_Point = "PBE/def2-QZVPPD/PCM(G2) Single Point"
    PBE_def2_QZVPPD_PCM_G2__Force = "PBE/def2-QZVPPD/PCM(G2) Force"
    PBE_def2_QZVPPD_PCM_G2__Geometry_Optimization = "PBE/def2-QZVPPD/PCM(G2) Geometry Optimization"
    PBE_def2_QZVPPD_PCM_G2__Frequency_Analysis = "PBE/def2-QZVPPD/PCM(G2) Frequency Analysis"
    PBE_def2_QZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_G2__Unknown = "PBE/def2-QZVPPD/PCM(G2) Unknown"
    PBE_def2_QZVPPD_PCM_DME__Single_Point = "PBE/def2-QZVPPD/PCM(DME) Single Point"
    PBE_def2_QZVPPD_PCM_DME__Force = "PBE/def2-QZVPPD/PCM(DME) Force"
    PBE_def2_QZVPPD_PCM_DME__Geometry_Optimization = "PBE/def2-QZVPPD/PCM(DME) Geometry Optimization"
    PBE_def2_QZVPPD_PCM_DME__Frequency_Analysis = "PBE/def2-QZVPPD/PCM(DME) Frequency Analysis"
    PBE_def2_QZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_DME__Unknown = "PBE/def2-QZVPPD/PCM(DME) Unknown"
    PBE_def2_QZVPPD_PCM_EC__Single_Point = "PBE/def2-QZVPPD/PCM(EC) Single Point"
    PBE_def2_QZVPPD_PCM_EC__Force = "PBE/def2-QZVPPD/PCM(EC) Force"
    PBE_def2_QZVPPD_PCM_EC__Geometry_Optimization = "PBE/def2-QZVPPD/PCM(EC) Geometry Optimization"
    PBE_def2_QZVPPD_PCM_EC__Frequency_Analysis = "PBE/def2-QZVPPD/PCM(EC) Frequency Analysis"
    PBE_def2_QZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_EC__Unknown = "PBE/def2-QZVPPD/PCM(EC) Unknown"
    PBE_def2_QZVPPD_PCM_EC_EMC__Single_Point = "PBE/def2-QZVPPD/PCM(EC/EMC) Single Point"
    PBE_def2_QZVPPD_PCM_EC_EMC__Force = "PBE/def2-QZVPPD/PCM(EC/EMC) Force"
    PBE_def2_QZVPPD_PCM_EC_EMC__Geometry_Optimization = "PBE/def2-QZVPPD/PCM(EC/EMC) Geometry Optimization"
    PBE_def2_QZVPPD_PCM_EC_EMC__Frequency_Analysis = "PBE/def2-QZVPPD/PCM(EC/EMC) Frequency Analysis"
    PBE_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_EC_EMC__Unknown = "PBE/def2-QZVPPD/PCM(EC/EMC) Unknown"
    PBE_def2_QZVPPD_PCM_PC__Single_Point = "PBE/def2-QZVPPD/PCM(PC) Single Point"
    PBE_def2_QZVPPD_PCM_PC__Force = "PBE/def2-QZVPPD/PCM(PC) Force"
    PBE_def2_QZVPPD_PCM_PC__Geometry_Optimization = "PBE/def2-QZVPPD/PCM(PC) Geometry Optimization"
    PBE_def2_QZVPPD_PCM_PC__Frequency_Analysis = "PBE/def2-QZVPPD/PCM(PC) Frequency Analysis"
    PBE_def2_QZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_PCM_PC__Unknown = "PBE/def2-QZVPPD/PCM(PC) Unknown"
    PBE_def2_QZVPPD_ISOSVP_WATER__Single_Point = "PBE/def2-QZVPPD/ISOSVP(WATER) Single Point"
    PBE_def2_QZVPPD_ISOSVP_WATER__Force = "PBE/def2-QZVPPD/ISOSVP(WATER) Force"
    PBE_def2_QZVPPD_ISOSVP_WATER__Geometry_Optimization = "PBE/def2-QZVPPD/ISOSVP(WATER) Geometry Optimization"
    PBE_def2_QZVPPD_ISOSVP_WATER__Frequency_Analysis = "PBE/def2-QZVPPD/ISOSVP(WATER) Frequency Analysis"
    PBE_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_WATER__Unknown = "PBE/def2-QZVPPD/ISOSVP(WATER) Unknown"
    PBE_def2_QZVPPD_ISOSVP_THF__Single_Point = "PBE/def2-QZVPPD/ISOSVP(THF) Single Point"
    PBE_def2_QZVPPD_ISOSVP_THF__Force = "PBE/def2-QZVPPD/ISOSVP(THF) Force"
    PBE_def2_QZVPPD_ISOSVP_THF__Geometry_Optimization = "PBE/def2-QZVPPD/ISOSVP(THF) Geometry Optimization"
    PBE_def2_QZVPPD_ISOSVP_THF__Frequency_Analysis = "PBE/def2-QZVPPD/ISOSVP(THF) Frequency Analysis"
    PBE_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_THF__Unknown = "PBE/def2-QZVPPD/ISOSVP(THF) Unknown"
    PBE_def2_QZVPPD_ISOSVP_G2__Single_Point = "PBE/def2-QZVPPD/ISOSVP(G2) Single Point"
    PBE_def2_QZVPPD_ISOSVP_G2__Force = "PBE/def2-QZVPPD/ISOSVP(G2) Force"
    PBE_def2_QZVPPD_ISOSVP_G2__Geometry_Optimization = "PBE/def2-QZVPPD/ISOSVP(G2) Geometry Optimization"
    PBE_def2_QZVPPD_ISOSVP_G2__Frequency_Analysis = "PBE/def2-QZVPPD/ISOSVP(G2) Frequency Analysis"
    PBE_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_G2__Unknown = "PBE/def2-QZVPPD/ISOSVP(G2) Unknown"
    PBE_def2_QZVPPD_ISOSVP_DME__Single_Point = "PBE/def2-QZVPPD/ISOSVP(DME) Single Point"
    PBE_def2_QZVPPD_ISOSVP_DME__Force = "PBE/def2-QZVPPD/ISOSVP(DME) Force"
    PBE_def2_QZVPPD_ISOSVP_DME__Geometry_Optimization = "PBE/def2-QZVPPD/ISOSVP(DME) Geometry Optimization"
    PBE_def2_QZVPPD_ISOSVP_DME__Frequency_Analysis = "PBE/def2-QZVPPD/ISOSVP(DME) Frequency Analysis"
    PBE_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_DME__Unknown = "PBE/def2-QZVPPD/ISOSVP(DME) Unknown"
    PBE_def2_QZVPPD_ISOSVP_EC__Single_Point = "PBE/def2-QZVPPD/ISOSVP(EC) Single Point"
    PBE_def2_QZVPPD_ISOSVP_EC__Force = "PBE/def2-QZVPPD/ISOSVP(EC) Force"
    PBE_def2_QZVPPD_ISOSVP_EC__Geometry_Optimization = "PBE/def2-QZVPPD/ISOSVP(EC) Geometry Optimization"
    PBE_def2_QZVPPD_ISOSVP_EC__Frequency_Analysis = "PBE/def2-QZVPPD/ISOSVP(EC) Frequency Analysis"
    PBE_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_EC__Unknown = "PBE/def2-QZVPPD/ISOSVP(EC) Unknown"
    PBE_def2_QZVPPD_ISOSVP_EC_EMC__Single_Point = "PBE/def2-QZVPPD/ISOSVP(EC/EMC) Single Point"
    PBE_def2_QZVPPD_ISOSVP_EC_EMC__Force = "PBE/def2-QZVPPD/ISOSVP(EC/EMC) Force"
    PBE_def2_QZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = "PBE/def2-QZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    PBE_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "PBE/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    PBE_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_EC_EMC__Unknown = "PBE/def2-QZVPPD/ISOSVP(EC/EMC) Unknown"
    PBE_def2_QZVPPD_ISOSVP_PC__Single_Point = "PBE/def2-QZVPPD/ISOSVP(PC) Single Point"
    PBE_def2_QZVPPD_ISOSVP_PC__Force = "PBE/def2-QZVPPD/ISOSVP(PC) Force"
    PBE_def2_QZVPPD_ISOSVP_PC__Geometry_Optimization = "PBE/def2-QZVPPD/ISOSVP(PC) Geometry Optimization"
    PBE_def2_QZVPPD_ISOSVP_PC__Frequency_Analysis = "PBE/def2-QZVPPD/ISOSVP(PC) Frequency Analysis"
    PBE_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_ISOSVP_PC__Unknown = "PBE/def2-QZVPPD/ISOSVP(PC) Unknown"
    PBE_def2_QZVPPD_SMD_WATER__Single_Point = "PBE/def2-QZVPPD/SMD(WATER) Single Point"
    PBE_def2_QZVPPD_SMD_WATER__Force = "PBE/def2-QZVPPD/SMD(WATER) Force"
    PBE_def2_QZVPPD_SMD_WATER__Geometry_Optimization = "PBE/def2-QZVPPD/SMD(WATER) Geometry Optimization"
    PBE_def2_QZVPPD_SMD_WATER__Frequency_Analysis = "PBE/def2-QZVPPD/SMD(WATER) Frequency Analysis"
    PBE_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_WATER__Unknown = "PBE/def2-QZVPPD/SMD(WATER) Unknown"
    PBE_def2_QZVPPD_SMD_THF__Single_Point = "PBE/def2-QZVPPD/SMD(THF) Single Point"
    PBE_def2_QZVPPD_SMD_THF__Force = "PBE/def2-QZVPPD/SMD(THF) Force"
    PBE_def2_QZVPPD_SMD_THF__Geometry_Optimization = "PBE/def2-QZVPPD/SMD(THF) Geometry Optimization"
    PBE_def2_QZVPPD_SMD_THF__Frequency_Analysis = "PBE/def2-QZVPPD/SMD(THF) Frequency Analysis"
    PBE_def2_QZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_THF__Unknown = "PBE/def2-QZVPPD/SMD(THF) Unknown"
    PBE_def2_QZVPPD_SMD_G2__Single_Point = "PBE/def2-QZVPPD/SMD(G2) Single Point"
    PBE_def2_QZVPPD_SMD_G2__Force = "PBE/def2-QZVPPD/SMD(G2) Force"
    PBE_def2_QZVPPD_SMD_G2__Geometry_Optimization = "PBE/def2-QZVPPD/SMD(G2) Geometry Optimization"
    PBE_def2_QZVPPD_SMD_G2__Frequency_Analysis = "PBE/def2-QZVPPD/SMD(G2) Frequency Analysis"
    PBE_def2_QZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_G2__Unknown = "PBE/def2-QZVPPD/SMD(G2) Unknown"
    PBE_def2_QZVPPD_SMD_DME__Single_Point = "PBE/def2-QZVPPD/SMD(DME) Single Point"
    PBE_def2_QZVPPD_SMD_DME__Force = "PBE/def2-QZVPPD/SMD(DME) Force"
    PBE_def2_QZVPPD_SMD_DME__Geometry_Optimization = "PBE/def2-QZVPPD/SMD(DME) Geometry Optimization"
    PBE_def2_QZVPPD_SMD_DME__Frequency_Analysis = "PBE/def2-QZVPPD/SMD(DME) Frequency Analysis"
    PBE_def2_QZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_DME__Unknown = "PBE/def2-QZVPPD/SMD(DME) Unknown"
    PBE_def2_QZVPPD_SMD_EC__Single_Point = "PBE/def2-QZVPPD/SMD(EC) Single Point"
    PBE_def2_QZVPPD_SMD_EC__Force = "PBE/def2-QZVPPD/SMD(EC) Force"
    PBE_def2_QZVPPD_SMD_EC__Geometry_Optimization = "PBE/def2-QZVPPD/SMD(EC) Geometry Optimization"
    PBE_def2_QZVPPD_SMD_EC__Frequency_Analysis = "PBE/def2-QZVPPD/SMD(EC) Frequency Analysis"
    PBE_def2_QZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_EC__Unknown = "PBE/def2-QZVPPD/SMD(EC) Unknown"
    PBE_def2_QZVPPD_SMD_EC_EMC__Single_Point = "PBE/def2-QZVPPD/SMD(EC/EMC) Single Point"
    PBE_def2_QZVPPD_SMD_EC_EMC__Force = "PBE/def2-QZVPPD/SMD(EC/EMC) Force"
    PBE_def2_QZVPPD_SMD_EC_EMC__Geometry_Optimization = "PBE/def2-QZVPPD/SMD(EC/EMC) Geometry Optimization"
    PBE_def2_QZVPPD_SMD_EC_EMC__Frequency_Analysis = "PBE/def2-QZVPPD/SMD(EC/EMC) Frequency Analysis"
    PBE_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_EC_EMC__Unknown = "PBE/def2-QZVPPD/SMD(EC/EMC) Unknown"
    PBE_def2_QZVPPD_SMD_PC__Single_Point = "PBE/def2-QZVPPD/SMD(PC) Single Point"
    PBE_def2_QZVPPD_SMD_PC__Force = "PBE/def2-QZVPPD/SMD(PC) Force"
    PBE_def2_QZVPPD_SMD_PC__Geometry_Optimization = "PBE/def2-QZVPPD/SMD(PC) Geometry Optimization"
    PBE_def2_QZVPPD_SMD_PC__Frequency_Analysis = "PBE/def2-QZVPPD/SMD(PC) Frequency Analysis"
    PBE_def2_QZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_SMD_PC__Unknown = "PBE/def2-QZVPPD/SMD(PC) Unknown"
    PBE_def2_QZVPPD_CMIRS_WATER__Single_Point = "PBE/def2-QZVPPD/CMIRS(WATER) Single Point"
    PBE_def2_QZVPPD_CMIRS_WATER__Force = "PBE/def2-QZVPPD/CMIRS(WATER) Force"
    PBE_def2_QZVPPD_CMIRS_WATER__Geometry_Optimization = "PBE/def2-QZVPPD/CMIRS(WATER) Geometry Optimization"
    PBE_def2_QZVPPD_CMIRS_WATER__Frequency_Analysis = "PBE/def2-QZVPPD/CMIRS(WATER) Frequency Analysis"
    PBE_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_WATER__Unknown = "PBE/def2-QZVPPD/CMIRS(WATER) Unknown"
    PBE_def2_QZVPPD_CMIRS_THF__Single_Point = "PBE/def2-QZVPPD/CMIRS(THF) Single Point"
    PBE_def2_QZVPPD_CMIRS_THF__Force = "PBE/def2-QZVPPD/CMIRS(THF) Force"
    PBE_def2_QZVPPD_CMIRS_THF__Geometry_Optimization = "PBE/def2-QZVPPD/CMIRS(THF) Geometry Optimization"
    PBE_def2_QZVPPD_CMIRS_THF__Frequency_Analysis = "PBE/def2-QZVPPD/CMIRS(THF) Frequency Analysis"
    PBE_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_THF__Unknown = "PBE/def2-QZVPPD/CMIRS(THF) Unknown"
    PBE_def2_QZVPPD_CMIRS_G2__Single_Point = "PBE/def2-QZVPPD/CMIRS(G2) Single Point"
    PBE_def2_QZVPPD_CMIRS_G2__Force = "PBE/def2-QZVPPD/CMIRS(G2) Force"
    PBE_def2_QZVPPD_CMIRS_G2__Geometry_Optimization = "PBE/def2-QZVPPD/CMIRS(G2) Geometry Optimization"
    PBE_def2_QZVPPD_CMIRS_G2__Frequency_Analysis = "PBE/def2-QZVPPD/CMIRS(G2) Frequency Analysis"
    PBE_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_G2__Unknown = "PBE/def2-QZVPPD/CMIRS(G2) Unknown"
    PBE_def2_QZVPPD_CMIRS_DME__Single_Point = "PBE/def2-QZVPPD/CMIRS(DME) Single Point"
    PBE_def2_QZVPPD_CMIRS_DME__Force = "PBE/def2-QZVPPD/CMIRS(DME) Force"
    PBE_def2_QZVPPD_CMIRS_DME__Geometry_Optimization = "PBE/def2-QZVPPD/CMIRS(DME) Geometry Optimization"
    PBE_def2_QZVPPD_CMIRS_DME__Frequency_Analysis = "PBE/def2-QZVPPD/CMIRS(DME) Frequency Analysis"
    PBE_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_DME__Unknown = "PBE/def2-QZVPPD/CMIRS(DME) Unknown"
    PBE_def2_QZVPPD_CMIRS_EC__Single_Point = "PBE/def2-QZVPPD/CMIRS(EC) Single Point"
    PBE_def2_QZVPPD_CMIRS_EC__Force = "PBE/def2-QZVPPD/CMIRS(EC) Force"
    PBE_def2_QZVPPD_CMIRS_EC__Geometry_Optimization = "PBE/def2-QZVPPD/CMIRS(EC) Geometry Optimization"
    PBE_def2_QZVPPD_CMIRS_EC__Frequency_Analysis = "PBE/def2-QZVPPD/CMIRS(EC) Frequency Analysis"
    PBE_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_EC__Unknown = "PBE/def2-QZVPPD/CMIRS(EC) Unknown"
    PBE_def2_QZVPPD_CMIRS_EC_EMC__Single_Point = "PBE/def2-QZVPPD/CMIRS(EC/EMC) Single Point"
    PBE_def2_QZVPPD_CMIRS_EC_EMC__Force = "PBE/def2-QZVPPD/CMIRS(EC/EMC) Force"
    PBE_def2_QZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "PBE/def2-QZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    PBE_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "PBE/def2-QZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    PBE_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_EC_EMC__Unknown = "PBE/def2-QZVPPD/CMIRS(EC/EMC) Unknown"
    PBE_def2_QZVPPD_CMIRS_PC__Single_Point = "PBE/def2-QZVPPD/CMIRS(PC) Single Point"
    PBE_def2_QZVPPD_CMIRS_PC__Force = "PBE/def2-QZVPPD/CMIRS(PC) Force"
    PBE_def2_QZVPPD_CMIRS_PC__Geometry_Optimization = "PBE/def2-QZVPPD/CMIRS(PC) Geometry Optimization"
    PBE_def2_QZVPPD_CMIRS_PC__Frequency_Analysis = "PBE/def2-QZVPPD/CMIRS(PC) Frequency Analysis"
    PBE_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "PBE/def2-QZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    PBE_def2_QZVPPD_CMIRS_PC__Unknown = "PBE/def2-QZVPPD/CMIRS(PC) Unknown"
    B97_D_6_31g_d_VACUUM_Single_Point = "B97-D/6-31g*/VACUUM Single Point"
    B97_D_6_31g_d_VACUUM_Force = "B97-D/6-31g*/VACUUM Force"
    B97_D_6_31g_d_VACUUM_Geometry_Optimization = "B97-D/6-31g*/VACUUM Geometry Optimization"
    B97_D_6_31g_d_VACUUM_Frequency_Analysis = "B97-D/6-31g*/VACUUM Frequency Analysis"
    B97_D_6_31g_d_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/VACUUM Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_VACUUM_Unknown = "B97-D/6-31g*/VACUUM Unknown"
    B97_D_6_31g_d_PCM_WATER__Single_Point = "B97-D/6-31g*/PCM(WATER) Single Point"
    B97_D_6_31g_d_PCM_WATER__Force = "B97-D/6-31g*/PCM(WATER) Force"
    B97_D_6_31g_d_PCM_WATER__Geometry_Optimization = "B97-D/6-31g*/PCM(WATER) Geometry Optimization"
    B97_D_6_31g_d_PCM_WATER__Frequency_Analysis = "B97-D/6-31g*/PCM(WATER) Frequency Analysis"
    B97_D_6_31g_d_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_WATER__Unknown = "B97-D/6-31g*/PCM(WATER) Unknown"
    B97_D_6_31g_d_PCM_THF__Single_Point = "B97-D/6-31g*/PCM(THF) Single Point"
    B97_D_6_31g_d_PCM_THF__Force = "B97-D/6-31g*/PCM(THF) Force"
    B97_D_6_31g_d_PCM_THF__Geometry_Optimization = "B97-D/6-31g*/PCM(THF) Geometry Optimization"
    B97_D_6_31g_d_PCM_THF__Frequency_Analysis = "B97-D/6-31g*/PCM(THF) Frequency Analysis"
    B97_D_6_31g_d_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_THF__Unknown = "B97-D/6-31g*/PCM(THF) Unknown"
    B97_D_6_31g_d_PCM_G2__Single_Point = "B97-D/6-31g*/PCM(G2) Single Point"
    B97_D_6_31g_d_PCM_G2__Force = "B97-D/6-31g*/PCM(G2) Force"
    B97_D_6_31g_d_PCM_G2__Geometry_Optimization = "B97-D/6-31g*/PCM(G2) Geometry Optimization"
    B97_D_6_31g_d_PCM_G2__Frequency_Analysis = "B97-D/6-31g*/PCM(G2) Frequency Analysis"
    B97_D_6_31g_d_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_G2__Unknown = "B97-D/6-31g*/PCM(G2) Unknown"
    B97_D_6_31g_d_PCM_DME__Single_Point = "B97-D/6-31g*/PCM(DME) Single Point"
    B97_D_6_31g_d_PCM_DME__Force = "B97-D/6-31g*/PCM(DME) Force"
    B97_D_6_31g_d_PCM_DME__Geometry_Optimization = "B97-D/6-31g*/PCM(DME) Geometry Optimization"
    B97_D_6_31g_d_PCM_DME__Frequency_Analysis = "B97-D/6-31g*/PCM(DME) Frequency Analysis"
    B97_D_6_31g_d_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_DME__Unknown = "B97-D/6-31g*/PCM(DME) Unknown"
    B97_D_6_31g_d_PCM_EC__Single_Point = "B97-D/6-31g*/PCM(EC) Single Point"
    B97_D_6_31g_d_PCM_EC__Force = "B97-D/6-31g*/PCM(EC) Force"
    B97_D_6_31g_d_PCM_EC__Geometry_Optimization = "B97-D/6-31g*/PCM(EC) Geometry Optimization"
    B97_D_6_31g_d_PCM_EC__Frequency_Analysis = "B97-D/6-31g*/PCM(EC) Frequency Analysis"
    B97_D_6_31g_d_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_EC__Unknown = "B97-D/6-31g*/PCM(EC) Unknown"
    B97_D_6_31g_d_PCM_EC_EMC__Single_Point = "B97-D/6-31g*/PCM(EC/EMC) Single Point"
    B97_D_6_31g_d_PCM_EC_EMC__Force = "B97-D/6-31g*/PCM(EC/EMC) Force"
    B97_D_6_31g_d_PCM_EC_EMC__Geometry_Optimization = "B97-D/6-31g*/PCM(EC/EMC) Geometry Optimization"
    B97_D_6_31g_d_PCM_EC_EMC__Frequency_Analysis = "B97-D/6-31g*/PCM(EC/EMC) Frequency Analysis"
    B97_D_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_EC_EMC__Unknown = "B97-D/6-31g*/PCM(EC/EMC) Unknown"
    B97_D_6_31g_d_PCM_PC__Single_Point = "B97-D/6-31g*/PCM(PC) Single Point"
    B97_D_6_31g_d_PCM_PC__Force = "B97-D/6-31g*/PCM(PC) Force"
    B97_D_6_31g_d_PCM_PC__Geometry_Optimization = "B97-D/6-31g*/PCM(PC) Geometry Optimization"
    B97_D_6_31g_d_PCM_PC__Frequency_Analysis = "B97-D/6-31g*/PCM(PC) Frequency Analysis"
    B97_D_6_31g_d_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_PCM_PC__Unknown = "B97-D/6-31g*/PCM(PC) Unknown"
    B97_D_6_31g_d_ISOSVP_WATER__Single_Point = "B97-D/6-31g*/ISOSVP(WATER) Single Point"
    B97_D_6_31g_d_ISOSVP_WATER__Force = "B97-D/6-31g*/ISOSVP(WATER) Force"
    B97_D_6_31g_d_ISOSVP_WATER__Geometry_Optimization = "B97-D/6-31g*/ISOSVP(WATER) Geometry Optimization"
    B97_D_6_31g_d_ISOSVP_WATER__Frequency_Analysis = "B97-D/6-31g*/ISOSVP(WATER) Frequency Analysis"
    B97_D_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_WATER__Unknown = "B97-D/6-31g*/ISOSVP(WATER) Unknown"
    B97_D_6_31g_d_ISOSVP_THF__Single_Point = "B97-D/6-31g*/ISOSVP(THF) Single Point"
    B97_D_6_31g_d_ISOSVP_THF__Force = "B97-D/6-31g*/ISOSVP(THF) Force"
    B97_D_6_31g_d_ISOSVP_THF__Geometry_Optimization = "B97-D/6-31g*/ISOSVP(THF) Geometry Optimization"
    B97_D_6_31g_d_ISOSVP_THF__Frequency_Analysis = "B97-D/6-31g*/ISOSVP(THF) Frequency Analysis"
    B97_D_6_31g_d_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_THF__Unknown = "B97-D/6-31g*/ISOSVP(THF) Unknown"
    B97_D_6_31g_d_ISOSVP_G2__Single_Point = "B97-D/6-31g*/ISOSVP(G2) Single Point"
    B97_D_6_31g_d_ISOSVP_G2__Force = "B97-D/6-31g*/ISOSVP(G2) Force"
    B97_D_6_31g_d_ISOSVP_G2__Geometry_Optimization = "B97-D/6-31g*/ISOSVP(G2) Geometry Optimization"
    B97_D_6_31g_d_ISOSVP_G2__Frequency_Analysis = "B97-D/6-31g*/ISOSVP(G2) Frequency Analysis"
    B97_D_6_31g_d_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_G2__Unknown = "B97-D/6-31g*/ISOSVP(G2) Unknown"
    B97_D_6_31g_d_ISOSVP_DME__Single_Point = "B97-D/6-31g*/ISOSVP(DME) Single Point"
    B97_D_6_31g_d_ISOSVP_DME__Force = "B97-D/6-31g*/ISOSVP(DME) Force"
    B97_D_6_31g_d_ISOSVP_DME__Geometry_Optimization = "B97-D/6-31g*/ISOSVP(DME) Geometry Optimization"
    B97_D_6_31g_d_ISOSVP_DME__Frequency_Analysis = "B97-D/6-31g*/ISOSVP(DME) Frequency Analysis"
    B97_D_6_31g_d_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_DME__Unknown = "B97-D/6-31g*/ISOSVP(DME) Unknown"
    B97_D_6_31g_d_ISOSVP_EC__Single_Point = "B97-D/6-31g*/ISOSVP(EC) Single Point"
    B97_D_6_31g_d_ISOSVP_EC__Force = "B97-D/6-31g*/ISOSVP(EC) Force"
    B97_D_6_31g_d_ISOSVP_EC__Geometry_Optimization = "B97-D/6-31g*/ISOSVP(EC) Geometry Optimization"
    B97_D_6_31g_d_ISOSVP_EC__Frequency_Analysis = "B97-D/6-31g*/ISOSVP(EC) Frequency Analysis"
    B97_D_6_31g_d_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_EC__Unknown = "B97-D/6-31g*/ISOSVP(EC) Unknown"
    B97_D_6_31g_d_ISOSVP_EC_EMC__Single_Point = "B97-D/6-31g*/ISOSVP(EC/EMC) Single Point"
    B97_D_6_31g_d_ISOSVP_EC_EMC__Force = "B97-D/6-31g*/ISOSVP(EC/EMC) Force"
    B97_D_6_31g_d_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D/6-31g*/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D_6_31g_d_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D/6-31g*/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_EC_EMC__Unknown = "B97-D/6-31g*/ISOSVP(EC/EMC) Unknown"
    B97_D_6_31g_d_ISOSVP_PC__Single_Point = "B97-D/6-31g*/ISOSVP(PC) Single Point"
    B97_D_6_31g_d_ISOSVP_PC__Force = "B97-D/6-31g*/ISOSVP(PC) Force"
    B97_D_6_31g_d_ISOSVP_PC__Geometry_Optimization = "B97-D/6-31g*/ISOSVP(PC) Geometry Optimization"
    B97_D_6_31g_d_ISOSVP_PC__Frequency_Analysis = "B97-D/6-31g*/ISOSVP(PC) Frequency Analysis"
    B97_D_6_31g_d_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_ISOSVP_PC__Unknown = "B97-D/6-31g*/ISOSVP(PC) Unknown"
    B97_D_6_31g_d_SMD_WATER__Single_Point = "B97-D/6-31g*/SMD(WATER) Single Point"
    B97_D_6_31g_d_SMD_WATER__Force = "B97-D/6-31g*/SMD(WATER) Force"
    B97_D_6_31g_d_SMD_WATER__Geometry_Optimization = "B97-D/6-31g*/SMD(WATER) Geometry Optimization"
    B97_D_6_31g_d_SMD_WATER__Frequency_Analysis = "B97-D/6-31g*/SMD(WATER) Frequency Analysis"
    B97_D_6_31g_d_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_WATER__Unknown = "B97-D/6-31g*/SMD(WATER) Unknown"
    B97_D_6_31g_d_SMD_THF__Single_Point = "B97-D/6-31g*/SMD(THF) Single Point"
    B97_D_6_31g_d_SMD_THF__Force = "B97-D/6-31g*/SMD(THF) Force"
    B97_D_6_31g_d_SMD_THF__Geometry_Optimization = "B97-D/6-31g*/SMD(THF) Geometry Optimization"
    B97_D_6_31g_d_SMD_THF__Frequency_Analysis = "B97-D/6-31g*/SMD(THF) Frequency Analysis"
    B97_D_6_31g_d_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_THF__Unknown = "B97-D/6-31g*/SMD(THF) Unknown"
    B97_D_6_31g_d_SMD_G2__Single_Point = "B97-D/6-31g*/SMD(G2) Single Point"
    B97_D_6_31g_d_SMD_G2__Force = "B97-D/6-31g*/SMD(G2) Force"
    B97_D_6_31g_d_SMD_G2__Geometry_Optimization = "B97-D/6-31g*/SMD(G2) Geometry Optimization"
    B97_D_6_31g_d_SMD_G2__Frequency_Analysis = "B97-D/6-31g*/SMD(G2) Frequency Analysis"
    B97_D_6_31g_d_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_G2__Unknown = "B97-D/6-31g*/SMD(G2) Unknown"
    B97_D_6_31g_d_SMD_DME__Single_Point = "B97-D/6-31g*/SMD(DME) Single Point"
    B97_D_6_31g_d_SMD_DME__Force = "B97-D/6-31g*/SMD(DME) Force"
    B97_D_6_31g_d_SMD_DME__Geometry_Optimization = "B97-D/6-31g*/SMD(DME) Geometry Optimization"
    B97_D_6_31g_d_SMD_DME__Frequency_Analysis = "B97-D/6-31g*/SMD(DME) Frequency Analysis"
    B97_D_6_31g_d_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_DME__Unknown = "B97-D/6-31g*/SMD(DME) Unknown"
    B97_D_6_31g_d_SMD_EC__Single_Point = "B97-D/6-31g*/SMD(EC) Single Point"
    B97_D_6_31g_d_SMD_EC__Force = "B97-D/6-31g*/SMD(EC) Force"
    B97_D_6_31g_d_SMD_EC__Geometry_Optimization = "B97-D/6-31g*/SMD(EC) Geometry Optimization"
    B97_D_6_31g_d_SMD_EC__Frequency_Analysis = "B97-D/6-31g*/SMD(EC) Frequency Analysis"
    B97_D_6_31g_d_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_EC__Unknown = "B97-D/6-31g*/SMD(EC) Unknown"
    B97_D_6_31g_d_SMD_EC_EMC__Single_Point = "B97-D/6-31g*/SMD(EC/EMC) Single Point"
    B97_D_6_31g_d_SMD_EC_EMC__Force = "B97-D/6-31g*/SMD(EC/EMC) Force"
    B97_D_6_31g_d_SMD_EC_EMC__Geometry_Optimization = "B97-D/6-31g*/SMD(EC/EMC) Geometry Optimization"
    B97_D_6_31g_d_SMD_EC_EMC__Frequency_Analysis = "B97-D/6-31g*/SMD(EC/EMC) Frequency Analysis"
    B97_D_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_EC_EMC__Unknown = "B97-D/6-31g*/SMD(EC/EMC) Unknown"
    B97_D_6_31g_d_SMD_PC__Single_Point = "B97-D/6-31g*/SMD(PC) Single Point"
    B97_D_6_31g_d_SMD_PC__Force = "B97-D/6-31g*/SMD(PC) Force"
    B97_D_6_31g_d_SMD_PC__Geometry_Optimization = "B97-D/6-31g*/SMD(PC) Geometry Optimization"
    B97_D_6_31g_d_SMD_PC__Frequency_Analysis = "B97-D/6-31g*/SMD(PC) Frequency Analysis"
    B97_D_6_31g_d_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_SMD_PC__Unknown = "B97-D/6-31g*/SMD(PC) Unknown"
    B97_D_6_31g_d_CMIRS_WATER__Single_Point = "B97-D/6-31g*/CMIRS(WATER) Single Point"
    B97_D_6_31g_d_CMIRS_WATER__Force = "B97-D/6-31g*/CMIRS(WATER) Force"
    B97_D_6_31g_d_CMIRS_WATER__Geometry_Optimization = "B97-D/6-31g*/CMIRS(WATER) Geometry Optimization"
    B97_D_6_31g_d_CMIRS_WATER__Frequency_Analysis = "B97-D/6-31g*/CMIRS(WATER) Frequency Analysis"
    B97_D_6_31g_d_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_WATER__Unknown = "B97-D/6-31g*/CMIRS(WATER) Unknown"
    B97_D_6_31g_d_CMIRS_THF__Single_Point = "B97-D/6-31g*/CMIRS(THF) Single Point"
    B97_D_6_31g_d_CMIRS_THF__Force = "B97-D/6-31g*/CMIRS(THF) Force"
    B97_D_6_31g_d_CMIRS_THF__Geometry_Optimization = "B97-D/6-31g*/CMIRS(THF) Geometry Optimization"
    B97_D_6_31g_d_CMIRS_THF__Frequency_Analysis = "B97-D/6-31g*/CMIRS(THF) Frequency Analysis"
    B97_D_6_31g_d_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_THF__Unknown = "B97-D/6-31g*/CMIRS(THF) Unknown"
    B97_D_6_31g_d_CMIRS_G2__Single_Point = "B97-D/6-31g*/CMIRS(G2) Single Point"
    B97_D_6_31g_d_CMIRS_G2__Force = "B97-D/6-31g*/CMIRS(G2) Force"
    B97_D_6_31g_d_CMIRS_G2__Geometry_Optimization = "B97-D/6-31g*/CMIRS(G2) Geometry Optimization"
    B97_D_6_31g_d_CMIRS_G2__Frequency_Analysis = "B97-D/6-31g*/CMIRS(G2) Frequency Analysis"
    B97_D_6_31g_d_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_G2__Unknown = "B97-D/6-31g*/CMIRS(G2) Unknown"
    B97_D_6_31g_d_CMIRS_DME__Single_Point = "B97-D/6-31g*/CMIRS(DME) Single Point"
    B97_D_6_31g_d_CMIRS_DME__Force = "B97-D/6-31g*/CMIRS(DME) Force"
    B97_D_6_31g_d_CMIRS_DME__Geometry_Optimization = "B97-D/6-31g*/CMIRS(DME) Geometry Optimization"
    B97_D_6_31g_d_CMIRS_DME__Frequency_Analysis = "B97-D/6-31g*/CMIRS(DME) Frequency Analysis"
    B97_D_6_31g_d_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_DME__Unknown = "B97-D/6-31g*/CMIRS(DME) Unknown"
    B97_D_6_31g_d_CMIRS_EC__Single_Point = "B97-D/6-31g*/CMIRS(EC) Single Point"
    B97_D_6_31g_d_CMIRS_EC__Force = "B97-D/6-31g*/CMIRS(EC) Force"
    B97_D_6_31g_d_CMIRS_EC__Geometry_Optimization = "B97-D/6-31g*/CMIRS(EC) Geometry Optimization"
    B97_D_6_31g_d_CMIRS_EC__Frequency_Analysis = "B97-D/6-31g*/CMIRS(EC) Frequency Analysis"
    B97_D_6_31g_d_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_EC__Unknown = "B97-D/6-31g*/CMIRS(EC) Unknown"
    B97_D_6_31g_d_CMIRS_EC_EMC__Single_Point = "B97-D/6-31g*/CMIRS(EC/EMC) Single Point"
    B97_D_6_31g_d_CMIRS_EC_EMC__Force = "B97-D/6-31g*/CMIRS(EC/EMC) Force"
    B97_D_6_31g_d_CMIRS_EC_EMC__Geometry_Optimization = "B97-D/6-31g*/CMIRS(EC/EMC) Geometry Optimization"
    B97_D_6_31g_d_CMIRS_EC_EMC__Frequency_Analysis = "B97-D/6-31g*/CMIRS(EC/EMC) Frequency Analysis"
    B97_D_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_EC_EMC__Unknown = "B97-D/6-31g*/CMIRS(EC/EMC) Unknown"
    B97_D_6_31g_d_CMIRS_PC__Single_Point = "B97-D/6-31g*/CMIRS(PC) Single Point"
    B97_D_6_31g_d_CMIRS_PC__Force = "B97-D/6-31g*/CMIRS(PC) Force"
    B97_D_6_31g_d_CMIRS_PC__Geometry_Optimization = "B97-D/6-31g*/CMIRS(PC) Geometry Optimization"
    B97_D_6_31g_d_CMIRS_PC__Frequency_Analysis = "B97-D/6-31g*/CMIRS(PC) Frequency Analysis"
    B97_D_6_31g_d_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/6-31g*/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_6_31g_d_CMIRS_PC__Unknown = "B97-D/6-31g*/CMIRS(PC) Unknown"
    B97_D_def2_SVPD_VACUUM_Single_Point = "B97-D/def2-SVPD/VACUUM Single Point"
    B97_D_def2_SVPD_VACUUM_Force = "B97-D/def2-SVPD/VACUUM Force"
    B97_D_def2_SVPD_VACUUM_Geometry_Optimization = "B97-D/def2-SVPD/VACUUM Geometry Optimization"
    B97_D_def2_SVPD_VACUUM_Frequency_Analysis = "B97-D/def2-SVPD/VACUUM Frequency Analysis"
    B97_D_def2_SVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/VACUUM Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_VACUUM_Unknown = "B97-D/def2-SVPD/VACUUM Unknown"
    B97_D_def2_SVPD_PCM_WATER__Single_Point = "B97-D/def2-SVPD/PCM(WATER) Single Point"
    B97_D_def2_SVPD_PCM_WATER__Force = "B97-D/def2-SVPD/PCM(WATER) Force"
    B97_D_def2_SVPD_PCM_WATER__Geometry_Optimization = "B97-D/def2-SVPD/PCM(WATER) Geometry Optimization"
    B97_D_def2_SVPD_PCM_WATER__Frequency_Analysis = "B97-D/def2-SVPD/PCM(WATER) Frequency Analysis"
    B97_D_def2_SVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_WATER__Unknown = "B97-D/def2-SVPD/PCM(WATER) Unknown"
    B97_D_def2_SVPD_PCM_THF__Single_Point = "B97-D/def2-SVPD/PCM(THF) Single Point"
    B97_D_def2_SVPD_PCM_THF__Force = "B97-D/def2-SVPD/PCM(THF) Force"
    B97_D_def2_SVPD_PCM_THF__Geometry_Optimization = "B97-D/def2-SVPD/PCM(THF) Geometry Optimization"
    B97_D_def2_SVPD_PCM_THF__Frequency_Analysis = "B97-D/def2-SVPD/PCM(THF) Frequency Analysis"
    B97_D_def2_SVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_THF__Unknown = "B97-D/def2-SVPD/PCM(THF) Unknown"
    B97_D_def2_SVPD_PCM_G2__Single_Point = "B97-D/def2-SVPD/PCM(G2) Single Point"
    B97_D_def2_SVPD_PCM_G2__Force = "B97-D/def2-SVPD/PCM(G2) Force"
    B97_D_def2_SVPD_PCM_G2__Geometry_Optimization = "B97-D/def2-SVPD/PCM(G2) Geometry Optimization"
    B97_D_def2_SVPD_PCM_G2__Frequency_Analysis = "B97-D/def2-SVPD/PCM(G2) Frequency Analysis"
    B97_D_def2_SVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_G2__Unknown = "B97-D/def2-SVPD/PCM(G2) Unknown"
    B97_D_def2_SVPD_PCM_DME__Single_Point = "B97-D/def2-SVPD/PCM(DME) Single Point"
    B97_D_def2_SVPD_PCM_DME__Force = "B97-D/def2-SVPD/PCM(DME) Force"
    B97_D_def2_SVPD_PCM_DME__Geometry_Optimization = "B97-D/def2-SVPD/PCM(DME) Geometry Optimization"
    B97_D_def2_SVPD_PCM_DME__Frequency_Analysis = "B97-D/def2-SVPD/PCM(DME) Frequency Analysis"
    B97_D_def2_SVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_DME__Unknown = "B97-D/def2-SVPD/PCM(DME) Unknown"
    B97_D_def2_SVPD_PCM_EC__Single_Point = "B97-D/def2-SVPD/PCM(EC) Single Point"
    B97_D_def2_SVPD_PCM_EC__Force = "B97-D/def2-SVPD/PCM(EC) Force"
    B97_D_def2_SVPD_PCM_EC__Geometry_Optimization = "B97-D/def2-SVPD/PCM(EC) Geometry Optimization"
    B97_D_def2_SVPD_PCM_EC__Frequency_Analysis = "B97-D/def2-SVPD/PCM(EC) Frequency Analysis"
    B97_D_def2_SVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_EC__Unknown = "B97-D/def2-SVPD/PCM(EC) Unknown"
    B97_D_def2_SVPD_PCM_EC_EMC__Single_Point = "B97-D/def2-SVPD/PCM(EC/EMC) Single Point"
    B97_D_def2_SVPD_PCM_EC_EMC__Force = "B97-D/def2-SVPD/PCM(EC/EMC) Force"
    B97_D_def2_SVPD_PCM_EC_EMC__Geometry_Optimization = "B97-D/def2-SVPD/PCM(EC/EMC) Geometry Optimization"
    B97_D_def2_SVPD_PCM_EC_EMC__Frequency_Analysis = "B97-D/def2-SVPD/PCM(EC/EMC) Frequency Analysis"
    B97_D_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_EC_EMC__Unknown = "B97-D/def2-SVPD/PCM(EC/EMC) Unknown"
    B97_D_def2_SVPD_PCM_PC__Single_Point = "B97-D/def2-SVPD/PCM(PC) Single Point"
    B97_D_def2_SVPD_PCM_PC__Force = "B97-D/def2-SVPD/PCM(PC) Force"
    B97_D_def2_SVPD_PCM_PC__Geometry_Optimization = "B97-D/def2-SVPD/PCM(PC) Geometry Optimization"
    B97_D_def2_SVPD_PCM_PC__Frequency_Analysis = "B97-D/def2-SVPD/PCM(PC) Frequency Analysis"
    B97_D_def2_SVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_PCM_PC__Unknown = "B97-D/def2-SVPD/PCM(PC) Unknown"
    B97_D_def2_SVPD_ISOSVP_WATER__Single_Point = "B97-D/def2-SVPD/ISOSVP(WATER) Single Point"
    B97_D_def2_SVPD_ISOSVP_WATER__Force = "B97-D/def2-SVPD/ISOSVP(WATER) Force"
    B97_D_def2_SVPD_ISOSVP_WATER__Geometry_Optimization = "B97-D/def2-SVPD/ISOSVP(WATER) Geometry Optimization"
    B97_D_def2_SVPD_ISOSVP_WATER__Frequency_Analysis = "B97-D/def2-SVPD/ISOSVP(WATER) Frequency Analysis"
    B97_D_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_WATER__Unknown = "B97-D/def2-SVPD/ISOSVP(WATER) Unknown"
    B97_D_def2_SVPD_ISOSVP_THF__Single_Point = "B97-D/def2-SVPD/ISOSVP(THF) Single Point"
    B97_D_def2_SVPD_ISOSVP_THF__Force = "B97-D/def2-SVPD/ISOSVP(THF) Force"
    B97_D_def2_SVPD_ISOSVP_THF__Geometry_Optimization = "B97-D/def2-SVPD/ISOSVP(THF) Geometry Optimization"
    B97_D_def2_SVPD_ISOSVP_THF__Frequency_Analysis = "B97-D/def2-SVPD/ISOSVP(THF) Frequency Analysis"
    B97_D_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_THF__Unknown = "B97-D/def2-SVPD/ISOSVP(THF) Unknown"
    B97_D_def2_SVPD_ISOSVP_G2__Single_Point = "B97-D/def2-SVPD/ISOSVP(G2) Single Point"
    B97_D_def2_SVPD_ISOSVP_G2__Force = "B97-D/def2-SVPD/ISOSVP(G2) Force"
    B97_D_def2_SVPD_ISOSVP_G2__Geometry_Optimization = "B97-D/def2-SVPD/ISOSVP(G2) Geometry Optimization"
    B97_D_def2_SVPD_ISOSVP_G2__Frequency_Analysis = "B97-D/def2-SVPD/ISOSVP(G2) Frequency Analysis"
    B97_D_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_G2__Unknown = "B97-D/def2-SVPD/ISOSVP(G2) Unknown"
    B97_D_def2_SVPD_ISOSVP_DME__Single_Point = "B97-D/def2-SVPD/ISOSVP(DME) Single Point"
    B97_D_def2_SVPD_ISOSVP_DME__Force = "B97-D/def2-SVPD/ISOSVP(DME) Force"
    B97_D_def2_SVPD_ISOSVP_DME__Geometry_Optimization = "B97-D/def2-SVPD/ISOSVP(DME) Geometry Optimization"
    B97_D_def2_SVPD_ISOSVP_DME__Frequency_Analysis = "B97-D/def2-SVPD/ISOSVP(DME) Frequency Analysis"
    B97_D_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_DME__Unknown = "B97-D/def2-SVPD/ISOSVP(DME) Unknown"
    B97_D_def2_SVPD_ISOSVP_EC__Single_Point = "B97-D/def2-SVPD/ISOSVP(EC) Single Point"
    B97_D_def2_SVPD_ISOSVP_EC__Force = "B97-D/def2-SVPD/ISOSVP(EC) Force"
    B97_D_def2_SVPD_ISOSVP_EC__Geometry_Optimization = "B97-D/def2-SVPD/ISOSVP(EC) Geometry Optimization"
    B97_D_def2_SVPD_ISOSVP_EC__Frequency_Analysis = "B97-D/def2-SVPD/ISOSVP(EC) Frequency Analysis"
    B97_D_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_EC__Unknown = "B97-D/def2-SVPD/ISOSVP(EC) Unknown"
    B97_D_def2_SVPD_ISOSVP_EC_EMC__Single_Point = "B97-D/def2-SVPD/ISOSVP(EC/EMC) Single Point"
    B97_D_def2_SVPD_ISOSVP_EC_EMC__Force = "B97-D/def2-SVPD/ISOSVP(EC/EMC) Force"
    B97_D_def2_SVPD_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D/def2-SVPD/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D_def2_SVPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D/def2-SVPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_EC_EMC__Unknown = "B97-D/def2-SVPD/ISOSVP(EC/EMC) Unknown"
    B97_D_def2_SVPD_ISOSVP_PC__Single_Point = "B97-D/def2-SVPD/ISOSVP(PC) Single Point"
    B97_D_def2_SVPD_ISOSVP_PC__Force = "B97-D/def2-SVPD/ISOSVP(PC) Force"
    B97_D_def2_SVPD_ISOSVP_PC__Geometry_Optimization = "B97-D/def2-SVPD/ISOSVP(PC) Geometry Optimization"
    B97_D_def2_SVPD_ISOSVP_PC__Frequency_Analysis = "B97-D/def2-SVPD/ISOSVP(PC) Frequency Analysis"
    B97_D_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_ISOSVP_PC__Unknown = "B97-D/def2-SVPD/ISOSVP(PC) Unknown"
    B97_D_def2_SVPD_SMD_WATER__Single_Point = "B97-D/def2-SVPD/SMD(WATER) Single Point"
    B97_D_def2_SVPD_SMD_WATER__Force = "B97-D/def2-SVPD/SMD(WATER) Force"
    B97_D_def2_SVPD_SMD_WATER__Geometry_Optimization = "B97-D/def2-SVPD/SMD(WATER) Geometry Optimization"
    B97_D_def2_SVPD_SMD_WATER__Frequency_Analysis = "B97-D/def2-SVPD/SMD(WATER) Frequency Analysis"
    B97_D_def2_SVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_WATER__Unknown = "B97-D/def2-SVPD/SMD(WATER) Unknown"
    B97_D_def2_SVPD_SMD_THF__Single_Point = "B97-D/def2-SVPD/SMD(THF) Single Point"
    B97_D_def2_SVPD_SMD_THF__Force = "B97-D/def2-SVPD/SMD(THF) Force"
    B97_D_def2_SVPD_SMD_THF__Geometry_Optimization = "B97-D/def2-SVPD/SMD(THF) Geometry Optimization"
    B97_D_def2_SVPD_SMD_THF__Frequency_Analysis = "B97-D/def2-SVPD/SMD(THF) Frequency Analysis"
    B97_D_def2_SVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_THF__Unknown = "B97-D/def2-SVPD/SMD(THF) Unknown"
    B97_D_def2_SVPD_SMD_G2__Single_Point = "B97-D/def2-SVPD/SMD(G2) Single Point"
    B97_D_def2_SVPD_SMD_G2__Force = "B97-D/def2-SVPD/SMD(G2) Force"
    B97_D_def2_SVPD_SMD_G2__Geometry_Optimization = "B97-D/def2-SVPD/SMD(G2) Geometry Optimization"
    B97_D_def2_SVPD_SMD_G2__Frequency_Analysis = "B97-D/def2-SVPD/SMD(G2) Frequency Analysis"
    B97_D_def2_SVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_G2__Unknown = "B97-D/def2-SVPD/SMD(G2) Unknown"
    B97_D_def2_SVPD_SMD_DME__Single_Point = "B97-D/def2-SVPD/SMD(DME) Single Point"
    B97_D_def2_SVPD_SMD_DME__Force = "B97-D/def2-SVPD/SMD(DME) Force"
    B97_D_def2_SVPD_SMD_DME__Geometry_Optimization = "B97-D/def2-SVPD/SMD(DME) Geometry Optimization"
    B97_D_def2_SVPD_SMD_DME__Frequency_Analysis = "B97-D/def2-SVPD/SMD(DME) Frequency Analysis"
    B97_D_def2_SVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_DME__Unknown = "B97-D/def2-SVPD/SMD(DME) Unknown"
    B97_D_def2_SVPD_SMD_EC__Single_Point = "B97-D/def2-SVPD/SMD(EC) Single Point"
    B97_D_def2_SVPD_SMD_EC__Force = "B97-D/def2-SVPD/SMD(EC) Force"
    B97_D_def2_SVPD_SMD_EC__Geometry_Optimization = "B97-D/def2-SVPD/SMD(EC) Geometry Optimization"
    B97_D_def2_SVPD_SMD_EC__Frequency_Analysis = "B97-D/def2-SVPD/SMD(EC) Frequency Analysis"
    B97_D_def2_SVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_EC__Unknown = "B97-D/def2-SVPD/SMD(EC) Unknown"
    B97_D_def2_SVPD_SMD_EC_EMC__Single_Point = "B97-D/def2-SVPD/SMD(EC/EMC) Single Point"
    B97_D_def2_SVPD_SMD_EC_EMC__Force = "B97-D/def2-SVPD/SMD(EC/EMC) Force"
    B97_D_def2_SVPD_SMD_EC_EMC__Geometry_Optimization = "B97-D/def2-SVPD/SMD(EC/EMC) Geometry Optimization"
    B97_D_def2_SVPD_SMD_EC_EMC__Frequency_Analysis = "B97-D/def2-SVPD/SMD(EC/EMC) Frequency Analysis"
    B97_D_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_EC_EMC__Unknown = "B97-D/def2-SVPD/SMD(EC/EMC) Unknown"
    B97_D_def2_SVPD_SMD_PC__Single_Point = "B97-D/def2-SVPD/SMD(PC) Single Point"
    B97_D_def2_SVPD_SMD_PC__Force = "B97-D/def2-SVPD/SMD(PC) Force"
    B97_D_def2_SVPD_SMD_PC__Geometry_Optimization = "B97-D/def2-SVPD/SMD(PC) Geometry Optimization"
    B97_D_def2_SVPD_SMD_PC__Frequency_Analysis = "B97-D/def2-SVPD/SMD(PC) Frequency Analysis"
    B97_D_def2_SVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_SMD_PC__Unknown = "B97-D/def2-SVPD/SMD(PC) Unknown"
    B97_D_def2_SVPD_CMIRS_WATER__Single_Point = "B97-D/def2-SVPD/CMIRS(WATER) Single Point"
    B97_D_def2_SVPD_CMIRS_WATER__Force = "B97-D/def2-SVPD/CMIRS(WATER) Force"
    B97_D_def2_SVPD_CMIRS_WATER__Geometry_Optimization = "B97-D/def2-SVPD/CMIRS(WATER) Geometry Optimization"
    B97_D_def2_SVPD_CMIRS_WATER__Frequency_Analysis = "B97-D/def2-SVPD/CMIRS(WATER) Frequency Analysis"
    B97_D_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_WATER__Unknown = "B97-D/def2-SVPD/CMIRS(WATER) Unknown"
    B97_D_def2_SVPD_CMIRS_THF__Single_Point = "B97-D/def2-SVPD/CMIRS(THF) Single Point"
    B97_D_def2_SVPD_CMIRS_THF__Force = "B97-D/def2-SVPD/CMIRS(THF) Force"
    B97_D_def2_SVPD_CMIRS_THF__Geometry_Optimization = "B97-D/def2-SVPD/CMIRS(THF) Geometry Optimization"
    B97_D_def2_SVPD_CMIRS_THF__Frequency_Analysis = "B97-D/def2-SVPD/CMIRS(THF) Frequency Analysis"
    B97_D_def2_SVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_THF__Unknown = "B97-D/def2-SVPD/CMIRS(THF) Unknown"
    B97_D_def2_SVPD_CMIRS_G2__Single_Point = "B97-D/def2-SVPD/CMIRS(G2) Single Point"
    B97_D_def2_SVPD_CMIRS_G2__Force = "B97-D/def2-SVPD/CMIRS(G2) Force"
    B97_D_def2_SVPD_CMIRS_G2__Geometry_Optimization = "B97-D/def2-SVPD/CMIRS(G2) Geometry Optimization"
    B97_D_def2_SVPD_CMIRS_G2__Frequency_Analysis = "B97-D/def2-SVPD/CMIRS(G2) Frequency Analysis"
    B97_D_def2_SVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_G2__Unknown = "B97-D/def2-SVPD/CMIRS(G2) Unknown"
    B97_D_def2_SVPD_CMIRS_DME__Single_Point = "B97-D/def2-SVPD/CMIRS(DME) Single Point"
    B97_D_def2_SVPD_CMIRS_DME__Force = "B97-D/def2-SVPD/CMIRS(DME) Force"
    B97_D_def2_SVPD_CMIRS_DME__Geometry_Optimization = "B97-D/def2-SVPD/CMIRS(DME) Geometry Optimization"
    B97_D_def2_SVPD_CMIRS_DME__Frequency_Analysis = "B97-D/def2-SVPD/CMIRS(DME) Frequency Analysis"
    B97_D_def2_SVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_DME__Unknown = "B97-D/def2-SVPD/CMIRS(DME) Unknown"
    B97_D_def2_SVPD_CMIRS_EC__Single_Point = "B97-D/def2-SVPD/CMIRS(EC) Single Point"
    B97_D_def2_SVPD_CMIRS_EC__Force = "B97-D/def2-SVPD/CMIRS(EC) Force"
    B97_D_def2_SVPD_CMIRS_EC__Geometry_Optimization = "B97-D/def2-SVPD/CMIRS(EC) Geometry Optimization"
    B97_D_def2_SVPD_CMIRS_EC__Frequency_Analysis = "B97-D/def2-SVPD/CMIRS(EC) Frequency Analysis"
    B97_D_def2_SVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_EC__Unknown = "B97-D/def2-SVPD/CMIRS(EC) Unknown"
    B97_D_def2_SVPD_CMIRS_EC_EMC__Single_Point = "B97-D/def2-SVPD/CMIRS(EC/EMC) Single Point"
    B97_D_def2_SVPD_CMIRS_EC_EMC__Force = "B97-D/def2-SVPD/CMIRS(EC/EMC) Force"
    B97_D_def2_SVPD_CMIRS_EC_EMC__Geometry_Optimization = "B97-D/def2-SVPD/CMIRS(EC/EMC) Geometry Optimization"
    B97_D_def2_SVPD_CMIRS_EC_EMC__Frequency_Analysis = "B97-D/def2-SVPD/CMIRS(EC/EMC) Frequency Analysis"
    B97_D_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_EC_EMC__Unknown = "B97-D/def2-SVPD/CMIRS(EC/EMC) Unknown"
    B97_D_def2_SVPD_CMIRS_PC__Single_Point = "B97-D/def2-SVPD/CMIRS(PC) Single Point"
    B97_D_def2_SVPD_CMIRS_PC__Force = "B97-D/def2-SVPD/CMIRS(PC) Force"
    B97_D_def2_SVPD_CMIRS_PC__Geometry_Optimization = "B97-D/def2-SVPD/CMIRS(PC) Geometry Optimization"
    B97_D_def2_SVPD_CMIRS_PC__Frequency_Analysis = "B97-D/def2-SVPD/CMIRS(PC) Frequency Analysis"
    B97_D_def2_SVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-SVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_SVPD_CMIRS_PC__Unknown = "B97-D/def2-SVPD/CMIRS(PC) Unknown"
    B97_D_def2_TZVP_VACUUM_Single_Point = "B97-D/def2-TZVP/VACUUM Single Point"
    B97_D_def2_TZVP_VACUUM_Force = "B97-D/def2-TZVP/VACUUM Force"
    B97_D_def2_TZVP_VACUUM_Geometry_Optimization = "B97-D/def2-TZVP/VACUUM Geometry Optimization"
    B97_D_def2_TZVP_VACUUM_Frequency_Analysis = "B97-D/def2-TZVP/VACUUM Frequency Analysis"
    B97_D_def2_TZVP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/VACUUM Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_VACUUM_Unknown = "B97-D/def2-TZVP/VACUUM Unknown"
    B97_D_def2_TZVP_PCM_WATER__Single_Point = "B97-D/def2-TZVP/PCM(WATER) Single Point"
    B97_D_def2_TZVP_PCM_WATER__Force = "B97-D/def2-TZVP/PCM(WATER) Force"
    B97_D_def2_TZVP_PCM_WATER__Geometry_Optimization = "B97-D/def2-TZVP/PCM(WATER) Geometry Optimization"
    B97_D_def2_TZVP_PCM_WATER__Frequency_Analysis = "B97-D/def2-TZVP/PCM(WATER) Frequency Analysis"
    B97_D_def2_TZVP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_WATER__Unknown = "B97-D/def2-TZVP/PCM(WATER) Unknown"
    B97_D_def2_TZVP_PCM_THF__Single_Point = "B97-D/def2-TZVP/PCM(THF) Single Point"
    B97_D_def2_TZVP_PCM_THF__Force = "B97-D/def2-TZVP/PCM(THF) Force"
    B97_D_def2_TZVP_PCM_THF__Geometry_Optimization = "B97-D/def2-TZVP/PCM(THF) Geometry Optimization"
    B97_D_def2_TZVP_PCM_THF__Frequency_Analysis = "B97-D/def2-TZVP/PCM(THF) Frequency Analysis"
    B97_D_def2_TZVP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_THF__Unknown = "B97-D/def2-TZVP/PCM(THF) Unknown"
    B97_D_def2_TZVP_PCM_G2__Single_Point = "B97-D/def2-TZVP/PCM(G2) Single Point"
    B97_D_def2_TZVP_PCM_G2__Force = "B97-D/def2-TZVP/PCM(G2) Force"
    B97_D_def2_TZVP_PCM_G2__Geometry_Optimization = "B97-D/def2-TZVP/PCM(G2) Geometry Optimization"
    B97_D_def2_TZVP_PCM_G2__Frequency_Analysis = "B97-D/def2-TZVP/PCM(G2) Frequency Analysis"
    B97_D_def2_TZVP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_G2__Unknown = "B97-D/def2-TZVP/PCM(G2) Unknown"
    B97_D_def2_TZVP_PCM_DME__Single_Point = "B97-D/def2-TZVP/PCM(DME) Single Point"
    B97_D_def2_TZVP_PCM_DME__Force = "B97-D/def2-TZVP/PCM(DME) Force"
    B97_D_def2_TZVP_PCM_DME__Geometry_Optimization = "B97-D/def2-TZVP/PCM(DME) Geometry Optimization"
    B97_D_def2_TZVP_PCM_DME__Frequency_Analysis = "B97-D/def2-TZVP/PCM(DME) Frequency Analysis"
    B97_D_def2_TZVP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_DME__Unknown = "B97-D/def2-TZVP/PCM(DME) Unknown"
    B97_D_def2_TZVP_PCM_EC__Single_Point = "B97-D/def2-TZVP/PCM(EC) Single Point"
    B97_D_def2_TZVP_PCM_EC__Force = "B97-D/def2-TZVP/PCM(EC) Force"
    B97_D_def2_TZVP_PCM_EC__Geometry_Optimization = "B97-D/def2-TZVP/PCM(EC) Geometry Optimization"
    B97_D_def2_TZVP_PCM_EC__Frequency_Analysis = "B97-D/def2-TZVP/PCM(EC) Frequency Analysis"
    B97_D_def2_TZVP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_EC__Unknown = "B97-D/def2-TZVP/PCM(EC) Unknown"
    B97_D_def2_TZVP_PCM_EC_EMC__Single_Point = "B97-D/def2-TZVP/PCM(EC/EMC) Single Point"
    B97_D_def2_TZVP_PCM_EC_EMC__Force = "B97-D/def2-TZVP/PCM(EC/EMC) Force"
    B97_D_def2_TZVP_PCM_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVP/PCM(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVP_PCM_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVP/PCM(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_EC_EMC__Unknown = "B97-D/def2-TZVP/PCM(EC/EMC) Unknown"
    B97_D_def2_TZVP_PCM_PC__Single_Point = "B97-D/def2-TZVP/PCM(PC) Single Point"
    B97_D_def2_TZVP_PCM_PC__Force = "B97-D/def2-TZVP/PCM(PC) Force"
    B97_D_def2_TZVP_PCM_PC__Geometry_Optimization = "B97-D/def2-TZVP/PCM(PC) Geometry Optimization"
    B97_D_def2_TZVP_PCM_PC__Frequency_Analysis = "B97-D/def2-TZVP/PCM(PC) Frequency Analysis"
    B97_D_def2_TZVP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_PCM_PC__Unknown = "B97-D/def2-TZVP/PCM(PC) Unknown"
    B97_D_def2_TZVP_ISOSVP_WATER__Single_Point = "B97-D/def2-TZVP/ISOSVP(WATER) Single Point"
    B97_D_def2_TZVP_ISOSVP_WATER__Force = "B97-D/def2-TZVP/ISOSVP(WATER) Force"
    B97_D_def2_TZVP_ISOSVP_WATER__Geometry_Optimization = "B97-D/def2-TZVP/ISOSVP(WATER) Geometry Optimization"
    B97_D_def2_TZVP_ISOSVP_WATER__Frequency_Analysis = "B97-D/def2-TZVP/ISOSVP(WATER) Frequency Analysis"
    B97_D_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_WATER__Unknown = "B97-D/def2-TZVP/ISOSVP(WATER) Unknown"
    B97_D_def2_TZVP_ISOSVP_THF__Single_Point = "B97-D/def2-TZVP/ISOSVP(THF) Single Point"
    B97_D_def2_TZVP_ISOSVP_THF__Force = "B97-D/def2-TZVP/ISOSVP(THF) Force"
    B97_D_def2_TZVP_ISOSVP_THF__Geometry_Optimization = "B97-D/def2-TZVP/ISOSVP(THF) Geometry Optimization"
    B97_D_def2_TZVP_ISOSVP_THF__Frequency_Analysis = "B97-D/def2-TZVP/ISOSVP(THF) Frequency Analysis"
    B97_D_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_THF__Unknown = "B97-D/def2-TZVP/ISOSVP(THF) Unknown"
    B97_D_def2_TZVP_ISOSVP_G2__Single_Point = "B97-D/def2-TZVP/ISOSVP(G2) Single Point"
    B97_D_def2_TZVP_ISOSVP_G2__Force = "B97-D/def2-TZVP/ISOSVP(G2) Force"
    B97_D_def2_TZVP_ISOSVP_G2__Geometry_Optimization = "B97-D/def2-TZVP/ISOSVP(G2) Geometry Optimization"
    B97_D_def2_TZVP_ISOSVP_G2__Frequency_Analysis = "B97-D/def2-TZVP/ISOSVP(G2) Frequency Analysis"
    B97_D_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_G2__Unknown = "B97-D/def2-TZVP/ISOSVP(G2) Unknown"
    B97_D_def2_TZVP_ISOSVP_DME__Single_Point = "B97-D/def2-TZVP/ISOSVP(DME) Single Point"
    B97_D_def2_TZVP_ISOSVP_DME__Force = "B97-D/def2-TZVP/ISOSVP(DME) Force"
    B97_D_def2_TZVP_ISOSVP_DME__Geometry_Optimization = "B97-D/def2-TZVP/ISOSVP(DME) Geometry Optimization"
    B97_D_def2_TZVP_ISOSVP_DME__Frequency_Analysis = "B97-D/def2-TZVP/ISOSVP(DME) Frequency Analysis"
    B97_D_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_DME__Unknown = "B97-D/def2-TZVP/ISOSVP(DME) Unknown"
    B97_D_def2_TZVP_ISOSVP_EC__Single_Point = "B97-D/def2-TZVP/ISOSVP(EC) Single Point"
    B97_D_def2_TZVP_ISOSVP_EC__Force = "B97-D/def2-TZVP/ISOSVP(EC) Force"
    B97_D_def2_TZVP_ISOSVP_EC__Geometry_Optimization = "B97-D/def2-TZVP/ISOSVP(EC) Geometry Optimization"
    B97_D_def2_TZVP_ISOSVP_EC__Frequency_Analysis = "B97-D/def2-TZVP/ISOSVP(EC) Frequency Analysis"
    B97_D_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_EC__Unknown = "B97-D/def2-TZVP/ISOSVP(EC) Unknown"
    B97_D_def2_TZVP_ISOSVP_EC_EMC__Single_Point = "B97-D/def2-TZVP/ISOSVP(EC/EMC) Single Point"
    B97_D_def2_TZVP_ISOSVP_EC_EMC__Force = "B97-D/def2-TZVP/ISOSVP(EC/EMC) Force"
    B97_D_def2_TZVP_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVP/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVP_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVP/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_EC_EMC__Unknown = "B97-D/def2-TZVP/ISOSVP(EC/EMC) Unknown"
    B97_D_def2_TZVP_ISOSVP_PC__Single_Point = "B97-D/def2-TZVP/ISOSVP(PC) Single Point"
    B97_D_def2_TZVP_ISOSVP_PC__Force = "B97-D/def2-TZVP/ISOSVP(PC) Force"
    B97_D_def2_TZVP_ISOSVP_PC__Geometry_Optimization = "B97-D/def2-TZVP/ISOSVP(PC) Geometry Optimization"
    B97_D_def2_TZVP_ISOSVP_PC__Frequency_Analysis = "B97-D/def2-TZVP/ISOSVP(PC) Frequency Analysis"
    B97_D_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_ISOSVP_PC__Unknown = "B97-D/def2-TZVP/ISOSVP(PC) Unknown"
    B97_D_def2_TZVP_SMD_WATER__Single_Point = "B97-D/def2-TZVP/SMD(WATER) Single Point"
    B97_D_def2_TZVP_SMD_WATER__Force = "B97-D/def2-TZVP/SMD(WATER) Force"
    B97_D_def2_TZVP_SMD_WATER__Geometry_Optimization = "B97-D/def2-TZVP/SMD(WATER) Geometry Optimization"
    B97_D_def2_TZVP_SMD_WATER__Frequency_Analysis = "B97-D/def2-TZVP/SMD(WATER) Frequency Analysis"
    B97_D_def2_TZVP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_WATER__Unknown = "B97-D/def2-TZVP/SMD(WATER) Unknown"
    B97_D_def2_TZVP_SMD_THF__Single_Point = "B97-D/def2-TZVP/SMD(THF) Single Point"
    B97_D_def2_TZVP_SMD_THF__Force = "B97-D/def2-TZVP/SMD(THF) Force"
    B97_D_def2_TZVP_SMD_THF__Geometry_Optimization = "B97-D/def2-TZVP/SMD(THF) Geometry Optimization"
    B97_D_def2_TZVP_SMD_THF__Frequency_Analysis = "B97-D/def2-TZVP/SMD(THF) Frequency Analysis"
    B97_D_def2_TZVP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_THF__Unknown = "B97-D/def2-TZVP/SMD(THF) Unknown"
    B97_D_def2_TZVP_SMD_G2__Single_Point = "B97-D/def2-TZVP/SMD(G2) Single Point"
    B97_D_def2_TZVP_SMD_G2__Force = "B97-D/def2-TZVP/SMD(G2) Force"
    B97_D_def2_TZVP_SMD_G2__Geometry_Optimization = "B97-D/def2-TZVP/SMD(G2) Geometry Optimization"
    B97_D_def2_TZVP_SMD_G2__Frequency_Analysis = "B97-D/def2-TZVP/SMD(G2) Frequency Analysis"
    B97_D_def2_TZVP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_G2__Unknown = "B97-D/def2-TZVP/SMD(G2) Unknown"
    B97_D_def2_TZVP_SMD_DME__Single_Point = "B97-D/def2-TZVP/SMD(DME) Single Point"
    B97_D_def2_TZVP_SMD_DME__Force = "B97-D/def2-TZVP/SMD(DME) Force"
    B97_D_def2_TZVP_SMD_DME__Geometry_Optimization = "B97-D/def2-TZVP/SMD(DME) Geometry Optimization"
    B97_D_def2_TZVP_SMD_DME__Frequency_Analysis = "B97-D/def2-TZVP/SMD(DME) Frequency Analysis"
    B97_D_def2_TZVP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_DME__Unknown = "B97-D/def2-TZVP/SMD(DME) Unknown"
    B97_D_def2_TZVP_SMD_EC__Single_Point = "B97-D/def2-TZVP/SMD(EC) Single Point"
    B97_D_def2_TZVP_SMD_EC__Force = "B97-D/def2-TZVP/SMD(EC) Force"
    B97_D_def2_TZVP_SMD_EC__Geometry_Optimization = "B97-D/def2-TZVP/SMD(EC) Geometry Optimization"
    B97_D_def2_TZVP_SMD_EC__Frequency_Analysis = "B97-D/def2-TZVP/SMD(EC) Frequency Analysis"
    B97_D_def2_TZVP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_EC__Unknown = "B97-D/def2-TZVP/SMD(EC) Unknown"
    B97_D_def2_TZVP_SMD_EC_EMC__Single_Point = "B97-D/def2-TZVP/SMD(EC/EMC) Single Point"
    B97_D_def2_TZVP_SMD_EC_EMC__Force = "B97-D/def2-TZVP/SMD(EC/EMC) Force"
    B97_D_def2_TZVP_SMD_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVP/SMD(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVP_SMD_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVP/SMD(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_EC_EMC__Unknown = "B97-D/def2-TZVP/SMD(EC/EMC) Unknown"
    B97_D_def2_TZVP_SMD_PC__Single_Point = "B97-D/def2-TZVP/SMD(PC) Single Point"
    B97_D_def2_TZVP_SMD_PC__Force = "B97-D/def2-TZVP/SMD(PC) Force"
    B97_D_def2_TZVP_SMD_PC__Geometry_Optimization = "B97-D/def2-TZVP/SMD(PC) Geometry Optimization"
    B97_D_def2_TZVP_SMD_PC__Frequency_Analysis = "B97-D/def2-TZVP/SMD(PC) Frequency Analysis"
    B97_D_def2_TZVP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_SMD_PC__Unknown = "B97-D/def2-TZVP/SMD(PC) Unknown"
    B97_D_def2_TZVP_CMIRS_WATER__Single_Point = "B97-D/def2-TZVP/CMIRS(WATER) Single Point"
    B97_D_def2_TZVP_CMIRS_WATER__Force = "B97-D/def2-TZVP/CMIRS(WATER) Force"
    B97_D_def2_TZVP_CMIRS_WATER__Geometry_Optimization = "B97-D/def2-TZVP/CMIRS(WATER) Geometry Optimization"
    B97_D_def2_TZVP_CMIRS_WATER__Frequency_Analysis = "B97-D/def2-TZVP/CMIRS(WATER) Frequency Analysis"
    B97_D_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_WATER__Unknown = "B97-D/def2-TZVP/CMIRS(WATER) Unknown"
    B97_D_def2_TZVP_CMIRS_THF__Single_Point = "B97-D/def2-TZVP/CMIRS(THF) Single Point"
    B97_D_def2_TZVP_CMIRS_THF__Force = "B97-D/def2-TZVP/CMIRS(THF) Force"
    B97_D_def2_TZVP_CMIRS_THF__Geometry_Optimization = "B97-D/def2-TZVP/CMIRS(THF) Geometry Optimization"
    B97_D_def2_TZVP_CMIRS_THF__Frequency_Analysis = "B97-D/def2-TZVP/CMIRS(THF) Frequency Analysis"
    B97_D_def2_TZVP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_THF__Unknown = "B97-D/def2-TZVP/CMIRS(THF) Unknown"
    B97_D_def2_TZVP_CMIRS_G2__Single_Point = "B97-D/def2-TZVP/CMIRS(G2) Single Point"
    B97_D_def2_TZVP_CMIRS_G2__Force = "B97-D/def2-TZVP/CMIRS(G2) Force"
    B97_D_def2_TZVP_CMIRS_G2__Geometry_Optimization = "B97-D/def2-TZVP/CMIRS(G2) Geometry Optimization"
    B97_D_def2_TZVP_CMIRS_G2__Frequency_Analysis = "B97-D/def2-TZVP/CMIRS(G2) Frequency Analysis"
    B97_D_def2_TZVP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_G2__Unknown = "B97-D/def2-TZVP/CMIRS(G2) Unknown"
    B97_D_def2_TZVP_CMIRS_DME__Single_Point = "B97-D/def2-TZVP/CMIRS(DME) Single Point"
    B97_D_def2_TZVP_CMIRS_DME__Force = "B97-D/def2-TZVP/CMIRS(DME) Force"
    B97_D_def2_TZVP_CMIRS_DME__Geometry_Optimization = "B97-D/def2-TZVP/CMIRS(DME) Geometry Optimization"
    B97_D_def2_TZVP_CMIRS_DME__Frequency_Analysis = "B97-D/def2-TZVP/CMIRS(DME) Frequency Analysis"
    B97_D_def2_TZVP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_DME__Unknown = "B97-D/def2-TZVP/CMIRS(DME) Unknown"
    B97_D_def2_TZVP_CMIRS_EC__Single_Point = "B97-D/def2-TZVP/CMIRS(EC) Single Point"
    B97_D_def2_TZVP_CMIRS_EC__Force = "B97-D/def2-TZVP/CMIRS(EC) Force"
    B97_D_def2_TZVP_CMIRS_EC__Geometry_Optimization = "B97-D/def2-TZVP/CMIRS(EC) Geometry Optimization"
    B97_D_def2_TZVP_CMIRS_EC__Frequency_Analysis = "B97-D/def2-TZVP/CMIRS(EC) Frequency Analysis"
    B97_D_def2_TZVP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_EC__Unknown = "B97-D/def2-TZVP/CMIRS(EC) Unknown"
    B97_D_def2_TZVP_CMIRS_EC_EMC__Single_Point = "B97-D/def2-TZVP/CMIRS(EC/EMC) Single Point"
    B97_D_def2_TZVP_CMIRS_EC_EMC__Force = "B97-D/def2-TZVP/CMIRS(EC/EMC) Force"
    B97_D_def2_TZVP_CMIRS_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVP/CMIRS(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVP_CMIRS_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVP/CMIRS(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_EC_EMC__Unknown = "B97-D/def2-TZVP/CMIRS(EC/EMC) Unknown"
    B97_D_def2_TZVP_CMIRS_PC__Single_Point = "B97-D/def2-TZVP/CMIRS(PC) Single Point"
    B97_D_def2_TZVP_CMIRS_PC__Force = "B97-D/def2-TZVP/CMIRS(PC) Force"
    B97_D_def2_TZVP_CMIRS_PC__Geometry_Optimization = "B97-D/def2-TZVP/CMIRS(PC) Geometry Optimization"
    B97_D_def2_TZVP_CMIRS_PC__Frequency_Analysis = "B97-D/def2-TZVP/CMIRS(PC) Frequency Analysis"
    B97_D_def2_TZVP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVP_CMIRS_PC__Unknown = "B97-D/def2-TZVP/CMIRS(PC) Unknown"
    B97_D_def2_TZVPD_VACUUM_Single_Point = "B97-D/def2-TZVPD/VACUUM Single Point"
    B97_D_def2_TZVPD_VACUUM_Force = "B97-D/def2-TZVPD/VACUUM Force"
    B97_D_def2_TZVPD_VACUUM_Geometry_Optimization = "B97-D/def2-TZVPD/VACUUM Geometry Optimization"
    B97_D_def2_TZVPD_VACUUM_Frequency_Analysis = "B97-D/def2-TZVPD/VACUUM Frequency Analysis"
    B97_D_def2_TZVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/VACUUM Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_VACUUM_Unknown = "B97-D/def2-TZVPD/VACUUM Unknown"
    B97_D_def2_TZVPD_PCM_WATER__Single_Point = "B97-D/def2-TZVPD/PCM(WATER) Single Point"
    B97_D_def2_TZVPD_PCM_WATER__Force = "B97-D/def2-TZVPD/PCM(WATER) Force"
    B97_D_def2_TZVPD_PCM_WATER__Geometry_Optimization = "B97-D/def2-TZVPD/PCM(WATER) Geometry Optimization"
    B97_D_def2_TZVPD_PCM_WATER__Frequency_Analysis = "B97-D/def2-TZVPD/PCM(WATER) Frequency Analysis"
    B97_D_def2_TZVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_WATER__Unknown = "B97-D/def2-TZVPD/PCM(WATER) Unknown"
    B97_D_def2_TZVPD_PCM_THF__Single_Point = "B97-D/def2-TZVPD/PCM(THF) Single Point"
    B97_D_def2_TZVPD_PCM_THF__Force = "B97-D/def2-TZVPD/PCM(THF) Force"
    B97_D_def2_TZVPD_PCM_THF__Geometry_Optimization = "B97-D/def2-TZVPD/PCM(THF) Geometry Optimization"
    B97_D_def2_TZVPD_PCM_THF__Frequency_Analysis = "B97-D/def2-TZVPD/PCM(THF) Frequency Analysis"
    B97_D_def2_TZVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_THF__Unknown = "B97-D/def2-TZVPD/PCM(THF) Unknown"
    B97_D_def2_TZVPD_PCM_G2__Single_Point = "B97-D/def2-TZVPD/PCM(G2) Single Point"
    B97_D_def2_TZVPD_PCM_G2__Force = "B97-D/def2-TZVPD/PCM(G2) Force"
    B97_D_def2_TZVPD_PCM_G2__Geometry_Optimization = "B97-D/def2-TZVPD/PCM(G2) Geometry Optimization"
    B97_D_def2_TZVPD_PCM_G2__Frequency_Analysis = "B97-D/def2-TZVPD/PCM(G2) Frequency Analysis"
    B97_D_def2_TZVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_G2__Unknown = "B97-D/def2-TZVPD/PCM(G2) Unknown"
    B97_D_def2_TZVPD_PCM_DME__Single_Point = "B97-D/def2-TZVPD/PCM(DME) Single Point"
    B97_D_def2_TZVPD_PCM_DME__Force = "B97-D/def2-TZVPD/PCM(DME) Force"
    B97_D_def2_TZVPD_PCM_DME__Geometry_Optimization = "B97-D/def2-TZVPD/PCM(DME) Geometry Optimization"
    B97_D_def2_TZVPD_PCM_DME__Frequency_Analysis = "B97-D/def2-TZVPD/PCM(DME) Frequency Analysis"
    B97_D_def2_TZVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_DME__Unknown = "B97-D/def2-TZVPD/PCM(DME) Unknown"
    B97_D_def2_TZVPD_PCM_EC__Single_Point = "B97-D/def2-TZVPD/PCM(EC) Single Point"
    B97_D_def2_TZVPD_PCM_EC__Force = "B97-D/def2-TZVPD/PCM(EC) Force"
    B97_D_def2_TZVPD_PCM_EC__Geometry_Optimization = "B97-D/def2-TZVPD/PCM(EC) Geometry Optimization"
    B97_D_def2_TZVPD_PCM_EC__Frequency_Analysis = "B97-D/def2-TZVPD/PCM(EC) Frequency Analysis"
    B97_D_def2_TZVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_EC__Unknown = "B97-D/def2-TZVPD/PCM(EC) Unknown"
    B97_D_def2_TZVPD_PCM_EC_EMC__Single_Point = "B97-D/def2-TZVPD/PCM(EC/EMC) Single Point"
    B97_D_def2_TZVPD_PCM_EC_EMC__Force = "B97-D/def2-TZVPD/PCM(EC/EMC) Force"
    B97_D_def2_TZVPD_PCM_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPD/PCM(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPD_PCM_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPD/PCM(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_EC_EMC__Unknown = "B97-D/def2-TZVPD/PCM(EC/EMC) Unknown"
    B97_D_def2_TZVPD_PCM_PC__Single_Point = "B97-D/def2-TZVPD/PCM(PC) Single Point"
    B97_D_def2_TZVPD_PCM_PC__Force = "B97-D/def2-TZVPD/PCM(PC) Force"
    B97_D_def2_TZVPD_PCM_PC__Geometry_Optimization = "B97-D/def2-TZVPD/PCM(PC) Geometry Optimization"
    B97_D_def2_TZVPD_PCM_PC__Frequency_Analysis = "B97-D/def2-TZVPD/PCM(PC) Frequency Analysis"
    B97_D_def2_TZVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_PCM_PC__Unknown = "B97-D/def2-TZVPD/PCM(PC) Unknown"
    B97_D_def2_TZVPD_ISOSVP_WATER__Single_Point = "B97-D/def2-TZVPD/ISOSVP(WATER) Single Point"
    B97_D_def2_TZVPD_ISOSVP_WATER__Force = "B97-D/def2-TZVPD/ISOSVP(WATER) Force"
    B97_D_def2_TZVPD_ISOSVP_WATER__Geometry_Optimization = "B97-D/def2-TZVPD/ISOSVP(WATER) Geometry Optimization"
    B97_D_def2_TZVPD_ISOSVP_WATER__Frequency_Analysis = "B97-D/def2-TZVPD/ISOSVP(WATER) Frequency Analysis"
    B97_D_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_WATER__Unknown = "B97-D/def2-TZVPD/ISOSVP(WATER) Unknown"
    B97_D_def2_TZVPD_ISOSVP_THF__Single_Point = "B97-D/def2-TZVPD/ISOSVP(THF) Single Point"
    B97_D_def2_TZVPD_ISOSVP_THF__Force = "B97-D/def2-TZVPD/ISOSVP(THF) Force"
    B97_D_def2_TZVPD_ISOSVP_THF__Geometry_Optimization = "B97-D/def2-TZVPD/ISOSVP(THF) Geometry Optimization"
    B97_D_def2_TZVPD_ISOSVP_THF__Frequency_Analysis = "B97-D/def2-TZVPD/ISOSVP(THF) Frequency Analysis"
    B97_D_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_THF__Unknown = "B97-D/def2-TZVPD/ISOSVP(THF) Unknown"
    B97_D_def2_TZVPD_ISOSVP_G2__Single_Point = "B97-D/def2-TZVPD/ISOSVP(G2) Single Point"
    B97_D_def2_TZVPD_ISOSVP_G2__Force = "B97-D/def2-TZVPD/ISOSVP(G2) Force"
    B97_D_def2_TZVPD_ISOSVP_G2__Geometry_Optimization = "B97-D/def2-TZVPD/ISOSVP(G2) Geometry Optimization"
    B97_D_def2_TZVPD_ISOSVP_G2__Frequency_Analysis = "B97-D/def2-TZVPD/ISOSVP(G2) Frequency Analysis"
    B97_D_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_G2__Unknown = "B97-D/def2-TZVPD/ISOSVP(G2) Unknown"
    B97_D_def2_TZVPD_ISOSVP_DME__Single_Point = "B97-D/def2-TZVPD/ISOSVP(DME) Single Point"
    B97_D_def2_TZVPD_ISOSVP_DME__Force = "B97-D/def2-TZVPD/ISOSVP(DME) Force"
    B97_D_def2_TZVPD_ISOSVP_DME__Geometry_Optimization = "B97-D/def2-TZVPD/ISOSVP(DME) Geometry Optimization"
    B97_D_def2_TZVPD_ISOSVP_DME__Frequency_Analysis = "B97-D/def2-TZVPD/ISOSVP(DME) Frequency Analysis"
    B97_D_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_DME__Unknown = "B97-D/def2-TZVPD/ISOSVP(DME) Unknown"
    B97_D_def2_TZVPD_ISOSVP_EC__Single_Point = "B97-D/def2-TZVPD/ISOSVP(EC) Single Point"
    B97_D_def2_TZVPD_ISOSVP_EC__Force = "B97-D/def2-TZVPD/ISOSVP(EC) Force"
    B97_D_def2_TZVPD_ISOSVP_EC__Geometry_Optimization = "B97-D/def2-TZVPD/ISOSVP(EC) Geometry Optimization"
    B97_D_def2_TZVPD_ISOSVP_EC__Frequency_Analysis = "B97-D/def2-TZVPD/ISOSVP(EC) Frequency Analysis"
    B97_D_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_EC__Unknown = "B97-D/def2-TZVPD/ISOSVP(EC) Unknown"
    B97_D_def2_TZVPD_ISOSVP_EC_EMC__Single_Point = "B97-D/def2-TZVPD/ISOSVP(EC/EMC) Single Point"
    B97_D_def2_TZVPD_ISOSVP_EC_EMC__Force = "B97-D/def2-TZVPD/ISOSVP(EC/EMC) Force"
    B97_D_def2_TZVPD_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPD/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_EC_EMC__Unknown = "B97-D/def2-TZVPD/ISOSVP(EC/EMC) Unknown"
    B97_D_def2_TZVPD_ISOSVP_PC__Single_Point = "B97-D/def2-TZVPD/ISOSVP(PC) Single Point"
    B97_D_def2_TZVPD_ISOSVP_PC__Force = "B97-D/def2-TZVPD/ISOSVP(PC) Force"
    B97_D_def2_TZVPD_ISOSVP_PC__Geometry_Optimization = "B97-D/def2-TZVPD/ISOSVP(PC) Geometry Optimization"
    B97_D_def2_TZVPD_ISOSVP_PC__Frequency_Analysis = "B97-D/def2-TZVPD/ISOSVP(PC) Frequency Analysis"
    B97_D_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_ISOSVP_PC__Unknown = "B97-D/def2-TZVPD/ISOSVP(PC) Unknown"
    B97_D_def2_TZVPD_SMD_WATER__Single_Point = "B97-D/def2-TZVPD/SMD(WATER) Single Point"
    B97_D_def2_TZVPD_SMD_WATER__Force = "B97-D/def2-TZVPD/SMD(WATER) Force"
    B97_D_def2_TZVPD_SMD_WATER__Geometry_Optimization = "B97-D/def2-TZVPD/SMD(WATER) Geometry Optimization"
    B97_D_def2_TZVPD_SMD_WATER__Frequency_Analysis = "B97-D/def2-TZVPD/SMD(WATER) Frequency Analysis"
    B97_D_def2_TZVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_WATER__Unknown = "B97-D/def2-TZVPD/SMD(WATER) Unknown"
    B97_D_def2_TZVPD_SMD_THF__Single_Point = "B97-D/def2-TZVPD/SMD(THF) Single Point"
    B97_D_def2_TZVPD_SMD_THF__Force = "B97-D/def2-TZVPD/SMD(THF) Force"
    B97_D_def2_TZVPD_SMD_THF__Geometry_Optimization = "B97-D/def2-TZVPD/SMD(THF) Geometry Optimization"
    B97_D_def2_TZVPD_SMD_THF__Frequency_Analysis = "B97-D/def2-TZVPD/SMD(THF) Frequency Analysis"
    B97_D_def2_TZVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_THF__Unknown = "B97-D/def2-TZVPD/SMD(THF) Unknown"
    B97_D_def2_TZVPD_SMD_G2__Single_Point = "B97-D/def2-TZVPD/SMD(G2) Single Point"
    B97_D_def2_TZVPD_SMD_G2__Force = "B97-D/def2-TZVPD/SMD(G2) Force"
    B97_D_def2_TZVPD_SMD_G2__Geometry_Optimization = "B97-D/def2-TZVPD/SMD(G2) Geometry Optimization"
    B97_D_def2_TZVPD_SMD_G2__Frequency_Analysis = "B97-D/def2-TZVPD/SMD(G2) Frequency Analysis"
    B97_D_def2_TZVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_G2__Unknown = "B97-D/def2-TZVPD/SMD(G2) Unknown"
    B97_D_def2_TZVPD_SMD_DME__Single_Point = "B97-D/def2-TZVPD/SMD(DME) Single Point"
    B97_D_def2_TZVPD_SMD_DME__Force = "B97-D/def2-TZVPD/SMD(DME) Force"
    B97_D_def2_TZVPD_SMD_DME__Geometry_Optimization = "B97-D/def2-TZVPD/SMD(DME) Geometry Optimization"
    B97_D_def2_TZVPD_SMD_DME__Frequency_Analysis = "B97-D/def2-TZVPD/SMD(DME) Frequency Analysis"
    B97_D_def2_TZVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_DME__Unknown = "B97-D/def2-TZVPD/SMD(DME) Unknown"
    B97_D_def2_TZVPD_SMD_EC__Single_Point = "B97-D/def2-TZVPD/SMD(EC) Single Point"
    B97_D_def2_TZVPD_SMD_EC__Force = "B97-D/def2-TZVPD/SMD(EC) Force"
    B97_D_def2_TZVPD_SMD_EC__Geometry_Optimization = "B97-D/def2-TZVPD/SMD(EC) Geometry Optimization"
    B97_D_def2_TZVPD_SMD_EC__Frequency_Analysis = "B97-D/def2-TZVPD/SMD(EC) Frequency Analysis"
    B97_D_def2_TZVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_EC__Unknown = "B97-D/def2-TZVPD/SMD(EC) Unknown"
    B97_D_def2_TZVPD_SMD_EC_EMC__Single_Point = "B97-D/def2-TZVPD/SMD(EC/EMC) Single Point"
    B97_D_def2_TZVPD_SMD_EC_EMC__Force = "B97-D/def2-TZVPD/SMD(EC/EMC) Force"
    B97_D_def2_TZVPD_SMD_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPD/SMD(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPD_SMD_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPD/SMD(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_EC_EMC__Unknown = "B97-D/def2-TZVPD/SMD(EC/EMC) Unknown"
    B97_D_def2_TZVPD_SMD_PC__Single_Point = "B97-D/def2-TZVPD/SMD(PC) Single Point"
    B97_D_def2_TZVPD_SMD_PC__Force = "B97-D/def2-TZVPD/SMD(PC) Force"
    B97_D_def2_TZVPD_SMD_PC__Geometry_Optimization = "B97-D/def2-TZVPD/SMD(PC) Geometry Optimization"
    B97_D_def2_TZVPD_SMD_PC__Frequency_Analysis = "B97-D/def2-TZVPD/SMD(PC) Frequency Analysis"
    B97_D_def2_TZVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_SMD_PC__Unknown = "B97-D/def2-TZVPD/SMD(PC) Unknown"
    B97_D_def2_TZVPD_CMIRS_WATER__Single_Point = "B97-D/def2-TZVPD/CMIRS(WATER) Single Point"
    B97_D_def2_TZVPD_CMIRS_WATER__Force = "B97-D/def2-TZVPD/CMIRS(WATER) Force"
    B97_D_def2_TZVPD_CMIRS_WATER__Geometry_Optimization = "B97-D/def2-TZVPD/CMIRS(WATER) Geometry Optimization"
    B97_D_def2_TZVPD_CMIRS_WATER__Frequency_Analysis = "B97-D/def2-TZVPD/CMIRS(WATER) Frequency Analysis"
    B97_D_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_WATER__Unknown = "B97-D/def2-TZVPD/CMIRS(WATER) Unknown"
    B97_D_def2_TZVPD_CMIRS_THF__Single_Point = "B97-D/def2-TZVPD/CMIRS(THF) Single Point"
    B97_D_def2_TZVPD_CMIRS_THF__Force = "B97-D/def2-TZVPD/CMIRS(THF) Force"
    B97_D_def2_TZVPD_CMIRS_THF__Geometry_Optimization = "B97-D/def2-TZVPD/CMIRS(THF) Geometry Optimization"
    B97_D_def2_TZVPD_CMIRS_THF__Frequency_Analysis = "B97-D/def2-TZVPD/CMIRS(THF) Frequency Analysis"
    B97_D_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_THF__Unknown = "B97-D/def2-TZVPD/CMIRS(THF) Unknown"
    B97_D_def2_TZVPD_CMIRS_G2__Single_Point = "B97-D/def2-TZVPD/CMIRS(G2) Single Point"
    B97_D_def2_TZVPD_CMIRS_G2__Force = "B97-D/def2-TZVPD/CMIRS(G2) Force"
    B97_D_def2_TZVPD_CMIRS_G2__Geometry_Optimization = "B97-D/def2-TZVPD/CMIRS(G2) Geometry Optimization"
    B97_D_def2_TZVPD_CMIRS_G2__Frequency_Analysis = "B97-D/def2-TZVPD/CMIRS(G2) Frequency Analysis"
    B97_D_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_G2__Unknown = "B97-D/def2-TZVPD/CMIRS(G2) Unknown"
    B97_D_def2_TZVPD_CMIRS_DME__Single_Point = "B97-D/def2-TZVPD/CMIRS(DME) Single Point"
    B97_D_def2_TZVPD_CMIRS_DME__Force = "B97-D/def2-TZVPD/CMIRS(DME) Force"
    B97_D_def2_TZVPD_CMIRS_DME__Geometry_Optimization = "B97-D/def2-TZVPD/CMIRS(DME) Geometry Optimization"
    B97_D_def2_TZVPD_CMIRS_DME__Frequency_Analysis = "B97-D/def2-TZVPD/CMIRS(DME) Frequency Analysis"
    B97_D_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_DME__Unknown = "B97-D/def2-TZVPD/CMIRS(DME) Unknown"
    B97_D_def2_TZVPD_CMIRS_EC__Single_Point = "B97-D/def2-TZVPD/CMIRS(EC) Single Point"
    B97_D_def2_TZVPD_CMIRS_EC__Force = "B97-D/def2-TZVPD/CMIRS(EC) Force"
    B97_D_def2_TZVPD_CMIRS_EC__Geometry_Optimization = "B97-D/def2-TZVPD/CMIRS(EC) Geometry Optimization"
    B97_D_def2_TZVPD_CMIRS_EC__Frequency_Analysis = "B97-D/def2-TZVPD/CMIRS(EC) Frequency Analysis"
    B97_D_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_EC__Unknown = "B97-D/def2-TZVPD/CMIRS(EC) Unknown"
    B97_D_def2_TZVPD_CMIRS_EC_EMC__Single_Point = "B97-D/def2-TZVPD/CMIRS(EC/EMC) Single Point"
    B97_D_def2_TZVPD_CMIRS_EC_EMC__Force = "B97-D/def2-TZVPD/CMIRS(EC/EMC) Force"
    B97_D_def2_TZVPD_CMIRS_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPD/CMIRS(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPD_CMIRS_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPD/CMIRS(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_EC_EMC__Unknown = "B97-D/def2-TZVPD/CMIRS(EC/EMC) Unknown"
    B97_D_def2_TZVPD_CMIRS_PC__Single_Point = "B97-D/def2-TZVPD/CMIRS(PC) Single Point"
    B97_D_def2_TZVPD_CMIRS_PC__Force = "B97-D/def2-TZVPD/CMIRS(PC) Force"
    B97_D_def2_TZVPD_CMIRS_PC__Geometry_Optimization = "B97-D/def2-TZVPD/CMIRS(PC) Geometry Optimization"
    B97_D_def2_TZVPD_CMIRS_PC__Frequency_Analysis = "B97-D/def2-TZVPD/CMIRS(PC) Frequency Analysis"
    B97_D_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPD_CMIRS_PC__Unknown = "B97-D/def2-TZVPD/CMIRS(PC) Unknown"
    B97_D_def2_TZVPP_VACUUM_Single_Point = "B97-D/def2-TZVPP/VACUUM Single Point"
    B97_D_def2_TZVPP_VACUUM_Force = "B97-D/def2-TZVPP/VACUUM Force"
    B97_D_def2_TZVPP_VACUUM_Geometry_Optimization = "B97-D/def2-TZVPP/VACUUM Geometry Optimization"
    B97_D_def2_TZVPP_VACUUM_Frequency_Analysis = "B97-D/def2-TZVPP/VACUUM Frequency Analysis"
    B97_D_def2_TZVPP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/VACUUM Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_VACUUM_Unknown = "B97-D/def2-TZVPP/VACUUM Unknown"
    B97_D_def2_TZVPP_PCM_WATER__Single_Point = "B97-D/def2-TZVPP/PCM(WATER) Single Point"
    B97_D_def2_TZVPP_PCM_WATER__Force = "B97-D/def2-TZVPP/PCM(WATER) Force"
    B97_D_def2_TZVPP_PCM_WATER__Geometry_Optimization = "B97-D/def2-TZVPP/PCM(WATER) Geometry Optimization"
    B97_D_def2_TZVPP_PCM_WATER__Frequency_Analysis = "B97-D/def2-TZVPP/PCM(WATER) Frequency Analysis"
    B97_D_def2_TZVPP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_WATER__Unknown = "B97-D/def2-TZVPP/PCM(WATER) Unknown"
    B97_D_def2_TZVPP_PCM_THF__Single_Point = "B97-D/def2-TZVPP/PCM(THF) Single Point"
    B97_D_def2_TZVPP_PCM_THF__Force = "B97-D/def2-TZVPP/PCM(THF) Force"
    B97_D_def2_TZVPP_PCM_THF__Geometry_Optimization = "B97-D/def2-TZVPP/PCM(THF) Geometry Optimization"
    B97_D_def2_TZVPP_PCM_THF__Frequency_Analysis = "B97-D/def2-TZVPP/PCM(THF) Frequency Analysis"
    B97_D_def2_TZVPP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_THF__Unknown = "B97-D/def2-TZVPP/PCM(THF) Unknown"
    B97_D_def2_TZVPP_PCM_G2__Single_Point = "B97-D/def2-TZVPP/PCM(G2) Single Point"
    B97_D_def2_TZVPP_PCM_G2__Force = "B97-D/def2-TZVPP/PCM(G2) Force"
    B97_D_def2_TZVPP_PCM_G2__Geometry_Optimization = "B97-D/def2-TZVPP/PCM(G2) Geometry Optimization"
    B97_D_def2_TZVPP_PCM_G2__Frequency_Analysis = "B97-D/def2-TZVPP/PCM(G2) Frequency Analysis"
    B97_D_def2_TZVPP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_G2__Unknown = "B97-D/def2-TZVPP/PCM(G2) Unknown"
    B97_D_def2_TZVPP_PCM_DME__Single_Point = "B97-D/def2-TZVPP/PCM(DME) Single Point"
    B97_D_def2_TZVPP_PCM_DME__Force = "B97-D/def2-TZVPP/PCM(DME) Force"
    B97_D_def2_TZVPP_PCM_DME__Geometry_Optimization = "B97-D/def2-TZVPP/PCM(DME) Geometry Optimization"
    B97_D_def2_TZVPP_PCM_DME__Frequency_Analysis = "B97-D/def2-TZVPP/PCM(DME) Frequency Analysis"
    B97_D_def2_TZVPP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_DME__Unknown = "B97-D/def2-TZVPP/PCM(DME) Unknown"
    B97_D_def2_TZVPP_PCM_EC__Single_Point = "B97-D/def2-TZVPP/PCM(EC) Single Point"
    B97_D_def2_TZVPP_PCM_EC__Force = "B97-D/def2-TZVPP/PCM(EC) Force"
    B97_D_def2_TZVPP_PCM_EC__Geometry_Optimization = "B97-D/def2-TZVPP/PCM(EC) Geometry Optimization"
    B97_D_def2_TZVPP_PCM_EC__Frequency_Analysis = "B97-D/def2-TZVPP/PCM(EC) Frequency Analysis"
    B97_D_def2_TZVPP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_EC__Unknown = "B97-D/def2-TZVPP/PCM(EC) Unknown"
    B97_D_def2_TZVPP_PCM_EC_EMC__Single_Point = "B97-D/def2-TZVPP/PCM(EC/EMC) Single Point"
    B97_D_def2_TZVPP_PCM_EC_EMC__Force = "B97-D/def2-TZVPP/PCM(EC/EMC) Force"
    B97_D_def2_TZVPP_PCM_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPP/PCM(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPP_PCM_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPP/PCM(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_EC_EMC__Unknown = "B97-D/def2-TZVPP/PCM(EC/EMC) Unknown"
    B97_D_def2_TZVPP_PCM_PC__Single_Point = "B97-D/def2-TZVPP/PCM(PC) Single Point"
    B97_D_def2_TZVPP_PCM_PC__Force = "B97-D/def2-TZVPP/PCM(PC) Force"
    B97_D_def2_TZVPP_PCM_PC__Geometry_Optimization = "B97-D/def2-TZVPP/PCM(PC) Geometry Optimization"
    B97_D_def2_TZVPP_PCM_PC__Frequency_Analysis = "B97-D/def2-TZVPP/PCM(PC) Frequency Analysis"
    B97_D_def2_TZVPP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_PCM_PC__Unknown = "B97-D/def2-TZVPP/PCM(PC) Unknown"
    B97_D_def2_TZVPP_ISOSVP_WATER__Single_Point = "B97-D/def2-TZVPP/ISOSVP(WATER) Single Point"
    B97_D_def2_TZVPP_ISOSVP_WATER__Force = "B97-D/def2-TZVPP/ISOSVP(WATER) Force"
    B97_D_def2_TZVPP_ISOSVP_WATER__Geometry_Optimization = "B97-D/def2-TZVPP/ISOSVP(WATER) Geometry Optimization"
    B97_D_def2_TZVPP_ISOSVP_WATER__Frequency_Analysis = "B97-D/def2-TZVPP/ISOSVP(WATER) Frequency Analysis"
    B97_D_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_WATER__Unknown = "B97-D/def2-TZVPP/ISOSVP(WATER) Unknown"
    B97_D_def2_TZVPP_ISOSVP_THF__Single_Point = "B97-D/def2-TZVPP/ISOSVP(THF) Single Point"
    B97_D_def2_TZVPP_ISOSVP_THF__Force = "B97-D/def2-TZVPP/ISOSVP(THF) Force"
    B97_D_def2_TZVPP_ISOSVP_THF__Geometry_Optimization = "B97-D/def2-TZVPP/ISOSVP(THF) Geometry Optimization"
    B97_D_def2_TZVPP_ISOSVP_THF__Frequency_Analysis = "B97-D/def2-TZVPP/ISOSVP(THF) Frequency Analysis"
    B97_D_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_THF__Unknown = "B97-D/def2-TZVPP/ISOSVP(THF) Unknown"
    B97_D_def2_TZVPP_ISOSVP_G2__Single_Point = "B97-D/def2-TZVPP/ISOSVP(G2) Single Point"
    B97_D_def2_TZVPP_ISOSVP_G2__Force = "B97-D/def2-TZVPP/ISOSVP(G2) Force"
    B97_D_def2_TZVPP_ISOSVP_G2__Geometry_Optimization = "B97-D/def2-TZVPP/ISOSVP(G2) Geometry Optimization"
    B97_D_def2_TZVPP_ISOSVP_G2__Frequency_Analysis = "B97-D/def2-TZVPP/ISOSVP(G2) Frequency Analysis"
    B97_D_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_G2__Unknown = "B97-D/def2-TZVPP/ISOSVP(G2) Unknown"
    B97_D_def2_TZVPP_ISOSVP_DME__Single_Point = "B97-D/def2-TZVPP/ISOSVP(DME) Single Point"
    B97_D_def2_TZVPP_ISOSVP_DME__Force = "B97-D/def2-TZVPP/ISOSVP(DME) Force"
    B97_D_def2_TZVPP_ISOSVP_DME__Geometry_Optimization = "B97-D/def2-TZVPP/ISOSVP(DME) Geometry Optimization"
    B97_D_def2_TZVPP_ISOSVP_DME__Frequency_Analysis = "B97-D/def2-TZVPP/ISOSVP(DME) Frequency Analysis"
    B97_D_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_DME__Unknown = "B97-D/def2-TZVPP/ISOSVP(DME) Unknown"
    B97_D_def2_TZVPP_ISOSVP_EC__Single_Point = "B97-D/def2-TZVPP/ISOSVP(EC) Single Point"
    B97_D_def2_TZVPP_ISOSVP_EC__Force = "B97-D/def2-TZVPP/ISOSVP(EC) Force"
    B97_D_def2_TZVPP_ISOSVP_EC__Geometry_Optimization = "B97-D/def2-TZVPP/ISOSVP(EC) Geometry Optimization"
    B97_D_def2_TZVPP_ISOSVP_EC__Frequency_Analysis = "B97-D/def2-TZVPP/ISOSVP(EC) Frequency Analysis"
    B97_D_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_EC__Unknown = "B97-D/def2-TZVPP/ISOSVP(EC) Unknown"
    B97_D_def2_TZVPP_ISOSVP_EC_EMC__Single_Point = "B97-D/def2-TZVPP/ISOSVP(EC/EMC) Single Point"
    B97_D_def2_TZVPP_ISOSVP_EC_EMC__Force = "B97-D/def2-TZVPP/ISOSVP(EC/EMC) Force"
    B97_D_def2_TZVPP_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPP/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPP/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_EC_EMC__Unknown = "B97-D/def2-TZVPP/ISOSVP(EC/EMC) Unknown"
    B97_D_def2_TZVPP_ISOSVP_PC__Single_Point = "B97-D/def2-TZVPP/ISOSVP(PC) Single Point"
    B97_D_def2_TZVPP_ISOSVP_PC__Force = "B97-D/def2-TZVPP/ISOSVP(PC) Force"
    B97_D_def2_TZVPP_ISOSVP_PC__Geometry_Optimization = "B97-D/def2-TZVPP/ISOSVP(PC) Geometry Optimization"
    B97_D_def2_TZVPP_ISOSVP_PC__Frequency_Analysis = "B97-D/def2-TZVPP/ISOSVP(PC) Frequency Analysis"
    B97_D_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_ISOSVP_PC__Unknown = "B97-D/def2-TZVPP/ISOSVP(PC) Unknown"
    B97_D_def2_TZVPP_SMD_WATER__Single_Point = "B97-D/def2-TZVPP/SMD(WATER) Single Point"
    B97_D_def2_TZVPP_SMD_WATER__Force = "B97-D/def2-TZVPP/SMD(WATER) Force"
    B97_D_def2_TZVPP_SMD_WATER__Geometry_Optimization = "B97-D/def2-TZVPP/SMD(WATER) Geometry Optimization"
    B97_D_def2_TZVPP_SMD_WATER__Frequency_Analysis = "B97-D/def2-TZVPP/SMD(WATER) Frequency Analysis"
    B97_D_def2_TZVPP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_WATER__Unknown = "B97-D/def2-TZVPP/SMD(WATER) Unknown"
    B97_D_def2_TZVPP_SMD_THF__Single_Point = "B97-D/def2-TZVPP/SMD(THF) Single Point"
    B97_D_def2_TZVPP_SMD_THF__Force = "B97-D/def2-TZVPP/SMD(THF) Force"
    B97_D_def2_TZVPP_SMD_THF__Geometry_Optimization = "B97-D/def2-TZVPP/SMD(THF) Geometry Optimization"
    B97_D_def2_TZVPP_SMD_THF__Frequency_Analysis = "B97-D/def2-TZVPP/SMD(THF) Frequency Analysis"
    B97_D_def2_TZVPP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_THF__Unknown = "B97-D/def2-TZVPP/SMD(THF) Unknown"
    B97_D_def2_TZVPP_SMD_G2__Single_Point = "B97-D/def2-TZVPP/SMD(G2) Single Point"
    B97_D_def2_TZVPP_SMD_G2__Force = "B97-D/def2-TZVPP/SMD(G2) Force"
    B97_D_def2_TZVPP_SMD_G2__Geometry_Optimization = "B97-D/def2-TZVPP/SMD(G2) Geometry Optimization"
    B97_D_def2_TZVPP_SMD_G2__Frequency_Analysis = "B97-D/def2-TZVPP/SMD(G2) Frequency Analysis"
    B97_D_def2_TZVPP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_G2__Unknown = "B97-D/def2-TZVPP/SMD(G2) Unknown"
    B97_D_def2_TZVPP_SMD_DME__Single_Point = "B97-D/def2-TZVPP/SMD(DME) Single Point"
    B97_D_def2_TZVPP_SMD_DME__Force = "B97-D/def2-TZVPP/SMD(DME) Force"
    B97_D_def2_TZVPP_SMD_DME__Geometry_Optimization = "B97-D/def2-TZVPP/SMD(DME) Geometry Optimization"
    B97_D_def2_TZVPP_SMD_DME__Frequency_Analysis = "B97-D/def2-TZVPP/SMD(DME) Frequency Analysis"
    B97_D_def2_TZVPP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_DME__Unknown = "B97-D/def2-TZVPP/SMD(DME) Unknown"
    B97_D_def2_TZVPP_SMD_EC__Single_Point = "B97-D/def2-TZVPP/SMD(EC) Single Point"
    B97_D_def2_TZVPP_SMD_EC__Force = "B97-D/def2-TZVPP/SMD(EC) Force"
    B97_D_def2_TZVPP_SMD_EC__Geometry_Optimization = "B97-D/def2-TZVPP/SMD(EC) Geometry Optimization"
    B97_D_def2_TZVPP_SMD_EC__Frequency_Analysis = "B97-D/def2-TZVPP/SMD(EC) Frequency Analysis"
    B97_D_def2_TZVPP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_EC__Unknown = "B97-D/def2-TZVPP/SMD(EC) Unknown"
    B97_D_def2_TZVPP_SMD_EC_EMC__Single_Point = "B97-D/def2-TZVPP/SMD(EC/EMC) Single Point"
    B97_D_def2_TZVPP_SMD_EC_EMC__Force = "B97-D/def2-TZVPP/SMD(EC/EMC) Force"
    B97_D_def2_TZVPP_SMD_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPP/SMD(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPP_SMD_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPP/SMD(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_EC_EMC__Unknown = "B97-D/def2-TZVPP/SMD(EC/EMC) Unknown"
    B97_D_def2_TZVPP_SMD_PC__Single_Point = "B97-D/def2-TZVPP/SMD(PC) Single Point"
    B97_D_def2_TZVPP_SMD_PC__Force = "B97-D/def2-TZVPP/SMD(PC) Force"
    B97_D_def2_TZVPP_SMD_PC__Geometry_Optimization = "B97-D/def2-TZVPP/SMD(PC) Geometry Optimization"
    B97_D_def2_TZVPP_SMD_PC__Frequency_Analysis = "B97-D/def2-TZVPP/SMD(PC) Frequency Analysis"
    B97_D_def2_TZVPP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_SMD_PC__Unknown = "B97-D/def2-TZVPP/SMD(PC) Unknown"
    B97_D_def2_TZVPP_CMIRS_WATER__Single_Point = "B97-D/def2-TZVPP/CMIRS(WATER) Single Point"
    B97_D_def2_TZVPP_CMIRS_WATER__Force = "B97-D/def2-TZVPP/CMIRS(WATER) Force"
    B97_D_def2_TZVPP_CMIRS_WATER__Geometry_Optimization = "B97-D/def2-TZVPP/CMIRS(WATER) Geometry Optimization"
    B97_D_def2_TZVPP_CMIRS_WATER__Frequency_Analysis = "B97-D/def2-TZVPP/CMIRS(WATER) Frequency Analysis"
    B97_D_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_WATER__Unknown = "B97-D/def2-TZVPP/CMIRS(WATER) Unknown"
    B97_D_def2_TZVPP_CMIRS_THF__Single_Point = "B97-D/def2-TZVPP/CMIRS(THF) Single Point"
    B97_D_def2_TZVPP_CMIRS_THF__Force = "B97-D/def2-TZVPP/CMIRS(THF) Force"
    B97_D_def2_TZVPP_CMIRS_THF__Geometry_Optimization = "B97-D/def2-TZVPP/CMIRS(THF) Geometry Optimization"
    B97_D_def2_TZVPP_CMIRS_THF__Frequency_Analysis = "B97-D/def2-TZVPP/CMIRS(THF) Frequency Analysis"
    B97_D_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_THF__Unknown = "B97-D/def2-TZVPP/CMIRS(THF) Unknown"
    B97_D_def2_TZVPP_CMIRS_G2__Single_Point = "B97-D/def2-TZVPP/CMIRS(G2) Single Point"
    B97_D_def2_TZVPP_CMIRS_G2__Force = "B97-D/def2-TZVPP/CMIRS(G2) Force"
    B97_D_def2_TZVPP_CMIRS_G2__Geometry_Optimization = "B97-D/def2-TZVPP/CMIRS(G2) Geometry Optimization"
    B97_D_def2_TZVPP_CMIRS_G2__Frequency_Analysis = "B97-D/def2-TZVPP/CMIRS(G2) Frequency Analysis"
    B97_D_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_G2__Unknown = "B97-D/def2-TZVPP/CMIRS(G2) Unknown"
    B97_D_def2_TZVPP_CMIRS_DME__Single_Point = "B97-D/def2-TZVPP/CMIRS(DME) Single Point"
    B97_D_def2_TZVPP_CMIRS_DME__Force = "B97-D/def2-TZVPP/CMIRS(DME) Force"
    B97_D_def2_TZVPP_CMIRS_DME__Geometry_Optimization = "B97-D/def2-TZVPP/CMIRS(DME) Geometry Optimization"
    B97_D_def2_TZVPP_CMIRS_DME__Frequency_Analysis = "B97-D/def2-TZVPP/CMIRS(DME) Frequency Analysis"
    B97_D_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_DME__Unknown = "B97-D/def2-TZVPP/CMIRS(DME) Unknown"
    B97_D_def2_TZVPP_CMIRS_EC__Single_Point = "B97-D/def2-TZVPP/CMIRS(EC) Single Point"
    B97_D_def2_TZVPP_CMIRS_EC__Force = "B97-D/def2-TZVPP/CMIRS(EC) Force"
    B97_D_def2_TZVPP_CMIRS_EC__Geometry_Optimization = "B97-D/def2-TZVPP/CMIRS(EC) Geometry Optimization"
    B97_D_def2_TZVPP_CMIRS_EC__Frequency_Analysis = "B97-D/def2-TZVPP/CMIRS(EC) Frequency Analysis"
    B97_D_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_EC__Unknown = "B97-D/def2-TZVPP/CMIRS(EC) Unknown"
    B97_D_def2_TZVPP_CMIRS_EC_EMC__Single_Point = "B97-D/def2-TZVPP/CMIRS(EC/EMC) Single Point"
    B97_D_def2_TZVPP_CMIRS_EC_EMC__Force = "B97-D/def2-TZVPP/CMIRS(EC/EMC) Force"
    B97_D_def2_TZVPP_CMIRS_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPP/CMIRS(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPP_CMIRS_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPP/CMIRS(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_EC_EMC__Unknown = "B97-D/def2-TZVPP/CMIRS(EC/EMC) Unknown"
    B97_D_def2_TZVPP_CMIRS_PC__Single_Point = "B97-D/def2-TZVPP/CMIRS(PC) Single Point"
    B97_D_def2_TZVPP_CMIRS_PC__Force = "B97-D/def2-TZVPP/CMIRS(PC) Force"
    B97_D_def2_TZVPP_CMIRS_PC__Geometry_Optimization = "B97-D/def2-TZVPP/CMIRS(PC) Geometry Optimization"
    B97_D_def2_TZVPP_CMIRS_PC__Frequency_Analysis = "B97-D/def2-TZVPP/CMIRS(PC) Frequency Analysis"
    B97_D_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPP_CMIRS_PC__Unknown = "B97-D/def2-TZVPP/CMIRS(PC) Unknown"
    B97_D_def2_TZVPPD_VACUUM_Single_Point = "B97-D/def2-TZVPPD/VACUUM Single Point"
    B97_D_def2_TZVPPD_VACUUM_Force = "B97-D/def2-TZVPPD/VACUUM Force"
    B97_D_def2_TZVPPD_VACUUM_Geometry_Optimization = "B97-D/def2-TZVPPD/VACUUM Geometry Optimization"
    B97_D_def2_TZVPPD_VACUUM_Frequency_Analysis = "B97-D/def2-TZVPPD/VACUUM Frequency Analysis"
    B97_D_def2_TZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/VACUUM Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_VACUUM_Unknown = "B97-D/def2-TZVPPD/VACUUM Unknown"
    B97_D_def2_TZVPPD_PCM_WATER__Single_Point = "B97-D/def2-TZVPPD/PCM(WATER) Single Point"
    B97_D_def2_TZVPPD_PCM_WATER__Force = "B97-D/def2-TZVPPD/PCM(WATER) Force"
    B97_D_def2_TZVPPD_PCM_WATER__Geometry_Optimization = "B97-D/def2-TZVPPD/PCM(WATER) Geometry Optimization"
    B97_D_def2_TZVPPD_PCM_WATER__Frequency_Analysis = "B97-D/def2-TZVPPD/PCM(WATER) Frequency Analysis"
    B97_D_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_WATER__Unknown = "B97-D/def2-TZVPPD/PCM(WATER) Unknown"
    B97_D_def2_TZVPPD_PCM_THF__Single_Point = "B97-D/def2-TZVPPD/PCM(THF) Single Point"
    B97_D_def2_TZVPPD_PCM_THF__Force = "B97-D/def2-TZVPPD/PCM(THF) Force"
    B97_D_def2_TZVPPD_PCM_THF__Geometry_Optimization = "B97-D/def2-TZVPPD/PCM(THF) Geometry Optimization"
    B97_D_def2_TZVPPD_PCM_THF__Frequency_Analysis = "B97-D/def2-TZVPPD/PCM(THF) Frequency Analysis"
    B97_D_def2_TZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_THF__Unknown = "B97-D/def2-TZVPPD/PCM(THF) Unknown"
    B97_D_def2_TZVPPD_PCM_G2__Single_Point = "B97-D/def2-TZVPPD/PCM(G2) Single Point"
    B97_D_def2_TZVPPD_PCM_G2__Force = "B97-D/def2-TZVPPD/PCM(G2) Force"
    B97_D_def2_TZVPPD_PCM_G2__Geometry_Optimization = "B97-D/def2-TZVPPD/PCM(G2) Geometry Optimization"
    B97_D_def2_TZVPPD_PCM_G2__Frequency_Analysis = "B97-D/def2-TZVPPD/PCM(G2) Frequency Analysis"
    B97_D_def2_TZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_G2__Unknown = "B97-D/def2-TZVPPD/PCM(G2) Unknown"
    B97_D_def2_TZVPPD_PCM_DME__Single_Point = "B97-D/def2-TZVPPD/PCM(DME) Single Point"
    B97_D_def2_TZVPPD_PCM_DME__Force = "B97-D/def2-TZVPPD/PCM(DME) Force"
    B97_D_def2_TZVPPD_PCM_DME__Geometry_Optimization = "B97-D/def2-TZVPPD/PCM(DME) Geometry Optimization"
    B97_D_def2_TZVPPD_PCM_DME__Frequency_Analysis = "B97-D/def2-TZVPPD/PCM(DME) Frequency Analysis"
    B97_D_def2_TZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_DME__Unknown = "B97-D/def2-TZVPPD/PCM(DME) Unknown"
    B97_D_def2_TZVPPD_PCM_EC__Single_Point = "B97-D/def2-TZVPPD/PCM(EC) Single Point"
    B97_D_def2_TZVPPD_PCM_EC__Force = "B97-D/def2-TZVPPD/PCM(EC) Force"
    B97_D_def2_TZVPPD_PCM_EC__Geometry_Optimization = "B97-D/def2-TZVPPD/PCM(EC) Geometry Optimization"
    B97_D_def2_TZVPPD_PCM_EC__Frequency_Analysis = "B97-D/def2-TZVPPD/PCM(EC) Frequency Analysis"
    B97_D_def2_TZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_EC__Unknown = "B97-D/def2-TZVPPD/PCM(EC) Unknown"
    B97_D_def2_TZVPPD_PCM_EC_EMC__Single_Point = "B97-D/def2-TZVPPD/PCM(EC/EMC) Single Point"
    B97_D_def2_TZVPPD_PCM_EC_EMC__Force = "B97-D/def2-TZVPPD/PCM(EC/EMC) Force"
    B97_D_def2_TZVPPD_PCM_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPPD/PCM(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPPD_PCM_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPPD/PCM(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_EC_EMC__Unknown = "B97-D/def2-TZVPPD/PCM(EC/EMC) Unknown"
    B97_D_def2_TZVPPD_PCM_PC__Single_Point = "B97-D/def2-TZVPPD/PCM(PC) Single Point"
    B97_D_def2_TZVPPD_PCM_PC__Force = "B97-D/def2-TZVPPD/PCM(PC) Force"
    B97_D_def2_TZVPPD_PCM_PC__Geometry_Optimization = "B97-D/def2-TZVPPD/PCM(PC) Geometry Optimization"
    B97_D_def2_TZVPPD_PCM_PC__Frequency_Analysis = "B97-D/def2-TZVPPD/PCM(PC) Frequency Analysis"
    B97_D_def2_TZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_PCM_PC__Unknown = "B97-D/def2-TZVPPD/PCM(PC) Unknown"
    B97_D_def2_TZVPPD_ISOSVP_WATER__Single_Point = "B97-D/def2-TZVPPD/ISOSVP(WATER) Single Point"
    B97_D_def2_TZVPPD_ISOSVP_WATER__Force = "B97-D/def2-TZVPPD/ISOSVP(WATER) Force"
    B97_D_def2_TZVPPD_ISOSVP_WATER__Geometry_Optimization = "B97-D/def2-TZVPPD/ISOSVP(WATER) Geometry Optimization"
    B97_D_def2_TZVPPD_ISOSVP_WATER__Frequency_Analysis = "B97-D/def2-TZVPPD/ISOSVP(WATER) Frequency Analysis"
    B97_D_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_WATER__Unknown = "B97-D/def2-TZVPPD/ISOSVP(WATER) Unknown"
    B97_D_def2_TZVPPD_ISOSVP_THF__Single_Point = "B97-D/def2-TZVPPD/ISOSVP(THF) Single Point"
    B97_D_def2_TZVPPD_ISOSVP_THF__Force = "B97-D/def2-TZVPPD/ISOSVP(THF) Force"
    B97_D_def2_TZVPPD_ISOSVP_THF__Geometry_Optimization = "B97-D/def2-TZVPPD/ISOSVP(THF) Geometry Optimization"
    B97_D_def2_TZVPPD_ISOSVP_THF__Frequency_Analysis = "B97-D/def2-TZVPPD/ISOSVP(THF) Frequency Analysis"
    B97_D_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_THF__Unknown = "B97-D/def2-TZVPPD/ISOSVP(THF) Unknown"
    B97_D_def2_TZVPPD_ISOSVP_G2__Single_Point = "B97-D/def2-TZVPPD/ISOSVP(G2) Single Point"
    B97_D_def2_TZVPPD_ISOSVP_G2__Force = "B97-D/def2-TZVPPD/ISOSVP(G2) Force"
    B97_D_def2_TZVPPD_ISOSVP_G2__Geometry_Optimization = "B97-D/def2-TZVPPD/ISOSVP(G2) Geometry Optimization"
    B97_D_def2_TZVPPD_ISOSVP_G2__Frequency_Analysis = "B97-D/def2-TZVPPD/ISOSVP(G2) Frequency Analysis"
    B97_D_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_G2__Unknown = "B97-D/def2-TZVPPD/ISOSVP(G2) Unknown"
    B97_D_def2_TZVPPD_ISOSVP_DME__Single_Point = "B97-D/def2-TZVPPD/ISOSVP(DME) Single Point"
    B97_D_def2_TZVPPD_ISOSVP_DME__Force = "B97-D/def2-TZVPPD/ISOSVP(DME) Force"
    B97_D_def2_TZVPPD_ISOSVP_DME__Geometry_Optimization = "B97-D/def2-TZVPPD/ISOSVP(DME) Geometry Optimization"
    B97_D_def2_TZVPPD_ISOSVP_DME__Frequency_Analysis = "B97-D/def2-TZVPPD/ISOSVP(DME) Frequency Analysis"
    B97_D_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_DME__Unknown = "B97-D/def2-TZVPPD/ISOSVP(DME) Unknown"
    B97_D_def2_TZVPPD_ISOSVP_EC__Single_Point = "B97-D/def2-TZVPPD/ISOSVP(EC) Single Point"
    B97_D_def2_TZVPPD_ISOSVP_EC__Force = "B97-D/def2-TZVPPD/ISOSVP(EC) Force"
    B97_D_def2_TZVPPD_ISOSVP_EC__Geometry_Optimization = "B97-D/def2-TZVPPD/ISOSVP(EC) Geometry Optimization"
    B97_D_def2_TZVPPD_ISOSVP_EC__Frequency_Analysis = "B97-D/def2-TZVPPD/ISOSVP(EC) Frequency Analysis"
    B97_D_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_EC__Unknown = "B97-D/def2-TZVPPD/ISOSVP(EC) Unknown"
    B97_D_def2_TZVPPD_ISOSVP_EC_EMC__Single_Point = "B97-D/def2-TZVPPD/ISOSVP(EC/EMC) Single Point"
    B97_D_def2_TZVPPD_ISOSVP_EC_EMC__Force = "B97-D/def2-TZVPPD/ISOSVP(EC/EMC) Force"
    B97_D_def2_TZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_EC_EMC__Unknown = "B97-D/def2-TZVPPD/ISOSVP(EC/EMC) Unknown"
    B97_D_def2_TZVPPD_ISOSVP_PC__Single_Point = "B97-D/def2-TZVPPD/ISOSVP(PC) Single Point"
    B97_D_def2_TZVPPD_ISOSVP_PC__Force = "B97-D/def2-TZVPPD/ISOSVP(PC) Force"
    B97_D_def2_TZVPPD_ISOSVP_PC__Geometry_Optimization = "B97-D/def2-TZVPPD/ISOSVP(PC) Geometry Optimization"
    B97_D_def2_TZVPPD_ISOSVP_PC__Frequency_Analysis = "B97-D/def2-TZVPPD/ISOSVP(PC) Frequency Analysis"
    B97_D_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_ISOSVP_PC__Unknown = "B97-D/def2-TZVPPD/ISOSVP(PC) Unknown"
    B97_D_def2_TZVPPD_SMD_WATER__Single_Point = "B97-D/def2-TZVPPD/SMD(WATER) Single Point"
    B97_D_def2_TZVPPD_SMD_WATER__Force = "B97-D/def2-TZVPPD/SMD(WATER) Force"
    B97_D_def2_TZVPPD_SMD_WATER__Geometry_Optimization = "B97-D/def2-TZVPPD/SMD(WATER) Geometry Optimization"
    B97_D_def2_TZVPPD_SMD_WATER__Frequency_Analysis = "B97-D/def2-TZVPPD/SMD(WATER) Frequency Analysis"
    B97_D_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_WATER__Unknown = "B97-D/def2-TZVPPD/SMD(WATER) Unknown"
    B97_D_def2_TZVPPD_SMD_THF__Single_Point = "B97-D/def2-TZVPPD/SMD(THF) Single Point"
    B97_D_def2_TZVPPD_SMD_THF__Force = "B97-D/def2-TZVPPD/SMD(THF) Force"
    B97_D_def2_TZVPPD_SMD_THF__Geometry_Optimization = "B97-D/def2-TZVPPD/SMD(THF) Geometry Optimization"
    B97_D_def2_TZVPPD_SMD_THF__Frequency_Analysis = "B97-D/def2-TZVPPD/SMD(THF) Frequency Analysis"
    B97_D_def2_TZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_THF__Unknown = "B97-D/def2-TZVPPD/SMD(THF) Unknown"
    B97_D_def2_TZVPPD_SMD_G2__Single_Point = "B97-D/def2-TZVPPD/SMD(G2) Single Point"
    B97_D_def2_TZVPPD_SMD_G2__Force = "B97-D/def2-TZVPPD/SMD(G2) Force"
    B97_D_def2_TZVPPD_SMD_G2__Geometry_Optimization = "B97-D/def2-TZVPPD/SMD(G2) Geometry Optimization"
    B97_D_def2_TZVPPD_SMD_G2__Frequency_Analysis = "B97-D/def2-TZVPPD/SMD(G2) Frequency Analysis"
    B97_D_def2_TZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_G2__Unknown = "B97-D/def2-TZVPPD/SMD(G2) Unknown"
    B97_D_def2_TZVPPD_SMD_DME__Single_Point = "B97-D/def2-TZVPPD/SMD(DME) Single Point"
    B97_D_def2_TZVPPD_SMD_DME__Force = "B97-D/def2-TZVPPD/SMD(DME) Force"
    B97_D_def2_TZVPPD_SMD_DME__Geometry_Optimization = "B97-D/def2-TZVPPD/SMD(DME) Geometry Optimization"
    B97_D_def2_TZVPPD_SMD_DME__Frequency_Analysis = "B97-D/def2-TZVPPD/SMD(DME) Frequency Analysis"
    B97_D_def2_TZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_DME__Unknown = "B97-D/def2-TZVPPD/SMD(DME) Unknown"
    B97_D_def2_TZVPPD_SMD_EC__Single_Point = "B97-D/def2-TZVPPD/SMD(EC) Single Point"
    B97_D_def2_TZVPPD_SMD_EC__Force = "B97-D/def2-TZVPPD/SMD(EC) Force"
    B97_D_def2_TZVPPD_SMD_EC__Geometry_Optimization = "B97-D/def2-TZVPPD/SMD(EC) Geometry Optimization"
    B97_D_def2_TZVPPD_SMD_EC__Frequency_Analysis = "B97-D/def2-TZVPPD/SMD(EC) Frequency Analysis"
    B97_D_def2_TZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_EC__Unknown = "B97-D/def2-TZVPPD/SMD(EC) Unknown"
    B97_D_def2_TZVPPD_SMD_EC_EMC__Single_Point = "B97-D/def2-TZVPPD/SMD(EC/EMC) Single Point"
    B97_D_def2_TZVPPD_SMD_EC_EMC__Force = "B97-D/def2-TZVPPD/SMD(EC/EMC) Force"
    B97_D_def2_TZVPPD_SMD_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPPD/SMD(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPPD_SMD_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPPD/SMD(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_EC_EMC__Unknown = "B97-D/def2-TZVPPD/SMD(EC/EMC) Unknown"
    B97_D_def2_TZVPPD_SMD_PC__Single_Point = "B97-D/def2-TZVPPD/SMD(PC) Single Point"
    B97_D_def2_TZVPPD_SMD_PC__Force = "B97-D/def2-TZVPPD/SMD(PC) Force"
    B97_D_def2_TZVPPD_SMD_PC__Geometry_Optimization = "B97-D/def2-TZVPPD/SMD(PC) Geometry Optimization"
    B97_D_def2_TZVPPD_SMD_PC__Frequency_Analysis = "B97-D/def2-TZVPPD/SMD(PC) Frequency Analysis"
    B97_D_def2_TZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_SMD_PC__Unknown = "B97-D/def2-TZVPPD/SMD(PC) Unknown"
    B97_D_def2_TZVPPD_CMIRS_WATER__Single_Point = "B97-D/def2-TZVPPD/CMIRS(WATER) Single Point"
    B97_D_def2_TZVPPD_CMIRS_WATER__Force = "B97-D/def2-TZVPPD/CMIRS(WATER) Force"
    B97_D_def2_TZVPPD_CMIRS_WATER__Geometry_Optimization = "B97-D/def2-TZVPPD/CMIRS(WATER) Geometry Optimization"
    B97_D_def2_TZVPPD_CMIRS_WATER__Frequency_Analysis = "B97-D/def2-TZVPPD/CMIRS(WATER) Frequency Analysis"
    B97_D_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_WATER__Unknown = "B97-D/def2-TZVPPD/CMIRS(WATER) Unknown"
    B97_D_def2_TZVPPD_CMIRS_THF__Single_Point = "B97-D/def2-TZVPPD/CMIRS(THF) Single Point"
    B97_D_def2_TZVPPD_CMIRS_THF__Force = "B97-D/def2-TZVPPD/CMIRS(THF) Force"
    B97_D_def2_TZVPPD_CMIRS_THF__Geometry_Optimization = "B97-D/def2-TZVPPD/CMIRS(THF) Geometry Optimization"
    B97_D_def2_TZVPPD_CMIRS_THF__Frequency_Analysis = "B97-D/def2-TZVPPD/CMIRS(THF) Frequency Analysis"
    B97_D_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_THF__Unknown = "B97-D/def2-TZVPPD/CMIRS(THF) Unknown"
    B97_D_def2_TZVPPD_CMIRS_G2__Single_Point = "B97-D/def2-TZVPPD/CMIRS(G2) Single Point"
    B97_D_def2_TZVPPD_CMIRS_G2__Force = "B97-D/def2-TZVPPD/CMIRS(G2) Force"
    B97_D_def2_TZVPPD_CMIRS_G2__Geometry_Optimization = "B97-D/def2-TZVPPD/CMIRS(G2) Geometry Optimization"
    B97_D_def2_TZVPPD_CMIRS_G2__Frequency_Analysis = "B97-D/def2-TZVPPD/CMIRS(G2) Frequency Analysis"
    B97_D_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_G2__Unknown = "B97-D/def2-TZVPPD/CMIRS(G2) Unknown"
    B97_D_def2_TZVPPD_CMIRS_DME__Single_Point = "B97-D/def2-TZVPPD/CMIRS(DME) Single Point"
    B97_D_def2_TZVPPD_CMIRS_DME__Force = "B97-D/def2-TZVPPD/CMIRS(DME) Force"
    B97_D_def2_TZVPPD_CMIRS_DME__Geometry_Optimization = "B97-D/def2-TZVPPD/CMIRS(DME) Geometry Optimization"
    B97_D_def2_TZVPPD_CMIRS_DME__Frequency_Analysis = "B97-D/def2-TZVPPD/CMIRS(DME) Frequency Analysis"
    B97_D_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_DME__Unknown = "B97-D/def2-TZVPPD/CMIRS(DME) Unknown"
    B97_D_def2_TZVPPD_CMIRS_EC__Single_Point = "B97-D/def2-TZVPPD/CMIRS(EC) Single Point"
    B97_D_def2_TZVPPD_CMIRS_EC__Force = "B97-D/def2-TZVPPD/CMIRS(EC) Force"
    B97_D_def2_TZVPPD_CMIRS_EC__Geometry_Optimization = "B97-D/def2-TZVPPD/CMIRS(EC) Geometry Optimization"
    B97_D_def2_TZVPPD_CMIRS_EC__Frequency_Analysis = "B97-D/def2-TZVPPD/CMIRS(EC) Frequency Analysis"
    B97_D_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_EC__Unknown = "B97-D/def2-TZVPPD/CMIRS(EC) Unknown"
    B97_D_def2_TZVPPD_CMIRS_EC_EMC__Single_Point = "B97-D/def2-TZVPPD/CMIRS(EC/EMC) Single Point"
    B97_D_def2_TZVPPD_CMIRS_EC_EMC__Force = "B97-D/def2-TZVPPD/CMIRS(EC/EMC) Force"
    B97_D_def2_TZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "B97-D/def2-TZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    B97_D_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "B97-D/def2-TZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    B97_D_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_EC_EMC__Unknown = "B97-D/def2-TZVPPD/CMIRS(EC/EMC) Unknown"
    B97_D_def2_TZVPPD_CMIRS_PC__Single_Point = "B97-D/def2-TZVPPD/CMIRS(PC) Single Point"
    B97_D_def2_TZVPPD_CMIRS_PC__Force = "B97-D/def2-TZVPPD/CMIRS(PC) Force"
    B97_D_def2_TZVPPD_CMIRS_PC__Geometry_Optimization = "B97-D/def2-TZVPPD/CMIRS(PC) Geometry Optimization"
    B97_D_def2_TZVPPD_CMIRS_PC__Frequency_Analysis = "B97-D/def2-TZVPPD/CMIRS(PC) Frequency Analysis"
    B97_D_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-TZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_TZVPPD_CMIRS_PC__Unknown = "B97-D/def2-TZVPPD/CMIRS(PC) Unknown"
    B97_D_def2_QZVPPD_VACUUM_Single_Point = "B97-D/def2-QZVPPD/VACUUM Single Point"
    B97_D_def2_QZVPPD_VACUUM_Force = "B97-D/def2-QZVPPD/VACUUM Force"
    B97_D_def2_QZVPPD_VACUUM_Geometry_Optimization = "B97-D/def2-QZVPPD/VACUUM Geometry Optimization"
    B97_D_def2_QZVPPD_VACUUM_Frequency_Analysis = "B97-D/def2-QZVPPD/VACUUM Frequency Analysis"
    B97_D_def2_QZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/VACUUM Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_VACUUM_Unknown = "B97-D/def2-QZVPPD/VACUUM Unknown"
    B97_D_def2_QZVPPD_PCM_WATER__Single_Point = "B97-D/def2-QZVPPD/PCM(WATER) Single Point"
    B97_D_def2_QZVPPD_PCM_WATER__Force = "B97-D/def2-QZVPPD/PCM(WATER) Force"
    B97_D_def2_QZVPPD_PCM_WATER__Geometry_Optimization = "B97-D/def2-QZVPPD/PCM(WATER) Geometry Optimization"
    B97_D_def2_QZVPPD_PCM_WATER__Frequency_Analysis = "B97-D/def2-QZVPPD/PCM(WATER) Frequency Analysis"
    B97_D_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_WATER__Unknown = "B97-D/def2-QZVPPD/PCM(WATER) Unknown"
    B97_D_def2_QZVPPD_PCM_THF__Single_Point = "B97-D/def2-QZVPPD/PCM(THF) Single Point"
    B97_D_def2_QZVPPD_PCM_THF__Force = "B97-D/def2-QZVPPD/PCM(THF) Force"
    B97_D_def2_QZVPPD_PCM_THF__Geometry_Optimization = "B97-D/def2-QZVPPD/PCM(THF) Geometry Optimization"
    B97_D_def2_QZVPPD_PCM_THF__Frequency_Analysis = "B97-D/def2-QZVPPD/PCM(THF) Frequency Analysis"
    B97_D_def2_QZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_THF__Unknown = "B97-D/def2-QZVPPD/PCM(THF) Unknown"
    B97_D_def2_QZVPPD_PCM_G2__Single_Point = "B97-D/def2-QZVPPD/PCM(G2) Single Point"
    B97_D_def2_QZVPPD_PCM_G2__Force = "B97-D/def2-QZVPPD/PCM(G2) Force"
    B97_D_def2_QZVPPD_PCM_G2__Geometry_Optimization = "B97-D/def2-QZVPPD/PCM(G2) Geometry Optimization"
    B97_D_def2_QZVPPD_PCM_G2__Frequency_Analysis = "B97-D/def2-QZVPPD/PCM(G2) Frequency Analysis"
    B97_D_def2_QZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_G2__Unknown = "B97-D/def2-QZVPPD/PCM(G2) Unknown"
    B97_D_def2_QZVPPD_PCM_DME__Single_Point = "B97-D/def2-QZVPPD/PCM(DME) Single Point"
    B97_D_def2_QZVPPD_PCM_DME__Force = "B97-D/def2-QZVPPD/PCM(DME) Force"
    B97_D_def2_QZVPPD_PCM_DME__Geometry_Optimization = "B97-D/def2-QZVPPD/PCM(DME) Geometry Optimization"
    B97_D_def2_QZVPPD_PCM_DME__Frequency_Analysis = "B97-D/def2-QZVPPD/PCM(DME) Frequency Analysis"
    B97_D_def2_QZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_DME__Unknown = "B97-D/def2-QZVPPD/PCM(DME) Unknown"
    B97_D_def2_QZVPPD_PCM_EC__Single_Point = "B97-D/def2-QZVPPD/PCM(EC) Single Point"
    B97_D_def2_QZVPPD_PCM_EC__Force = "B97-D/def2-QZVPPD/PCM(EC) Force"
    B97_D_def2_QZVPPD_PCM_EC__Geometry_Optimization = "B97-D/def2-QZVPPD/PCM(EC) Geometry Optimization"
    B97_D_def2_QZVPPD_PCM_EC__Frequency_Analysis = "B97-D/def2-QZVPPD/PCM(EC) Frequency Analysis"
    B97_D_def2_QZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_EC__Unknown = "B97-D/def2-QZVPPD/PCM(EC) Unknown"
    B97_D_def2_QZVPPD_PCM_EC_EMC__Single_Point = "B97-D/def2-QZVPPD/PCM(EC/EMC) Single Point"
    B97_D_def2_QZVPPD_PCM_EC_EMC__Force = "B97-D/def2-QZVPPD/PCM(EC/EMC) Force"
    B97_D_def2_QZVPPD_PCM_EC_EMC__Geometry_Optimization = "B97-D/def2-QZVPPD/PCM(EC/EMC) Geometry Optimization"
    B97_D_def2_QZVPPD_PCM_EC_EMC__Frequency_Analysis = "B97-D/def2-QZVPPD/PCM(EC/EMC) Frequency Analysis"
    B97_D_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_EC_EMC__Unknown = "B97-D/def2-QZVPPD/PCM(EC/EMC) Unknown"
    B97_D_def2_QZVPPD_PCM_PC__Single_Point = "B97-D/def2-QZVPPD/PCM(PC) Single Point"
    B97_D_def2_QZVPPD_PCM_PC__Force = "B97-D/def2-QZVPPD/PCM(PC) Force"
    B97_D_def2_QZVPPD_PCM_PC__Geometry_Optimization = "B97-D/def2-QZVPPD/PCM(PC) Geometry Optimization"
    B97_D_def2_QZVPPD_PCM_PC__Frequency_Analysis = "B97-D/def2-QZVPPD/PCM(PC) Frequency Analysis"
    B97_D_def2_QZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_PCM_PC__Unknown = "B97-D/def2-QZVPPD/PCM(PC) Unknown"
    B97_D_def2_QZVPPD_ISOSVP_WATER__Single_Point = "B97-D/def2-QZVPPD/ISOSVP(WATER) Single Point"
    B97_D_def2_QZVPPD_ISOSVP_WATER__Force = "B97-D/def2-QZVPPD/ISOSVP(WATER) Force"
    B97_D_def2_QZVPPD_ISOSVP_WATER__Geometry_Optimization = "B97-D/def2-QZVPPD/ISOSVP(WATER) Geometry Optimization"
    B97_D_def2_QZVPPD_ISOSVP_WATER__Frequency_Analysis = "B97-D/def2-QZVPPD/ISOSVP(WATER) Frequency Analysis"
    B97_D_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_WATER__Unknown = "B97-D/def2-QZVPPD/ISOSVP(WATER) Unknown"
    B97_D_def2_QZVPPD_ISOSVP_THF__Single_Point = "B97-D/def2-QZVPPD/ISOSVP(THF) Single Point"
    B97_D_def2_QZVPPD_ISOSVP_THF__Force = "B97-D/def2-QZVPPD/ISOSVP(THF) Force"
    B97_D_def2_QZVPPD_ISOSVP_THF__Geometry_Optimization = "B97-D/def2-QZVPPD/ISOSVP(THF) Geometry Optimization"
    B97_D_def2_QZVPPD_ISOSVP_THF__Frequency_Analysis = "B97-D/def2-QZVPPD/ISOSVP(THF) Frequency Analysis"
    B97_D_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_THF__Unknown = "B97-D/def2-QZVPPD/ISOSVP(THF) Unknown"
    B97_D_def2_QZVPPD_ISOSVP_G2__Single_Point = "B97-D/def2-QZVPPD/ISOSVP(G2) Single Point"
    B97_D_def2_QZVPPD_ISOSVP_G2__Force = "B97-D/def2-QZVPPD/ISOSVP(G2) Force"
    B97_D_def2_QZVPPD_ISOSVP_G2__Geometry_Optimization = "B97-D/def2-QZVPPD/ISOSVP(G2) Geometry Optimization"
    B97_D_def2_QZVPPD_ISOSVP_G2__Frequency_Analysis = "B97-D/def2-QZVPPD/ISOSVP(G2) Frequency Analysis"
    B97_D_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_G2__Unknown = "B97-D/def2-QZVPPD/ISOSVP(G2) Unknown"
    B97_D_def2_QZVPPD_ISOSVP_DME__Single_Point = "B97-D/def2-QZVPPD/ISOSVP(DME) Single Point"
    B97_D_def2_QZVPPD_ISOSVP_DME__Force = "B97-D/def2-QZVPPD/ISOSVP(DME) Force"
    B97_D_def2_QZVPPD_ISOSVP_DME__Geometry_Optimization = "B97-D/def2-QZVPPD/ISOSVP(DME) Geometry Optimization"
    B97_D_def2_QZVPPD_ISOSVP_DME__Frequency_Analysis = "B97-D/def2-QZVPPD/ISOSVP(DME) Frequency Analysis"
    B97_D_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_DME__Unknown = "B97-D/def2-QZVPPD/ISOSVP(DME) Unknown"
    B97_D_def2_QZVPPD_ISOSVP_EC__Single_Point = "B97-D/def2-QZVPPD/ISOSVP(EC) Single Point"
    B97_D_def2_QZVPPD_ISOSVP_EC__Force = "B97-D/def2-QZVPPD/ISOSVP(EC) Force"
    B97_D_def2_QZVPPD_ISOSVP_EC__Geometry_Optimization = "B97-D/def2-QZVPPD/ISOSVP(EC) Geometry Optimization"
    B97_D_def2_QZVPPD_ISOSVP_EC__Frequency_Analysis = "B97-D/def2-QZVPPD/ISOSVP(EC) Frequency Analysis"
    B97_D_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_EC__Unknown = "B97-D/def2-QZVPPD/ISOSVP(EC) Unknown"
    B97_D_def2_QZVPPD_ISOSVP_EC_EMC__Single_Point = "B97-D/def2-QZVPPD/ISOSVP(EC/EMC) Single Point"
    B97_D_def2_QZVPPD_ISOSVP_EC_EMC__Force = "B97-D/def2-QZVPPD/ISOSVP(EC/EMC) Force"
    B97_D_def2_QZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D/def2-QZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_EC_EMC__Unknown = "B97-D/def2-QZVPPD/ISOSVP(EC/EMC) Unknown"
    B97_D_def2_QZVPPD_ISOSVP_PC__Single_Point = "B97-D/def2-QZVPPD/ISOSVP(PC) Single Point"
    B97_D_def2_QZVPPD_ISOSVP_PC__Force = "B97-D/def2-QZVPPD/ISOSVP(PC) Force"
    B97_D_def2_QZVPPD_ISOSVP_PC__Geometry_Optimization = "B97-D/def2-QZVPPD/ISOSVP(PC) Geometry Optimization"
    B97_D_def2_QZVPPD_ISOSVP_PC__Frequency_Analysis = "B97-D/def2-QZVPPD/ISOSVP(PC) Frequency Analysis"
    B97_D_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_ISOSVP_PC__Unknown = "B97-D/def2-QZVPPD/ISOSVP(PC) Unknown"
    B97_D_def2_QZVPPD_SMD_WATER__Single_Point = "B97-D/def2-QZVPPD/SMD(WATER) Single Point"
    B97_D_def2_QZVPPD_SMD_WATER__Force = "B97-D/def2-QZVPPD/SMD(WATER) Force"
    B97_D_def2_QZVPPD_SMD_WATER__Geometry_Optimization = "B97-D/def2-QZVPPD/SMD(WATER) Geometry Optimization"
    B97_D_def2_QZVPPD_SMD_WATER__Frequency_Analysis = "B97-D/def2-QZVPPD/SMD(WATER) Frequency Analysis"
    B97_D_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_WATER__Unknown = "B97-D/def2-QZVPPD/SMD(WATER) Unknown"
    B97_D_def2_QZVPPD_SMD_THF__Single_Point = "B97-D/def2-QZVPPD/SMD(THF) Single Point"
    B97_D_def2_QZVPPD_SMD_THF__Force = "B97-D/def2-QZVPPD/SMD(THF) Force"
    B97_D_def2_QZVPPD_SMD_THF__Geometry_Optimization = "B97-D/def2-QZVPPD/SMD(THF) Geometry Optimization"
    B97_D_def2_QZVPPD_SMD_THF__Frequency_Analysis = "B97-D/def2-QZVPPD/SMD(THF) Frequency Analysis"
    B97_D_def2_QZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_THF__Unknown = "B97-D/def2-QZVPPD/SMD(THF) Unknown"
    B97_D_def2_QZVPPD_SMD_G2__Single_Point = "B97-D/def2-QZVPPD/SMD(G2) Single Point"
    B97_D_def2_QZVPPD_SMD_G2__Force = "B97-D/def2-QZVPPD/SMD(G2) Force"
    B97_D_def2_QZVPPD_SMD_G2__Geometry_Optimization = "B97-D/def2-QZVPPD/SMD(G2) Geometry Optimization"
    B97_D_def2_QZVPPD_SMD_G2__Frequency_Analysis = "B97-D/def2-QZVPPD/SMD(G2) Frequency Analysis"
    B97_D_def2_QZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_G2__Unknown = "B97-D/def2-QZVPPD/SMD(G2) Unknown"
    B97_D_def2_QZVPPD_SMD_DME__Single_Point = "B97-D/def2-QZVPPD/SMD(DME) Single Point"
    B97_D_def2_QZVPPD_SMD_DME__Force = "B97-D/def2-QZVPPD/SMD(DME) Force"
    B97_D_def2_QZVPPD_SMD_DME__Geometry_Optimization = "B97-D/def2-QZVPPD/SMD(DME) Geometry Optimization"
    B97_D_def2_QZVPPD_SMD_DME__Frequency_Analysis = "B97-D/def2-QZVPPD/SMD(DME) Frequency Analysis"
    B97_D_def2_QZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_DME__Unknown = "B97-D/def2-QZVPPD/SMD(DME) Unknown"
    B97_D_def2_QZVPPD_SMD_EC__Single_Point = "B97-D/def2-QZVPPD/SMD(EC) Single Point"
    B97_D_def2_QZVPPD_SMD_EC__Force = "B97-D/def2-QZVPPD/SMD(EC) Force"
    B97_D_def2_QZVPPD_SMD_EC__Geometry_Optimization = "B97-D/def2-QZVPPD/SMD(EC) Geometry Optimization"
    B97_D_def2_QZVPPD_SMD_EC__Frequency_Analysis = "B97-D/def2-QZVPPD/SMD(EC) Frequency Analysis"
    B97_D_def2_QZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_EC__Unknown = "B97-D/def2-QZVPPD/SMD(EC) Unknown"
    B97_D_def2_QZVPPD_SMD_EC_EMC__Single_Point = "B97-D/def2-QZVPPD/SMD(EC/EMC) Single Point"
    B97_D_def2_QZVPPD_SMD_EC_EMC__Force = "B97-D/def2-QZVPPD/SMD(EC/EMC) Force"
    B97_D_def2_QZVPPD_SMD_EC_EMC__Geometry_Optimization = "B97-D/def2-QZVPPD/SMD(EC/EMC) Geometry Optimization"
    B97_D_def2_QZVPPD_SMD_EC_EMC__Frequency_Analysis = "B97-D/def2-QZVPPD/SMD(EC/EMC) Frequency Analysis"
    B97_D_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_EC_EMC__Unknown = "B97-D/def2-QZVPPD/SMD(EC/EMC) Unknown"
    B97_D_def2_QZVPPD_SMD_PC__Single_Point = "B97-D/def2-QZVPPD/SMD(PC) Single Point"
    B97_D_def2_QZVPPD_SMD_PC__Force = "B97-D/def2-QZVPPD/SMD(PC) Force"
    B97_D_def2_QZVPPD_SMD_PC__Geometry_Optimization = "B97-D/def2-QZVPPD/SMD(PC) Geometry Optimization"
    B97_D_def2_QZVPPD_SMD_PC__Frequency_Analysis = "B97-D/def2-QZVPPD/SMD(PC) Frequency Analysis"
    B97_D_def2_QZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_SMD_PC__Unknown = "B97-D/def2-QZVPPD/SMD(PC) Unknown"
    B97_D_def2_QZVPPD_CMIRS_WATER__Single_Point = "B97-D/def2-QZVPPD/CMIRS(WATER) Single Point"
    B97_D_def2_QZVPPD_CMIRS_WATER__Force = "B97-D/def2-QZVPPD/CMIRS(WATER) Force"
    B97_D_def2_QZVPPD_CMIRS_WATER__Geometry_Optimization = "B97-D/def2-QZVPPD/CMIRS(WATER) Geometry Optimization"
    B97_D_def2_QZVPPD_CMIRS_WATER__Frequency_Analysis = "B97-D/def2-QZVPPD/CMIRS(WATER) Frequency Analysis"
    B97_D_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_WATER__Unknown = "B97-D/def2-QZVPPD/CMIRS(WATER) Unknown"
    B97_D_def2_QZVPPD_CMIRS_THF__Single_Point = "B97-D/def2-QZVPPD/CMIRS(THF) Single Point"
    B97_D_def2_QZVPPD_CMIRS_THF__Force = "B97-D/def2-QZVPPD/CMIRS(THF) Force"
    B97_D_def2_QZVPPD_CMIRS_THF__Geometry_Optimization = "B97-D/def2-QZVPPD/CMIRS(THF) Geometry Optimization"
    B97_D_def2_QZVPPD_CMIRS_THF__Frequency_Analysis = "B97-D/def2-QZVPPD/CMIRS(THF) Frequency Analysis"
    B97_D_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_THF__Unknown = "B97-D/def2-QZVPPD/CMIRS(THF) Unknown"
    B97_D_def2_QZVPPD_CMIRS_G2__Single_Point = "B97-D/def2-QZVPPD/CMIRS(G2) Single Point"
    B97_D_def2_QZVPPD_CMIRS_G2__Force = "B97-D/def2-QZVPPD/CMIRS(G2) Force"
    B97_D_def2_QZVPPD_CMIRS_G2__Geometry_Optimization = "B97-D/def2-QZVPPD/CMIRS(G2) Geometry Optimization"
    B97_D_def2_QZVPPD_CMIRS_G2__Frequency_Analysis = "B97-D/def2-QZVPPD/CMIRS(G2) Frequency Analysis"
    B97_D_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_G2__Unknown = "B97-D/def2-QZVPPD/CMIRS(G2) Unknown"
    B97_D_def2_QZVPPD_CMIRS_DME__Single_Point = "B97-D/def2-QZVPPD/CMIRS(DME) Single Point"
    B97_D_def2_QZVPPD_CMIRS_DME__Force = "B97-D/def2-QZVPPD/CMIRS(DME) Force"
    B97_D_def2_QZVPPD_CMIRS_DME__Geometry_Optimization = "B97-D/def2-QZVPPD/CMIRS(DME) Geometry Optimization"
    B97_D_def2_QZVPPD_CMIRS_DME__Frequency_Analysis = "B97-D/def2-QZVPPD/CMIRS(DME) Frequency Analysis"
    B97_D_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_DME__Unknown = "B97-D/def2-QZVPPD/CMIRS(DME) Unknown"
    B97_D_def2_QZVPPD_CMIRS_EC__Single_Point = "B97-D/def2-QZVPPD/CMIRS(EC) Single Point"
    B97_D_def2_QZVPPD_CMIRS_EC__Force = "B97-D/def2-QZVPPD/CMIRS(EC) Force"
    B97_D_def2_QZVPPD_CMIRS_EC__Geometry_Optimization = "B97-D/def2-QZVPPD/CMIRS(EC) Geometry Optimization"
    B97_D_def2_QZVPPD_CMIRS_EC__Frequency_Analysis = "B97-D/def2-QZVPPD/CMIRS(EC) Frequency Analysis"
    B97_D_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_EC__Unknown = "B97-D/def2-QZVPPD/CMIRS(EC) Unknown"
    B97_D_def2_QZVPPD_CMIRS_EC_EMC__Single_Point = "B97-D/def2-QZVPPD/CMIRS(EC/EMC) Single Point"
    B97_D_def2_QZVPPD_CMIRS_EC_EMC__Force = "B97-D/def2-QZVPPD/CMIRS(EC/EMC) Force"
    B97_D_def2_QZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "B97-D/def2-QZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    B97_D_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "B97-D/def2-QZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    B97_D_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_EC_EMC__Unknown = "B97-D/def2-QZVPPD/CMIRS(EC/EMC) Unknown"
    B97_D_def2_QZVPPD_CMIRS_PC__Single_Point = "B97-D/def2-QZVPPD/CMIRS(PC) Single Point"
    B97_D_def2_QZVPPD_CMIRS_PC__Force = "B97-D/def2-QZVPPD/CMIRS(PC) Force"
    B97_D_def2_QZVPPD_CMIRS_PC__Geometry_Optimization = "B97-D/def2-QZVPPD/CMIRS(PC) Geometry Optimization"
    B97_D_def2_QZVPPD_CMIRS_PC__Frequency_Analysis = "B97-D/def2-QZVPPD/CMIRS(PC) Frequency Analysis"
    B97_D_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D/def2-QZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D_def2_QZVPPD_CMIRS_PC__Unknown = "B97-D/def2-QZVPPD/CMIRS(PC) Unknown"
    B97_D3_6_31g_d_VACUUM_Single_Point = "B97-D3/6-31g*/VACUUM Single Point"
    B97_D3_6_31g_d_VACUUM_Force = "B97-D3/6-31g*/VACUUM Force"
    B97_D3_6_31g_d_VACUUM_Geometry_Optimization = "B97-D3/6-31g*/VACUUM Geometry Optimization"
    B97_D3_6_31g_d_VACUUM_Frequency_Analysis = "B97-D3/6-31g*/VACUUM Frequency Analysis"
    B97_D3_6_31g_d_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/VACUUM Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_VACUUM_Unknown = "B97-D3/6-31g*/VACUUM Unknown"
    B97_D3_6_31g_d_PCM_WATER__Single_Point = "B97-D3/6-31g*/PCM(WATER) Single Point"
    B97_D3_6_31g_d_PCM_WATER__Force = "B97-D3/6-31g*/PCM(WATER) Force"
    B97_D3_6_31g_d_PCM_WATER__Geometry_Optimization = "B97-D3/6-31g*/PCM(WATER) Geometry Optimization"
    B97_D3_6_31g_d_PCM_WATER__Frequency_Analysis = "B97-D3/6-31g*/PCM(WATER) Frequency Analysis"
    B97_D3_6_31g_d_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_WATER__Unknown = "B97-D3/6-31g*/PCM(WATER) Unknown"
    B97_D3_6_31g_d_PCM_THF__Single_Point = "B97-D3/6-31g*/PCM(THF) Single Point"
    B97_D3_6_31g_d_PCM_THF__Force = "B97-D3/6-31g*/PCM(THF) Force"
    B97_D3_6_31g_d_PCM_THF__Geometry_Optimization = "B97-D3/6-31g*/PCM(THF) Geometry Optimization"
    B97_D3_6_31g_d_PCM_THF__Frequency_Analysis = "B97-D3/6-31g*/PCM(THF) Frequency Analysis"
    B97_D3_6_31g_d_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_THF__Unknown = "B97-D3/6-31g*/PCM(THF) Unknown"
    B97_D3_6_31g_d_PCM_G2__Single_Point = "B97-D3/6-31g*/PCM(G2) Single Point"
    B97_D3_6_31g_d_PCM_G2__Force = "B97-D3/6-31g*/PCM(G2) Force"
    B97_D3_6_31g_d_PCM_G2__Geometry_Optimization = "B97-D3/6-31g*/PCM(G2) Geometry Optimization"
    B97_D3_6_31g_d_PCM_G2__Frequency_Analysis = "B97-D3/6-31g*/PCM(G2) Frequency Analysis"
    B97_D3_6_31g_d_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_G2__Unknown = "B97-D3/6-31g*/PCM(G2) Unknown"
    B97_D3_6_31g_d_PCM_DME__Single_Point = "B97-D3/6-31g*/PCM(DME) Single Point"
    B97_D3_6_31g_d_PCM_DME__Force = "B97-D3/6-31g*/PCM(DME) Force"
    B97_D3_6_31g_d_PCM_DME__Geometry_Optimization = "B97-D3/6-31g*/PCM(DME) Geometry Optimization"
    B97_D3_6_31g_d_PCM_DME__Frequency_Analysis = "B97-D3/6-31g*/PCM(DME) Frequency Analysis"
    B97_D3_6_31g_d_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_DME__Unknown = "B97-D3/6-31g*/PCM(DME) Unknown"
    B97_D3_6_31g_d_PCM_EC__Single_Point = "B97-D3/6-31g*/PCM(EC) Single Point"
    B97_D3_6_31g_d_PCM_EC__Force = "B97-D3/6-31g*/PCM(EC) Force"
    B97_D3_6_31g_d_PCM_EC__Geometry_Optimization = "B97-D3/6-31g*/PCM(EC) Geometry Optimization"
    B97_D3_6_31g_d_PCM_EC__Frequency_Analysis = "B97-D3/6-31g*/PCM(EC) Frequency Analysis"
    B97_D3_6_31g_d_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_EC__Unknown = "B97-D3/6-31g*/PCM(EC) Unknown"
    B97_D3_6_31g_d_PCM_EC_EMC__Single_Point = "B97-D3/6-31g*/PCM(EC/EMC) Single Point"
    B97_D3_6_31g_d_PCM_EC_EMC__Force = "B97-D3/6-31g*/PCM(EC/EMC) Force"
    B97_D3_6_31g_d_PCM_EC_EMC__Geometry_Optimization = "B97-D3/6-31g*/PCM(EC/EMC) Geometry Optimization"
    B97_D3_6_31g_d_PCM_EC_EMC__Frequency_Analysis = "B97-D3/6-31g*/PCM(EC/EMC) Frequency Analysis"
    B97_D3_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_EC_EMC__Unknown = "B97-D3/6-31g*/PCM(EC/EMC) Unknown"
    B97_D3_6_31g_d_PCM_PC__Single_Point = "B97-D3/6-31g*/PCM(PC) Single Point"
    B97_D3_6_31g_d_PCM_PC__Force = "B97-D3/6-31g*/PCM(PC) Force"
    B97_D3_6_31g_d_PCM_PC__Geometry_Optimization = "B97-D3/6-31g*/PCM(PC) Geometry Optimization"
    B97_D3_6_31g_d_PCM_PC__Frequency_Analysis = "B97-D3/6-31g*/PCM(PC) Frequency Analysis"
    B97_D3_6_31g_d_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_PCM_PC__Unknown = "B97-D3/6-31g*/PCM(PC) Unknown"
    B97_D3_6_31g_d_ISOSVP_WATER__Single_Point = "B97-D3/6-31g*/ISOSVP(WATER) Single Point"
    B97_D3_6_31g_d_ISOSVP_WATER__Force = "B97-D3/6-31g*/ISOSVP(WATER) Force"
    B97_D3_6_31g_d_ISOSVP_WATER__Geometry_Optimization = "B97-D3/6-31g*/ISOSVP(WATER) Geometry Optimization"
    B97_D3_6_31g_d_ISOSVP_WATER__Frequency_Analysis = "B97-D3/6-31g*/ISOSVP(WATER) Frequency Analysis"
    B97_D3_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_WATER__Unknown = "B97-D3/6-31g*/ISOSVP(WATER) Unknown"
    B97_D3_6_31g_d_ISOSVP_THF__Single_Point = "B97-D3/6-31g*/ISOSVP(THF) Single Point"
    B97_D3_6_31g_d_ISOSVP_THF__Force = "B97-D3/6-31g*/ISOSVP(THF) Force"
    B97_D3_6_31g_d_ISOSVP_THF__Geometry_Optimization = "B97-D3/6-31g*/ISOSVP(THF) Geometry Optimization"
    B97_D3_6_31g_d_ISOSVP_THF__Frequency_Analysis = "B97-D3/6-31g*/ISOSVP(THF) Frequency Analysis"
    B97_D3_6_31g_d_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_THF__Unknown = "B97-D3/6-31g*/ISOSVP(THF) Unknown"
    B97_D3_6_31g_d_ISOSVP_G2__Single_Point = "B97-D3/6-31g*/ISOSVP(G2) Single Point"
    B97_D3_6_31g_d_ISOSVP_G2__Force = "B97-D3/6-31g*/ISOSVP(G2) Force"
    B97_D3_6_31g_d_ISOSVP_G2__Geometry_Optimization = "B97-D3/6-31g*/ISOSVP(G2) Geometry Optimization"
    B97_D3_6_31g_d_ISOSVP_G2__Frequency_Analysis = "B97-D3/6-31g*/ISOSVP(G2) Frequency Analysis"
    B97_D3_6_31g_d_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_G2__Unknown = "B97-D3/6-31g*/ISOSVP(G2) Unknown"
    B97_D3_6_31g_d_ISOSVP_DME__Single_Point = "B97-D3/6-31g*/ISOSVP(DME) Single Point"
    B97_D3_6_31g_d_ISOSVP_DME__Force = "B97-D3/6-31g*/ISOSVP(DME) Force"
    B97_D3_6_31g_d_ISOSVP_DME__Geometry_Optimization = "B97-D3/6-31g*/ISOSVP(DME) Geometry Optimization"
    B97_D3_6_31g_d_ISOSVP_DME__Frequency_Analysis = "B97-D3/6-31g*/ISOSVP(DME) Frequency Analysis"
    B97_D3_6_31g_d_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_DME__Unknown = "B97-D3/6-31g*/ISOSVP(DME) Unknown"
    B97_D3_6_31g_d_ISOSVP_EC__Single_Point = "B97-D3/6-31g*/ISOSVP(EC) Single Point"
    B97_D3_6_31g_d_ISOSVP_EC__Force = "B97-D3/6-31g*/ISOSVP(EC) Force"
    B97_D3_6_31g_d_ISOSVP_EC__Geometry_Optimization = "B97-D3/6-31g*/ISOSVP(EC) Geometry Optimization"
    B97_D3_6_31g_d_ISOSVP_EC__Frequency_Analysis = "B97-D3/6-31g*/ISOSVP(EC) Frequency Analysis"
    B97_D3_6_31g_d_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_EC__Unknown = "B97-D3/6-31g*/ISOSVP(EC) Unknown"
    B97_D3_6_31g_d_ISOSVP_EC_EMC__Single_Point = "B97-D3/6-31g*/ISOSVP(EC/EMC) Single Point"
    B97_D3_6_31g_d_ISOSVP_EC_EMC__Force = "B97-D3/6-31g*/ISOSVP(EC/EMC) Force"
    B97_D3_6_31g_d_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D3/6-31g*/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D3_6_31g_d_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D3/6-31g*/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D3_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_EC_EMC__Unknown = "B97-D3/6-31g*/ISOSVP(EC/EMC) Unknown"
    B97_D3_6_31g_d_ISOSVP_PC__Single_Point = "B97-D3/6-31g*/ISOSVP(PC) Single Point"
    B97_D3_6_31g_d_ISOSVP_PC__Force = "B97-D3/6-31g*/ISOSVP(PC) Force"
    B97_D3_6_31g_d_ISOSVP_PC__Geometry_Optimization = "B97-D3/6-31g*/ISOSVP(PC) Geometry Optimization"
    B97_D3_6_31g_d_ISOSVP_PC__Frequency_Analysis = "B97-D3/6-31g*/ISOSVP(PC) Frequency Analysis"
    B97_D3_6_31g_d_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_ISOSVP_PC__Unknown = "B97-D3/6-31g*/ISOSVP(PC) Unknown"
    B97_D3_6_31g_d_SMD_WATER__Single_Point = "B97-D3/6-31g*/SMD(WATER) Single Point"
    B97_D3_6_31g_d_SMD_WATER__Force = "B97-D3/6-31g*/SMD(WATER) Force"
    B97_D3_6_31g_d_SMD_WATER__Geometry_Optimization = "B97-D3/6-31g*/SMD(WATER) Geometry Optimization"
    B97_D3_6_31g_d_SMD_WATER__Frequency_Analysis = "B97-D3/6-31g*/SMD(WATER) Frequency Analysis"
    B97_D3_6_31g_d_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_WATER__Unknown = "B97-D3/6-31g*/SMD(WATER) Unknown"
    B97_D3_6_31g_d_SMD_THF__Single_Point = "B97-D3/6-31g*/SMD(THF) Single Point"
    B97_D3_6_31g_d_SMD_THF__Force = "B97-D3/6-31g*/SMD(THF) Force"
    B97_D3_6_31g_d_SMD_THF__Geometry_Optimization = "B97-D3/6-31g*/SMD(THF) Geometry Optimization"
    B97_D3_6_31g_d_SMD_THF__Frequency_Analysis = "B97-D3/6-31g*/SMD(THF) Frequency Analysis"
    B97_D3_6_31g_d_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_THF__Unknown = "B97-D3/6-31g*/SMD(THF) Unknown"
    B97_D3_6_31g_d_SMD_G2__Single_Point = "B97-D3/6-31g*/SMD(G2) Single Point"
    B97_D3_6_31g_d_SMD_G2__Force = "B97-D3/6-31g*/SMD(G2) Force"
    B97_D3_6_31g_d_SMD_G2__Geometry_Optimization = "B97-D3/6-31g*/SMD(G2) Geometry Optimization"
    B97_D3_6_31g_d_SMD_G2__Frequency_Analysis = "B97-D3/6-31g*/SMD(G2) Frequency Analysis"
    B97_D3_6_31g_d_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_G2__Unknown = "B97-D3/6-31g*/SMD(G2) Unknown"
    B97_D3_6_31g_d_SMD_DME__Single_Point = "B97-D3/6-31g*/SMD(DME) Single Point"
    B97_D3_6_31g_d_SMD_DME__Force = "B97-D3/6-31g*/SMD(DME) Force"
    B97_D3_6_31g_d_SMD_DME__Geometry_Optimization = "B97-D3/6-31g*/SMD(DME) Geometry Optimization"
    B97_D3_6_31g_d_SMD_DME__Frequency_Analysis = "B97-D3/6-31g*/SMD(DME) Frequency Analysis"
    B97_D3_6_31g_d_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_DME__Unknown = "B97-D3/6-31g*/SMD(DME) Unknown"
    B97_D3_6_31g_d_SMD_EC__Single_Point = "B97-D3/6-31g*/SMD(EC) Single Point"
    B97_D3_6_31g_d_SMD_EC__Force = "B97-D3/6-31g*/SMD(EC) Force"
    B97_D3_6_31g_d_SMD_EC__Geometry_Optimization = "B97-D3/6-31g*/SMD(EC) Geometry Optimization"
    B97_D3_6_31g_d_SMD_EC__Frequency_Analysis = "B97-D3/6-31g*/SMD(EC) Frequency Analysis"
    B97_D3_6_31g_d_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_EC__Unknown = "B97-D3/6-31g*/SMD(EC) Unknown"
    B97_D3_6_31g_d_SMD_EC_EMC__Single_Point = "B97-D3/6-31g*/SMD(EC/EMC) Single Point"
    B97_D3_6_31g_d_SMD_EC_EMC__Force = "B97-D3/6-31g*/SMD(EC/EMC) Force"
    B97_D3_6_31g_d_SMD_EC_EMC__Geometry_Optimization = "B97-D3/6-31g*/SMD(EC/EMC) Geometry Optimization"
    B97_D3_6_31g_d_SMD_EC_EMC__Frequency_Analysis = "B97-D3/6-31g*/SMD(EC/EMC) Frequency Analysis"
    B97_D3_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_EC_EMC__Unknown = "B97-D3/6-31g*/SMD(EC/EMC) Unknown"
    B97_D3_6_31g_d_SMD_PC__Single_Point = "B97-D3/6-31g*/SMD(PC) Single Point"
    B97_D3_6_31g_d_SMD_PC__Force = "B97-D3/6-31g*/SMD(PC) Force"
    B97_D3_6_31g_d_SMD_PC__Geometry_Optimization = "B97-D3/6-31g*/SMD(PC) Geometry Optimization"
    B97_D3_6_31g_d_SMD_PC__Frequency_Analysis = "B97-D3/6-31g*/SMD(PC) Frequency Analysis"
    B97_D3_6_31g_d_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_SMD_PC__Unknown = "B97-D3/6-31g*/SMD(PC) Unknown"
    B97_D3_6_31g_d_CMIRS_WATER__Single_Point = "B97-D3/6-31g*/CMIRS(WATER) Single Point"
    B97_D3_6_31g_d_CMIRS_WATER__Force = "B97-D3/6-31g*/CMIRS(WATER) Force"
    B97_D3_6_31g_d_CMIRS_WATER__Geometry_Optimization = "B97-D3/6-31g*/CMIRS(WATER) Geometry Optimization"
    B97_D3_6_31g_d_CMIRS_WATER__Frequency_Analysis = "B97-D3/6-31g*/CMIRS(WATER) Frequency Analysis"
    B97_D3_6_31g_d_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_WATER__Unknown = "B97-D3/6-31g*/CMIRS(WATER) Unknown"
    B97_D3_6_31g_d_CMIRS_THF__Single_Point = "B97-D3/6-31g*/CMIRS(THF) Single Point"
    B97_D3_6_31g_d_CMIRS_THF__Force = "B97-D3/6-31g*/CMIRS(THF) Force"
    B97_D3_6_31g_d_CMIRS_THF__Geometry_Optimization = "B97-D3/6-31g*/CMIRS(THF) Geometry Optimization"
    B97_D3_6_31g_d_CMIRS_THF__Frequency_Analysis = "B97-D3/6-31g*/CMIRS(THF) Frequency Analysis"
    B97_D3_6_31g_d_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_THF__Unknown = "B97-D3/6-31g*/CMIRS(THF) Unknown"
    B97_D3_6_31g_d_CMIRS_G2__Single_Point = "B97-D3/6-31g*/CMIRS(G2) Single Point"
    B97_D3_6_31g_d_CMIRS_G2__Force = "B97-D3/6-31g*/CMIRS(G2) Force"
    B97_D3_6_31g_d_CMIRS_G2__Geometry_Optimization = "B97-D3/6-31g*/CMIRS(G2) Geometry Optimization"
    B97_D3_6_31g_d_CMIRS_G2__Frequency_Analysis = "B97-D3/6-31g*/CMIRS(G2) Frequency Analysis"
    B97_D3_6_31g_d_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_G2__Unknown = "B97-D3/6-31g*/CMIRS(G2) Unknown"
    B97_D3_6_31g_d_CMIRS_DME__Single_Point = "B97-D3/6-31g*/CMIRS(DME) Single Point"
    B97_D3_6_31g_d_CMIRS_DME__Force = "B97-D3/6-31g*/CMIRS(DME) Force"
    B97_D3_6_31g_d_CMIRS_DME__Geometry_Optimization = "B97-D3/6-31g*/CMIRS(DME) Geometry Optimization"
    B97_D3_6_31g_d_CMIRS_DME__Frequency_Analysis = "B97-D3/6-31g*/CMIRS(DME) Frequency Analysis"
    B97_D3_6_31g_d_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_DME__Unknown = "B97-D3/6-31g*/CMIRS(DME) Unknown"
    B97_D3_6_31g_d_CMIRS_EC__Single_Point = "B97-D3/6-31g*/CMIRS(EC) Single Point"
    B97_D3_6_31g_d_CMIRS_EC__Force = "B97-D3/6-31g*/CMIRS(EC) Force"
    B97_D3_6_31g_d_CMIRS_EC__Geometry_Optimization = "B97-D3/6-31g*/CMIRS(EC) Geometry Optimization"
    B97_D3_6_31g_d_CMIRS_EC__Frequency_Analysis = "B97-D3/6-31g*/CMIRS(EC) Frequency Analysis"
    B97_D3_6_31g_d_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_EC__Unknown = "B97-D3/6-31g*/CMIRS(EC) Unknown"
    B97_D3_6_31g_d_CMIRS_EC_EMC__Single_Point = "B97-D3/6-31g*/CMIRS(EC/EMC) Single Point"
    B97_D3_6_31g_d_CMIRS_EC_EMC__Force = "B97-D3/6-31g*/CMIRS(EC/EMC) Force"
    B97_D3_6_31g_d_CMIRS_EC_EMC__Geometry_Optimization = "B97-D3/6-31g*/CMIRS(EC/EMC) Geometry Optimization"
    B97_D3_6_31g_d_CMIRS_EC_EMC__Frequency_Analysis = "B97-D3/6-31g*/CMIRS(EC/EMC) Frequency Analysis"
    B97_D3_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_EC_EMC__Unknown = "B97-D3/6-31g*/CMIRS(EC/EMC) Unknown"
    B97_D3_6_31g_d_CMIRS_PC__Single_Point = "B97-D3/6-31g*/CMIRS(PC) Single Point"
    B97_D3_6_31g_d_CMIRS_PC__Force = "B97-D3/6-31g*/CMIRS(PC) Force"
    B97_D3_6_31g_d_CMIRS_PC__Geometry_Optimization = "B97-D3/6-31g*/CMIRS(PC) Geometry Optimization"
    B97_D3_6_31g_d_CMIRS_PC__Frequency_Analysis = "B97-D3/6-31g*/CMIRS(PC) Frequency Analysis"
    B97_D3_6_31g_d_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/6-31g*/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_6_31g_d_CMIRS_PC__Unknown = "B97-D3/6-31g*/CMIRS(PC) Unknown"
    B97_D3_def2_SVPD_VACUUM_Single_Point = "B97-D3/def2-SVPD/VACUUM Single Point"
    B97_D3_def2_SVPD_VACUUM_Force = "B97-D3/def2-SVPD/VACUUM Force"
    B97_D3_def2_SVPD_VACUUM_Geometry_Optimization = "B97-D3/def2-SVPD/VACUUM Geometry Optimization"
    B97_D3_def2_SVPD_VACUUM_Frequency_Analysis = "B97-D3/def2-SVPD/VACUUM Frequency Analysis"
    B97_D3_def2_SVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/VACUUM Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_VACUUM_Unknown = "B97-D3/def2-SVPD/VACUUM Unknown"
    B97_D3_def2_SVPD_PCM_WATER__Single_Point = "B97-D3/def2-SVPD/PCM(WATER) Single Point"
    B97_D3_def2_SVPD_PCM_WATER__Force = "B97-D3/def2-SVPD/PCM(WATER) Force"
    B97_D3_def2_SVPD_PCM_WATER__Geometry_Optimization = "B97-D3/def2-SVPD/PCM(WATER) Geometry Optimization"
    B97_D3_def2_SVPD_PCM_WATER__Frequency_Analysis = "B97-D3/def2-SVPD/PCM(WATER) Frequency Analysis"
    B97_D3_def2_SVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_WATER__Unknown = "B97-D3/def2-SVPD/PCM(WATER) Unknown"
    B97_D3_def2_SVPD_PCM_THF__Single_Point = "B97-D3/def2-SVPD/PCM(THF) Single Point"
    B97_D3_def2_SVPD_PCM_THF__Force = "B97-D3/def2-SVPD/PCM(THF) Force"
    B97_D3_def2_SVPD_PCM_THF__Geometry_Optimization = "B97-D3/def2-SVPD/PCM(THF) Geometry Optimization"
    B97_D3_def2_SVPD_PCM_THF__Frequency_Analysis = "B97-D3/def2-SVPD/PCM(THF) Frequency Analysis"
    B97_D3_def2_SVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_THF__Unknown = "B97-D3/def2-SVPD/PCM(THF) Unknown"
    B97_D3_def2_SVPD_PCM_G2__Single_Point = "B97-D3/def2-SVPD/PCM(G2) Single Point"
    B97_D3_def2_SVPD_PCM_G2__Force = "B97-D3/def2-SVPD/PCM(G2) Force"
    B97_D3_def2_SVPD_PCM_G2__Geometry_Optimization = "B97-D3/def2-SVPD/PCM(G2) Geometry Optimization"
    B97_D3_def2_SVPD_PCM_G2__Frequency_Analysis = "B97-D3/def2-SVPD/PCM(G2) Frequency Analysis"
    B97_D3_def2_SVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_G2__Unknown = "B97-D3/def2-SVPD/PCM(G2) Unknown"
    B97_D3_def2_SVPD_PCM_DME__Single_Point = "B97-D3/def2-SVPD/PCM(DME) Single Point"
    B97_D3_def2_SVPD_PCM_DME__Force = "B97-D3/def2-SVPD/PCM(DME) Force"
    B97_D3_def2_SVPD_PCM_DME__Geometry_Optimization = "B97-D3/def2-SVPD/PCM(DME) Geometry Optimization"
    B97_D3_def2_SVPD_PCM_DME__Frequency_Analysis = "B97-D3/def2-SVPD/PCM(DME) Frequency Analysis"
    B97_D3_def2_SVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_DME__Unknown = "B97-D3/def2-SVPD/PCM(DME) Unknown"
    B97_D3_def2_SVPD_PCM_EC__Single_Point = "B97-D3/def2-SVPD/PCM(EC) Single Point"
    B97_D3_def2_SVPD_PCM_EC__Force = "B97-D3/def2-SVPD/PCM(EC) Force"
    B97_D3_def2_SVPD_PCM_EC__Geometry_Optimization = "B97-D3/def2-SVPD/PCM(EC) Geometry Optimization"
    B97_D3_def2_SVPD_PCM_EC__Frequency_Analysis = "B97-D3/def2-SVPD/PCM(EC) Frequency Analysis"
    B97_D3_def2_SVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_EC__Unknown = "B97-D3/def2-SVPD/PCM(EC) Unknown"
    B97_D3_def2_SVPD_PCM_EC_EMC__Single_Point = "B97-D3/def2-SVPD/PCM(EC/EMC) Single Point"
    B97_D3_def2_SVPD_PCM_EC_EMC__Force = "B97-D3/def2-SVPD/PCM(EC/EMC) Force"
    B97_D3_def2_SVPD_PCM_EC_EMC__Geometry_Optimization = "B97-D3/def2-SVPD/PCM(EC/EMC) Geometry Optimization"
    B97_D3_def2_SVPD_PCM_EC_EMC__Frequency_Analysis = "B97-D3/def2-SVPD/PCM(EC/EMC) Frequency Analysis"
    B97_D3_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_EC_EMC__Unknown = "B97-D3/def2-SVPD/PCM(EC/EMC) Unknown"
    B97_D3_def2_SVPD_PCM_PC__Single_Point = "B97-D3/def2-SVPD/PCM(PC) Single Point"
    B97_D3_def2_SVPD_PCM_PC__Force = "B97-D3/def2-SVPD/PCM(PC) Force"
    B97_D3_def2_SVPD_PCM_PC__Geometry_Optimization = "B97-D3/def2-SVPD/PCM(PC) Geometry Optimization"
    B97_D3_def2_SVPD_PCM_PC__Frequency_Analysis = "B97-D3/def2-SVPD/PCM(PC) Frequency Analysis"
    B97_D3_def2_SVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_PCM_PC__Unknown = "B97-D3/def2-SVPD/PCM(PC) Unknown"
    B97_D3_def2_SVPD_ISOSVP_WATER__Single_Point = "B97-D3/def2-SVPD/ISOSVP(WATER) Single Point"
    B97_D3_def2_SVPD_ISOSVP_WATER__Force = "B97-D3/def2-SVPD/ISOSVP(WATER) Force"
    B97_D3_def2_SVPD_ISOSVP_WATER__Geometry_Optimization = "B97-D3/def2-SVPD/ISOSVP(WATER) Geometry Optimization"
    B97_D3_def2_SVPD_ISOSVP_WATER__Frequency_Analysis = "B97-D3/def2-SVPD/ISOSVP(WATER) Frequency Analysis"
    B97_D3_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_WATER__Unknown = "B97-D3/def2-SVPD/ISOSVP(WATER) Unknown"
    B97_D3_def2_SVPD_ISOSVP_THF__Single_Point = "B97-D3/def2-SVPD/ISOSVP(THF) Single Point"
    B97_D3_def2_SVPD_ISOSVP_THF__Force = "B97-D3/def2-SVPD/ISOSVP(THF) Force"
    B97_D3_def2_SVPD_ISOSVP_THF__Geometry_Optimization = "B97-D3/def2-SVPD/ISOSVP(THF) Geometry Optimization"
    B97_D3_def2_SVPD_ISOSVP_THF__Frequency_Analysis = "B97-D3/def2-SVPD/ISOSVP(THF) Frequency Analysis"
    B97_D3_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_THF__Unknown = "B97-D3/def2-SVPD/ISOSVP(THF) Unknown"
    B97_D3_def2_SVPD_ISOSVP_G2__Single_Point = "B97-D3/def2-SVPD/ISOSVP(G2) Single Point"
    B97_D3_def2_SVPD_ISOSVP_G2__Force = "B97-D3/def2-SVPD/ISOSVP(G2) Force"
    B97_D3_def2_SVPD_ISOSVP_G2__Geometry_Optimization = "B97-D3/def2-SVPD/ISOSVP(G2) Geometry Optimization"
    B97_D3_def2_SVPD_ISOSVP_G2__Frequency_Analysis = "B97-D3/def2-SVPD/ISOSVP(G2) Frequency Analysis"
    B97_D3_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_G2__Unknown = "B97-D3/def2-SVPD/ISOSVP(G2) Unknown"
    B97_D3_def2_SVPD_ISOSVP_DME__Single_Point = "B97-D3/def2-SVPD/ISOSVP(DME) Single Point"
    B97_D3_def2_SVPD_ISOSVP_DME__Force = "B97-D3/def2-SVPD/ISOSVP(DME) Force"
    B97_D3_def2_SVPD_ISOSVP_DME__Geometry_Optimization = "B97-D3/def2-SVPD/ISOSVP(DME) Geometry Optimization"
    B97_D3_def2_SVPD_ISOSVP_DME__Frequency_Analysis = "B97-D3/def2-SVPD/ISOSVP(DME) Frequency Analysis"
    B97_D3_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_DME__Unknown = "B97-D3/def2-SVPD/ISOSVP(DME) Unknown"
    B97_D3_def2_SVPD_ISOSVP_EC__Single_Point = "B97-D3/def2-SVPD/ISOSVP(EC) Single Point"
    B97_D3_def2_SVPD_ISOSVP_EC__Force = "B97-D3/def2-SVPD/ISOSVP(EC) Force"
    B97_D3_def2_SVPD_ISOSVP_EC__Geometry_Optimization = "B97-D3/def2-SVPD/ISOSVP(EC) Geometry Optimization"
    B97_D3_def2_SVPD_ISOSVP_EC__Frequency_Analysis = "B97-D3/def2-SVPD/ISOSVP(EC) Frequency Analysis"
    B97_D3_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_EC__Unknown = "B97-D3/def2-SVPD/ISOSVP(EC) Unknown"
    B97_D3_def2_SVPD_ISOSVP_EC_EMC__Single_Point = "B97-D3/def2-SVPD/ISOSVP(EC/EMC) Single Point"
    B97_D3_def2_SVPD_ISOSVP_EC_EMC__Force = "B97-D3/def2-SVPD/ISOSVP(EC/EMC) Force"
    B97_D3_def2_SVPD_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D3/def2-SVPD/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D3_def2_SVPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D3/def2-SVPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D3_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_EC_EMC__Unknown = "B97-D3/def2-SVPD/ISOSVP(EC/EMC) Unknown"
    B97_D3_def2_SVPD_ISOSVP_PC__Single_Point = "B97-D3/def2-SVPD/ISOSVP(PC) Single Point"
    B97_D3_def2_SVPD_ISOSVP_PC__Force = "B97-D3/def2-SVPD/ISOSVP(PC) Force"
    B97_D3_def2_SVPD_ISOSVP_PC__Geometry_Optimization = "B97-D3/def2-SVPD/ISOSVP(PC) Geometry Optimization"
    B97_D3_def2_SVPD_ISOSVP_PC__Frequency_Analysis = "B97-D3/def2-SVPD/ISOSVP(PC) Frequency Analysis"
    B97_D3_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_ISOSVP_PC__Unknown = "B97-D3/def2-SVPD/ISOSVP(PC) Unknown"
    B97_D3_def2_SVPD_SMD_WATER__Single_Point = "B97-D3/def2-SVPD/SMD(WATER) Single Point"
    B97_D3_def2_SVPD_SMD_WATER__Force = "B97-D3/def2-SVPD/SMD(WATER) Force"
    B97_D3_def2_SVPD_SMD_WATER__Geometry_Optimization = "B97-D3/def2-SVPD/SMD(WATER) Geometry Optimization"
    B97_D3_def2_SVPD_SMD_WATER__Frequency_Analysis = "B97-D3/def2-SVPD/SMD(WATER) Frequency Analysis"
    B97_D3_def2_SVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_WATER__Unknown = "B97-D3/def2-SVPD/SMD(WATER) Unknown"
    B97_D3_def2_SVPD_SMD_THF__Single_Point = "B97-D3/def2-SVPD/SMD(THF) Single Point"
    B97_D3_def2_SVPD_SMD_THF__Force = "B97-D3/def2-SVPD/SMD(THF) Force"
    B97_D3_def2_SVPD_SMD_THF__Geometry_Optimization = "B97-D3/def2-SVPD/SMD(THF) Geometry Optimization"
    B97_D3_def2_SVPD_SMD_THF__Frequency_Analysis = "B97-D3/def2-SVPD/SMD(THF) Frequency Analysis"
    B97_D3_def2_SVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_THF__Unknown = "B97-D3/def2-SVPD/SMD(THF) Unknown"
    B97_D3_def2_SVPD_SMD_G2__Single_Point = "B97-D3/def2-SVPD/SMD(G2) Single Point"
    B97_D3_def2_SVPD_SMD_G2__Force = "B97-D3/def2-SVPD/SMD(G2) Force"
    B97_D3_def2_SVPD_SMD_G2__Geometry_Optimization = "B97-D3/def2-SVPD/SMD(G2) Geometry Optimization"
    B97_D3_def2_SVPD_SMD_G2__Frequency_Analysis = "B97-D3/def2-SVPD/SMD(G2) Frequency Analysis"
    B97_D3_def2_SVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_G2__Unknown = "B97-D3/def2-SVPD/SMD(G2) Unknown"
    B97_D3_def2_SVPD_SMD_DME__Single_Point = "B97-D3/def2-SVPD/SMD(DME) Single Point"
    B97_D3_def2_SVPD_SMD_DME__Force = "B97-D3/def2-SVPD/SMD(DME) Force"
    B97_D3_def2_SVPD_SMD_DME__Geometry_Optimization = "B97-D3/def2-SVPD/SMD(DME) Geometry Optimization"
    B97_D3_def2_SVPD_SMD_DME__Frequency_Analysis = "B97-D3/def2-SVPD/SMD(DME) Frequency Analysis"
    B97_D3_def2_SVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_DME__Unknown = "B97-D3/def2-SVPD/SMD(DME) Unknown"
    B97_D3_def2_SVPD_SMD_EC__Single_Point = "B97-D3/def2-SVPD/SMD(EC) Single Point"
    B97_D3_def2_SVPD_SMD_EC__Force = "B97-D3/def2-SVPD/SMD(EC) Force"
    B97_D3_def2_SVPD_SMD_EC__Geometry_Optimization = "B97-D3/def2-SVPD/SMD(EC) Geometry Optimization"
    B97_D3_def2_SVPD_SMD_EC__Frequency_Analysis = "B97-D3/def2-SVPD/SMD(EC) Frequency Analysis"
    B97_D3_def2_SVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_EC__Unknown = "B97-D3/def2-SVPD/SMD(EC) Unknown"
    B97_D3_def2_SVPD_SMD_EC_EMC__Single_Point = "B97-D3/def2-SVPD/SMD(EC/EMC) Single Point"
    B97_D3_def2_SVPD_SMD_EC_EMC__Force = "B97-D3/def2-SVPD/SMD(EC/EMC) Force"
    B97_D3_def2_SVPD_SMD_EC_EMC__Geometry_Optimization = "B97-D3/def2-SVPD/SMD(EC/EMC) Geometry Optimization"
    B97_D3_def2_SVPD_SMD_EC_EMC__Frequency_Analysis = "B97-D3/def2-SVPD/SMD(EC/EMC) Frequency Analysis"
    B97_D3_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_EC_EMC__Unknown = "B97-D3/def2-SVPD/SMD(EC/EMC) Unknown"
    B97_D3_def2_SVPD_SMD_PC__Single_Point = "B97-D3/def2-SVPD/SMD(PC) Single Point"
    B97_D3_def2_SVPD_SMD_PC__Force = "B97-D3/def2-SVPD/SMD(PC) Force"
    B97_D3_def2_SVPD_SMD_PC__Geometry_Optimization = "B97-D3/def2-SVPD/SMD(PC) Geometry Optimization"
    B97_D3_def2_SVPD_SMD_PC__Frequency_Analysis = "B97-D3/def2-SVPD/SMD(PC) Frequency Analysis"
    B97_D3_def2_SVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_SMD_PC__Unknown = "B97-D3/def2-SVPD/SMD(PC) Unknown"
    B97_D3_def2_SVPD_CMIRS_WATER__Single_Point = "B97-D3/def2-SVPD/CMIRS(WATER) Single Point"
    B97_D3_def2_SVPD_CMIRS_WATER__Force = "B97-D3/def2-SVPD/CMIRS(WATER) Force"
    B97_D3_def2_SVPD_CMIRS_WATER__Geometry_Optimization = "B97-D3/def2-SVPD/CMIRS(WATER) Geometry Optimization"
    B97_D3_def2_SVPD_CMIRS_WATER__Frequency_Analysis = "B97-D3/def2-SVPD/CMIRS(WATER) Frequency Analysis"
    B97_D3_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_WATER__Unknown = "B97-D3/def2-SVPD/CMIRS(WATER) Unknown"
    B97_D3_def2_SVPD_CMIRS_THF__Single_Point = "B97-D3/def2-SVPD/CMIRS(THF) Single Point"
    B97_D3_def2_SVPD_CMIRS_THF__Force = "B97-D3/def2-SVPD/CMIRS(THF) Force"
    B97_D3_def2_SVPD_CMIRS_THF__Geometry_Optimization = "B97-D3/def2-SVPD/CMIRS(THF) Geometry Optimization"
    B97_D3_def2_SVPD_CMIRS_THF__Frequency_Analysis = "B97-D3/def2-SVPD/CMIRS(THF) Frequency Analysis"
    B97_D3_def2_SVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_THF__Unknown = "B97-D3/def2-SVPD/CMIRS(THF) Unknown"
    B97_D3_def2_SVPD_CMIRS_G2__Single_Point = "B97-D3/def2-SVPD/CMIRS(G2) Single Point"
    B97_D3_def2_SVPD_CMIRS_G2__Force = "B97-D3/def2-SVPD/CMIRS(G2) Force"
    B97_D3_def2_SVPD_CMIRS_G2__Geometry_Optimization = "B97-D3/def2-SVPD/CMIRS(G2) Geometry Optimization"
    B97_D3_def2_SVPD_CMIRS_G2__Frequency_Analysis = "B97-D3/def2-SVPD/CMIRS(G2) Frequency Analysis"
    B97_D3_def2_SVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_G2__Unknown = "B97-D3/def2-SVPD/CMIRS(G2) Unknown"
    B97_D3_def2_SVPD_CMIRS_DME__Single_Point = "B97-D3/def2-SVPD/CMIRS(DME) Single Point"
    B97_D3_def2_SVPD_CMIRS_DME__Force = "B97-D3/def2-SVPD/CMIRS(DME) Force"
    B97_D3_def2_SVPD_CMIRS_DME__Geometry_Optimization = "B97-D3/def2-SVPD/CMIRS(DME) Geometry Optimization"
    B97_D3_def2_SVPD_CMIRS_DME__Frequency_Analysis = "B97-D3/def2-SVPD/CMIRS(DME) Frequency Analysis"
    B97_D3_def2_SVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_DME__Unknown = "B97-D3/def2-SVPD/CMIRS(DME) Unknown"
    B97_D3_def2_SVPD_CMIRS_EC__Single_Point = "B97-D3/def2-SVPD/CMIRS(EC) Single Point"
    B97_D3_def2_SVPD_CMIRS_EC__Force = "B97-D3/def2-SVPD/CMIRS(EC) Force"
    B97_D3_def2_SVPD_CMIRS_EC__Geometry_Optimization = "B97-D3/def2-SVPD/CMIRS(EC) Geometry Optimization"
    B97_D3_def2_SVPD_CMIRS_EC__Frequency_Analysis = "B97-D3/def2-SVPD/CMIRS(EC) Frequency Analysis"
    B97_D3_def2_SVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_EC__Unknown = "B97-D3/def2-SVPD/CMIRS(EC) Unknown"
    B97_D3_def2_SVPD_CMIRS_EC_EMC__Single_Point = "B97-D3/def2-SVPD/CMIRS(EC/EMC) Single Point"
    B97_D3_def2_SVPD_CMIRS_EC_EMC__Force = "B97-D3/def2-SVPD/CMIRS(EC/EMC) Force"
    B97_D3_def2_SVPD_CMIRS_EC_EMC__Geometry_Optimization = "B97-D3/def2-SVPD/CMIRS(EC/EMC) Geometry Optimization"
    B97_D3_def2_SVPD_CMIRS_EC_EMC__Frequency_Analysis = "B97-D3/def2-SVPD/CMIRS(EC/EMC) Frequency Analysis"
    B97_D3_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_EC_EMC__Unknown = "B97-D3/def2-SVPD/CMIRS(EC/EMC) Unknown"
    B97_D3_def2_SVPD_CMIRS_PC__Single_Point = "B97-D3/def2-SVPD/CMIRS(PC) Single Point"
    B97_D3_def2_SVPD_CMIRS_PC__Force = "B97-D3/def2-SVPD/CMIRS(PC) Force"
    B97_D3_def2_SVPD_CMIRS_PC__Geometry_Optimization = "B97-D3/def2-SVPD/CMIRS(PC) Geometry Optimization"
    B97_D3_def2_SVPD_CMIRS_PC__Frequency_Analysis = "B97-D3/def2-SVPD/CMIRS(PC) Frequency Analysis"
    B97_D3_def2_SVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-SVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_SVPD_CMIRS_PC__Unknown = "B97-D3/def2-SVPD/CMIRS(PC) Unknown"
    B97_D3_def2_TZVP_VACUUM_Single_Point = "B97-D3/def2-TZVP/VACUUM Single Point"
    B97_D3_def2_TZVP_VACUUM_Force = "B97-D3/def2-TZVP/VACUUM Force"
    B97_D3_def2_TZVP_VACUUM_Geometry_Optimization = "B97-D3/def2-TZVP/VACUUM Geometry Optimization"
    B97_D3_def2_TZVP_VACUUM_Frequency_Analysis = "B97-D3/def2-TZVP/VACUUM Frequency Analysis"
    B97_D3_def2_TZVP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/VACUUM Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_VACUUM_Unknown = "B97-D3/def2-TZVP/VACUUM Unknown"
    B97_D3_def2_TZVP_PCM_WATER__Single_Point = "B97-D3/def2-TZVP/PCM(WATER) Single Point"
    B97_D3_def2_TZVP_PCM_WATER__Force = "B97-D3/def2-TZVP/PCM(WATER) Force"
    B97_D3_def2_TZVP_PCM_WATER__Geometry_Optimization = "B97-D3/def2-TZVP/PCM(WATER) Geometry Optimization"
    B97_D3_def2_TZVP_PCM_WATER__Frequency_Analysis = "B97-D3/def2-TZVP/PCM(WATER) Frequency Analysis"
    B97_D3_def2_TZVP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_WATER__Unknown = "B97-D3/def2-TZVP/PCM(WATER) Unknown"
    B97_D3_def2_TZVP_PCM_THF__Single_Point = "B97-D3/def2-TZVP/PCM(THF) Single Point"
    B97_D3_def2_TZVP_PCM_THF__Force = "B97-D3/def2-TZVP/PCM(THF) Force"
    B97_D3_def2_TZVP_PCM_THF__Geometry_Optimization = "B97-D3/def2-TZVP/PCM(THF) Geometry Optimization"
    B97_D3_def2_TZVP_PCM_THF__Frequency_Analysis = "B97-D3/def2-TZVP/PCM(THF) Frequency Analysis"
    B97_D3_def2_TZVP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_THF__Unknown = "B97-D3/def2-TZVP/PCM(THF) Unknown"
    B97_D3_def2_TZVP_PCM_G2__Single_Point = "B97-D3/def2-TZVP/PCM(G2) Single Point"
    B97_D3_def2_TZVP_PCM_G2__Force = "B97-D3/def2-TZVP/PCM(G2) Force"
    B97_D3_def2_TZVP_PCM_G2__Geometry_Optimization = "B97-D3/def2-TZVP/PCM(G2) Geometry Optimization"
    B97_D3_def2_TZVP_PCM_G2__Frequency_Analysis = "B97-D3/def2-TZVP/PCM(G2) Frequency Analysis"
    B97_D3_def2_TZVP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_G2__Unknown = "B97-D3/def2-TZVP/PCM(G2) Unknown"
    B97_D3_def2_TZVP_PCM_DME__Single_Point = "B97-D3/def2-TZVP/PCM(DME) Single Point"
    B97_D3_def2_TZVP_PCM_DME__Force = "B97-D3/def2-TZVP/PCM(DME) Force"
    B97_D3_def2_TZVP_PCM_DME__Geometry_Optimization = "B97-D3/def2-TZVP/PCM(DME) Geometry Optimization"
    B97_D3_def2_TZVP_PCM_DME__Frequency_Analysis = "B97-D3/def2-TZVP/PCM(DME) Frequency Analysis"
    B97_D3_def2_TZVP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_DME__Unknown = "B97-D3/def2-TZVP/PCM(DME) Unknown"
    B97_D3_def2_TZVP_PCM_EC__Single_Point = "B97-D3/def2-TZVP/PCM(EC) Single Point"
    B97_D3_def2_TZVP_PCM_EC__Force = "B97-D3/def2-TZVP/PCM(EC) Force"
    B97_D3_def2_TZVP_PCM_EC__Geometry_Optimization = "B97-D3/def2-TZVP/PCM(EC) Geometry Optimization"
    B97_D3_def2_TZVP_PCM_EC__Frequency_Analysis = "B97-D3/def2-TZVP/PCM(EC) Frequency Analysis"
    B97_D3_def2_TZVP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_EC__Unknown = "B97-D3/def2-TZVP/PCM(EC) Unknown"
    B97_D3_def2_TZVP_PCM_EC_EMC__Single_Point = "B97-D3/def2-TZVP/PCM(EC/EMC) Single Point"
    B97_D3_def2_TZVP_PCM_EC_EMC__Force = "B97-D3/def2-TZVP/PCM(EC/EMC) Force"
    B97_D3_def2_TZVP_PCM_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVP/PCM(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVP_PCM_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVP/PCM(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_EC_EMC__Unknown = "B97-D3/def2-TZVP/PCM(EC/EMC) Unknown"
    B97_D3_def2_TZVP_PCM_PC__Single_Point = "B97-D3/def2-TZVP/PCM(PC) Single Point"
    B97_D3_def2_TZVP_PCM_PC__Force = "B97-D3/def2-TZVP/PCM(PC) Force"
    B97_D3_def2_TZVP_PCM_PC__Geometry_Optimization = "B97-D3/def2-TZVP/PCM(PC) Geometry Optimization"
    B97_D3_def2_TZVP_PCM_PC__Frequency_Analysis = "B97-D3/def2-TZVP/PCM(PC) Frequency Analysis"
    B97_D3_def2_TZVP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_PCM_PC__Unknown = "B97-D3/def2-TZVP/PCM(PC) Unknown"
    B97_D3_def2_TZVP_ISOSVP_WATER__Single_Point = "B97-D3/def2-TZVP/ISOSVP(WATER) Single Point"
    B97_D3_def2_TZVP_ISOSVP_WATER__Force = "B97-D3/def2-TZVP/ISOSVP(WATER) Force"
    B97_D3_def2_TZVP_ISOSVP_WATER__Geometry_Optimization = "B97-D3/def2-TZVP/ISOSVP(WATER) Geometry Optimization"
    B97_D3_def2_TZVP_ISOSVP_WATER__Frequency_Analysis = "B97-D3/def2-TZVP/ISOSVP(WATER) Frequency Analysis"
    B97_D3_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_WATER__Unknown = "B97-D3/def2-TZVP/ISOSVP(WATER) Unknown"
    B97_D3_def2_TZVP_ISOSVP_THF__Single_Point = "B97-D3/def2-TZVP/ISOSVP(THF) Single Point"
    B97_D3_def2_TZVP_ISOSVP_THF__Force = "B97-D3/def2-TZVP/ISOSVP(THF) Force"
    B97_D3_def2_TZVP_ISOSVP_THF__Geometry_Optimization = "B97-D3/def2-TZVP/ISOSVP(THF) Geometry Optimization"
    B97_D3_def2_TZVP_ISOSVP_THF__Frequency_Analysis = "B97-D3/def2-TZVP/ISOSVP(THF) Frequency Analysis"
    B97_D3_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_THF__Unknown = "B97-D3/def2-TZVP/ISOSVP(THF) Unknown"
    B97_D3_def2_TZVP_ISOSVP_G2__Single_Point = "B97-D3/def2-TZVP/ISOSVP(G2) Single Point"
    B97_D3_def2_TZVP_ISOSVP_G2__Force = "B97-D3/def2-TZVP/ISOSVP(G2) Force"
    B97_D3_def2_TZVP_ISOSVP_G2__Geometry_Optimization = "B97-D3/def2-TZVP/ISOSVP(G2) Geometry Optimization"
    B97_D3_def2_TZVP_ISOSVP_G2__Frequency_Analysis = "B97-D3/def2-TZVP/ISOSVP(G2) Frequency Analysis"
    B97_D3_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_G2__Unknown = "B97-D3/def2-TZVP/ISOSVP(G2) Unknown"
    B97_D3_def2_TZVP_ISOSVP_DME__Single_Point = "B97-D3/def2-TZVP/ISOSVP(DME) Single Point"
    B97_D3_def2_TZVP_ISOSVP_DME__Force = "B97-D3/def2-TZVP/ISOSVP(DME) Force"
    B97_D3_def2_TZVP_ISOSVP_DME__Geometry_Optimization = "B97-D3/def2-TZVP/ISOSVP(DME) Geometry Optimization"
    B97_D3_def2_TZVP_ISOSVP_DME__Frequency_Analysis = "B97-D3/def2-TZVP/ISOSVP(DME) Frequency Analysis"
    B97_D3_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_DME__Unknown = "B97-D3/def2-TZVP/ISOSVP(DME) Unknown"
    B97_D3_def2_TZVP_ISOSVP_EC__Single_Point = "B97-D3/def2-TZVP/ISOSVP(EC) Single Point"
    B97_D3_def2_TZVP_ISOSVP_EC__Force = "B97-D3/def2-TZVP/ISOSVP(EC) Force"
    B97_D3_def2_TZVP_ISOSVP_EC__Geometry_Optimization = "B97-D3/def2-TZVP/ISOSVP(EC) Geometry Optimization"
    B97_D3_def2_TZVP_ISOSVP_EC__Frequency_Analysis = "B97-D3/def2-TZVP/ISOSVP(EC) Frequency Analysis"
    B97_D3_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_EC__Unknown = "B97-D3/def2-TZVP/ISOSVP(EC) Unknown"
    B97_D3_def2_TZVP_ISOSVP_EC_EMC__Single_Point = "B97-D3/def2-TZVP/ISOSVP(EC/EMC) Single Point"
    B97_D3_def2_TZVP_ISOSVP_EC_EMC__Force = "B97-D3/def2-TZVP/ISOSVP(EC/EMC) Force"
    B97_D3_def2_TZVP_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVP/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVP_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVP/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_EC_EMC__Unknown = "B97-D3/def2-TZVP/ISOSVP(EC/EMC) Unknown"
    B97_D3_def2_TZVP_ISOSVP_PC__Single_Point = "B97-D3/def2-TZVP/ISOSVP(PC) Single Point"
    B97_D3_def2_TZVP_ISOSVP_PC__Force = "B97-D3/def2-TZVP/ISOSVP(PC) Force"
    B97_D3_def2_TZVP_ISOSVP_PC__Geometry_Optimization = "B97-D3/def2-TZVP/ISOSVP(PC) Geometry Optimization"
    B97_D3_def2_TZVP_ISOSVP_PC__Frequency_Analysis = "B97-D3/def2-TZVP/ISOSVP(PC) Frequency Analysis"
    B97_D3_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_ISOSVP_PC__Unknown = "B97-D3/def2-TZVP/ISOSVP(PC) Unknown"
    B97_D3_def2_TZVP_SMD_WATER__Single_Point = "B97-D3/def2-TZVP/SMD(WATER) Single Point"
    B97_D3_def2_TZVP_SMD_WATER__Force = "B97-D3/def2-TZVP/SMD(WATER) Force"
    B97_D3_def2_TZVP_SMD_WATER__Geometry_Optimization = "B97-D3/def2-TZVP/SMD(WATER) Geometry Optimization"
    B97_D3_def2_TZVP_SMD_WATER__Frequency_Analysis = "B97-D3/def2-TZVP/SMD(WATER) Frequency Analysis"
    B97_D3_def2_TZVP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_WATER__Unknown = "B97-D3/def2-TZVP/SMD(WATER) Unknown"
    B97_D3_def2_TZVP_SMD_THF__Single_Point = "B97-D3/def2-TZVP/SMD(THF) Single Point"
    B97_D3_def2_TZVP_SMD_THF__Force = "B97-D3/def2-TZVP/SMD(THF) Force"
    B97_D3_def2_TZVP_SMD_THF__Geometry_Optimization = "B97-D3/def2-TZVP/SMD(THF) Geometry Optimization"
    B97_D3_def2_TZVP_SMD_THF__Frequency_Analysis = "B97-D3/def2-TZVP/SMD(THF) Frequency Analysis"
    B97_D3_def2_TZVP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_THF__Unknown = "B97-D3/def2-TZVP/SMD(THF) Unknown"
    B97_D3_def2_TZVP_SMD_G2__Single_Point = "B97-D3/def2-TZVP/SMD(G2) Single Point"
    B97_D3_def2_TZVP_SMD_G2__Force = "B97-D3/def2-TZVP/SMD(G2) Force"
    B97_D3_def2_TZVP_SMD_G2__Geometry_Optimization = "B97-D3/def2-TZVP/SMD(G2) Geometry Optimization"
    B97_D3_def2_TZVP_SMD_G2__Frequency_Analysis = "B97-D3/def2-TZVP/SMD(G2) Frequency Analysis"
    B97_D3_def2_TZVP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_G2__Unknown = "B97-D3/def2-TZVP/SMD(G2) Unknown"
    B97_D3_def2_TZVP_SMD_DME__Single_Point = "B97-D3/def2-TZVP/SMD(DME) Single Point"
    B97_D3_def2_TZVP_SMD_DME__Force = "B97-D3/def2-TZVP/SMD(DME) Force"
    B97_D3_def2_TZVP_SMD_DME__Geometry_Optimization = "B97-D3/def2-TZVP/SMD(DME) Geometry Optimization"
    B97_D3_def2_TZVP_SMD_DME__Frequency_Analysis = "B97-D3/def2-TZVP/SMD(DME) Frequency Analysis"
    B97_D3_def2_TZVP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_DME__Unknown = "B97-D3/def2-TZVP/SMD(DME) Unknown"
    B97_D3_def2_TZVP_SMD_EC__Single_Point = "B97-D3/def2-TZVP/SMD(EC) Single Point"
    B97_D3_def2_TZVP_SMD_EC__Force = "B97-D3/def2-TZVP/SMD(EC) Force"
    B97_D3_def2_TZVP_SMD_EC__Geometry_Optimization = "B97-D3/def2-TZVP/SMD(EC) Geometry Optimization"
    B97_D3_def2_TZVP_SMD_EC__Frequency_Analysis = "B97-D3/def2-TZVP/SMD(EC) Frequency Analysis"
    B97_D3_def2_TZVP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_EC__Unknown = "B97-D3/def2-TZVP/SMD(EC) Unknown"
    B97_D3_def2_TZVP_SMD_EC_EMC__Single_Point = "B97-D3/def2-TZVP/SMD(EC/EMC) Single Point"
    B97_D3_def2_TZVP_SMD_EC_EMC__Force = "B97-D3/def2-TZVP/SMD(EC/EMC) Force"
    B97_D3_def2_TZVP_SMD_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVP/SMD(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVP_SMD_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVP/SMD(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_EC_EMC__Unknown = "B97-D3/def2-TZVP/SMD(EC/EMC) Unknown"
    B97_D3_def2_TZVP_SMD_PC__Single_Point = "B97-D3/def2-TZVP/SMD(PC) Single Point"
    B97_D3_def2_TZVP_SMD_PC__Force = "B97-D3/def2-TZVP/SMD(PC) Force"
    B97_D3_def2_TZVP_SMD_PC__Geometry_Optimization = "B97-D3/def2-TZVP/SMD(PC) Geometry Optimization"
    B97_D3_def2_TZVP_SMD_PC__Frequency_Analysis = "B97-D3/def2-TZVP/SMD(PC) Frequency Analysis"
    B97_D3_def2_TZVP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_SMD_PC__Unknown = "B97-D3/def2-TZVP/SMD(PC) Unknown"
    B97_D3_def2_TZVP_CMIRS_WATER__Single_Point = "B97-D3/def2-TZVP/CMIRS(WATER) Single Point"
    B97_D3_def2_TZVP_CMIRS_WATER__Force = "B97-D3/def2-TZVP/CMIRS(WATER) Force"
    B97_D3_def2_TZVP_CMIRS_WATER__Geometry_Optimization = "B97-D3/def2-TZVP/CMIRS(WATER) Geometry Optimization"
    B97_D3_def2_TZVP_CMIRS_WATER__Frequency_Analysis = "B97-D3/def2-TZVP/CMIRS(WATER) Frequency Analysis"
    B97_D3_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_WATER__Unknown = "B97-D3/def2-TZVP/CMIRS(WATER) Unknown"
    B97_D3_def2_TZVP_CMIRS_THF__Single_Point = "B97-D3/def2-TZVP/CMIRS(THF) Single Point"
    B97_D3_def2_TZVP_CMIRS_THF__Force = "B97-D3/def2-TZVP/CMIRS(THF) Force"
    B97_D3_def2_TZVP_CMIRS_THF__Geometry_Optimization = "B97-D3/def2-TZVP/CMIRS(THF) Geometry Optimization"
    B97_D3_def2_TZVP_CMIRS_THF__Frequency_Analysis = "B97-D3/def2-TZVP/CMIRS(THF) Frequency Analysis"
    B97_D3_def2_TZVP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_THF__Unknown = "B97-D3/def2-TZVP/CMIRS(THF) Unknown"
    B97_D3_def2_TZVP_CMIRS_G2__Single_Point = "B97-D3/def2-TZVP/CMIRS(G2) Single Point"
    B97_D3_def2_TZVP_CMIRS_G2__Force = "B97-D3/def2-TZVP/CMIRS(G2) Force"
    B97_D3_def2_TZVP_CMIRS_G2__Geometry_Optimization = "B97-D3/def2-TZVP/CMIRS(G2) Geometry Optimization"
    B97_D3_def2_TZVP_CMIRS_G2__Frequency_Analysis = "B97-D3/def2-TZVP/CMIRS(G2) Frequency Analysis"
    B97_D3_def2_TZVP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_G2__Unknown = "B97-D3/def2-TZVP/CMIRS(G2) Unknown"
    B97_D3_def2_TZVP_CMIRS_DME__Single_Point = "B97-D3/def2-TZVP/CMIRS(DME) Single Point"
    B97_D3_def2_TZVP_CMIRS_DME__Force = "B97-D3/def2-TZVP/CMIRS(DME) Force"
    B97_D3_def2_TZVP_CMIRS_DME__Geometry_Optimization = "B97-D3/def2-TZVP/CMIRS(DME) Geometry Optimization"
    B97_D3_def2_TZVP_CMIRS_DME__Frequency_Analysis = "B97-D3/def2-TZVP/CMIRS(DME) Frequency Analysis"
    B97_D3_def2_TZVP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_DME__Unknown = "B97-D3/def2-TZVP/CMIRS(DME) Unknown"
    B97_D3_def2_TZVP_CMIRS_EC__Single_Point = "B97-D3/def2-TZVP/CMIRS(EC) Single Point"
    B97_D3_def2_TZVP_CMIRS_EC__Force = "B97-D3/def2-TZVP/CMIRS(EC) Force"
    B97_D3_def2_TZVP_CMIRS_EC__Geometry_Optimization = "B97-D3/def2-TZVP/CMIRS(EC) Geometry Optimization"
    B97_D3_def2_TZVP_CMIRS_EC__Frequency_Analysis = "B97-D3/def2-TZVP/CMIRS(EC) Frequency Analysis"
    B97_D3_def2_TZVP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_EC__Unknown = "B97-D3/def2-TZVP/CMIRS(EC) Unknown"
    B97_D3_def2_TZVP_CMIRS_EC_EMC__Single_Point = "B97-D3/def2-TZVP/CMIRS(EC/EMC) Single Point"
    B97_D3_def2_TZVP_CMIRS_EC_EMC__Force = "B97-D3/def2-TZVP/CMIRS(EC/EMC) Force"
    B97_D3_def2_TZVP_CMIRS_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVP/CMIRS(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVP_CMIRS_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVP/CMIRS(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_EC_EMC__Unknown = "B97-D3/def2-TZVP/CMIRS(EC/EMC) Unknown"
    B97_D3_def2_TZVP_CMIRS_PC__Single_Point = "B97-D3/def2-TZVP/CMIRS(PC) Single Point"
    B97_D3_def2_TZVP_CMIRS_PC__Force = "B97-D3/def2-TZVP/CMIRS(PC) Force"
    B97_D3_def2_TZVP_CMIRS_PC__Geometry_Optimization = "B97-D3/def2-TZVP/CMIRS(PC) Geometry Optimization"
    B97_D3_def2_TZVP_CMIRS_PC__Frequency_Analysis = "B97-D3/def2-TZVP/CMIRS(PC) Frequency Analysis"
    B97_D3_def2_TZVP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVP_CMIRS_PC__Unknown = "B97-D3/def2-TZVP/CMIRS(PC) Unknown"
    B97_D3_def2_TZVPD_VACUUM_Single_Point = "B97-D3/def2-TZVPD/VACUUM Single Point"
    B97_D3_def2_TZVPD_VACUUM_Force = "B97-D3/def2-TZVPD/VACUUM Force"
    B97_D3_def2_TZVPD_VACUUM_Geometry_Optimization = "B97-D3/def2-TZVPD/VACUUM Geometry Optimization"
    B97_D3_def2_TZVPD_VACUUM_Frequency_Analysis = "B97-D3/def2-TZVPD/VACUUM Frequency Analysis"
    B97_D3_def2_TZVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/VACUUM Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_VACUUM_Unknown = "B97-D3/def2-TZVPD/VACUUM Unknown"
    B97_D3_def2_TZVPD_PCM_WATER__Single_Point = "B97-D3/def2-TZVPD/PCM(WATER) Single Point"
    B97_D3_def2_TZVPD_PCM_WATER__Force = "B97-D3/def2-TZVPD/PCM(WATER) Force"
    B97_D3_def2_TZVPD_PCM_WATER__Geometry_Optimization = "B97-D3/def2-TZVPD/PCM(WATER) Geometry Optimization"
    B97_D3_def2_TZVPD_PCM_WATER__Frequency_Analysis = "B97-D3/def2-TZVPD/PCM(WATER) Frequency Analysis"
    B97_D3_def2_TZVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_WATER__Unknown = "B97-D3/def2-TZVPD/PCM(WATER) Unknown"
    B97_D3_def2_TZVPD_PCM_THF__Single_Point = "B97-D3/def2-TZVPD/PCM(THF) Single Point"
    B97_D3_def2_TZVPD_PCM_THF__Force = "B97-D3/def2-TZVPD/PCM(THF) Force"
    B97_D3_def2_TZVPD_PCM_THF__Geometry_Optimization = "B97-D3/def2-TZVPD/PCM(THF) Geometry Optimization"
    B97_D3_def2_TZVPD_PCM_THF__Frequency_Analysis = "B97-D3/def2-TZVPD/PCM(THF) Frequency Analysis"
    B97_D3_def2_TZVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_THF__Unknown = "B97-D3/def2-TZVPD/PCM(THF) Unknown"
    B97_D3_def2_TZVPD_PCM_G2__Single_Point = "B97-D3/def2-TZVPD/PCM(G2) Single Point"
    B97_D3_def2_TZVPD_PCM_G2__Force = "B97-D3/def2-TZVPD/PCM(G2) Force"
    B97_D3_def2_TZVPD_PCM_G2__Geometry_Optimization = "B97-D3/def2-TZVPD/PCM(G2) Geometry Optimization"
    B97_D3_def2_TZVPD_PCM_G2__Frequency_Analysis = "B97-D3/def2-TZVPD/PCM(G2) Frequency Analysis"
    B97_D3_def2_TZVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_G2__Unknown = "B97-D3/def2-TZVPD/PCM(G2) Unknown"
    B97_D3_def2_TZVPD_PCM_DME__Single_Point = "B97-D3/def2-TZVPD/PCM(DME) Single Point"
    B97_D3_def2_TZVPD_PCM_DME__Force = "B97-D3/def2-TZVPD/PCM(DME) Force"
    B97_D3_def2_TZVPD_PCM_DME__Geometry_Optimization = "B97-D3/def2-TZVPD/PCM(DME) Geometry Optimization"
    B97_D3_def2_TZVPD_PCM_DME__Frequency_Analysis = "B97-D3/def2-TZVPD/PCM(DME) Frequency Analysis"
    B97_D3_def2_TZVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_DME__Unknown = "B97-D3/def2-TZVPD/PCM(DME) Unknown"
    B97_D3_def2_TZVPD_PCM_EC__Single_Point = "B97-D3/def2-TZVPD/PCM(EC) Single Point"
    B97_D3_def2_TZVPD_PCM_EC__Force = "B97-D3/def2-TZVPD/PCM(EC) Force"
    B97_D3_def2_TZVPD_PCM_EC__Geometry_Optimization = "B97-D3/def2-TZVPD/PCM(EC) Geometry Optimization"
    B97_D3_def2_TZVPD_PCM_EC__Frequency_Analysis = "B97-D3/def2-TZVPD/PCM(EC) Frequency Analysis"
    B97_D3_def2_TZVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_EC__Unknown = "B97-D3/def2-TZVPD/PCM(EC) Unknown"
    B97_D3_def2_TZVPD_PCM_EC_EMC__Single_Point = "B97-D3/def2-TZVPD/PCM(EC/EMC) Single Point"
    B97_D3_def2_TZVPD_PCM_EC_EMC__Force = "B97-D3/def2-TZVPD/PCM(EC/EMC) Force"
    B97_D3_def2_TZVPD_PCM_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPD/PCM(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPD_PCM_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPD/PCM(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_EC_EMC__Unknown = "B97-D3/def2-TZVPD/PCM(EC/EMC) Unknown"
    B97_D3_def2_TZVPD_PCM_PC__Single_Point = "B97-D3/def2-TZVPD/PCM(PC) Single Point"
    B97_D3_def2_TZVPD_PCM_PC__Force = "B97-D3/def2-TZVPD/PCM(PC) Force"
    B97_D3_def2_TZVPD_PCM_PC__Geometry_Optimization = "B97-D3/def2-TZVPD/PCM(PC) Geometry Optimization"
    B97_D3_def2_TZVPD_PCM_PC__Frequency_Analysis = "B97-D3/def2-TZVPD/PCM(PC) Frequency Analysis"
    B97_D3_def2_TZVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_PCM_PC__Unknown = "B97-D3/def2-TZVPD/PCM(PC) Unknown"
    B97_D3_def2_TZVPD_ISOSVP_WATER__Single_Point = "B97-D3/def2-TZVPD/ISOSVP(WATER) Single Point"
    B97_D3_def2_TZVPD_ISOSVP_WATER__Force = "B97-D3/def2-TZVPD/ISOSVP(WATER) Force"
    B97_D3_def2_TZVPD_ISOSVP_WATER__Geometry_Optimization = "B97-D3/def2-TZVPD/ISOSVP(WATER) Geometry Optimization"
    B97_D3_def2_TZVPD_ISOSVP_WATER__Frequency_Analysis = "B97-D3/def2-TZVPD/ISOSVP(WATER) Frequency Analysis"
    B97_D3_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_WATER__Unknown = "B97-D3/def2-TZVPD/ISOSVP(WATER) Unknown"
    B97_D3_def2_TZVPD_ISOSVP_THF__Single_Point = "B97-D3/def2-TZVPD/ISOSVP(THF) Single Point"
    B97_D3_def2_TZVPD_ISOSVP_THF__Force = "B97-D3/def2-TZVPD/ISOSVP(THF) Force"
    B97_D3_def2_TZVPD_ISOSVP_THF__Geometry_Optimization = "B97-D3/def2-TZVPD/ISOSVP(THF) Geometry Optimization"
    B97_D3_def2_TZVPD_ISOSVP_THF__Frequency_Analysis = "B97-D3/def2-TZVPD/ISOSVP(THF) Frequency Analysis"
    B97_D3_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_THF__Unknown = "B97-D3/def2-TZVPD/ISOSVP(THF) Unknown"
    B97_D3_def2_TZVPD_ISOSVP_G2__Single_Point = "B97-D3/def2-TZVPD/ISOSVP(G2) Single Point"
    B97_D3_def2_TZVPD_ISOSVP_G2__Force = "B97-D3/def2-TZVPD/ISOSVP(G2) Force"
    B97_D3_def2_TZVPD_ISOSVP_G2__Geometry_Optimization = "B97-D3/def2-TZVPD/ISOSVP(G2) Geometry Optimization"
    B97_D3_def2_TZVPD_ISOSVP_G2__Frequency_Analysis = "B97-D3/def2-TZVPD/ISOSVP(G2) Frequency Analysis"
    B97_D3_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_G2__Unknown = "B97-D3/def2-TZVPD/ISOSVP(G2) Unknown"
    B97_D3_def2_TZVPD_ISOSVP_DME__Single_Point = "B97-D3/def2-TZVPD/ISOSVP(DME) Single Point"
    B97_D3_def2_TZVPD_ISOSVP_DME__Force = "B97-D3/def2-TZVPD/ISOSVP(DME) Force"
    B97_D3_def2_TZVPD_ISOSVP_DME__Geometry_Optimization = "B97-D3/def2-TZVPD/ISOSVP(DME) Geometry Optimization"
    B97_D3_def2_TZVPD_ISOSVP_DME__Frequency_Analysis = "B97-D3/def2-TZVPD/ISOSVP(DME) Frequency Analysis"
    B97_D3_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_DME__Unknown = "B97-D3/def2-TZVPD/ISOSVP(DME) Unknown"
    B97_D3_def2_TZVPD_ISOSVP_EC__Single_Point = "B97-D3/def2-TZVPD/ISOSVP(EC) Single Point"
    B97_D3_def2_TZVPD_ISOSVP_EC__Force = "B97-D3/def2-TZVPD/ISOSVP(EC) Force"
    B97_D3_def2_TZVPD_ISOSVP_EC__Geometry_Optimization = "B97-D3/def2-TZVPD/ISOSVP(EC) Geometry Optimization"
    B97_D3_def2_TZVPD_ISOSVP_EC__Frequency_Analysis = "B97-D3/def2-TZVPD/ISOSVP(EC) Frequency Analysis"
    B97_D3_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_EC__Unknown = "B97-D3/def2-TZVPD/ISOSVP(EC) Unknown"
    B97_D3_def2_TZVPD_ISOSVP_EC_EMC__Single_Point = "B97-D3/def2-TZVPD/ISOSVP(EC/EMC) Single Point"
    B97_D3_def2_TZVPD_ISOSVP_EC_EMC__Force = "B97-D3/def2-TZVPD/ISOSVP(EC/EMC) Force"
    B97_D3_def2_TZVPD_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPD/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_EC_EMC__Unknown = "B97-D3/def2-TZVPD/ISOSVP(EC/EMC) Unknown"
    B97_D3_def2_TZVPD_ISOSVP_PC__Single_Point = "B97-D3/def2-TZVPD/ISOSVP(PC) Single Point"
    B97_D3_def2_TZVPD_ISOSVP_PC__Force = "B97-D3/def2-TZVPD/ISOSVP(PC) Force"
    B97_D3_def2_TZVPD_ISOSVP_PC__Geometry_Optimization = "B97-D3/def2-TZVPD/ISOSVP(PC) Geometry Optimization"
    B97_D3_def2_TZVPD_ISOSVP_PC__Frequency_Analysis = "B97-D3/def2-TZVPD/ISOSVP(PC) Frequency Analysis"
    B97_D3_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_ISOSVP_PC__Unknown = "B97-D3/def2-TZVPD/ISOSVP(PC) Unknown"
    B97_D3_def2_TZVPD_SMD_WATER__Single_Point = "B97-D3/def2-TZVPD/SMD(WATER) Single Point"
    B97_D3_def2_TZVPD_SMD_WATER__Force = "B97-D3/def2-TZVPD/SMD(WATER) Force"
    B97_D3_def2_TZVPD_SMD_WATER__Geometry_Optimization = "B97-D3/def2-TZVPD/SMD(WATER) Geometry Optimization"
    B97_D3_def2_TZVPD_SMD_WATER__Frequency_Analysis = "B97-D3/def2-TZVPD/SMD(WATER) Frequency Analysis"
    B97_D3_def2_TZVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_WATER__Unknown = "B97-D3/def2-TZVPD/SMD(WATER) Unknown"
    B97_D3_def2_TZVPD_SMD_THF__Single_Point = "B97-D3/def2-TZVPD/SMD(THF) Single Point"
    B97_D3_def2_TZVPD_SMD_THF__Force = "B97-D3/def2-TZVPD/SMD(THF) Force"
    B97_D3_def2_TZVPD_SMD_THF__Geometry_Optimization = "B97-D3/def2-TZVPD/SMD(THF) Geometry Optimization"
    B97_D3_def2_TZVPD_SMD_THF__Frequency_Analysis = "B97-D3/def2-TZVPD/SMD(THF) Frequency Analysis"
    B97_D3_def2_TZVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_THF__Unknown = "B97-D3/def2-TZVPD/SMD(THF) Unknown"
    B97_D3_def2_TZVPD_SMD_G2__Single_Point = "B97-D3/def2-TZVPD/SMD(G2) Single Point"
    B97_D3_def2_TZVPD_SMD_G2__Force = "B97-D3/def2-TZVPD/SMD(G2) Force"
    B97_D3_def2_TZVPD_SMD_G2__Geometry_Optimization = "B97-D3/def2-TZVPD/SMD(G2) Geometry Optimization"
    B97_D3_def2_TZVPD_SMD_G2__Frequency_Analysis = "B97-D3/def2-TZVPD/SMD(G2) Frequency Analysis"
    B97_D3_def2_TZVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_G2__Unknown = "B97-D3/def2-TZVPD/SMD(G2) Unknown"
    B97_D3_def2_TZVPD_SMD_DME__Single_Point = "B97-D3/def2-TZVPD/SMD(DME) Single Point"
    B97_D3_def2_TZVPD_SMD_DME__Force = "B97-D3/def2-TZVPD/SMD(DME) Force"
    B97_D3_def2_TZVPD_SMD_DME__Geometry_Optimization = "B97-D3/def2-TZVPD/SMD(DME) Geometry Optimization"
    B97_D3_def2_TZVPD_SMD_DME__Frequency_Analysis = "B97-D3/def2-TZVPD/SMD(DME) Frequency Analysis"
    B97_D3_def2_TZVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_DME__Unknown = "B97-D3/def2-TZVPD/SMD(DME) Unknown"
    B97_D3_def2_TZVPD_SMD_EC__Single_Point = "B97-D3/def2-TZVPD/SMD(EC) Single Point"
    B97_D3_def2_TZVPD_SMD_EC__Force = "B97-D3/def2-TZVPD/SMD(EC) Force"
    B97_D3_def2_TZVPD_SMD_EC__Geometry_Optimization = "B97-D3/def2-TZVPD/SMD(EC) Geometry Optimization"
    B97_D3_def2_TZVPD_SMD_EC__Frequency_Analysis = "B97-D3/def2-TZVPD/SMD(EC) Frequency Analysis"
    B97_D3_def2_TZVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_EC__Unknown = "B97-D3/def2-TZVPD/SMD(EC) Unknown"
    B97_D3_def2_TZVPD_SMD_EC_EMC__Single_Point = "B97-D3/def2-TZVPD/SMD(EC/EMC) Single Point"
    B97_D3_def2_TZVPD_SMD_EC_EMC__Force = "B97-D3/def2-TZVPD/SMD(EC/EMC) Force"
    B97_D3_def2_TZVPD_SMD_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPD/SMD(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPD_SMD_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPD/SMD(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_EC_EMC__Unknown = "B97-D3/def2-TZVPD/SMD(EC/EMC) Unknown"
    B97_D3_def2_TZVPD_SMD_PC__Single_Point = "B97-D3/def2-TZVPD/SMD(PC) Single Point"
    B97_D3_def2_TZVPD_SMD_PC__Force = "B97-D3/def2-TZVPD/SMD(PC) Force"
    B97_D3_def2_TZVPD_SMD_PC__Geometry_Optimization = "B97-D3/def2-TZVPD/SMD(PC) Geometry Optimization"
    B97_D3_def2_TZVPD_SMD_PC__Frequency_Analysis = "B97-D3/def2-TZVPD/SMD(PC) Frequency Analysis"
    B97_D3_def2_TZVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_SMD_PC__Unknown = "B97-D3/def2-TZVPD/SMD(PC) Unknown"
    B97_D3_def2_TZVPD_CMIRS_WATER__Single_Point = "B97-D3/def2-TZVPD/CMIRS(WATER) Single Point"
    B97_D3_def2_TZVPD_CMIRS_WATER__Force = "B97-D3/def2-TZVPD/CMIRS(WATER) Force"
    B97_D3_def2_TZVPD_CMIRS_WATER__Geometry_Optimization = "B97-D3/def2-TZVPD/CMIRS(WATER) Geometry Optimization"
    B97_D3_def2_TZVPD_CMIRS_WATER__Frequency_Analysis = "B97-D3/def2-TZVPD/CMIRS(WATER) Frequency Analysis"
    B97_D3_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_WATER__Unknown = "B97-D3/def2-TZVPD/CMIRS(WATER) Unknown"
    B97_D3_def2_TZVPD_CMIRS_THF__Single_Point = "B97-D3/def2-TZVPD/CMIRS(THF) Single Point"
    B97_D3_def2_TZVPD_CMIRS_THF__Force = "B97-D3/def2-TZVPD/CMIRS(THF) Force"
    B97_D3_def2_TZVPD_CMIRS_THF__Geometry_Optimization = "B97-D3/def2-TZVPD/CMIRS(THF) Geometry Optimization"
    B97_D3_def2_TZVPD_CMIRS_THF__Frequency_Analysis = "B97-D3/def2-TZVPD/CMIRS(THF) Frequency Analysis"
    B97_D3_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_THF__Unknown = "B97-D3/def2-TZVPD/CMIRS(THF) Unknown"
    B97_D3_def2_TZVPD_CMIRS_G2__Single_Point = "B97-D3/def2-TZVPD/CMIRS(G2) Single Point"
    B97_D3_def2_TZVPD_CMIRS_G2__Force = "B97-D3/def2-TZVPD/CMIRS(G2) Force"
    B97_D3_def2_TZVPD_CMIRS_G2__Geometry_Optimization = "B97-D3/def2-TZVPD/CMIRS(G2) Geometry Optimization"
    B97_D3_def2_TZVPD_CMIRS_G2__Frequency_Analysis = "B97-D3/def2-TZVPD/CMIRS(G2) Frequency Analysis"
    B97_D3_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_G2__Unknown = "B97-D3/def2-TZVPD/CMIRS(G2) Unknown"
    B97_D3_def2_TZVPD_CMIRS_DME__Single_Point = "B97-D3/def2-TZVPD/CMIRS(DME) Single Point"
    B97_D3_def2_TZVPD_CMIRS_DME__Force = "B97-D3/def2-TZVPD/CMIRS(DME) Force"
    B97_D3_def2_TZVPD_CMIRS_DME__Geometry_Optimization = "B97-D3/def2-TZVPD/CMIRS(DME) Geometry Optimization"
    B97_D3_def2_TZVPD_CMIRS_DME__Frequency_Analysis = "B97-D3/def2-TZVPD/CMIRS(DME) Frequency Analysis"
    B97_D3_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_DME__Unknown = "B97-D3/def2-TZVPD/CMIRS(DME) Unknown"
    B97_D3_def2_TZVPD_CMIRS_EC__Single_Point = "B97-D3/def2-TZVPD/CMIRS(EC) Single Point"
    B97_D3_def2_TZVPD_CMIRS_EC__Force = "B97-D3/def2-TZVPD/CMIRS(EC) Force"
    B97_D3_def2_TZVPD_CMIRS_EC__Geometry_Optimization = "B97-D3/def2-TZVPD/CMIRS(EC) Geometry Optimization"
    B97_D3_def2_TZVPD_CMIRS_EC__Frequency_Analysis = "B97-D3/def2-TZVPD/CMIRS(EC) Frequency Analysis"
    B97_D3_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_EC__Unknown = "B97-D3/def2-TZVPD/CMIRS(EC) Unknown"
    B97_D3_def2_TZVPD_CMIRS_EC_EMC__Single_Point = "B97-D3/def2-TZVPD/CMIRS(EC/EMC) Single Point"
    B97_D3_def2_TZVPD_CMIRS_EC_EMC__Force = "B97-D3/def2-TZVPD/CMIRS(EC/EMC) Force"
    B97_D3_def2_TZVPD_CMIRS_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPD/CMIRS(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPD_CMIRS_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPD/CMIRS(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_EC_EMC__Unknown = "B97-D3/def2-TZVPD/CMIRS(EC/EMC) Unknown"
    B97_D3_def2_TZVPD_CMIRS_PC__Single_Point = "B97-D3/def2-TZVPD/CMIRS(PC) Single Point"
    B97_D3_def2_TZVPD_CMIRS_PC__Force = "B97-D3/def2-TZVPD/CMIRS(PC) Force"
    B97_D3_def2_TZVPD_CMIRS_PC__Geometry_Optimization = "B97-D3/def2-TZVPD/CMIRS(PC) Geometry Optimization"
    B97_D3_def2_TZVPD_CMIRS_PC__Frequency_Analysis = "B97-D3/def2-TZVPD/CMIRS(PC) Frequency Analysis"
    B97_D3_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPD_CMIRS_PC__Unknown = "B97-D3/def2-TZVPD/CMIRS(PC) Unknown"
    B97_D3_def2_TZVPP_VACUUM_Single_Point = "B97-D3/def2-TZVPP/VACUUM Single Point"
    B97_D3_def2_TZVPP_VACUUM_Force = "B97-D3/def2-TZVPP/VACUUM Force"
    B97_D3_def2_TZVPP_VACUUM_Geometry_Optimization = "B97-D3/def2-TZVPP/VACUUM Geometry Optimization"
    B97_D3_def2_TZVPP_VACUUM_Frequency_Analysis = "B97-D3/def2-TZVPP/VACUUM Frequency Analysis"
    B97_D3_def2_TZVPP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/VACUUM Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_VACUUM_Unknown = "B97-D3/def2-TZVPP/VACUUM Unknown"
    B97_D3_def2_TZVPP_PCM_WATER__Single_Point = "B97-D3/def2-TZVPP/PCM(WATER) Single Point"
    B97_D3_def2_TZVPP_PCM_WATER__Force = "B97-D3/def2-TZVPP/PCM(WATER) Force"
    B97_D3_def2_TZVPP_PCM_WATER__Geometry_Optimization = "B97-D3/def2-TZVPP/PCM(WATER) Geometry Optimization"
    B97_D3_def2_TZVPP_PCM_WATER__Frequency_Analysis = "B97-D3/def2-TZVPP/PCM(WATER) Frequency Analysis"
    B97_D3_def2_TZVPP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_WATER__Unknown = "B97-D3/def2-TZVPP/PCM(WATER) Unknown"
    B97_D3_def2_TZVPP_PCM_THF__Single_Point = "B97-D3/def2-TZVPP/PCM(THF) Single Point"
    B97_D3_def2_TZVPP_PCM_THF__Force = "B97-D3/def2-TZVPP/PCM(THF) Force"
    B97_D3_def2_TZVPP_PCM_THF__Geometry_Optimization = "B97-D3/def2-TZVPP/PCM(THF) Geometry Optimization"
    B97_D3_def2_TZVPP_PCM_THF__Frequency_Analysis = "B97-D3/def2-TZVPP/PCM(THF) Frequency Analysis"
    B97_D3_def2_TZVPP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_THF__Unknown = "B97-D3/def2-TZVPP/PCM(THF) Unknown"
    B97_D3_def2_TZVPP_PCM_G2__Single_Point = "B97-D3/def2-TZVPP/PCM(G2) Single Point"
    B97_D3_def2_TZVPP_PCM_G2__Force = "B97-D3/def2-TZVPP/PCM(G2) Force"
    B97_D3_def2_TZVPP_PCM_G2__Geometry_Optimization = "B97-D3/def2-TZVPP/PCM(G2) Geometry Optimization"
    B97_D3_def2_TZVPP_PCM_G2__Frequency_Analysis = "B97-D3/def2-TZVPP/PCM(G2) Frequency Analysis"
    B97_D3_def2_TZVPP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_G2__Unknown = "B97-D3/def2-TZVPP/PCM(G2) Unknown"
    B97_D3_def2_TZVPP_PCM_DME__Single_Point = "B97-D3/def2-TZVPP/PCM(DME) Single Point"
    B97_D3_def2_TZVPP_PCM_DME__Force = "B97-D3/def2-TZVPP/PCM(DME) Force"
    B97_D3_def2_TZVPP_PCM_DME__Geometry_Optimization = "B97-D3/def2-TZVPP/PCM(DME) Geometry Optimization"
    B97_D3_def2_TZVPP_PCM_DME__Frequency_Analysis = "B97-D3/def2-TZVPP/PCM(DME) Frequency Analysis"
    B97_D3_def2_TZVPP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_DME__Unknown = "B97-D3/def2-TZVPP/PCM(DME) Unknown"
    B97_D3_def2_TZVPP_PCM_EC__Single_Point = "B97-D3/def2-TZVPP/PCM(EC) Single Point"
    B97_D3_def2_TZVPP_PCM_EC__Force = "B97-D3/def2-TZVPP/PCM(EC) Force"
    B97_D3_def2_TZVPP_PCM_EC__Geometry_Optimization = "B97-D3/def2-TZVPP/PCM(EC) Geometry Optimization"
    B97_D3_def2_TZVPP_PCM_EC__Frequency_Analysis = "B97-D3/def2-TZVPP/PCM(EC) Frequency Analysis"
    B97_D3_def2_TZVPP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_EC__Unknown = "B97-D3/def2-TZVPP/PCM(EC) Unknown"
    B97_D3_def2_TZVPP_PCM_EC_EMC__Single_Point = "B97-D3/def2-TZVPP/PCM(EC/EMC) Single Point"
    B97_D3_def2_TZVPP_PCM_EC_EMC__Force = "B97-D3/def2-TZVPP/PCM(EC/EMC) Force"
    B97_D3_def2_TZVPP_PCM_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPP/PCM(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPP_PCM_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPP/PCM(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_EC_EMC__Unknown = "B97-D3/def2-TZVPP/PCM(EC/EMC) Unknown"
    B97_D3_def2_TZVPP_PCM_PC__Single_Point = "B97-D3/def2-TZVPP/PCM(PC) Single Point"
    B97_D3_def2_TZVPP_PCM_PC__Force = "B97-D3/def2-TZVPP/PCM(PC) Force"
    B97_D3_def2_TZVPP_PCM_PC__Geometry_Optimization = "B97-D3/def2-TZVPP/PCM(PC) Geometry Optimization"
    B97_D3_def2_TZVPP_PCM_PC__Frequency_Analysis = "B97-D3/def2-TZVPP/PCM(PC) Frequency Analysis"
    B97_D3_def2_TZVPP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_PCM_PC__Unknown = "B97-D3/def2-TZVPP/PCM(PC) Unknown"
    B97_D3_def2_TZVPP_ISOSVP_WATER__Single_Point = "B97-D3/def2-TZVPP/ISOSVP(WATER) Single Point"
    B97_D3_def2_TZVPP_ISOSVP_WATER__Force = "B97-D3/def2-TZVPP/ISOSVP(WATER) Force"
    B97_D3_def2_TZVPP_ISOSVP_WATER__Geometry_Optimization = "B97-D3/def2-TZVPP/ISOSVP(WATER) Geometry Optimization"
    B97_D3_def2_TZVPP_ISOSVP_WATER__Frequency_Analysis = "B97-D3/def2-TZVPP/ISOSVP(WATER) Frequency Analysis"
    B97_D3_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_WATER__Unknown = "B97-D3/def2-TZVPP/ISOSVP(WATER) Unknown"
    B97_D3_def2_TZVPP_ISOSVP_THF__Single_Point = "B97-D3/def2-TZVPP/ISOSVP(THF) Single Point"
    B97_D3_def2_TZVPP_ISOSVP_THF__Force = "B97-D3/def2-TZVPP/ISOSVP(THF) Force"
    B97_D3_def2_TZVPP_ISOSVP_THF__Geometry_Optimization = "B97-D3/def2-TZVPP/ISOSVP(THF) Geometry Optimization"
    B97_D3_def2_TZVPP_ISOSVP_THF__Frequency_Analysis = "B97-D3/def2-TZVPP/ISOSVP(THF) Frequency Analysis"
    B97_D3_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_THF__Unknown = "B97-D3/def2-TZVPP/ISOSVP(THF) Unknown"
    B97_D3_def2_TZVPP_ISOSVP_G2__Single_Point = "B97-D3/def2-TZVPP/ISOSVP(G2) Single Point"
    B97_D3_def2_TZVPP_ISOSVP_G2__Force = "B97-D3/def2-TZVPP/ISOSVP(G2) Force"
    B97_D3_def2_TZVPP_ISOSVP_G2__Geometry_Optimization = "B97-D3/def2-TZVPP/ISOSVP(G2) Geometry Optimization"
    B97_D3_def2_TZVPP_ISOSVP_G2__Frequency_Analysis = "B97-D3/def2-TZVPP/ISOSVP(G2) Frequency Analysis"
    B97_D3_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_G2__Unknown = "B97-D3/def2-TZVPP/ISOSVP(G2) Unknown"
    B97_D3_def2_TZVPP_ISOSVP_DME__Single_Point = "B97-D3/def2-TZVPP/ISOSVP(DME) Single Point"
    B97_D3_def2_TZVPP_ISOSVP_DME__Force = "B97-D3/def2-TZVPP/ISOSVP(DME) Force"
    B97_D3_def2_TZVPP_ISOSVP_DME__Geometry_Optimization = "B97-D3/def2-TZVPP/ISOSVP(DME) Geometry Optimization"
    B97_D3_def2_TZVPP_ISOSVP_DME__Frequency_Analysis = "B97-D3/def2-TZVPP/ISOSVP(DME) Frequency Analysis"
    B97_D3_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_DME__Unknown = "B97-D3/def2-TZVPP/ISOSVP(DME) Unknown"
    B97_D3_def2_TZVPP_ISOSVP_EC__Single_Point = "B97-D3/def2-TZVPP/ISOSVP(EC) Single Point"
    B97_D3_def2_TZVPP_ISOSVP_EC__Force = "B97-D3/def2-TZVPP/ISOSVP(EC) Force"
    B97_D3_def2_TZVPP_ISOSVP_EC__Geometry_Optimization = "B97-D3/def2-TZVPP/ISOSVP(EC) Geometry Optimization"
    B97_D3_def2_TZVPP_ISOSVP_EC__Frequency_Analysis = "B97-D3/def2-TZVPP/ISOSVP(EC) Frequency Analysis"
    B97_D3_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_EC__Unknown = "B97-D3/def2-TZVPP/ISOSVP(EC) Unknown"
    B97_D3_def2_TZVPP_ISOSVP_EC_EMC__Single_Point = "B97-D3/def2-TZVPP/ISOSVP(EC/EMC) Single Point"
    B97_D3_def2_TZVPP_ISOSVP_EC_EMC__Force = "B97-D3/def2-TZVPP/ISOSVP(EC/EMC) Force"
    B97_D3_def2_TZVPP_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPP/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPP/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_EC_EMC__Unknown = "B97-D3/def2-TZVPP/ISOSVP(EC/EMC) Unknown"
    B97_D3_def2_TZVPP_ISOSVP_PC__Single_Point = "B97-D3/def2-TZVPP/ISOSVP(PC) Single Point"
    B97_D3_def2_TZVPP_ISOSVP_PC__Force = "B97-D3/def2-TZVPP/ISOSVP(PC) Force"
    B97_D3_def2_TZVPP_ISOSVP_PC__Geometry_Optimization = "B97-D3/def2-TZVPP/ISOSVP(PC) Geometry Optimization"
    B97_D3_def2_TZVPP_ISOSVP_PC__Frequency_Analysis = "B97-D3/def2-TZVPP/ISOSVP(PC) Frequency Analysis"
    B97_D3_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_ISOSVP_PC__Unknown = "B97-D3/def2-TZVPP/ISOSVP(PC) Unknown"
    B97_D3_def2_TZVPP_SMD_WATER__Single_Point = "B97-D3/def2-TZVPP/SMD(WATER) Single Point"
    B97_D3_def2_TZVPP_SMD_WATER__Force = "B97-D3/def2-TZVPP/SMD(WATER) Force"
    B97_D3_def2_TZVPP_SMD_WATER__Geometry_Optimization = "B97-D3/def2-TZVPP/SMD(WATER) Geometry Optimization"
    B97_D3_def2_TZVPP_SMD_WATER__Frequency_Analysis = "B97-D3/def2-TZVPP/SMD(WATER) Frequency Analysis"
    B97_D3_def2_TZVPP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_WATER__Unknown = "B97-D3/def2-TZVPP/SMD(WATER) Unknown"
    B97_D3_def2_TZVPP_SMD_THF__Single_Point = "B97-D3/def2-TZVPP/SMD(THF) Single Point"
    B97_D3_def2_TZVPP_SMD_THF__Force = "B97-D3/def2-TZVPP/SMD(THF) Force"
    B97_D3_def2_TZVPP_SMD_THF__Geometry_Optimization = "B97-D3/def2-TZVPP/SMD(THF) Geometry Optimization"
    B97_D3_def2_TZVPP_SMD_THF__Frequency_Analysis = "B97-D3/def2-TZVPP/SMD(THF) Frequency Analysis"
    B97_D3_def2_TZVPP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_THF__Unknown = "B97-D3/def2-TZVPP/SMD(THF) Unknown"
    B97_D3_def2_TZVPP_SMD_G2__Single_Point = "B97-D3/def2-TZVPP/SMD(G2) Single Point"
    B97_D3_def2_TZVPP_SMD_G2__Force = "B97-D3/def2-TZVPP/SMD(G2) Force"
    B97_D3_def2_TZVPP_SMD_G2__Geometry_Optimization = "B97-D3/def2-TZVPP/SMD(G2) Geometry Optimization"
    B97_D3_def2_TZVPP_SMD_G2__Frequency_Analysis = "B97-D3/def2-TZVPP/SMD(G2) Frequency Analysis"
    B97_D3_def2_TZVPP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_G2__Unknown = "B97-D3/def2-TZVPP/SMD(G2) Unknown"
    B97_D3_def2_TZVPP_SMD_DME__Single_Point = "B97-D3/def2-TZVPP/SMD(DME) Single Point"
    B97_D3_def2_TZVPP_SMD_DME__Force = "B97-D3/def2-TZVPP/SMD(DME) Force"
    B97_D3_def2_TZVPP_SMD_DME__Geometry_Optimization = "B97-D3/def2-TZVPP/SMD(DME) Geometry Optimization"
    B97_D3_def2_TZVPP_SMD_DME__Frequency_Analysis = "B97-D3/def2-TZVPP/SMD(DME) Frequency Analysis"
    B97_D3_def2_TZVPP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_DME__Unknown = "B97-D3/def2-TZVPP/SMD(DME) Unknown"
    B97_D3_def2_TZVPP_SMD_EC__Single_Point = "B97-D3/def2-TZVPP/SMD(EC) Single Point"
    B97_D3_def2_TZVPP_SMD_EC__Force = "B97-D3/def2-TZVPP/SMD(EC) Force"
    B97_D3_def2_TZVPP_SMD_EC__Geometry_Optimization = "B97-D3/def2-TZVPP/SMD(EC) Geometry Optimization"
    B97_D3_def2_TZVPP_SMD_EC__Frequency_Analysis = "B97-D3/def2-TZVPP/SMD(EC) Frequency Analysis"
    B97_D3_def2_TZVPP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_EC__Unknown = "B97-D3/def2-TZVPP/SMD(EC) Unknown"
    B97_D3_def2_TZVPP_SMD_EC_EMC__Single_Point = "B97-D3/def2-TZVPP/SMD(EC/EMC) Single Point"
    B97_D3_def2_TZVPP_SMD_EC_EMC__Force = "B97-D3/def2-TZVPP/SMD(EC/EMC) Force"
    B97_D3_def2_TZVPP_SMD_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPP/SMD(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPP_SMD_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPP/SMD(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_EC_EMC__Unknown = "B97-D3/def2-TZVPP/SMD(EC/EMC) Unknown"
    B97_D3_def2_TZVPP_SMD_PC__Single_Point = "B97-D3/def2-TZVPP/SMD(PC) Single Point"
    B97_D3_def2_TZVPP_SMD_PC__Force = "B97-D3/def2-TZVPP/SMD(PC) Force"
    B97_D3_def2_TZVPP_SMD_PC__Geometry_Optimization = "B97-D3/def2-TZVPP/SMD(PC) Geometry Optimization"
    B97_D3_def2_TZVPP_SMD_PC__Frequency_Analysis = "B97-D3/def2-TZVPP/SMD(PC) Frequency Analysis"
    B97_D3_def2_TZVPP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_SMD_PC__Unknown = "B97-D3/def2-TZVPP/SMD(PC) Unknown"
    B97_D3_def2_TZVPP_CMIRS_WATER__Single_Point = "B97-D3/def2-TZVPP/CMIRS(WATER) Single Point"
    B97_D3_def2_TZVPP_CMIRS_WATER__Force = "B97-D3/def2-TZVPP/CMIRS(WATER) Force"
    B97_D3_def2_TZVPP_CMIRS_WATER__Geometry_Optimization = "B97-D3/def2-TZVPP/CMIRS(WATER) Geometry Optimization"
    B97_D3_def2_TZVPP_CMIRS_WATER__Frequency_Analysis = "B97-D3/def2-TZVPP/CMIRS(WATER) Frequency Analysis"
    B97_D3_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_WATER__Unknown = "B97-D3/def2-TZVPP/CMIRS(WATER) Unknown"
    B97_D3_def2_TZVPP_CMIRS_THF__Single_Point = "B97-D3/def2-TZVPP/CMIRS(THF) Single Point"
    B97_D3_def2_TZVPP_CMIRS_THF__Force = "B97-D3/def2-TZVPP/CMIRS(THF) Force"
    B97_D3_def2_TZVPP_CMIRS_THF__Geometry_Optimization = "B97-D3/def2-TZVPP/CMIRS(THF) Geometry Optimization"
    B97_D3_def2_TZVPP_CMIRS_THF__Frequency_Analysis = "B97-D3/def2-TZVPP/CMIRS(THF) Frequency Analysis"
    B97_D3_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_THF__Unknown = "B97-D3/def2-TZVPP/CMIRS(THF) Unknown"
    B97_D3_def2_TZVPP_CMIRS_G2__Single_Point = "B97-D3/def2-TZVPP/CMIRS(G2) Single Point"
    B97_D3_def2_TZVPP_CMIRS_G2__Force = "B97-D3/def2-TZVPP/CMIRS(G2) Force"
    B97_D3_def2_TZVPP_CMIRS_G2__Geometry_Optimization = "B97-D3/def2-TZVPP/CMIRS(G2) Geometry Optimization"
    B97_D3_def2_TZVPP_CMIRS_G2__Frequency_Analysis = "B97-D3/def2-TZVPP/CMIRS(G2) Frequency Analysis"
    B97_D3_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_G2__Unknown = "B97-D3/def2-TZVPP/CMIRS(G2) Unknown"
    B97_D3_def2_TZVPP_CMIRS_DME__Single_Point = "B97-D3/def2-TZVPP/CMIRS(DME) Single Point"
    B97_D3_def2_TZVPP_CMIRS_DME__Force = "B97-D3/def2-TZVPP/CMIRS(DME) Force"
    B97_D3_def2_TZVPP_CMIRS_DME__Geometry_Optimization = "B97-D3/def2-TZVPP/CMIRS(DME) Geometry Optimization"
    B97_D3_def2_TZVPP_CMIRS_DME__Frequency_Analysis = "B97-D3/def2-TZVPP/CMIRS(DME) Frequency Analysis"
    B97_D3_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_DME__Unknown = "B97-D3/def2-TZVPP/CMIRS(DME) Unknown"
    B97_D3_def2_TZVPP_CMIRS_EC__Single_Point = "B97-D3/def2-TZVPP/CMIRS(EC) Single Point"
    B97_D3_def2_TZVPP_CMIRS_EC__Force = "B97-D3/def2-TZVPP/CMIRS(EC) Force"
    B97_D3_def2_TZVPP_CMIRS_EC__Geometry_Optimization = "B97-D3/def2-TZVPP/CMIRS(EC) Geometry Optimization"
    B97_D3_def2_TZVPP_CMIRS_EC__Frequency_Analysis = "B97-D3/def2-TZVPP/CMIRS(EC) Frequency Analysis"
    B97_D3_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_EC__Unknown = "B97-D3/def2-TZVPP/CMIRS(EC) Unknown"
    B97_D3_def2_TZVPP_CMIRS_EC_EMC__Single_Point = "B97-D3/def2-TZVPP/CMIRS(EC/EMC) Single Point"
    B97_D3_def2_TZVPP_CMIRS_EC_EMC__Force = "B97-D3/def2-TZVPP/CMIRS(EC/EMC) Force"
    B97_D3_def2_TZVPP_CMIRS_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPP/CMIRS(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPP_CMIRS_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPP/CMIRS(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_EC_EMC__Unknown = "B97-D3/def2-TZVPP/CMIRS(EC/EMC) Unknown"
    B97_D3_def2_TZVPP_CMIRS_PC__Single_Point = "B97-D3/def2-TZVPP/CMIRS(PC) Single Point"
    B97_D3_def2_TZVPP_CMIRS_PC__Force = "B97-D3/def2-TZVPP/CMIRS(PC) Force"
    B97_D3_def2_TZVPP_CMIRS_PC__Geometry_Optimization = "B97-D3/def2-TZVPP/CMIRS(PC) Geometry Optimization"
    B97_D3_def2_TZVPP_CMIRS_PC__Frequency_Analysis = "B97-D3/def2-TZVPP/CMIRS(PC) Frequency Analysis"
    B97_D3_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPP_CMIRS_PC__Unknown = "B97-D3/def2-TZVPP/CMIRS(PC) Unknown"
    B97_D3_def2_TZVPPD_VACUUM_Single_Point = "B97-D3/def2-TZVPPD/VACUUM Single Point"
    B97_D3_def2_TZVPPD_VACUUM_Force = "B97-D3/def2-TZVPPD/VACUUM Force"
    B97_D3_def2_TZVPPD_VACUUM_Geometry_Optimization = "B97-D3/def2-TZVPPD/VACUUM Geometry Optimization"
    B97_D3_def2_TZVPPD_VACUUM_Frequency_Analysis = "B97-D3/def2-TZVPPD/VACUUM Frequency Analysis"
    B97_D3_def2_TZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/VACUUM Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_VACUUM_Unknown = "B97-D3/def2-TZVPPD/VACUUM Unknown"
    B97_D3_def2_TZVPPD_PCM_WATER__Single_Point = "B97-D3/def2-TZVPPD/PCM(WATER) Single Point"
    B97_D3_def2_TZVPPD_PCM_WATER__Force = "B97-D3/def2-TZVPPD/PCM(WATER) Force"
    B97_D3_def2_TZVPPD_PCM_WATER__Geometry_Optimization = "B97-D3/def2-TZVPPD/PCM(WATER) Geometry Optimization"
    B97_D3_def2_TZVPPD_PCM_WATER__Frequency_Analysis = "B97-D3/def2-TZVPPD/PCM(WATER) Frequency Analysis"
    B97_D3_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_WATER__Unknown = "B97-D3/def2-TZVPPD/PCM(WATER) Unknown"
    B97_D3_def2_TZVPPD_PCM_THF__Single_Point = "B97-D3/def2-TZVPPD/PCM(THF) Single Point"
    B97_D3_def2_TZVPPD_PCM_THF__Force = "B97-D3/def2-TZVPPD/PCM(THF) Force"
    B97_D3_def2_TZVPPD_PCM_THF__Geometry_Optimization = "B97-D3/def2-TZVPPD/PCM(THF) Geometry Optimization"
    B97_D3_def2_TZVPPD_PCM_THF__Frequency_Analysis = "B97-D3/def2-TZVPPD/PCM(THF) Frequency Analysis"
    B97_D3_def2_TZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_THF__Unknown = "B97-D3/def2-TZVPPD/PCM(THF) Unknown"
    B97_D3_def2_TZVPPD_PCM_G2__Single_Point = "B97-D3/def2-TZVPPD/PCM(G2) Single Point"
    B97_D3_def2_TZVPPD_PCM_G2__Force = "B97-D3/def2-TZVPPD/PCM(G2) Force"
    B97_D3_def2_TZVPPD_PCM_G2__Geometry_Optimization = "B97-D3/def2-TZVPPD/PCM(G2) Geometry Optimization"
    B97_D3_def2_TZVPPD_PCM_G2__Frequency_Analysis = "B97-D3/def2-TZVPPD/PCM(G2) Frequency Analysis"
    B97_D3_def2_TZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_G2__Unknown = "B97-D3/def2-TZVPPD/PCM(G2) Unknown"
    B97_D3_def2_TZVPPD_PCM_DME__Single_Point = "B97-D3/def2-TZVPPD/PCM(DME) Single Point"
    B97_D3_def2_TZVPPD_PCM_DME__Force = "B97-D3/def2-TZVPPD/PCM(DME) Force"
    B97_D3_def2_TZVPPD_PCM_DME__Geometry_Optimization = "B97-D3/def2-TZVPPD/PCM(DME) Geometry Optimization"
    B97_D3_def2_TZVPPD_PCM_DME__Frequency_Analysis = "B97-D3/def2-TZVPPD/PCM(DME) Frequency Analysis"
    B97_D3_def2_TZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_DME__Unknown = "B97-D3/def2-TZVPPD/PCM(DME) Unknown"
    B97_D3_def2_TZVPPD_PCM_EC__Single_Point = "B97-D3/def2-TZVPPD/PCM(EC) Single Point"
    B97_D3_def2_TZVPPD_PCM_EC__Force = "B97-D3/def2-TZVPPD/PCM(EC) Force"
    B97_D3_def2_TZVPPD_PCM_EC__Geometry_Optimization = "B97-D3/def2-TZVPPD/PCM(EC) Geometry Optimization"
    B97_D3_def2_TZVPPD_PCM_EC__Frequency_Analysis = "B97-D3/def2-TZVPPD/PCM(EC) Frequency Analysis"
    B97_D3_def2_TZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_EC__Unknown = "B97-D3/def2-TZVPPD/PCM(EC) Unknown"
    B97_D3_def2_TZVPPD_PCM_EC_EMC__Single_Point = "B97-D3/def2-TZVPPD/PCM(EC/EMC) Single Point"
    B97_D3_def2_TZVPPD_PCM_EC_EMC__Force = "B97-D3/def2-TZVPPD/PCM(EC/EMC) Force"
    B97_D3_def2_TZVPPD_PCM_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPPD/PCM(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPPD_PCM_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPPD/PCM(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_EC_EMC__Unknown = "B97-D3/def2-TZVPPD/PCM(EC/EMC) Unknown"
    B97_D3_def2_TZVPPD_PCM_PC__Single_Point = "B97-D3/def2-TZVPPD/PCM(PC) Single Point"
    B97_D3_def2_TZVPPD_PCM_PC__Force = "B97-D3/def2-TZVPPD/PCM(PC) Force"
    B97_D3_def2_TZVPPD_PCM_PC__Geometry_Optimization = "B97-D3/def2-TZVPPD/PCM(PC) Geometry Optimization"
    B97_D3_def2_TZVPPD_PCM_PC__Frequency_Analysis = "B97-D3/def2-TZVPPD/PCM(PC) Frequency Analysis"
    B97_D3_def2_TZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_PCM_PC__Unknown = "B97-D3/def2-TZVPPD/PCM(PC) Unknown"
    B97_D3_def2_TZVPPD_ISOSVP_WATER__Single_Point = "B97-D3/def2-TZVPPD/ISOSVP(WATER) Single Point"
    B97_D3_def2_TZVPPD_ISOSVP_WATER__Force = "B97-D3/def2-TZVPPD/ISOSVP(WATER) Force"
    B97_D3_def2_TZVPPD_ISOSVP_WATER__Geometry_Optimization = "B97-D3/def2-TZVPPD/ISOSVP(WATER) Geometry Optimization"
    B97_D3_def2_TZVPPD_ISOSVP_WATER__Frequency_Analysis = "B97-D3/def2-TZVPPD/ISOSVP(WATER) Frequency Analysis"
    B97_D3_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_WATER__Unknown = "B97-D3/def2-TZVPPD/ISOSVP(WATER) Unknown"
    B97_D3_def2_TZVPPD_ISOSVP_THF__Single_Point = "B97-D3/def2-TZVPPD/ISOSVP(THF) Single Point"
    B97_D3_def2_TZVPPD_ISOSVP_THF__Force = "B97-D3/def2-TZVPPD/ISOSVP(THF) Force"
    B97_D3_def2_TZVPPD_ISOSVP_THF__Geometry_Optimization = "B97-D3/def2-TZVPPD/ISOSVP(THF) Geometry Optimization"
    B97_D3_def2_TZVPPD_ISOSVP_THF__Frequency_Analysis = "B97-D3/def2-TZVPPD/ISOSVP(THF) Frequency Analysis"
    B97_D3_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_THF__Unknown = "B97-D3/def2-TZVPPD/ISOSVP(THF) Unknown"
    B97_D3_def2_TZVPPD_ISOSVP_G2__Single_Point = "B97-D3/def2-TZVPPD/ISOSVP(G2) Single Point"
    B97_D3_def2_TZVPPD_ISOSVP_G2__Force = "B97-D3/def2-TZVPPD/ISOSVP(G2) Force"
    B97_D3_def2_TZVPPD_ISOSVP_G2__Geometry_Optimization = "B97-D3/def2-TZVPPD/ISOSVP(G2) Geometry Optimization"
    B97_D3_def2_TZVPPD_ISOSVP_G2__Frequency_Analysis = "B97-D3/def2-TZVPPD/ISOSVP(G2) Frequency Analysis"
    B97_D3_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_G2__Unknown = "B97-D3/def2-TZVPPD/ISOSVP(G2) Unknown"
    B97_D3_def2_TZVPPD_ISOSVP_DME__Single_Point = "B97-D3/def2-TZVPPD/ISOSVP(DME) Single Point"
    B97_D3_def2_TZVPPD_ISOSVP_DME__Force = "B97-D3/def2-TZVPPD/ISOSVP(DME) Force"
    B97_D3_def2_TZVPPD_ISOSVP_DME__Geometry_Optimization = "B97-D3/def2-TZVPPD/ISOSVP(DME) Geometry Optimization"
    B97_D3_def2_TZVPPD_ISOSVP_DME__Frequency_Analysis = "B97-D3/def2-TZVPPD/ISOSVP(DME) Frequency Analysis"
    B97_D3_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_DME__Unknown = "B97-D3/def2-TZVPPD/ISOSVP(DME) Unknown"
    B97_D3_def2_TZVPPD_ISOSVP_EC__Single_Point = "B97-D3/def2-TZVPPD/ISOSVP(EC) Single Point"
    B97_D3_def2_TZVPPD_ISOSVP_EC__Force = "B97-D3/def2-TZVPPD/ISOSVP(EC) Force"
    B97_D3_def2_TZVPPD_ISOSVP_EC__Geometry_Optimization = "B97-D3/def2-TZVPPD/ISOSVP(EC) Geometry Optimization"
    B97_D3_def2_TZVPPD_ISOSVP_EC__Frequency_Analysis = "B97-D3/def2-TZVPPD/ISOSVP(EC) Frequency Analysis"
    B97_D3_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_EC__Unknown = "B97-D3/def2-TZVPPD/ISOSVP(EC) Unknown"
    B97_D3_def2_TZVPPD_ISOSVP_EC_EMC__Single_Point = "B97-D3/def2-TZVPPD/ISOSVP(EC/EMC) Single Point"
    B97_D3_def2_TZVPPD_ISOSVP_EC_EMC__Force = "B97-D3/def2-TZVPPD/ISOSVP(EC/EMC) Force"
    B97_D3_def2_TZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_EC_EMC__Unknown = "B97-D3/def2-TZVPPD/ISOSVP(EC/EMC) Unknown"
    B97_D3_def2_TZVPPD_ISOSVP_PC__Single_Point = "B97-D3/def2-TZVPPD/ISOSVP(PC) Single Point"
    B97_D3_def2_TZVPPD_ISOSVP_PC__Force = "B97-D3/def2-TZVPPD/ISOSVP(PC) Force"
    B97_D3_def2_TZVPPD_ISOSVP_PC__Geometry_Optimization = "B97-D3/def2-TZVPPD/ISOSVP(PC) Geometry Optimization"
    B97_D3_def2_TZVPPD_ISOSVP_PC__Frequency_Analysis = "B97-D3/def2-TZVPPD/ISOSVP(PC) Frequency Analysis"
    B97_D3_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_ISOSVP_PC__Unknown = "B97-D3/def2-TZVPPD/ISOSVP(PC) Unknown"
    B97_D3_def2_TZVPPD_SMD_WATER__Single_Point = "B97-D3/def2-TZVPPD/SMD(WATER) Single Point"
    B97_D3_def2_TZVPPD_SMD_WATER__Force = "B97-D3/def2-TZVPPD/SMD(WATER) Force"
    B97_D3_def2_TZVPPD_SMD_WATER__Geometry_Optimization = "B97-D3/def2-TZVPPD/SMD(WATER) Geometry Optimization"
    B97_D3_def2_TZVPPD_SMD_WATER__Frequency_Analysis = "B97-D3/def2-TZVPPD/SMD(WATER) Frequency Analysis"
    B97_D3_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_WATER__Unknown = "B97-D3/def2-TZVPPD/SMD(WATER) Unknown"
    B97_D3_def2_TZVPPD_SMD_THF__Single_Point = "B97-D3/def2-TZVPPD/SMD(THF) Single Point"
    B97_D3_def2_TZVPPD_SMD_THF__Force = "B97-D3/def2-TZVPPD/SMD(THF) Force"
    B97_D3_def2_TZVPPD_SMD_THF__Geometry_Optimization = "B97-D3/def2-TZVPPD/SMD(THF) Geometry Optimization"
    B97_D3_def2_TZVPPD_SMD_THF__Frequency_Analysis = "B97-D3/def2-TZVPPD/SMD(THF) Frequency Analysis"
    B97_D3_def2_TZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_THF__Unknown = "B97-D3/def2-TZVPPD/SMD(THF) Unknown"
    B97_D3_def2_TZVPPD_SMD_G2__Single_Point = "B97-D3/def2-TZVPPD/SMD(G2) Single Point"
    B97_D3_def2_TZVPPD_SMD_G2__Force = "B97-D3/def2-TZVPPD/SMD(G2) Force"
    B97_D3_def2_TZVPPD_SMD_G2__Geometry_Optimization = "B97-D3/def2-TZVPPD/SMD(G2) Geometry Optimization"
    B97_D3_def2_TZVPPD_SMD_G2__Frequency_Analysis = "B97-D3/def2-TZVPPD/SMD(G2) Frequency Analysis"
    B97_D3_def2_TZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_G2__Unknown = "B97-D3/def2-TZVPPD/SMD(G2) Unknown"
    B97_D3_def2_TZVPPD_SMD_DME__Single_Point = "B97-D3/def2-TZVPPD/SMD(DME) Single Point"
    B97_D3_def2_TZVPPD_SMD_DME__Force = "B97-D3/def2-TZVPPD/SMD(DME) Force"
    B97_D3_def2_TZVPPD_SMD_DME__Geometry_Optimization = "B97-D3/def2-TZVPPD/SMD(DME) Geometry Optimization"
    B97_D3_def2_TZVPPD_SMD_DME__Frequency_Analysis = "B97-D3/def2-TZVPPD/SMD(DME) Frequency Analysis"
    B97_D3_def2_TZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_DME__Unknown = "B97-D3/def2-TZVPPD/SMD(DME) Unknown"
    B97_D3_def2_TZVPPD_SMD_EC__Single_Point = "B97-D3/def2-TZVPPD/SMD(EC) Single Point"
    B97_D3_def2_TZVPPD_SMD_EC__Force = "B97-D3/def2-TZVPPD/SMD(EC) Force"
    B97_D3_def2_TZVPPD_SMD_EC__Geometry_Optimization = "B97-D3/def2-TZVPPD/SMD(EC) Geometry Optimization"
    B97_D3_def2_TZVPPD_SMD_EC__Frequency_Analysis = "B97-D3/def2-TZVPPD/SMD(EC) Frequency Analysis"
    B97_D3_def2_TZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_EC__Unknown = "B97-D3/def2-TZVPPD/SMD(EC) Unknown"
    B97_D3_def2_TZVPPD_SMD_EC_EMC__Single_Point = "B97-D3/def2-TZVPPD/SMD(EC/EMC) Single Point"
    B97_D3_def2_TZVPPD_SMD_EC_EMC__Force = "B97-D3/def2-TZVPPD/SMD(EC/EMC) Force"
    B97_D3_def2_TZVPPD_SMD_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPPD/SMD(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPPD_SMD_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPPD/SMD(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_EC_EMC__Unknown = "B97-D3/def2-TZVPPD/SMD(EC/EMC) Unknown"
    B97_D3_def2_TZVPPD_SMD_PC__Single_Point = "B97-D3/def2-TZVPPD/SMD(PC) Single Point"
    B97_D3_def2_TZVPPD_SMD_PC__Force = "B97-D3/def2-TZVPPD/SMD(PC) Force"
    B97_D3_def2_TZVPPD_SMD_PC__Geometry_Optimization = "B97-D3/def2-TZVPPD/SMD(PC) Geometry Optimization"
    B97_D3_def2_TZVPPD_SMD_PC__Frequency_Analysis = "B97-D3/def2-TZVPPD/SMD(PC) Frequency Analysis"
    B97_D3_def2_TZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_SMD_PC__Unknown = "B97-D3/def2-TZVPPD/SMD(PC) Unknown"
    B97_D3_def2_TZVPPD_CMIRS_WATER__Single_Point = "B97-D3/def2-TZVPPD/CMIRS(WATER) Single Point"
    B97_D3_def2_TZVPPD_CMIRS_WATER__Force = "B97-D3/def2-TZVPPD/CMIRS(WATER) Force"
    B97_D3_def2_TZVPPD_CMIRS_WATER__Geometry_Optimization = "B97-D3/def2-TZVPPD/CMIRS(WATER) Geometry Optimization"
    B97_D3_def2_TZVPPD_CMIRS_WATER__Frequency_Analysis = "B97-D3/def2-TZVPPD/CMIRS(WATER) Frequency Analysis"
    B97_D3_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_WATER__Unknown = "B97-D3/def2-TZVPPD/CMIRS(WATER) Unknown"
    B97_D3_def2_TZVPPD_CMIRS_THF__Single_Point = "B97-D3/def2-TZVPPD/CMIRS(THF) Single Point"
    B97_D3_def2_TZVPPD_CMIRS_THF__Force = "B97-D3/def2-TZVPPD/CMIRS(THF) Force"
    B97_D3_def2_TZVPPD_CMIRS_THF__Geometry_Optimization = "B97-D3/def2-TZVPPD/CMIRS(THF) Geometry Optimization"
    B97_D3_def2_TZVPPD_CMIRS_THF__Frequency_Analysis = "B97-D3/def2-TZVPPD/CMIRS(THF) Frequency Analysis"
    B97_D3_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_THF__Unknown = "B97-D3/def2-TZVPPD/CMIRS(THF) Unknown"
    B97_D3_def2_TZVPPD_CMIRS_G2__Single_Point = "B97-D3/def2-TZVPPD/CMIRS(G2) Single Point"
    B97_D3_def2_TZVPPD_CMIRS_G2__Force = "B97-D3/def2-TZVPPD/CMIRS(G2) Force"
    B97_D3_def2_TZVPPD_CMIRS_G2__Geometry_Optimization = "B97-D3/def2-TZVPPD/CMIRS(G2) Geometry Optimization"
    B97_D3_def2_TZVPPD_CMIRS_G2__Frequency_Analysis = "B97-D3/def2-TZVPPD/CMIRS(G2) Frequency Analysis"
    B97_D3_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_G2__Unknown = "B97-D3/def2-TZVPPD/CMIRS(G2) Unknown"
    B97_D3_def2_TZVPPD_CMIRS_DME__Single_Point = "B97-D3/def2-TZVPPD/CMIRS(DME) Single Point"
    B97_D3_def2_TZVPPD_CMIRS_DME__Force = "B97-D3/def2-TZVPPD/CMIRS(DME) Force"
    B97_D3_def2_TZVPPD_CMIRS_DME__Geometry_Optimization = "B97-D3/def2-TZVPPD/CMIRS(DME) Geometry Optimization"
    B97_D3_def2_TZVPPD_CMIRS_DME__Frequency_Analysis = "B97-D3/def2-TZVPPD/CMIRS(DME) Frequency Analysis"
    B97_D3_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_DME__Unknown = "B97-D3/def2-TZVPPD/CMIRS(DME) Unknown"
    B97_D3_def2_TZVPPD_CMIRS_EC__Single_Point = "B97-D3/def2-TZVPPD/CMIRS(EC) Single Point"
    B97_D3_def2_TZVPPD_CMIRS_EC__Force = "B97-D3/def2-TZVPPD/CMIRS(EC) Force"
    B97_D3_def2_TZVPPD_CMIRS_EC__Geometry_Optimization = "B97-D3/def2-TZVPPD/CMIRS(EC) Geometry Optimization"
    B97_D3_def2_TZVPPD_CMIRS_EC__Frequency_Analysis = "B97-D3/def2-TZVPPD/CMIRS(EC) Frequency Analysis"
    B97_D3_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_EC__Unknown = "B97-D3/def2-TZVPPD/CMIRS(EC) Unknown"
    B97_D3_def2_TZVPPD_CMIRS_EC_EMC__Single_Point = "B97-D3/def2-TZVPPD/CMIRS(EC/EMC) Single Point"
    B97_D3_def2_TZVPPD_CMIRS_EC_EMC__Force = "B97-D3/def2-TZVPPD/CMIRS(EC/EMC) Force"
    B97_D3_def2_TZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "B97-D3/def2-TZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    B97_D3_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "B97-D3/def2-TZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    B97_D3_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_EC_EMC__Unknown = "B97-D3/def2-TZVPPD/CMIRS(EC/EMC) Unknown"
    B97_D3_def2_TZVPPD_CMIRS_PC__Single_Point = "B97-D3/def2-TZVPPD/CMIRS(PC) Single Point"
    B97_D3_def2_TZVPPD_CMIRS_PC__Force = "B97-D3/def2-TZVPPD/CMIRS(PC) Force"
    B97_D3_def2_TZVPPD_CMIRS_PC__Geometry_Optimization = "B97-D3/def2-TZVPPD/CMIRS(PC) Geometry Optimization"
    B97_D3_def2_TZVPPD_CMIRS_PC__Frequency_Analysis = "B97-D3/def2-TZVPPD/CMIRS(PC) Frequency Analysis"
    B97_D3_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-TZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_TZVPPD_CMIRS_PC__Unknown = "B97-D3/def2-TZVPPD/CMIRS(PC) Unknown"
    B97_D3_def2_QZVPPD_VACUUM_Single_Point = "B97-D3/def2-QZVPPD/VACUUM Single Point"
    B97_D3_def2_QZVPPD_VACUUM_Force = "B97-D3/def2-QZVPPD/VACUUM Force"
    B97_D3_def2_QZVPPD_VACUUM_Geometry_Optimization = "B97-D3/def2-QZVPPD/VACUUM Geometry Optimization"
    B97_D3_def2_QZVPPD_VACUUM_Frequency_Analysis = "B97-D3/def2-QZVPPD/VACUUM Frequency Analysis"
    B97_D3_def2_QZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/VACUUM Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_VACUUM_Unknown = "B97-D3/def2-QZVPPD/VACUUM Unknown"
    B97_D3_def2_QZVPPD_PCM_WATER__Single_Point = "B97-D3/def2-QZVPPD/PCM(WATER) Single Point"
    B97_D3_def2_QZVPPD_PCM_WATER__Force = "B97-D3/def2-QZVPPD/PCM(WATER) Force"
    B97_D3_def2_QZVPPD_PCM_WATER__Geometry_Optimization = "B97-D3/def2-QZVPPD/PCM(WATER) Geometry Optimization"
    B97_D3_def2_QZVPPD_PCM_WATER__Frequency_Analysis = "B97-D3/def2-QZVPPD/PCM(WATER) Frequency Analysis"
    B97_D3_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_WATER__Unknown = "B97-D3/def2-QZVPPD/PCM(WATER) Unknown"
    B97_D3_def2_QZVPPD_PCM_THF__Single_Point = "B97-D3/def2-QZVPPD/PCM(THF) Single Point"
    B97_D3_def2_QZVPPD_PCM_THF__Force = "B97-D3/def2-QZVPPD/PCM(THF) Force"
    B97_D3_def2_QZVPPD_PCM_THF__Geometry_Optimization = "B97-D3/def2-QZVPPD/PCM(THF) Geometry Optimization"
    B97_D3_def2_QZVPPD_PCM_THF__Frequency_Analysis = "B97-D3/def2-QZVPPD/PCM(THF) Frequency Analysis"
    B97_D3_def2_QZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_THF__Unknown = "B97-D3/def2-QZVPPD/PCM(THF) Unknown"
    B97_D3_def2_QZVPPD_PCM_G2__Single_Point = "B97-D3/def2-QZVPPD/PCM(G2) Single Point"
    B97_D3_def2_QZVPPD_PCM_G2__Force = "B97-D3/def2-QZVPPD/PCM(G2) Force"
    B97_D3_def2_QZVPPD_PCM_G2__Geometry_Optimization = "B97-D3/def2-QZVPPD/PCM(G2) Geometry Optimization"
    B97_D3_def2_QZVPPD_PCM_G2__Frequency_Analysis = "B97-D3/def2-QZVPPD/PCM(G2) Frequency Analysis"
    B97_D3_def2_QZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_G2__Unknown = "B97-D3/def2-QZVPPD/PCM(G2) Unknown"
    B97_D3_def2_QZVPPD_PCM_DME__Single_Point = "B97-D3/def2-QZVPPD/PCM(DME) Single Point"
    B97_D3_def2_QZVPPD_PCM_DME__Force = "B97-D3/def2-QZVPPD/PCM(DME) Force"
    B97_D3_def2_QZVPPD_PCM_DME__Geometry_Optimization = "B97-D3/def2-QZVPPD/PCM(DME) Geometry Optimization"
    B97_D3_def2_QZVPPD_PCM_DME__Frequency_Analysis = "B97-D3/def2-QZVPPD/PCM(DME) Frequency Analysis"
    B97_D3_def2_QZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_DME__Unknown = "B97-D3/def2-QZVPPD/PCM(DME) Unknown"
    B97_D3_def2_QZVPPD_PCM_EC__Single_Point = "B97-D3/def2-QZVPPD/PCM(EC) Single Point"
    B97_D3_def2_QZVPPD_PCM_EC__Force = "B97-D3/def2-QZVPPD/PCM(EC) Force"
    B97_D3_def2_QZVPPD_PCM_EC__Geometry_Optimization = "B97-D3/def2-QZVPPD/PCM(EC) Geometry Optimization"
    B97_D3_def2_QZVPPD_PCM_EC__Frequency_Analysis = "B97-D3/def2-QZVPPD/PCM(EC) Frequency Analysis"
    B97_D3_def2_QZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_EC__Unknown = "B97-D3/def2-QZVPPD/PCM(EC) Unknown"
    B97_D3_def2_QZVPPD_PCM_EC_EMC__Single_Point = "B97-D3/def2-QZVPPD/PCM(EC/EMC) Single Point"
    B97_D3_def2_QZVPPD_PCM_EC_EMC__Force = "B97-D3/def2-QZVPPD/PCM(EC/EMC) Force"
    B97_D3_def2_QZVPPD_PCM_EC_EMC__Geometry_Optimization = "B97-D3/def2-QZVPPD/PCM(EC/EMC) Geometry Optimization"
    B97_D3_def2_QZVPPD_PCM_EC_EMC__Frequency_Analysis = "B97-D3/def2-QZVPPD/PCM(EC/EMC) Frequency Analysis"
    B97_D3_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_EC_EMC__Unknown = "B97-D3/def2-QZVPPD/PCM(EC/EMC) Unknown"
    B97_D3_def2_QZVPPD_PCM_PC__Single_Point = "B97-D3/def2-QZVPPD/PCM(PC) Single Point"
    B97_D3_def2_QZVPPD_PCM_PC__Force = "B97-D3/def2-QZVPPD/PCM(PC) Force"
    B97_D3_def2_QZVPPD_PCM_PC__Geometry_Optimization = "B97-D3/def2-QZVPPD/PCM(PC) Geometry Optimization"
    B97_D3_def2_QZVPPD_PCM_PC__Frequency_Analysis = "B97-D3/def2-QZVPPD/PCM(PC) Frequency Analysis"
    B97_D3_def2_QZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_PCM_PC__Unknown = "B97-D3/def2-QZVPPD/PCM(PC) Unknown"
    B97_D3_def2_QZVPPD_ISOSVP_WATER__Single_Point = "B97-D3/def2-QZVPPD/ISOSVP(WATER) Single Point"
    B97_D3_def2_QZVPPD_ISOSVP_WATER__Force = "B97-D3/def2-QZVPPD/ISOSVP(WATER) Force"
    B97_D3_def2_QZVPPD_ISOSVP_WATER__Geometry_Optimization = "B97-D3/def2-QZVPPD/ISOSVP(WATER) Geometry Optimization"
    B97_D3_def2_QZVPPD_ISOSVP_WATER__Frequency_Analysis = "B97-D3/def2-QZVPPD/ISOSVP(WATER) Frequency Analysis"
    B97_D3_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_WATER__Unknown = "B97-D3/def2-QZVPPD/ISOSVP(WATER) Unknown"
    B97_D3_def2_QZVPPD_ISOSVP_THF__Single_Point = "B97-D3/def2-QZVPPD/ISOSVP(THF) Single Point"
    B97_D3_def2_QZVPPD_ISOSVP_THF__Force = "B97-D3/def2-QZVPPD/ISOSVP(THF) Force"
    B97_D3_def2_QZVPPD_ISOSVP_THF__Geometry_Optimization = "B97-D3/def2-QZVPPD/ISOSVP(THF) Geometry Optimization"
    B97_D3_def2_QZVPPD_ISOSVP_THF__Frequency_Analysis = "B97-D3/def2-QZVPPD/ISOSVP(THF) Frequency Analysis"
    B97_D3_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_THF__Unknown = "B97-D3/def2-QZVPPD/ISOSVP(THF) Unknown"
    B97_D3_def2_QZVPPD_ISOSVP_G2__Single_Point = "B97-D3/def2-QZVPPD/ISOSVP(G2) Single Point"
    B97_D3_def2_QZVPPD_ISOSVP_G2__Force = "B97-D3/def2-QZVPPD/ISOSVP(G2) Force"
    B97_D3_def2_QZVPPD_ISOSVP_G2__Geometry_Optimization = "B97-D3/def2-QZVPPD/ISOSVP(G2) Geometry Optimization"
    B97_D3_def2_QZVPPD_ISOSVP_G2__Frequency_Analysis = "B97-D3/def2-QZVPPD/ISOSVP(G2) Frequency Analysis"
    B97_D3_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_G2__Unknown = "B97-D3/def2-QZVPPD/ISOSVP(G2) Unknown"
    B97_D3_def2_QZVPPD_ISOSVP_DME__Single_Point = "B97-D3/def2-QZVPPD/ISOSVP(DME) Single Point"
    B97_D3_def2_QZVPPD_ISOSVP_DME__Force = "B97-D3/def2-QZVPPD/ISOSVP(DME) Force"
    B97_D3_def2_QZVPPD_ISOSVP_DME__Geometry_Optimization = "B97-D3/def2-QZVPPD/ISOSVP(DME) Geometry Optimization"
    B97_D3_def2_QZVPPD_ISOSVP_DME__Frequency_Analysis = "B97-D3/def2-QZVPPD/ISOSVP(DME) Frequency Analysis"
    B97_D3_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_DME__Unknown = "B97-D3/def2-QZVPPD/ISOSVP(DME) Unknown"
    B97_D3_def2_QZVPPD_ISOSVP_EC__Single_Point = "B97-D3/def2-QZVPPD/ISOSVP(EC) Single Point"
    B97_D3_def2_QZVPPD_ISOSVP_EC__Force = "B97-D3/def2-QZVPPD/ISOSVP(EC) Force"
    B97_D3_def2_QZVPPD_ISOSVP_EC__Geometry_Optimization = "B97-D3/def2-QZVPPD/ISOSVP(EC) Geometry Optimization"
    B97_D3_def2_QZVPPD_ISOSVP_EC__Frequency_Analysis = "B97-D3/def2-QZVPPD/ISOSVP(EC) Frequency Analysis"
    B97_D3_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_EC__Unknown = "B97-D3/def2-QZVPPD/ISOSVP(EC) Unknown"
    B97_D3_def2_QZVPPD_ISOSVP_EC_EMC__Single_Point = "B97-D3/def2-QZVPPD/ISOSVP(EC/EMC) Single Point"
    B97_D3_def2_QZVPPD_ISOSVP_EC_EMC__Force = "B97-D3/def2-QZVPPD/ISOSVP(EC/EMC) Force"
    B97_D3_def2_QZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = "B97-D3/def2-QZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    B97_D3_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97-D3/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97_D3_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_EC_EMC__Unknown = "B97-D3/def2-QZVPPD/ISOSVP(EC/EMC) Unknown"
    B97_D3_def2_QZVPPD_ISOSVP_PC__Single_Point = "B97-D3/def2-QZVPPD/ISOSVP(PC) Single Point"
    B97_D3_def2_QZVPPD_ISOSVP_PC__Force = "B97-D3/def2-QZVPPD/ISOSVP(PC) Force"
    B97_D3_def2_QZVPPD_ISOSVP_PC__Geometry_Optimization = "B97-D3/def2-QZVPPD/ISOSVP(PC) Geometry Optimization"
    B97_D3_def2_QZVPPD_ISOSVP_PC__Frequency_Analysis = "B97-D3/def2-QZVPPD/ISOSVP(PC) Frequency Analysis"
    B97_D3_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_ISOSVP_PC__Unknown = "B97-D3/def2-QZVPPD/ISOSVP(PC) Unknown"
    B97_D3_def2_QZVPPD_SMD_WATER__Single_Point = "B97-D3/def2-QZVPPD/SMD(WATER) Single Point"
    B97_D3_def2_QZVPPD_SMD_WATER__Force = "B97-D3/def2-QZVPPD/SMD(WATER) Force"
    B97_D3_def2_QZVPPD_SMD_WATER__Geometry_Optimization = "B97-D3/def2-QZVPPD/SMD(WATER) Geometry Optimization"
    B97_D3_def2_QZVPPD_SMD_WATER__Frequency_Analysis = "B97-D3/def2-QZVPPD/SMD(WATER) Frequency Analysis"
    B97_D3_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_WATER__Unknown = "B97-D3/def2-QZVPPD/SMD(WATER) Unknown"
    B97_D3_def2_QZVPPD_SMD_THF__Single_Point = "B97-D3/def2-QZVPPD/SMD(THF) Single Point"
    B97_D3_def2_QZVPPD_SMD_THF__Force = "B97-D3/def2-QZVPPD/SMD(THF) Force"
    B97_D3_def2_QZVPPD_SMD_THF__Geometry_Optimization = "B97-D3/def2-QZVPPD/SMD(THF) Geometry Optimization"
    B97_D3_def2_QZVPPD_SMD_THF__Frequency_Analysis = "B97-D3/def2-QZVPPD/SMD(THF) Frequency Analysis"
    B97_D3_def2_QZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_THF__Unknown = "B97-D3/def2-QZVPPD/SMD(THF) Unknown"
    B97_D3_def2_QZVPPD_SMD_G2__Single_Point = "B97-D3/def2-QZVPPD/SMD(G2) Single Point"
    B97_D3_def2_QZVPPD_SMD_G2__Force = "B97-D3/def2-QZVPPD/SMD(G2) Force"
    B97_D3_def2_QZVPPD_SMD_G2__Geometry_Optimization = "B97-D3/def2-QZVPPD/SMD(G2) Geometry Optimization"
    B97_D3_def2_QZVPPD_SMD_G2__Frequency_Analysis = "B97-D3/def2-QZVPPD/SMD(G2) Frequency Analysis"
    B97_D3_def2_QZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_G2__Unknown = "B97-D3/def2-QZVPPD/SMD(G2) Unknown"
    B97_D3_def2_QZVPPD_SMD_DME__Single_Point = "B97-D3/def2-QZVPPD/SMD(DME) Single Point"
    B97_D3_def2_QZVPPD_SMD_DME__Force = "B97-D3/def2-QZVPPD/SMD(DME) Force"
    B97_D3_def2_QZVPPD_SMD_DME__Geometry_Optimization = "B97-D3/def2-QZVPPD/SMD(DME) Geometry Optimization"
    B97_D3_def2_QZVPPD_SMD_DME__Frequency_Analysis = "B97-D3/def2-QZVPPD/SMD(DME) Frequency Analysis"
    B97_D3_def2_QZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_DME__Unknown = "B97-D3/def2-QZVPPD/SMD(DME) Unknown"
    B97_D3_def2_QZVPPD_SMD_EC__Single_Point = "B97-D3/def2-QZVPPD/SMD(EC) Single Point"
    B97_D3_def2_QZVPPD_SMD_EC__Force = "B97-D3/def2-QZVPPD/SMD(EC) Force"
    B97_D3_def2_QZVPPD_SMD_EC__Geometry_Optimization = "B97-D3/def2-QZVPPD/SMD(EC) Geometry Optimization"
    B97_D3_def2_QZVPPD_SMD_EC__Frequency_Analysis = "B97-D3/def2-QZVPPD/SMD(EC) Frequency Analysis"
    B97_D3_def2_QZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_EC__Unknown = "B97-D3/def2-QZVPPD/SMD(EC) Unknown"
    B97_D3_def2_QZVPPD_SMD_EC_EMC__Single_Point = "B97-D3/def2-QZVPPD/SMD(EC/EMC) Single Point"
    B97_D3_def2_QZVPPD_SMD_EC_EMC__Force = "B97-D3/def2-QZVPPD/SMD(EC/EMC) Force"
    B97_D3_def2_QZVPPD_SMD_EC_EMC__Geometry_Optimization = "B97-D3/def2-QZVPPD/SMD(EC/EMC) Geometry Optimization"
    B97_D3_def2_QZVPPD_SMD_EC_EMC__Frequency_Analysis = "B97-D3/def2-QZVPPD/SMD(EC/EMC) Frequency Analysis"
    B97_D3_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_EC_EMC__Unknown = "B97-D3/def2-QZVPPD/SMD(EC/EMC) Unknown"
    B97_D3_def2_QZVPPD_SMD_PC__Single_Point = "B97-D3/def2-QZVPPD/SMD(PC) Single Point"
    B97_D3_def2_QZVPPD_SMD_PC__Force = "B97-D3/def2-QZVPPD/SMD(PC) Force"
    B97_D3_def2_QZVPPD_SMD_PC__Geometry_Optimization = "B97-D3/def2-QZVPPD/SMD(PC) Geometry Optimization"
    B97_D3_def2_QZVPPD_SMD_PC__Frequency_Analysis = "B97-D3/def2-QZVPPD/SMD(PC) Frequency Analysis"
    B97_D3_def2_QZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_SMD_PC__Unknown = "B97-D3/def2-QZVPPD/SMD(PC) Unknown"
    B97_D3_def2_QZVPPD_CMIRS_WATER__Single_Point = "B97-D3/def2-QZVPPD/CMIRS(WATER) Single Point"
    B97_D3_def2_QZVPPD_CMIRS_WATER__Force = "B97-D3/def2-QZVPPD/CMIRS(WATER) Force"
    B97_D3_def2_QZVPPD_CMIRS_WATER__Geometry_Optimization = "B97-D3/def2-QZVPPD/CMIRS(WATER) Geometry Optimization"
    B97_D3_def2_QZVPPD_CMIRS_WATER__Frequency_Analysis = "B97-D3/def2-QZVPPD/CMIRS(WATER) Frequency Analysis"
    B97_D3_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_WATER__Unknown = "B97-D3/def2-QZVPPD/CMIRS(WATER) Unknown"
    B97_D3_def2_QZVPPD_CMIRS_THF__Single_Point = "B97-D3/def2-QZVPPD/CMIRS(THF) Single Point"
    B97_D3_def2_QZVPPD_CMIRS_THF__Force = "B97-D3/def2-QZVPPD/CMIRS(THF) Force"
    B97_D3_def2_QZVPPD_CMIRS_THF__Geometry_Optimization = "B97-D3/def2-QZVPPD/CMIRS(THF) Geometry Optimization"
    B97_D3_def2_QZVPPD_CMIRS_THF__Frequency_Analysis = "B97-D3/def2-QZVPPD/CMIRS(THF) Frequency Analysis"
    B97_D3_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_THF__Unknown = "B97-D3/def2-QZVPPD/CMIRS(THF) Unknown"
    B97_D3_def2_QZVPPD_CMIRS_G2__Single_Point = "B97-D3/def2-QZVPPD/CMIRS(G2) Single Point"
    B97_D3_def2_QZVPPD_CMIRS_G2__Force = "B97-D3/def2-QZVPPD/CMIRS(G2) Force"
    B97_D3_def2_QZVPPD_CMIRS_G2__Geometry_Optimization = "B97-D3/def2-QZVPPD/CMIRS(G2) Geometry Optimization"
    B97_D3_def2_QZVPPD_CMIRS_G2__Frequency_Analysis = "B97-D3/def2-QZVPPD/CMIRS(G2) Frequency Analysis"
    B97_D3_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_G2__Unknown = "B97-D3/def2-QZVPPD/CMIRS(G2) Unknown"
    B97_D3_def2_QZVPPD_CMIRS_DME__Single_Point = "B97-D3/def2-QZVPPD/CMIRS(DME) Single Point"
    B97_D3_def2_QZVPPD_CMIRS_DME__Force = "B97-D3/def2-QZVPPD/CMIRS(DME) Force"
    B97_D3_def2_QZVPPD_CMIRS_DME__Geometry_Optimization = "B97-D3/def2-QZVPPD/CMIRS(DME) Geometry Optimization"
    B97_D3_def2_QZVPPD_CMIRS_DME__Frequency_Analysis = "B97-D3/def2-QZVPPD/CMIRS(DME) Frequency Analysis"
    B97_D3_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_DME__Unknown = "B97-D3/def2-QZVPPD/CMIRS(DME) Unknown"
    B97_D3_def2_QZVPPD_CMIRS_EC__Single_Point = "B97-D3/def2-QZVPPD/CMIRS(EC) Single Point"
    B97_D3_def2_QZVPPD_CMIRS_EC__Force = "B97-D3/def2-QZVPPD/CMIRS(EC) Force"
    B97_D3_def2_QZVPPD_CMIRS_EC__Geometry_Optimization = "B97-D3/def2-QZVPPD/CMIRS(EC) Geometry Optimization"
    B97_D3_def2_QZVPPD_CMIRS_EC__Frequency_Analysis = "B97-D3/def2-QZVPPD/CMIRS(EC) Frequency Analysis"
    B97_D3_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_EC__Unknown = "B97-D3/def2-QZVPPD/CMIRS(EC) Unknown"
    B97_D3_def2_QZVPPD_CMIRS_EC_EMC__Single_Point = "B97-D3/def2-QZVPPD/CMIRS(EC/EMC) Single Point"
    B97_D3_def2_QZVPPD_CMIRS_EC_EMC__Force = "B97-D3/def2-QZVPPD/CMIRS(EC/EMC) Force"
    B97_D3_def2_QZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "B97-D3/def2-QZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    B97_D3_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "B97-D3/def2-QZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    B97_D3_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_EC_EMC__Unknown = "B97-D3/def2-QZVPPD/CMIRS(EC/EMC) Unknown"
    B97_D3_def2_QZVPPD_CMIRS_PC__Single_Point = "B97-D3/def2-QZVPPD/CMIRS(PC) Single Point"
    B97_D3_def2_QZVPPD_CMIRS_PC__Force = "B97-D3/def2-QZVPPD/CMIRS(PC) Force"
    B97_D3_def2_QZVPPD_CMIRS_PC__Geometry_Optimization = "B97-D3/def2-QZVPPD/CMIRS(PC) Geometry Optimization"
    B97_D3_def2_QZVPPD_CMIRS_PC__Frequency_Analysis = "B97-D3/def2-QZVPPD/CMIRS(PC) Frequency Analysis"
    B97_D3_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97-D3/def2-QZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97_D3_def2_QZVPPD_CMIRS_PC__Unknown = "B97-D3/def2-QZVPPD/CMIRS(PC) Unknown"
    B97M_rV_6_31g_d_VACUUM_Single_Point = "B97M-rV/6-31g*/VACUUM Single Point"
    B97M_rV_6_31g_d_VACUUM_Force = "B97M-rV/6-31g*/VACUUM Force"
    B97M_rV_6_31g_d_VACUUM_Geometry_Optimization = "B97M-rV/6-31g*/VACUUM Geometry Optimization"
    B97M_rV_6_31g_d_VACUUM_Frequency_Analysis = "B97M-rV/6-31g*/VACUUM Frequency Analysis"
    B97M_rV_6_31g_d_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_VACUUM_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/VACUUM Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_VACUUM_Unknown = "B97M-rV/6-31g*/VACUUM Unknown"
    B97M_rV_6_31g_d_PCM_WATER__Single_Point = "B97M-rV/6-31g*/PCM(WATER) Single Point"
    B97M_rV_6_31g_d_PCM_WATER__Force = "B97M-rV/6-31g*/PCM(WATER) Force"
    B97M_rV_6_31g_d_PCM_WATER__Geometry_Optimization = "B97M-rV/6-31g*/PCM(WATER) Geometry Optimization"
    B97M_rV_6_31g_d_PCM_WATER__Frequency_Analysis = "B97M-rV/6-31g*/PCM(WATER) Frequency Analysis"
    B97M_rV_6_31g_d_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_WATER__Unknown = "B97M-rV/6-31g*/PCM(WATER) Unknown"
    B97M_rV_6_31g_d_PCM_THF__Single_Point = "B97M-rV/6-31g*/PCM(THF) Single Point"
    B97M_rV_6_31g_d_PCM_THF__Force = "B97M-rV/6-31g*/PCM(THF) Force"
    B97M_rV_6_31g_d_PCM_THF__Geometry_Optimization = "B97M-rV/6-31g*/PCM(THF) Geometry Optimization"
    B97M_rV_6_31g_d_PCM_THF__Frequency_Analysis = "B97M-rV/6-31g*/PCM(THF) Frequency Analysis"
    B97M_rV_6_31g_d_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(THF) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_THF__Unknown = "B97M-rV/6-31g*/PCM(THF) Unknown"
    B97M_rV_6_31g_d_PCM_G2__Single_Point = "B97M-rV/6-31g*/PCM(G2) Single Point"
    B97M_rV_6_31g_d_PCM_G2__Force = "B97M-rV/6-31g*/PCM(G2) Force"
    B97M_rV_6_31g_d_PCM_G2__Geometry_Optimization = "B97M-rV/6-31g*/PCM(G2) Geometry Optimization"
    B97M_rV_6_31g_d_PCM_G2__Frequency_Analysis = "B97M-rV/6-31g*/PCM(G2) Frequency Analysis"
    B97M_rV_6_31g_d_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(G2) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_G2__Unknown = "B97M-rV/6-31g*/PCM(G2) Unknown"
    B97M_rV_6_31g_d_PCM_DME__Single_Point = "B97M-rV/6-31g*/PCM(DME) Single Point"
    B97M_rV_6_31g_d_PCM_DME__Force = "B97M-rV/6-31g*/PCM(DME) Force"
    B97M_rV_6_31g_d_PCM_DME__Geometry_Optimization = "B97M-rV/6-31g*/PCM(DME) Geometry Optimization"
    B97M_rV_6_31g_d_PCM_DME__Frequency_Analysis = "B97M-rV/6-31g*/PCM(DME) Frequency Analysis"
    B97M_rV_6_31g_d_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(DME) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_DME__Unknown = "B97M-rV/6-31g*/PCM(DME) Unknown"
    B97M_rV_6_31g_d_PCM_EC__Single_Point = "B97M-rV/6-31g*/PCM(EC) Single Point"
    B97M_rV_6_31g_d_PCM_EC__Force = "B97M-rV/6-31g*/PCM(EC) Force"
    B97M_rV_6_31g_d_PCM_EC__Geometry_Optimization = "B97M-rV/6-31g*/PCM(EC) Geometry Optimization"
    B97M_rV_6_31g_d_PCM_EC__Frequency_Analysis = "B97M-rV/6-31g*/PCM(EC) Frequency Analysis"
    B97M_rV_6_31g_d_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(EC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_EC__Unknown = "B97M-rV/6-31g*/PCM(EC) Unknown"
    B97M_rV_6_31g_d_PCM_EC_EMC__Single_Point = "B97M-rV/6-31g*/PCM(EC/EMC) Single Point"
    B97M_rV_6_31g_d_PCM_EC_EMC__Force = "B97M-rV/6-31g*/PCM(EC/EMC) Force"
    B97M_rV_6_31g_d_PCM_EC_EMC__Geometry_Optimization = "B97M-rV/6-31g*/PCM(EC/EMC) Geometry Optimization"
    B97M_rV_6_31g_d_PCM_EC_EMC__Frequency_Analysis = "B97M-rV/6-31g*/PCM(EC/EMC) Frequency Analysis"
    B97M_rV_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_EC_EMC__Unknown = "B97M-rV/6-31g*/PCM(EC/EMC) Unknown"
    B97M_rV_6_31g_d_PCM_PC__Single_Point = "B97M-rV/6-31g*/PCM(PC) Single Point"
    B97M_rV_6_31g_d_PCM_PC__Force = "B97M-rV/6-31g*/PCM(PC) Force"
    B97M_rV_6_31g_d_PCM_PC__Geometry_Optimization = "B97M-rV/6-31g*/PCM(PC) Geometry Optimization"
    B97M_rV_6_31g_d_PCM_PC__Frequency_Analysis = "B97M-rV/6-31g*/PCM(PC) Frequency Analysis"
    B97M_rV_6_31g_d_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(PC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_PCM_PC__Unknown = "B97M-rV/6-31g*/PCM(PC) Unknown"
    B97M_rV_6_31g_d_ISOSVP_WATER__Single_Point = "B97M-rV/6-31g*/ISOSVP(WATER) Single Point"
    B97M_rV_6_31g_d_ISOSVP_WATER__Force = "B97M-rV/6-31g*/ISOSVP(WATER) Force"
    B97M_rV_6_31g_d_ISOSVP_WATER__Geometry_Optimization = "B97M-rV/6-31g*/ISOSVP(WATER) Geometry Optimization"
    B97M_rV_6_31g_d_ISOSVP_WATER__Frequency_Analysis = "B97M-rV/6-31g*/ISOSVP(WATER) Frequency Analysis"
    B97M_rV_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_WATER__Unknown = "B97M-rV/6-31g*/ISOSVP(WATER) Unknown"
    B97M_rV_6_31g_d_ISOSVP_THF__Single_Point = "B97M-rV/6-31g*/ISOSVP(THF) Single Point"
    B97M_rV_6_31g_d_ISOSVP_THF__Force = "B97M-rV/6-31g*/ISOSVP(THF) Force"
    B97M_rV_6_31g_d_ISOSVP_THF__Geometry_Optimization = "B97M-rV/6-31g*/ISOSVP(THF) Geometry Optimization"
    B97M_rV_6_31g_d_ISOSVP_THF__Frequency_Analysis = "B97M-rV/6-31g*/ISOSVP(THF) Frequency Analysis"
    B97M_rV_6_31g_d_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_THF__Unknown = "B97M-rV/6-31g*/ISOSVP(THF) Unknown"
    B97M_rV_6_31g_d_ISOSVP_G2__Single_Point = "B97M-rV/6-31g*/ISOSVP(G2) Single Point"
    B97M_rV_6_31g_d_ISOSVP_G2__Force = "B97M-rV/6-31g*/ISOSVP(G2) Force"
    B97M_rV_6_31g_d_ISOSVP_G2__Geometry_Optimization = "B97M-rV/6-31g*/ISOSVP(G2) Geometry Optimization"
    B97M_rV_6_31g_d_ISOSVP_G2__Frequency_Analysis = "B97M-rV/6-31g*/ISOSVP(G2) Frequency Analysis"
    B97M_rV_6_31g_d_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_G2__Unknown = "B97M-rV/6-31g*/ISOSVP(G2) Unknown"
    B97M_rV_6_31g_d_ISOSVP_DME__Single_Point = "B97M-rV/6-31g*/ISOSVP(DME) Single Point"
    B97M_rV_6_31g_d_ISOSVP_DME__Force = "B97M-rV/6-31g*/ISOSVP(DME) Force"
    B97M_rV_6_31g_d_ISOSVP_DME__Geometry_Optimization = "B97M-rV/6-31g*/ISOSVP(DME) Geometry Optimization"
    B97M_rV_6_31g_d_ISOSVP_DME__Frequency_Analysis = "B97M-rV/6-31g*/ISOSVP(DME) Frequency Analysis"
    B97M_rV_6_31g_d_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_DME__Unknown = "B97M-rV/6-31g*/ISOSVP(DME) Unknown"
    B97M_rV_6_31g_d_ISOSVP_EC__Single_Point = "B97M-rV/6-31g*/ISOSVP(EC) Single Point"
    B97M_rV_6_31g_d_ISOSVP_EC__Force = "B97M-rV/6-31g*/ISOSVP(EC) Force"
    B97M_rV_6_31g_d_ISOSVP_EC__Geometry_Optimization = "B97M-rV/6-31g*/ISOSVP(EC) Geometry Optimization"
    B97M_rV_6_31g_d_ISOSVP_EC__Frequency_Analysis = "B97M-rV/6-31g*/ISOSVP(EC) Frequency Analysis"
    B97M_rV_6_31g_d_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_EC__Unknown = "B97M-rV/6-31g*/ISOSVP(EC) Unknown"
    B97M_rV_6_31g_d_ISOSVP_EC_EMC__Single_Point = "B97M-rV/6-31g*/ISOSVP(EC/EMC) Single Point"
    B97M_rV_6_31g_d_ISOSVP_EC_EMC__Force = "B97M-rV/6-31g*/ISOSVP(EC/EMC) Force"
    B97M_rV_6_31g_d_ISOSVP_EC_EMC__Geometry_Optimization = "B97M-rV/6-31g*/ISOSVP(EC/EMC) Geometry Optimization"
    B97M_rV_6_31g_d_ISOSVP_EC_EMC__Frequency_Analysis = "B97M-rV/6-31g*/ISOSVP(EC/EMC) Frequency Analysis"
    B97M_rV_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_EC_EMC__Unknown = "B97M-rV/6-31g*/ISOSVP(EC/EMC) Unknown"
    B97M_rV_6_31g_d_ISOSVP_PC__Single_Point = "B97M-rV/6-31g*/ISOSVP(PC) Single Point"
    B97M_rV_6_31g_d_ISOSVP_PC__Force = "B97M-rV/6-31g*/ISOSVP(PC) Force"
    B97M_rV_6_31g_d_ISOSVP_PC__Geometry_Optimization = "B97M-rV/6-31g*/ISOSVP(PC) Geometry Optimization"
    B97M_rV_6_31g_d_ISOSVP_PC__Frequency_Analysis = "B97M-rV/6-31g*/ISOSVP(PC) Frequency Analysis"
    B97M_rV_6_31g_d_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_ISOSVP_PC__Unknown = "B97M-rV/6-31g*/ISOSVP(PC) Unknown"
    B97M_rV_6_31g_d_SMD_WATER__Single_Point = "B97M-rV/6-31g*/SMD(WATER) Single Point"
    B97M_rV_6_31g_d_SMD_WATER__Force = "B97M-rV/6-31g*/SMD(WATER) Force"
    B97M_rV_6_31g_d_SMD_WATER__Geometry_Optimization = "B97M-rV/6-31g*/SMD(WATER) Geometry Optimization"
    B97M_rV_6_31g_d_SMD_WATER__Frequency_Analysis = "B97M-rV/6-31g*/SMD(WATER) Frequency Analysis"
    B97M_rV_6_31g_d_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_WATER__Unknown = "B97M-rV/6-31g*/SMD(WATER) Unknown"
    B97M_rV_6_31g_d_SMD_THF__Single_Point = "B97M-rV/6-31g*/SMD(THF) Single Point"
    B97M_rV_6_31g_d_SMD_THF__Force = "B97M-rV/6-31g*/SMD(THF) Force"
    B97M_rV_6_31g_d_SMD_THF__Geometry_Optimization = "B97M-rV/6-31g*/SMD(THF) Geometry Optimization"
    B97M_rV_6_31g_d_SMD_THF__Frequency_Analysis = "B97M-rV/6-31g*/SMD(THF) Frequency Analysis"
    B97M_rV_6_31g_d_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(THF) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_THF__Unknown = "B97M-rV/6-31g*/SMD(THF) Unknown"
    B97M_rV_6_31g_d_SMD_G2__Single_Point = "B97M-rV/6-31g*/SMD(G2) Single Point"
    B97M_rV_6_31g_d_SMD_G2__Force = "B97M-rV/6-31g*/SMD(G2) Force"
    B97M_rV_6_31g_d_SMD_G2__Geometry_Optimization = "B97M-rV/6-31g*/SMD(G2) Geometry Optimization"
    B97M_rV_6_31g_d_SMD_G2__Frequency_Analysis = "B97M-rV/6-31g*/SMD(G2) Frequency Analysis"
    B97M_rV_6_31g_d_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(G2) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_G2__Unknown = "B97M-rV/6-31g*/SMD(G2) Unknown"
    B97M_rV_6_31g_d_SMD_DME__Single_Point = "B97M-rV/6-31g*/SMD(DME) Single Point"
    B97M_rV_6_31g_d_SMD_DME__Force = "B97M-rV/6-31g*/SMD(DME) Force"
    B97M_rV_6_31g_d_SMD_DME__Geometry_Optimization = "B97M-rV/6-31g*/SMD(DME) Geometry Optimization"
    B97M_rV_6_31g_d_SMD_DME__Frequency_Analysis = "B97M-rV/6-31g*/SMD(DME) Frequency Analysis"
    B97M_rV_6_31g_d_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(DME) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_DME__Unknown = "B97M-rV/6-31g*/SMD(DME) Unknown"
    B97M_rV_6_31g_d_SMD_EC__Single_Point = "B97M-rV/6-31g*/SMD(EC) Single Point"
    B97M_rV_6_31g_d_SMD_EC__Force = "B97M-rV/6-31g*/SMD(EC) Force"
    B97M_rV_6_31g_d_SMD_EC__Geometry_Optimization = "B97M-rV/6-31g*/SMD(EC) Geometry Optimization"
    B97M_rV_6_31g_d_SMD_EC__Frequency_Analysis = "B97M-rV/6-31g*/SMD(EC) Frequency Analysis"
    B97M_rV_6_31g_d_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(EC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_EC__Unknown = "B97M-rV/6-31g*/SMD(EC) Unknown"
    B97M_rV_6_31g_d_SMD_EC_EMC__Single_Point = "B97M-rV/6-31g*/SMD(EC/EMC) Single Point"
    B97M_rV_6_31g_d_SMD_EC_EMC__Force = "B97M-rV/6-31g*/SMD(EC/EMC) Force"
    B97M_rV_6_31g_d_SMD_EC_EMC__Geometry_Optimization = "B97M-rV/6-31g*/SMD(EC/EMC) Geometry Optimization"
    B97M_rV_6_31g_d_SMD_EC_EMC__Frequency_Analysis = "B97M-rV/6-31g*/SMD(EC/EMC) Frequency Analysis"
    B97M_rV_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_EC_EMC__Unknown = "B97M-rV/6-31g*/SMD(EC/EMC) Unknown"
    B97M_rV_6_31g_d_SMD_PC__Single_Point = "B97M-rV/6-31g*/SMD(PC) Single Point"
    B97M_rV_6_31g_d_SMD_PC__Force = "B97M-rV/6-31g*/SMD(PC) Force"
    B97M_rV_6_31g_d_SMD_PC__Geometry_Optimization = "B97M-rV/6-31g*/SMD(PC) Geometry Optimization"
    B97M_rV_6_31g_d_SMD_PC__Frequency_Analysis = "B97M-rV/6-31g*/SMD(PC) Frequency Analysis"
    B97M_rV_6_31g_d_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(PC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_SMD_PC__Unknown = "B97M-rV/6-31g*/SMD(PC) Unknown"
    B97M_rV_6_31g_d_CMIRS_WATER__Single_Point = "B97M-rV/6-31g*/CMIRS(WATER) Single Point"
    B97M_rV_6_31g_d_CMIRS_WATER__Force = "B97M-rV/6-31g*/CMIRS(WATER) Force"
    B97M_rV_6_31g_d_CMIRS_WATER__Geometry_Optimization = "B97M-rV/6-31g*/CMIRS(WATER) Geometry Optimization"
    B97M_rV_6_31g_d_CMIRS_WATER__Frequency_Analysis = "B97M-rV/6-31g*/CMIRS(WATER) Frequency Analysis"
    B97M_rV_6_31g_d_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_WATER__Unknown = "B97M-rV/6-31g*/CMIRS(WATER) Unknown"
    B97M_rV_6_31g_d_CMIRS_THF__Single_Point = "B97M-rV/6-31g*/CMIRS(THF) Single Point"
    B97M_rV_6_31g_d_CMIRS_THF__Force = "B97M-rV/6-31g*/CMIRS(THF) Force"
    B97M_rV_6_31g_d_CMIRS_THF__Geometry_Optimization = "B97M-rV/6-31g*/CMIRS(THF) Geometry Optimization"
    B97M_rV_6_31g_d_CMIRS_THF__Frequency_Analysis = "B97M-rV/6-31g*/CMIRS(THF) Frequency Analysis"
    B97M_rV_6_31g_d_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_THF__Unknown = "B97M-rV/6-31g*/CMIRS(THF) Unknown"
    B97M_rV_6_31g_d_CMIRS_G2__Single_Point = "B97M-rV/6-31g*/CMIRS(G2) Single Point"
    B97M_rV_6_31g_d_CMIRS_G2__Force = "B97M-rV/6-31g*/CMIRS(G2) Force"
    B97M_rV_6_31g_d_CMIRS_G2__Geometry_Optimization = "B97M-rV/6-31g*/CMIRS(G2) Geometry Optimization"
    B97M_rV_6_31g_d_CMIRS_G2__Frequency_Analysis = "B97M-rV/6-31g*/CMIRS(G2) Frequency Analysis"
    B97M_rV_6_31g_d_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_G2__Unknown = "B97M-rV/6-31g*/CMIRS(G2) Unknown"
    B97M_rV_6_31g_d_CMIRS_DME__Single_Point = "B97M-rV/6-31g*/CMIRS(DME) Single Point"
    B97M_rV_6_31g_d_CMIRS_DME__Force = "B97M-rV/6-31g*/CMIRS(DME) Force"
    B97M_rV_6_31g_d_CMIRS_DME__Geometry_Optimization = "B97M-rV/6-31g*/CMIRS(DME) Geometry Optimization"
    B97M_rV_6_31g_d_CMIRS_DME__Frequency_Analysis = "B97M-rV/6-31g*/CMIRS(DME) Frequency Analysis"
    B97M_rV_6_31g_d_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_DME__Unknown = "B97M-rV/6-31g*/CMIRS(DME) Unknown"
    B97M_rV_6_31g_d_CMIRS_EC__Single_Point = "B97M-rV/6-31g*/CMIRS(EC) Single Point"
    B97M_rV_6_31g_d_CMIRS_EC__Force = "B97M-rV/6-31g*/CMIRS(EC) Force"
    B97M_rV_6_31g_d_CMIRS_EC__Geometry_Optimization = "B97M-rV/6-31g*/CMIRS(EC) Geometry Optimization"
    B97M_rV_6_31g_d_CMIRS_EC__Frequency_Analysis = "B97M-rV/6-31g*/CMIRS(EC) Frequency Analysis"
    B97M_rV_6_31g_d_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_EC__Unknown = "B97M-rV/6-31g*/CMIRS(EC) Unknown"
    B97M_rV_6_31g_d_CMIRS_EC_EMC__Single_Point = "B97M-rV/6-31g*/CMIRS(EC/EMC) Single Point"
    B97M_rV_6_31g_d_CMIRS_EC_EMC__Force = "B97M-rV/6-31g*/CMIRS(EC/EMC) Force"
    B97M_rV_6_31g_d_CMIRS_EC_EMC__Geometry_Optimization = "B97M-rV/6-31g*/CMIRS(EC/EMC) Geometry Optimization"
    B97M_rV_6_31g_d_CMIRS_EC_EMC__Frequency_Analysis = "B97M-rV/6-31g*/CMIRS(EC/EMC) Frequency Analysis"
    B97M_rV_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_EC_EMC__Unknown = "B97M-rV/6-31g*/CMIRS(EC/EMC) Unknown"
    B97M_rV_6_31g_d_CMIRS_PC__Single_Point = "B97M-rV/6-31g*/CMIRS(PC) Single Point"
    B97M_rV_6_31g_d_CMIRS_PC__Force = "B97M-rV/6-31g*/CMIRS(PC) Force"
    B97M_rV_6_31g_d_CMIRS_PC__Geometry_Optimization = "B97M-rV/6-31g*/CMIRS(PC) Geometry Optimization"
    B97M_rV_6_31g_d_CMIRS_PC__Frequency_Analysis = "B97M-rV/6-31g*/CMIRS(PC) Frequency Analysis"
    B97M_rV_6_31g_d_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/6-31g*/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_6_31g_d_CMIRS_PC__Unknown = "B97M-rV/6-31g*/CMIRS(PC) Unknown"
    B97M_rV_def2_SVPD_VACUUM_Single_Point = "B97M-rV/def2-SVPD/VACUUM Single Point"
    B97M_rV_def2_SVPD_VACUUM_Force = "B97M-rV/def2-SVPD/VACUUM Force"
    B97M_rV_def2_SVPD_VACUUM_Geometry_Optimization = "B97M-rV/def2-SVPD/VACUUM Geometry Optimization"
    B97M_rV_def2_SVPD_VACUUM_Frequency_Analysis = "B97M-rV/def2-SVPD/VACUUM Frequency Analysis"
    B97M_rV_def2_SVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/VACUUM Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_VACUUM_Unknown = "B97M-rV/def2-SVPD/VACUUM Unknown"
    B97M_rV_def2_SVPD_PCM_WATER__Single_Point = "B97M-rV/def2-SVPD/PCM(WATER) Single Point"
    B97M_rV_def2_SVPD_PCM_WATER__Force = "B97M-rV/def2-SVPD/PCM(WATER) Force"
    B97M_rV_def2_SVPD_PCM_WATER__Geometry_Optimization = "B97M-rV/def2-SVPD/PCM(WATER) Geometry Optimization"
    B97M_rV_def2_SVPD_PCM_WATER__Frequency_Analysis = "B97M-rV/def2-SVPD/PCM(WATER) Frequency Analysis"
    B97M_rV_def2_SVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_WATER__Unknown = "B97M-rV/def2-SVPD/PCM(WATER) Unknown"
    B97M_rV_def2_SVPD_PCM_THF__Single_Point = "B97M-rV/def2-SVPD/PCM(THF) Single Point"
    B97M_rV_def2_SVPD_PCM_THF__Force = "B97M-rV/def2-SVPD/PCM(THF) Force"
    B97M_rV_def2_SVPD_PCM_THF__Geometry_Optimization = "B97M-rV/def2-SVPD/PCM(THF) Geometry Optimization"
    B97M_rV_def2_SVPD_PCM_THF__Frequency_Analysis = "B97M-rV/def2-SVPD/PCM(THF) Frequency Analysis"
    B97M_rV_def2_SVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_THF__Unknown = "B97M-rV/def2-SVPD/PCM(THF) Unknown"
    B97M_rV_def2_SVPD_PCM_G2__Single_Point = "B97M-rV/def2-SVPD/PCM(G2) Single Point"
    B97M_rV_def2_SVPD_PCM_G2__Force = "B97M-rV/def2-SVPD/PCM(G2) Force"
    B97M_rV_def2_SVPD_PCM_G2__Geometry_Optimization = "B97M-rV/def2-SVPD/PCM(G2) Geometry Optimization"
    B97M_rV_def2_SVPD_PCM_G2__Frequency_Analysis = "B97M-rV/def2-SVPD/PCM(G2) Frequency Analysis"
    B97M_rV_def2_SVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_G2__Unknown = "B97M-rV/def2-SVPD/PCM(G2) Unknown"
    B97M_rV_def2_SVPD_PCM_DME__Single_Point = "B97M-rV/def2-SVPD/PCM(DME) Single Point"
    B97M_rV_def2_SVPD_PCM_DME__Force = "B97M-rV/def2-SVPD/PCM(DME) Force"
    B97M_rV_def2_SVPD_PCM_DME__Geometry_Optimization = "B97M-rV/def2-SVPD/PCM(DME) Geometry Optimization"
    B97M_rV_def2_SVPD_PCM_DME__Frequency_Analysis = "B97M-rV/def2-SVPD/PCM(DME) Frequency Analysis"
    B97M_rV_def2_SVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_DME__Unknown = "B97M-rV/def2-SVPD/PCM(DME) Unknown"
    B97M_rV_def2_SVPD_PCM_EC__Single_Point = "B97M-rV/def2-SVPD/PCM(EC) Single Point"
    B97M_rV_def2_SVPD_PCM_EC__Force = "B97M-rV/def2-SVPD/PCM(EC) Force"
    B97M_rV_def2_SVPD_PCM_EC__Geometry_Optimization = "B97M-rV/def2-SVPD/PCM(EC) Geometry Optimization"
    B97M_rV_def2_SVPD_PCM_EC__Frequency_Analysis = "B97M-rV/def2-SVPD/PCM(EC) Frequency Analysis"
    B97M_rV_def2_SVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_EC__Unknown = "B97M-rV/def2-SVPD/PCM(EC) Unknown"
    B97M_rV_def2_SVPD_PCM_EC_EMC__Single_Point = "B97M-rV/def2-SVPD/PCM(EC/EMC) Single Point"
    B97M_rV_def2_SVPD_PCM_EC_EMC__Force = "B97M-rV/def2-SVPD/PCM(EC/EMC) Force"
    B97M_rV_def2_SVPD_PCM_EC_EMC__Geometry_Optimization = "B97M-rV/def2-SVPD/PCM(EC/EMC) Geometry Optimization"
    B97M_rV_def2_SVPD_PCM_EC_EMC__Frequency_Analysis = "B97M-rV/def2-SVPD/PCM(EC/EMC) Frequency Analysis"
    B97M_rV_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_EC_EMC__Unknown = "B97M-rV/def2-SVPD/PCM(EC/EMC) Unknown"
    B97M_rV_def2_SVPD_PCM_PC__Single_Point = "B97M-rV/def2-SVPD/PCM(PC) Single Point"
    B97M_rV_def2_SVPD_PCM_PC__Force = "B97M-rV/def2-SVPD/PCM(PC) Force"
    B97M_rV_def2_SVPD_PCM_PC__Geometry_Optimization = "B97M-rV/def2-SVPD/PCM(PC) Geometry Optimization"
    B97M_rV_def2_SVPD_PCM_PC__Frequency_Analysis = "B97M-rV/def2-SVPD/PCM(PC) Frequency Analysis"
    B97M_rV_def2_SVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_PCM_PC__Unknown = "B97M-rV/def2-SVPD/PCM(PC) Unknown"
    B97M_rV_def2_SVPD_ISOSVP_WATER__Single_Point = "B97M-rV/def2-SVPD/ISOSVP(WATER) Single Point"
    B97M_rV_def2_SVPD_ISOSVP_WATER__Force = "B97M-rV/def2-SVPD/ISOSVP(WATER) Force"
    B97M_rV_def2_SVPD_ISOSVP_WATER__Geometry_Optimization = "B97M-rV/def2-SVPD/ISOSVP(WATER) Geometry Optimization"
    B97M_rV_def2_SVPD_ISOSVP_WATER__Frequency_Analysis = "B97M-rV/def2-SVPD/ISOSVP(WATER) Frequency Analysis"
    B97M_rV_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_WATER__Unknown = "B97M-rV/def2-SVPD/ISOSVP(WATER) Unknown"
    B97M_rV_def2_SVPD_ISOSVP_THF__Single_Point = "B97M-rV/def2-SVPD/ISOSVP(THF) Single Point"
    B97M_rV_def2_SVPD_ISOSVP_THF__Force = "B97M-rV/def2-SVPD/ISOSVP(THF) Force"
    B97M_rV_def2_SVPD_ISOSVP_THF__Geometry_Optimization = "B97M-rV/def2-SVPD/ISOSVP(THF) Geometry Optimization"
    B97M_rV_def2_SVPD_ISOSVP_THF__Frequency_Analysis = "B97M-rV/def2-SVPD/ISOSVP(THF) Frequency Analysis"
    B97M_rV_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_THF__Unknown = "B97M-rV/def2-SVPD/ISOSVP(THF) Unknown"
    B97M_rV_def2_SVPD_ISOSVP_G2__Single_Point = "B97M-rV/def2-SVPD/ISOSVP(G2) Single Point"
    B97M_rV_def2_SVPD_ISOSVP_G2__Force = "B97M-rV/def2-SVPD/ISOSVP(G2) Force"
    B97M_rV_def2_SVPD_ISOSVP_G2__Geometry_Optimization = "B97M-rV/def2-SVPD/ISOSVP(G2) Geometry Optimization"
    B97M_rV_def2_SVPD_ISOSVP_G2__Frequency_Analysis = "B97M-rV/def2-SVPD/ISOSVP(G2) Frequency Analysis"
    B97M_rV_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_G2__Unknown = "B97M-rV/def2-SVPD/ISOSVP(G2) Unknown"
    B97M_rV_def2_SVPD_ISOSVP_DME__Single_Point = "B97M-rV/def2-SVPD/ISOSVP(DME) Single Point"
    B97M_rV_def2_SVPD_ISOSVP_DME__Force = "B97M-rV/def2-SVPD/ISOSVP(DME) Force"
    B97M_rV_def2_SVPD_ISOSVP_DME__Geometry_Optimization = "B97M-rV/def2-SVPD/ISOSVP(DME) Geometry Optimization"
    B97M_rV_def2_SVPD_ISOSVP_DME__Frequency_Analysis = "B97M-rV/def2-SVPD/ISOSVP(DME) Frequency Analysis"
    B97M_rV_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_DME__Unknown = "B97M-rV/def2-SVPD/ISOSVP(DME) Unknown"
    B97M_rV_def2_SVPD_ISOSVP_EC__Single_Point = "B97M-rV/def2-SVPD/ISOSVP(EC) Single Point"
    B97M_rV_def2_SVPD_ISOSVP_EC__Force = "B97M-rV/def2-SVPD/ISOSVP(EC) Force"
    B97M_rV_def2_SVPD_ISOSVP_EC__Geometry_Optimization = "B97M-rV/def2-SVPD/ISOSVP(EC) Geometry Optimization"
    B97M_rV_def2_SVPD_ISOSVP_EC__Frequency_Analysis = "B97M-rV/def2-SVPD/ISOSVP(EC) Frequency Analysis"
    B97M_rV_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_EC__Unknown = "B97M-rV/def2-SVPD/ISOSVP(EC) Unknown"
    B97M_rV_def2_SVPD_ISOSVP_EC_EMC__Single_Point = "B97M-rV/def2-SVPD/ISOSVP(EC/EMC) Single Point"
    B97M_rV_def2_SVPD_ISOSVP_EC_EMC__Force = "B97M-rV/def2-SVPD/ISOSVP(EC/EMC) Force"
    B97M_rV_def2_SVPD_ISOSVP_EC_EMC__Geometry_Optimization = "B97M-rV/def2-SVPD/ISOSVP(EC/EMC) Geometry Optimization"
    B97M_rV_def2_SVPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97M-rV/def2-SVPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97M_rV_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_EC_EMC__Unknown = "B97M-rV/def2-SVPD/ISOSVP(EC/EMC) Unknown"
    B97M_rV_def2_SVPD_ISOSVP_PC__Single_Point = "B97M-rV/def2-SVPD/ISOSVP(PC) Single Point"
    B97M_rV_def2_SVPD_ISOSVP_PC__Force = "B97M-rV/def2-SVPD/ISOSVP(PC) Force"
    B97M_rV_def2_SVPD_ISOSVP_PC__Geometry_Optimization = "B97M-rV/def2-SVPD/ISOSVP(PC) Geometry Optimization"
    B97M_rV_def2_SVPD_ISOSVP_PC__Frequency_Analysis = "B97M-rV/def2-SVPD/ISOSVP(PC) Frequency Analysis"
    B97M_rV_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_ISOSVP_PC__Unknown = "B97M-rV/def2-SVPD/ISOSVP(PC) Unknown"
    B97M_rV_def2_SVPD_SMD_WATER__Single_Point = "B97M-rV/def2-SVPD/SMD(WATER) Single Point"
    B97M_rV_def2_SVPD_SMD_WATER__Force = "B97M-rV/def2-SVPD/SMD(WATER) Force"
    B97M_rV_def2_SVPD_SMD_WATER__Geometry_Optimization = "B97M-rV/def2-SVPD/SMD(WATER) Geometry Optimization"
    B97M_rV_def2_SVPD_SMD_WATER__Frequency_Analysis = "B97M-rV/def2-SVPD/SMD(WATER) Frequency Analysis"
    B97M_rV_def2_SVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_WATER__Unknown = "B97M-rV/def2-SVPD/SMD(WATER) Unknown"
    B97M_rV_def2_SVPD_SMD_THF__Single_Point = "B97M-rV/def2-SVPD/SMD(THF) Single Point"
    B97M_rV_def2_SVPD_SMD_THF__Force = "B97M-rV/def2-SVPD/SMD(THF) Force"
    B97M_rV_def2_SVPD_SMD_THF__Geometry_Optimization = "B97M-rV/def2-SVPD/SMD(THF) Geometry Optimization"
    B97M_rV_def2_SVPD_SMD_THF__Frequency_Analysis = "B97M-rV/def2-SVPD/SMD(THF) Frequency Analysis"
    B97M_rV_def2_SVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_THF__Unknown = "B97M-rV/def2-SVPD/SMD(THF) Unknown"
    B97M_rV_def2_SVPD_SMD_G2__Single_Point = "B97M-rV/def2-SVPD/SMD(G2) Single Point"
    B97M_rV_def2_SVPD_SMD_G2__Force = "B97M-rV/def2-SVPD/SMD(G2) Force"
    B97M_rV_def2_SVPD_SMD_G2__Geometry_Optimization = "B97M-rV/def2-SVPD/SMD(G2) Geometry Optimization"
    B97M_rV_def2_SVPD_SMD_G2__Frequency_Analysis = "B97M-rV/def2-SVPD/SMD(G2) Frequency Analysis"
    B97M_rV_def2_SVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_G2__Unknown = "B97M-rV/def2-SVPD/SMD(G2) Unknown"
    B97M_rV_def2_SVPD_SMD_DME__Single_Point = "B97M-rV/def2-SVPD/SMD(DME) Single Point"
    B97M_rV_def2_SVPD_SMD_DME__Force = "B97M-rV/def2-SVPD/SMD(DME) Force"
    B97M_rV_def2_SVPD_SMD_DME__Geometry_Optimization = "B97M-rV/def2-SVPD/SMD(DME) Geometry Optimization"
    B97M_rV_def2_SVPD_SMD_DME__Frequency_Analysis = "B97M-rV/def2-SVPD/SMD(DME) Frequency Analysis"
    B97M_rV_def2_SVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_DME__Unknown = "B97M-rV/def2-SVPD/SMD(DME) Unknown"
    B97M_rV_def2_SVPD_SMD_EC__Single_Point = "B97M-rV/def2-SVPD/SMD(EC) Single Point"
    B97M_rV_def2_SVPD_SMD_EC__Force = "B97M-rV/def2-SVPD/SMD(EC) Force"
    B97M_rV_def2_SVPD_SMD_EC__Geometry_Optimization = "B97M-rV/def2-SVPD/SMD(EC) Geometry Optimization"
    B97M_rV_def2_SVPD_SMD_EC__Frequency_Analysis = "B97M-rV/def2-SVPD/SMD(EC) Frequency Analysis"
    B97M_rV_def2_SVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_EC__Unknown = "B97M-rV/def2-SVPD/SMD(EC) Unknown"
    B97M_rV_def2_SVPD_SMD_EC_EMC__Single_Point = "B97M-rV/def2-SVPD/SMD(EC/EMC) Single Point"
    B97M_rV_def2_SVPD_SMD_EC_EMC__Force = "B97M-rV/def2-SVPD/SMD(EC/EMC) Force"
    B97M_rV_def2_SVPD_SMD_EC_EMC__Geometry_Optimization = "B97M-rV/def2-SVPD/SMD(EC/EMC) Geometry Optimization"
    B97M_rV_def2_SVPD_SMD_EC_EMC__Frequency_Analysis = "B97M-rV/def2-SVPD/SMD(EC/EMC) Frequency Analysis"
    B97M_rV_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_EC_EMC__Unknown = "B97M-rV/def2-SVPD/SMD(EC/EMC) Unknown"
    B97M_rV_def2_SVPD_SMD_PC__Single_Point = "B97M-rV/def2-SVPD/SMD(PC) Single Point"
    B97M_rV_def2_SVPD_SMD_PC__Force = "B97M-rV/def2-SVPD/SMD(PC) Force"
    B97M_rV_def2_SVPD_SMD_PC__Geometry_Optimization = "B97M-rV/def2-SVPD/SMD(PC) Geometry Optimization"
    B97M_rV_def2_SVPD_SMD_PC__Frequency_Analysis = "B97M-rV/def2-SVPD/SMD(PC) Frequency Analysis"
    B97M_rV_def2_SVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_SMD_PC__Unknown = "B97M-rV/def2-SVPD/SMD(PC) Unknown"
    B97M_rV_def2_SVPD_CMIRS_WATER__Single_Point = "B97M-rV/def2-SVPD/CMIRS(WATER) Single Point"
    B97M_rV_def2_SVPD_CMIRS_WATER__Force = "B97M-rV/def2-SVPD/CMIRS(WATER) Force"
    B97M_rV_def2_SVPD_CMIRS_WATER__Geometry_Optimization = "B97M-rV/def2-SVPD/CMIRS(WATER) Geometry Optimization"
    B97M_rV_def2_SVPD_CMIRS_WATER__Frequency_Analysis = "B97M-rV/def2-SVPD/CMIRS(WATER) Frequency Analysis"
    B97M_rV_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_WATER__Unknown = "B97M-rV/def2-SVPD/CMIRS(WATER) Unknown"
    B97M_rV_def2_SVPD_CMIRS_THF__Single_Point = "B97M-rV/def2-SVPD/CMIRS(THF) Single Point"
    B97M_rV_def2_SVPD_CMIRS_THF__Force = "B97M-rV/def2-SVPD/CMIRS(THF) Force"
    B97M_rV_def2_SVPD_CMIRS_THF__Geometry_Optimization = "B97M-rV/def2-SVPD/CMIRS(THF) Geometry Optimization"
    B97M_rV_def2_SVPD_CMIRS_THF__Frequency_Analysis = "B97M-rV/def2-SVPD/CMIRS(THF) Frequency Analysis"
    B97M_rV_def2_SVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_THF__Unknown = "B97M-rV/def2-SVPD/CMIRS(THF) Unknown"
    B97M_rV_def2_SVPD_CMIRS_G2__Single_Point = "B97M-rV/def2-SVPD/CMIRS(G2) Single Point"
    B97M_rV_def2_SVPD_CMIRS_G2__Force = "B97M-rV/def2-SVPD/CMIRS(G2) Force"
    B97M_rV_def2_SVPD_CMIRS_G2__Geometry_Optimization = "B97M-rV/def2-SVPD/CMIRS(G2) Geometry Optimization"
    B97M_rV_def2_SVPD_CMIRS_G2__Frequency_Analysis = "B97M-rV/def2-SVPD/CMIRS(G2) Frequency Analysis"
    B97M_rV_def2_SVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_G2__Unknown = "B97M-rV/def2-SVPD/CMIRS(G2) Unknown"
    B97M_rV_def2_SVPD_CMIRS_DME__Single_Point = "B97M-rV/def2-SVPD/CMIRS(DME) Single Point"
    B97M_rV_def2_SVPD_CMIRS_DME__Force = "B97M-rV/def2-SVPD/CMIRS(DME) Force"
    B97M_rV_def2_SVPD_CMIRS_DME__Geometry_Optimization = "B97M-rV/def2-SVPD/CMIRS(DME) Geometry Optimization"
    B97M_rV_def2_SVPD_CMIRS_DME__Frequency_Analysis = "B97M-rV/def2-SVPD/CMIRS(DME) Frequency Analysis"
    B97M_rV_def2_SVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_DME__Unknown = "B97M-rV/def2-SVPD/CMIRS(DME) Unknown"
    B97M_rV_def2_SVPD_CMIRS_EC__Single_Point = "B97M-rV/def2-SVPD/CMIRS(EC) Single Point"
    B97M_rV_def2_SVPD_CMIRS_EC__Force = "B97M-rV/def2-SVPD/CMIRS(EC) Force"
    B97M_rV_def2_SVPD_CMIRS_EC__Geometry_Optimization = "B97M-rV/def2-SVPD/CMIRS(EC) Geometry Optimization"
    B97M_rV_def2_SVPD_CMIRS_EC__Frequency_Analysis = "B97M-rV/def2-SVPD/CMIRS(EC) Frequency Analysis"
    B97M_rV_def2_SVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_EC__Unknown = "B97M-rV/def2-SVPD/CMIRS(EC) Unknown"
    B97M_rV_def2_SVPD_CMIRS_EC_EMC__Single_Point = "B97M-rV/def2-SVPD/CMIRS(EC/EMC) Single Point"
    B97M_rV_def2_SVPD_CMIRS_EC_EMC__Force = "B97M-rV/def2-SVPD/CMIRS(EC/EMC) Force"
    B97M_rV_def2_SVPD_CMIRS_EC_EMC__Geometry_Optimization = "B97M-rV/def2-SVPD/CMIRS(EC/EMC) Geometry Optimization"
    B97M_rV_def2_SVPD_CMIRS_EC_EMC__Frequency_Analysis = "B97M-rV/def2-SVPD/CMIRS(EC/EMC) Frequency Analysis"
    B97M_rV_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_EC_EMC__Unknown = "B97M-rV/def2-SVPD/CMIRS(EC/EMC) Unknown"
    B97M_rV_def2_SVPD_CMIRS_PC__Single_Point = "B97M-rV/def2-SVPD/CMIRS(PC) Single Point"
    B97M_rV_def2_SVPD_CMIRS_PC__Force = "B97M-rV/def2-SVPD/CMIRS(PC) Force"
    B97M_rV_def2_SVPD_CMIRS_PC__Geometry_Optimization = "B97M-rV/def2-SVPD/CMIRS(PC) Geometry Optimization"
    B97M_rV_def2_SVPD_CMIRS_PC__Frequency_Analysis = "B97M-rV/def2-SVPD/CMIRS(PC) Frequency Analysis"
    B97M_rV_def2_SVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-SVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_SVPD_CMIRS_PC__Unknown = "B97M-rV/def2-SVPD/CMIRS(PC) Unknown"
    B97M_rV_def2_TZVP_VACUUM_Single_Point = "B97M-rV/def2-TZVP/VACUUM Single Point"
    B97M_rV_def2_TZVP_VACUUM_Force = "B97M-rV/def2-TZVP/VACUUM Force"
    B97M_rV_def2_TZVP_VACUUM_Geometry_Optimization = "B97M-rV/def2-TZVP/VACUUM Geometry Optimization"
    B97M_rV_def2_TZVP_VACUUM_Frequency_Analysis = "B97M-rV/def2-TZVP/VACUUM Frequency Analysis"
    B97M_rV_def2_TZVP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_VACUUM_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/VACUUM Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_VACUUM_Unknown = "B97M-rV/def2-TZVP/VACUUM Unknown"
    B97M_rV_def2_TZVP_PCM_WATER__Single_Point = "B97M-rV/def2-TZVP/PCM(WATER) Single Point"
    B97M_rV_def2_TZVP_PCM_WATER__Force = "B97M-rV/def2-TZVP/PCM(WATER) Force"
    B97M_rV_def2_TZVP_PCM_WATER__Geometry_Optimization = "B97M-rV/def2-TZVP/PCM(WATER) Geometry Optimization"
    B97M_rV_def2_TZVP_PCM_WATER__Frequency_Analysis = "B97M-rV/def2-TZVP/PCM(WATER) Frequency Analysis"
    B97M_rV_def2_TZVP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_WATER__Unknown = "B97M-rV/def2-TZVP/PCM(WATER) Unknown"
    B97M_rV_def2_TZVP_PCM_THF__Single_Point = "B97M-rV/def2-TZVP/PCM(THF) Single Point"
    B97M_rV_def2_TZVP_PCM_THF__Force = "B97M-rV/def2-TZVP/PCM(THF) Force"
    B97M_rV_def2_TZVP_PCM_THF__Geometry_Optimization = "B97M-rV/def2-TZVP/PCM(THF) Geometry Optimization"
    B97M_rV_def2_TZVP_PCM_THF__Frequency_Analysis = "B97M-rV/def2-TZVP/PCM(THF) Frequency Analysis"
    B97M_rV_def2_TZVP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_THF__Unknown = "B97M-rV/def2-TZVP/PCM(THF) Unknown"
    B97M_rV_def2_TZVP_PCM_G2__Single_Point = "B97M-rV/def2-TZVP/PCM(G2) Single Point"
    B97M_rV_def2_TZVP_PCM_G2__Force = "B97M-rV/def2-TZVP/PCM(G2) Force"
    B97M_rV_def2_TZVP_PCM_G2__Geometry_Optimization = "B97M-rV/def2-TZVP/PCM(G2) Geometry Optimization"
    B97M_rV_def2_TZVP_PCM_G2__Frequency_Analysis = "B97M-rV/def2-TZVP/PCM(G2) Frequency Analysis"
    B97M_rV_def2_TZVP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_G2__Unknown = "B97M-rV/def2-TZVP/PCM(G2) Unknown"
    B97M_rV_def2_TZVP_PCM_DME__Single_Point = "B97M-rV/def2-TZVP/PCM(DME) Single Point"
    B97M_rV_def2_TZVP_PCM_DME__Force = "B97M-rV/def2-TZVP/PCM(DME) Force"
    B97M_rV_def2_TZVP_PCM_DME__Geometry_Optimization = "B97M-rV/def2-TZVP/PCM(DME) Geometry Optimization"
    B97M_rV_def2_TZVP_PCM_DME__Frequency_Analysis = "B97M-rV/def2-TZVP/PCM(DME) Frequency Analysis"
    B97M_rV_def2_TZVP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_DME__Unknown = "B97M-rV/def2-TZVP/PCM(DME) Unknown"
    B97M_rV_def2_TZVP_PCM_EC__Single_Point = "B97M-rV/def2-TZVP/PCM(EC) Single Point"
    B97M_rV_def2_TZVP_PCM_EC__Force = "B97M-rV/def2-TZVP/PCM(EC) Force"
    B97M_rV_def2_TZVP_PCM_EC__Geometry_Optimization = "B97M-rV/def2-TZVP/PCM(EC) Geometry Optimization"
    B97M_rV_def2_TZVP_PCM_EC__Frequency_Analysis = "B97M-rV/def2-TZVP/PCM(EC) Frequency Analysis"
    B97M_rV_def2_TZVP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_EC__Unknown = "B97M-rV/def2-TZVP/PCM(EC) Unknown"
    B97M_rV_def2_TZVP_PCM_EC_EMC__Single_Point = "B97M-rV/def2-TZVP/PCM(EC/EMC) Single Point"
    B97M_rV_def2_TZVP_PCM_EC_EMC__Force = "B97M-rV/def2-TZVP/PCM(EC/EMC) Force"
    B97M_rV_def2_TZVP_PCM_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVP/PCM(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVP_PCM_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVP/PCM(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_EC_EMC__Unknown = "B97M-rV/def2-TZVP/PCM(EC/EMC) Unknown"
    B97M_rV_def2_TZVP_PCM_PC__Single_Point = "B97M-rV/def2-TZVP/PCM(PC) Single Point"
    B97M_rV_def2_TZVP_PCM_PC__Force = "B97M-rV/def2-TZVP/PCM(PC) Force"
    B97M_rV_def2_TZVP_PCM_PC__Geometry_Optimization = "B97M-rV/def2-TZVP/PCM(PC) Geometry Optimization"
    B97M_rV_def2_TZVP_PCM_PC__Frequency_Analysis = "B97M-rV/def2-TZVP/PCM(PC) Frequency Analysis"
    B97M_rV_def2_TZVP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_PCM_PC__Unknown = "B97M-rV/def2-TZVP/PCM(PC) Unknown"
    B97M_rV_def2_TZVP_ISOSVP_WATER__Single_Point = "B97M-rV/def2-TZVP/ISOSVP(WATER) Single Point"
    B97M_rV_def2_TZVP_ISOSVP_WATER__Force = "B97M-rV/def2-TZVP/ISOSVP(WATER) Force"
    B97M_rV_def2_TZVP_ISOSVP_WATER__Geometry_Optimization = "B97M-rV/def2-TZVP/ISOSVP(WATER) Geometry Optimization"
    B97M_rV_def2_TZVP_ISOSVP_WATER__Frequency_Analysis = "B97M-rV/def2-TZVP/ISOSVP(WATER) Frequency Analysis"
    B97M_rV_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_WATER__Unknown = "B97M-rV/def2-TZVP/ISOSVP(WATER) Unknown"
    B97M_rV_def2_TZVP_ISOSVP_THF__Single_Point = "B97M-rV/def2-TZVP/ISOSVP(THF) Single Point"
    B97M_rV_def2_TZVP_ISOSVP_THF__Force = "B97M-rV/def2-TZVP/ISOSVP(THF) Force"
    B97M_rV_def2_TZVP_ISOSVP_THF__Geometry_Optimization = "B97M-rV/def2-TZVP/ISOSVP(THF) Geometry Optimization"
    B97M_rV_def2_TZVP_ISOSVP_THF__Frequency_Analysis = "B97M-rV/def2-TZVP/ISOSVP(THF) Frequency Analysis"
    B97M_rV_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_THF__Unknown = "B97M-rV/def2-TZVP/ISOSVP(THF) Unknown"
    B97M_rV_def2_TZVP_ISOSVP_G2__Single_Point = "B97M-rV/def2-TZVP/ISOSVP(G2) Single Point"
    B97M_rV_def2_TZVP_ISOSVP_G2__Force = "B97M-rV/def2-TZVP/ISOSVP(G2) Force"
    B97M_rV_def2_TZVP_ISOSVP_G2__Geometry_Optimization = "B97M-rV/def2-TZVP/ISOSVP(G2) Geometry Optimization"
    B97M_rV_def2_TZVP_ISOSVP_G2__Frequency_Analysis = "B97M-rV/def2-TZVP/ISOSVP(G2) Frequency Analysis"
    B97M_rV_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_G2__Unknown = "B97M-rV/def2-TZVP/ISOSVP(G2) Unknown"
    B97M_rV_def2_TZVP_ISOSVP_DME__Single_Point = "B97M-rV/def2-TZVP/ISOSVP(DME) Single Point"
    B97M_rV_def2_TZVP_ISOSVP_DME__Force = "B97M-rV/def2-TZVP/ISOSVP(DME) Force"
    B97M_rV_def2_TZVP_ISOSVP_DME__Geometry_Optimization = "B97M-rV/def2-TZVP/ISOSVP(DME) Geometry Optimization"
    B97M_rV_def2_TZVP_ISOSVP_DME__Frequency_Analysis = "B97M-rV/def2-TZVP/ISOSVP(DME) Frequency Analysis"
    B97M_rV_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_DME__Unknown = "B97M-rV/def2-TZVP/ISOSVP(DME) Unknown"
    B97M_rV_def2_TZVP_ISOSVP_EC__Single_Point = "B97M-rV/def2-TZVP/ISOSVP(EC) Single Point"
    B97M_rV_def2_TZVP_ISOSVP_EC__Force = "B97M-rV/def2-TZVP/ISOSVP(EC) Force"
    B97M_rV_def2_TZVP_ISOSVP_EC__Geometry_Optimization = "B97M-rV/def2-TZVP/ISOSVP(EC) Geometry Optimization"
    B97M_rV_def2_TZVP_ISOSVP_EC__Frequency_Analysis = "B97M-rV/def2-TZVP/ISOSVP(EC) Frequency Analysis"
    B97M_rV_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_EC__Unknown = "B97M-rV/def2-TZVP/ISOSVP(EC) Unknown"
    B97M_rV_def2_TZVP_ISOSVP_EC_EMC__Single_Point = "B97M-rV/def2-TZVP/ISOSVP(EC/EMC) Single Point"
    B97M_rV_def2_TZVP_ISOSVP_EC_EMC__Force = "B97M-rV/def2-TZVP/ISOSVP(EC/EMC) Force"
    B97M_rV_def2_TZVP_ISOSVP_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVP/ISOSVP(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVP_ISOSVP_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVP/ISOSVP(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_EC_EMC__Unknown = "B97M-rV/def2-TZVP/ISOSVP(EC/EMC) Unknown"
    B97M_rV_def2_TZVP_ISOSVP_PC__Single_Point = "B97M-rV/def2-TZVP/ISOSVP(PC) Single Point"
    B97M_rV_def2_TZVP_ISOSVP_PC__Force = "B97M-rV/def2-TZVP/ISOSVP(PC) Force"
    B97M_rV_def2_TZVP_ISOSVP_PC__Geometry_Optimization = "B97M-rV/def2-TZVP/ISOSVP(PC) Geometry Optimization"
    B97M_rV_def2_TZVP_ISOSVP_PC__Frequency_Analysis = "B97M-rV/def2-TZVP/ISOSVP(PC) Frequency Analysis"
    B97M_rV_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_ISOSVP_PC__Unknown = "B97M-rV/def2-TZVP/ISOSVP(PC) Unknown"
    B97M_rV_def2_TZVP_SMD_WATER__Single_Point = "B97M-rV/def2-TZVP/SMD(WATER) Single Point"
    B97M_rV_def2_TZVP_SMD_WATER__Force = "B97M-rV/def2-TZVP/SMD(WATER) Force"
    B97M_rV_def2_TZVP_SMD_WATER__Geometry_Optimization = "B97M-rV/def2-TZVP/SMD(WATER) Geometry Optimization"
    B97M_rV_def2_TZVP_SMD_WATER__Frequency_Analysis = "B97M-rV/def2-TZVP/SMD(WATER) Frequency Analysis"
    B97M_rV_def2_TZVP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_WATER__Unknown = "B97M-rV/def2-TZVP/SMD(WATER) Unknown"
    B97M_rV_def2_TZVP_SMD_THF__Single_Point = "B97M-rV/def2-TZVP/SMD(THF) Single Point"
    B97M_rV_def2_TZVP_SMD_THF__Force = "B97M-rV/def2-TZVP/SMD(THF) Force"
    B97M_rV_def2_TZVP_SMD_THF__Geometry_Optimization = "B97M-rV/def2-TZVP/SMD(THF) Geometry Optimization"
    B97M_rV_def2_TZVP_SMD_THF__Frequency_Analysis = "B97M-rV/def2-TZVP/SMD(THF) Frequency Analysis"
    B97M_rV_def2_TZVP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_THF__Unknown = "B97M-rV/def2-TZVP/SMD(THF) Unknown"
    B97M_rV_def2_TZVP_SMD_G2__Single_Point = "B97M-rV/def2-TZVP/SMD(G2) Single Point"
    B97M_rV_def2_TZVP_SMD_G2__Force = "B97M-rV/def2-TZVP/SMD(G2) Force"
    B97M_rV_def2_TZVP_SMD_G2__Geometry_Optimization = "B97M-rV/def2-TZVP/SMD(G2) Geometry Optimization"
    B97M_rV_def2_TZVP_SMD_G2__Frequency_Analysis = "B97M-rV/def2-TZVP/SMD(G2) Frequency Analysis"
    B97M_rV_def2_TZVP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_G2__Unknown = "B97M-rV/def2-TZVP/SMD(G2) Unknown"
    B97M_rV_def2_TZVP_SMD_DME__Single_Point = "B97M-rV/def2-TZVP/SMD(DME) Single Point"
    B97M_rV_def2_TZVP_SMD_DME__Force = "B97M-rV/def2-TZVP/SMD(DME) Force"
    B97M_rV_def2_TZVP_SMD_DME__Geometry_Optimization = "B97M-rV/def2-TZVP/SMD(DME) Geometry Optimization"
    B97M_rV_def2_TZVP_SMD_DME__Frequency_Analysis = "B97M-rV/def2-TZVP/SMD(DME) Frequency Analysis"
    B97M_rV_def2_TZVP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_DME__Unknown = "B97M-rV/def2-TZVP/SMD(DME) Unknown"
    B97M_rV_def2_TZVP_SMD_EC__Single_Point = "B97M-rV/def2-TZVP/SMD(EC) Single Point"
    B97M_rV_def2_TZVP_SMD_EC__Force = "B97M-rV/def2-TZVP/SMD(EC) Force"
    B97M_rV_def2_TZVP_SMD_EC__Geometry_Optimization = "B97M-rV/def2-TZVP/SMD(EC) Geometry Optimization"
    B97M_rV_def2_TZVP_SMD_EC__Frequency_Analysis = "B97M-rV/def2-TZVP/SMD(EC) Frequency Analysis"
    B97M_rV_def2_TZVP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_EC__Unknown = "B97M-rV/def2-TZVP/SMD(EC) Unknown"
    B97M_rV_def2_TZVP_SMD_EC_EMC__Single_Point = "B97M-rV/def2-TZVP/SMD(EC/EMC) Single Point"
    B97M_rV_def2_TZVP_SMD_EC_EMC__Force = "B97M-rV/def2-TZVP/SMD(EC/EMC) Force"
    B97M_rV_def2_TZVP_SMD_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVP/SMD(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVP_SMD_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVP/SMD(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_EC_EMC__Unknown = "B97M-rV/def2-TZVP/SMD(EC/EMC) Unknown"
    B97M_rV_def2_TZVP_SMD_PC__Single_Point = "B97M-rV/def2-TZVP/SMD(PC) Single Point"
    B97M_rV_def2_TZVP_SMD_PC__Force = "B97M-rV/def2-TZVP/SMD(PC) Force"
    B97M_rV_def2_TZVP_SMD_PC__Geometry_Optimization = "B97M-rV/def2-TZVP/SMD(PC) Geometry Optimization"
    B97M_rV_def2_TZVP_SMD_PC__Frequency_Analysis = "B97M-rV/def2-TZVP/SMD(PC) Frequency Analysis"
    B97M_rV_def2_TZVP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_SMD_PC__Unknown = "B97M-rV/def2-TZVP/SMD(PC) Unknown"
    B97M_rV_def2_TZVP_CMIRS_WATER__Single_Point = "B97M-rV/def2-TZVP/CMIRS(WATER) Single Point"
    B97M_rV_def2_TZVP_CMIRS_WATER__Force = "B97M-rV/def2-TZVP/CMIRS(WATER) Force"
    B97M_rV_def2_TZVP_CMIRS_WATER__Geometry_Optimization = "B97M-rV/def2-TZVP/CMIRS(WATER) Geometry Optimization"
    B97M_rV_def2_TZVP_CMIRS_WATER__Frequency_Analysis = "B97M-rV/def2-TZVP/CMIRS(WATER) Frequency Analysis"
    B97M_rV_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_WATER__Unknown = "B97M-rV/def2-TZVP/CMIRS(WATER) Unknown"
    B97M_rV_def2_TZVP_CMIRS_THF__Single_Point = "B97M-rV/def2-TZVP/CMIRS(THF) Single Point"
    B97M_rV_def2_TZVP_CMIRS_THF__Force = "B97M-rV/def2-TZVP/CMIRS(THF) Force"
    B97M_rV_def2_TZVP_CMIRS_THF__Geometry_Optimization = "B97M-rV/def2-TZVP/CMIRS(THF) Geometry Optimization"
    B97M_rV_def2_TZVP_CMIRS_THF__Frequency_Analysis = "B97M-rV/def2-TZVP/CMIRS(THF) Frequency Analysis"
    B97M_rV_def2_TZVP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_THF__Unknown = "B97M-rV/def2-TZVP/CMIRS(THF) Unknown"
    B97M_rV_def2_TZVP_CMIRS_G2__Single_Point = "B97M-rV/def2-TZVP/CMIRS(G2) Single Point"
    B97M_rV_def2_TZVP_CMIRS_G2__Force = "B97M-rV/def2-TZVP/CMIRS(G2) Force"
    B97M_rV_def2_TZVP_CMIRS_G2__Geometry_Optimization = "B97M-rV/def2-TZVP/CMIRS(G2) Geometry Optimization"
    B97M_rV_def2_TZVP_CMIRS_G2__Frequency_Analysis = "B97M-rV/def2-TZVP/CMIRS(G2) Frequency Analysis"
    B97M_rV_def2_TZVP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_G2__Unknown = "B97M-rV/def2-TZVP/CMIRS(G2) Unknown"
    B97M_rV_def2_TZVP_CMIRS_DME__Single_Point = "B97M-rV/def2-TZVP/CMIRS(DME) Single Point"
    B97M_rV_def2_TZVP_CMIRS_DME__Force = "B97M-rV/def2-TZVP/CMIRS(DME) Force"
    B97M_rV_def2_TZVP_CMIRS_DME__Geometry_Optimization = "B97M-rV/def2-TZVP/CMIRS(DME) Geometry Optimization"
    B97M_rV_def2_TZVP_CMIRS_DME__Frequency_Analysis = "B97M-rV/def2-TZVP/CMIRS(DME) Frequency Analysis"
    B97M_rV_def2_TZVP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_DME__Unknown = "B97M-rV/def2-TZVP/CMIRS(DME) Unknown"
    B97M_rV_def2_TZVP_CMIRS_EC__Single_Point = "B97M-rV/def2-TZVP/CMIRS(EC) Single Point"
    B97M_rV_def2_TZVP_CMIRS_EC__Force = "B97M-rV/def2-TZVP/CMIRS(EC) Force"
    B97M_rV_def2_TZVP_CMIRS_EC__Geometry_Optimization = "B97M-rV/def2-TZVP/CMIRS(EC) Geometry Optimization"
    B97M_rV_def2_TZVP_CMIRS_EC__Frequency_Analysis = "B97M-rV/def2-TZVP/CMIRS(EC) Frequency Analysis"
    B97M_rV_def2_TZVP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_EC__Unknown = "B97M-rV/def2-TZVP/CMIRS(EC) Unknown"
    B97M_rV_def2_TZVP_CMIRS_EC_EMC__Single_Point = "B97M-rV/def2-TZVP/CMIRS(EC/EMC) Single Point"
    B97M_rV_def2_TZVP_CMIRS_EC_EMC__Force = "B97M-rV/def2-TZVP/CMIRS(EC/EMC) Force"
    B97M_rV_def2_TZVP_CMIRS_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVP/CMIRS(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVP_CMIRS_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVP/CMIRS(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_EC_EMC__Unknown = "B97M-rV/def2-TZVP/CMIRS(EC/EMC) Unknown"
    B97M_rV_def2_TZVP_CMIRS_PC__Single_Point = "B97M-rV/def2-TZVP/CMIRS(PC) Single Point"
    B97M_rV_def2_TZVP_CMIRS_PC__Force = "B97M-rV/def2-TZVP/CMIRS(PC) Force"
    B97M_rV_def2_TZVP_CMIRS_PC__Geometry_Optimization = "B97M-rV/def2-TZVP/CMIRS(PC) Geometry Optimization"
    B97M_rV_def2_TZVP_CMIRS_PC__Frequency_Analysis = "B97M-rV/def2-TZVP/CMIRS(PC) Frequency Analysis"
    B97M_rV_def2_TZVP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVP_CMIRS_PC__Unknown = "B97M-rV/def2-TZVP/CMIRS(PC) Unknown"
    B97M_rV_def2_TZVPD_VACUUM_Single_Point = "B97M-rV/def2-TZVPD/VACUUM Single Point"
    B97M_rV_def2_TZVPD_VACUUM_Force = "B97M-rV/def2-TZVPD/VACUUM Force"
    B97M_rV_def2_TZVPD_VACUUM_Geometry_Optimization = "B97M-rV/def2-TZVPD/VACUUM Geometry Optimization"
    B97M_rV_def2_TZVPD_VACUUM_Frequency_Analysis = "B97M-rV/def2-TZVPD/VACUUM Frequency Analysis"
    B97M_rV_def2_TZVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/VACUUM Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_VACUUM_Unknown = "B97M-rV/def2-TZVPD/VACUUM Unknown"
    B97M_rV_def2_TZVPD_PCM_WATER__Single_Point = "B97M-rV/def2-TZVPD/PCM(WATER) Single Point"
    B97M_rV_def2_TZVPD_PCM_WATER__Force = "B97M-rV/def2-TZVPD/PCM(WATER) Force"
    B97M_rV_def2_TZVPD_PCM_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPD/PCM(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPD_PCM_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPD/PCM(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_WATER__Unknown = "B97M-rV/def2-TZVPD/PCM(WATER) Unknown"
    B97M_rV_def2_TZVPD_PCM_THF__Single_Point = "B97M-rV/def2-TZVPD/PCM(THF) Single Point"
    B97M_rV_def2_TZVPD_PCM_THF__Force = "B97M-rV/def2-TZVPD/PCM(THF) Force"
    B97M_rV_def2_TZVPD_PCM_THF__Geometry_Optimization = "B97M-rV/def2-TZVPD/PCM(THF) Geometry Optimization"
    B97M_rV_def2_TZVPD_PCM_THF__Frequency_Analysis = "B97M-rV/def2-TZVPD/PCM(THF) Frequency Analysis"
    B97M_rV_def2_TZVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_THF__Unknown = "B97M-rV/def2-TZVPD/PCM(THF) Unknown"
    B97M_rV_def2_TZVPD_PCM_G2__Single_Point = "B97M-rV/def2-TZVPD/PCM(G2) Single Point"
    B97M_rV_def2_TZVPD_PCM_G2__Force = "B97M-rV/def2-TZVPD/PCM(G2) Force"
    B97M_rV_def2_TZVPD_PCM_G2__Geometry_Optimization = "B97M-rV/def2-TZVPD/PCM(G2) Geometry Optimization"
    B97M_rV_def2_TZVPD_PCM_G2__Frequency_Analysis = "B97M-rV/def2-TZVPD/PCM(G2) Frequency Analysis"
    B97M_rV_def2_TZVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_G2__Unknown = "B97M-rV/def2-TZVPD/PCM(G2) Unknown"
    B97M_rV_def2_TZVPD_PCM_DME__Single_Point = "B97M-rV/def2-TZVPD/PCM(DME) Single Point"
    B97M_rV_def2_TZVPD_PCM_DME__Force = "B97M-rV/def2-TZVPD/PCM(DME) Force"
    B97M_rV_def2_TZVPD_PCM_DME__Geometry_Optimization = "B97M-rV/def2-TZVPD/PCM(DME) Geometry Optimization"
    B97M_rV_def2_TZVPD_PCM_DME__Frequency_Analysis = "B97M-rV/def2-TZVPD/PCM(DME) Frequency Analysis"
    B97M_rV_def2_TZVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_DME__Unknown = "B97M-rV/def2-TZVPD/PCM(DME) Unknown"
    B97M_rV_def2_TZVPD_PCM_EC__Single_Point = "B97M-rV/def2-TZVPD/PCM(EC) Single Point"
    B97M_rV_def2_TZVPD_PCM_EC__Force = "B97M-rV/def2-TZVPD/PCM(EC) Force"
    B97M_rV_def2_TZVPD_PCM_EC__Geometry_Optimization = "B97M-rV/def2-TZVPD/PCM(EC) Geometry Optimization"
    B97M_rV_def2_TZVPD_PCM_EC__Frequency_Analysis = "B97M-rV/def2-TZVPD/PCM(EC) Frequency Analysis"
    B97M_rV_def2_TZVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_EC__Unknown = "B97M-rV/def2-TZVPD/PCM(EC) Unknown"
    B97M_rV_def2_TZVPD_PCM_EC_EMC__Single_Point = "B97M-rV/def2-TZVPD/PCM(EC/EMC) Single Point"
    B97M_rV_def2_TZVPD_PCM_EC_EMC__Force = "B97M-rV/def2-TZVPD/PCM(EC/EMC) Force"
    B97M_rV_def2_TZVPD_PCM_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVPD/PCM(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVPD_PCM_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPD/PCM(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_EC_EMC__Unknown = "B97M-rV/def2-TZVPD/PCM(EC/EMC) Unknown"
    B97M_rV_def2_TZVPD_PCM_PC__Single_Point = "B97M-rV/def2-TZVPD/PCM(PC) Single Point"
    B97M_rV_def2_TZVPD_PCM_PC__Force = "B97M-rV/def2-TZVPD/PCM(PC) Force"
    B97M_rV_def2_TZVPD_PCM_PC__Geometry_Optimization = "B97M-rV/def2-TZVPD/PCM(PC) Geometry Optimization"
    B97M_rV_def2_TZVPD_PCM_PC__Frequency_Analysis = "B97M-rV/def2-TZVPD/PCM(PC) Frequency Analysis"
    B97M_rV_def2_TZVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_PCM_PC__Unknown = "B97M-rV/def2-TZVPD/PCM(PC) Unknown"
    B97M_rV_def2_TZVPD_ISOSVP_WATER__Single_Point = "B97M-rV/def2-TZVPD/ISOSVP(WATER) Single Point"
    B97M_rV_def2_TZVPD_ISOSVP_WATER__Force = "B97M-rV/def2-TZVPD/ISOSVP(WATER) Force"
    B97M_rV_def2_TZVPD_ISOSVP_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPD/ISOSVP(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPD_ISOSVP_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPD/ISOSVP(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_WATER__Unknown = "B97M-rV/def2-TZVPD/ISOSVP(WATER) Unknown"
    B97M_rV_def2_TZVPD_ISOSVP_THF__Single_Point = "B97M-rV/def2-TZVPD/ISOSVP(THF) Single Point"
    B97M_rV_def2_TZVPD_ISOSVP_THF__Force = "B97M-rV/def2-TZVPD/ISOSVP(THF) Force"
    B97M_rV_def2_TZVPD_ISOSVP_THF__Geometry_Optimization = "B97M-rV/def2-TZVPD/ISOSVP(THF) Geometry Optimization"
    B97M_rV_def2_TZVPD_ISOSVP_THF__Frequency_Analysis = "B97M-rV/def2-TZVPD/ISOSVP(THF) Frequency Analysis"
    B97M_rV_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_THF__Unknown = "B97M-rV/def2-TZVPD/ISOSVP(THF) Unknown"
    B97M_rV_def2_TZVPD_ISOSVP_G2__Single_Point = "B97M-rV/def2-TZVPD/ISOSVP(G2) Single Point"
    B97M_rV_def2_TZVPD_ISOSVP_G2__Force = "B97M-rV/def2-TZVPD/ISOSVP(G2) Force"
    B97M_rV_def2_TZVPD_ISOSVP_G2__Geometry_Optimization = "B97M-rV/def2-TZVPD/ISOSVP(G2) Geometry Optimization"
    B97M_rV_def2_TZVPD_ISOSVP_G2__Frequency_Analysis = "B97M-rV/def2-TZVPD/ISOSVP(G2) Frequency Analysis"
    B97M_rV_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_G2__Unknown = "B97M-rV/def2-TZVPD/ISOSVP(G2) Unknown"
    B97M_rV_def2_TZVPD_ISOSVP_DME__Single_Point = "B97M-rV/def2-TZVPD/ISOSVP(DME) Single Point"
    B97M_rV_def2_TZVPD_ISOSVP_DME__Force = "B97M-rV/def2-TZVPD/ISOSVP(DME) Force"
    B97M_rV_def2_TZVPD_ISOSVP_DME__Geometry_Optimization = "B97M-rV/def2-TZVPD/ISOSVP(DME) Geometry Optimization"
    B97M_rV_def2_TZVPD_ISOSVP_DME__Frequency_Analysis = "B97M-rV/def2-TZVPD/ISOSVP(DME) Frequency Analysis"
    B97M_rV_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_DME__Unknown = "B97M-rV/def2-TZVPD/ISOSVP(DME) Unknown"
    B97M_rV_def2_TZVPD_ISOSVP_EC__Single_Point = "B97M-rV/def2-TZVPD/ISOSVP(EC) Single Point"
    B97M_rV_def2_TZVPD_ISOSVP_EC__Force = "B97M-rV/def2-TZVPD/ISOSVP(EC) Force"
    B97M_rV_def2_TZVPD_ISOSVP_EC__Geometry_Optimization = "B97M-rV/def2-TZVPD/ISOSVP(EC) Geometry Optimization"
    B97M_rV_def2_TZVPD_ISOSVP_EC__Frequency_Analysis = "B97M-rV/def2-TZVPD/ISOSVP(EC) Frequency Analysis"
    B97M_rV_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_EC__Unknown = "B97M-rV/def2-TZVPD/ISOSVP(EC) Unknown"
    B97M_rV_def2_TZVPD_ISOSVP_EC_EMC__Single_Point = "B97M-rV/def2-TZVPD/ISOSVP(EC/EMC) Single Point"
    B97M_rV_def2_TZVPD_ISOSVP_EC_EMC__Force = "B97M-rV/def2-TZVPD/ISOSVP(EC/EMC) Force"
    B97M_rV_def2_TZVPD_ISOSVP_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVPD/ISOSVP(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_EC_EMC__Unknown = "B97M-rV/def2-TZVPD/ISOSVP(EC/EMC) Unknown"
    B97M_rV_def2_TZVPD_ISOSVP_PC__Single_Point = "B97M-rV/def2-TZVPD/ISOSVP(PC) Single Point"
    B97M_rV_def2_TZVPD_ISOSVP_PC__Force = "B97M-rV/def2-TZVPD/ISOSVP(PC) Force"
    B97M_rV_def2_TZVPD_ISOSVP_PC__Geometry_Optimization = "B97M-rV/def2-TZVPD/ISOSVP(PC) Geometry Optimization"
    B97M_rV_def2_TZVPD_ISOSVP_PC__Frequency_Analysis = "B97M-rV/def2-TZVPD/ISOSVP(PC) Frequency Analysis"
    B97M_rV_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_ISOSVP_PC__Unknown = "B97M-rV/def2-TZVPD/ISOSVP(PC) Unknown"
    B97M_rV_def2_TZVPD_SMD_WATER__Single_Point = "B97M-rV/def2-TZVPD/SMD(WATER) Single Point"
    B97M_rV_def2_TZVPD_SMD_WATER__Force = "B97M-rV/def2-TZVPD/SMD(WATER) Force"
    B97M_rV_def2_TZVPD_SMD_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPD/SMD(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPD_SMD_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPD/SMD(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_WATER__Unknown = "B97M-rV/def2-TZVPD/SMD(WATER) Unknown"
    B97M_rV_def2_TZVPD_SMD_THF__Single_Point = "B97M-rV/def2-TZVPD/SMD(THF) Single Point"
    B97M_rV_def2_TZVPD_SMD_THF__Force = "B97M-rV/def2-TZVPD/SMD(THF) Force"
    B97M_rV_def2_TZVPD_SMD_THF__Geometry_Optimization = "B97M-rV/def2-TZVPD/SMD(THF) Geometry Optimization"
    B97M_rV_def2_TZVPD_SMD_THF__Frequency_Analysis = "B97M-rV/def2-TZVPD/SMD(THF) Frequency Analysis"
    B97M_rV_def2_TZVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_THF__Unknown = "B97M-rV/def2-TZVPD/SMD(THF) Unknown"
    B97M_rV_def2_TZVPD_SMD_G2__Single_Point = "B97M-rV/def2-TZVPD/SMD(G2) Single Point"
    B97M_rV_def2_TZVPD_SMD_G2__Force = "B97M-rV/def2-TZVPD/SMD(G2) Force"
    B97M_rV_def2_TZVPD_SMD_G2__Geometry_Optimization = "B97M-rV/def2-TZVPD/SMD(G2) Geometry Optimization"
    B97M_rV_def2_TZVPD_SMD_G2__Frequency_Analysis = "B97M-rV/def2-TZVPD/SMD(G2) Frequency Analysis"
    B97M_rV_def2_TZVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_G2__Unknown = "B97M-rV/def2-TZVPD/SMD(G2) Unknown"
    B97M_rV_def2_TZVPD_SMD_DME__Single_Point = "B97M-rV/def2-TZVPD/SMD(DME) Single Point"
    B97M_rV_def2_TZVPD_SMD_DME__Force = "B97M-rV/def2-TZVPD/SMD(DME) Force"
    B97M_rV_def2_TZVPD_SMD_DME__Geometry_Optimization = "B97M-rV/def2-TZVPD/SMD(DME) Geometry Optimization"
    B97M_rV_def2_TZVPD_SMD_DME__Frequency_Analysis = "B97M-rV/def2-TZVPD/SMD(DME) Frequency Analysis"
    B97M_rV_def2_TZVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_DME__Unknown = "B97M-rV/def2-TZVPD/SMD(DME) Unknown"
    B97M_rV_def2_TZVPD_SMD_EC__Single_Point = "B97M-rV/def2-TZVPD/SMD(EC) Single Point"
    B97M_rV_def2_TZVPD_SMD_EC__Force = "B97M-rV/def2-TZVPD/SMD(EC) Force"
    B97M_rV_def2_TZVPD_SMD_EC__Geometry_Optimization = "B97M-rV/def2-TZVPD/SMD(EC) Geometry Optimization"
    B97M_rV_def2_TZVPD_SMD_EC__Frequency_Analysis = "B97M-rV/def2-TZVPD/SMD(EC) Frequency Analysis"
    B97M_rV_def2_TZVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_EC__Unknown = "B97M-rV/def2-TZVPD/SMD(EC) Unknown"
    B97M_rV_def2_TZVPD_SMD_EC_EMC__Single_Point = "B97M-rV/def2-TZVPD/SMD(EC/EMC) Single Point"
    B97M_rV_def2_TZVPD_SMD_EC_EMC__Force = "B97M-rV/def2-TZVPD/SMD(EC/EMC) Force"
    B97M_rV_def2_TZVPD_SMD_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVPD/SMD(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVPD_SMD_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPD/SMD(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_EC_EMC__Unknown = "B97M-rV/def2-TZVPD/SMD(EC/EMC) Unknown"
    B97M_rV_def2_TZVPD_SMD_PC__Single_Point = "B97M-rV/def2-TZVPD/SMD(PC) Single Point"
    B97M_rV_def2_TZVPD_SMD_PC__Force = "B97M-rV/def2-TZVPD/SMD(PC) Force"
    B97M_rV_def2_TZVPD_SMD_PC__Geometry_Optimization = "B97M-rV/def2-TZVPD/SMD(PC) Geometry Optimization"
    B97M_rV_def2_TZVPD_SMD_PC__Frequency_Analysis = "B97M-rV/def2-TZVPD/SMD(PC) Frequency Analysis"
    B97M_rV_def2_TZVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_SMD_PC__Unknown = "B97M-rV/def2-TZVPD/SMD(PC) Unknown"
    B97M_rV_def2_TZVPD_CMIRS_WATER__Single_Point = "B97M-rV/def2-TZVPD/CMIRS(WATER) Single Point"
    B97M_rV_def2_TZVPD_CMIRS_WATER__Force = "B97M-rV/def2-TZVPD/CMIRS(WATER) Force"
    B97M_rV_def2_TZVPD_CMIRS_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPD/CMIRS(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPD_CMIRS_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPD/CMIRS(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_WATER__Unknown = "B97M-rV/def2-TZVPD/CMIRS(WATER) Unknown"
    B97M_rV_def2_TZVPD_CMIRS_THF__Single_Point = "B97M-rV/def2-TZVPD/CMIRS(THF) Single Point"
    B97M_rV_def2_TZVPD_CMIRS_THF__Force = "B97M-rV/def2-TZVPD/CMIRS(THF) Force"
    B97M_rV_def2_TZVPD_CMIRS_THF__Geometry_Optimization = "B97M-rV/def2-TZVPD/CMIRS(THF) Geometry Optimization"
    B97M_rV_def2_TZVPD_CMIRS_THF__Frequency_Analysis = "B97M-rV/def2-TZVPD/CMIRS(THF) Frequency Analysis"
    B97M_rV_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_THF__Unknown = "B97M-rV/def2-TZVPD/CMIRS(THF) Unknown"
    B97M_rV_def2_TZVPD_CMIRS_G2__Single_Point = "B97M-rV/def2-TZVPD/CMIRS(G2) Single Point"
    B97M_rV_def2_TZVPD_CMIRS_G2__Force = "B97M-rV/def2-TZVPD/CMIRS(G2) Force"
    B97M_rV_def2_TZVPD_CMIRS_G2__Geometry_Optimization = "B97M-rV/def2-TZVPD/CMIRS(G2) Geometry Optimization"
    B97M_rV_def2_TZVPD_CMIRS_G2__Frequency_Analysis = "B97M-rV/def2-TZVPD/CMIRS(G2) Frequency Analysis"
    B97M_rV_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_G2__Unknown = "B97M-rV/def2-TZVPD/CMIRS(G2) Unknown"
    B97M_rV_def2_TZVPD_CMIRS_DME__Single_Point = "B97M-rV/def2-TZVPD/CMIRS(DME) Single Point"
    B97M_rV_def2_TZVPD_CMIRS_DME__Force = "B97M-rV/def2-TZVPD/CMIRS(DME) Force"
    B97M_rV_def2_TZVPD_CMIRS_DME__Geometry_Optimization = "B97M-rV/def2-TZVPD/CMIRS(DME) Geometry Optimization"
    B97M_rV_def2_TZVPD_CMIRS_DME__Frequency_Analysis = "B97M-rV/def2-TZVPD/CMIRS(DME) Frequency Analysis"
    B97M_rV_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_DME__Unknown = "B97M-rV/def2-TZVPD/CMIRS(DME) Unknown"
    B97M_rV_def2_TZVPD_CMIRS_EC__Single_Point = "B97M-rV/def2-TZVPD/CMIRS(EC) Single Point"
    B97M_rV_def2_TZVPD_CMIRS_EC__Force = "B97M-rV/def2-TZVPD/CMIRS(EC) Force"
    B97M_rV_def2_TZVPD_CMIRS_EC__Geometry_Optimization = "B97M-rV/def2-TZVPD/CMIRS(EC) Geometry Optimization"
    B97M_rV_def2_TZVPD_CMIRS_EC__Frequency_Analysis = "B97M-rV/def2-TZVPD/CMIRS(EC) Frequency Analysis"
    B97M_rV_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_EC__Unknown = "B97M-rV/def2-TZVPD/CMIRS(EC) Unknown"
    B97M_rV_def2_TZVPD_CMIRS_EC_EMC__Single_Point = "B97M-rV/def2-TZVPD/CMIRS(EC/EMC) Single Point"
    B97M_rV_def2_TZVPD_CMIRS_EC_EMC__Force = "B97M-rV/def2-TZVPD/CMIRS(EC/EMC) Force"
    B97M_rV_def2_TZVPD_CMIRS_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVPD/CMIRS(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVPD_CMIRS_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPD/CMIRS(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_EC_EMC__Unknown = "B97M-rV/def2-TZVPD/CMIRS(EC/EMC) Unknown"
    B97M_rV_def2_TZVPD_CMIRS_PC__Single_Point = "B97M-rV/def2-TZVPD/CMIRS(PC) Single Point"
    B97M_rV_def2_TZVPD_CMIRS_PC__Force = "B97M-rV/def2-TZVPD/CMIRS(PC) Force"
    B97M_rV_def2_TZVPD_CMIRS_PC__Geometry_Optimization = "B97M-rV/def2-TZVPD/CMIRS(PC) Geometry Optimization"
    B97M_rV_def2_TZVPD_CMIRS_PC__Frequency_Analysis = "B97M-rV/def2-TZVPD/CMIRS(PC) Frequency Analysis"
    B97M_rV_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPD_CMIRS_PC__Unknown = "B97M-rV/def2-TZVPD/CMIRS(PC) Unknown"
    B97M_rV_def2_TZVPP_VACUUM_Single_Point = "B97M-rV/def2-TZVPP/VACUUM Single Point"
    B97M_rV_def2_TZVPP_VACUUM_Force = "B97M-rV/def2-TZVPP/VACUUM Force"
    B97M_rV_def2_TZVPP_VACUUM_Geometry_Optimization = "B97M-rV/def2-TZVPP/VACUUM Geometry Optimization"
    B97M_rV_def2_TZVPP_VACUUM_Frequency_Analysis = "B97M-rV/def2-TZVPP/VACUUM Frequency Analysis"
    B97M_rV_def2_TZVPP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_VACUUM_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/VACUUM Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_VACUUM_Unknown = "B97M-rV/def2-TZVPP/VACUUM Unknown"
    B97M_rV_def2_TZVPP_PCM_WATER__Single_Point = "B97M-rV/def2-TZVPP/PCM(WATER) Single Point"
    B97M_rV_def2_TZVPP_PCM_WATER__Force = "B97M-rV/def2-TZVPP/PCM(WATER) Force"
    B97M_rV_def2_TZVPP_PCM_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPP/PCM(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPP_PCM_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPP/PCM(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_WATER__Unknown = "B97M-rV/def2-TZVPP/PCM(WATER) Unknown"
    B97M_rV_def2_TZVPP_PCM_THF__Single_Point = "B97M-rV/def2-TZVPP/PCM(THF) Single Point"
    B97M_rV_def2_TZVPP_PCM_THF__Force = "B97M-rV/def2-TZVPP/PCM(THF) Force"
    B97M_rV_def2_TZVPP_PCM_THF__Geometry_Optimization = "B97M-rV/def2-TZVPP/PCM(THF) Geometry Optimization"
    B97M_rV_def2_TZVPP_PCM_THF__Frequency_Analysis = "B97M-rV/def2-TZVPP/PCM(THF) Frequency Analysis"
    B97M_rV_def2_TZVPP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_THF__Unknown = "B97M-rV/def2-TZVPP/PCM(THF) Unknown"
    B97M_rV_def2_TZVPP_PCM_G2__Single_Point = "B97M-rV/def2-TZVPP/PCM(G2) Single Point"
    B97M_rV_def2_TZVPP_PCM_G2__Force = "B97M-rV/def2-TZVPP/PCM(G2) Force"
    B97M_rV_def2_TZVPP_PCM_G2__Geometry_Optimization = "B97M-rV/def2-TZVPP/PCM(G2) Geometry Optimization"
    B97M_rV_def2_TZVPP_PCM_G2__Frequency_Analysis = "B97M-rV/def2-TZVPP/PCM(G2) Frequency Analysis"
    B97M_rV_def2_TZVPP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_G2__Unknown = "B97M-rV/def2-TZVPP/PCM(G2) Unknown"
    B97M_rV_def2_TZVPP_PCM_DME__Single_Point = "B97M-rV/def2-TZVPP/PCM(DME) Single Point"
    B97M_rV_def2_TZVPP_PCM_DME__Force = "B97M-rV/def2-TZVPP/PCM(DME) Force"
    B97M_rV_def2_TZVPP_PCM_DME__Geometry_Optimization = "B97M-rV/def2-TZVPP/PCM(DME) Geometry Optimization"
    B97M_rV_def2_TZVPP_PCM_DME__Frequency_Analysis = "B97M-rV/def2-TZVPP/PCM(DME) Frequency Analysis"
    B97M_rV_def2_TZVPP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_DME__Unknown = "B97M-rV/def2-TZVPP/PCM(DME) Unknown"
    B97M_rV_def2_TZVPP_PCM_EC__Single_Point = "B97M-rV/def2-TZVPP/PCM(EC) Single Point"
    B97M_rV_def2_TZVPP_PCM_EC__Force = "B97M-rV/def2-TZVPP/PCM(EC) Force"
    B97M_rV_def2_TZVPP_PCM_EC__Geometry_Optimization = "B97M-rV/def2-TZVPP/PCM(EC) Geometry Optimization"
    B97M_rV_def2_TZVPP_PCM_EC__Frequency_Analysis = "B97M-rV/def2-TZVPP/PCM(EC) Frequency Analysis"
    B97M_rV_def2_TZVPP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_EC__Unknown = "B97M-rV/def2-TZVPP/PCM(EC) Unknown"
    B97M_rV_def2_TZVPP_PCM_EC_EMC__Single_Point = "B97M-rV/def2-TZVPP/PCM(EC/EMC) Single Point"
    B97M_rV_def2_TZVPP_PCM_EC_EMC__Force = "B97M-rV/def2-TZVPP/PCM(EC/EMC) Force"
    B97M_rV_def2_TZVPP_PCM_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVPP/PCM(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVPP_PCM_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPP/PCM(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_EC_EMC__Unknown = "B97M-rV/def2-TZVPP/PCM(EC/EMC) Unknown"
    B97M_rV_def2_TZVPP_PCM_PC__Single_Point = "B97M-rV/def2-TZVPP/PCM(PC) Single Point"
    B97M_rV_def2_TZVPP_PCM_PC__Force = "B97M-rV/def2-TZVPP/PCM(PC) Force"
    B97M_rV_def2_TZVPP_PCM_PC__Geometry_Optimization = "B97M-rV/def2-TZVPP/PCM(PC) Geometry Optimization"
    B97M_rV_def2_TZVPP_PCM_PC__Frequency_Analysis = "B97M-rV/def2-TZVPP/PCM(PC) Frequency Analysis"
    B97M_rV_def2_TZVPP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_PCM_PC__Unknown = "B97M-rV/def2-TZVPP/PCM(PC) Unknown"
    B97M_rV_def2_TZVPP_ISOSVP_WATER__Single_Point = "B97M-rV/def2-TZVPP/ISOSVP(WATER) Single Point"
    B97M_rV_def2_TZVPP_ISOSVP_WATER__Force = "B97M-rV/def2-TZVPP/ISOSVP(WATER) Force"
    B97M_rV_def2_TZVPP_ISOSVP_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPP/ISOSVP(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPP_ISOSVP_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPP/ISOSVP(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_WATER__Unknown = "B97M-rV/def2-TZVPP/ISOSVP(WATER) Unknown"
    B97M_rV_def2_TZVPP_ISOSVP_THF__Single_Point = "B97M-rV/def2-TZVPP/ISOSVP(THF) Single Point"
    B97M_rV_def2_TZVPP_ISOSVP_THF__Force = "B97M-rV/def2-TZVPP/ISOSVP(THF) Force"
    B97M_rV_def2_TZVPP_ISOSVP_THF__Geometry_Optimization = "B97M-rV/def2-TZVPP/ISOSVP(THF) Geometry Optimization"
    B97M_rV_def2_TZVPP_ISOSVP_THF__Frequency_Analysis = "B97M-rV/def2-TZVPP/ISOSVP(THF) Frequency Analysis"
    B97M_rV_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_THF__Unknown = "B97M-rV/def2-TZVPP/ISOSVP(THF) Unknown"
    B97M_rV_def2_TZVPP_ISOSVP_G2__Single_Point = "B97M-rV/def2-TZVPP/ISOSVP(G2) Single Point"
    B97M_rV_def2_TZVPP_ISOSVP_G2__Force = "B97M-rV/def2-TZVPP/ISOSVP(G2) Force"
    B97M_rV_def2_TZVPP_ISOSVP_G2__Geometry_Optimization = "B97M-rV/def2-TZVPP/ISOSVP(G2) Geometry Optimization"
    B97M_rV_def2_TZVPP_ISOSVP_G2__Frequency_Analysis = "B97M-rV/def2-TZVPP/ISOSVP(G2) Frequency Analysis"
    B97M_rV_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_G2__Unknown = "B97M-rV/def2-TZVPP/ISOSVP(G2) Unknown"
    B97M_rV_def2_TZVPP_ISOSVP_DME__Single_Point = "B97M-rV/def2-TZVPP/ISOSVP(DME) Single Point"
    B97M_rV_def2_TZVPP_ISOSVP_DME__Force = "B97M-rV/def2-TZVPP/ISOSVP(DME) Force"
    B97M_rV_def2_TZVPP_ISOSVP_DME__Geometry_Optimization = "B97M-rV/def2-TZVPP/ISOSVP(DME) Geometry Optimization"
    B97M_rV_def2_TZVPP_ISOSVP_DME__Frequency_Analysis = "B97M-rV/def2-TZVPP/ISOSVP(DME) Frequency Analysis"
    B97M_rV_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_DME__Unknown = "B97M-rV/def2-TZVPP/ISOSVP(DME) Unknown"
    B97M_rV_def2_TZVPP_ISOSVP_EC__Single_Point = "B97M-rV/def2-TZVPP/ISOSVP(EC) Single Point"
    B97M_rV_def2_TZVPP_ISOSVP_EC__Force = "B97M-rV/def2-TZVPP/ISOSVP(EC) Force"
    B97M_rV_def2_TZVPP_ISOSVP_EC__Geometry_Optimization = "B97M-rV/def2-TZVPP/ISOSVP(EC) Geometry Optimization"
    B97M_rV_def2_TZVPP_ISOSVP_EC__Frequency_Analysis = "B97M-rV/def2-TZVPP/ISOSVP(EC) Frequency Analysis"
    B97M_rV_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_EC__Unknown = "B97M-rV/def2-TZVPP/ISOSVP(EC) Unknown"
    B97M_rV_def2_TZVPP_ISOSVP_EC_EMC__Single_Point = "B97M-rV/def2-TZVPP/ISOSVP(EC/EMC) Single Point"
    B97M_rV_def2_TZVPP_ISOSVP_EC_EMC__Force = "B97M-rV/def2-TZVPP/ISOSVP(EC/EMC) Force"
    B97M_rV_def2_TZVPP_ISOSVP_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVPP/ISOSVP(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPP/ISOSVP(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_EC_EMC__Unknown = "B97M-rV/def2-TZVPP/ISOSVP(EC/EMC) Unknown"
    B97M_rV_def2_TZVPP_ISOSVP_PC__Single_Point = "B97M-rV/def2-TZVPP/ISOSVP(PC) Single Point"
    B97M_rV_def2_TZVPP_ISOSVP_PC__Force = "B97M-rV/def2-TZVPP/ISOSVP(PC) Force"
    B97M_rV_def2_TZVPP_ISOSVP_PC__Geometry_Optimization = "B97M-rV/def2-TZVPP/ISOSVP(PC) Geometry Optimization"
    B97M_rV_def2_TZVPP_ISOSVP_PC__Frequency_Analysis = "B97M-rV/def2-TZVPP/ISOSVP(PC) Frequency Analysis"
    B97M_rV_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_ISOSVP_PC__Unknown = "B97M-rV/def2-TZVPP/ISOSVP(PC) Unknown"
    B97M_rV_def2_TZVPP_SMD_WATER__Single_Point = "B97M-rV/def2-TZVPP/SMD(WATER) Single Point"
    B97M_rV_def2_TZVPP_SMD_WATER__Force = "B97M-rV/def2-TZVPP/SMD(WATER) Force"
    B97M_rV_def2_TZVPP_SMD_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPP/SMD(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPP_SMD_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPP/SMD(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_WATER__Unknown = "B97M-rV/def2-TZVPP/SMD(WATER) Unknown"
    B97M_rV_def2_TZVPP_SMD_THF__Single_Point = "B97M-rV/def2-TZVPP/SMD(THF) Single Point"
    B97M_rV_def2_TZVPP_SMD_THF__Force = "B97M-rV/def2-TZVPP/SMD(THF) Force"
    B97M_rV_def2_TZVPP_SMD_THF__Geometry_Optimization = "B97M-rV/def2-TZVPP/SMD(THF) Geometry Optimization"
    B97M_rV_def2_TZVPP_SMD_THF__Frequency_Analysis = "B97M-rV/def2-TZVPP/SMD(THF) Frequency Analysis"
    B97M_rV_def2_TZVPP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_THF__Unknown = "B97M-rV/def2-TZVPP/SMD(THF) Unknown"
    B97M_rV_def2_TZVPP_SMD_G2__Single_Point = "B97M-rV/def2-TZVPP/SMD(G2) Single Point"
    B97M_rV_def2_TZVPP_SMD_G2__Force = "B97M-rV/def2-TZVPP/SMD(G2) Force"
    B97M_rV_def2_TZVPP_SMD_G2__Geometry_Optimization = "B97M-rV/def2-TZVPP/SMD(G2) Geometry Optimization"
    B97M_rV_def2_TZVPP_SMD_G2__Frequency_Analysis = "B97M-rV/def2-TZVPP/SMD(G2) Frequency Analysis"
    B97M_rV_def2_TZVPP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_G2__Unknown = "B97M-rV/def2-TZVPP/SMD(G2) Unknown"
    B97M_rV_def2_TZVPP_SMD_DME__Single_Point = "B97M-rV/def2-TZVPP/SMD(DME) Single Point"
    B97M_rV_def2_TZVPP_SMD_DME__Force = "B97M-rV/def2-TZVPP/SMD(DME) Force"
    B97M_rV_def2_TZVPP_SMD_DME__Geometry_Optimization = "B97M-rV/def2-TZVPP/SMD(DME) Geometry Optimization"
    B97M_rV_def2_TZVPP_SMD_DME__Frequency_Analysis = "B97M-rV/def2-TZVPP/SMD(DME) Frequency Analysis"
    B97M_rV_def2_TZVPP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_DME__Unknown = "B97M-rV/def2-TZVPP/SMD(DME) Unknown"
    B97M_rV_def2_TZVPP_SMD_EC__Single_Point = "B97M-rV/def2-TZVPP/SMD(EC) Single Point"
    B97M_rV_def2_TZVPP_SMD_EC__Force = "B97M-rV/def2-TZVPP/SMD(EC) Force"
    B97M_rV_def2_TZVPP_SMD_EC__Geometry_Optimization = "B97M-rV/def2-TZVPP/SMD(EC) Geometry Optimization"
    B97M_rV_def2_TZVPP_SMD_EC__Frequency_Analysis = "B97M-rV/def2-TZVPP/SMD(EC) Frequency Analysis"
    B97M_rV_def2_TZVPP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_EC__Unknown = "B97M-rV/def2-TZVPP/SMD(EC) Unknown"
    B97M_rV_def2_TZVPP_SMD_EC_EMC__Single_Point = "B97M-rV/def2-TZVPP/SMD(EC/EMC) Single Point"
    B97M_rV_def2_TZVPP_SMD_EC_EMC__Force = "B97M-rV/def2-TZVPP/SMD(EC/EMC) Force"
    B97M_rV_def2_TZVPP_SMD_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVPP/SMD(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVPP_SMD_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPP/SMD(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_EC_EMC__Unknown = "B97M-rV/def2-TZVPP/SMD(EC/EMC) Unknown"
    B97M_rV_def2_TZVPP_SMD_PC__Single_Point = "B97M-rV/def2-TZVPP/SMD(PC) Single Point"
    B97M_rV_def2_TZVPP_SMD_PC__Force = "B97M-rV/def2-TZVPP/SMD(PC) Force"
    B97M_rV_def2_TZVPP_SMD_PC__Geometry_Optimization = "B97M-rV/def2-TZVPP/SMD(PC) Geometry Optimization"
    B97M_rV_def2_TZVPP_SMD_PC__Frequency_Analysis = "B97M-rV/def2-TZVPP/SMD(PC) Frequency Analysis"
    B97M_rV_def2_TZVPP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_SMD_PC__Unknown = "B97M-rV/def2-TZVPP/SMD(PC) Unknown"
    B97M_rV_def2_TZVPP_CMIRS_WATER__Single_Point = "B97M-rV/def2-TZVPP/CMIRS(WATER) Single Point"
    B97M_rV_def2_TZVPP_CMIRS_WATER__Force = "B97M-rV/def2-TZVPP/CMIRS(WATER) Force"
    B97M_rV_def2_TZVPP_CMIRS_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPP/CMIRS(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPP_CMIRS_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPP/CMIRS(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_WATER__Unknown = "B97M-rV/def2-TZVPP/CMIRS(WATER) Unknown"
    B97M_rV_def2_TZVPP_CMIRS_THF__Single_Point = "B97M-rV/def2-TZVPP/CMIRS(THF) Single Point"
    B97M_rV_def2_TZVPP_CMIRS_THF__Force = "B97M-rV/def2-TZVPP/CMIRS(THF) Force"
    B97M_rV_def2_TZVPP_CMIRS_THF__Geometry_Optimization = "B97M-rV/def2-TZVPP/CMIRS(THF) Geometry Optimization"
    B97M_rV_def2_TZVPP_CMIRS_THF__Frequency_Analysis = "B97M-rV/def2-TZVPP/CMIRS(THF) Frequency Analysis"
    B97M_rV_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_THF__Unknown = "B97M-rV/def2-TZVPP/CMIRS(THF) Unknown"
    B97M_rV_def2_TZVPP_CMIRS_G2__Single_Point = "B97M-rV/def2-TZVPP/CMIRS(G2) Single Point"
    B97M_rV_def2_TZVPP_CMIRS_G2__Force = "B97M-rV/def2-TZVPP/CMIRS(G2) Force"
    B97M_rV_def2_TZVPP_CMIRS_G2__Geometry_Optimization = "B97M-rV/def2-TZVPP/CMIRS(G2) Geometry Optimization"
    B97M_rV_def2_TZVPP_CMIRS_G2__Frequency_Analysis = "B97M-rV/def2-TZVPP/CMIRS(G2) Frequency Analysis"
    B97M_rV_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_G2__Unknown = "B97M-rV/def2-TZVPP/CMIRS(G2) Unknown"
    B97M_rV_def2_TZVPP_CMIRS_DME__Single_Point = "B97M-rV/def2-TZVPP/CMIRS(DME) Single Point"
    B97M_rV_def2_TZVPP_CMIRS_DME__Force = "B97M-rV/def2-TZVPP/CMIRS(DME) Force"
    B97M_rV_def2_TZVPP_CMIRS_DME__Geometry_Optimization = "B97M-rV/def2-TZVPP/CMIRS(DME) Geometry Optimization"
    B97M_rV_def2_TZVPP_CMIRS_DME__Frequency_Analysis = "B97M-rV/def2-TZVPP/CMIRS(DME) Frequency Analysis"
    B97M_rV_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_DME__Unknown = "B97M-rV/def2-TZVPP/CMIRS(DME) Unknown"
    B97M_rV_def2_TZVPP_CMIRS_EC__Single_Point = "B97M-rV/def2-TZVPP/CMIRS(EC) Single Point"
    B97M_rV_def2_TZVPP_CMIRS_EC__Force = "B97M-rV/def2-TZVPP/CMIRS(EC) Force"
    B97M_rV_def2_TZVPP_CMIRS_EC__Geometry_Optimization = "B97M-rV/def2-TZVPP/CMIRS(EC) Geometry Optimization"
    B97M_rV_def2_TZVPP_CMIRS_EC__Frequency_Analysis = "B97M-rV/def2-TZVPP/CMIRS(EC) Frequency Analysis"
    B97M_rV_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_EC__Unknown = "B97M-rV/def2-TZVPP/CMIRS(EC) Unknown"
    B97M_rV_def2_TZVPP_CMIRS_EC_EMC__Single_Point = "B97M-rV/def2-TZVPP/CMIRS(EC/EMC) Single Point"
    B97M_rV_def2_TZVPP_CMIRS_EC_EMC__Force = "B97M-rV/def2-TZVPP/CMIRS(EC/EMC) Force"
    B97M_rV_def2_TZVPP_CMIRS_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVPP/CMIRS(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVPP_CMIRS_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPP/CMIRS(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_EC_EMC__Unknown = "B97M-rV/def2-TZVPP/CMIRS(EC/EMC) Unknown"
    B97M_rV_def2_TZVPP_CMIRS_PC__Single_Point = "B97M-rV/def2-TZVPP/CMIRS(PC) Single Point"
    B97M_rV_def2_TZVPP_CMIRS_PC__Force = "B97M-rV/def2-TZVPP/CMIRS(PC) Force"
    B97M_rV_def2_TZVPP_CMIRS_PC__Geometry_Optimization = "B97M-rV/def2-TZVPP/CMIRS(PC) Geometry Optimization"
    B97M_rV_def2_TZVPP_CMIRS_PC__Frequency_Analysis = "B97M-rV/def2-TZVPP/CMIRS(PC) Frequency Analysis"
    B97M_rV_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPP_CMIRS_PC__Unknown = "B97M-rV/def2-TZVPP/CMIRS(PC) Unknown"
    B97M_rV_def2_TZVPPD_VACUUM_Single_Point = "B97M-rV/def2-TZVPPD/VACUUM Single Point"
    B97M_rV_def2_TZVPPD_VACUUM_Force = "B97M-rV/def2-TZVPPD/VACUUM Force"
    B97M_rV_def2_TZVPPD_VACUUM_Geometry_Optimization = "B97M-rV/def2-TZVPPD/VACUUM Geometry Optimization"
    B97M_rV_def2_TZVPPD_VACUUM_Frequency_Analysis = "B97M-rV/def2-TZVPPD/VACUUM Frequency Analysis"
    B97M_rV_def2_TZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/VACUUM Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_VACUUM_Unknown = "B97M-rV/def2-TZVPPD/VACUUM Unknown"
    B97M_rV_def2_TZVPPD_PCM_WATER__Single_Point = "B97M-rV/def2-TZVPPD/PCM(WATER) Single Point"
    B97M_rV_def2_TZVPPD_PCM_WATER__Force = "B97M-rV/def2-TZVPPD/PCM(WATER) Force"
    B97M_rV_def2_TZVPPD_PCM_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPPD/PCM(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPPD_PCM_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPPD/PCM(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_WATER__Unknown = "B97M-rV/def2-TZVPPD/PCM(WATER) Unknown"
    B97M_rV_def2_TZVPPD_PCM_THF__Single_Point = "B97M-rV/def2-TZVPPD/PCM(THF) Single Point"
    B97M_rV_def2_TZVPPD_PCM_THF__Force = "B97M-rV/def2-TZVPPD/PCM(THF) Force"
    B97M_rV_def2_TZVPPD_PCM_THF__Geometry_Optimization = "B97M-rV/def2-TZVPPD/PCM(THF) Geometry Optimization"
    B97M_rV_def2_TZVPPD_PCM_THF__Frequency_Analysis = "B97M-rV/def2-TZVPPD/PCM(THF) Frequency Analysis"
    B97M_rV_def2_TZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_THF__Unknown = "B97M-rV/def2-TZVPPD/PCM(THF) Unknown"
    B97M_rV_def2_TZVPPD_PCM_G2__Single_Point = "B97M-rV/def2-TZVPPD/PCM(G2) Single Point"
    B97M_rV_def2_TZVPPD_PCM_G2__Force = "B97M-rV/def2-TZVPPD/PCM(G2) Force"
    B97M_rV_def2_TZVPPD_PCM_G2__Geometry_Optimization = "B97M-rV/def2-TZVPPD/PCM(G2) Geometry Optimization"
    B97M_rV_def2_TZVPPD_PCM_G2__Frequency_Analysis = "B97M-rV/def2-TZVPPD/PCM(G2) Frequency Analysis"
    B97M_rV_def2_TZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_G2__Unknown = "B97M-rV/def2-TZVPPD/PCM(G2) Unknown"
    B97M_rV_def2_TZVPPD_PCM_DME__Single_Point = "B97M-rV/def2-TZVPPD/PCM(DME) Single Point"
    B97M_rV_def2_TZVPPD_PCM_DME__Force = "B97M-rV/def2-TZVPPD/PCM(DME) Force"
    B97M_rV_def2_TZVPPD_PCM_DME__Geometry_Optimization = "B97M-rV/def2-TZVPPD/PCM(DME) Geometry Optimization"
    B97M_rV_def2_TZVPPD_PCM_DME__Frequency_Analysis = "B97M-rV/def2-TZVPPD/PCM(DME) Frequency Analysis"
    B97M_rV_def2_TZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_DME__Unknown = "B97M-rV/def2-TZVPPD/PCM(DME) Unknown"
    B97M_rV_def2_TZVPPD_PCM_EC__Single_Point = "B97M-rV/def2-TZVPPD/PCM(EC) Single Point"
    B97M_rV_def2_TZVPPD_PCM_EC__Force = "B97M-rV/def2-TZVPPD/PCM(EC) Force"
    B97M_rV_def2_TZVPPD_PCM_EC__Geometry_Optimization = "B97M-rV/def2-TZVPPD/PCM(EC) Geometry Optimization"
    B97M_rV_def2_TZVPPD_PCM_EC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/PCM(EC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_EC__Unknown = "B97M-rV/def2-TZVPPD/PCM(EC) Unknown"
    B97M_rV_def2_TZVPPD_PCM_EC_EMC__Single_Point = "B97M-rV/def2-TZVPPD/PCM(EC/EMC) Single Point"
    B97M_rV_def2_TZVPPD_PCM_EC_EMC__Force = "B97M-rV/def2-TZVPPD/PCM(EC/EMC) Force"
    B97M_rV_def2_TZVPPD_PCM_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVPPD/PCM(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVPPD_PCM_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/PCM(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_EC_EMC__Unknown = "B97M-rV/def2-TZVPPD/PCM(EC/EMC) Unknown"
    B97M_rV_def2_TZVPPD_PCM_PC__Single_Point = "B97M-rV/def2-TZVPPD/PCM(PC) Single Point"
    B97M_rV_def2_TZVPPD_PCM_PC__Force = "B97M-rV/def2-TZVPPD/PCM(PC) Force"
    B97M_rV_def2_TZVPPD_PCM_PC__Geometry_Optimization = "B97M-rV/def2-TZVPPD/PCM(PC) Geometry Optimization"
    B97M_rV_def2_TZVPPD_PCM_PC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/PCM(PC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_PCM_PC__Unknown = "B97M-rV/def2-TZVPPD/PCM(PC) Unknown"
    B97M_rV_def2_TZVPPD_ISOSVP_WATER__Single_Point = "B97M-rV/def2-TZVPPD/ISOSVP(WATER) Single Point"
    B97M_rV_def2_TZVPPD_ISOSVP_WATER__Force = "B97M-rV/def2-TZVPPD/ISOSVP(WATER) Force"
    B97M_rV_def2_TZVPPD_ISOSVP_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPPD/ISOSVP(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPPD_ISOSVP_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPPD/ISOSVP(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_WATER__Unknown = "B97M-rV/def2-TZVPPD/ISOSVP(WATER) Unknown"
    B97M_rV_def2_TZVPPD_ISOSVP_THF__Single_Point = "B97M-rV/def2-TZVPPD/ISOSVP(THF) Single Point"
    B97M_rV_def2_TZVPPD_ISOSVP_THF__Force = "B97M-rV/def2-TZVPPD/ISOSVP(THF) Force"
    B97M_rV_def2_TZVPPD_ISOSVP_THF__Geometry_Optimization = "B97M-rV/def2-TZVPPD/ISOSVP(THF) Geometry Optimization"
    B97M_rV_def2_TZVPPD_ISOSVP_THF__Frequency_Analysis = "B97M-rV/def2-TZVPPD/ISOSVP(THF) Frequency Analysis"
    B97M_rV_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_THF__Unknown = "B97M-rV/def2-TZVPPD/ISOSVP(THF) Unknown"
    B97M_rV_def2_TZVPPD_ISOSVP_G2__Single_Point = "B97M-rV/def2-TZVPPD/ISOSVP(G2) Single Point"
    B97M_rV_def2_TZVPPD_ISOSVP_G2__Force = "B97M-rV/def2-TZVPPD/ISOSVP(G2) Force"
    B97M_rV_def2_TZVPPD_ISOSVP_G2__Geometry_Optimization = "B97M-rV/def2-TZVPPD/ISOSVP(G2) Geometry Optimization"
    B97M_rV_def2_TZVPPD_ISOSVP_G2__Frequency_Analysis = "B97M-rV/def2-TZVPPD/ISOSVP(G2) Frequency Analysis"
    B97M_rV_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_G2__Unknown = "B97M-rV/def2-TZVPPD/ISOSVP(G2) Unknown"
    B97M_rV_def2_TZVPPD_ISOSVP_DME__Single_Point = "B97M-rV/def2-TZVPPD/ISOSVP(DME) Single Point"
    B97M_rV_def2_TZVPPD_ISOSVP_DME__Force = "B97M-rV/def2-TZVPPD/ISOSVP(DME) Force"
    B97M_rV_def2_TZVPPD_ISOSVP_DME__Geometry_Optimization = "B97M-rV/def2-TZVPPD/ISOSVP(DME) Geometry Optimization"
    B97M_rV_def2_TZVPPD_ISOSVP_DME__Frequency_Analysis = "B97M-rV/def2-TZVPPD/ISOSVP(DME) Frequency Analysis"
    B97M_rV_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_DME__Unknown = "B97M-rV/def2-TZVPPD/ISOSVP(DME) Unknown"
    B97M_rV_def2_TZVPPD_ISOSVP_EC__Single_Point = "B97M-rV/def2-TZVPPD/ISOSVP(EC) Single Point"
    B97M_rV_def2_TZVPPD_ISOSVP_EC__Force = "B97M-rV/def2-TZVPPD/ISOSVP(EC) Force"
    B97M_rV_def2_TZVPPD_ISOSVP_EC__Geometry_Optimization = "B97M-rV/def2-TZVPPD/ISOSVP(EC) Geometry Optimization"
    B97M_rV_def2_TZVPPD_ISOSVP_EC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/ISOSVP(EC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_EC__Unknown = "B97M-rV/def2-TZVPPD/ISOSVP(EC) Unknown"
    B97M_rV_def2_TZVPPD_ISOSVP_EC_EMC__Single_Point = "B97M-rV/def2-TZVPPD/ISOSVP(EC/EMC) Single Point"
    B97M_rV_def2_TZVPPD_ISOSVP_EC_EMC__Force = "B97M-rV/def2-TZVPPD/ISOSVP(EC/EMC) Force"
    B97M_rV_def2_TZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_EC_EMC__Unknown = "B97M-rV/def2-TZVPPD/ISOSVP(EC/EMC) Unknown"
    B97M_rV_def2_TZVPPD_ISOSVP_PC__Single_Point = "B97M-rV/def2-TZVPPD/ISOSVP(PC) Single Point"
    B97M_rV_def2_TZVPPD_ISOSVP_PC__Force = "B97M-rV/def2-TZVPPD/ISOSVP(PC) Force"
    B97M_rV_def2_TZVPPD_ISOSVP_PC__Geometry_Optimization = "B97M-rV/def2-TZVPPD/ISOSVP(PC) Geometry Optimization"
    B97M_rV_def2_TZVPPD_ISOSVP_PC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/ISOSVP(PC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_ISOSVP_PC__Unknown = "B97M-rV/def2-TZVPPD/ISOSVP(PC) Unknown"
    B97M_rV_def2_TZVPPD_SMD_WATER__Single_Point = "B97M-rV/def2-TZVPPD/SMD(WATER) Single Point"
    B97M_rV_def2_TZVPPD_SMD_WATER__Force = "B97M-rV/def2-TZVPPD/SMD(WATER) Force"
    B97M_rV_def2_TZVPPD_SMD_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPPD/SMD(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPPD_SMD_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPPD/SMD(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_WATER__Unknown = "B97M-rV/def2-TZVPPD/SMD(WATER) Unknown"
    B97M_rV_def2_TZVPPD_SMD_THF__Single_Point = "B97M-rV/def2-TZVPPD/SMD(THF) Single Point"
    B97M_rV_def2_TZVPPD_SMD_THF__Force = "B97M-rV/def2-TZVPPD/SMD(THF) Force"
    B97M_rV_def2_TZVPPD_SMD_THF__Geometry_Optimization = "B97M-rV/def2-TZVPPD/SMD(THF) Geometry Optimization"
    B97M_rV_def2_TZVPPD_SMD_THF__Frequency_Analysis = "B97M-rV/def2-TZVPPD/SMD(THF) Frequency Analysis"
    B97M_rV_def2_TZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_THF__Unknown = "B97M-rV/def2-TZVPPD/SMD(THF) Unknown"
    B97M_rV_def2_TZVPPD_SMD_G2__Single_Point = "B97M-rV/def2-TZVPPD/SMD(G2) Single Point"
    B97M_rV_def2_TZVPPD_SMD_G2__Force = "B97M-rV/def2-TZVPPD/SMD(G2) Force"
    B97M_rV_def2_TZVPPD_SMD_G2__Geometry_Optimization = "B97M-rV/def2-TZVPPD/SMD(G2) Geometry Optimization"
    B97M_rV_def2_TZVPPD_SMD_G2__Frequency_Analysis = "B97M-rV/def2-TZVPPD/SMD(G2) Frequency Analysis"
    B97M_rV_def2_TZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_G2__Unknown = "B97M-rV/def2-TZVPPD/SMD(G2) Unknown"
    B97M_rV_def2_TZVPPD_SMD_DME__Single_Point = "B97M-rV/def2-TZVPPD/SMD(DME) Single Point"
    B97M_rV_def2_TZVPPD_SMD_DME__Force = "B97M-rV/def2-TZVPPD/SMD(DME) Force"
    B97M_rV_def2_TZVPPD_SMD_DME__Geometry_Optimization = "B97M-rV/def2-TZVPPD/SMD(DME) Geometry Optimization"
    B97M_rV_def2_TZVPPD_SMD_DME__Frequency_Analysis = "B97M-rV/def2-TZVPPD/SMD(DME) Frequency Analysis"
    B97M_rV_def2_TZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_DME__Unknown = "B97M-rV/def2-TZVPPD/SMD(DME) Unknown"
    B97M_rV_def2_TZVPPD_SMD_EC__Single_Point = "B97M-rV/def2-TZVPPD/SMD(EC) Single Point"
    B97M_rV_def2_TZVPPD_SMD_EC__Force = "B97M-rV/def2-TZVPPD/SMD(EC) Force"
    B97M_rV_def2_TZVPPD_SMD_EC__Geometry_Optimization = "B97M-rV/def2-TZVPPD/SMD(EC) Geometry Optimization"
    B97M_rV_def2_TZVPPD_SMD_EC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/SMD(EC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_EC__Unknown = "B97M-rV/def2-TZVPPD/SMD(EC) Unknown"
    B97M_rV_def2_TZVPPD_SMD_EC_EMC__Single_Point = "B97M-rV/def2-TZVPPD/SMD(EC/EMC) Single Point"
    B97M_rV_def2_TZVPPD_SMD_EC_EMC__Force = "B97M-rV/def2-TZVPPD/SMD(EC/EMC) Force"
    B97M_rV_def2_TZVPPD_SMD_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVPPD/SMD(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVPPD_SMD_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/SMD(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_EC_EMC__Unknown = "B97M-rV/def2-TZVPPD/SMD(EC/EMC) Unknown"
    B97M_rV_def2_TZVPPD_SMD_PC__Single_Point = "B97M-rV/def2-TZVPPD/SMD(PC) Single Point"
    B97M_rV_def2_TZVPPD_SMD_PC__Force = "B97M-rV/def2-TZVPPD/SMD(PC) Force"
    B97M_rV_def2_TZVPPD_SMD_PC__Geometry_Optimization = "B97M-rV/def2-TZVPPD/SMD(PC) Geometry Optimization"
    B97M_rV_def2_TZVPPD_SMD_PC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/SMD(PC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_SMD_PC__Unknown = "B97M-rV/def2-TZVPPD/SMD(PC) Unknown"
    B97M_rV_def2_TZVPPD_CMIRS_WATER__Single_Point = "B97M-rV/def2-TZVPPD/CMIRS(WATER) Single Point"
    B97M_rV_def2_TZVPPD_CMIRS_WATER__Force = "B97M-rV/def2-TZVPPD/CMIRS(WATER) Force"
    B97M_rV_def2_TZVPPD_CMIRS_WATER__Geometry_Optimization = "B97M-rV/def2-TZVPPD/CMIRS(WATER) Geometry Optimization"
    B97M_rV_def2_TZVPPD_CMIRS_WATER__Frequency_Analysis = "B97M-rV/def2-TZVPPD/CMIRS(WATER) Frequency Analysis"
    B97M_rV_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_WATER__Unknown = "B97M-rV/def2-TZVPPD/CMIRS(WATER) Unknown"
    B97M_rV_def2_TZVPPD_CMIRS_THF__Single_Point = "B97M-rV/def2-TZVPPD/CMIRS(THF) Single Point"
    B97M_rV_def2_TZVPPD_CMIRS_THF__Force = "B97M-rV/def2-TZVPPD/CMIRS(THF) Force"
    B97M_rV_def2_TZVPPD_CMIRS_THF__Geometry_Optimization = "B97M-rV/def2-TZVPPD/CMIRS(THF) Geometry Optimization"
    B97M_rV_def2_TZVPPD_CMIRS_THF__Frequency_Analysis = "B97M-rV/def2-TZVPPD/CMIRS(THF) Frequency Analysis"
    B97M_rV_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_THF__Unknown = "B97M-rV/def2-TZVPPD/CMIRS(THF) Unknown"
    B97M_rV_def2_TZVPPD_CMIRS_G2__Single_Point = "B97M-rV/def2-TZVPPD/CMIRS(G2) Single Point"
    B97M_rV_def2_TZVPPD_CMIRS_G2__Force = "B97M-rV/def2-TZVPPD/CMIRS(G2) Force"
    B97M_rV_def2_TZVPPD_CMIRS_G2__Geometry_Optimization = "B97M-rV/def2-TZVPPD/CMIRS(G2) Geometry Optimization"
    B97M_rV_def2_TZVPPD_CMIRS_G2__Frequency_Analysis = "B97M-rV/def2-TZVPPD/CMIRS(G2) Frequency Analysis"
    B97M_rV_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_G2__Unknown = "B97M-rV/def2-TZVPPD/CMIRS(G2) Unknown"
    B97M_rV_def2_TZVPPD_CMIRS_DME__Single_Point = "B97M-rV/def2-TZVPPD/CMIRS(DME) Single Point"
    B97M_rV_def2_TZVPPD_CMIRS_DME__Force = "B97M-rV/def2-TZVPPD/CMIRS(DME) Force"
    B97M_rV_def2_TZVPPD_CMIRS_DME__Geometry_Optimization = "B97M-rV/def2-TZVPPD/CMIRS(DME) Geometry Optimization"
    B97M_rV_def2_TZVPPD_CMIRS_DME__Frequency_Analysis = "B97M-rV/def2-TZVPPD/CMIRS(DME) Frequency Analysis"
    B97M_rV_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_DME__Unknown = "B97M-rV/def2-TZVPPD/CMIRS(DME) Unknown"
    B97M_rV_def2_TZVPPD_CMIRS_EC__Single_Point = "B97M-rV/def2-TZVPPD/CMIRS(EC) Single Point"
    B97M_rV_def2_TZVPPD_CMIRS_EC__Force = "B97M-rV/def2-TZVPPD/CMIRS(EC) Force"
    B97M_rV_def2_TZVPPD_CMIRS_EC__Geometry_Optimization = "B97M-rV/def2-TZVPPD/CMIRS(EC) Geometry Optimization"
    B97M_rV_def2_TZVPPD_CMIRS_EC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/CMIRS(EC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_EC__Unknown = "B97M-rV/def2-TZVPPD/CMIRS(EC) Unknown"
    B97M_rV_def2_TZVPPD_CMIRS_EC_EMC__Single_Point = "B97M-rV/def2-TZVPPD/CMIRS(EC/EMC) Single Point"
    B97M_rV_def2_TZVPPD_CMIRS_EC_EMC__Force = "B97M-rV/def2-TZVPPD/CMIRS(EC/EMC) Force"
    B97M_rV_def2_TZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "B97M-rV/def2-TZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    B97M_rV_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_EC_EMC__Unknown = "B97M-rV/def2-TZVPPD/CMIRS(EC/EMC) Unknown"
    B97M_rV_def2_TZVPPD_CMIRS_PC__Single_Point = "B97M-rV/def2-TZVPPD/CMIRS(PC) Single Point"
    B97M_rV_def2_TZVPPD_CMIRS_PC__Force = "B97M-rV/def2-TZVPPD/CMIRS(PC) Force"
    B97M_rV_def2_TZVPPD_CMIRS_PC__Geometry_Optimization = "B97M-rV/def2-TZVPPD/CMIRS(PC) Geometry Optimization"
    B97M_rV_def2_TZVPPD_CMIRS_PC__Frequency_Analysis = "B97M-rV/def2-TZVPPD/CMIRS(PC) Frequency Analysis"
    B97M_rV_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-TZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_TZVPPD_CMIRS_PC__Unknown = "B97M-rV/def2-TZVPPD/CMIRS(PC) Unknown"
    B97M_rV_def2_QZVPPD_VACUUM_Single_Point = "B97M-rV/def2-QZVPPD/VACUUM Single Point"
    B97M_rV_def2_QZVPPD_VACUUM_Force = "B97M-rV/def2-QZVPPD/VACUUM Force"
    B97M_rV_def2_QZVPPD_VACUUM_Geometry_Optimization = "B97M-rV/def2-QZVPPD/VACUUM Geometry Optimization"
    B97M_rV_def2_QZVPPD_VACUUM_Frequency_Analysis = "B97M-rV/def2-QZVPPD/VACUUM Frequency Analysis"
    B97M_rV_def2_QZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/VACUUM Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_VACUUM_Unknown = "B97M-rV/def2-QZVPPD/VACUUM Unknown"
    B97M_rV_def2_QZVPPD_PCM_WATER__Single_Point = "B97M-rV/def2-QZVPPD/PCM(WATER) Single Point"
    B97M_rV_def2_QZVPPD_PCM_WATER__Force = "B97M-rV/def2-QZVPPD/PCM(WATER) Force"
    B97M_rV_def2_QZVPPD_PCM_WATER__Geometry_Optimization = "B97M-rV/def2-QZVPPD/PCM(WATER) Geometry Optimization"
    B97M_rV_def2_QZVPPD_PCM_WATER__Frequency_Analysis = "B97M-rV/def2-QZVPPD/PCM(WATER) Frequency Analysis"
    B97M_rV_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_WATER__Unknown = "B97M-rV/def2-QZVPPD/PCM(WATER) Unknown"
    B97M_rV_def2_QZVPPD_PCM_THF__Single_Point = "B97M-rV/def2-QZVPPD/PCM(THF) Single Point"
    B97M_rV_def2_QZVPPD_PCM_THF__Force = "B97M-rV/def2-QZVPPD/PCM(THF) Force"
    B97M_rV_def2_QZVPPD_PCM_THF__Geometry_Optimization = "B97M-rV/def2-QZVPPD/PCM(THF) Geometry Optimization"
    B97M_rV_def2_QZVPPD_PCM_THF__Frequency_Analysis = "B97M-rV/def2-QZVPPD/PCM(THF) Frequency Analysis"
    B97M_rV_def2_QZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_THF__Unknown = "B97M-rV/def2-QZVPPD/PCM(THF) Unknown"
    B97M_rV_def2_QZVPPD_PCM_G2__Single_Point = "B97M-rV/def2-QZVPPD/PCM(G2) Single Point"
    B97M_rV_def2_QZVPPD_PCM_G2__Force = "B97M-rV/def2-QZVPPD/PCM(G2) Force"
    B97M_rV_def2_QZVPPD_PCM_G2__Geometry_Optimization = "B97M-rV/def2-QZVPPD/PCM(G2) Geometry Optimization"
    B97M_rV_def2_QZVPPD_PCM_G2__Frequency_Analysis = "B97M-rV/def2-QZVPPD/PCM(G2) Frequency Analysis"
    B97M_rV_def2_QZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_G2__Unknown = "B97M-rV/def2-QZVPPD/PCM(G2) Unknown"
    B97M_rV_def2_QZVPPD_PCM_DME__Single_Point = "B97M-rV/def2-QZVPPD/PCM(DME) Single Point"
    B97M_rV_def2_QZVPPD_PCM_DME__Force = "B97M-rV/def2-QZVPPD/PCM(DME) Force"
    B97M_rV_def2_QZVPPD_PCM_DME__Geometry_Optimization = "B97M-rV/def2-QZVPPD/PCM(DME) Geometry Optimization"
    B97M_rV_def2_QZVPPD_PCM_DME__Frequency_Analysis = "B97M-rV/def2-QZVPPD/PCM(DME) Frequency Analysis"
    B97M_rV_def2_QZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_DME__Unknown = "B97M-rV/def2-QZVPPD/PCM(DME) Unknown"
    B97M_rV_def2_QZVPPD_PCM_EC__Single_Point = "B97M-rV/def2-QZVPPD/PCM(EC) Single Point"
    B97M_rV_def2_QZVPPD_PCM_EC__Force = "B97M-rV/def2-QZVPPD/PCM(EC) Force"
    B97M_rV_def2_QZVPPD_PCM_EC__Geometry_Optimization = "B97M-rV/def2-QZVPPD/PCM(EC) Geometry Optimization"
    B97M_rV_def2_QZVPPD_PCM_EC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/PCM(EC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_EC__Unknown = "B97M-rV/def2-QZVPPD/PCM(EC) Unknown"
    B97M_rV_def2_QZVPPD_PCM_EC_EMC__Single_Point = "B97M-rV/def2-QZVPPD/PCM(EC/EMC) Single Point"
    B97M_rV_def2_QZVPPD_PCM_EC_EMC__Force = "B97M-rV/def2-QZVPPD/PCM(EC/EMC) Force"
    B97M_rV_def2_QZVPPD_PCM_EC_EMC__Geometry_Optimization = "B97M-rV/def2-QZVPPD/PCM(EC/EMC) Geometry Optimization"
    B97M_rV_def2_QZVPPD_PCM_EC_EMC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/PCM(EC/EMC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_EC_EMC__Unknown = "B97M-rV/def2-QZVPPD/PCM(EC/EMC) Unknown"
    B97M_rV_def2_QZVPPD_PCM_PC__Single_Point = "B97M-rV/def2-QZVPPD/PCM(PC) Single Point"
    B97M_rV_def2_QZVPPD_PCM_PC__Force = "B97M-rV/def2-QZVPPD/PCM(PC) Force"
    B97M_rV_def2_QZVPPD_PCM_PC__Geometry_Optimization = "B97M-rV/def2-QZVPPD/PCM(PC) Geometry Optimization"
    B97M_rV_def2_QZVPPD_PCM_PC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/PCM(PC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_PCM_PC__Unknown = "B97M-rV/def2-QZVPPD/PCM(PC) Unknown"
    B97M_rV_def2_QZVPPD_ISOSVP_WATER__Single_Point = "B97M-rV/def2-QZVPPD/ISOSVP(WATER) Single Point"
    B97M_rV_def2_QZVPPD_ISOSVP_WATER__Force = "B97M-rV/def2-QZVPPD/ISOSVP(WATER) Force"
    B97M_rV_def2_QZVPPD_ISOSVP_WATER__Geometry_Optimization = "B97M-rV/def2-QZVPPD/ISOSVP(WATER) Geometry Optimization"
    B97M_rV_def2_QZVPPD_ISOSVP_WATER__Frequency_Analysis = "B97M-rV/def2-QZVPPD/ISOSVP(WATER) Frequency Analysis"
    B97M_rV_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_WATER__Unknown = "B97M-rV/def2-QZVPPD/ISOSVP(WATER) Unknown"
    B97M_rV_def2_QZVPPD_ISOSVP_THF__Single_Point = "B97M-rV/def2-QZVPPD/ISOSVP(THF) Single Point"
    B97M_rV_def2_QZVPPD_ISOSVP_THF__Force = "B97M-rV/def2-QZVPPD/ISOSVP(THF) Force"
    B97M_rV_def2_QZVPPD_ISOSVP_THF__Geometry_Optimization = "B97M-rV/def2-QZVPPD/ISOSVP(THF) Geometry Optimization"
    B97M_rV_def2_QZVPPD_ISOSVP_THF__Frequency_Analysis = "B97M-rV/def2-QZVPPD/ISOSVP(THF) Frequency Analysis"
    B97M_rV_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_THF__Unknown = "B97M-rV/def2-QZVPPD/ISOSVP(THF) Unknown"
    B97M_rV_def2_QZVPPD_ISOSVP_G2__Single_Point = "B97M-rV/def2-QZVPPD/ISOSVP(G2) Single Point"
    B97M_rV_def2_QZVPPD_ISOSVP_G2__Force = "B97M-rV/def2-QZVPPD/ISOSVP(G2) Force"
    B97M_rV_def2_QZVPPD_ISOSVP_G2__Geometry_Optimization = "B97M-rV/def2-QZVPPD/ISOSVP(G2) Geometry Optimization"
    B97M_rV_def2_QZVPPD_ISOSVP_G2__Frequency_Analysis = "B97M-rV/def2-QZVPPD/ISOSVP(G2) Frequency Analysis"
    B97M_rV_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_G2__Unknown = "B97M-rV/def2-QZVPPD/ISOSVP(G2) Unknown"
    B97M_rV_def2_QZVPPD_ISOSVP_DME__Single_Point = "B97M-rV/def2-QZVPPD/ISOSVP(DME) Single Point"
    B97M_rV_def2_QZVPPD_ISOSVP_DME__Force = "B97M-rV/def2-QZVPPD/ISOSVP(DME) Force"
    B97M_rV_def2_QZVPPD_ISOSVP_DME__Geometry_Optimization = "B97M-rV/def2-QZVPPD/ISOSVP(DME) Geometry Optimization"
    B97M_rV_def2_QZVPPD_ISOSVP_DME__Frequency_Analysis = "B97M-rV/def2-QZVPPD/ISOSVP(DME) Frequency Analysis"
    B97M_rV_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_DME__Unknown = "B97M-rV/def2-QZVPPD/ISOSVP(DME) Unknown"
    B97M_rV_def2_QZVPPD_ISOSVP_EC__Single_Point = "B97M-rV/def2-QZVPPD/ISOSVP(EC) Single Point"
    B97M_rV_def2_QZVPPD_ISOSVP_EC__Force = "B97M-rV/def2-QZVPPD/ISOSVP(EC) Force"
    B97M_rV_def2_QZVPPD_ISOSVP_EC__Geometry_Optimization = "B97M-rV/def2-QZVPPD/ISOSVP(EC) Geometry Optimization"
    B97M_rV_def2_QZVPPD_ISOSVP_EC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/ISOSVP(EC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_EC__Unknown = "B97M-rV/def2-QZVPPD/ISOSVP(EC) Unknown"
    B97M_rV_def2_QZVPPD_ISOSVP_EC_EMC__Single_Point = "B97M-rV/def2-QZVPPD/ISOSVP(EC/EMC) Single Point"
    B97M_rV_def2_QZVPPD_ISOSVP_EC_EMC__Force = "B97M-rV/def2-QZVPPD/ISOSVP(EC/EMC) Force"
    B97M_rV_def2_QZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_EC_EMC__Unknown = "B97M-rV/def2-QZVPPD/ISOSVP(EC/EMC) Unknown"
    B97M_rV_def2_QZVPPD_ISOSVP_PC__Single_Point = "B97M-rV/def2-QZVPPD/ISOSVP(PC) Single Point"
    B97M_rV_def2_QZVPPD_ISOSVP_PC__Force = "B97M-rV/def2-QZVPPD/ISOSVP(PC) Force"
    B97M_rV_def2_QZVPPD_ISOSVP_PC__Geometry_Optimization = "B97M-rV/def2-QZVPPD/ISOSVP(PC) Geometry Optimization"
    B97M_rV_def2_QZVPPD_ISOSVP_PC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/ISOSVP(PC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_ISOSVP_PC__Unknown = "B97M-rV/def2-QZVPPD/ISOSVP(PC) Unknown"
    B97M_rV_def2_QZVPPD_SMD_WATER__Single_Point = "B97M-rV/def2-QZVPPD/SMD(WATER) Single Point"
    B97M_rV_def2_QZVPPD_SMD_WATER__Force = "B97M-rV/def2-QZVPPD/SMD(WATER) Force"
    B97M_rV_def2_QZVPPD_SMD_WATER__Geometry_Optimization = "B97M-rV/def2-QZVPPD/SMD(WATER) Geometry Optimization"
    B97M_rV_def2_QZVPPD_SMD_WATER__Frequency_Analysis = "B97M-rV/def2-QZVPPD/SMD(WATER) Frequency Analysis"
    B97M_rV_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_WATER__Unknown = "B97M-rV/def2-QZVPPD/SMD(WATER) Unknown"
    B97M_rV_def2_QZVPPD_SMD_THF__Single_Point = "B97M-rV/def2-QZVPPD/SMD(THF) Single Point"
    B97M_rV_def2_QZVPPD_SMD_THF__Force = "B97M-rV/def2-QZVPPD/SMD(THF) Force"
    B97M_rV_def2_QZVPPD_SMD_THF__Geometry_Optimization = "B97M-rV/def2-QZVPPD/SMD(THF) Geometry Optimization"
    B97M_rV_def2_QZVPPD_SMD_THF__Frequency_Analysis = "B97M-rV/def2-QZVPPD/SMD(THF) Frequency Analysis"
    B97M_rV_def2_QZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_THF__Unknown = "B97M-rV/def2-QZVPPD/SMD(THF) Unknown"
    B97M_rV_def2_QZVPPD_SMD_G2__Single_Point = "B97M-rV/def2-QZVPPD/SMD(G2) Single Point"
    B97M_rV_def2_QZVPPD_SMD_G2__Force = "B97M-rV/def2-QZVPPD/SMD(G2) Force"
    B97M_rV_def2_QZVPPD_SMD_G2__Geometry_Optimization = "B97M-rV/def2-QZVPPD/SMD(G2) Geometry Optimization"
    B97M_rV_def2_QZVPPD_SMD_G2__Frequency_Analysis = "B97M-rV/def2-QZVPPD/SMD(G2) Frequency Analysis"
    B97M_rV_def2_QZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_G2__Unknown = "B97M-rV/def2-QZVPPD/SMD(G2) Unknown"
    B97M_rV_def2_QZVPPD_SMD_DME__Single_Point = "B97M-rV/def2-QZVPPD/SMD(DME) Single Point"
    B97M_rV_def2_QZVPPD_SMD_DME__Force = "B97M-rV/def2-QZVPPD/SMD(DME) Force"
    B97M_rV_def2_QZVPPD_SMD_DME__Geometry_Optimization = "B97M-rV/def2-QZVPPD/SMD(DME) Geometry Optimization"
    B97M_rV_def2_QZVPPD_SMD_DME__Frequency_Analysis = "B97M-rV/def2-QZVPPD/SMD(DME) Frequency Analysis"
    B97M_rV_def2_QZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_DME__Unknown = "B97M-rV/def2-QZVPPD/SMD(DME) Unknown"
    B97M_rV_def2_QZVPPD_SMD_EC__Single_Point = "B97M-rV/def2-QZVPPD/SMD(EC) Single Point"
    B97M_rV_def2_QZVPPD_SMD_EC__Force = "B97M-rV/def2-QZVPPD/SMD(EC) Force"
    B97M_rV_def2_QZVPPD_SMD_EC__Geometry_Optimization = "B97M-rV/def2-QZVPPD/SMD(EC) Geometry Optimization"
    B97M_rV_def2_QZVPPD_SMD_EC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/SMD(EC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_EC__Unknown = "B97M-rV/def2-QZVPPD/SMD(EC) Unknown"
    B97M_rV_def2_QZVPPD_SMD_EC_EMC__Single_Point = "B97M-rV/def2-QZVPPD/SMD(EC/EMC) Single Point"
    B97M_rV_def2_QZVPPD_SMD_EC_EMC__Force = "B97M-rV/def2-QZVPPD/SMD(EC/EMC) Force"
    B97M_rV_def2_QZVPPD_SMD_EC_EMC__Geometry_Optimization = "B97M-rV/def2-QZVPPD/SMD(EC/EMC) Geometry Optimization"
    B97M_rV_def2_QZVPPD_SMD_EC_EMC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/SMD(EC/EMC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_EC_EMC__Unknown = "B97M-rV/def2-QZVPPD/SMD(EC/EMC) Unknown"
    B97M_rV_def2_QZVPPD_SMD_PC__Single_Point = "B97M-rV/def2-QZVPPD/SMD(PC) Single Point"
    B97M_rV_def2_QZVPPD_SMD_PC__Force = "B97M-rV/def2-QZVPPD/SMD(PC) Force"
    B97M_rV_def2_QZVPPD_SMD_PC__Geometry_Optimization = "B97M-rV/def2-QZVPPD/SMD(PC) Geometry Optimization"
    B97M_rV_def2_QZVPPD_SMD_PC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/SMD(PC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_SMD_PC__Unknown = "B97M-rV/def2-QZVPPD/SMD(PC) Unknown"
    B97M_rV_def2_QZVPPD_CMIRS_WATER__Single_Point = "B97M-rV/def2-QZVPPD/CMIRS(WATER) Single Point"
    B97M_rV_def2_QZVPPD_CMIRS_WATER__Force = "B97M-rV/def2-QZVPPD/CMIRS(WATER) Force"
    B97M_rV_def2_QZVPPD_CMIRS_WATER__Geometry_Optimization = "B97M-rV/def2-QZVPPD/CMIRS(WATER) Geometry Optimization"
    B97M_rV_def2_QZVPPD_CMIRS_WATER__Frequency_Analysis = "B97M-rV/def2-QZVPPD/CMIRS(WATER) Frequency Analysis"
    B97M_rV_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_WATER__Unknown = "B97M-rV/def2-QZVPPD/CMIRS(WATER) Unknown"
    B97M_rV_def2_QZVPPD_CMIRS_THF__Single_Point = "B97M-rV/def2-QZVPPD/CMIRS(THF) Single Point"
    B97M_rV_def2_QZVPPD_CMIRS_THF__Force = "B97M-rV/def2-QZVPPD/CMIRS(THF) Force"
    B97M_rV_def2_QZVPPD_CMIRS_THF__Geometry_Optimization = "B97M-rV/def2-QZVPPD/CMIRS(THF) Geometry Optimization"
    B97M_rV_def2_QZVPPD_CMIRS_THF__Frequency_Analysis = "B97M-rV/def2-QZVPPD/CMIRS(THF) Frequency Analysis"
    B97M_rV_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_THF__Unknown = "B97M-rV/def2-QZVPPD/CMIRS(THF) Unknown"
    B97M_rV_def2_QZVPPD_CMIRS_G2__Single_Point = "B97M-rV/def2-QZVPPD/CMIRS(G2) Single Point"
    B97M_rV_def2_QZVPPD_CMIRS_G2__Force = "B97M-rV/def2-QZVPPD/CMIRS(G2) Force"
    B97M_rV_def2_QZVPPD_CMIRS_G2__Geometry_Optimization = "B97M-rV/def2-QZVPPD/CMIRS(G2) Geometry Optimization"
    B97M_rV_def2_QZVPPD_CMIRS_G2__Frequency_Analysis = "B97M-rV/def2-QZVPPD/CMIRS(G2) Frequency Analysis"
    B97M_rV_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_G2__Unknown = "B97M-rV/def2-QZVPPD/CMIRS(G2) Unknown"
    B97M_rV_def2_QZVPPD_CMIRS_DME__Single_Point = "B97M-rV/def2-QZVPPD/CMIRS(DME) Single Point"
    B97M_rV_def2_QZVPPD_CMIRS_DME__Force = "B97M-rV/def2-QZVPPD/CMIRS(DME) Force"
    B97M_rV_def2_QZVPPD_CMIRS_DME__Geometry_Optimization = "B97M-rV/def2-QZVPPD/CMIRS(DME) Geometry Optimization"
    B97M_rV_def2_QZVPPD_CMIRS_DME__Frequency_Analysis = "B97M-rV/def2-QZVPPD/CMIRS(DME) Frequency Analysis"
    B97M_rV_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_DME__Unknown = "B97M-rV/def2-QZVPPD/CMIRS(DME) Unknown"
    B97M_rV_def2_QZVPPD_CMIRS_EC__Single_Point = "B97M-rV/def2-QZVPPD/CMIRS(EC) Single Point"
    B97M_rV_def2_QZVPPD_CMIRS_EC__Force = "B97M-rV/def2-QZVPPD/CMIRS(EC) Force"
    B97M_rV_def2_QZVPPD_CMIRS_EC__Geometry_Optimization = "B97M-rV/def2-QZVPPD/CMIRS(EC) Geometry Optimization"
    B97M_rV_def2_QZVPPD_CMIRS_EC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/CMIRS(EC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_EC__Unknown = "B97M-rV/def2-QZVPPD/CMIRS(EC) Unknown"
    B97M_rV_def2_QZVPPD_CMIRS_EC_EMC__Single_Point = "B97M-rV/def2-QZVPPD/CMIRS(EC/EMC) Single Point"
    B97M_rV_def2_QZVPPD_CMIRS_EC_EMC__Force = "B97M-rV/def2-QZVPPD/CMIRS(EC/EMC) Force"
    B97M_rV_def2_QZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "B97M-rV/def2-QZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    B97M_rV_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_EC_EMC__Unknown = "B97M-rV/def2-QZVPPD/CMIRS(EC/EMC) Unknown"
    B97M_rV_def2_QZVPPD_CMIRS_PC__Single_Point = "B97M-rV/def2-QZVPPD/CMIRS(PC) Single Point"
    B97M_rV_def2_QZVPPD_CMIRS_PC__Force = "B97M-rV/def2-QZVPPD/CMIRS(PC) Force"
    B97M_rV_def2_QZVPPD_CMIRS_PC__Geometry_Optimization = "B97M-rV/def2-QZVPPD/CMIRS(PC) Geometry Optimization"
    B97M_rV_def2_QZVPPD_CMIRS_PC__Frequency_Analysis = "B97M-rV/def2-QZVPPD/CMIRS(PC) Frequency Analysis"
    B97M_rV_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B97M-rV/def2-QZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B97M_rV_def2_QZVPPD_CMIRS_PC__Unknown = "B97M-rV/def2-QZVPPD/CMIRS(PC) Unknown"
    B3LYP_6_31g_d_VACUUM_Single_Point = "B3LYP/6-31g*/VACUUM Single Point"
    B3LYP_6_31g_d_VACUUM_Force = "B3LYP/6-31g*/VACUUM Force"
    B3LYP_6_31g_d_VACUUM_Geometry_Optimization = "B3LYP/6-31g*/VACUUM Geometry Optimization"
    B3LYP_6_31g_d_VACUUM_Frequency_Analysis = "B3LYP/6-31g*/VACUUM Frequency Analysis"
    B3LYP_6_31g_d_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/VACUUM Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_VACUUM_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/VACUUM Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_VACUUM_Unknown = "B3LYP/6-31g*/VACUUM Unknown"
    B3LYP_6_31g_d_PCM_WATER__Single_Point = "B3LYP/6-31g*/PCM(WATER) Single Point"
    B3LYP_6_31g_d_PCM_WATER__Force = "B3LYP/6-31g*/PCM(WATER) Force"
    B3LYP_6_31g_d_PCM_WATER__Geometry_Optimization = "B3LYP/6-31g*/PCM(WATER) Geometry Optimization"
    B3LYP_6_31g_d_PCM_WATER__Frequency_Analysis = "B3LYP/6-31g*/PCM(WATER) Frequency Analysis"
    B3LYP_6_31g_d_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(WATER) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_WATER__Unknown = "B3LYP/6-31g*/PCM(WATER) Unknown"
    B3LYP_6_31g_d_PCM_THF__Single_Point = "B3LYP/6-31g*/PCM(THF) Single Point"
    B3LYP_6_31g_d_PCM_THF__Force = "B3LYP/6-31g*/PCM(THF) Force"
    B3LYP_6_31g_d_PCM_THF__Geometry_Optimization = "B3LYP/6-31g*/PCM(THF) Geometry Optimization"
    B3LYP_6_31g_d_PCM_THF__Frequency_Analysis = "B3LYP/6-31g*/PCM(THF) Frequency Analysis"
    B3LYP_6_31g_d_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(THF) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_THF__Unknown = "B3LYP/6-31g*/PCM(THF) Unknown"
    B3LYP_6_31g_d_PCM_G2__Single_Point = "B3LYP/6-31g*/PCM(G2) Single Point"
    B3LYP_6_31g_d_PCM_G2__Force = "B3LYP/6-31g*/PCM(G2) Force"
    B3LYP_6_31g_d_PCM_G2__Geometry_Optimization = "B3LYP/6-31g*/PCM(G2) Geometry Optimization"
    B3LYP_6_31g_d_PCM_G2__Frequency_Analysis = "B3LYP/6-31g*/PCM(G2) Frequency Analysis"
    B3LYP_6_31g_d_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(G2) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_G2__Unknown = "B3LYP/6-31g*/PCM(G2) Unknown"
    B3LYP_6_31g_d_PCM_DME__Single_Point = "B3LYP/6-31g*/PCM(DME) Single Point"
    B3LYP_6_31g_d_PCM_DME__Force = "B3LYP/6-31g*/PCM(DME) Force"
    B3LYP_6_31g_d_PCM_DME__Geometry_Optimization = "B3LYP/6-31g*/PCM(DME) Geometry Optimization"
    B3LYP_6_31g_d_PCM_DME__Frequency_Analysis = "B3LYP/6-31g*/PCM(DME) Frequency Analysis"
    B3LYP_6_31g_d_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(DME) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_DME__Unknown = "B3LYP/6-31g*/PCM(DME) Unknown"
    B3LYP_6_31g_d_PCM_EC__Single_Point = "B3LYP/6-31g*/PCM(EC) Single Point"
    B3LYP_6_31g_d_PCM_EC__Force = "B3LYP/6-31g*/PCM(EC) Force"
    B3LYP_6_31g_d_PCM_EC__Geometry_Optimization = "B3LYP/6-31g*/PCM(EC) Geometry Optimization"
    B3LYP_6_31g_d_PCM_EC__Frequency_Analysis = "B3LYP/6-31g*/PCM(EC) Frequency Analysis"
    B3LYP_6_31g_d_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(EC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_EC__Unknown = "B3LYP/6-31g*/PCM(EC) Unknown"
    B3LYP_6_31g_d_PCM_EC_EMC__Single_Point = "B3LYP/6-31g*/PCM(EC/EMC) Single Point"
    B3LYP_6_31g_d_PCM_EC_EMC__Force = "B3LYP/6-31g*/PCM(EC/EMC) Force"
    B3LYP_6_31g_d_PCM_EC_EMC__Geometry_Optimization = "B3LYP/6-31g*/PCM(EC/EMC) Geometry Optimization"
    B3LYP_6_31g_d_PCM_EC_EMC__Frequency_Analysis = "B3LYP/6-31g*/PCM(EC/EMC) Frequency Analysis"
    B3LYP_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_EC_EMC__Unknown = "B3LYP/6-31g*/PCM(EC/EMC) Unknown"
    B3LYP_6_31g_d_PCM_PC__Single_Point = "B3LYP/6-31g*/PCM(PC) Single Point"
    B3LYP_6_31g_d_PCM_PC__Force = "B3LYP/6-31g*/PCM(PC) Force"
    B3LYP_6_31g_d_PCM_PC__Geometry_Optimization = "B3LYP/6-31g*/PCM(PC) Geometry Optimization"
    B3LYP_6_31g_d_PCM_PC__Frequency_Analysis = "B3LYP/6-31g*/PCM(PC) Frequency Analysis"
    B3LYP_6_31g_d_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(PC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_PCM_PC__Unknown = "B3LYP/6-31g*/PCM(PC) Unknown"
    B3LYP_6_31g_d_ISOSVP_WATER__Single_Point = "B3LYP/6-31g*/ISOSVP(WATER) Single Point"
    B3LYP_6_31g_d_ISOSVP_WATER__Force = "B3LYP/6-31g*/ISOSVP(WATER) Force"
    B3LYP_6_31g_d_ISOSVP_WATER__Geometry_Optimization = "B3LYP/6-31g*/ISOSVP(WATER) Geometry Optimization"
    B3LYP_6_31g_d_ISOSVP_WATER__Frequency_Analysis = "B3LYP/6-31g*/ISOSVP(WATER) Frequency Analysis"
    B3LYP_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_WATER__Unknown = "B3LYP/6-31g*/ISOSVP(WATER) Unknown"
    B3LYP_6_31g_d_ISOSVP_THF__Single_Point = "B3LYP/6-31g*/ISOSVP(THF) Single Point"
    B3LYP_6_31g_d_ISOSVP_THF__Force = "B3LYP/6-31g*/ISOSVP(THF) Force"
    B3LYP_6_31g_d_ISOSVP_THF__Geometry_Optimization = "B3LYP/6-31g*/ISOSVP(THF) Geometry Optimization"
    B3LYP_6_31g_d_ISOSVP_THF__Frequency_Analysis = "B3LYP/6-31g*/ISOSVP(THF) Frequency Analysis"
    B3LYP_6_31g_d_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_THF__Unknown = "B3LYP/6-31g*/ISOSVP(THF) Unknown"
    B3LYP_6_31g_d_ISOSVP_G2__Single_Point = "B3LYP/6-31g*/ISOSVP(G2) Single Point"
    B3LYP_6_31g_d_ISOSVP_G2__Force = "B3LYP/6-31g*/ISOSVP(G2) Force"
    B3LYP_6_31g_d_ISOSVP_G2__Geometry_Optimization = "B3LYP/6-31g*/ISOSVP(G2) Geometry Optimization"
    B3LYP_6_31g_d_ISOSVP_G2__Frequency_Analysis = "B3LYP/6-31g*/ISOSVP(G2) Frequency Analysis"
    B3LYP_6_31g_d_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_G2__Unknown = "B3LYP/6-31g*/ISOSVP(G2) Unknown"
    B3LYP_6_31g_d_ISOSVP_DME__Single_Point = "B3LYP/6-31g*/ISOSVP(DME) Single Point"
    B3LYP_6_31g_d_ISOSVP_DME__Force = "B3LYP/6-31g*/ISOSVP(DME) Force"
    B3LYP_6_31g_d_ISOSVP_DME__Geometry_Optimization = "B3LYP/6-31g*/ISOSVP(DME) Geometry Optimization"
    B3LYP_6_31g_d_ISOSVP_DME__Frequency_Analysis = "B3LYP/6-31g*/ISOSVP(DME) Frequency Analysis"
    B3LYP_6_31g_d_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_DME__Unknown = "B3LYP/6-31g*/ISOSVP(DME) Unknown"
    B3LYP_6_31g_d_ISOSVP_EC__Single_Point = "B3LYP/6-31g*/ISOSVP(EC) Single Point"
    B3LYP_6_31g_d_ISOSVP_EC__Force = "B3LYP/6-31g*/ISOSVP(EC) Force"
    B3LYP_6_31g_d_ISOSVP_EC__Geometry_Optimization = "B3LYP/6-31g*/ISOSVP(EC) Geometry Optimization"
    B3LYP_6_31g_d_ISOSVP_EC__Frequency_Analysis = "B3LYP/6-31g*/ISOSVP(EC) Frequency Analysis"
    B3LYP_6_31g_d_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_EC__Unknown = "B3LYP/6-31g*/ISOSVP(EC) Unknown"
    B3LYP_6_31g_d_ISOSVP_EC_EMC__Single_Point = "B3LYP/6-31g*/ISOSVP(EC/EMC) Single Point"
    B3LYP_6_31g_d_ISOSVP_EC_EMC__Force = "B3LYP/6-31g*/ISOSVP(EC/EMC) Force"
    B3LYP_6_31g_d_ISOSVP_EC_EMC__Geometry_Optimization = "B3LYP/6-31g*/ISOSVP(EC/EMC) Geometry Optimization"
    B3LYP_6_31g_d_ISOSVP_EC_EMC__Frequency_Analysis = "B3LYP/6-31g*/ISOSVP(EC/EMC) Frequency Analysis"
    B3LYP_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_EC_EMC__Unknown = "B3LYP/6-31g*/ISOSVP(EC/EMC) Unknown"
    B3LYP_6_31g_d_ISOSVP_PC__Single_Point = "B3LYP/6-31g*/ISOSVP(PC) Single Point"
    B3LYP_6_31g_d_ISOSVP_PC__Force = "B3LYP/6-31g*/ISOSVP(PC) Force"
    B3LYP_6_31g_d_ISOSVP_PC__Geometry_Optimization = "B3LYP/6-31g*/ISOSVP(PC) Geometry Optimization"
    B3LYP_6_31g_d_ISOSVP_PC__Frequency_Analysis = "B3LYP/6-31g*/ISOSVP(PC) Frequency Analysis"
    B3LYP_6_31g_d_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_ISOSVP_PC__Unknown = "B3LYP/6-31g*/ISOSVP(PC) Unknown"
    B3LYP_6_31g_d_SMD_WATER__Single_Point = "B3LYP/6-31g*/SMD(WATER) Single Point"
    B3LYP_6_31g_d_SMD_WATER__Force = "B3LYP/6-31g*/SMD(WATER) Force"
    B3LYP_6_31g_d_SMD_WATER__Geometry_Optimization = "B3LYP/6-31g*/SMD(WATER) Geometry Optimization"
    B3LYP_6_31g_d_SMD_WATER__Frequency_Analysis = "B3LYP/6-31g*/SMD(WATER) Frequency Analysis"
    B3LYP_6_31g_d_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(WATER) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_WATER__Unknown = "B3LYP/6-31g*/SMD(WATER) Unknown"
    B3LYP_6_31g_d_SMD_THF__Single_Point = "B3LYP/6-31g*/SMD(THF) Single Point"
    B3LYP_6_31g_d_SMD_THF__Force = "B3LYP/6-31g*/SMD(THF) Force"
    B3LYP_6_31g_d_SMD_THF__Geometry_Optimization = "B3LYP/6-31g*/SMD(THF) Geometry Optimization"
    B3LYP_6_31g_d_SMD_THF__Frequency_Analysis = "B3LYP/6-31g*/SMD(THF) Frequency Analysis"
    B3LYP_6_31g_d_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(THF) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_THF__Unknown = "B3LYP/6-31g*/SMD(THF) Unknown"
    B3LYP_6_31g_d_SMD_G2__Single_Point = "B3LYP/6-31g*/SMD(G2) Single Point"
    B3LYP_6_31g_d_SMD_G2__Force = "B3LYP/6-31g*/SMD(G2) Force"
    B3LYP_6_31g_d_SMD_G2__Geometry_Optimization = "B3LYP/6-31g*/SMD(G2) Geometry Optimization"
    B3LYP_6_31g_d_SMD_G2__Frequency_Analysis = "B3LYP/6-31g*/SMD(G2) Frequency Analysis"
    B3LYP_6_31g_d_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(G2) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_G2__Unknown = "B3LYP/6-31g*/SMD(G2) Unknown"
    B3LYP_6_31g_d_SMD_DME__Single_Point = "B3LYP/6-31g*/SMD(DME) Single Point"
    B3LYP_6_31g_d_SMD_DME__Force = "B3LYP/6-31g*/SMD(DME) Force"
    B3LYP_6_31g_d_SMD_DME__Geometry_Optimization = "B3LYP/6-31g*/SMD(DME) Geometry Optimization"
    B3LYP_6_31g_d_SMD_DME__Frequency_Analysis = "B3LYP/6-31g*/SMD(DME) Frequency Analysis"
    B3LYP_6_31g_d_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(DME) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_DME__Unknown = "B3LYP/6-31g*/SMD(DME) Unknown"
    B3LYP_6_31g_d_SMD_EC__Single_Point = "B3LYP/6-31g*/SMD(EC) Single Point"
    B3LYP_6_31g_d_SMD_EC__Force = "B3LYP/6-31g*/SMD(EC) Force"
    B3LYP_6_31g_d_SMD_EC__Geometry_Optimization = "B3LYP/6-31g*/SMD(EC) Geometry Optimization"
    B3LYP_6_31g_d_SMD_EC__Frequency_Analysis = "B3LYP/6-31g*/SMD(EC) Frequency Analysis"
    B3LYP_6_31g_d_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(EC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_EC__Unknown = "B3LYP/6-31g*/SMD(EC) Unknown"
    B3LYP_6_31g_d_SMD_EC_EMC__Single_Point = "B3LYP/6-31g*/SMD(EC/EMC) Single Point"
    B3LYP_6_31g_d_SMD_EC_EMC__Force = "B3LYP/6-31g*/SMD(EC/EMC) Force"
    B3LYP_6_31g_d_SMD_EC_EMC__Geometry_Optimization = "B3LYP/6-31g*/SMD(EC/EMC) Geometry Optimization"
    B3LYP_6_31g_d_SMD_EC_EMC__Frequency_Analysis = "B3LYP/6-31g*/SMD(EC/EMC) Frequency Analysis"
    B3LYP_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_EC_EMC__Unknown = "B3LYP/6-31g*/SMD(EC/EMC) Unknown"
    B3LYP_6_31g_d_SMD_PC__Single_Point = "B3LYP/6-31g*/SMD(PC) Single Point"
    B3LYP_6_31g_d_SMD_PC__Force = "B3LYP/6-31g*/SMD(PC) Force"
    B3LYP_6_31g_d_SMD_PC__Geometry_Optimization = "B3LYP/6-31g*/SMD(PC) Geometry Optimization"
    B3LYP_6_31g_d_SMD_PC__Frequency_Analysis = "B3LYP/6-31g*/SMD(PC) Frequency Analysis"
    B3LYP_6_31g_d_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(PC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_SMD_PC__Unknown = "B3LYP/6-31g*/SMD(PC) Unknown"
    B3LYP_6_31g_d_CMIRS_WATER__Single_Point = "B3LYP/6-31g*/CMIRS(WATER) Single Point"
    B3LYP_6_31g_d_CMIRS_WATER__Force = "B3LYP/6-31g*/CMIRS(WATER) Force"
    B3LYP_6_31g_d_CMIRS_WATER__Geometry_Optimization = "B3LYP/6-31g*/CMIRS(WATER) Geometry Optimization"
    B3LYP_6_31g_d_CMIRS_WATER__Frequency_Analysis = "B3LYP/6-31g*/CMIRS(WATER) Frequency Analysis"
    B3LYP_6_31g_d_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_WATER__Unknown = "B3LYP/6-31g*/CMIRS(WATER) Unknown"
    B3LYP_6_31g_d_CMIRS_THF__Single_Point = "B3LYP/6-31g*/CMIRS(THF) Single Point"
    B3LYP_6_31g_d_CMIRS_THF__Force = "B3LYP/6-31g*/CMIRS(THF) Force"
    B3LYP_6_31g_d_CMIRS_THF__Geometry_Optimization = "B3LYP/6-31g*/CMIRS(THF) Geometry Optimization"
    B3LYP_6_31g_d_CMIRS_THF__Frequency_Analysis = "B3LYP/6-31g*/CMIRS(THF) Frequency Analysis"
    B3LYP_6_31g_d_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(THF) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_THF__Unknown = "B3LYP/6-31g*/CMIRS(THF) Unknown"
    B3LYP_6_31g_d_CMIRS_G2__Single_Point = "B3LYP/6-31g*/CMIRS(G2) Single Point"
    B3LYP_6_31g_d_CMIRS_G2__Force = "B3LYP/6-31g*/CMIRS(G2) Force"
    B3LYP_6_31g_d_CMIRS_G2__Geometry_Optimization = "B3LYP/6-31g*/CMIRS(G2) Geometry Optimization"
    B3LYP_6_31g_d_CMIRS_G2__Frequency_Analysis = "B3LYP/6-31g*/CMIRS(G2) Frequency Analysis"
    B3LYP_6_31g_d_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(G2) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_G2__Unknown = "B3LYP/6-31g*/CMIRS(G2) Unknown"
    B3LYP_6_31g_d_CMIRS_DME__Single_Point = "B3LYP/6-31g*/CMIRS(DME) Single Point"
    B3LYP_6_31g_d_CMIRS_DME__Force = "B3LYP/6-31g*/CMIRS(DME) Force"
    B3LYP_6_31g_d_CMIRS_DME__Geometry_Optimization = "B3LYP/6-31g*/CMIRS(DME) Geometry Optimization"
    B3LYP_6_31g_d_CMIRS_DME__Frequency_Analysis = "B3LYP/6-31g*/CMIRS(DME) Frequency Analysis"
    B3LYP_6_31g_d_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(DME) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_DME__Unknown = "B3LYP/6-31g*/CMIRS(DME) Unknown"
    B3LYP_6_31g_d_CMIRS_EC__Single_Point = "B3LYP/6-31g*/CMIRS(EC) Single Point"
    B3LYP_6_31g_d_CMIRS_EC__Force = "B3LYP/6-31g*/CMIRS(EC) Force"
    B3LYP_6_31g_d_CMIRS_EC__Geometry_Optimization = "B3LYP/6-31g*/CMIRS(EC) Geometry Optimization"
    B3LYP_6_31g_d_CMIRS_EC__Frequency_Analysis = "B3LYP/6-31g*/CMIRS(EC) Frequency Analysis"
    B3LYP_6_31g_d_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(EC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_EC__Unknown = "B3LYP/6-31g*/CMIRS(EC) Unknown"
    B3LYP_6_31g_d_CMIRS_EC_EMC__Single_Point = "B3LYP/6-31g*/CMIRS(EC/EMC) Single Point"
    B3LYP_6_31g_d_CMIRS_EC_EMC__Force = "B3LYP/6-31g*/CMIRS(EC/EMC) Force"
    B3LYP_6_31g_d_CMIRS_EC_EMC__Geometry_Optimization = "B3LYP/6-31g*/CMIRS(EC/EMC) Geometry Optimization"
    B3LYP_6_31g_d_CMIRS_EC_EMC__Frequency_Analysis = "B3LYP/6-31g*/CMIRS(EC/EMC) Frequency Analysis"
    B3LYP_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_EC_EMC__Unknown = "B3LYP/6-31g*/CMIRS(EC/EMC) Unknown"
    B3LYP_6_31g_d_CMIRS_PC__Single_Point = "B3LYP/6-31g*/CMIRS(PC) Single Point"
    B3LYP_6_31g_d_CMIRS_PC__Force = "B3LYP/6-31g*/CMIRS(PC) Force"
    B3LYP_6_31g_d_CMIRS_PC__Geometry_Optimization = "B3LYP/6-31g*/CMIRS(PC) Geometry Optimization"
    B3LYP_6_31g_d_CMIRS_PC__Frequency_Analysis = "B3LYP/6-31g*/CMIRS(PC) Frequency Analysis"
    B3LYP_6_31g_d_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(PC) Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/6-31g*/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_6_31g_d_CMIRS_PC__Unknown = "B3LYP/6-31g*/CMIRS(PC) Unknown"
    B3LYP_def2_SVPD_VACUUM_Single_Point = "B3LYP/def2-SVPD/VACUUM Single Point"
    B3LYP_def2_SVPD_VACUUM_Force = "B3LYP/def2-SVPD/VACUUM Force"
    B3LYP_def2_SVPD_VACUUM_Geometry_Optimization = "B3LYP/def2-SVPD/VACUUM Geometry Optimization"
    B3LYP_def2_SVPD_VACUUM_Frequency_Analysis = "B3LYP/def2-SVPD/VACUUM Frequency Analysis"
    B3LYP_def2_SVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/VACUUM Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_VACUUM_Unknown = "B3LYP/def2-SVPD/VACUUM Unknown"
    B3LYP_def2_SVPD_PCM_WATER__Single_Point = "B3LYP/def2-SVPD/PCM(WATER) Single Point"
    B3LYP_def2_SVPD_PCM_WATER__Force = "B3LYP/def2-SVPD/PCM(WATER) Force"
    B3LYP_def2_SVPD_PCM_WATER__Geometry_Optimization = "B3LYP/def2-SVPD/PCM(WATER) Geometry Optimization"
    B3LYP_def2_SVPD_PCM_WATER__Frequency_Analysis = "B3LYP/def2-SVPD/PCM(WATER) Frequency Analysis"
    B3LYP_def2_SVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_WATER__Unknown = "B3LYP/def2-SVPD/PCM(WATER) Unknown"
    B3LYP_def2_SVPD_PCM_THF__Single_Point = "B3LYP/def2-SVPD/PCM(THF) Single Point"
    B3LYP_def2_SVPD_PCM_THF__Force = "B3LYP/def2-SVPD/PCM(THF) Force"
    B3LYP_def2_SVPD_PCM_THF__Geometry_Optimization = "B3LYP/def2-SVPD/PCM(THF) Geometry Optimization"
    B3LYP_def2_SVPD_PCM_THF__Frequency_Analysis = "B3LYP/def2-SVPD/PCM(THF) Frequency Analysis"
    B3LYP_def2_SVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_THF__Unknown = "B3LYP/def2-SVPD/PCM(THF) Unknown"
    B3LYP_def2_SVPD_PCM_G2__Single_Point = "B3LYP/def2-SVPD/PCM(G2) Single Point"
    B3LYP_def2_SVPD_PCM_G2__Force = "B3LYP/def2-SVPD/PCM(G2) Force"
    B3LYP_def2_SVPD_PCM_G2__Geometry_Optimization = "B3LYP/def2-SVPD/PCM(G2) Geometry Optimization"
    B3LYP_def2_SVPD_PCM_G2__Frequency_Analysis = "B3LYP/def2-SVPD/PCM(G2) Frequency Analysis"
    B3LYP_def2_SVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_G2__Unknown = "B3LYP/def2-SVPD/PCM(G2) Unknown"
    B3LYP_def2_SVPD_PCM_DME__Single_Point = "B3LYP/def2-SVPD/PCM(DME) Single Point"
    B3LYP_def2_SVPD_PCM_DME__Force = "B3LYP/def2-SVPD/PCM(DME) Force"
    B3LYP_def2_SVPD_PCM_DME__Geometry_Optimization = "B3LYP/def2-SVPD/PCM(DME) Geometry Optimization"
    B3LYP_def2_SVPD_PCM_DME__Frequency_Analysis = "B3LYP/def2-SVPD/PCM(DME) Frequency Analysis"
    B3LYP_def2_SVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_DME__Unknown = "B3LYP/def2-SVPD/PCM(DME) Unknown"
    B3LYP_def2_SVPD_PCM_EC__Single_Point = "B3LYP/def2-SVPD/PCM(EC) Single Point"
    B3LYP_def2_SVPD_PCM_EC__Force = "B3LYP/def2-SVPD/PCM(EC) Force"
    B3LYP_def2_SVPD_PCM_EC__Geometry_Optimization = "B3LYP/def2-SVPD/PCM(EC) Geometry Optimization"
    B3LYP_def2_SVPD_PCM_EC__Frequency_Analysis = "B3LYP/def2-SVPD/PCM(EC) Frequency Analysis"
    B3LYP_def2_SVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_EC__Unknown = "B3LYP/def2-SVPD/PCM(EC) Unknown"
    B3LYP_def2_SVPD_PCM_EC_EMC__Single_Point = "B3LYP/def2-SVPD/PCM(EC/EMC) Single Point"
    B3LYP_def2_SVPD_PCM_EC_EMC__Force = "B3LYP/def2-SVPD/PCM(EC/EMC) Force"
    B3LYP_def2_SVPD_PCM_EC_EMC__Geometry_Optimization = "B3LYP/def2-SVPD/PCM(EC/EMC) Geometry Optimization"
    B3LYP_def2_SVPD_PCM_EC_EMC__Frequency_Analysis = "B3LYP/def2-SVPD/PCM(EC/EMC) Frequency Analysis"
    B3LYP_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_EC_EMC__Unknown = "B3LYP/def2-SVPD/PCM(EC/EMC) Unknown"
    B3LYP_def2_SVPD_PCM_PC__Single_Point = "B3LYP/def2-SVPD/PCM(PC) Single Point"
    B3LYP_def2_SVPD_PCM_PC__Force = "B3LYP/def2-SVPD/PCM(PC) Force"
    B3LYP_def2_SVPD_PCM_PC__Geometry_Optimization = "B3LYP/def2-SVPD/PCM(PC) Geometry Optimization"
    B3LYP_def2_SVPD_PCM_PC__Frequency_Analysis = "B3LYP/def2-SVPD/PCM(PC) Frequency Analysis"
    B3LYP_def2_SVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_PCM_PC__Unknown = "B3LYP/def2-SVPD/PCM(PC) Unknown"
    B3LYP_def2_SVPD_ISOSVP_WATER__Single_Point = "B3LYP/def2-SVPD/ISOSVP(WATER) Single Point"
    B3LYP_def2_SVPD_ISOSVP_WATER__Force = "B3LYP/def2-SVPD/ISOSVP(WATER) Force"
    B3LYP_def2_SVPD_ISOSVP_WATER__Geometry_Optimization = "B3LYP/def2-SVPD/ISOSVP(WATER) Geometry Optimization"
    B3LYP_def2_SVPD_ISOSVP_WATER__Frequency_Analysis = "B3LYP/def2-SVPD/ISOSVP(WATER) Frequency Analysis"
    B3LYP_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_WATER__Unknown = "B3LYP/def2-SVPD/ISOSVP(WATER) Unknown"
    B3LYP_def2_SVPD_ISOSVP_THF__Single_Point = "B3LYP/def2-SVPD/ISOSVP(THF) Single Point"
    B3LYP_def2_SVPD_ISOSVP_THF__Force = "B3LYP/def2-SVPD/ISOSVP(THF) Force"
    B3LYP_def2_SVPD_ISOSVP_THF__Geometry_Optimization = "B3LYP/def2-SVPD/ISOSVP(THF) Geometry Optimization"
    B3LYP_def2_SVPD_ISOSVP_THF__Frequency_Analysis = "B3LYP/def2-SVPD/ISOSVP(THF) Frequency Analysis"
    B3LYP_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_THF__Unknown = "B3LYP/def2-SVPD/ISOSVP(THF) Unknown"
    B3LYP_def2_SVPD_ISOSVP_G2__Single_Point = "B3LYP/def2-SVPD/ISOSVP(G2) Single Point"
    B3LYP_def2_SVPD_ISOSVP_G2__Force = "B3LYP/def2-SVPD/ISOSVP(G2) Force"
    B3LYP_def2_SVPD_ISOSVP_G2__Geometry_Optimization = "B3LYP/def2-SVPD/ISOSVP(G2) Geometry Optimization"
    B3LYP_def2_SVPD_ISOSVP_G2__Frequency_Analysis = "B3LYP/def2-SVPD/ISOSVP(G2) Frequency Analysis"
    B3LYP_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_G2__Unknown = "B3LYP/def2-SVPD/ISOSVP(G2) Unknown"
    B3LYP_def2_SVPD_ISOSVP_DME__Single_Point = "B3LYP/def2-SVPD/ISOSVP(DME) Single Point"
    B3LYP_def2_SVPD_ISOSVP_DME__Force = "B3LYP/def2-SVPD/ISOSVP(DME) Force"
    B3LYP_def2_SVPD_ISOSVP_DME__Geometry_Optimization = "B3LYP/def2-SVPD/ISOSVP(DME) Geometry Optimization"
    B3LYP_def2_SVPD_ISOSVP_DME__Frequency_Analysis = "B3LYP/def2-SVPD/ISOSVP(DME) Frequency Analysis"
    B3LYP_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_DME__Unknown = "B3LYP/def2-SVPD/ISOSVP(DME) Unknown"
    B3LYP_def2_SVPD_ISOSVP_EC__Single_Point = "B3LYP/def2-SVPD/ISOSVP(EC) Single Point"
    B3LYP_def2_SVPD_ISOSVP_EC__Force = "B3LYP/def2-SVPD/ISOSVP(EC) Force"
    B3LYP_def2_SVPD_ISOSVP_EC__Geometry_Optimization = "B3LYP/def2-SVPD/ISOSVP(EC) Geometry Optimization"
    B3LYP_def2_SVPD_ISOSVP_EC__Frequency_Analysis = "B3LYP/def2-SVPD/ISOSVP(EC) Frequency Analysis"
    B3LYP_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_EC__Unknown = "B3LYP/def2-SVPD/ISOSVP(EC) Unknown"
    B3LYP_def2_SVPD_ISOSVP_EC_EMC__Single_Point = "B3LYP/def2-SVPD/ISOSVP(EC/EMC) Single Point"
    B3LYP_def2_SVPD_ISOSVP_EC_EMC__Force = "B3LYP/def2-SVPD/ISOSVP(EC/EMC) Force"
    B3LYP_def2_SVPD_ISOSVP_EC_EMC__Geometry_Optimization = "B3LYP/def2-SVPD/ISOSVP(EC/EMC) Geometry Optimization"
    B3LYP_def2_SVPD_ISOSVP_EC_EMC__Frequency_Analysis = "B3LYP/def2-SVPD/ISOSVP(EC/EMC) Frequency Analysis"
    B3LYP_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_EC_EMC__Unknown = "B3LYP/def2-SVPD/ISOSVP(EC/EMC) Unknown"
    B3LYP_def2_SVPD_ISOSVP_PC__Single_Point = "B3LYP/def2-SVPD/ISOSVP(PC) Single Point"
    B3LYP_def2_SVPD_ISOSVP_PC__Force = "B3LYP/def2-SVPD/ISOSVP(PC) Force"
    B3LYP_def2_SVPD_ISOSVP_PC__Geometry_Optimization = "B3LYP/def2-SVPD/ISOSVP(PC) Geometry Optimization"
    B3LYP_def2_SVPD_ISOSVP_PC__Frequency_Analysis = "B3LYP/def2-SVPD/ISOSVP(PC) Frequency Analysis"
    B3LYP_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_ISOSVP_PC__Unknown = "B3LYP/def2-SVPD/ISOSVP(PC) Unknown"
    B3LYP_def2_SVPD_SMD_WATER__Single_Point = "B3LYP/def2-SVPD/SMD(WATER) Single Point"
    B3LYP_def2_SVPD_SMD_WATER__Force = "B3LYP/def2-SVPD/SMD(WATER) Force"
    B3LYP_def2_SVPD_SMD_WATER__Geometry_Optimization = "B3LYP/def2-SVPD/SMD(WATER) Geometry Optimization"
    B3LYP_def2_SVPD_SMD_WATER__Frequency_Analysis = "B3LYP/def2-SVPD/SMD(WATER) Frequency Analysis"
    B3LYP_def2_SVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_WATER__Unknown = "B3LYP/def2-SVPD/SMD(WATER) Unknown"
    B3LYP_def2_SVPD_SMD_THF__Single_Point = "B3LYP/def2-SVPD/SMD(THF) Single Point"
    B3LYP_def2_SVPD_SMD_THF__Force = "B3LYP/def2-SVPD/SMD(THF) Force"
    B3LYP_def2_SVPD_SMD_THF__Geometry_Optimization = "B3LYP/def2-SVPD/SMD(THF) Geometry Optimization"
    B3LYP_def2_SVPD_SMD_THF__Frequency_Analysis = "B3LYP/def2-SVPD/SMD(THF) Frequency Analysis"
    B3LYP_def2_SVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_THF__Unknown = "B3LYP/def2-SVPD/SMD(THF) Unknown"
    B3LYP_def2_SVPD_SMD_G2__Single_Point = "B3LYP/def2-SVPD/SMD(G2) Single Point"
    B3LYP_def2_SVPD_SMD_G2__Force = "B3LYP/def2-SVPD/SMD(G2) Force"
    B3LYP_def2_SVPD_SMD_G2__Geometry_Optimization = "B3LYP/def2-SVPD/SMD(G2) Geometry Optimization"
    B3LYP_def2_SVPD_SMD_G2__Frequency_Analysis = "B3LYP/def2-SVPD/SMD(G2) Frequency Analysis"
    B3LYP_def2_SVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_G2__Unknown = "B3LYP/def2-SVPD/SMD(G2) Unknown"
    B3LYP_def2_SVPD_SMD_DME__Single_Point = "B3LYP/def2-SVPD/SMD(DME) Single Point"
    B3LYP_def2_SVPD_SMD_DME__Force = "B3LYP/def2-SVPD/SMD(DME) Force"
    B3LYP_def2_SVPD_SMD_DME__Geometry_Optimization = "B3LYP/def2-SVPD/SMD(DME) Geometry Optimization"
    B3LYP_def2_SVPD_SMD_DME__Frequency_Analysis = "B3LYP/def2-SVPD/SMD(DME) Frequency Analysis"
    B3LYP_def2_SVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_DME__Unknown = "B3LYP/def2-SVPD/SMD(DME) Unknown"
    B3LYP_def2_SVPD_SMD_EC__Single_Point = "B3LYP/def2-SVPD/SMD(EC) Single Point"
    B3LYP_def2_SVPD_SMD_EC__Force = "B3LYP/def2-SVPD/SMD(EC) Force"
    B3LYP_def2_SVPD_SMD_EC__Geometry_Optimization = "B3LYP/def2-SVPD/SMD(EC) Geometry Optimization"
    B3LYP_def2_SVPD_SMD_EC__Frequency_Analysis = "B3LYP/def2-SVPD/SMD(EC) Frequency Analysis"
    B3LYP_def2_SVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_EC__Unknown = "B3LYP/def2-SVPD/SMD(EC) Unknown"
    B3LYP_def2_SVPD_SMD_EC_EMC__Single_Point = "B3LYP/def2-SVPD/SMD(EC/EMC) Single Point"
    B3LYP_def2_SVPD_SMD_EC_EMC__Force = "B3LYP/def2-SVPD/SMD(EC/EMC) Force"
    B3LYP_def2_SVPD_SMD_EC_EMC__Geometry_Optimization = "B3LYP/def2-SVPD/SMD(EC/EMC) Geometry Optimization"
    B3LYP_def2_SVPD_SMD_EC_EMC__Frequency_Analysis = "B3LYP/def2-SVPD/SMD(EC/EMC) Frequency Analysis"
    B3LYP_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_EC_EMC__Unknown = "B3LYP/def2-SVPD/SMD(EC/EMC) Unknown"
    B3LYP_def2_SVPD_SMD_PC__Single_Point = "B3LYP/def2-SVPD/SMD(PC) Single Point"
    B3LYP_def2_SVPD_SMD_PC__Force = "B3LYP/def2-SVPD/SMD(PC) Force"
    B3LYP_def2_SVPD_SMD_PC__Geometry_Optimization = "B3LYP/def2-SVPD/SMD(PC) Geometry Optimization"
    B3LYP_def2_SVPD_SMD_PC__Frequency_Analysis = "B3LYP/def2-SVPD/SMD(PC) Frequency Analysis"
    B3LYP_def2_SVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_SMD_PC__Unknown = "B3LYP/def2-SVPD/SMD(PC) Unknown"
    B3LYP_def2_SVPD_CMIRS_WATER__Single_Point = "B3LYP/def2-SVPD/CMIRS(WATER) Single Point"
    B3LYP_def2_SVPD_CMIRS_WATER__Force = "B3LYP/def2-SVPD/CMIRS(WATER) Force"
    B3LYP_def2_SVPD_CMIRS_WATER__Geometry_Optimization = "B3LYP/def2-SVPD/CMIRS(WATER) Geometry Optimization"
    B3LYP_def2_SVPD_CMIRS_WATER__Frequency_Analysis = "B3LYP/def2-SVPD/CMIRS(WATER) Frequency Analysis"
    B3LYP_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_WATER__Unknown = "B3LYP/def2-SVPD/CMIRS(WATER) Unknown"
    B3LYP_def2_SVPD_CMIRS_THF__Single_Point = "B3LYP/def2-SVPD/CMIRS(THF) Single Point"
    B3LYP_def2_SVPD_CMIRS_THF__Force = "B3LYP/def2-SVPD/CMIRS(THF) Force"
    B3LYP_def2_SVPD_CMIRS_THF__Geometry_Optimization = "B3LYP/def2-SVPD/CMIRS(THF) Geometry Optimization"
    B3LYP_def2_SVPD_CMIRS_THF__Frequency_Analysis = "B3LYP/def2-SVPD/CMIRS(THF) Frequency Analysis"
    B3LYP_def2_SVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_THF__Unknown = "B3LYP/def2-SVPD/CMIRS(THF) Unknown"
    B3LYP_def2_SVPD_CMIRS_G2__Single_Point = "B3LYP/def2-SVPD/CMIRS(G2) Single Point"
    B3LYP_def2_SVPD_CMIRS_G2__Force = "B3LYP/def2-SVPD/CMIRS(G2) Force"
    B3LYP_def2_SVPD_CMIRS_G2__Geometry_Optimization = "B3LYP/def2-SVPD/CMIRS(G2) Geometry Optimization"
    B3LYP_def2_SVPD_CMIRS_G2__Frequency_Analysis = "B3LYP/def2-SVPD/CMIRS(G2) Frequency Analysis"
    B3LYP_def2_SVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_G2__Unknown = "B3LYP/def2-SVPD/CMIRS(G2) Unknown"
    B3LYP_def2_SVPD_CMIRS_DME__Single_Point = "B3LYP/def2-SVPD/CMIRS(DME) Single Point"
    B3LYP_def2_SVPD_CMIRS_DME__Force = "B3LYP/def2-SVPD/CMIRS(DME) Force"
    B3LYP_def2_SVPD_CMIRS_DME__Geometry_Optimization = "B3LYP/def2-SVPD/CMIRS(DME) Geometry Optimization"
    B3LYP_def2_SVPD_CMIRS_DME__Frequency_Analysis = "B3LYP/def2-SVPD/CMIRS(DME) Frequency Analysis"
    B3LYP_def2_SVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_DME__Unknown = "B3LYP/def2-SVPD/CMIRS(DME) Unknown"
    B3LYP_def2_SVPD_CMIRS_EC__Single_Point = "B3LYP/def2-SVPD/CMIRS(EC) Single Point"
    B3LYP_def2_SVPD_CMIRS_EC__Force = "B3LYP/def2-SVPD/CMIRS(EC) Force"
    B3LYP_def2_SVPD_CMIRS_EC__Geometry_Optimization = "B3LYP/def2-SVPD/CMIRS(EC) Geometry Optimization"
    B3LYP_def2_SVPD_CMIRS_EC__Frequency_Analysis = "B3LYP/def2-SVPD/CMIRS(EC) Frequency Analysis"
    B3LYP_def2_SVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_EC__Unknown = "B3LYP/def2-SVPD/CMIRS(EC) Unknown"
    B3LYP_def2_SVPD_CMIRS_EC_EMC__Single_Point = "B3LYP/def2-SVPD/CMIRS(EC/EMC) Single Point"
    B3LYP_def2_SVPD_CMIRS_EC_EMC__Force = "B3LYP/def2-SVPD/CMIRS(EC/EMC) Force"
    B3LYP_def2_SVPD_CMIRS_EC_EMC__Geometry_Optimization = "B3LYP/def2-SVPD/CMIRS(EC/EMC) Geometry Optimization"
    B3LYP_def2_SVPD_CMIRS_EC_EMC__Frequency_Analysis = "B3LYP/def2-SVPD/CMIRS(EC/EMC) Frequency Analysis"
    B3LYP_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_EC_EMC__Unknown = "B3LYP/def2-SVPD/CMIRS(EC/EMC) Unknown"
    B3LYP_def2_SVPD_CMIRS_PC__Single_Point = "B3LYP/def2-SVPD/CMIRS(PC) Single Point"
    B3LYP_def2_SVPD_CMIRS_PC__Force = "B3LYP/def2-SVPD/CMIRS(PC) Force"
    B3LYP_def2_SVPD_CMIRS_PC__Geometry_Optimization = "B3LYP/def2-SVPD/CMIRS(PC) Geometry Optimization"
    B3LYP_def2_SVPD_CMIRS_PC__Frequency_Analysis = "B3LYP/def2-SVPD/CMIRS(PC) Frequency Analysis"
    B3LYP_def2_SVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-SVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_SVPD_CMIRS_PC__Unknown = "B3LYP/def2-SVPD/CMIRS(PC) Unknown"
    B3LYP_def2_TZVP_VACUUM_Single_Point = "B3LYP/def2-TZVP/VACUUM Single Point"
    B3LYP_def2_TZVP_VACUUM_Force = "B3LYP/def2-TZVP/VACUUM Force"
    B3LYP_def2_TZVP_VACUUM_Geometry_Optimization = "B3LYP/def2-TZVP/VACUUM Geometry Optimization"
    B3LYP_def2_TZVP_VACUUM_Frequency_Analysis = "B3LYP/def2-TZVP/VACUUM Frequency Analysis"
    B3LYP_def2_TZVP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/VACUUM Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_VACUUM_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/VACUUM Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_VACUUM_Unknown = "B3LYP/def2-TZVP/VACUUM Unknown"
    B3LYP_def2_TZVP_PCM_WATER__Single_Point = "B3LYP/def2-TZVP/PCM(WATER) Single Point"
    B3LYP_def2_TZVP_PCM_WATER__Force = "B3LYP/def2-TZVP/PCM(WATER) Force"
    B3LYP_def2_TZVP_PCM_WATER__Geometry_Optimization = "B3LYP/def2-TZVP/PCM(WATER) Geometry Optimization"
    B3LYP_def2_TZVP_PCM_WATER__Frequency_Analysis = "B3LYP/def2-TZVP/PCM(WATER) Frequency Analysis"
    B3LYP_def2_TZVP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_WATER__Unknown = "B3LYP/def2-TZVP/PCM(WATER) Unknown"
    B3LYP_def2_TZVP_PCM_THF__Single_Point = "B3LYP/def2-TZVP/PCM(THF) Single Point"
    B3LYP_def2_TZVP_PCM_THF__Force = "B3LYP/def2-TZVP/PCM(THF) Force"
    B3LYP_def2_TZVP_PCM_THF__Geometry_Optimization = "B3LYP/def2-TZVP/PCM(THF) Geometry Optimization"
    B3LYP_def2_TZVP_PCM_THF__Frequency_Analysis = "B3LYP/def2-TZVP/PCM(THF) Frequency Analysis"
    B3LYP_def2_TZVP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_THF__Unknown = "B3LYP/def2-TZVP/PCM(THF) Unknown"
    B3LYP_def2_TZVP_PCM_G2__Single_Point = "B3LYP/def2-TZVP/PCM(G2) Single Point"
    B3LYP_def2_TZVP_PCM_G2__Force = "B3LYP/def2-TZVP/PCM(G2) Force"
    B3LYP_def2_TZVP_PCM_G2__Geometry_Optimization = "B3LYP/def2-TZVP/PCM(G2) Geometry Optimization"
    B3LYP_def2_TZVP_PCM_G2__Frequency_Analysis = "B3LYP/def2-TZVP/PCM(G2) Frequency Analysis"
    B3LYP_def2_TZVP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_G2__Unknown = "B3LYP/def2-TZVP/PCM(G2) Unknown"
    B3LYP_def2_TZVP_PCM_DME__Single_Point = "B3LYP/def2-TZVP/PCM(DME) Single Point"
    B3LYP_def2_TZVP_PCM_DME__Force = "B3LYP/def2-TZVP/PCM(DME) Force"
    B3LYP_def2_TZVP_PCM_DME__Geometry_Optimization = "B3LYP/def2-TZVP/PCM(DME) Geometry Optimization"
    B3LYP_def2_TZVP_PCM_DME__Frequency_Analysis = "B3LYP/def2-TZVP/PCM(DME) Frequency Analysis"
    B3LYP_def2_TZVP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_DME__Unknown = "B3LYP/def2-TZVP/PCM(DME) Unknown"
    B3LYP_def2_TZVP_PCM_EC__Single_Point = "B3LYP/def2-TZVP/PCM(EC) Single Point"
    B3LYP_def2_TZVP_PCM_EC__Force = "B3LYP/def2-TZVP/PCM(EC) Force"
    B3LYP_def2_TZVP_PCM_EC__Geometry_Optimization = "B3LYP/def2-TZVP/PCM(EC) Geometry Optimization"
    B3LYP_def2_TZVP_PCM_EC__Frequency_Analysis = "B3LYP/def2-TZVP/PCM(EC) Frequency Analysis"
    B3LYP_def2_TZVP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_EC__Unknown = "B3LYP/def2-TZVP/PCM(EC) Unknown"
    B3LYP_def2_TZVP_PCM_EC_EMC__Single_Point = "B3LYP/def2-TZVP/PCM(EC/EMC) Single Point"
    B3LYP_def2_TZVP_PCM_EC_EMC__Force = "B3LYP/def2-TZVP/PCM(EC/EMC) Force"
    B3LYP_def2_TZVP_PCM_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVP/PCM(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVP_PCM_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVP/PCM(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_EC_EMC__Unknown = "B3LYP/def2-TZVP/PCM(EC/EMC) Unknown"
    B3LYP_def2_TZVP_PCM_PC__Single_Point = "B3LYP/def2-TZVP/PCM(PC) Single Point"
    B3LYP_def2_TZVP_PCM_PC__Force = "B3LYP/def2-TZVP/PCM(PC) Force"
    B3LYP_def2_TZVP_PCM_PC__Geometry_Optimization = "B3LYP/def2-TZVP/PCM(PC) Geometry Optimization"
    B3LYP_def2_TZVP_PCM_PC__Frequency_Analysis = "B3LYP/def2-TZVP/PCM(PC) Frequency Analysis"
    B3LYP_def2_TZVP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_PCM_PC__Unknown = "B3LYP/def2-TZVP/PCM(PC) Unknown"
    B3LYP_def2_TZVP_ISOSVP_WATER__Single_Point = "B3LYP/def2-TZVP/ISOSVP(WATER) Single Point"
    B3LYP_def2_TZVP_ISOSVP_WATER__Force = "B3LYP/def2-TZVP/ISOSVP(WATER) Force"
    B3LYP_def2_TZVP_ISOSVP_WATER__Geometry_Optimization = "B3LYP/def2-TZVP/ISOSVP(WATER) Geometry Optimization"
    B3LYP_def2_TZVP_ISOSVP_WATER__Frequency_Analysis = "B3LYP/def2-TZVP/ISOSVP(WATER) Frequency Analysis"
    B3LYP_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_WATER__Unknown = "B3LYP/def2-TZVP/ISOSVP(WATER) Unknown"
    B3LYP_def2_TZVP_ISOSVP_THF__Single_Point = "B3LYP/def2-TZVP/ISOSVP(THF) Single Point"
    B3LYP_def2_TZVP_ISOSVP_THF__Force = "B3LYP/def2-TZVP/ISOSVP(THF) Force"
    B3LYP_def2_TZVP_ISOSVP_THF__Geometry_Optimization = "B3LYP/def2-TZVP/ISOSVP(THF) Geometry Optimization"
    B3LYP_def2_TZVP_ISOSVP_THF__Frequency_Analysis = "B3LYP/def2-TZVP/ISOSVP(THF) Frequency Analysis"
    B3LYP_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_THF__Unknown = "B3LYP/def2-TZVP/ISOSVP(THF) Unknown"
    B3LYP_def2_TZVP_ISOSVP_G2__Single_Point = "B3LYP/def2-TZVP/ISOSVP(G2) Single Point"
    B3LYP_def2_TZVP_ISOSVP_G2__Force = "B3LYP/def2-TZVP/ISOSVP(G2) Force"
    B3LYP_def2_TZVP_ISOSVP_G2__Geometry_Optimization = "B3LYP/def2-TZVP/ISOSVP(G2) Geometry Optimization"
    B3LYP_def2_TZVP_ISOSVP_G2__Frequency_Analysis = "B3LYP/def2-TZVP/ISOSVP(G2) Frequency Analysis"
    B3LYP_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_G2__Unknown = "B3LYP/def2-TZVP/ISOSVP(G2) Unknown"
    B3LYP_def2_TZVP_ISOSVP_DME__Single_Point = "B3LYP/def2-TZVP/ISOSVP(DME) Single Point"
    B3LYP_def2_TZVP_ISOSVP_DME__Force = "B3LYP/def2-TZVP/ISOSVP(DME) Force"
    B3LYP_def2_TZVP_ISOSVP_DME__Geometry_Optimization = "B3LYP/def2-TZVP/ISOSVP(DME) Geometry Optimization"
    B3LYP_def2_TZVP_ISOSVP_DME__Frequency_Analysis = "B3LYP/def2-TZVP/ISOSVP(DME) Frequency Analysis"
    B3LYP_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_DME__Unknown = "B3LYP/def2-TZVP/ISOSVP(DME) Unknown"
    B3LYP_def2_TZVP_ISOSVP_EC__Single_Point = "B3LYP/def2-TZVP/ISOSVP(EC) Single Point"
    B3LYP_def2_TZVP_ISOSVP_EC__Force = "B3LYP/def2-TZVP/ISOSVP(EC) Force"
    B3LYP_def2_TZVP_ISOSVP_EC__Geometry_Optimization = "B3LYP/def2-TZVP/ISOSVP(EC) Geometry Optimization"
    B3LYP_def2_TZVP_ISOSVP_EC__Frequency_Analysis = "B3LYP/def2-TZVP/ISOSVP(EC) Frequency Analysis"
    B3LYP_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_EC__Unknown = "B3LYP/def2-TZVP/ISOSVP(EC) Unknown"
    B3LYP_def2_TZVP_ISOSVP_EC_EMC__Single_Point = "B3LYP/def2-TZVP/ISOSVP(EC/EMC) Single Point"
    B3LYP_def2_TZVP_ISOSVP_EC_EMC__Force = "B3LYP/def2-TZVP/ISOSVP(EC/EMC) Force"
    B3LYP_def2_TZVP_ISOSVP_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVP/ISOSVP(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVP_ISOSVP_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVP/ISOSVP(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_EC_EMC__Unknown = "B3LYP/def2-TZVP/ISOSVP(EC/EMC) Unknown"
    B3LYP_def2_TZVP_ISOSVP_PC__Single_Point = "B3LYP/def2-TZVP/ISOSVP(PC) Single Point"
    B3LYP_def2_TZVP_ISOSVP_PC__Force = "B3LYP/def2-TZVP/ISOSVP(PC) Force"
    B3LYP_def2_TZVP_ISOSVP_PC__Geometry_Optimization = "B3LYP/def2-TZVP/ISOSVP(PC) Geometry Optimization"
    B3LYP_def2_TZVP_ISOSVP_PC__Frequency_Analysis = "B3LYP/def2-TZVP/ISOSVP(PC) Frequency Analysis"
    B3LYP_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_ISOSVP_PC__Unknown = "B3LYP/def2-TZVP/ISOSVP(PC) Unknown"
    B3LYP_def2_TZVP_SMD_WATER__Single_Point = "B3LYP/def2-TZVP/SMD(WATER) Single Point"
    B3LYP_def2_TZVP_SMD_WATER__Force = "B3LYP/def2-TZVP/SMD(WATER) Force"
    B3LYP_def2_TZVP_SMD_WATER__Geometry_Optimization = "B3LYP/def2-TZVP/SMD(WATER) Geometry Optimization"
    B3LYP_def2_TZVP_SMD_WATER__Frequency_Analysis = "B3LYP/def2-TZVP/SMD(WATER) Frequency Analysis"
    B3LYP_def2_TZVP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_WATER__Unknown = "B3LYP/def2-TZVP/SMD(WATER) Unknown"
    B3LYP_def2_TZVP_SMD_THF__Single_Point = "B3LYP/def2-TZVP/SMD(THF) Single Point"
    B3LYP_def2_TZVP_SMD_THF__Force = "B3LYP/def2-TZVP/SMD(THF) Force"
    B3LYP_def2_TZVP_SMD_THF__Geometry_Optimization = "B3LYP/def2-TZVP/SMD(THF) Geometry Optimization"
    B3LYP_def2_TZVP_SMD_THF__Frequency_Analysis = "B3LYP/def2-TZVP/SMD(THF) Frequency Analysis"
    B3LYP_def2_TZVP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_THF__Unknown = "B3LYP/def2-TZVP/SMD(THF) Unknown"
    B3LYP_def2_TZVP_SMD_G2__Single_Point = "B3LYP/def2-TZVP/SMD(G2) Single Point"
    B3LYP_def2_TZVP_SMD_G2__Force = "B3LYP/def2-TZVP/SMD(G2) Force"
    B3LYP_def2_TZVP_SMD_G2__Geometry_Optimization = "B3LYP/def2-TZVP/SMD(G2) Geometry Optimization"
    B3LYP_def2_TZVP_SMD_G2__Frequency_Analysis = "B3LYP/def2-TZVP/SMD(G2) Frequency Analysis"
    B3LYP_def2_TZVP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_G2__Unknown = "B3LYP/def2-TZVP/SMD(G2) Unknown"
    B3LYP_def2_TZVP_SMD_DME__Single_Point = "B3LYP/def2-TZVP/SMD(DME) Single Point"
    B3LYP_def2_TZVP_SMD_DME__Force = "B3LYP/def2-TZVP/SMD(DME) Force"
    B3LYP_def2_TZVP_SMD_DME__Geometry_Optimization = "B3LYP/def2-TZVP/SMD(DME) Geometry Optimization"
    B3LYP_def2_TZVP_SMD_DME__Frequency_Analysis = "B3LYP/def2-TZVP/SMD(DME) Frequency Analysis"
    B3LYP_def2_TZVP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_DME__Unknown = "B3LYP/def2-TZVP/SMD(DME) Unknown"
    B3LYP_def2_TZVP_SMD_EC__Single_Point = "B3LYP/def2-TZVP/SMD(EC) Single Point"
    B3LYP_def2_TZVP_SMD_EC__Force = "B3LYP/def2-TZVP/SMD(EC) Force"
    B3LYP_def2_TZVP_SMD_EC__Geometry_Optimization = "B3LYP/def2-TZVP/SMD(EC) Geometry Optimization"
    B3LYP_def2_TZVP_SMD_EC__Frequency_Analysis = "B3LYP/def2-TZVP/SMD(EC) Frequency Analysis"
    B3LYP_def2_TZVP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_EC__Unknown = "B3LYP/def2-TZVP/SMD(EC) Unknown"
    B3LYP_def2_TZVP_SMD_EC_EMC__Single_Point = "B3LYP/def2-TZVP/SMD(EC/EMC) Single Point"
    B3LYP_def2_TZVP_SMD_EC_EMC__Force = "B3LYP/def2-TZVP/SMD(EC/EMC) Force"
    B3LYP_def2_TZVP_SMD_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVP/SMD(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVP_SMD_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVP/SMD(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_EC_EMC__Unknown = "B3LYP/def2-TZVP/SMD(EC/EMC) Unknown"
    B3LYP_def2_TZVP_SMD_PC__Single_Point = "B3LYP/def2-TZVP/SMD(PC) Single Point"
    B3LYP_def2_TZVP_SMD_PC__Force = "B3LYP/def2-TZVP/SMD(PC) Force"
    B3LYP_def2_TZVP_SMD_PC__Geometry_Optimization = "B3LYP/def2-TZVP/SMD(PC) Geometry Optimization"
    B3LYP_def2_TZVP_SMD_PC__Frequency_Analysis = "B3LYP/def2-TZVP/SMD(PC) Frequency Analysis"
    B3LYP_def2_TZVP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_SMD_PC__Unknown = "B3LYP/def2-TZVP/SMD(PC) Unknown"
    B3LYP_def2_TZVP_CMIRS_WATER__Single_Point = "B3LYP/def2-TZVP/CMIRS(WATER) Single Point"
    B3LYP_def2_TZVP_CMIRS_WATER__Force = "B3LYP/def2-TZVP/CMIRS(WATER) Force"
    B3LYP_def2_TZVP_CMIRS_WATER__Geometry_Optimization = "B3LYP/def2-TZVP/CMIRS(WATER) Geometry Optimization"
    B3LYP_def2_TZVP_CMIRS_WATER__Frequency_Analysis = "B3LYP/def2-TZVP/CMIRS(WATER) Frequency Analysis"
    B3LYP_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_WATER__Unknown = "B3LYP/def2-TZVP/CMIRS(WATER) Unknown"
    B3LYP_def2_TZVP_CMIRS_THF__Single_Point = "B3LYP/def2-TZVP/CMIRS(THF) Single Point"
    B3LYP_def2_TZVP_CMIRS_THF__Force = "B3LYP/def2-TZVP/CMIRS(THF) Force"
    B3LYP_def2_TZVP_CMIRS_THF__Geometry_Optimization = "B3LYP/def2-TZVP/CMIRS(THF) Geometry Optimization"
    B3LYP_def2_TZVP_CMIRS_THF__Frequency_Analysis = "B3LYP/def2-TZVP/CMIRS(THF) Frequency Analysis"
    B3LYP_def2_TZVP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_THF__Unknown = "B3LYP/def2-TZVP/CMIRS(THF) Unknown"
    B3LYP_def2_TZVP_CMIRS_G2__Single_Point = "B3LYP/def2-TZVP/CMIRS(G2) Single Point"
    B3LYP_def2_TZVP_CMIRS_G2__Force = "B3LYP/def2-TZVP/CMIRS(G2) Force"
    B3LYP_def2_TZVP_CMIRS_G2__Geometry_Optimization = "B3LYP/def2-TZVP/CMIRS(G2) Geometry Optimization"
    B3LYP_def2_TZVP_CMIRS_G2__Frequency_Analysis = "B3LYP/def2-TZVP/CMIRS(G2) Frequency Analysis"
    B3LYP_def2_TZVP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_G2__Unknown = "B3LYP/def2-TZVP/CMIRS(G2) Unknown"
    B3LYP_def2_TZVP_CMIRS_DME__Single_Point = "B3LYP/def2-TZVP/CMIRS(DME) Single Point"
    B3LYP_def2_TZVP_CMIRS_DME__Force = "B3LYP/def2-TZVP/CMIRS(DME) Force"
    B3LYP_def2_TZVP_CMIRS_DME__Geometry_Optimization = "B3LYP/def2-TZVP/CMIRS(DME) Geometry Optimization"
    B3LYP_def2_TZVP_CMIRS_DME__Frequency_Analysis = "B3LYP/def2-TZVP/CMIRS(DME) Frequency Analysis"
    B3LYP_def2_TZVP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_DME__Unknown = "B3LYP/def2-TZVP/CMIRS(DME) Unknown"
    B3LYP_def2_TZVP_CMIRS_EC__Single_Point = "B3LYP/def2-TZVP/CMIRS(EC) Single Point"
    B3LYP_def2_TZVP_CMIRS_EC__Force = "B3LYP/def2-TZVP/CMIRS(EC) Force"
    B3LYP_def2_TZVP_CMIRS_EC__Geometry_Optimization = "B3LYP/def2-TZVP/CMIRS(EC) Geometry Optimization"
    B3LYP_def2_TZVP_CMIRS_EC__Frequency_Analysis = "B3LYP/def2-TZVP/CMIRS(EC) Frequency Analysis"
    B3LYP_def2_TZVP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_EC__Unknown = "B3LYP/def2-TZVP/CMIRS(EC) Unknown"
    B3LYP_def2_TZVP_CMIRS_EC_EMC__Single_Point = "B3LYP/def2-TZVP/CMIRS(EC/EMC) Single Point"
    B3LYP_def2_TZVP_CMIRS_EC_EMC__Force = "B3LYP/def2-TZVP/CMIRS(EC/EMC) Force"
    B3LYP_def2_TZVP_CMIRS_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVP/CMIRS(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVP_CMIRS_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVP/CMIRS(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_EC_EMC__Unknown = "B3LYP/def2-TZVP/CMIRS(EC/EMC) Unknown"
    B3LYP_def2_TZVP_CMIRS_PC__Single_Point = "B3LYP/def2-TZVP/CMIRS(PC) Single Point"
    B3LYP_def2_TZVP_CMIRS_PC__Force = "B3LYP/def2-TZVP/CMIRS(PC) Force"
    B3LYP_def2_TZVP_CMIRS_PC__Geometry_Optimization = "B3LYP/def2-TZVP/CMIRS(PC) Geometry Optimization"
    B3LYP_def2_TZVP_CMIRS_PC__Frequency_Analysis = "B3LYP/def2-TZVP/CMIRS(PC) Frequency Analysis"
    B3LYP_def2_TZVP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVP_CMIRS_PC__Unknown = "B3LYP/def2-TZVP/CMIRS(PC) Unknown"
    B3LYP_def2_TZVPD_VACUUM_Single_Point = "B3LYP/def2-TZVPD/VACUUM Single Point"
    B3LYP_def2_TZVPD_VACUUM_Force = "B3LYP/def2-TZVPD/VACUUM Force"
    B3LYP_def2_TZVPD_VACUUM_Geometry_Optimization = "B3LYP/def2-TZVPD/VACUUM Geometry Optimization"
    B3LYP_def2_TZVPD_VACUUM_Frequency_Analysis = "B3LYP/def2-TZVPD/VACUUM Frequency Analysis"
    B3LYP_def2_TZVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/VACUUM Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_VACUUM_Unknown = "B3LYP/def2-TZVPD/VACUUM Unknown"
    B3LYP_def2_TZVPD_PCM_WATER__Single_Point = "B3LYP/def2-TZVPD/PCM(WATER) Single Point"
    B3LYP_def2_TZVPD_PCM_WATER__Force = "B3LYP/def2-TZVPD/PCM(WATER) Force"
    B3LYP_def2_TZVPD_PCM_WATER__Geometry_Optimization = "B3LYP/def2-TZVPD/PCM(WATER) Geometry Optimization"
    B3LYP_def2_TZVPD_PCM_WATER__Frequency_Analysis = "B3LYP/def2-TZVPD/PCM(WATER) Frequency Analysis"
    B3LYP_def2_TZVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_WATER__Unknown = "B3LYP/def2-TZVPD/PCM(WATER) Unknown"
    B3LYP_def2_TZVPD_PCM_THF__Single_Point = "B3LYP/def2-TZVPD/PCM(THF) Single Point"
    B3LYP_def2_TZVPD_PCM_THF__Force = "B3LYP/def2-TZVPD/PCM(THF) Force"
    B3LYP_def2_TZVPD_PCM_THF__Geometry_Optimization = "B3LYP/def2-TZVPD/PCM(THF) Geometry Optimization"
    B3LYP_def2_TZVPD_PCM_THF__Frequency_Analysis = "B3LYP/def2-TZVPD/PCM(THF) Frequency Analysis"
    B3LYP_def2_TZVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_THF__Unknown = "B3LYP/def2-TZVPD/PCM(THF) Unknown"
    B3LYP_def2_TZVPD_PCM_G2__Single_Point = "B3LYP/def2-TZVPD/PCM(G2) Single Point"
    B3LYP_def2_TZVPD_PCM_G2__Force = "B3LYP/def2-TZVPD/PCM(G2) Force"
    B3LYP_def2_TZVPD_PCM_G2__Geometry_Optimization = "B3LYP/def2-TZVPD/PCM(G2) Geometry Optimization"
    B3LYP_def2_TZVPD_PCM_G2__Frequency_Analysis = "B3LYP/def2-TZVPD/PCM(G2) Frequency Analysis"
    B3LYP_def2_TZVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_G2__Unknown = "B3LYP/def2-TZVPD/PCM(G2) Unknown"
    B3LYP_def2_TZVPD_PCM_DME__Single_Point = "B3LYP/def2-TZVPD/PCM(DME) Single Point"
    B3LYP_def2_TZVPD_PCM_DME__Force = "B3LYP/def2-TZVPD/PCM(DME) Force"
    B3LYP_def2_TZVPD_PCM_DME__Geometry_Optimization = "B3LYP/def2-TZVPD/PCM(DME) Geometry Optimization"
    B3LYP_def2_TZVPD_PCM_DME__Frequency_Analysis = "B3LYP/def2-TZVPD/PCM(DME) Frequency Analysis"
    B3LYP_def2_TZVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_DME__Unknown = "B3LYP/def2-TZVPD/PCM(DME) Unknown"
    B3LYP_def2_TZVPD_PCM_EC__Single_Point = "B3LYP/def2-TZVPD/PCM(EC) Single Point"
    B3LYP_def2_TZVPD_PCM_EC__Force = "B3LYP/def2-TZVPD/PCM(EC) Force"
    B3LYP_def2_TZVPD_PCM_EC__Geometry_Optimization = "B3LYP/def2-TZVPD/PCM(EC) Geometry Optimization"
    B3LYP_def2_TZVPD_PCM_EC__Frequency_Analysis = "B3LYP/def2-TZVPD/PCM(EC) Frequency Analysis"
    B3LYP_def2_TZVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_EC__Unknown = "B3LYP/def2-TZVPD/PCM(EC) Unknown"
    B3LYP_def2_TZVPD_PCM_EC_EMC__Single_Point = "B3LYP/def2-TZVPD/PCM(EC/EMC) Single Point"
    B3LYP_def2_TZVPD_PCM_EC_EMC__Force = "B3LYP/def2-TZVPD/PCM(EC/EMC) Force"
    B3LYP_def2_TZVPD_PCM_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPD/PCM(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPD_PCM_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPD/PCM(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_EC_EMC__Unknown = "B3LYP/def2-TZVPD/PCM(EC/EMC) Unknown"
    B3LYP_def2_TZVPD_PCM_PC__Single_Point = "B3LYP/def2-TZVPD/PCM(PC) Single Point"
    B3LYP_def2_TZVPD_PCM_PC__Force = "B3LYP/def2-TZVPD/PCM(PC) Force"
    B3LYP_def2_TZVPD_PCM_PC__Geometry_Optimization = "B3LYP/def2-TZVPD/PCM(PC) Geometry Optimization"
    B3LYP_def2_TZVPD_PCM_PC__Frequency_Analysis = "B3LYP/def2-TZVPD/PCM(PC) Frequency Analysis"
    B3LYP_def2_TZVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_PCM_PC__Unknown = "B3LYP/def2-TZVPD/PCM(PC) Unknown"
    B3LYP_def2_TZVPD_ISOSVP_WATER__Single_Point = "B3LYP/def2-TZVPD/ISOSVP(WATER) Single Point"
    B3LYP_def2_TZVPD_ISOSVP_WATER__Force = "B3LYP/def2-TZVPD/ISOSVP(WATER) Force"
    B3LYP_def2_TZVPD_ISOSVP_WATER__Geometry_Optimization = "B3LYP/def2-TZVPD/ISOSVP(WATER) Geometry Optimization"
    B3LYP_def2_TZVPD_ISOSVP_WATER__Frequency_Analysis = "B3LYP/def2-TZVPD/ISOSVP(WATER) Frequency Analysis"
    B3LYP_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_WATER__Unknown = "B3LYP/def2-TZVPD/ISOSVP(WATER) Unknown"
    B3LYP_def2_TZVPD_ISOSVP_THF__Single_Point = "B3LYP/def2-TZVPD/ISOSVP(THF) Single Point"
    B3LYP_def2_TZVPD_ISOSVP_THF__Force = "B3LYP/def2-TZVPD/ISOSVP(THF) Force"
    B3LYP_def2_TZVPD_ISOSVP_THF__Geometry_Optimization = "B3LYP/def2-TZVPD/ISOSVP(THF) Geometry Optimization"
    B3LYP_def2_TZVPD_ISOSVP_THF__Frequency_Analysis = "B3LYP/def2-TZVPD/ISOSVP(THF) Frequency Analysis"
    B3LYP_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_THF__Unknown = "B3LYP/def2-TZVPD/ISOSVP(THF) Unknown"
    B3LYP_def2_TZVPD_ISOSVP_G2__Single_Point = "B3LYP/def2-TZVPD/ISOSVP(G2) Single Point"
    B3LYP_def2_TZVPD_ISOSVP_G2__Force = "B3LYP/def2-TZVPD/ISOSVP(G2) Force"
    B3LYP_def2_TZVPD_ISOSVP_G2__Geometry_Optimization = "B3LYP/def2-TZVPD/ISOSVP(G2) Geometry Optimization"
    B3LYP_def2_TZVPD_ISOSVP_G2__Frequency_Analysis = "B3LYP/def2-TZVPD/ISOSVP(G2) Frequency Analysis"
    B3LYP_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_G2__Unknown = "B3LYP/def2-TZVPD/ISOSVP(G2) Unknown"
    B3LYP_def2_TZVPD_ISOSVP_DME__Single_Point = "B3LYP/def2-TZVPD/ISOSVP(DME) Single Point"
    B3LYP_def2_TZVPD_ISOSVP_DME__Force = "B3LYP/def2-TZVPD/ISOSVP(DME) Force"
    B3LYP_def2_TZVPD_ISOSVP_DME__Geometry_Optimization = "B3LYP/def2-TZVPD/ISOSVP(DME) Geometry Optimization"
    B3LYP_def2_TZVPD_ISOSVP_DME__Frequency_Analysis = "B3LYP/def2-TZVPD/ISOSVP(DME) Frequency Analysis"
    B3LYP_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_DME__Unknown = "B3LYP/def2-TZVPD/ISOSVP(DME) Unknown"
    B3LYP_def2_TZVPD_ISOSVP_EC__Single_Point = "B3LYP/def2-TZVPD/ISOSVP(EC) Single Point"
    B3LYP_def2_TZVPD_ISOSVP_EC__Force = "B3LYP/def2-TZVPD/ISOSVP(EC) Force"
    B3LYP_def2_TZVPD_ISOSVP_EC__Geometry_Optimization = "B3LYP/def2-TZVPD/ISOSVP(EC) Geometry Optimization"
    B3LYP_def2_TZVPD_ISOSVP_EC__Frequency_Analysis = "B3LYP/def2-TZVPD/ISOSVP(EC) Frequency Analysis"
    B3LYP_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_EC__Unknown = "B3LYP/def2-TZVPD/ISOSVP(EC) Unknown"
    B3LYP_def2_TZVPD_ISOSVP_EC_EMC__Single_Point = "B3LYP/def2-TZVPD/ISOSVP(EC/EMC) Single Point"
    B3LYP_def2_TZVPD_ISOSVP_EC_EMC__Force = "B3LYP/def2-TZVPD/ISOSVP(EC/EMC) Force"
    B3LYP_def2_TZVPD_ISOSVP_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPD/ISOSVP(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPD/ISOSVP(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_EC_EMC__Unknown = "B3LYP/def2-TZVPD/ISOSVP(EC/EMC) Unknown"
    B3LYP_def2_TZVPD_ISOSVP_PC__Single_Point = "B3LYP/def2-TZVPD/ISOSVP(PC) Single Point"
    B3LYP_def2_TZVPD_ISOSVP_PC__Force = "B3LYP/def2-TZVPD/ISOSVP(PC) Force"
    B3LYP_def2_TZVPD_ISOSVP_PC__Geometry_Optimization = "B3LYP/def2-TZVPD/ISOSVP(PC) Geometry Optimization"
    B3LYP_def2_TZVPD_ISOSVP_PC__Frequency_Analysis = "B3LYP/def2-TZVPD/ISOSVP(PC) Frequency Analysis"
    B3LYP_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_ISOSVP_PC__Unknown = "B3LYP/def2-TZVPD/ISOSVP(PC) Unknown"
    B3LYP_def2_TZVPD_SMD_WATER__Single_Point = "B3LYP/def2-TZVPD/SMD(WATER) Single Point"
    B3LYP_def2_TZVPD_SMD_WATER__Force = "B3LYP/def2-TZVPD/SMD(WATER) Force"
    B3LYP_def2_TZVPD_SMD_WATER__Geometry_Optimization = "B3LYP/def2-TZVPD/SMD(WATER) Geometry Optimization"
    B3LYP_def2_TZVPD_SMD_WATER__Frequency_Analysis = "B3LYP/def2-TZVPD/SMD(WATER) Frequency Analysis"
    B3LYP_def2_TZVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_WATER__Unknown = "B3LYP/def2-TZVPD/SMD(WATER) Unknown"
    B3LYP_def2_TZVPD_SMD_THF__Single_Point = "B3LYP/def2-TZVPD/SMD(THF) Single Point"
    B3LYP_def2_TZVPD_SMD_THF__Force = "B3LYP/def2-TZVPD/SMD(THF) Force"
    B3LYP_def2_TZVPD_SMD_THF__Geometry_Optimization = "B3LYP/def2-TZVPD/SMD(THF) Geometry Optimization"
    B3LYP_def2_TZVPD_SMD_THF__Frequency_Analysis = "B3LYP/def2-TZVPD/SMD(THF) Frequency Analysis"
    B3LYP_def2_TZVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_THF__Unknown = "B3LYP/def2-TZVPD/SMD(THF) Unknown"
    B3LYP_def2_TZVPD_SMD_G2__Single_Point = "B3LYP/def2-TZVPD/SMD(G2) Single Point"
    B3LYP_def2_TZVPD_SMD_G2__Force = "B3LYP/def2-TZVPD/SMD(G2) Force"
    B3LYP_def2_TZVPD_SMD_G2__Geometry_Optimization = "B3LYP/def2-TZVPD/SMD(G2) Geometry Optimization"
    B3LYP_def2_TZVPD_SMD_G2__Frequency_Analysis = "B3LYP/def2-TZVPD/SMD(G2) Frequency Analysis"
    B3LYP_def2_TZVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_G2__Unknown = "B3LYP/def2-TZVPD/SMD(G2) Unknown"
    B3LYP_def2_TZVPD_SMD_DME__Single_Point = "B3LYP/def2-TZVPD/SMD(DME) Single Point"
    B3LYP_def2_TZVPD_SMD_DME__Force = "B3LYP/def2-TZVPD/SMD(DME) Force"
    B3LYP_def2_TZVPD_SMD_DME__Geometry_Optimization = "B3LYP/def2-TZVPD/SMD(DME) Geometry Optimization"
    B3LYP_def2_TZVPD_SMD_DME__Frequency_Analysis = "B3LYP/def2-TZVPD/SMD(DME) Frequency Analysis"
    B3LYP_def2_TZVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_DME__Unknown = "B3LYP/def2-TZVPD/SMD(DME) Unknown"
    B3LYP_def2_TZVPD_SMD_EC__Single_Point = "B3LYP/def2-TZVPD/SMD(EC) Single Point"
    B3LYP_def2_TZVPD_SMD_EC__Force = "B3LYP/def2-TZVPD/SMD(EC) Force"
    B3LYP_def2_TZVPD_SMD_EC__Geometry_Optimization = "B3LYP/def2-TZVPD/SMD(EC) Geometry Optimization"
    B3LYP_def2_TZVPD_SMD_EC__Frequency_Analysis = "B3LYP/def2-TZVPD/SMD(EC) Frequency Analysis"
    B3LYP_def2_TZVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_EC__Unknown = "B3LYP/def2-TZVPD/SMD(EC) Unknown"
    B3LYP_def2_TZVPD_SMD_EC_EMC__Single_Point = "B3LYP/def2-TZVPD/SMD(EC/EMC) Single Point"
    B3LYP_def2_TZVPD_SMD_EC_EMC__Force = "B3LYP/def2-TZVPD/SMD(EC/EMC) Force"
    B3LYP_def2_TZVPD_SMD_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPD/SMD(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPD_SMD_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPD/SMD(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_EC_EMC__Unknown = "B3LYP/def2-TZVPD/SMD(EC/EMC) Unknown"
    B3LYP_def2_TZVPD_SMD_PC__Single_Point = "B3LYP/def2-TZVPD/SMD(PC) Single Point"
    B3LYP_def2_TZVPD_SMD_PC__Force = "B3LYP/def2-TZVPD/SMD(PC) Force"
    B3LYP_def2_TZVPD_SMD_PC__Geometry_Optimization = "B3LYP/def2-TZVPD/SMD(PC) Geometry Optimization"
    B3LYP_def2_TZVPD_SMD_PC__Frequency_Analysis = "B3LYP/def2-TZVPD/SMD(PC) Frequency Analysis"
    B3LYP_def2_TZVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_SMD_PC__Unknown = "B3LYP/def2-TZVPD/SMD(PC) Unknown"
    B3LYP_def2_TZVPD_CMIRS_WATER__Single_Point = "B3LYP/def2-TZVPD/CMIRS(WATER) Single Point"
    B3LYP_def2_TZVPD_CMIRS_WATER__Force = "B3LYP/def2-TZVPD/CMIRS(WATER) Force"
    B3LYP_def2_TZVPD_CMIRS_WATER__Geometry_Optimization = "B3LYP/def2-TZVPD/CMIRS(WATER) Geometry Optimization"
    B3LYP_def2_TZVPD_CMIRS_WATER__Frequency_Analysis = "B3LYP/def2-TZVPD/CMIRS(WATER) Frequency Analysis"
    B3LYP_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_WATER__Unknown = "B3LYP/def2-TZVPD/CMIRS(WATER) Unknown"
    B3LYP_def2_TZVPD_CMIRS_THF__Single_Point = "B3LYP/def2-TZVPD/CMIRS(THF) Single Point"
    B3LYP_def2_TZVPD_CMIRS_THF__Force = "B3LYP/def2-TZVPD/CMIRS(THF) Force"
    B3LYP_def2_TZVPD_CMIRS_THF__Geometry_Optimization = "B3LYP/def2-TZVPD/CMIRS(THF) Geometry Optimization"
    B3LYP_def2_TZVPD_CMIRS_THF__Frequency_Analysis = "B3LYP/def2-TZVPD/CMIRS(THF) Frequency Analysis"
    B3LYP_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_THF__Unknown = "B3LYP/def2-TZVPD/CMIRS(THF) Unknown"
    B3LYP_def2_TZVPD_CMIRS_G2__Single_Point = "B3LYP/def2-TZVPD/CMIRS(G2) Single Point"
    B3LYP_def2_TZVPD_CMIRS_G2__Force = "B3LYP/def2-TZVPD/CMIRS(G2) Force"
    B3LYP_def2_TZVPD_CMIRS_G2__Geometry_Optimization = "B3LYP/def2-TZVPD/CMIRS(G2) Geometry Optimization"
    B3LYP_def2_TZVPD_CMIRS_G2__Frequency_Analysis = "B3LYP/def2-TZVPD/CMIRS(G2) Frequency Analysis"
    B3LYP_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_G2__Unknown = "B3LYP/def2-TZVPD/CMIRS(G2) Unknown"
    B3LYP_def2_TZVPD_CMIRS_DME__Single_Point = "B3LYP/def2-TZVPD/CMIRS(DME) Single Point"
    B3LYP_def2_TZVPD_CMIRS_DME__Force = "B3LYP/def2-TZVPD/CMIRS(DME) Force"
    B3LYP_def2_TZVPD_CMIRS_DME__Geometry_Optimization = "B3LYP/def2-TZVPD/CMIRS(DME) Geometry Optimization"
    B3LYP_def2_TZVPD_CMIRS_DME__Frequency_Analysis = "B3LYP/def2-TZVPD/CMIRS(DME) Frequency Analysis"
    B3LYP_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_DME__Unknown = "B3LYP/def2-TZVPD/CMIRS(DME) Unknown"
    B3LYP_def2_TZVPD_CMIRS_EC__Single_Point = "B3LYP/def2-TZVPD/CMIRS(EC) Single Point"
    B3LYP_def2_TZVPD_CMIRS_EC__Force = "B3LYP/def2-TZVPD/CMIRS(EC) Force"
    B3LYP_def2_TZVPD_CMIRS_EC__Geometry_Optimization = "B3LYP/def2-TZVPD/CMIRS(EC) Geometry Optimization"
    B3LYP_def2_TZVPD_CMIRS_EC__Frequency_Analysis = "B3LYP/def2-TZVPD/CMIRS(EC) Frequency Analysis"
    B3LYP_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_EC__Unknown = "B3LYP/def2-TZVPD/CMIRS(EC) Unknown"
    B3LYP_def2_TZVPD_CMIRS_EC_EMC__Single_Point = "B3LYP/def2-TZVPD/CMIRS(EC/EMC) Single Point"
    B3LYP_def2_TZVPD_CMIRS_EC_EMC__Force = "B3LYP/def2-TZVPD/CMIRS(EC/EMC) Force"
    B3LYP_def2_TZVPD_CMIRS_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPD/CMIRS(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPD_CMIRS_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPD/CMIRS(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_EC_EMC__Unknown = "B3LYP/def2-TZVPD/CMIRS(EC/EMC) Unknown"
    B3LYP_def2_TZVPD_CMIRS_PC__Single_Point = "B3LYP/def2-TZVPD/CMIRS(PC) Single Point"
    B3LYP_def2_TZVPD_CMIRS_PC__Force = "B3LYP/def2-TZVPD/CMIRS(PC) Force"
    B3LYP_def2_TZVPD_CMIRS_PC__Geometry_Optimization = "B3LYP/def2-TZVPD/CMIRS(PC) Geometry Optimization"
    B3LYP_def2_TZVPD_CMIRS_PC__Frequency_Analysis = "B3LYP/def2-TZVPD/CMIRS(PC) Frequency Analysis"
    B3LYP_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPD_CMIRS_PC__Unknown = "B3LYP/def2-TZVPD/CMIRS(PC) Unknown"
    B3LYP_def2_TZVPP_VACUUM_Single_Point = "B3LYP/def2-TZVPP/VACUUM Single Point"
    B3LYP_def2_TZVPP_VACUUM_Force = "B3LYP/def2-TZVPP/VACUUM Force"
    B3LYP_def2_TZVPP_VACUUM_Geometry_Optimization = "B3LYP/def2-TZVPP/VACUUM Geometry Optimization"
    B3LYP_def2_TZVPP_VACUUM_Frequency_Analysis = "B3LYP/def2-TZVPP/VACUUM Frequency Analysis"
    B3LYP_def2_TZVPP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/VACUUM Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_VACUUM_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/VACUUM Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_VACUUM_Unknown = "B3LYP/def2-TZVPP/VACUUM Unknown"
    B3LYP_def2_TZVPP_PCM_WATER__Single_Point = "B3LYP/def2-TZVPP/PCM(WATER) Single Point"
    B3LYP_def2_TZVPP_PCM_WATER__Force = "B3LYP/def2-TZVPP/PCM(WATER) Force"
    B3LYP_def2_TZVPP_PCM_WATER__Geometry_Optimization = "B3LYP/def2-TZVPP/PCM(WATER) Geometry Optimization"
    B3LYP_def2_TZVPP_PCM_WATER__Frequency_Analysis = "B3LYP/def2-TZVPP/PCM(WATER) Frequency Analysis"
    B3LYP_def2_TZVPP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_WATER__Unknown = "B3LYP/def2-TZVPP/PCM(WATER) Unknown"
    B3LYP_def2_TZVPP_PCM_THF__Single_Point = "B3LYP/def2-TZVPP/PCM(THF) Single Point"
    B3LYP_def2_TZVPP_PCM_THF__Force = "B3LYP/def2-TZVPP/PCM(THF) Force"
    B3LYP_def2_TZVPP_PCM_THF__Geometry_Optimization = "B3LYP/def2-TZVPP/PCM(THF) Geometry Optimization"
    B3LYP_def2_TZVPP_PCM_THF__Frequency_Analysis = "B3LYP/def2-TZVPP/PCM(THF) Frequency Analysis"
    B3LYP_def2_TZVPP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_THF__Unknown = "B3LYP/def2-TZVPP/PCM(THF) Unknown"
    B3LYP_def2_TZVPP_PCM_G2__Single_Point = "B3LYP/def2-TZVPP/PCM(G2) Single Point"
    B3LYP_def2_TZVPP_PCM_G2__Force = "B3LYP/def2-TZVPP/PCM(G2) Force"
    B3LYP_def2_TZVPP_PCM_G2__Geometry_Optimization = "B3LYP/def2-TZVPP/PCM(G2) Geometry Optimization"
    B3LYP_def2_TZVPP_PCM_G2__Frequency_Analysis = "B3LYP/def2-TZVPP/PCM(G2) Frequency Analysis"
    B3LYP_def2_TZVPP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_G2__Unknown = "B3LYP/def2-TZVPP/PCM(G2) Unknown"
    B3LYP_def2_TZVPP_PCM_DME__Single_Point = "B3LYP/def2-TZVPP/PCM(DME) Single Point"
    B3LYP_def2_TZVPP_PCM_DME__Force = "B3LYP/def2-TZVPP/PCM(DME) Force"
    B3LYP_def2_TZVPP_PCM_DME__Geometry_Optimization = "B3LYP/def2-TZVPP/PCM(DME) Geometry Optimization"
    B3LYP_def2_TZVPP_PCM_DME__Frequency_Analysis = "B3LYP/def2-TZVPP/PCM(DME) Frequency Analysis"
    B3LYP_def2_TZVPP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_DME__Unknown = "B3LYP/def2-TZVPP/PCM(DME) Unknown"
    B3LYP_def2_TZVPP_PCM_EC__Single_Point = "B3LYP/def2-TZVPP/PCM(EC) Single Point"
    B3LYP_def2_TZVPP_PCM_EC__Force = "B3LYP/def2-TZVPP/PCM(EC) Force"
    B3LYP_def2_TZVPP_PCM_EC__Geometry_Optimization = "B3LYP/def2-TZVPP/PCM(EC) Geometry Optimization"
    B3LYP_def2_TZVPP_PCM_EC__Frequency_Analysis = "B3LYP/def2-TZVPP/PCM(EC) Frequency Analysis"
    B3LYP_def2_TZVPP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_EC__Unknown = "B3LYP/def2-TZVPP/PCM(EC) Unknown"
    B3LYP_def2_TZVPP_PCM_EC_EMC__Single_Point = "B3LYP/def2-TZVPP/PCM(EC/EMC) Single Point"
    B3LYP_def2_TZVPP_PCM_EC_EMC__Force = "B3LYP/def2-TZVPP/PCM(EC/EMC) Force"
    B3LYP_def2_TZVPP_PCM_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPP/PCM(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPP_PCM_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPP/PCM(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_EC_EMC__Unknown = "B3LYP/def2-TZVPP/PCM(EC/EMC) Unknown"
    B3LYP_def2_TZVPP_PCM_PC__Single_Point = "B3LYP/def2-TZVPP/PCM(PC) Single Point"
    B3LYP_def2_TZVPP_PCM_PC__Force = "B3LYP/def2-TZVPP/PCM(PC) Force"
    B3LYP_def2_TZVPP_PCM_PC__Geometry_Optimization = "B3LYP/def2-TZVPP/PCM(PC) Geometry Optimization"
    B3LYP_def2_TZVPP_PCM_PC__Frequency_Analysis = "B3LYP/def2-TZVPP/PCM(PC) Frequency Analysis"
    B3LYP_def2_TZVPP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_PCM_PC__Unknown = "B3LYP/def2-TZVPP/PCM(PC) Unknown"
    B3LYP_def2_TZVPP_ISOSVP_WATER__Single_Point = "B3LYP/def2-TZVPP/ISOSVP(WATER) Single Point"
    B3LYP_def2_TZVPP_ISOSVP_WATER__Force = "B3LYP/def2-TZVPP/ISOSVP(WATER) Force"
    B3LYP_def2_TZVPP_ISOSVP_WATER__Geometry_Optimization = "B3LYP/def2-TZVPP/ISOSVP(WATER) Geometry Optimization"
    B3LYP_def2_TZVPP_ISOSVP_WATER__Frequency_Analysis = "B3LYP/def2-TZVPP/ISOSVP(WATER) Frequency Analysis"
    B3LYP_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_WATER__Unknown = "B3LYP/def2-TZVPP/ISOSVP(WATER) Unknown"
    B3LYP_def2_TZVPP_ISOSVP_THF__Single_Point = "B3LYP/def2-TZVPP/ISOSVP(THF) Single Point"
    B3LYP_def2_TZVPP_ISOSVP_THF__Force = "B3LYP/def2-TZVPP/ISOSVP(THF) Force"
    B3LYP_def2_TZVPP_ISOSVP_THF__Geometry_Optimization = "B3LYP/def2-TZVPP/ISOSVP(THF) Geometry Optimization"
    B3LYP_def2_TZVPP_ISOSVP_THF__Frequency_Analysis = "B3LYP/def2-TZVPP/ISOSVP(THF) Frequency Analysis"
    B3LYP_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_THF__Unknown = "B3LYP/def2-TZVPP/ISOSVP(THF) Unknown"
    B3LYP_def2_TZVPP_ISOSVP_G2__Single_Point = "B3LYP/def2-TZVPP/ISOSVP(G2) Single Point"
    B3LYP_def2_TZVPP_ISOSVP_G2__Force = "B3LYP/def2-TZVPP/ISOSVP(G2) Force"
    B3LYP_def2_TZVPP_ISOSVP_G2__Geometry_Optimization = "B3LYP/def2-TZVPP/ISOSVP(G2) Geometry Optimization"
    B3LYP_def2_TZVPP_ISOSVP_G2__Frequency_Analysis = "B3LYP/def2-TZVPP/ISOSVP(G2) Frequency Analysis"
    B3LYP_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_G2__Unknown = "B3LYP/def2-TZVPP/ISOSVP(G2) Unknown"
    B3LYP_def2_TZVPP_ISOSVP_DME__Single_Point = "B3LYP/def2-TZVPP/ISOSVP(DME) Single Point"
    B3LYP_def2_TZVPP_ISOSVP_DME__Force = "B3LYP/def2-TZVPP/ISOSVP(DME) Force"
    B3LYP_def2_TZVPP_ISOSVP_DME__Geometry_Optimization = "B3LYP/def2-TZVPP/ISOSVP(DME) Geometry Optimization"
    B3LYP_def2_TZVPP_ISOSVP_DME__Frequency_Analysis = "B3LYP/def2-TZVPP/ISOSVP(DME) Frequency Analysis"
    B3LYP_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_DME__Unknown = "B3LYP/def2-TZVPP/ISOSVP(DME) Unknown"
    B3LYP_def2_TZVPP_ISOSVP_EC__Single_Point = "B3LYP/def2-TZVPP/ISOSVP(EC) Single Point"
    B3LYP_def2_TZVPP_ISOSVP_EC__Force = "B3LYP/def2-TZVPP/ISOSVP(EC) Force"
    B3LYP_def2_TZVPP_ISOSVP_EC__Geometry_Optimization = "B3LYP/def2-TZVPP/ISOSVP(EC) Geometry Optimization"
    B3LYP_def2_TZVPP_ISOSVP_EC__Frequency_Analysis = "B3LYP/def2-TZVPP/ISOSVP(EC) Frequency Analysis"
    B3LYP_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_EC__Unknown = "B3LYP/def2-TZVPP/ISOSVP(EC) Unknown"
    B3LYP_def2_TZVPP_ISOSVP_EC_EMC__Single_Point = "B3LYP/def2-TZVPP/ISOSVP(EC/EMC) Single Point"
    B3LYP_def2_TZVPP_ISOSVP_EC_EMC__Force = "B3LYP/def2-TZVPP/ISOSVP(EC/EMC) Force"
    B3LYP_def2_TZVPP_ISOSVP_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPP/ISOSVP(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPP/ISOSVP(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_EC_EMC__Unknown = "B3LYP/def2-TZVPP/ISOSVP(EC/EMC) Unknown"
    B3LYP_def2_TZVPP_ISOSVP_PC__Single_Point = "B3LYP/def2-TZVPP/ISOSVP(PC) Single Point"
    B3LYP_def2_TZVPP_ISOSVP_PC__Force = "B3LYP/def2-TZVPP/ISOSVP(PC) Force"
    B3LYP_def2_TZVPP_ISOSVP_PC__Geometry_Optimization = "B3LYP/def2-TZVPP/ISOSVP(PC) Geometry Optimization"
    B3LYP_def2_TZVPP_ISOSVP_PC__Frequency_Analysis = "B3LYP/def2-TZVPP/ISOSVP(PC) Frequency Analysis"
    B3LYP_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_ISOSVP_PC__Unknown = "B3LYP/def2-TZVPP/ISOSVP(PC) Unknown"
    B3LYP_def2_TZVPP_SMD_WATER__Single_Point = "B3LYP/def2-TZVPP/SMD(WATER) Single Point"
    B3LYP_def2_TZVPP_SMD_WATER__Force = "B3LYP/def2-TZVPP/SMD(WATER) Force"
    B3LYP_def2_TZVPP_SMD_WATER__Geometry_Optimization = "B3LYP/def2-TZVPP/SMD(WATER) Geometry Optimization"
    B3LYP_def2_TZVPP_SMD_WATER__Frequency_Analysis = "B3LYP/def2-TZVPP/SMD(WATER) Frequency Analysis"
    B3LYP_def2_TZVPP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_WATER__Unknown = "B3LYP/def2-TZVPP/SMD(WATER) Unknown"
    B3LYP_def2_TZVPP_SMD_THF__Single_Point = "B3LYP/def2-TZVPP/SMD(THF) Single Point"
    B3LYP_def2_TZVPP_SMD_THF__Force = "B3LYP/def2-TZVPP/SMD(THF) Force"
    B3LYP_def2_TZVPP_SMD_THF__Geometry_Optimization = "B3LYP/def2-TZVPP/SMD(THF) Geometry Optimization"
    B3LYP_def2_TZVPP_SMD_THF__Frequency_Analysis = "B3LYP/def2-TZVPP/SMD(THF) Frequency Analysis"
    B3LYP_def2_TZVPP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_THF__Unknown = "B3LYP/def2-TZVPP/SMD(THF) Unknown"
    B3LYP_def2_TZVPP_SMD_G2__Single_Point = "B3LYP/def2-TZVPP/SMD(G2) Single Point"
    B3LYP_def2_TZVPP_SMD_G2__Force = "B3LYP/def2-TZVPP/SMD(G2) Force"
    B3LYP_def2_TZVPP_SMD_G2__Geometry_Optimization = "B3LYP/def2-TZVPP/SMD(G2) Geometry Optimization"
    B3LYP_def2_TZVPP_SMD_G2__Frequency_Analysis = "B3LYP/def2-TZVPP/SMD(G2) Frequency Analysis"
    B3LYP_def2_TZVPP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_G2__Unknown = "B3LYP/def2-TZVPP/SMD(G2) Unknown"
    B3LYP_def2_TZVPP_SMD_DME__Single_Point = "B3LYP/def2-TZVPP/SMD(DME) Single Point"
    B3LYP_def2_TZVPP_SMD_DME__Force = "B3LYP/def2-TZVPP/SMD(DME) Force"
    B3LYP_def2_TZVPP_SMD_DME__Geometry_Optimization = "B3LYP/def2-TZVPP/SMD(DME) Geometry Optimization"
    B3LYP_def2_TZVPP_SMD_DME__Frequency_Analysis = "B3LYP/def2-TZVPP/SMD(DME) Frequency Analysis"
    B3LYP_def2_TZVPP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_DME__Unknown = "B3LYP/def2-TZVPP/SMD(DME) Unknown"
    B3LYP_def2_TZVPP_SMD_EC__Single_Point = "B3LYP/def2-TZVPP/SMD(EC) Single Point"
    B3LYP_def2_TZVPP_SMD_EC__Force = "B3LYP/def2-TZVPP/SMD(EC) Force"
    B3LYP_def2_TZVPP_SMD_EC__Geometry_Optimization = "B3LYP/def2-TZVPP/SMD(EC) Geometry Optimization"
    B3LYP_def2_TZVPP_SMD_EC__Frequency_Analysis = "B3LYP/def2-TZVPP/SMD(EC) Frequency Analysis"
    B3LYP_def2_TZVPP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_EC__Unknown = "B3LYP/def2-TZVPP/SMD(EC) Unknown"
    B3LYP_def2_TZVPP_SMD_EC_EMC__Single_Point = "B3LYP/def2-TZVPP/SMD(EC/EMC) Single Point"
    B3LYP_def2_TZVPP_SMD_EC_EMC__Force = "B3LYP/def2-TZVPP/SMD(EC/EMC) Force"
    B3LYP_def2_TZVPP_SMD_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPP/SMD(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPP_SMD_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPP/SMD(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_EC_EMC__Unknown = "B3LYP/def2-TZVPP/SMD(EC/EMC) Unknown"
    B3LYP_def2_TZVPP_SMD_PC__Single_Point = "B3LYP/def2-TZVPP/SMD(PC) Single Point"
    B3LYP_def2_TZVPP_SMD_PC__Force = "B3LYP/def2-TZVPP/SMD(PC) Force"
    B3LYP_def2_TZVPP_SMD_PC__Geometry_Optimization = "B3LYP/def2-TZVPP/SMD(PC) Geometry Optimization"
    B3LYP_def2_TZVPP_SMD_PC__Frequency_Analysis = "B3LYP/def2-TZVPP/SMD(PC) Frequency Analysis"
    B3LYP_def2_TZVPP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_SMD_PC__Unknown = "B3LYP/def2-TZVPP/SMD(PC) Unknown"
    B3LYP_def2_TZVPP_CMIRS_WATER__Single_Point = "B3LYP/def2-TZVPP/CMIRS(WATER) Single Point"
    B3LYP_def2_TZVPP_CMIRS_WATER__Force = "B3LYP/def2-TZVPP/CMIRS(WATER) Force"
    B3LYP_def2_TZVPP_CMIRS_WATER__Geometry_Optimization = "B3LYP/def2-TZVPP/CMIRS(WATER) Geometry Optimization"
    B3LYP_def2_TZVPP_CMIRS_WATER__Frequency_Analysis = "B3LYP/def2-TZVPP/CMIRS(WATER) Frequency Analysis"
    B3LYP_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_WATER__Unknown = "B3LYP/def2-TZVPP/CMIRS(WATER) Unknown"
    B3LYP_def2_TZVPP_CMIRS_THF__Single_Point = "B3LYP/def2-TZVPP/CMIRS(THF) Single Point"
    B3LYP_def2_TZVPP_CMIRS_THF__Force = "B3LYP/def2-TZVPP/CMIRS(THF) Force"
    B3LYP_def2_TZVPP_CMIRS_THF__Geometry_Optimization = "B3LYP/def2-TZVPP/CMIRS(THF) Geometry Optimization"
    B3LYP_def2_TZVPP_CMIRS_THF__Frequency_Analysis = "B3LYP/def2-TZVPP/CMIRS(THF) Frequency Analysis"
    B3LYP_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_THF__Unknown = "B3LYP/def2-TZVPP/CMIRS(THF) Unknown"
    B3LYP_def2_TZVPP_CMIRS_G2__Single_Point = "B3LYP/def2-TZVPP/CMIRS(G2) Single Point"
    B3LYP_def2_TZVPP_CMIRS_G2__Force = "B3LYP/def2-TZVPP/CMIRS(G2) Force"
    B3LYP_def2_TZVPP_CMIRS_G2__Geometry_Optimization = "B3LYP/def2-TZVPP/CMIRS(G2) Geometry Optimization"
    B3LYP_def2_TZVPP_CMIRS_G2__Frequency_Analysis = "B3LYP/def2-TZVPP/CMIRS(G2) Frequency Analysis"
    B3LYP_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_G2__Unknown = "B3LYP/def2-TZVPP/CMIRS(G2) Unknown"
    B3LYP_def2_TZVPP_CMIRS_DME__Single_Point = "B3LYP/def2-TZVPP/CMIRS(DME) Single Point"
    B3LYP_def2_TZVPP_CMIRS_DME__Force = "B3LYP/def2-TZVPP/CMIRS(DME) Force"
    B3LYP_def2_TZVPP_CMIRS_DME__Geometry_Optimization = "B3LYP/def2-TZVPP/CMIRS(DME) Geometry Optimization"
    B3LYP_def2_TZVPP_CMIRS_DME__Frequency_Analysis = "B3LYP/def2-TZVPP/CMIRS(DME) Frequency Analysis"
    B3LYP_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_DME__Unknown = "B3LYP/def2-TZVPP/CMIRS(DME) Unknown"
    B3LYP_def2_TZVPP_CMIRS_EC__Single_Point = "B3LYP/def2-TZVPP/CMIRS(EC) Single Point"
    B3LYP_def2_TZVPP_CMIRS_EC__Force = "B3LYP/def2-TZVPP/CMIRS(EC) Force"
    B3LYP_def2_TZVPP_CMIRS_EC__Geometry_Optimization = "B3LYP/def2-TZVPP/CMIRS(EC) Geometry Optimization"
    B3LYP_def2_TZVPP_CMIRS_EC__Frequency_Analysis = "B3LYP/def2-TZVPP/CMIRS(EC) Frequency Analysis"
    B3LYP_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_EC__Unknown = "B3LYP/def2-TZVPP/CMIRS(EC) Unknown"
    B3LYP_def2_TZVPP_CMIRS_EC_EMC__Single_Point = "B3LYP/def2-TZVPP/CMIRS(EC/EMC) Single Point"
    B3LYP_def2_TZVPP_CMIRS_EC_EMC__Force = "B3LYP/def2-TZVPP/CMIRS(EC/EMC) Force"
    B3LYP_def2_TZVPP_CMIRS_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPP/CMIRS(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPP_CMIRS_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPP/CMIRS(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_EC_EMC__Unknown = "B3LYP/def2-TZVPP/CMIRS(EC/EMC) Unknown"
    B3LYP_def2_TZVPP_CMIRS_PC__Single_Point = "B3LYP/def2-TZVPP/CMIRS(PC) Single Point"
    B3LYP_def2_TZVPP_CMIRS_PC__Force = "B3LYP/def2-TZVPP/CMIRS(PC) Force"
    B3LYP_def2_TZVPP_CMIRS_PC__Geometry_Optimization = "B3LYP/def2-TZVPP/CMIRS(PC) Geometry Optimization"
    B3LYP_def2_TZVPP_CMIRS_PC__Frequency_Analysis = "B3LYP/def2-TZVPP/CMIRS(PC) Frequency Analysis"
    B3LYP_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPP_CMIRS_PC__Unknown = "B3LYP/def2-TZVPP/CMIRS(PC) Unknown"
    B3LYP_def2_TZVPPD_VACUUM_Single_Point = "B3LYP/def2-TZVPPD/VACUUM Single Point"
    B3LYP_def2_TZVPPD_VACUUM_Force = "B3LYP/def2-TZVPPD/VACUUM Force"
    B3LYP_def2_TZVPPD_VACUUM_Geometry_Optimization = "B3LYP/def2-TZVPPD/VACUUM Geometry Optimization"
    B3LYP_def2_TZVPPD_VACUUM_Frequency_Analysis = "B3LYP/def2-TZVPPD/VACUUM Frequency Analysis"
    B3LYP_def2_TZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/VACUUM Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_VACUUM_Unknown = "B3LYP/def2-TZVPPD/VACUUM Unknown"
    B3LYP_def2_TZVPPD_PCM_WATER__Single_Point = "B3LYP/def2-TZVPPD/PCM(WATER) Single Point"
    B3LYP_def2_TZVPPD_PCM_WATER__Force = "B3LYP/def2-TZVPPD/PCM(WATER) Force"
    B3LYP_def2_TZVPPD_PCM_WATER__Geometry_Optimization = "B3LYP/def2-TZVPPD/PCM(WATER) Geometry Optimization"
    B3LYP_def2_TZVPPD_PCM_WATER__Frequency_Analysis = "B3LYP/def2-TZVPPD/PCM(WATER) Frequency Analysis"
    B3LYP_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_WATER__Unknown = "B3LYP/def2-TZVPPD/PCM(WATER) Unknown"
    B3LYP_def2_TZVPPD_PCM_THF__Single_Point = "B3LYP/def2-TZVPPD/PCM(THF) Single Point"
    B3LYP_def2_TZVPPD_PCM_THF__Force = "B3LYP/def2-TZVPPD/PCM(THF) Force"
    B3LYP_def2_TZVPPD_PCM_THF__Geometry_Optimization = "B3LYP/def2-TZVPPD/PCM(THF) Geometry Optimization"
    B3LYP_def2_TZVPPD_PCM_THF__Frequency_Analysis = "B3LYP/def2-TZVPPD/PCM(THF) Frequency Analysis"
    B3LYP_def2_TZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_THF__Unknown = "B3LYP/def2-TZVPPD/PCM(THF) Unknown"
    B3LYP_def2_TZVPPD_PCM_G2__Single_Point = "B3LYP/def2-TZVPPD/PCM(G2) Single Point"
    B3LYP_def2_TZVPPD_PCM_G2__Force = "B3LYP/def2-TZVPPD/PCM(G2) Force"
    B3LYP_def2_TZVPPD_PCM_G2__Geometry_Optimization = "B3LYP/def2-TZVPPD/PCM(G2) Geometry Optimization"
    B3LYP_def2_TZVPPD_PCM_G2__Frequency_Analysis = "B3LYP/def2-TZVPPD/PCM(G2) Frequency Analysis"
    B3LYP_def2_TZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_G2__Unknown = "B3LYP/def2-TZVPPD/PCM(G2) Unknown"
    B3LYP_def2_TZVPPD_PCM_DME__Single_Point = "B3LYP/def2-TZVPPD/PCM(DME) Single Point"
    B3LYP_def2_TZVPPD_PCM_DME__Force = "B3LYP/def2-TZVPPD/PCM(DME) Force"
    B3LYP_def2_TZVPPD_PCM_DME__Geometry_Optimization = "B3LYP/def2-TZVPPD/PCM(DME) Geometry Optimization"
    B3LYP_def2_TZVPPD_PCM_DME__Frequency_Analysis = "B3LYP/def2-TZVPPD/PCM(DME) Frequency Analysis"
    B3LYP_def2_TZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_DME__Unknown = "B3LYP/def2-TZVPPD/PCM(DME) Unknown"
    B3LYP_def2_TZVPPD_PCM_EC__Single_Point = "B3LYP/def2-TZVPPD/PCM(EC) Single Point"
    B3LYP_def2_TZVPPD_PCM_EC__Force = "B3LYP/def2-TZVPPD/PCM(EC) Force"
    B3LYP_def2_TZVPPD_PCM_EC__Geometry_Optimization = "B3LYP/def2-TZVPPD/PCM(EC) Geometry Optimization"
    B3LYP_def2_TZVPPD_PCM_EC__Frequency_Analysis = "B3LYP/def2-TZVPPD/PCM(EC) Frequency Analysis"
    B3LYP_def2_TZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_EC__Unknown = "B3LYP/def2-TZVPPD/PCM(EC) Unknown"
    B3LYP_def2_TZVPPD_PCM_EC_EMC__Single_Point = "B3LYP/def2-TZVPPD/PCM(EC/EMC) Single Point"
    B3LYP_def2_TZVPPD_PCM_EC_EMC__Force = "B3LYP/def2-TZVPPD/PCM(EC/EMC) Force"
    B3LYP_def2_TZVPPD_PCM_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPPD/PCM(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPPD_PCM_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPPD/PCM(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_EC_EMC__Unknown = "B3LYP/def2-TZVPPD/PCM(EC/EMC) Unknown"
    B3LYP_def2_TZVPPD_PCM_PC__Single_Point = "B3LYP/def2-TZVPPD/PCM(PC) Single Point"
    B3LYP_def2_TZVPPD_PCM_PC__Force = "B3LYP/def2-TZVPPD/PCM(PC) Force"
    B3LYP_def2_TZVPPD_PCM_PC__Geometry_Optimization = "B3LYP/def2-TZVPPD/PCM(PC) Geometry Optimization"
    B3LYP_def2_TZVPPD_PCM_PC__Frequency_Analysis = "B3LYP/def2-TZVPPD/PCM(PC) Frequency Analysis"
    B3LYP_def2_TZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_PCM_PC__Unknown = "B3LYP/def2-TZVPPD/PCM(PC) Unknown"
    B3LYP_def2_TZVPPD_ISOSVP_WATER__Single_Point = "B3LYP/def2-TZVPPD/ISOSVP(WATER) Single Point"
    B3LYP_def2_TZVPPD_ISOSVP_WATER__Force = "B3LYP/def2-TZVPPD/ISOSVP(WATER) Force"
    B3LYP_def2_TZVPPD_ISOSVP_WATER__Geometry_Optimization = "B3LYP/def2-TZVPPD/ISOSVP(WATER) Geometry Optimization"
    B3LYP_def2_TZVPPD_ISOSVP_WATER__Frequency_Analysis = "B3LYP/def2-TZVPPD/ISOSVP(WATER) Frequency Analysis"
    B3LYP_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_WATER__Unknown = "B3LYP/def2-TZVPPD/ISOSVP(WATER) Unknown"
    B3LYP_def2_TZVPPD_ISOSVP_THF__Single_Point = "B3LYP/def2-TZVPPD/ISOSVP(THF) Single Point"
    B3LYP_def2_TZVPPD_ISOSVP_THF__Force = "B3LYP/def2-TZVPPD/ISOSVP(THF) Force"
    B3LYP_def2_TZVPPD_ISOSVP_THF__Geometry_Optimization = "B3LYP/def2-TZVPPD/ISOSVP(THF) Geometry Optimization"
    B3LYP_def2_TZVPPD_ISOSVP_THF__Frequency_Analysis = "B3LYP/def2-TZVPPD/ISOSVP(THF) Frequency Analysis"
    B3LYP_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_THF__Unknown = "B3LYP/def2-TZVPPD/ISOSVP(THF) Unknown"
    B3LYP_def2_TZVPPD_ISOSVP_G2__Single_Point = "B3LYP/def2-TZVPPD/ISOSVP(G2) Single Point"
    B3LYP_def2_TZVPPD_ISOSVP_G2__Force = "B3LYP/def2-TZVPPD/ISOSVP(G2) Force"
    B3LYP_def2_TZVPPD_ISOSVP_G2__Geometry_Optimization = "B3LYP/def2-TZVPPD/ISOSVP(G2) Geometry Optimization"
    B3LYP_def2_TZVPPD_ISOSVP_G2__Frequency_Analysis = "B3LYP/def2-TZVPPD/ISOSVP(G2) Frequency Analysis"
    B3LYP_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_G2__Unknown = "B3LYP/def2-TZVPPD/ISOSVP(G2) Unknown"
    B3LYP_def2_TZVPPD_ISOSVP_DME__Single_Point = "B3LYP/def2-TZVPPD/ISOSVP(DME) Single Point"
    B3LYP_def2_TZVPPD_ISOSVP_DME__Force = "B3LYP/def2-TZVPPD/ISOSVP(DME) Force"
    B3LYP_def2_TZVPPD_ISOSVP_DME__Geometry_Optimization = "B3LYP/def2-TZVPPD/ISOSVP(DME) Geometry Optimization"
    B3LYP_def2_TZVPPD_ISOSVP_DME__Frequency_Analysis = "B3LYP/def2-TZVPPD/ISOSVP(DME) Frequency Analysis"
    B3LYP_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_DME__Unknown = "B3LYP/def2-TZVPPD/ISOSVP(DME) Unknown"
    B3LYP_def2_TZVPPD_ISOSVP_EC__Single_Point = "B3LYP/def2-TZVPPD/ISOSVP(EC) Single Point"
    B3LYP_def2_TZVPPD_ISOSVP_EC__Force = "B3LYP/def2-TZVPPD/ISOSVP(EC) Force"
    B3LYP_def2_TZVPPD_ISOSVP_EC__Geometry_Optimization = "B3LYP/def2-TZVPPD/ISOSVP(EC) Geometry Optimization"
    B3LYP_def2_TZVPPD_ISOSVP_EC__Frequency_Analysis = "B3LYP/def2-TZVPPD/ISOSVP(EC) Frequency Analysis"
    B3LYP_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_EC__Unknown = "B3LYP/def2-TZVPPD/ISOSVP(EC) Unknown"
    B3LYP_def2_TZVPPD_ISOSVP_EC_EMC__Single_Point = "B3LYP/def2-TZVPPD/ISOSVP(EC/EMC) Single Point"
    B3LYP_def2_TZVPPD_ISOSVP_EC_EMC__Force = "B3LYP/def2-TZVPPD/ISOSVP(EC/EMC) Force"
    B3LYP_def2_TZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_EC_EMC__Unknown = "B3LYP/def2-TZVPPD/ISOSVP(EC/EMC) Unknown"
    B3LYP_def2_TZVPPD_ISOSVP_PC__Single_Point = "B3LYP/def2-TZVPPD/ISOSVP(PC) Single Point"
    B3LYP_def2_TZVPPD_ISOSVP_PC__Force = "B3LYP/def2-TZVPPD/ISOSVP(PC) Force"
    B3LYP_def2_TZVPPD_ISOSVP_PC__Geometry_Optimization = "B3LYP/def2-TZVPPD/ISOSVP(PC) Geometry Optimization"
    B3LYP_def2_TZVPPD_ISOSVP_PC__Frequency_Analysis = "B3LYP/def2-TZVPPD/ISOSVP(PC) Frequency Analysis"
    B3LYP_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_ISOSVP_PC__Unknown = "B3LYP/def2-TZVPPD/ISOSVP(PC) Unknown"
    B3LYP_def2_TZVPPD_SMD_WATER__Single_Point = "B3LYP/def2-TZVPPD/SMD(WATER) Single Point"
    B3LYP_def2_TZVPPD_SMD_WATER__Force = "B3LYP/def2-TZVPPD/SMD(WATER) Force"
    B3LYP_def2_TZVPPD_SMD_WATER__Geometry_Optimization = "B3LYP/def2-TZVPPD/SMD(WATER) Geometry Optimization"
    B3LYP_def2_TZVPPD_SMD_WATER__Frequency_Analysis = "B3LYP/def2-TZVPPD/SMD(WATER) Frequency Analysis"
    B3LYP_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_WATER__Unknown = "B3LYP/def2-TZVPPD/SMD(WATER) Unknown"
    B3LYP_def2_TZVPPD_SMD_THF__Single_Point = "B3LYP/def2-TZVPPD/SMD(THF) Single Point"
    B3LYP_def2_TZVPPD_SMD_THF__Force = "B3LYP/def2-TZVPPD/SMD(THF) Force"
    B3LYP_def2_TZVPPD_SMD_THF__Geometry_Optimization = "B3LYP/def2-TZVPPD/SMD(THF) Geometry Optimization"
    B3LYP_def2_TZVPPD_SMD_THF__Frequency_Analysis = "B3LYP/def2-TZVPPD/SMD(THF) Frequency Analysis"
    B3LYP_def2_TZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_THF__Unknown = "B3LYP/def2-TZVPPD/SMD(THF) Unknown"
    B3LYP_def2_TZVPPD_SMD_G2__Single_Point = "B3LYP/def2-TZVPPD/SMD(G2) Single Point"
    B3LYP_def2_TZVPPD_SMD_G2__Force = "B3LYP/def2-TZVPPD/SMD(G2) Force"
    B3LYP_def2_TZVPPD_SMD_G2__Geometry_Optimization = "B3LYP/def2-TZVPPD/SMD(G2) Geometry Optimization"
    B3LYP_def2_TZVPPD_SMD_G2__Frequency_Analysis = "B3LYP/def2-TZVPPD/SMD(G2) Frequency Analysis"
    B3LYP_def2_TZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_G2__Unknown = "B3LYP/def2-TZVPPD/SMD(G2) Unknown"
    B3LYP_def2_TZVPPD_SMD_DME__Single_Point = "B3LYP/def2-TZVPPD/SMD(DME) Single Point"
    B3LYP_def2_TZVPPD_SMD_DME__Force = "B3LYP/def2-TZVPPD/SMD(DME) Force"
    B3LYP_def2_TZVPPD_SMD_DME__Geometry_Optimization = "B3LYP/def2-TZVPPD/SMD(DME) Geometry Optimization"
    B3LYP_def2_TZVPPD_SMD_DME__Frequency_Analysis = "B3LYP/def2-TZVPPD/SMD(DME) Frequency Analysis"
    B3LYP_def2_TZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_DME__Unknown = "B3LYP/def2-TZVPPD/SMD(DME) Unknown"
    B3LYP_def2_TZVPPD_SMD_EC__Single_Point = "B3LYP/def2-TZVPPD/SMD(EC) Single Point"
    B3LYP_def2_TZVPPD_SMD_EC__Force = "B3LYP/def2-TZVPPD/SMD(EC) Force"
    B3LYP_def2_TZVPPD_SMD_EC__Geometry_Optimization = "B3LYP/def2-TZVPPD/SMD(EC) Geometry Optimization"
    B3LYP_def2_TZVPPD_SMD_EC__Frequency_Analysis = "B3LYP/def2-TZVPPD/SMD(EC) Frequency Analysis"
    B3LYP_def2_TZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_EC__Unknown = "B3LYP/def2-TZVPPD/SMD(EC) Unknown"
    B3LYP_def2_TZVPPD_SMD_EC_EMC__Single_Point = "B3LYP/def2-TZVPPD/SMD(EC/EMC) Single Point"
    B3LYP_def2_TZVPPD_SMD_EC_EMC__Force = "B3LYP/def2-TZVPPD/SMD(EC/EMC) Force"
    B3LYP_def2_TZVPPD_SMD_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPPD/SMD(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPPD_SMD_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPPD/SMD(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_EC_EMC__Unknown = "B3LYP/def2-TZVPPD/SMD(EC/EMC) Unknown"
    B3LYP_def2_TZVPPD_SMD_PC__Single_Point = "B3LYP/def2-TZVPPD/SMD(PC) Single Point"
    B3LYP_def2_TZVPPD_SMD_PC__Force = "B3LYP/def2-TZVPPD/SMD(PC) Force"
    B3LYP_def2_TZVPPD_SMD_PC__Geometry_Optimization = "B3LYP/def2-TZVPPD/SMD(PC) Geometry Optimization"
    B3LYP_def2_TZVPPD_SMD_PC__Frequency_Analysis = "B3LYP/def2-TZVPPD/SMD(PC) Frequency Analysis"
    B3LYP_def2_TZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_SMD_PC__Unknown = "B3LYP/def2-TZVPPD/SMD(PC) Unknown"
    B3LYP_def2_TZVPPD_CMIRS_WATER__Single_Point = "B3LYP/def2-TZVPPD/CMIRS(WATER) Single Point"
    B3LYP_def2_TZVPPD_CMIRS_WATER__Force = "B3LYP/def2-TZVPPD/CMIRS(WATER) Force"
    B3LYP_def2_TZVPPD_CMIRS_WATER__Geometry_Optimization = "B3LYP/def2-TZVPPD/CMIRS(WATER) Geometry Optimization"
    B3LYP_def2_TZVPPD_CMIRS_WATER__Frequency_Analysis = "B3LYP/def2-TZVPPD/CMIRS(WATER) Frequency Analysis"
    B3LYP_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_WATER__Unknown = "B3LYP/def2-TZVPPD/CMIRS(WATER) Unknown"
    B3LYP_def2_TZVPPD_CMIRS_THF__Single_Point = "B3LYP/def2-TZVPPD/CMIRS(THF) Single Point"
    B3LYP_def2_TZVPPD_CMIRS_THF__Force = "B3LYP/def2-TZVPPD/CMIRS(THF) Force"
    B3LYP_def2_TZVPPD_CMIRS_THF__Geometry_Optimization = "B3LYP/def2-TZVPPD/CMIRS(THF) Geometry Optimization"
    B3LYP_def2_TZVPPD_CMIRS_THF__Frequency_Analysis = "B3LYP/def2-TZVPPD/CMIRS(THF) Frequency Analysis"
    B3LYP_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_THF__Unknown = "B3LYP/def2-TZVPPD/CMIRS(THF) Unknown"
    B3LYP_def2_TZVPPD_CMIRS_G2__Single_Point = "B3LYP/def2-TZVPPD/CMIRS(G2) Single Point"
    B3LYP_def2_TZVPPD_CMIRS_G2__Force = "B3LYP/def2-TZVPPD/CMIRS(G2) Force"
    B3LYP_def2_TZVPPD_CMIRS_G2__Geometry_Optimization = "B3LYP/def2-TZVPPD/CMIRS(G2) Geometry Optimization"
    B3LYP_def2_TZVPPD_CMIRS_G2__Frequency_Analysis = "B3LYP/def2-TZVPPD/CMIRS(G2) Frequency Analysis"
    B3LYP_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_G2__Unknown = "B3LYP/def2-TZVPPD/CMIRS(G2) Unknown"
    B3LYP_def2_TZVPPD_CMIRS_DME__Single_Point = "B3LYP/def2-TZVPPD/CMIRS(DME) Single Point"
    B3LYP_def2_TZVPPD_CMIRS_DME__Force = "B3LYP/def2-TZVPPD/CMIRS(DME) Force"
    B3LYP_def2_TZVPPD_CMIRS_DME__Geometry_Optimization = "B3LYP/def2-TZVPPD/CMIRS(DME) Geometry Optimization"
    B3LYP_def2_TZVPPD_CMIRS_DME__Frequency_Analysis = "B3LYP/def2-TZVPPD/CMIRS(DME) Frequency Analysis"
    B3LYP_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_DME__Unknown = "B3LYP/def2-TZVPPD/CMIRS(DME) Unknown"
    B3LYP_def2_TZVPPD_CMIRS_EC__Single_Point = "B3LYP/def2-TZVPPD/CMIRS(EC) Single Point"
    B3LYP_def2_TZVPPD_CMIRS_EC__Force = "B3LYP/def2-TZVPPD/CMIRS(EC) Force"
    B3LYP_def2_TZVPPD_CMIRS_EC__Geometry_Optimization = "B3LYP/def2-TZVPPD/CMIRS(EC) Geometry Optimization"
    B3LYP_def2_TZVPPD_CMIRS_EC__Frequency_Analysis = "B3LYP/def2-TZVPPD/CMIRS(EC) Frequency Analysis"
    B3LYP_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_EC__Unknown = "B3LYP/def2-TZVPPD/CMIRS(EC) Unknown"
    B3LYP_def2_TZVPPD_CMIRS_EC_EMC__Single_Point = "B3LYP/def2-TZVPPD/CMIRS(EC/EMC) Single Point"
    B3LYP_def2_TZVPPD_CMIRS_EC_EMC__Force = "B3LYP/def2-TZVPPD/CMIRS(EC/EMC) Force"
    B3LYP_def2_TZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "B3LYP/def2-TZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    B3LYP_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "B3LYP/def2-TZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    B3LYP_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_EC_EMC__Unknown = "B3LYP/def2-TZVPPD/CMIRS(EC/EMC) Unknown"
    B3LYP_def2_TZVPPD_CMIRS_PC__Single_Point = "B3LYP/def2-TZVPPD/CMIRS(PC) Single Point"
    B3LYP_def2_TZVPPD_CMIRS_PC__Force = "B3LYP/def2-TZVPPD/CMIRS(PC) Force"
    B3LYP_def2_TZVPPD_CMIRS_PC__Geometry_Optimization = "B3LYP/def2-TZVPPD/CMIRS(PC) Geometry Optimization"
    B3LYP_def2_TZVPPD_CMIRS_PC__Frequency_Analysis = "B3LYP/def2-TZVPPD/CMIRS(PC) Frequency Analysis"
    B3LYP_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-TZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_TZVPPD_CMIRS_PC__Unknown = "B3LYP/def2-TZVPPD/CMIRS(PC) Unknown"
    B3LYP_def2_QZVPPD_VACUUM_Single_Point = "B3LYP/def2-QZVPPD/VACUUM Single Point"
    B3LYP_def2_QZVPPD_VACUUM_Force = "B3LYP/def2-QZVPPD/VACUUM Force"
    B3LYP_def2_QZVPPD_VACUUM_Geometry_Optimization = "B3LYP/def2-QZVPPD/VACUUM Geometry Optimization"
    B3LYP_def2_QZVPPD_VACUUM_Frequency_Analysis = "B3LYP/def2-QZVPPD/VACUUM Frequency Analysis"
    B3LYP_def2_QZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/VACUUM Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_VACUUM_Unknown = "B3LYP/def2-QZVPPD/VACUUM Unknown"
    B3LYP_def2_QZVPPD_PCM_WATER__Single_Point = "B3LYP/def2-QZVPPD/PCM(WATER) Single Point"
    B3LYP_def2_QZVPPD_PCM_WATER__Force = "B3LYP/def2-QZVPPD/PCM(WATER) Force"
    B3LYP_def2_QZVPPD_PCM_WATER__Geometry_Optimization = "B3LYP/def2-QZVPPD/PCM(WATER) Geometry Optimization"
    B3LYP_def2_QZVPPD_PCM_WATER__Frequency_Analysis = "B3LYP/def2-QZVPPD/PCM(WATER) Frequency Analysis"
    B3LYP_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_WATER__Unknown = "B3LYP/def2-QZVPPD/PCM(WATER) Unknown"
    B3LYP_def2_QZVPPD_PCM_THF__Single_Point = "B3LYP/def2-QZVPPD/PCM(THF) Single Point"
    B3LYP_def2_QZVPPD_PCM_THF__Force = "B3LYP/def2-QZVPPD/PCM(THF) Force"
    B3LYP_def2_QZVPPD_PCM_THF__Geometry_Optimization = "B3LYP/def2-QZVPPD/PCM(THF) Geometry Optimization"
    B3LYP_def2_QZVPPD_PCM_THF__Frequency_Analysis = "B3LYP/def2-QZVPPD/PCM(THF) Frequency Analysis"
    B3LYP_def2_QZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_THF__Unknown = "B3LYP/def2-QZVPPD/PCM(THF) Unknown"
    B3LYP_def2_QZVPPD_PCM_G2__Single_Point = "B3LYP/def2-QZVPPD/PCM(G2) Single Point"
    B3LYP_def2_QZVPPD_PCM_G2__Force = "B3LYP/def2-QZVPPD/PCM(G2) Force"
    B3LYP_def2_QZVPPD_PCM_G2__Geometry_Optimization = "B3LYP/def2-QZVPPD/PCM(G2) Geometry Optimization"
    B3LYP_def2_QZVPPD_PCM_G2__Frequency_Analysis = "B3LYP/def2-QZVPPD/PCM(G2) Frequency Analysis"
    B3LYP_def2_QZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_G2__Unknown = "B3LYP/def2-QZVPPD/PCM(G2) Unknown"
    B3LYP_def2_QZVPPD_PCM_DME__Single_Point = "B3LYP/def2-QZVPPD/PCM(DME) Single Point"
    B3LYP_def2_QZVPPD_PCM_DME__Force = "B3LYP/def2-QZVPPD/PCM(DME) Force"
    B3LYP_def2_QZVPPD_PCM_DME__Geometry_Optimization = "B3LYP/def2-QZVPPD/PCM(DME) Geometry Optimization"
    B3LYP_def2_QZVPPD_PCM_DME__Frequency_Analysis = "B3LYP/def2-QZVPPD/PCM(DME) Frequency Analysis"
    B3LYP_def2_QZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_DME__Unknown = "B3LYP/def2-QZVPPD/PCM(DME) Unknown"
    B3LYP_def2_QZVPPD_PCM_EC__Single_Point = "B3LYP/def2-QZVPPD/PCM(EC) Single Point"
    B3LYP_def2_QZVPPD_PCM_EC__Force = "B3LYP/def2-QZVPPD/PCM(EC) Force"
    B3LYP_def2_QZVPPD_PCM_EC__Geometry_Optimization = "B3LYP/def2-QZVPPD/PCM(EC) Geometry Optimization"
    B3LYP_def2_QZVPPD_PCM_EC__Frequency_Analysis = "B3LYP/def2-QZVPPD/PCM(EC) Frequency Analysis"
    B3LYP_def2_QZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_EC__Unknown = "B3LYP/def2-QZVPPD/PCM(EC) Unknown"
    B3LYP_def2_QZVPPD_PCM_EC_EMC__Single_Point = "B3LYP/def2-QZVPPD/PCM(EC/EMC) Single Point"
    B3LYP_def2_QZVPPD_PCM_EC_EMC__Force = "B3LYP/def2-QZVPPD/PCM(EC/EMC) Force"
    B3LYP_def2_QZVPPD_PCM_EC_EMC__Geometry_Optimization = "B3LYP/def2-QZVPPD/PCM(EC/EMC) Geometry Optimization"
    B3LYP_def2_QZVPPD_PCM_EC_EMC__Frequency_Analysis = "B3LYP/def2-QZVPPD/PCM(EC/EMC) Frequency Analysis"
    B3LYP_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_EC_EMC__Unknown = "B3LYP/def2-QZVPPD/PCM(EC/EMC) Unknown"
    B3LYP_def2_QZVPPD_PCM_PC__Single_Point = "B3LYP/def2-QZVPPD/PCM(PC) Single Point"
    B3LYP_def2_QZVPPD_PCM_PC__Force = "B3LYP/def2-QZVPPD/PCM(PC) Force"
    B3LYP_def2_QZVPPD_PCM_PC__Geometry_Optimization = "B3LYP/def2-QZVPPD/PCM(PC) Geometry Optimization"
    B3LYP_def2_QZVPPD_PCM_PC__Frequency_Analysis = "B3LYP/def2-QZVPPD/PCM(PC) Frequency Analysis"
    B3LYP_def2_QZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_PCM_PC__Unknown = "B3LYP/def2-QZVPPD/PCM(PC) Unknown"
    B3LYP_def2_QZVPPD_ISOSVP_WATER__Single_Point = "B3LYP/def2-QZVPPD/ISOSVP(WATER) Single Point"
    B3LYP_def2_QZVPPD_ISOSVP_WATER__Force = "B3LYP/def2-QZVPPD/ISOSVP(WATER) Force"
    B3LYP_def2_QZVPPD_ISOSVP_WATER__Geometry_Optimization = "B3LYP/def2-QZVPPD/ISOSVP(WATER) Geometry Optimization"
    B3LYP_def2_QZVPPD_ISOSVP_WATER__Frequency_Analysis = "B3LYP/def2-QZVPPD/ISOSVP(WATER) Frequency Analysis"
    B3LYP_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_WATER__Unknown = "B3LYP/def2-QZVPPD/ISOSVP(WATER) Unknown"
    B3LYP_def2_QZVPPD_ISOSVP_THF__Single_Point = "B3LYP/def2-QZVPPD/ISOSVP(THF) Single Point"
    B3LYP_def2_QZVPPD_ISOSVP_THF__Force = "B3LYP/def2-QZVPPD/ISOSVP(THF) Force"
    B3LYP_def2_QZVPPD_ISOSVP_THF__Geometry_Optimization = "B3LYP/def2-QZVPPD/ISOSVP(THF) Geometry Optimization"
    B3LYP_def2_QZVPPD_ISOSVP_THF__Frequency_Analysis = "B3LYP/def2-QZVPPD/ISOSVP(THF) Frequency Analysis"
    B3LYP_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_THF__Unknown = "B3LYP/def2-QZVPPD/ISOSVP(THF) Unknown"
    B3LYP_def2_QZVPPD_ISOSVP_G2__Single_Point = "B3LYP/def2-QZVPPD/ISOSVP(G2) Single Point"
    B3LYP_def2_QZVPPD_ISOSVP_G2__Force = "B3LYP/def2-QZVPPD/ISOSVP(G2) Force"
    B3LYP_def2_QZVPPD_ISOSVP_G2__Geometry_Optimization = "B3LYP/def2-QZVPPD/ISOSVP(G2) Geometry Optimization"
    B3LYP_def2_QZVPPD_ISOSVP_G2__Frequency_Analysis = "B3LYP/def2-QZVPPD/ISOSVP(G2) Frequency Analysis"
    B3LYP_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_G2__Unknown = "B3LYP/def2-QZVPPD/ISOSVP(G2) Unknown"
    B3LYP_def2_QZVPPD_ISOSVP_DME__Single_Point = "B3LYP/def2-QZVPPD/ISOSVP(DME) Single Point"
    B3LYP_def2_QZVPPD_ISOSVP_DME__Force = "B3LYP/def2-QZVPPD/ISOSVP(DME) Force"
    B3LYP_def2_QZVPPD_ISOSVP_DME__Geometry_Optimization = "B3LYP/def2-QZVPPD/ISOSVP(DME) Geometry Optimization"
    B3LYP_def2_QZVPPD_ISOSVP_DME__Frequency_Analysis = "B3LYP/def2-QZVPPD/ISOSVP(DME) Frequency Analysis"
    B3LYP_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_DME__Unknown = "B3LYP/def2-QZVPPD/ISOSVP(DME) Unknown"
    B3LYP_def2_QZVPPD_ISOSVP_EC__Single_Point = "B3LYP/def2-QZVPPD/ISOSVP(EC) Single Point"
    B3LYP_def2_QZVPPD_ISOSVP_EC__Force = "B3LYP/def2-QZVPPD/ISOSVP(EC) Force"
    B3LYP_def2_QZVPPD_ISOSVP_EC__Geometry_Optimization = "B3LYP/def2-QZVPPD/ISOSVP(EC) Geometry Optimization"
    B3LYP_def2_QZVPPD_ISOSVP_EC__Frequency_Analysis = "B3LYP/def2-QZVPPD/ISOSVP(EC) Frequency Analysis"
    B3LYP_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_EC__Unknown = "B3LYP/def2-QZVPPD/ISOSVP(EC) Unknown"
    B3LYP_def2_QZVPPD_ISOSVP_EC_EMC__Single_Point = "B3LYP/def2-QZVPPD/ISOSVP(EC/EMC) Single Point"
    B3LYP_def2_QZVPPD_ISOSVP_EC_EMC__Force = "B3LYP/def2-QZVPPD/ISOSVP(EC/EMC) Force"
    B3LYP_def2_QZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = "B3LYP/def2-QZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    B3LYP_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "B3LYP/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    B3LYP_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_EC_EMC__Unknown = "B3LYP/def2-QZVPPD/ISOSVP(EC/EMC) Unknown"
    B3LYP_def2_QZVPPD_ISOSVP_PC__Single_Point = "B3LYP/def2-QZVPPD/ISOSVP(PC) Single Point"
    B3LYP_def2_QZVPPD_ISOSVP_PC__Force = "B3LYP/def2-QZVPPD/ISOSVP(PC) Force"
    B3LYP_def2_QZVPPD_ISOSVP_PC__Geometry_Optimization = "B3LYP/def2-QZVPPD/ISOSVP(PC) Geometry Optimization"
    B3LYP_def2_QZVPPD_ISOSVP_PC__Frequency_Analysis = "B3LYP/def2-QZVPPD/ISOSVP(PC) Frequency Analysis"
    B3LYP_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_ISOSVP_PC__Unknown = "B3LYP/def2-QZVPPD/ISOSVP(PC) Unknown"
    B3LYP_def2_QZVPPD_SMD_WATER__Single_Point = "B3LYP/def2-QZVPPD/SMD(WATER) Single Point"
    B3LYP_def2_QZVPPD_SMD_WATER__Force = "B3LYP/def2-QZVPPD/SMD(WATER) Force"
    B3LYP_def2_QZVPPD_SMD_WATER__Geometry_Optimization = "B3LYP/def2-QZVPPD/SMD(WATER) Geometry Optimization"
    B3LYP_def2_QZVPPD_SMD_WATER__Frequency_Analysis = "B3LYP/def2-QZVPPD/SMD(WATER) Frequency Analysis"
    B3LYP_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_WATER__Unknown = "B3LYP/def2-QZVPPD/SMD(WATER) Unknown"
    B3LYP_def2_QZVPPD_SMD_THF__Single_Point = "B3LYP/def2-QZVPPD/SMD(THF) Single Point"
    B3LYP_def2_QZVPPD_SMD_THF__Force = "B3LYP/def2-QZVPPD/SMD(THF) Force"
    B3LYP_def2_QZVPPD_SMD_THF__Geometry_Optimization = "B3LYP/def2-QZVPPD/SMD(THF) Geometry Optimization"
    B3LYP_def2_QZVPPD_SMD_THF__Frequency_Analysis = "B3LYP/def2-QZVPPD/SMD(THF) Frequency Analysis"
    B3LYP_def2_QZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_THF__Unknown = "B3LYP/def2-QZVPPD/SMD(THF) Unknown"
    B3LYP_def2_QZVPPD_SMD_G2__Single_Point = "B3LYP/def2-QZVPPD/SMD(G2) Single Point"
    B3LYP_def2_QZVPPD_SMD_G2__Force = "B3LYP/def2-QZVPPD/SMD(G2) Force"
    B3LYP_def2_QZVPPD_SMD_G2__Geometry_Optimization = "B3LYP/def2-QZVPPD/SMD(G2) Geometry Optimization"
    B3LYP_def2_QZVPPD_SMD_G2__Frequency_Analysis = "B3LYP/def2-QZVPPD/SMD(G2) Frequency Analysis"
    B3LYP_def2_QZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_G2__Unknown = "B3LYP/def2-QZVPPD/SMD(G2) Unknown"
    B3LYP_def2_QZVPPD_SMD_DME__Single_Point = "B3LYP/def2-QZVPPD/SMD(DME) Single Point"
    B3LYP_def2_QZVPPD_SMD_DME__Force = "B3LYP/def2-QZVPPD/SMD(DME) Force"
    B3LYP_def2_QZVPPD_SMD_DME__Geometry_Optimization = "B3LYP/def2-QZVPPD/SMD(DME) Geometry Optimization"
    B3LYP_def2_QZVPPD_SMD_DME__Frequency_Analysis = "B3LYP/def2-QZVPPD/SMD(DME) Frequency Analysis"
    B3LYP_def2_QZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_DME__Unknown = "B3LYP/def2-QZVPPD/SMD(DME) Unknown"
    B3LYP_def2_QZVPPD_SMD_EC__Single_Point = "B3LYP/def2-QZVPPD/SMD(EC) Single Point"
    B3LYP_def2_QZVPPD_SMD_EC__Force = "B3LYP/def2-QZVPPD/SMD(EC) Force"
    B3LYP_def2_QZVPPD_SMD_EC__Geometry_Optimization = "B3LYP/def2-QZVPPD/SMD(EC) Geometry Optimization"
    B3LYP_def2_QZVPPD_SMD_EC__Frequency_Analysis = "B3LYP/def2-QZVPPD/SMD(EC) Frequency Analysis"
    B3LYP_def2_QZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_EC__Unknown = "B3LYP/def2-QZVPPD/SMD(EC) Unknown"
    B3LYP_def2_QZVPPD_SMD_EC_EMC__Single_Point = "B3LYP/def2-QZVPPD/SMD(EC/EMC) Single Point"
    B3LYP_def2_QZVPPD_SMD_EC_EMC__Force = "B3LYP/def2-QZVPPD/SMD(EC/EMC) Force"
    B3LYP_def2_QZVPPD_SMD_EC_EMC__Geometry_Optimization = "B3LYP/def2-QZVPPD/SMD(EC/EMC) Geometry Optimization"
    B3LYP_def2_QZVPPD_SMD_EC_EMC__Frequency_Analysis = "B3LYP/def2-QZVPPD/SMD(EC/EMC) Frequency Analysis"
    B3LYP_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_EC_EMC__Unknown = "B3LYP/def2-QZVPPD/SMD(EC/EMC) Unknown"
    B3LYP_def2_QZVPPD_SMD_PC__Single_Point = "B3LYP/def2-QZVPPD/SMD(PC) Single Point"
    B3LYP_def2_QZVPPD_SMD_PC__Force = "B3LYP/def2-QZVPPD/SMD(PC) Force"
    B3LYP_def2_QZVPPD_SMD_PC__Geometry_Optimization = "B3LYP/def2-QZVPPD/SMD(PC) Geometry Optimization"
    B3LYP_def2_QZVPPD_SMD_PC__Frequency_Analysis = "B3LYP/def2-QZVPPD/SMD(PC) Frequency Analysis"
    B3LYP_def2_QZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_SMD_PC__Unknown = "B3LYP/def2-QZVPPD/SMD(PC) Unknown"
    B3LYP_def2_QZVPPD_CMIRS_WATER__Single_Point = "B3LYP/def2-QZVPPD/CMIRS(WATER) Single Point"
    B3LYP_def2_QZVPPD_CMIRS_WATER__Force = "B3LYP/def2-QZVPPD/CMIRS(WATER) Force"
    B3LYP_def2_QZVPPD_CMIRS_WATER__Geometry_Optimization = "B3LYP/def2-QZVPPD/CMIRS(WATER) Geometry Optimization"
    B3LYP_def2_QZVPPD_CMIRS_WATER__Frequency_Analysis = "B3LYP/def2-QZVPPD/CMIRS(WATER) Frequency Analysis"
    B3LYP_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_WATER__Unknown = "B3LYP/def2-QZVPPD/CMIRS(WATER) Unknown"
    B3LYP_def2_QZVPPD_CMIRS_THF__Single_Point = "B3LYP/def2-QZVPPD/CMIRS(THF) Single Point"
    B3LYP_def2_QZVPPD_CMIRS_THF__Force = "B3LYP/def2-QZVPPD/CMIRS(THF) Force"
    B3LYP_def2_QZVPPD_CMIRS_THF__Geometry_Optimization = "B3LYP/def2-QZVPPD/CMIRS(THF) Geometry Optimization"
    B3LYP_def2_QZVPPD_CMIRS_THF__Frequency_Analysis = "B3LYP/def2-QZVPPD/CMIRS(THF) Frequency Analysis"
    B3LYP_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_THF__Unknown = "B3LYP/def2-QZVPPD/CMIRS(THF) Unknown"
    B3LYP_def2_QZVPPD_CMIRS_G2__Single_Point = "B3LYP/def2-QZVPPD/CMIRS(G2) Single Point"
    B3LYP_def2_QZVPPD_CMIRS_G2__Force = "B3LYP/def2-QZVPPD/CMIRS(G2) Force"
    B3LYP_def2_QZVPPD_CMIRS_G2__Geometry_Optimization = "B3LYP/def2-QZVPPD/CMIRS(G2) Geometry Optimization"
    B3LYP_def2_QZVPPD_CMIRS_G2__Frequency_Analysis = "B3LYP/def2-QZVPPD/CMIRS(G2) Frequency Analysis"
    B3LYP_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_G2__Unknown = "B3LYP/def2-QZVPPD/CMIRS(G2) Unknown"
    B3LYP_def2_QZVPPD_CMIRS_DME__Single_Point = "B3LYP/def2-QZVPPD/CMIRS(DME) Single Point"
    B3LYP_def2_QZVPPD_CMIRS_DME__Force = "B3LYP/def2-QZVPPD/CMIRS(DME) Force"
    B3LYP_def2_QZVPPD_CMIRS_DME__Geometry_Optimization = "B3LYP/def2-QZVPPD/CMIRS(DME) Geometry Optimization"
    B3LYP_def2_QZVPPD_CMIRS_DME__Frequency_Analysis = "B3LYP/def2-QZVPPD/CMIRS(DME) Frequency Analysis"
    B3LYP_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_DME__Unknown = "B3LYP/def2-QZVPPD/CMIRS(DME) Unknown"
    B3LYP_def2_QZVPPD_CMIRS_EC__Single_Point = "B3LYP/def2-QZVPPD/CMIRS(EC) Single Point"
    B3LYP_def2_QZVPPD_CMIRS_EC__Force = "B3LYP/def2-QZVPPD/CMIRS(EC) Force"
    B3LYP_def2_QZVPPD_CMIRS_EC__Geometry_Optimization = "B3LYP/def2-QZVPPD/CMIRS(EC) Geometry Optimization"
    B3LYP_def2_QZVPPD_CMIRS_EC__Frequency_Analysis = "B3LYP/def2-QZVPPD/CMIRS(EC) Frequency Analysis"
    B3LYP_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_EC__Unknown = "B3LYP/def2-QZVPPD/CMIRS(EC) Unknown"
    B3LYP_def2_QZVPPD_CMIRS_EC_EMC__Single_Point = "B3LYP/def2-QZVPPD/CMIRS(EC/EMC) Single Point"
    B3LYP_def2_QZVPPD_CMIRS_EC_EMC__Force = "B3LYP/def2-QZVPPD/CMIRS(EC/EMC) Force"
    B3LYP_def2_QZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "B3LYP/def2-QZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    B3LYP_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "B3LYP/def2-QZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    B3LYP_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_EC_EMC__Unknown = "B3LYP/def2-QZVPPD/CMIRS(EC/EMC) Unknown"
    B3LYP_def2_QZVPPD_CMIRS_PC__Single_Point = "B3LYP/def2-QZVPPD/CMIRS(PC) Single Point"
    B3LYP_def2_QZVPPD_CMIRS_PC__Force = "B3LYP/def2-QZVPPD/CMIRS(PC) Force"
    B3LYP_def2_QZVPPD_CMIRS_PC__Geometry_Optimization = "B3LYP/def2-QZVPPD/CMIRS(PC) Geometry Optimization"
    B3LYP_def2_QZVPPD_CMIRS_PC__Frequency_Analysis = "B3LYP/def2-QZVPPD/CMIRS(PC) Frequency Analysis"
    B3LYP_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "B3LYP/def2-QZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    B3LYP_def2_QZVPPD_CMIRS_PC__Unknown = "B3LYP/def2-QZVPPD/CMIRS(PC) Unknown"
    wB97X_D_6_31g_d_VACUUM_Single_Point = "wB97X-D/6-31g*/VACUUM Single Point"
    wB97X_D_6_31g_d_VACUUM_Force = "wB97X-D/6-31g*/VACUUM Force"
    wB97X_D_6_31g_d_VACUUM_Geometry_Optimization = "wB97X-D/6-31g*/VACUUM Geometry Optimization"
    wB97X_D_6_31g_d_VACUUM_Frequency_Analysis = "wB97X-D/6-31g*/VACUUM Frequency Analysis"
    wB97X_D_6_31g_d_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_VACUUM_Unknown = "wB97X-D/6-31g*/VACUUM Unknown"
    wB97X_D_6_31g_d_PCM_WATER__Single_Point = "wB97X-D/6-31g*/PCM(WATER) Single Point"
    wB97X_D_6_31g_d_PCM_WATER__Force = "wB97X-D/6-31g*/PCM(WATER) Force"
    wB97X_D_6_31g_d_PCM_WATER__Geometry_Optimization = "wB97X-D/6-31g*/PCM(WATER) Geometry Optimization"
    wB97X_D_6_31g_d_PCM_WATER__Frequency_Analysis = "wB97X-D/6-31g*/PCM(WATER) Frequency Analysis"
    wB97X_D_6_31g_d_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_WATER__Unknown = "wB97X-D/6-31g*/PCM(WATER) Unknown"
    wB97X_D_6_31g_d_PCM_THF__Single_Point = "wB97X-D/6-31g*/PCM(THF) Single Point"
    wB97X_D_6_31g_d_PCM_THF__Force = "wB97X-D/6-31g*/PCM(THF) Force"
    wB97X_D_6_31g_d_PCM_THF__Geometry_Optimization = "wB97X-D/6-31g*/PCM(THF) Geometry Optimization"
    wB97X_D_6_31g_d_PCM_THF__Frequency_Analysis = "wB97X-D/6-31g*/PCM(THF) Frequency Analysis"
    wB97X_D_6_31g_d_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_THF__Unknown = "wB97X-D/6-31g*/PCM(THF) Unknown"
    wB97X_D_6_31g_d_PCM_G2__Single_Point = "wB97X-D/6-31g*/PCM(G2) Single Point"
    wB97X_D_6_31g_d_PCM_G2__Force = "wB97X-D/6-31g*/PCM(G2) Force"
    wB97X_D_6_31g_d_PCM_G2__Geometry_Optimization = "wB97X-D/6-31g*/PCM(G2) Geometry Optimization"
    wB97X_D_6_31g_d_PCM_G2__Frequency_Analysis = "wB97X-D/6-31g*/PCM(G2) Frequency Analysis"
    wB97X_D_6_31g_d_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_G2__Unknown = "wB97X-D/6-31g*/PCM(G2) Unknown"
    wB97X_D_6_31g_d_PCM_DME__Single_Point = "wB97X-D/6-31g*/PCM(DME) Single Point"
    wB97X_D_6_31g_d_PCM_DME__Force = "wB97X-D/6-31g*/PCM(DME) Force"
    wB97X_D_6_31g_d_PCM_DME__Geometry_Optimization = "wB97X-D/6-31g*/PCM(DME) Geometry Optimization"
    wB97X_D_6_31g_d_PCM_DME__Frequency_Analysis = "wB97X-D/6-31g*/PCM(DME) Frequency Analysis"
    wB97X_D_6_31g_d_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_DME__Unknown = "wB97X-D/6-31g*/PCM(DME) Unknown"
    wB97X_D_6_31g_d_PCM_EC__Single_Point = "wB97X-D/6-31g*/PCM(EC) Single Point"
    wB97X_D_6_31g_d_PCM_EC__Force = "wB97X-D/6-31g*/PCM(EC) Force"
    wB97X_D_6_31g_d_PCM_EC__Geometry_Optimization = "wB97X-D/6-31g*/PCM(EC) Geometry Optimization"
    wB97X_D_6_31g_d_PCM_EC__Frequency_Analysis = "wB97X-D/6-31g*/PCM(EC) Frequency Analysis"
    wB97X_D_6_31g_d_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_EC__Unknown = "wB97X-D/6-31g*/PCM(EC) Unknown"
    wB97X_D_6_31g_d_PCM_EC_EMC__Single_Point = "wB97X-D/6-31g*/PCM(EC/EMC) Single Point"
    wB97X_D_6_31g_d_PCM_EC_EMC__Force = "wB97X-D/6-31g*/PCM(EC/EMC) Force"
    wB97X_D_6_31g_d_PCM_EC_EMC__Geometry_Optimization = "wB97X-D/6-31g*/PCM(EC/EMC) Geometry Optimization"
    wB97X_D_6_31g_d_PCM_EC_EMC__Frequency_Analysis = "wB97X-D/6-31g*/PCM(EC/EMC) Frequency Analysis"
    wB97X_D_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_EC_EMC__Unknown = "wB97X-D/6-31g*/PCM(EC/EMC) Unknown"
    wB97X_D_6_31g_d_PCM_PC__Single_Point = "wB97X-D/6-31g*/PCM(PC) Single Point"
    wB97X_D_6_31g_d_PCM_PC__Force = "wB97X-D/6-31g*/PCM(PC) Force"
    wB97X_D_6_31g_d_PCM_PC__Geometry_Optimization = "wB97X-D/6-31g*/PCM(PC) Geometry Optimization"
    wB97X_D_6_31g_d_PCM_PC__Frequency_Analysis = "wB97X-D/6-31g*/PCM(PC) Frequency Analysis"
    wB97X_D_6_31g_d_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_PCM_PC__Unknown = "wB97X-D/6-31g*/PCM(PC) Unknown"
    wB97X_D_6_31g_d_ISOSVP_WATER__Single_Point = "wB97X-D/6-31g*/ISOSVP(WATER) Single Point"
    wB97X_D_6_31g_d_ISOSVP_WATER__Force = "wB97X-D/6-31g*/ISOSVP(WATER) Force"
    wB97X_D_6_31g_d_ISOSVP_WATER__Geometry_Optimization = "wB97X-D/6-31g*/ISOSVP(WATER) Geometry Optimization"
    wB97X_D_6_31g_d_ISOSVP_WATER__Frequency_Analysis = "wB97X-D/6-31g*/ISOSVP(WATER) Frequency Analysis"
    wB97X_D_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_WATER__Unknown = "wB97X-D/6-31g*/ISOSVP(WATER) Unknown"
    wB97X_D_6_31g_d_ISOSVP_THF__Single_Point = "wB97X-D/6-31g*/ISOSVP(THF) Single Point"
    wB97X_D_6_31g_d_ISOSVP_THF__Force = "wB97X-D/6-31g*/ISOSVP(THF) Force"
    wB97X_D_6_31g_d_ISOSVP_THF__Geometry_Optimization = "wB97X-D/6-31g*/ISOSVP(THF) Geometry Optimization"
    wB97X_D_6_31g_d_ISOSVP_THF__Frequency_Analysis = "wB97X-D/6-31g*/ISOSVP(THF) Frequency Analysis"
    wB97X_D_6_31g_d_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_THF__Unknown = "wB97X-D/6-31g*/ISOSVP(THF) Unknown"
    wB97X_D_6_31g_d_ISOSVP_G2__Single_Point = "wB97X-D/6-31g*/ISOSVP(G2) Single Point"
    wB97X_D_6_31g_d_ISOSVP_G2__Force = "wB97X-D/6-31g*/ISOSVP(G2) Force"
    wB97X_D_6_31g_d_ISOSVP_G2__Geometry_Optimization = "wB97X-D/6-31g*/ISOSVP(G2) Geometry Optimization"
    wB97X_D_6_31g_d_ISOSVP_G2__Frequency_Analysis = "wB97X-D/6-31g*/ISOSVP(G2) Frequency Analysis"
    wB97X_D_6_31g_d_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_G2__Unknown = "wB97X-D/6-31g*/ISOSVP(G2) Unknown"
    wB97X_D_6_31g_d_ISOSVP_DME__Single_Point = "wB97X-D/6-31g*/ISOSVP(DME) Single Point"
    wB97X_D_6_31g_d_ISOSVP_DME__Force = "wB97X-D/6-31g*/ISOSVP(DME) Force"
    wB97X_D_6_31g_d_ISOSVP_DME__Geometry_Optimization = "wB97X-D/6-31g*/ISOSVP(DME) Geometry Optimization"
    wB97X_D_6_31g_d_ISOSVP_DME__Frequency_Analysis = "wB97X-D/6-31g*/ISOSVP(DME) Frequency Analysis"
    wB97X_D_6_31g_d_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_DME__Unknown = "wB97X-D/6-31g*/ISOSVP(DME) Unknown"
    wB97X_D_6_31g_d_ISOSVP_EC__Single_Point = "wB97X-D/6-31g*/ISOSVP(EC) Single Point"
    wB97X_D_6_31g_d_ISOSVP_EC__Force = "wB97X-D/6-31g*/ISOSVP(EC) Force"
    wB97X_D_6_31g_d_ISOSVP_EC__Geometry_Optimization = "wB97X-D/6-31g*/ISOSVP(EC) Geometry Optimization"
    wB97X_D_6_31g_d_ISOSVP_EC__Frequency_Analysis = "wB97X-D/6-31g*/ISOSVP(EC) Frequency Analysis"
    wB97X_D_6_31g_d_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_EC__Unknown = "wB97X-D/6-31g*/ISOSVP(EC) Unknown"
    wB97X_D_6_31g_d_ISOSVP_EC_EMC__Single_Point = "wB97X-D/6-31g*/ISOSVP(EC/EMC) Single Point"
    wB97X_D_6_31g_d_ISOSVP_EC_EMC__Force = "wB97X-D/6-31g*/ISOSVP(EC/EMC) Force"
    wB97X_D_6_31g_d_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-D/6-31g*/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_D_6_31g_d_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D/6-31g*/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_EC_EMC__Unknown = "wB97X-D/6-31g*/ISOSVP(EC/EMC) Unknown"
    wB97X_D_6_31g_d_ISOSVP_PC__Single_Point = "wB97X-D/6-31g*/ISOSVP(PC) Single Point"
    wB97X_D_6_31g_d_ISOSVP_PC__Force = "wB97X-D/6-31g*/ISOSVP(PC) Force"
    wB97X_D_6_31g_d_ISOSVP_PC__Geometry_Optimization = "wB97X-D/6-31g*/ISOSVP(PC) Geometry Optimization"
    wB97X_D_6_31g_d_ISOSVP_PC__Frequency_Analysis = "wB97X-D/6-31g*/ISOSVP(PC) Frequency Analysis"
    wB97X_D_6_31g_d_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_ISOSVP_PC__Unknown = "wB97X-D/6-31g*/ISOSVP(PC) Unknown"
    wB97X_D_6_31g_d_SMD_WATER__Single_Point = "wB97X-D/6-31g*/SMD(WATER) Single Point"
    wB97X_D_6_31g_d_SMD_WATER__Force = "wB97X-D/6-31g*/SMD(WATER) Force"
    wB97X_D_6_31g_d_SMD_WATER__Geometry_Optimization = "wB97X-D/6-31g*/SMD(WATER) Geometry Optimization"
    wB97X_D_6_31g_d_SMD_WATER__Frequency_Analysis = "wB97X-D/6-31g*/SMD(WATER) Frequency Analysis"
    wB97X_D_6_31g_d_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_WATER__Unknown = "wB97X-D/6-31g*/SMD(WATER) Unknown"
    wB97X_D_6_31g_d_SMD_THF__Single_Point = "wB97X-D/6-31g*/SMD(THF) Single Point"
    wB97X_D_6_31g_d_SMD_THF__Force = "wB97X-D/6-31g*/SMD(THF) Force"
    wB97X_D_6_31g_d_SMD_THF__Geometry_Optimization = "wB97X-D/6-31g*/SMD(THF) Geometry Optimization"
    wB97X_D_6_31g_d_SMD_THF__Frequency_Analysis = "wB97X-D/6-31g*/SMD(THF) Frequency Analysis"
    wB97X_D_6_31g_d_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_THF__Unknown = "wB97X-D/6-31g*/SMD(THF) Unknown"
    wB97X_D_6_31g_d_SMD_G2__Single_Point = "wB97X-D/6-31g*/SMD(G2) Single Point"
    wB97X_D_6_31g_d_SMD_G2__Force = "wB97X-D/6-31g*/SMD(G2) Force"
    wB97X_D_6_31g_d_SMD_G2__Geometry_Optimization = "wB97X-D/6-31g*/SMD(G2) Geometry Optimization"
    wB97X_D_6_31g_d_SMD_G2__Frequency_Analysis = "wB97X-D/6-31g*/SMD(G2) Frequency Analysis"
    wB97X_D_6_31g_d_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_G2__Unknown = "wB97X-D/6-31g*/SMD(G2) Unknown"
    wB97X_D_6_31g_d_SMD_DME__Single_Point = "wB97X-D/6-31g*/SMD(DME) Single Point"
    wB97X_D_6_31g_d_SMD_DME__Force = "wB97X-D/6-31g*/SMD(DME) Force"
    wB97X_D_6_31g_d_SMD_DME__Geometry_Optimization = "wB97X-D/6-31g*/SMD(DME) Geometry Optimization"
    wB97X_D_6_31g_d_SMD_DME__Frequency_Analysis = "wB97X-D/6-31g*/SMD(DME) Frequency Analysis"
    wB97X_D_6_31g_d_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_DME__Unknown = "wB97X-D/6-31g*/SMD(DME) Unknown"
    wB97X_D_6_31g_d_SMD_EC__Single_Point = "wB97X-D/6-31g*/SMD(EC) Single Point"
    wB97X_D_6_31g_d_SMD_EC__Force = "wB97X-D/6-31g*/SMD(EC) Force"
    wB97X_D_6_31g_d_SMD_EC__Geometry_Optimization = "wB97X-D/6-31g*/SMD(EC) Geometry Optimization"
    wB97X_D_6_31g_d_SMD_EC__Frequency_Analysis = "wB97X-D/6-31g*/SMD(EC) Frequency Analysis"
    wB97X_D_6_31g_d_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_EC__Unknown = "wB97X-D/6-31g*/SMD(EC) Unknown"
    wB97X_D_6_31g_d_SMD_EC_EMC__Single_Point = "wB97X-D/6-31g*/SMD(EC/EMC) Single Point"
    wB97X_D_6_31g_d_SMD_EC_EMC__Force = "wB97X-D/6-31g*/SMD(EC/EMC) Force"
    wB97X_D_6_31g_d_SMD_EC_EMC__Geometry_Optimization = "wB97X-D/6-31g*/SMD(EC/EMC) Geometry Optimization"
    wB97X_D_6_31g_d_SMD_EC_EMC__Frequency_Analysis = "wB97X-D/6-31g*/SMD(EC/EMC) Frequency Analysis"
    wB97X_D_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_EC_EMC__Unknown = "wB97X-D/6-31g*/SMD(EC/EMC) Unknown"
    wB97X_D_6_31g_d_SMD_PC__Single_Point = "wB97X-D/6-31g*/SMD(PC) Single Point"
    wB97X_D_6_31g_d_SMD_PC__Force = "wB97X-D/6-31g*/SMD(PC) Force"
    wB97X_D_6_31g_d_SMD_PC__Geometry_Optimization = "wB97X-D/6-31g*/SMD(PC) Geometry Optimization"
    wB97X_D_6_31g_d_SMD_PC__Frequency_Analysis = "wB97X-D/6-31g*/SMD(PC) Frequency Analysis"
    wB97X_D_6_31g_d_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_SMD_PC__Unknown = "wB97X-D/6-31g*/SMD(PC) Unknown"
    wB97X_D_6_31g_d_CMIRS_WATER__Single_Point = "wB97X-D/6-31g*/CMIRS(WATER) Single Point"
    wB97X_D_6_31g_d_CMIRS_WATER__Force = "wB97X-D/6-31g*/CMIRS(WATER) Force"
    wB97X_D_6_31g_d_CMIRS_WATER__Geometry_Optimization = "wB97X-D/6-31g*/CMIRS(WATER) Geometry Optimization"
    wB97X_D_6_31g_d_CMIRS_WATER__Frequency_Analysis = "wB97X-D/6-31g*/CMIRS(WATER) Frequency Analysis"
    wB97X_D_6_31g_d_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_WATER__Unknown = "wB97X-D/6-31g*/CMIRS(WATER) Unknown"
    wB97X_D_6_31g_d_CMIRS_THF__Single_Point = "wB97X-D/6-31g*/CMIRS(THF) Single Point"
    wB97X_D_6_31g_d_CMIRS_THF__Force = "wB97X-D/6-31g*/CMIRS(THF) Force"
    wB97X_D_6_31g_d_CMIRS_THF__Geometry_Optimization = "wB97X-D/6-31g*/CMIRS(THF) Geometry Optimization"
    wB97X_D_6_31g_d_CMIRS_THF__Frequency_Analysis = "wB97X-D/6-31g*/CMIRS(THF) Frequency Analysis"
    wB97X_D_6_31g_d_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_THF__Unknown = "wB97X-D/6-31g*/CMIRS(THF) Unknown"
    wB97X_D_6_31g_d_CMIRS_G2__Single_Point = "wB97X-D/6-31g*/CMIRS(G2) Single Point"
    wB97X_D_6_31g_d_CMIRS_G2__Force = "wB97X-D/6-31g*/CMIRS(G2) Force"
    wB97X_D_6_31g_d_CMIRS_G2__Geometry_Optimization = "wB97X-D/6-31g*/CMIRS(G2) Geometry Optimization"
    wB97X_D_6_31g_d_CMIRS_G2__Frequency_Analysis = "wB97X-D/6-31g*/CMIRS(G2) Frequency Analysis"
    wB97X_D_6_31g_d_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_G2__Unknown = "wB97X-D/6-31g*/CMIRS(G2) Unknown"
    wB97X_D_6_31g_d_CMIRS_DME__Single_Point = "wB97X-D/6-31g*/CMIRS(DME) Single Point"
    wB97X_D_6_31g_d_CMIRS_DME__Force = "wB97X-D/6-31g*/CMIRS(DME) Force"
    wB97X_D_6_31g_d_CMIRS_DME__Geometry_Optimization = "wB97X-D/6-31g*/CMIRS(DME) Geometry Optimization"
    wB97X_D_6_31g_d_CMIRS_DME__Frequency_Analysis = "wB97X-D/6-31g*/CMIRS(DME) Frequency Analysis"
    wB97X_D_6_31g_d_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_DME__Unknown = "wB97X-D/6-31g*/CMIRS(DME) Unknown"
    wB97X_D_6_31g_d_CMIRS_EC__Single_Point = "wB97X-D/6-31g*/CMIRS(EC) Single Point"
    wB97X_D_6_31g_d_CMIRS_EC__Force = "wB97X-D/6-31g*/CMIRS(EC) Force"
    wB97X_D_6_31g_d_CMIRS_EC__Geometry_Optimization = "wB97X-D/6-31g*/CMIRS(EC) Geometry Optimization"
    wB97X_D_6_31g_d_CMIRS_EC__Frequency_Analysis = "wB97X-D/6-31g*/CMIRS(EC) Frequency Analysis"
    wB97X_D_6_31g_d_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_EC__Unknown = "wB97X-D/6-31g*/CMIRS(EC) Unknown"
    wB97X_D_6_31g_d_CMIRS_EC_EMC__Single_Point = "wB97X-D/6-31g*/CMIRS(EC/EMC) Single Point"
    wB97X_D_6_31g_d_CMIRS_EC_EMC__Force = "wB97X-D/6-31g*/CMIRS(EC/EMC) Force"
    wB97X_D_6_31g_d_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D/6-31g*/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D_6_31g_d_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D/6-31g*/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_EC_EMC__Unknown = "wB97X-D/6-31g*/CMIRS(EC/EMC) Unknown"
    wB97X_D_6_31g_d_CMIRS_PC__Single_Point = "wB97X-D/6-31g*/CMIRS(PC) Single Point"
    wB97X_D_6_31g_d_CMIRS_PC__Force = "wB97X-D/6-31g*/CMIRS(PC) Force"
    wB97X_D_6_31g_d_CMIRS_PC__Geometry_Optimization = "wB97X-D/6-31g*/CMIRS(PC) Geometry Optimization"
    wB97X_D_6_31g_d_CMIRS_PC__Frequency_Analysis = "wB97X-D/6-31g*/CMIRS(PC) Frequency Analysis"
    wB97X_D_6_31g_d_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/6-31g*/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_6_31g_d_CMIRS_PC__Unknown = "wB97X-D/6-31g*/CMIRS(PC) Unknown"
    wB97X_D_def2_SVPD_VACUUM_Single_Point = "wB97X-D/def2-SVPD/VACUUM Single Point"
    wB97X_D_def2_SVPD_VACUUM_Force = "wB97X-D/def2-SVPD/VACUUM Force"
    wB97X_D_def2_SVPD_VACUUM_Geometry_Optimization = "wB97X-D/def2-SVPD/VACUUM Geometry Optimization"
    wB97X_D_def2_SVPD_VACUUM_Frequency_Analysis = "wB97X-D/def2-SVPD/VACUUM Frequency Analysis"
    wB97X_D_def2_SVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_VACUUM_Unknown = "wB97X-D/def2-SVPD/VACUUM Unknown"
    wB97X_D_def2_SVPD_PCM_WATER__Single_Point = "wB97X-D/def2-SVPD/PCM(WATER) Single Point"
    wB97X_D_def2_SVPD_PCM_WATER__Force = "wB97X-D/def2-SVPD/PCM(WATER) Force"
    wB97X_D_def2_SVPD_PCM_WATER__Geometry_Optimization = "wB97X-D/def2-SVPD/PCM(WATER) Geometry Optimization"
    wB97X_D_def2_SVPD_PCM_WATER__Frequency_Analysis = "wB97X-D/def2-SVPD/PCM(WATER) Frequency Analysis"
    wB97X_D_def2_SVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_WATER__Unknown = "wB97X-D/def2-SVPD/PCM(WATER) Unknown"
    wB97X_D_def2_SVPD_PCM_THF__Single_Point = "wB97X-D/def2-SVPD/PCM(THF) Single Point"
    wB97X_D_def2_SVPD_PCM_THF__Force = "wB97X-D/def2-SVPD/PCM(THF) Force"
    wB97X_D_def2_SVPD_PCM_THF__Geometry_Optimization = "wB97X-D/def2-SVPD/PCM(THF) Geometry Optimization"
    wB97X_D_def2_SVPD_PCM_THF__Frequency_Analysis = "wB97X-D/def2-SVPD/PCM(THF) Frequency Analysis"
    wB97X_D_def2_SVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_THF__Unknown = "wB97X-D/def2-SVPD/PCM(THF) Unknown"
    wB97X_D_def2_SVPD_PCM_G2__Single_Point = "wB97X-D/def2-SVPD/PCM(G2) Single Point"
    wB97X_D_def2_SVPD_PCM_G2__Force = "wB97X-D/def2-SVPD/PCM(G2) Force"
    wB97X_D_def2_SVPD_PCM_G2__Geometry_Optimization = "wB97X-D/def2-SVPD/PCM(G2) Geometry Optimization"
    wB97X_D_def2_SVPD_PCM_G2__Frequency_Analysis = "wB97X-D/def2-SVPD/PCM(G2) Frequency Analysis"
    wB97X_D_def2_SVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_G2__Unknown = "wB97X-D/def2-SVPD/PCM(G2) Unknown"
    wB97X_D_def2_SVPD_PCM_DME__Single_Point = "wB97X-D/def2-SVPD/PCM(DME) Single Point"
    wB97X_D_def2_SVPD_PCM_DME__Force = "wB97X-D/def2-SVPD/PCM(DME) Force"
    wB97X_D_def2_SVPD_PCM_DME__Geometry_Optimization = "wB97X-D/def2-SVPD/PCM(DME) Geometry Optimization"
    wB97X_D_def2_SVPD_PCM_DME__Frequency_Analysis = "wB97X-D/def2-SVPD/PCM(DME) Frequency Analysis"
    wB97X_D_def2_SVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_DME__Unknown = "wB97X-D/def2-SVPD/PCM(DME) Unknown"
    wB97X_D_def2_SVPD_PCM_EC__Single_Point = "wB97X-D/def2-SVPD/PCM(EC) Single Point"
    wB97X_D_def2_SVPD_PCM_EC__Force = "wB97X-D/def2-SVPD/PCM(EC) Force"
    wB97X_D_def2_SVPD_PCM_EC__Geometry_Optimization = "wB97X-D/def2-SVPD/PCM(EC) Geometry Optimization"
    wB97X_D_def2_SVPD_PCM_EC__Frequency_Analysis = "wB97X-D/def2-SVPD/PCM(EC) Frequency Analysis"
    wB97X_D_def2_SVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_EC__Unknown = "wB97X-D/def2-SVPD/PCM(EC) Unknown"
    wB97X_D_def2_SVPD_PCM_EC_EMC__Single_Point = "wB97X-D/def2-SVPD/PCM(EC/EMC) Single Point"
    wB97X_D_def2_SVPD_PCM_EC_EMC__Force = "wB97X-D/def2-SVPD/PCM(EC/EMC) Force"
    wB97X_D_def2_SVPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-D/def2-SVPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_D_def2_SVPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-D/def2-SVPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_D_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_EC_EMC__Unknown = "wB97X-D/def2-SVPD/PCM(EC/EMC) Unknown"
    wB97X_D_def2_SVPD_PCM_PC__Single_Point = "wB97X-D/def2-SVPD/PCM(PC) Single Point"
    wB97X_D_def2_SVPD_PCM_PC__Force = "wB97X-D/def2-SVPD/PCM(PC) Force"
    wB97X_D_def2_SVPD_PCM_PC__Geometry_Optimization = "wB97X-D/def2-SVPD/PCM(PC) Geometry Optimization"
    wB97X_D_def2_SVPD_PCM_PC__Frequency_Analysis = "wB97X-D/def2-SVPD/PCM(PC) Frequency Analysis"
    wB97X_D_def2_SVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_PCM_PC__Unknown = "wB97X-D/def2-SVPD/PCM(PC) Unknown"
    wB97X_D_def2_SVPD_ISOSVP_WATER__Single_Point = "wB97X-D/def2-SVPD/ISOSVP(WATER) Single Point"
    wB97X_D_def2_SVPD_ISOSVP_WATER__Force = "wB97X-D/def2-SVPD/ISOSVP(WATER) Force"
    wB97X_D_def2_SVPD_ISOSVP_WATER__Geometry_Optimization = "wB97X-D/def2-SVPD/ISOSVP(WATER) Geometry Optimization"
    wB97X_D_def2_SVPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-D/def2-SVPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_D_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_WATER__Unknown = "wB97X-D/def2-SVPD/ISOSVP(WATER) Unknown"
    wB97X_D_def2_SVPD_ISOSVP_THF__Single_Point = "wB97X-D/def2-SVPD/ISOSVP(THF) Single Point"
    wB97X_D_def2_SVPD_ISOSVP_THF__Force = "wB97X-D/def2-SVPD/ISOSVP(THF) Force"
    wB97X_D_def2_SVPD_ISOSVP_THF__Geometry_Optimization = "wB97X-D/def2-SVPD/ISOSVP(THF) Geometry Optimization"
    wB97X_D_def2_SVPD_ISOSVP_THF__Frequency_Analysis = "wB97X-D/def2-SVPD/ISOSVP(THF) Frequency Analysis"
    wB97X_D_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_THF__Unknown = "wB97X-D/def2-SVPD/ISOSVP(THF) Unknown"
    wB97X_D_def2_SVPD_ISOSVP_G2__Single_Point = "wB97X-D/def2-SVPD/ISOSVP(G2) Single Point"
    wB97X_D_def2_SVPD_ISOSVP_G2__Force = "wB97X-D/def2-SVPD/ISOSVP(G2) Force"
    wB97X_D_def2_SVPD_ISOSVP_G2__Geometry_Optimization = "wB97X-D/def2-SVPD/ISOSVP(G2) Geometry Optimization"
    wB97X_D_def2_SVPD_ISOSVP_G2__Frequency_Analysis = "wB97X-D/def2-SVPD/ISOSVP(G2) Frequency Analysis"
    wB97X_D_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_G2__Unknown = "wB97X-D/def2-SVPD/ISOSVP(G2) Unknown"
    wB97X_D_def2_SVPD_ISOSVP_DME__Single_Point = "wB97X-D/def2-SVPD/ISOSVP(DME) Single Point"
    wB97X_D_def2_SVPD_ISOSVP_DME__Force = "wB97X-D/def2-SVPD/ISOSVP(DME) Force"
    wB97X_D_def2_SVPD_ISOSVP_DME__Geometry_Optimization = "wB97X-D/def2-SVPD/ISOSVP(DME) Geometry Optimization"
    wB97X_D_def2_SVPD_ISOSVP_DME__Frequency_Analysis = "wB97X-D/def2-SVPD/ISOSVP(DME) Frequency Analysis"
    wB97X_D_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_DME__Unknown = "wB97X-D/def2-SVPD/ISOSVP(DME) Unknown"
    wB97X_D_def2_SVPD_ISOSVP_EC__Single_Point = "wB97X-D/def2-SVPD/ISOSVP(EC) Single Point"
    wB97X_D_def2_SVPD_ISOSVP_EC__Force = "wB97X-D/def2-SVPD/ISOSVP(EC) Force"
    wB97X_D_def2_SVPD_ISOSVP_EC__Geometry_Optimization = "wB97X-D/def2-SVPD/ISOSVP(EC) Geometry Optimization"
    wB97X_D_def2_SVPD_ISOSVP_EC__Frequency_Analysis = "wB97X-D/def2-SVPD/ISOSVP(EC) Frequency Analysis"
    wB97X_D_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_EC__Unknown = "wB97X-D/def2-SVPD/ISOSVP(EC) Unknown"
    wB97X_D_def2_SVPD_ISOSVP_EC_EMC__Single_Point = "wB97X-D/def2-SVPD/ISOSVP(EC/EMC) Single Point"
    wB97X_D_def2_SVPD_ISOSVP_EC_EMC__Force = "wB97X-D/def2-SVPD/ISOSVP(EC/EMC) Force"
    wB97X_D_def2_SVPD_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-D/def2-SVPD/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_D_def2_SVPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D/def2-SVPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_EC_EMC__Unknown = "wB97X-D/def2-SVPD/ISOSVP(EC/EMC) Unknown"
    wB97X_D_def2_SVPD_ISOSVP_PC__Single_Point = "wB97X-D/def2-SVPD/ISOSVP(PC) Single Point"
    wB97X_D_def2_SVPD_ISOSVP_PC__Force = "wB97X-D/def2-SVPD/ISOSVP(PC) Force"
    wB97X_D_def2_SVPD_ISOSVP_PC__Geometry_Optimization = "wB97X-D/def2-SVPD/ISOSVP(PC) Geometry Optimization"
    wB97X_D_def2_SVPD_ISOSVP_PC__Frequency_Analysis = "wB97X-D/def2-SVPD/ISOSVP(PC) Frequency Analysis"
    wB97X_D_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_ISOSVP_PC__Unknown = "wB97X-D/def2-SVPD/ISOSVP(PC) Unknown"
    wB97X_D_def2_SVPD_SMD_WATER__Single_Point = "wB97X-D/def2-SVPD/SMD(WATER) Single Point"
    wB97X_D_def2_SVPD_SMD_WATER__Force = "wB97X-D/def2-SVPD/SMD(WATER) Force"
    wB97X_D_def2_SVPD_SMD_WATER__Geometry_Optimization = "wB97X-D/def2-SVPD/SMD(WATER) Geometry Optimization"
    wB97X_D_def2_SVPD_SMD_WATER__Frequency_Analysis = "wB97X-D/def2-SVPD/SMD(WATER) Frequency Analysis"
    wB97X_D_def2_SVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_WATER__Unknown = "wB97X-D/def2-SVPD/SMD(WATER) Unknown"
    wB97X_D_def2_SVPD_SMD_THF__Single_Point = "wB97X-D/def2-SVPD/SMD(THF) Single Point"
    wB97X_D_def2_SVPD_SMD_THF__Force = "wB97X-D/def2-SVPD/SMD(THF) Force"
    wB97X_D_def2_SVPD_SMD_THF__Geometry_Optimization = "wB97X-D/def2-SVPD/SMD(THF) Geometry Optimization"
    wB97X_D_def2_SVPD_SMD_THF__Frequency_Analysis = "wB97X-D/def2-SVPD/SMD(THF) Frequency Analysis"
    wB97X_D_def2_SVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_THF__Unknown = "wB97X-D/def2-SVPD/SMD(THF) Unknown"
    wB97X_D_def2_SVPD_SMD_G2__Single_Point = "wB97X-D/def2-SVPD/SMD(G2) Single Point"
    wB97X_D_def2_SVPD_SMD_G2__Force = "wB97X-D/def2-SVPD/SMD(G2) Force"
    wB97X_D_def2_SVPD_SMD_G2__Geometry_Optimization = "wB97X-D/def2-SVPD/SMD(G2) Geometry Optimization"
    wB97X_D_def2_SVPD_SMD_G2__Frequency_Analysis = "wB97X-D/def2-SVPD/SMD(G2) Frequency Analysis"
    wB97X_D_def2_SVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_G2__Unknown = "wB97X-D/def2-SVPD/SMD(G2) Unknown"
    wB97X_D_def2_SVPD_SMD_DME__Single_Point = "wB97X-D/def2-SVPD/SMD(DME) Single Point"
    wB97X_D_def2_SVPD_SMD_DME__Force = "wB97X-D/def2-SVPD/SMD(DME) Force"
    wB97X_D_def2_SVPD_SMD_DME__Geometry_Optimization = "wB97X-D/def2-SVPD/SMD(DME) Geometry Optimization"
    wB97X_D_def2_SVPD_SMD_DME__Frequency_Analysis = "wB97X-D/def2-SVPD/SMD(DME) Frequency Analysis"
    wB97X_D_def2_SVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_DME__Unknown = "wB97X-D/def2-SVPD/SMD(DME) Unknown"
    wB97X_D_def2_SVPD_SMD_EC__Single_Point = "wB97X-D/def2-SVPD/SMD(EC) Single Point"
    wB97X_D_def2_SVPD_SMD_EC__Force = "wB97X-D/def2-SVPD/SMD(EC) Force"
    wB97X_D_def2_SVPD_SMD_EC__Geometry_Optimization = "wB97X-D/def2-SVPD/SMD(EC) Geometry Optimization"
    wB97X_D_def2_SVPD_SMD_EC__Frequency_Analysis = "wB97X-D/def2-SVPD/SMD(EC) Frequency Analysis"
    wB97X_D_def2_SVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_EC__Unknown = "wB97X-D/def2-SVPD/SMD(EC) Unknown"
    wB97X_D_def2_SVPD_SMD_EC_EMC__Single_Point = "wB97X-D/def2-SVPD/SMD(EC/EMC) Single Point"
    wB97X_D_def2_SVPD_SMD_EC_EMC__Force = "wB97X-D/def2-SVPD/SMD(EC/EMC) Force"
    wB97X_D_def2_SVPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-D/def2-SVPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_D_def2_SVPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-D/def2-SVPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_D_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_EC_EMC__Unknown = "wB97X-D/def2-SVPD/SMD(EC/EMC) Unknown"
    wB97X_D_def2_SVPD_SMD_PC__Single_Point = "wB97X-D/def2-SVPD/SMD(PC) Single Point"
    wB97X_D_def2_SVPD_SMD_PC__Force = "wB97X-D/def2-SVPD/SMD(PC) Force"
    wB97X_D_def2_SVPD_SMD_PC__Geometry_Optimization = "wB97X-D/def2-SVPD/SMD(PC) Geometry Optimization"
    wB97X_D_def2_SVPD_SMD_PC__Frequency_Analysis = "wB97X-D/def2-SVPD/SMD(PC) Frequency Analysis"
    wB97X_D_def2_SVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_SMD_PC__Unknown = "wB97X-D/def2-SVPD/SMD(PC) Unknown"
    wB97X_D_def2_SVPD_CMIRS_WATER__Single_Point = "wB97X-D/def2-SVPD/CMIRS(WATER) Single Point"
    wB97X_D_def2_SVPD_CMIRS_WATER__Force = "wB97X-D/def2-SVPD/CMIRS(WATER) Force"
    wB97X_D_def2_SVPD_CMIRS_WATER__Geometry_Optimization = "wB97X-D/def2-SVPD/CMIRS(WATER) Geometry Optimization"
    wB97X_D_def2_SVPD_CMIRS_WATER__Frequency_Analysis = "wB97X-D/def2-SVPD/CMIRS(WATER) Frequency Analysis"
    wB97X_D_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_WATER__Unknown = "wB97X-D/def2-SVPD/CMIRS(WATER) Unknown"
    wB97X_D_def2_SVPD_CMIRS_THF__Single_Point = "wB97X-D/def2-SVPD/CMIRS(THF) Single Point"
    wB97X_D_def2_SVPD_CMIRS_THF__Force = "wB97X-D/def2-SVPD/CMIRS(THF) Force"
    wB97X_D_def2_SVPD_CMIRS_THF__Geometry_Optimization = "wB97X-D/def2-SVPD/CMIRS(THF) Geometry Optimization"
    wB97X_D_def2_SVPD_CMIRS_THF__Frequency_Analysis = "wB97X-D/def2-SVPD/CMIRS(THF) Frequency Analysis"
    wB97X_D_def2_SVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_THF__Unknown = "wB97X-D/def2-SVPD/CMIRS(THF) Unknown"
    wB97X_D_def2_SVPD_CMIRS_G2__Single_Point = "wB97X-D/def2-SVPD/CMIRS(G2) Single Point"
    wB97X_D_def2_SVPD_CMIRS_G2__Force = "wB97X-D/def2-SVPD/CMIRS(G2) Force"
    wB97X_D_def2_SVPD_CMIRS_G2__Geometry_Optimization = "wB97X-D/def2-SVPD/CMIRS(G2) Geometry Optimization"
    wB97X_D_def2_SVPD_CMIRS_G2__Frequency_Analysis = "wB97X-D/def2-SVPD/CMIRS(G2) Frequency Analysis"
    wB97X_D_def2_SVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_G2__Unknown = "wB97X-D/def2-SVPD/CMIRS(G2) Unknown"
    wB97X_D_def2_SVPD_CMIRS_DME__Single_Point = "wB97X-D/def2-SVPD/CMIRS(DME) Single Point"
    wB97X_D_def2_SVPD_CMIRS_DME__Force = "wB97X-D/def2-SVPD/CMIRS(DME) Force"
    wB97X_D_def2_SVPD_CMIRS_DME__Geometry_Optimization = "wB97X-D/def2-SVPD/CMIRS(DME) Geometry Optimization"
    wB97X_D_def2_SVPD_CMIRS_DME__Frequency_Analysis = "wB97X-D/def2-SVPD/CMIRS(DME) Frequency Analysis"
    wB97X_D_def2_SVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_DME__Unknown = "wB97X-D/def2-SVPD/CMIRS(DME) Unknown"
    wB97X_D_def2_SVPD_CMIRS_EC__Single_Point = "wB97X-D/def2-SVPD/CMIRS(EC) Single Point"
    wB97X_D_def2_SVPD_CMIRS_EC__Force = "wB97X-D/def2-SVPD/CMIRS(EC) Force"
    wB97X_D_def2_SVPD_CMIRS_EC__Geometry_Optimization = "wB97X-D/def2-SVPD/CMIRS(EC) Geometry Optimization"
    wB97X_D_def2_SVPD_CMIRS_EC__Frequency_Analysis = "wB97X-D/def2-SVPD/CMIRS(EC) Frequency Analysis"
    wB97X_D_def2_SVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_EC__Unknown = "wB97X-D/def2-SVPD/CMIRS(EC) Unknown"
    wB97X_D_def2_SVPD_CMIRS_EC_EMC__Single_Point = "wB97X-D/def2-SVPD/CMIRS(EC/EMC) Single Point"
    wB97X_D_def2_SVPD_CMIRS_EC_EMC__Force = "wB97X-D/def2-SVPD/CMIRS(EC/EMC) Force"
    wB97X_D_def2_SVPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D/def2-SVPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D_def2_SVPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D/def2-SVPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_EC_EMC__Unknown = "wB97X-D/def2-SVPD/CMIRS(EC/EMC) Unknown"
    wB97X_D_def2_SVPD_CMIRS_PC__Single_Point = "wB97X-D/def2-SVPD/CMIRS(PC) Single Point"
    wB97X_D_def2_SVPD_CMIRS_PC__Force = "wB97X-D/def2-SVPD/CMIRS(PC) Force"
    wB97X_D_def2_SVPD_CMIRS_PC__Geometry_Optimization = "wB97X-D/def2-SVPD/CMIRS(PC) Geometry Optimization"
    wB97X_D_def2_SVPD_CMIRS_PC__Frequency_Analysis = "wB97X-D/def2-SVPD/CMIRS(PC) Frequency Analysis"
    wB97X_D_def2_SVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-SVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_SVPD_CMIRS_PC__Unknown = "wB97X-D/def2-SVPD/CMIRS(PC) Unknown"
    wB97X_D_def2_TZVP_VACUUM_Single_Point = "wB97X-D/def2-TZVP/VACUUM Single Point"
    wB97X_D_def2_TZVP_VACUUM_Force = "wB97X-D/def2-TZVP/VACUUM Force"
    wB97X_D_def2_TZVP_VACUUM_Geometry_Optimization = "wB97X-D/def2-TZVP/VACUUM Geometry Optimization"
    wB97X_D_def2_TZVP_VACUUM_Frequency_Analysis = "wB97X-D/def2-TZVP/VACUUM Frequency Analysis"
    wB97X_D_def2_TZVP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_VACUUM_Unknown = "wB97X-D/def2-TZVP/VACUUM Unknown"
    wB97X_D_def2_TZVP_PCM_WATER__Single_Point = "wB97X-D/def2-TZVP/PCM(WATER) Single Point"
    wB97X_D_def2_TZVP_PCM_WATER__Force = "wB97X-D/def2-TZVP/PCM(WATER) Force"
    wB97X_D_def2_TZVP_PCM_WATER__Geometry_Optimization = "wB97X-D/def2-TZVP/PCM(WATER) Geometry Optimization"
    wB97X_D_def2_TZVP_PCM_WATER__Frequency_Analysis = "wB97X-D/def2-TZVP/PCM(WATER) Frequency Analysis"
    wB97X_D_def2_TZVP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_WATER__Unknown = "wB97X-D/def2-TZVP/PCM(WATER) Unknown"
    wB97X_D_def2_TZVP_PCM_THF__Single_Point = "wB97X-D/def2-TZVP/PCM(THF) Single Point"
    wB97X_D_def2_TZVP_PCM_THF__Force = "wB97X-D/def2-TZVP/PCM(THF) Force"
    wB97X_D_def2_TZVP_PCM_THF__Geometry_Optimization = "wB97X-D/def2-TZVP/PCM(THF) Geometry Optimization"
    wB97X_D_def2_TZVP_PCM_THF__Frequency_Analysis = "wB97X-D/def2-TZVP/PCM(THF) Frequency Analysis"
    wB97X_D_def2_TZVP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_THF__Unknown = "wB97X-D/def2-TZVP/PCM(THF) Unknown"
    wB97X_D_def2_TZVP_PCM_G2__Single_Point = "wB97X-D/def2-TZVP/PCM(G2) Single Point"
    wB97X_D_def2_TZVP_PCM_G2__Force = "wB97X-D/def2-TZVP/PCM(G2) Force"
    wB97X_D_def2_TZVP_PCM_G2__Geometry_Optimization = "wB97X-D/def2-TZVP/PCM(G2) Geometry Optimization"
    wB97X_D_def2_TZVP_PCM_G2__Frequency_Analysis = "wB97X-D/def2-TZVP/PCM(G2) Frequency Analysis"
    wB97X_D_def2_TZVP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_G2__Unknown = "wB97X-D/def2-TZVP/PCM(G2) Unknown"
    wB97X_D_def2_TZVP_PCM_DME__Single_Point = "wB97X-D/def2-TZVP/PCM(DME) Single Point"
    wB97X_D_def2_TZVP_PCM_DME__Force = "wB97X-D/def2-TZVP/PCM(DME) Force"
    wB97X_D_def2_TZVP_PCM_DME__Geometry_Optimization = "wB97X-D/def2-TZVP/PCM(DME) Geometry Optimization"
    wB97X_D_def2_TZVP_PCM_DME__Frequency_Analysis = "wB97X-D/def2-TZVP/PCM(DME) Frequency Analysis"
    wB97X_D_def2_TZVP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_DME__Unknown = "wB97X-D/def2-TZVP/PCM(DME) Unknown"
    wB97X_D_def2_TZVP_PCM_EC__Single_Point = "wB97X-D/def2-TZVP/PCM(EC) Single Point"
    wB97X_D_def2_TZVP_PCM_EC__Force = "wB97X-D/def2-TZVP/PCM(EC) Force"
    wB97X_D_def2_TZVP_PCM_EC__Geometry_Optimization = "wB97X-D/def2-TZVP/PCM(EC) Geometry Optimization"
    wB97X_D_def2_TZVP_PCM_EC__Frequency_Analysis = "wB97X-D/def2-TZVP/PCM(EC) Frequency Analysis"
    wB97X_D_def2_TZVP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_EC__Unknown = "wB97X-D/def2-TZVP/PCM(EC) Unknown"
    wB97X_D_def2_TZVP_PCM_EC_EMC__Single_Point = "wB97X-D/def2-TZVP/PCM(EC/EMC) Single Point"
    wB97X_D_def2_TZVP_PCM_EC_EMC__Force = "wB97X-D/def2-TZVP/PCM(EC/EMC) Force"
    wB97X_D_def2_TZVP_PCM_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVP/PCM(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVP_PCM_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVP/PCM(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_EC_EMC__Unknown = "wB97X-D/def2-TZVP/PCM(EC/EMC) Unknown"
    wB97X_D_def2_TZVP_PCM_PC__Single_Point = "wB97X-D/def2-TZVP/PCM(PC) Single Point"
    wB97X_D_def2_TZVP_PCM_PC__Force = "wB97X-D/def2-TZVP/PCM(PC) Force"
    wB97X_D_def2_TZVP_PCM_PC__Geometry_Optimization = "wB97X-D/def2-TZVP/PCM(PC) Geometry Optimization"
    wB97X_D_def2_TZVP_PCM_PC__Frequency_Analysis = "wB97X-D/def2-TZVP/PCM(PC) Frequency Analysis"
    wB97X_D_def2_TZVP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_PCM_PC__Unknown = "wB97X-D/def2-TZVP/PCM(PC) Unknown"
    wB97X_D_def2_TZVP_ISOSVP_WATER__Single_Point = "wB97X-D/def2-TZVP/ISOSVP(WATER) Single Point"
    wB97X_D_def2_TZVP_ISOSVP_WATER__Force = "wB97X-D/def2-TZVP/ISOSVP(WATER) Force"
    wB97X_D_def2_TZVP_ISOSVP_WATER__Geometry_Optimization = "wB97X-D/def2-TZVP/ISOSVP(WATER) Geometry Optimization"
    wB97X_D_def2_TZVP_ISOSVP_WATER__Frequency_Analysis = "wB97X-D/def2-TZVP/ISOSVP(WATER) Frequency Analysis"
    wB97X_D_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_WATER__Unknown = "wB97X-D/def2-TZVP/ISOSVP(WATER) Unknown"
    wB97X_D_def2_TZVP_ISOSVP_THF__Single_Point = "wB97X-D/def2-TZVP/ISOSVP(THF) Single Point"
    wB97X_D_def2_TZVP_ISOSVP_THF__Force = "wB97X-D/def2-TZVP/ISOSVP(THF) Force"
    wB97X_D_def2_TZVP_ISOSVP_THF__Geometry_Optimization = "wB97X-D/def2-TZVP/ISOSVP(THF) Geometry Optimization"
    wB97X_D_def2_TZVP_ISOSVP_THF__Frequency_Analysis = "wB97X-D/def2-TZVP/ISOSVP(THF) Frequency Analysis"
    wB97X_D_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_THF__Unknown = "wB97X-D/def2-TZVP/ISOSVP(THF) Unknown"
    wB97X_D_def2_TZVP_ISOSVP_G2__Single_Point = "wB97X-D/def2-TZVP/ISOSVP(G2) Single Point"
    wB97X_D_def2_TZVP_ISOSVP_G2__Force = "wB97X-D/def2-TZVP/ISOSVP(G2) Force"
    wB97X_D_def2_TZVP_ISOSVP_G2__Geometry_Optimization = "wB97X-D/def2-TZVP/ISOSVP(G2) Geometry Optimization"
    wB97X_D_def2_TZVP_ISOSVP_G2__Frequency_Analysis = "wB97X-D/def2-TZVP/ISOSVP(G2) Frequency Analysis"
    wB97X_D_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_G2__Unknown = "wB97X-D/def2-TZVP/ISOSVP(G2) Unknown"
    wB97X_D_def2_TZVP_ISOSVP_DME__Single_Point = "wB97X-D/def2-TZVP/ISOSVP(DME) Single Point"
    wB97X_D_def2_TZVP_ISOSVP_DME__Force = "wB97X-D/def2-TZVP/ISOSVP(DME) Force"
    wB97X_D_def2_TZVP_ISOSVP_DME__Geometry_Optimization = "wB97X-D/def2-TZVP/ISOSVP(DME) Geometry Optimization"
    wB97X_D_def2_TZVP_ISOSVP_DME__Frequency_Analysis = "wB97X-D/def2-TZVP/ISOSVP(DME) Frequency Analysis"
    wB97X_D_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_DME__Unknown = "wB97X-D/def2-TZVP/ISOSVP(DME) Unknown"
    wB97X_D_def2_TZVP_ISOSVP_EC__Single_Point = "wB97X-D/def2-TZVP/ISOSVP(EC) Single Point"
    wB97X_D_def2_TZVP_ISOSVP_EC__Force = "wB97X-D/def2-TZVP/ISOSVP(EC) Force"
    wB97X_D_def2_TZVP_ISOSVP_EC__Geometry_Optimization = "wB97X-D/def2-TZVP/ISOSVP(EC) Geometry Optimization"
    wB97X_D_def2_TZVP_ISOSVP_EC__Frequency_Analysis = "wB97X-D/def2-TZVP/ISOSVP(EC) Frequency Analysis"
    wB97X_D_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_EC__Unknown = "wB97X-D/def2-TZVP/ISOSVP(EC) Unknown"
    wB97X_D_def2_TZVP_ISOSVP_EC_EMC__Single_Point = "wB97X-D/def2-TZVP/ISOSVP(EC/EMC) Single Point"
    wB97X_D_def2_TZVP_ISOSVP_EC_EMC__Force = "wB97X-D/def2-TZVP/ISOSVP(EC/EMC) Force"
    wB97X_D_def2_TZVP_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVP/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVP_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVP/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_EC_EMC__Unknown = "wB97X-D/def2-TZVP/ISOSVP(EC/EMC) Unknown"
    wB97X_D_def2_TZVP_ISOSVP_PC__Single_Point = "wB97X-D/def2-TZVP/ISOSVP(PC) Single Point"
    wB97X_D_def2_TZVP_ISOSVP_PC__Force = "wB97X-D/def2-TZVP/ISOSVP(PC) Force"
    wB97X_D_def2_TZVP_ISOSVP_PC__Geometry_Optimization = "wB97X-D/def2-TZVP/ISOSVP(PC) Geometry Optimization"
    wB97X_D_def2_TZVP_ISOSVP_PC__Frequency_Analysis = "wB97X-D/def2-TZVP/ISOSVP(PC) Frequency Analysis"
    wB97X_D_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_ISOSVP_PC__Unknown = "wB97X-D/def2-TZVP/ISOSVP(PC) Unknown"
    wB97X_D_def2_TZVP_SMD_WATER__Single_Point = "wB97X-D/def2-TZVP/SMD(WATER) Single Point"
    wB97X_D_def2_TZVP_SMD_WATER__Force = "wB97X-D/def2-TZVP/SMD(WATER) Force"
    wB97X_D_def2_TZVP_SMD_WATER__Geometry_Optimization = "wB97X-D/def2-TZVP/SMD(WATER) Geometry Optimization"
    wB97X_D_def2_TZVP_SMD_WATER__Frequency_Analysis = "wB97X-D/def2-TZVP/SMD(WATER) Frequency Analysis"
    wB97X_D_def2_TZVP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_WATER__Unknown = "wB97X-D/def2-TZVP/SMD(WATER) Unknown"
    wB97X_D_def2_TZVP_SMD_THF__Single_Point = "wB97X-D/def2-TZVP/SMD(THF) Single Point"
    wB97X_D_def2_TZVP_SMD_THF__Force = "wB97X-D/def2-TZVP/SMD(THF) Force"
    wB97X_D_def2_TZVP_SMD_THF__Geometry_Optimization = "wB97X-D/def2-TZVP/SMD(THF) Geometry Optimization"
    wB97X_D_def2_TZVP_SMD_THF__Frequency_Analysis = "wB97X-D/def2-TZVP/SMD(THF) Frequency Analysis"
    wB97X_D_def2_TZVP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_THF__Unknown = "wB97X-D/def2-TZVP/SMD(THF) Unknown"
    wB97X_D_def2_TZVP_SMD_G2__Single_Point = "wB97X-D/def2-TZVP/SMD(G2) Single Point"
    wB97X_D_def2_TZVP_SMD_G2__Force = "wB97X-D/def2-TZVP/SMD(G2) Force"
    wB97X_D_def2_TZVP_SMD_G2__Geometry_Optimization = "wB97X-D/def2-TZVP/SMD(G2) Geometry Optimization"
    wB97X_D_def2_TZVP_SMD_G2__Frequency_Analysis = "wB97X-D/def2-TZVP/SMD(G2) Frequency Analysis"
    wB97X_D_def2_TZVP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_G2__Unknown = "wB97X-D/def2-TZVP/SMD(G2) Unknown"
    wB97X_D_def2_TZVP_SMD_DME__Single_Point = "wB97X-D/def2-TZVP/SMD(DME) Single Point"
    wB97X_D_def2_TZVP_SMD_DME__Force = "wB97X-D/def2-TZVP/SMD(DME) Force"
    wB97X_D_def2_TZVP_SMD_DME__Geometry_Optimization = "wB97X-D/def2-TZVP/SMD(DME) Geometry Optimization"
    wB97X_D_def2_TZVP_SMD_DME__Frequency_Analysis = "wB97X-D/def2-TZVP/SMD(DME) Frequency Analysis"
    wB97X_D_def2_TZVP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_DME__Unknown = "wB97X-D/def2-TZVP/SMD(DME) Unknown"
    wB97X_D_def2_TZVP_SMD_EC__Single_Point = "wB97X-D/def2-TZVP/SMD(EC) Single Point"
    wB97X_D_def2_TZVP_SMD_EC__Force = "wB97X-D/def2-TZVP/SMD(EC) Force"
    wB97X_D_def2_TZVP_SMD_EC__Geometry_Optimization = "wB97X-D/def2-TZVP/SMD(EC) Geometry Optimization"
    wB97X_D_def2_TZVP_SMD_EC__Frequency_Analysis = "wB97X-D/def2-TZVP/SMD(EC) Frequency Analysis"
    wB97X_D_def2_TZVP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_EC__Unknown = "wB97X-D/def2-TZVP/SMD(EC) Unknown"
    wB97X_D_def2_TZVP_SMD_EC_EMC__Single_Point = "wB97X-D/def2-TZVP/SMD(EC/EMC) Single Point"
    wB97X_D_def2_TZVP_SMD_EC_EMC__Force = "wB97X-D/def2-TZVP/SMD(EC/EMC) Force"
    wB97X_D_def2_TZVP_SMD_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVP/SMD(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVP_SMD_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVP/SMD(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_EC_EMC__Unknown = "wB97X-D/def2-TZVP/SMD(EC/EMC) Unknown"
    wB97X_D_def2_TZVP_SMD_PC__Single_Point = "wB97X-D/def2-TZVP/SMD(PC) Single Point"
    wB97X_D_def2_TZVP_SMD_PC__Force = "wB97X-D/def2-TZVP/SMD(PC) Force"
    wB97X_D_def2_TZVP_SMD_PC__Geometry_Optimization = "wB97X-D/def2-TZVP/SMD(PC) Geometry Optimization"
    wB97X_D_def2_TZVP_SMD_PC__Frequency_Analysis = "wB97X-D/def2-TZVP/SMD(PC) Frequency Analysis"
    wB97X_D_def2_TZVP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_SMD_PC__Unknown = "wB97X-D/def2-TZVP/SMD(PC) Unknown"
    wB97X_D_def2_TZVP_CMIRS_WATER__Single_Point = "wB97X-D/def2-TZVP/CMIRS(WATER) Single Point"
    wB97X_D_def2_TZVP_CMIRS_WATER__Force = "wB97X-D/def2-TZVP/CMIRS(WATER) Force"
    wB97X_D_def2_TZVP_CMIRS_WATER__Geometry_Optimization = "wB97X-D/def2-TZVP/CMIRS(WATER) Geometry Optimization"
    wB97X_D_def2_TZVP_CMIRS_WATER__Frequency_Analysis = "wB97X-D/def2-TZVP/CMIRS(WATER) Frequency Analysis"
    wB97X_D_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_WATER__Unknown = "wB97X-D/def2-TZVP/CMIRS(WATER) Unknown"
    wB97X_D_def2_TZVP_CMIRS_THF__Single_Point = "wB97X-D/def2-TZVP/CMIRS(THF) Single Point"
    wB97X_D_def2_TZVP_CMIRS_THF__Force = "wB97X-D/def2-TZVP/CMIRS(THF) Force"
    wB97X_D_def2_TZVP_CMIRS_THF__Geometry_Optimization = "wB97X-D/def2-TZVP/CMIRS(THF) Geometry Optimization"
    wB97X_D_def2_TZVP_CMIRS_THF__Frequency_Analysis = "wB97X-D/def2-TZVP/CMIRS(THF) Frequency Analysis"
    wB97X_D_def2_TZVP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_THF__Unknown = "wB97X-D/def2-TZVP/CMIRS(THF) Unknown"
    wB97X_D_def2_TZVP_CMIRS_G2__Single_Point = "wB97X-D/def2-TZVP/CMIRS(G2) Single Point"
    wB97X_D_def2_TZVP_CMIRS_G2__Force = "wB97X-D/def2-TZVP/CMIRS(G2) Force"
    wB97X_D_def2_TZVP_CMIRS_G2__Geometry_Optimization = "wB97X-D/def2-TZVP/CMIRS(G2) Geometry Optimization"
    wB97X_D_def2_TZVP_CMIRS_G2__Frequency_Analysis = "wB97X-D/def2-TZVP/CMIRS(G2) Frequency Analysis"
    wB97X_D_def2_TZVP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_G2__Unknown = "wB97X-D/def2-TZVP/CMIRS(G2) Unknown"
    wB97X_D_def2_TZVP_CMIRS_DME__Single_Point = "wB97X-D/def2-TZVP/CMIRS(DME) Single Point"
    wB97X_D_def2_TZVP_CMIRS_DME__Force = "wB97X-D/def2-TZVP/CMIRS(DME) Force"
    wB97X_D_def2_TZVP_CMIRS_DME__Geometry_Optimization = "wB97X-D/def2-TZVP/CMIRS(DME) Geometry Optimization"
    wB97X_D_def2_TZVP_CMIRS_DME__Frequency_Analysis = "wB97X-D/def2-TZVP/CMIRS(DME) Frequency Analysis"
    wB97X_D_def2_TZVP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_DME__Unknown = "wB97X-D/def2-TZVP/CMIRS(DME) Unknown"
    wB97X_D_def2_TZVP_CMIRS_EC__Single_Point = "wB97X-D/def2-TZVP/CMIRS(EC) Single Point"
    wB97X_D_def2_TZVP_CMIRS_EC__Force = "wB97X-D/def2-TZVP/CMIRS(EC) Force"
    wB97X_D_def2_TZVP_CMIRS_EC__Geometry_Optimization = "wB97X-D/def2-TZVP/CMIRS(EC) Geometry Optimization"
    wB97X_D_def2_TZVP_CMIRS_EC__Frequency_Analysis = "wB97X-D/def2-TZVP/CMIRS(EC) Frequency Analysis"
    wB97X_D_def2_TZVP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_EC__Unknown = "wB97X-D/def2-TZVP/CMIRS(EC) Unknown"
    wB97X_D_def2_TZVP_CMIRS_EC_EMC__Single_Point = "wB97X-D/def2-TZVP/CMIRS(EC/EMC) Single Point"
    wB97X_D_def2_TZVP_CMIRS_EC_EMC__Force = "wB97X-D/def2-TZVP/CMIRS(EC/EMC) Force"
    wB97X_D_def2_TZVP_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVP/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVP_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVP/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_EC_EMC__Unknown = "wB97X-D/def2-TZVP/CMIRS(EC/EMC) Unknown"
    wB97X_D_def2_TZVP_CMIRS_PC__Single_Point = "wB97X-D/def2-TZVP/CMIRS(PC) Single Point"
    wB97X_D_def2_TZVP_CMIRS_PC__Force = "wB97X-D/def2-TZVP/CMIRS(PC) Force"
    wB97X_D_def2_TZVP_CMIRS_PC__Geometry_Optimization = "wB97X-D/def2-TZVP/CMIRS(PC) Geometry Optimization"
    wB97X_D_def2_TZVP_CMIRS_PC__Frequency_Analysis = "wB97X-D/def2-TZVP/CMIRS(PC) Frequency Analysis"
    wB97X_D_def2_TZVP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVP_CMIRS_PC__Unknown = "wB97X-D/def2-TZVP/CMIRS(PC) Unknown"
    wB97X_D_def2_TZVPD_VACUUM_Single_Point = "wB97X-D/def2-TZVPD/VACUUM Single Point"
    wB97X_D_def2_TZVPD_VACUUM_Force = "wB97X-D/def2-TZVPD/VACUUM Force"
    wB97X_D_def2_TZVPD_VACUUM_Geometry_Optimization = "wB97X-D/def2-TZVPD/VACUUM Geometry Optimization"
    wB97X_D_def2_TZVPD_VACUUM_Frequency_Analysis = "wB97X-D/def2-TZVPD/VACUUM Frequency Analysis"
    wB97X_D_def2_TZVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_VACUUM_Unknown = "wB97X-D/def2-TZVPD/VACUUM Unknown"
    wB97X_D_def2_TZVPD_PCM_WATER__Single_Point = "wB97X-D/def2-TZVPD/PCM(WATER) Single Point"
    wB97X_D_def2_TZVPD_PCM_WATER__Force = "wB97X-D/def2-TZVPD/PCM(WATER) Force"
    wB97X_D_def2_TZVPD_PCM_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPD/PCM(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPD_PCM_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPD/PCM(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_WATER__Unknown = "wB97X-D/def2-TZVPD/PCM(WATER) Unknown"
    wB97X_D_def2_TZVPD_PCM_THF__Single_Point = "wB97X-D/def2-TZVPD/PCM(THF) Single Point"
    wB97X_D_def2_TZVPD_PCM_THF__Force = "wB97X-D/def2-TZVPD/PCM(THF) Force"
    wB97X_D_def2_TZVPD_PCM_THF__Geometry_Optimization = "wB97X-D/def2-TZVPD/PCM(THF) Geometry Optimization"
    wB97X_D_def2_TZVPD_PCM_THF__Frequency_Analysis = "wB97X-D/def2-TZVPD/PCM(THF) Frequency Analysis"
    wB97X_D_def2_TZVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_THF__Unknown = "wB97X-D/def2-TZVPD/PCM(THF) Unknown"
    wB97X_D_def2_TZVPD_PCM_G2__Single_Point = "wB97X-D/def2-TZVPD/PCM(G2) Single Point"
    wB97X_D_def2_TZVPD_PCM_G2__Force = "wB97X-D/def2-TZVPD/PCM(G2) Force"
    wB97X_D_def2_TZVPD_PCM_G2__Geometry_Optimization = "wB97X-D/def2-TZVPD/PCM(G2) Geometry Optimization"
    wB97X_D_def2_TZVPD_PCM_G2__Frequency_Analysis = "wB97X-D/def2-TZVPD/PCM(G2) Frequency Analysis"
    wB97X_D_def2_TZVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_G2__Unknown = "wB97X-D/def2-TZVPD/PCM(G2) Unknown"
    wB97X_D_def2_TZVPD_PCM_DME__Single_Point = "wB97X-D/def2-TZVPD/PCM(DME) Single Point"
    wB97X_D_def2_TZVPD_PCM_DME__Force = "wB97X-D/def2-TZVPD/PCM(DME) Force"
    wB97X_D_def2_TZVPD_PCM_DME__Geometry_Optimization = "wB97X-D/def2-TZVPD/PCM(DME) Geometry Optimization"
    wB97X_D_def2_TZVPD_PCM_DME__Frequency_Analysis = "wB97X-D/def2-TZVPD/PCM(DME) Frequency Analysis"
    wB97X_D_def2_TZVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_DME__Unknown = "wB97X-D/def2-TZVPD/PCM(DME) Unknown"
    wB97X_D_def2_TZVPD_PCM_EC__Single_Point = "wB97X-D/def2-TZVPD/PCM(EC) Single Point"
    wB97X_D_def2_TZVPD_PCM_EC__Force = "wB97X-D/def2-TZVPD/PCM(EC) Force"
    wB97X_D_def2_TZVPD_PCM_EC__Geometry_Optimization = "wB97X-D/def2-TZVPD/PCM(EC) Geometry Optimization"
    wB97X_D_def2_TZVPD_PCM_EC__Frequency_Analysis = "wB97X-D/def2-TZVPD/PCM(EC) Frequency Analysis"
    wB97X_D_def2_TZVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_EC__Unknown = "wB97X-D/def2-TZVPD/PCM(EC) Unknown"
    wB97X_D_def2_TZVPD_PCM_EC_EMC__Single_Point = "wB97X-D/def2-TZVPD/PCM(EC/EMC) Single Point"
    wB97X_D_def2_TZVPD_PCM_EC_EMC__Force = "wB97X-D/def2-TZVPD/PCM(EC/EMC) Force"
    wB97X_D_def2_TZVPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_EC_EMC__Unknown = "wB97X-D/def2-TZVPD/PCM(EC/EMC) Unknown"
    wB97X_D_def2_TZVPD_PCM_PC__Single_Point = "wB97X-D/def2-TZVPD/PCM(PC) Single Point"
    wB97X_D_def2_TZVPD_PCM_PC__Force = "wB97X-D/def2-TZVPD/PCM(PC) Force"
    wB97X_D_def2_TZVPD_PCM_PC__Geometry_Optimization = "wB97X-D/def2-TZVPD/PCM(PC) Geometry Optimization"
    wB97X_D_def2_TZVPD_PCM_PC__Frequency_Analysis = "wB97X-D/def2-TZVPD/PCM(PC) Frequency Analysis"
    wB97X_D_def2_TZVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_PCM_PC__Unknown = "wB97X-D/def2-TZVPD/PCM(PC) Unknown"
    wB97X_D_def2_TZVPD_ISOSVP_WATER__Single_Point = "wB97X-D/def2-TZVPD/ISOSVP(WATER) Single Point"
    wB97X_D_def2_TZVPD_ISOSVP_WATER__Force = "wB97X-D/def2-TZVPD/ISOSVP(WATER) Force"
    wB97X_D_def2_TZVPD_ISOSVP_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPD/ISOSVP(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_WATER__Unknown = "wB97X-D/def2-TZVPD/ISOSVP(WATER) Unknown"
    wB97X_D_def2_TZVPD_ISOSVP_THF__Single_Point = "wB97X-D/def2-TZVPD/ISOSVP(THF) Single Point"
    wB97X_D_def2_TZVPD_ISOSVP_THF__Force = "wB97X-D/def2-TZVPD/ISOSVP(THF) Force"
    wB97X_D_def2_TZVPD_ISOSVP_THF__Geometry_Optimization = "wB97X-D/def2-TZVPD/ISOSVP(THF) Geometry Optimization"
    wB97X_D_def2_TZVPD_ISOSVP_THF__Frequency_Analysis = "wB97X-D/def2-TZVPD/ISOSVP(THF) Frequency Analysis"
    wB97X_D_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_THF__Unknown = "wB97X-D/def2-TZVPD/ISOSVP(THF) Unknown"
    wB97X_D_def2_TZVPD_ISOSVP_G2__Single_Point = "wB97X-D/def2-TZVPD/ISOSVP(G2) Single Point"
    wB97X_D_def2_TZVPD_ISOSVP_G2__Force = "wB97X-D/def2-TZVPD/ISOSVP(G2) Force"
    wB97X_D_def2_TZVPD_ISOSVP_G2__Geometry_Optimization = "wB97X-D/def2-TZVPD/ISOSVP(G2) Geometry Optimization"
    wB97X_D_def2_TZVPD_ISOSVP_G2__Frequency_Analysis = "wB97X-D/def2-TZVPD/ISOSVP(G2) Frequency Analysis"
    wB97X_D_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_G2__Unknown = "wB97X-D/def2-TZVPD/ISOSVP(G2) Unknown"
    wB97X_D_def2_TZVPD_ISOSVP_DME__Single_Point = "wB97X-D/def2-TZVPD/ISOSVP(DME) Single Point"
    wB97X_D_def2_TZVPD_ISOSVP_DME__Force = "wB97X-D/def2-TZVPD/ISOSVP(DME) Force"
    wB97X_D_def2_TZVPD_ISOSVP_DME__Geometry_Optimization = "wB97X-D/def2-TZVPD/ISOSVP(DME) Geometry Optimization"
    wB97X_D_def2_TZVPD_ISOSVP_DME__Frequency_Analysis = "wB97X-D/def2-TZVPD/ISOSVP(DME) Frequency Analysis"
    wB97X_D_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_DME__Unknown = "wB97X-D/def2-TZVPD/ISOSVP(DME) Unknown"
    wB97X_D_def2_TZVPD_ISOSVP_EC__Single_Point = "wB97X-D/def2-TZVPD/ISOSVP(EC) Single Point"
    wB97X_D_def2_TZVPD_ISOSVP_EC__Force = "wB97X-D/def2-TZVPD/ISOSVP(EC) Force"
    wB97X_D_def2_TZVPD_ISOSVP_EC__Geometry_Optimization = "wB97X-D/def2-TZVPD/ISOSVP(EC) Geometry Optimization"
    wB97X_D_def2_TZVPD_ISOSVP_EC__Frequency_Analysis = "wB97X-D/def2-TZVPD/ISOSVP(EC) Frequency Analysis"
    wB97X_D_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_EC__Unknown = "wB97X-D/def2-TZVPD/ISOSVP(EC) Unknown"
    wB97X_D_def2_TZVPD_ISOSVP_EC_EMC__Single_Point = "wB97X-D/def2-TZVPD/ISOSVP(EC/EMC) Single Point"
    wB97X_D_def2_TZVPD_ISOSVP_EC_EMC__Force = "wB97X-D/def2-TZVPD/ISOSVP(EC/EMC) Force"
    wB97X_D_def2_TZVPD_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVPD/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_EC_EMC__Unknown = "wB97X-D/def2-TZVPD/ISOSVP(EC/EMC) Unknown"
    wB97X_D_def2_TZVPD_ISOSVP_PC__Single_Point = "wB97X-D/def2-TZVPD/ISOSVP(PC) Single Point"
    wB97X_D_def2_TZVPD_ISOSVP_PC__Force = "wB97X-D/def2-TZVPD/ISOSVP(PC) Force"
    wB97X_D_def2_TZVPD_ISOSVP_PC__Geometry_Optimization = "wB97X-D/def2-TZVPD/ISOSVP(PC) Geometry Optimization"
    wB97X_D_def2_TZVPD_ISOSVP_PC__Frequency_Analysis = "wB97X-D/def2-TZVPD/ISOSVP(PC) Frequency Analysis"
    wB97X_D_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_ISOSVP_PC__Unknown = "wB97X-D/def2-TZVPD/ISOSVP(PC) Unknown"
    wB97X_D_def2_TZVPD_SMD_WATER__Single_Point = "wB97X-D/def2-TZVPD/SMD(WATER) Single Point"
    wB97X_D_def2_TZVPD_SMD_WATER__Force = "wB97X-D/def2-TZVPD/SMD(WATER) Force"
    wB97X_D_def2_TZVPD_SMD_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPD/SMD(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPD_SMD_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPD/SMD(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_WATER__Unknown = "wB97X-D/def2-TZVPD/SMD(WATER) Unknown"
    wB97X_D_def2_TZVPD_SMD_THF__Single_Point = "wB97X-D/def2-TZVPD/SMD(THF) Single Point"
    wB97X_D_def2_TZVPD_SMD_THF__Force = "wB97X-D/def2-TZVPD/SMD(THF) Force"
    wB97X_D_def2_TZVPD_SMD_THF__Geometry_Optimization = "wB97X-D/def2-TZVPD/SMD(THF) Geometry Optimization"
    wB97X_D_def2_TZVPD_SMD_THF__Frequency_Analysis = "wB97X-D/def2-TZVPD/SMD(THF) Frequency Analysis"
    wB97X_D_def2_TZVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_THF__Unknown = "wB97X-D/def2-TZVPD/SMD(THF) Unknown"
    wB97X_D_def2_TZVPD_SMD_G2__Single_Point = "wB97X-D/def2-TZVPD/SMD(G2) Single Point"
    wB97X_D_def2_TZVPD_SMD_G2__Force = "wB97X-D/def2-TZVPD/SMD(G2) Force"
    wB97X_D_def2_TZVPD_SMD_G2__Geometry_Optimization = "wB97X-D/def2-TZVPD/SMD(G2) Geometry Optimization"
    wB97X_D_def2_TZVPD_SMD_G2__Frequency_Analysis = "wB97X-D/def2-TZVPD/SMD(G2) Frequency Analysis"
    wB97X_D_def2_TZVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_G2__Unknown = "wB97X-D/def2-TZVPD/SMD(G2) Unknown"
    wB97X_D_def2_TZVPD_SMD_DME__Single_Point = "wB97X-D/def2-TZVPD/SMD(DME) Single Point"
    wB97X_D_def2_TZVPD_SMD_DME__Force = "wB97X-D/def2-TZVPD/SMD(DME) Force"
    wB97X_D_def2_TZVPD_SMD_DME__Geometry_Optimization = "wB97X-D/def2-TZVPD/SMD(DME) Geometry Optimization"
    wB97X_D_def2_TZVPD_SMD_DME__Frequency_Analysis = "wB97X-D/def2-TZVPD/SMD(DME) Frequency Analysis"
    wB97X_D_def2_TZVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_DME__Unknown = "wB97X-D/def2-TZVPD/SMD(DME) Unknown"
    wB97X_D_def2_TZVPD_SMD_EC__Single_Point = "wB97X-D/def2-TZVPD/SMD(EC) Single Point"
    wB97X_D_def2_TZVPD_SMD_EC__Force = "wB97X-D/def2-TZVPD/SMD(EC) Force"
    wB97X_D_def2_TZVPD_SMD_EC__Geometry_Optimization = "wB97X-D/def2-TZVPD/SMD(EC) Geometry Optimization"
    wB97X_D_def2_TZVPD_SMD_EC__Frequency_Analysis = "wB97X-D/def2-TZVPD/SMD(EC) Frequency Analysis"
    wB97X_D_def2_TZVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_EC__Unknown = "wB97X-D/def2-TZVPD/SMD(EC) Unknown"
    wB97X_D_def2_TZVPD_SMD_EC_EMC__Single_Point = "wB97X-D/def2-TZVPD/SMD(EC/EMC) Single Point"
    wB97X_D_def2_TZVPD_SMD_EC_EMC__Force = "wB97X-D/def2-TZVPD/SMD(EC/EMC) Force"
    wB97X_D_def2_TZVPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_EC_EMC__Unknown = "wB97X-D/def2-TZVPD/SMD(EC/EMC) Unknown"
    wB97X_D_def2_TZVPD_SMD_PC__Single_Point = "wB97X-D/def2-TZVPD/SMD(PC) Single Point"
    wB97X_D_def2_TZVPD_SMD_PC__Force = "wB97X-D/def2-TZVPD/SMD(PC) Force"
    wB97X_D_def2_TZVPD_SMD_PC__Geometry_Optimization = "wB97X-D/def2-TZVPD/SMD(PC) Geometry Optimization"
    wB97X_D_def2_TZVPD_SMD_PC__Frequency_Analysis = "wB97X-D/def2-TZVPD/SMD(PC) Frequency Analysis"
    wB97X_D_def2_TZVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_SMD_PC__Unknown = "wB97X-D/def2-TZVPD/SMD(PC) Unknown"
    wB97X_D_def2_TZVPD_CMIRS_WATER__Single_Point = "wB97X-D/def2-TZVPD/CMIRS(WATER) Single Point"
    wB97X_D_def2_TZVPD_CMIRS_WATER__Force = "wB97X-D/def2-TZVPD/CMIRS(WATER) Force"
    wB97X_D_def2_TZVPD_CMIRS_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPD/CMIRS(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPD_CMIRS_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPD/CMIRS(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_WATER__Unknown = "wB97X-D/def2-TZVPD/CMIRS(WATER) Unknown"
    wB97X_D_def2_TZVPD_CMIRS_THF__Single_Point = "wB97X-D/def2-TZVPD/CMIRS(THF) Single Point"
    wB97X_D_def2_TZVPD_CMIRS_THF__Force = "wB97X-D/def2-TZVPD/CMIRS(THF) Force"
    wB97X_D_def2_TZVPD_CMIRS_THF__Geometry_Optimization = "wB97X-D/def2-TZVPD/CMIRS(THF) Geometry Optimization"
    wB97X_D_def2_TZVPD_CMIRS_THF__Frequency_Analysis = "wB97X-D/def2-TZVPD/CMIRS(THF) Frequency Analysis"
    wB97X_D_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_THF__Unknown = "wB97X-D/def2-TZVPD/CMIRS(THF) Unknown"
    wB97X_D_def2_TZVPD_CMIRS_G2__Single_Point = "wB97X-D/def2-TZVPD/CMIRS(G2) Single Point"
    wB97X_D_def2_TZVPD_CMIRS_G2__Force = "wB97X-D/def2-TZVPD/CMIRS(G2) Force"
    wB97X_D_def2_TZVPD_CMIRS_G2__Geometry_Optimization = "wB97X-D/def2-TZVPD/CMIRS(G2) Geometry Optimization"
    wB97X_D_def2_TZVPD_CMIRS_G2__Frequency_Analysis = "wB97X-D/def2-TZVPD/CMIRS(G2) Frequency Analysis"
    wB97X_D_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_G2__Unknown = "wB97X-D/def2-TZVPD/CMIRS(G2) Unknown"
    wB97X_D_def2_TZVPD_CMIRS_DME__Single_Point = "wB97X-D/def2-TZVPD/CMIRS(DME) Single Point"
    wB97X_D_def2_TZVPD_CMIRS_DME__Force = "wB97X-D/def2-TZVPD/CMIRS(DME) Force"
    wB97X_D_def2_TZVPD_CMIRS_DME__Geometry_Optimization = "wB97X-D/def2-TZVPD/CMIRS(DME) Geometry Optimization"
    wB97X_D_def2_TZVPD_CMIRS_DME__Frequency_Analysis = "wB97X-D/def2-TZVPD/CMIRS(DME) Frequency Analysis"
    wB97X_D_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_DME__Unknown = "wB97X-D/def2-TZVPD/CMIRS(DME) Unknown"
    wB97X_D_def2_TZVPD_CMIRS_EC__Single_Point = "wB97X-D/def2-TZVPD/CMIRS(EC) Single Point"
    wB97X_D_def2_TZVPD_CMIRS_EC__Force = "wB97X-D/def2-TZVPD/CMIRS(EC) Force"
    wB97X_D_def2_TZVPD_CMIRS_EC__Geometry_Optimization = "wB97X-D/def2-TZVPD/CMIRS(EC) Geometry Optimization"
    wB97X_D_def2_TZVPD_CMIRS_EC__Frequency_Analysis = "wB97X-D/def2-TZVPD/CMIRS(EC) Frequency Analysis"
    wB97X_D_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_EC__Unknown = "wB97X-D/def2-TZVPD/CMIRS(EC) Unknown"
    wB97X_D_def2_TZVPD_CMIRS_EC_EMC__Single_Point = "wB97X-D/def2-TZVPD/CMIRS(EC/EMC) Single Point"
    wB97X_D_def2_TZVPD_CMIRS_EC_EMC__Force = "wB97X-D/def2-TZVPD/CMIRS(EC/EMC) Force"
    wB97X_D_def2_TZVPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_EC_EMC__Unknown = "wB97X-D/def2-TZVPD/CMIRS(EC/EMC) Unknown"
    wB97X_D_def2_TZVPD_CMIRS_PC__Single_Point = "wB97X-D/def2-TZVPD/CMIRS(PC) Single Point"
    wB97X_D_def2_TZVPD_CMIRS_PC__Force = "wB97X-D/def2-TZVPD/CMIRS(PC) Force"
    wB97X_D_def2_TZVPD_CMIRS_PC__Geometry_Optimization = "wB97X-D/def2-TZVPD/CMIRS(PC) Geometry Optimization"
    wB97X_D_def2_TZVPD_CMIRS_PC__Frequency_Analysis = "wB97X-D/def2-TZVPD/CMIRS(PC) Frequency Analysis"
    wB97X_D_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPD_CMIRS_PC__Unknown = "wB97X-D/def2-TZVPD/CMIRS(PC) Unknown"
    wB97X_D_def2_TZVPP_VACUUM_Single_Point = "wB97X-D/def2-TZVPP/VACUUM Single Point"
    wB97X_D_def2_TZVPP_VACUUM_Force = "wB97X-D/def2-TZVPP/VACUUM Force"
    wB97X_D_def2_TZVPP_VACUUM_Geometry_Optimization = "wB97X-D/def2-TZVPP/VACUUM Geometry Optimization"
    wB97X_D_def2_TZVPP_VACUUM_Frequency_Analysis = "wB97X-D/def2-TZVPP/VACUUM Frequency Analysis"
    wB97X_D_def2_TZVPP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_VACUUM_Unknown = "wB97X-D/def2-TZVPP/VACUUM Unknown"
    wB97X_D_def2_TZVPP_PCM_WATER__Single_Point = "wB97X-D/def2-TZVPP/PCM(WATER) Single Point"
    wB97X_D_def2_TZVPP_PCM_WATER__Force = "wB97X-D/def2-TZVPP/PCM(WATER) Force"
    wB97X_D_def2_TZVPP_PCM_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPP/PCM(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPP_PCM_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPP/PCM(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_WATER__Unknown = "wB97X-D/def2-TZVPP/PCM(WATER) Unknown"
    wB97X_D_def2_TZVPP_PCM_THF__Single_Point = "wB97X-D/def2-TZVPP/PCM(THF) Single Point"
    wB97X_D_def2_TZVPP_PCM_THF__Force = "wB97X-D/def2-TZVPP/PCM(THF) Force"
    wB97X_D_def2_TZVPP_PCM_THF__Geometry_Optimization = "wB97X-D/def2-TZVPP/PCM(THF) Geometry Optimization"
    wB97X_D_def2_TZVPP_PCM_THF__Frequency_Analysis = "wB97X-D/def2-TZVPP/PCM(THF) Frequency Analysis"
    wB97X_D_def2_TZVPP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_THF__Unknown = "wB97X-D/def2-TZVPP/PCM(THF) Unknown"
    wB97X_D_def2_TZVPP_PCM_G2__Single_Point = "wB97X-D/def2-TZVPP/PCM(G2) Single Point"
    wB97X_D_def2_TZVPP_PCM_G2__Force = "wB97X-D/def2-TZVPP/PCM(G2) Force"
    wB97X_D_def2_TZVPP_PCM_G2__Geometry_Optimization = "wB97X-D/def2-TZVPP/PCM(G2) Geometry Optimization"
    wB97X_D_def2_TZVPP_PCM_G2__Frequency_Analysis = "wB97X-D/def2-TZVPP/PCM(G2) Frequency Analysis"
    wB97X_D_def2_TZVPP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_G2__Unknown = "wB97X-D/def2-TZVPP/PCM(G2) Unknown"
    wB97X_D_def2_TZVPP_PCM_DME__Single_Point = "wB97X-D/def2-TZVPP/PCM(DME) Single Point"
    wB97X_D_def2_TZVPP_PCM_DME__Force = "wB97X-D/def2-TZVPP/PCM(DME) Force"
    wB97X_D_def2_TZVPP_PCM_DME__Geometry_Optimization = "wB97X-D/def2-TZVPP/PCM(DME) Geometry Optimization"
    wB97X_D_def2_TZVPP_PCM_DME__Frequency_Analysis = "wB97X-D/def2-TZVPP/PCM(DME) Frequency Analysis"
    wB97X_D_def2_TZVPP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_DME__Unknown = "wB97X-D/def2-TZVPP/PCM(DME) Unknown"
    wB97X_D_def2_TZVPP_PCM_EC__Single_Point = "wB97X-D/def2-TZVPP/PCM(EC) Single Point"
    wB97X_D_def2_TZVPP_PCM_EC__Force = "wB97X-D/def2-TZVPP/PCM(EC) Force"
    wB97X_D_def2_TZVPP_PCM_EC__Geometry_Optimization = "wB97X-D/def2-TZVPP/PCM(EC) Geometry Optimization"
    wB97X_D_def2_TZVPP_PCM_EC__Frequency_Analysis = "wB97X-D/def2-TZVPP/PCM(EC) Frequency Analysis"
    wB97X_D_def2_TZVPP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_EC__Unknown = "wB97X-D/def2-TZVPP/PCM(EC) Unknown"
    wB97X_D_def2_TZVPP_PCM_EC_EMC__Single_Point = "wB97X-D/def2-TZVPP/PCM(EC/EMC) Single Point"
    wB97X_D_def2_TZVPP_PCM_EC_EMC__Force = "wB97X-D/def2-TZVPP/PCM(EC/EMC) Force"
    wB97X_D_def2_TZVPP_PCM_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVPP/PCM(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVPP_PCM_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPP/PCM(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_EC_EMC__Unknown = "wB97X-D/def2-TZVPP/PCM(EC/EMC) Unknown"
    wB97X_D_def2_TZVPP_PCM_PC__Single_Point = "wB97X-D/def2-TZVPP/PCM(PC) Single Point"
    wB97X_D_def2_TZVPP_PCM_PC__Force = "wB97X-D/def2-TZVPP/PCM(PC) Force"
    wB97X_D_def2_TZVPP_PCM_PC__Geometry_Optimization = "wB97X-D/def2-TZVPP/PCM(PC) Geometry Optimization"
    wB97X_D_def2_TZVPP_PCM_PC__Frequency_Analysis = "wB97X-D/def2-TZVPP/PCM(PC) Frequency Analysis"
    wB97X_D_def2_TZVPP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_PCM_PC__Unknown = "wB97X-D/def2-TZVPP/PCM(PC) Unknown"
    wB97X_D_def2_TZVPP_ISOSVP_WATER__Single_Point = "wB97X-D/def2-TZVPP/ISOSVP(WATER) Single Point"
    wB97X_D_def2_TZVPP_ISOSVP_WATER__Force = "wB97X-D/def2-TZVPP/ISOSVP(WATER) Force"
    wB97X_D_def2_TZVPP_ISOSVP_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPP/ISOSVP(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPP_ISOSVP_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPP/ISOSVP(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_WATER__Unknown = "wB97X-D/def2-TZVPP/ISOSVP(WATER) Unknown"
    wB97X_D_def2_TZVPP_ISOSVP_THF__Single_Point = "wB97X-D/def2-TZVPP/ISOSVP(THF) Single Point"
    wB97X_D_def2_TZVPP_ISOSVP_THF__Force = "wB97X-D/def2-TZVPP/ISOSVP(THF) Force"
    wB97X_D_def2_TZVPP_ISOSVP_THF__Geometry_Optimization = "wB97X-D/def2-TZVPP/ISOSVP(THF) Geometry Optimization"
    wB97X_D_def2_TZVPP_ISOSVP_THF__Frequency_Analysis = "wB97X-D/def2-TZVPP/ISOSVP(THF) Frequency Analysis"
    wB97X_D_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_THF__Unknown = "wB97X-D/def2-TZVPP/ISOSVP(THF) Unknown"
    wB97X_D_def2_TZVPP_ISOSVP_G2__Single_Point = "wB97X-D/def2-TZVPP/ISOSVP(G2) Single Point"
    wB97X_D_def2_TZVPP_ISOSVP_G2__Force = "wB97X-D/def2-TZVPP/ISOSVP(G2) Force"
    wB97X_D_def2_TZVPP_ISOSVP_G2__Geometry_Optimization = "wB97X-D/def2-TZVPP/ISOSVP(G2) Geometry Optimization"
    wB97X_D_def2_TZVPP_ISOSVP_G2__Frequency_Analysis = "wB97X-D/def2-TZVPP/ISOSVP(G2) Frequency Analysis"
    wB97X_D_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_G2__Unknown = "wB97X-D/def2-TZVPP/ISOSVP(G2) Unknown"
    wB97X_D_def2_TZVPP_ISOSVP_DME__Single_Point = "wB97X-D/def2-TZVPP/ISOSVP(DME) Single Point"
    wB97X_D_def2_TZVPP_ISOSVP_DME__Force = "wB97X-D/def2-TZVPP/ISOSVP(DME) Force"
    wB97X_D_def2_TZVPP_ISOSVP_DME__Geometry_Optimization = "wB97X-D/def2-TZVPP/ISOSVP(DME) Geometry Optimization"
    wB97X_D_def2_TZVPP_ISOSVP_DME__Frequency_Analysis = "wB97X-D/def2-TZVPP/ISOSVP(DME) Frequency Analysis"
    wB97X_D_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_DME__Unknown = "wB97X-D/def2-TZVPP/ISOSVP(DME) Unknown"
    wB97X_D_def2_TZVPP_ISOSVP_EC__Single_Point = "wB97X-D/def2-TZVPP/ISOSVP(EC) Single Point"
    wB97X_D_def2_TZVPP_ISOSVP_EC__Force = "wB97X-D/def2-TZVPP/ISOSVP(EC) Force"
    wB97X_D_def2_TZVPP_ISOSVP_EC__Geometry_Optimization = "wB97X-D/def2-TZVPP/ISOSVP(EC) Geometry Optimization"
    wB97X_D_def2_TZVPP_ISOSVP_EC__Frequency_Analysis = "wB97X-D/def2-TZVPP/ISOSVP(EC) Frequency Analysis"
    wB97X_D_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_EC__Unknown = "wB97X-D/def2-TZVPP/ISOSVP(EC) Unknown"
    wB97X_D_def2_TZVPP_ISOSVP_EC_EMC__Single_Point = "wB97X-D/def2-TZVPP/ISOSVP(EC/EMC) Single Point"
    wB97X_D_def2_TZVPP_ISOSVP_EC_EMC__Force = "wB97X-D/def2-TZVPP/ISOSVP(EC/EMC) Force"
    wB97X_D_def2_TZVPP_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVPP/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPP/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_EC_EMC__Unknown = "wB97X-D/def2-TZVPP/ISOSVP(EC/EMC) Unknown"
    wB97X_D_def2_TZVPP_ISOSVP_PC__Single_Point = "wB97X-D/def2-TZVPP/ISOSVP(PC) Single Point"
    wB97X_D_def2_TZVPP_ISOSVP_PC__Force = "wB97X-D/def2-TZVPP/ISOSVP(PC) Force"
    wB97X_D_def2_TZVPP_ISOSVP_PC__Geometry_Optimization = "wB97X-D/def2-TZVPP/ISOSVP(PC) Geometry Optimization"
    wB97X_D_def2_TZVPP_ISOSVP_PC__Frequency_Analysis = "wB97X-D/def2-TZVPP/ISOSVP(PC) Frequency Analysis"
    wB97X_D_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_ISOSVP_PC__Unknown = "wB97X-D/def2-TZVPP/ISOSVP(PC) Unknown"
    wB97X_D_def2_TZVPP_SMD_WATER__Single_Point = "wB97X-D/def2-TZVPP/SMD(WATER) Single Point"
    wB97X_D_def2_TZVPP_SMD_WATER__Force = "wB97X-D/def2-TZVPP/SMD(WATER) Force"
    wB97X_D_def2_TZVPP_SMD_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPP/SMD(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPP_SMD_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPP/SMD(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_WATER__Unknown = "wB97X-D/def2-TZVPP/SMD(WATER) Unknown"
    wB97X_D_def2_TZVPP_SMD_THF__Single_Point = "wB97X-D/def2-TZVPP/SMD(THF) Single Point"
    wB97X_D_def2_TZVPP_SMD_THF__Force = "wB97X-D/def2-TZVPP/SMD(THF) Force"
    wB97X_D_def2_TZVPP_SMD_THF__Geometry_Optimization = "wB97X-D/def2-TZVPP/SMD(THF) Geometry Optimization"
    wB97X_D_def2_TZVPP_SMD_THF__Frequency_Analysis = "wB97X-D/def2-TZVPP/SMD(THF) Frequency Analysis"
    wB97X_D_def2_TZVPP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_THF__Unknown = "wB97X-D/def2-TZVPP/SMD(THF) Unknown"
    wB97X_D_def2_TZVPP_SMD_G2__Single_Point = "wB97X-D/def2-TZVPP/SMD(G2) Single Point"
    wB97X_D_def2_TZVPP_SMD_G2__Force = "wB97X-D/def2-TZVPP/SMD(G2) Force"
    wB97X_D_def2_TZVPP_SMD_G2__Geometry_Optimization = "wB97X-D/def2-TZVPP/SMD(G2) Geometry Optimization"
    wB97X_D_def2_TZVPP_SMD_G2__Frequency_Analysis = "wB97X-D/def2-TZVPP/SMD(G2) Frequency Analysis"
    wB97X_D_def2_TZVPP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_G2__Unknown = "wB97X-D/def2-TZVPP/SMD(G2) Unknown"
    wB97X_D_def2_TZVPP_SMD_DME__Single_Point = "wB97X-D/def2-TZVPP/SMD(DME) Single Point"
    wB97X_D_def2_TZVPP_SMD_DME__Force = "wB97X-D/def2-TZVPP/SMD(DME) Force"
    wB97X_D_def2_TZVPP_SMD_DME__Geometry_Optimization = "wB97X-D/def2-TZVPP/SMD(DME) Geometry Optimization"
    wB97X_D_def2_TZVPP_SMD_DME__Frequency_Analysis = "wB97X-D/def2-TZVPP/SMD(DME) Frequency Analysis"
    wB97X_D_def2_TZVPP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_DME__Unknown = "wB97X-D/def2-TZVPP/SMD(DME) Unknown"
    wB97X_D_def2_TZVPP_SMD_EC__Single_Point = "wB97X-D/def2-TZVPP/SMD(EC) Single Point"
    wB97X_D_def2_TZVPP_SMD_EC__Force = "wB97X-D/def2-TZVPP/SMD(EC) Force"
    wB97X_D_def2_TZVPP_SMD_EC__Geometry_Optimization = "wB97X-D/def2-TZVPP/SMD(EC) Geometry Optimization"
    wB97X_D_def2_TZVPP_SMD_EC__Frequency_Analysis = "wB97X-D/def2-TZVPP/SMD(EC) Frequency Analysis"
    wB97X_D_def2_TZVPP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_EC__Unknown = "wB97X-D/def2-TZVPP/SMD(EC) Unknown"
    wB97X_D_def2_TZVPP_SMD_EC_EMC__Single_Point = "wB97X-D/def2-TZVPP/SMD(EC/EMC) Single Point"
    wB97X_D_def2_TZVPP_SMD_EC_EMC__Force = "wB97X-D/def2-TZVPP/SMD(EC/EMC) Force"
    wB97X_D_def2_TZVPP_SMD_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVPP/SMD(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVPP_SMD_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPP/SMD(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_EC_EMC__Unknown = "wB97X-D/def2-TZVPP/SMD(EC/EMC) Unknown"
    wB97X_D_def2_TZVPP_SMD_PC__Single_Point = "wB97X-D/def2-TZVPP/SMD(PC) Single Point"
    wB97X_D_def2_TZVPP_SMD_PC__Force = "wB97X-D/def2-TZVPP/SMD(PC) Force"
    wB97X_D_def2_TZVPP_SMD_PC__Geometry_Optimization = "wB97X-D/def2-TZVPP/SMD(PC) Geometry Optimization"
    wB97X_D_def2_TZVPP_SMD_PC__Frequency_Analysis = "wB97X-D/def2-TZVPP/SMD(PC) Frequency Analysis"
    wB97X_D_def2_TZVPP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_SMD_PC__Unknown = "wB97X-D/def2-TZVPP/SMD(PC) Unknown"
    wB97X_D_def2_TZVPP_CMIRS_WATER__Single_Point = "wB97X-D/def2-TZVPP/CMIRS(WATER) Single Point"
    wB97X_D_def2_TZVPP_CMIRS_WATER__Force = "wB97X-D/def2-TZVPP/CMIRS(WATER) Force"
    wB97X_D_def2_TZVPP_CMIRS_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPP/CMIRS(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPP_CMIRS_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPP/CMIRS(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_WATER__Unknown = "wB97X-D/def2-TZVPP/CMIRS(WATER) Unknown"
    wB97X_D_def2_TZVPP_CMIRS_THF__Single_Point = "wB97X-D/def2-TZVPP/CMIRS(THF) Single Point"
    wB97X_D_def2_TZVPP_CMIRS_THF__Force = "wB97X-D/def2-TZVPP/CMIRS(THF) Force"
    wB97X_D_def2_TZVPP_CMIRS_THF__Geometry_Optimization = "wB97X-D/def2-TZVPP/CMIRS(THF) Geometry Optimization"
    wB97X_D_def2_TZVPP_CMIRS_THF__Frequency_Analysis = "wB97X-D/def2-TZVPP/CMIRS(THF) Frequency Analysis"
    wB97X_D_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_THF__Unknown = "wB97X-D/def2-TZVPP/CMIRS(THF) Unknown"
    wB97X_D_def2_TZVPP_CMIRS_G2__Single_Point = "wB97X-D/def2-TZVPP/CMIRS(G2) Single Point"
    wB97X_D_def2_TZVPP_CMIRS_G2__Force = "wB97X-D/def2-TZVPP/CMIRS(G2) Force"
    wB97X_D_def2_TZVPP_CMIRS_G2__Geometry_Optimization = "wB97X-D/def2-TZVPP/CMIRS(G2) Geometry Optimization"
    wB97X_D_def2_TZVPP_CMIRS_G2__Frequency_Analysis = "wB97X-D/def2-TZVPP/CMIRS(G2) Frequency Analysis"
    wB97X_D_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_G2__Unknown = "wB97X-D/def2-TZVPP/CMIRS(G2) Unknown"
    wB97X_D_def2_TZVPP_CMIRS_DME__Single_Point = "wB97X-D/def2-TZVPP/CMIRS(DME) Single Point"
    wB97X_D_def2_TZVPP_CMIRS_DME__Force = "wB97X-D/def2-TZVPP/CMIRS(DME) Force"
    wB97X_D_def2_TZVPP_CMIRS_DME__Geometry_Optimization = "wB97X-D/def2-TZVPP/CMIRS(DME) Geometry Optimization"
    wB97X_D_def2_TZVPP_CMIRS_DME__Frequency_Analysis = "wB97X-D/def2-TZVPP/CMIRS(DME) Frequency Analysis"
    wB97X_D_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_DME__Unknown = "wB97X-D/def2-TZVPP/CMIRS(DME) Unknown"
    wB97X_D_def2_TZVPP_CMIRS_EC__Single_Point = "wB97X-D/def2-TZVPP/CMIRS(EC) Single Point"
    wB97X_D_def2_TZVPP_CMIRS_EC__Force = "wB97X-D/def2-TZVPP/CMIRS(EC) Force"
    wB97X_D_def2_TZVPP_CMIRS_EC__Geometry_Optimization = "wB97X-D/def2-TZVPP/CMIRS(EC) Geometry Optimization"
    wB97X_D_def2_TZVPP_CMIRS_EC__Frequency_Analysis = "wB97X-D/def2-TZVPP/CMIRS(EC) Frequency Analysis"
    wB97X_D_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_EC__Unknown = "wB97X-D/def2-TZVPP/CMIRS(EC) Unknown"
    wB97X_D_def2_TZVPP_CMIRS_EC_EMC__Single_Point = "wB97X-D/def2-TZVPP/CMIRS(EC/EMC) Single Point"
    wB97X_D_def2_TZVPP_CMIRS_EC_EMC__Force = "wB97X-D/def2-TZVPP/CMIRS(EC/EMC) Force"
    wB97X_D_def2_TZVPP_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVPP/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVPP_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPP/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_EC_EMC__Unknown = "wB97X-D/def2-TZVPP/CMIRS(EC/EMC) Unknown"
    wB97X_D_def2_TZVPP_CMIRS_PC__Single_Point = "wB97X-D/def2-TZVPP/CMIRS(PC) Single Point"
    wB97X_D_def2_TZVPP_CMIRS_PC__Force = "wB97X-D/def2-TZVPP/CMIRS(PC) Force"
    wB97X_D_def2_TZVPP_CMIRS_PC__Geometry_Optimization = "wB97X-D/def2-TZVPP/CMIRS(PC) Geometry Optimization"
    wB97X_D_def2_TZVPP_CMIRS_PC__Frequency_Analysis = "wB97X-D/def2-TZVPP/CMIRS(PC) Frequency Analysis"
    wB97X_D_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPP_CMIRS_PC__Unknown = "wB97X-D/def2-TZVPP/CMIRS(PC) Unknown"
    wB97X_D_def2_TZVPPD_VACUUM_Single_Point = "wB97X-D/def2-TZVPPD/VACUUM Single Point"
    wB97X_D_def2_TZVPPD_VACUUM_Force = "wB97X-D/def2-TZVPPD/VACUUM Force"
    wB97X_D_def2_TZVPPD_VACUUM_Geometry_Optimization = "wB97X-D/def2-TZVPPD/VACUUM Geometry Optimization"
    wB97X_D_def2_TZVPPD_VACUUM_Frequency_Analysis = "wB97X-D/def2-TZVPPD/VACUUM Frequency Analysis"
    wB97X_D_def2_TZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_VACUUM_Unknown = "wB97X-D/def2-TZVPPD/VACUUM Unknown"
    wB97X_D_def2_TZVPPD_PCM_WATER__Single_Point = "wB97X-D/def2-TZVPPD/PCM(WATER) Single Point"
    wB97X_D_def2_TZVPPD_PCM_WATER__Force = "wB97X-D/def2-TZVPPD/PCM(WATER) Force"
    wB97X_D_def2_TZVPPD_PCM_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPPD/PCM(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPPD_PCM_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPPD/PCM(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_WATER__Unknown = "wB97X-D/def2-TZVPPD/PCM(WATER) Unknown"
    wB97X_D_def2_TZVPPD_PCM_THF__Single_Point = "wB97X-D/def2-TZVPPD/PCM(THF) Single Point"
    wB97X_D_def2_TZVPPD_PCM_THF__Force = "wB97X-D/def2-TZVPPD/PCM(THF) Force"
    wB97X_D_def2_TZVPPD_PCM_THF__Geometry_Optimization = "wB97X-D/def2-TZVPPD/PCM(THF) Geometry Optimization"
    wB97X_D_def2_TZVPPD_PCM_THF__Frequency_Analysis = "wB97X-D/def2-TZVPPD/PCM(THF) Frequency Analysis"
    wB97X_D_def2_TZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_THF__Unknown = "wB97X-D/def2-TZVPPD/PCM(THF) Unknown"
    wB97X_D_def2_TZVPPD_PCM_G2__Single_Point = "wB97X-D/def2-TZVPPD/PCM(G2) Single Point"
    wB97X_D_def2_TZVPPD_PCM_G2__Force = "wB97X-D/def2-TZVPPD/PCM(G2) Force"
    wB97X_D_def2_TZVPPD_PCM_G2__Geometry_Optimization = "wB97X-D/def2-TZVPPD/PCM(G2) Geometry Optimization"
    wB97X_D_def2_TZVPPD_PCM_G2__Frequency_Analysis = "wB97X-D/def2-TZVPPD/PCM(G2) Frequency Analysis"
    wB97X_D_def2_TZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_G2__Unknown = "wB97X-D/def2-TZVPPD/PCM(G2) Unknown"
    wB97X_D_def2_TZVPPD_PCM_DME__Single_Point = "wB97X-D/def2-TZVPPD/PCM(DME) Single Point"
    wB97X_D_def2_TZVPPD_PCM_DME__Force = "wB97X-D/def2-TZVPPD/PCM(DME) Force"
    wB97X_D_def2_TZVPPD_PCM_DME__Geometry_Optimization = "wB97X-D/def2-TZVPPD/PCM(DME) Geometry Optimization"
    wB97X_D_def2_TZVPPD_PCM_DME__Frequency_Analysis = "wB97X-D/def2-TZVPPD/PCM(DME) Frequency Analysis"
    wB97X_D_def2_TZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_DME__Unknown = "wB97X-D/def2-TZVPPD/PCM(DME) Unknown"
    wB97X_D_def2_TZVPPD_PCM_EC__Single_Point = "wB97X-D/def2-TZVPPD/PCM(EC) Single Point"
    wB97X_D_def2_TZVPPD_PCM_EC__Force = "wB97X-D/def2-TZVPPD/PCM(EC) Force"
    wB97X_D_def2_TZVPPD_PCM_EC__Geometry_Optimization = "wB97X-D/def2-TZVPPD/PCM(EC) Geometry Optimization"
    wB97X_D_def2_TZVPPD_PCM_EC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/PCM(EC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_EC__Unknown = "wB97X-D/def2-TZVPPD/PCM(EC) Unknown"
    wB97X_D_def2_TZVPPD_PCM_EC_EMC__Single_Point = "wB97X-D/def2-TZVPPD/PCM(EC/EMC) Single Point"
    wB97X_D_def2_TZVPPD_PCM_EC_EMC__Force = "wB97X-D/def2-TZVPPD/PCM(EC/EMC) Force"
    wB97X_D_def2_TZVPPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVPPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVPPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_EC_EMC__Unknown = "wB97X-D/def2-TZVPPD/PCM(EC/EMC) Unknown"
    wB97X_D_def2_TZVPPD_PCM_PC__Single_Point = "wB97X-D/def2-TZVPPD/PCM(PC) Single Point"
    wB97X_D_def2_TZVPPD_PCM_PC__Force = "wB97X-D/def2-TZVPPD/PCM(PC) Force"
    wB97X_D_def2_TZVPPD_PCM_PC__Geometry_Optimization = "wB97X-D/def2-TZVPPD/PCM(PC) Geometry Optimization"
    wB97X_D_def2_TZVPPD_PCM_PC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/PCM(PC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_PCM_PC__Unknown = "wB97X-D/def2-TZVPPD/PCM(PC) Unknown"
    wB97X_D_def2_TZVPPD_ISOSVP_WATER__Single_Point = "wB97X-D/def2-TZVPPD/ISOSVP(WATER) Single Point"
    wB97X_D_def2_TZVPPD_ISOSVP_WATER__Force = "wB97X-D/def2-TZVPPD/ISOSVP(WATER) Force"
    wB97X_D_def2_TZVPPD_ISOSVP_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPPD/ISOSVP(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_WATER__Unknown = "wB97X-D/def2-TZVPPD/ISOSVP(WATER) Unknown"
    wB97X_D_def2_TZVPPD_ISOSVP_THF__Single_Point = "wB97X-D/def2-TZVPPD/ISOSVP(THF) Single Point"
    wB97X_D_def2_TZVPPD_ISOSVP_THF__Force = "wB97X-D/def2-TZVPPD/ISOSVP(THF) Force"
    wB97X_D_def2_TZVPPD_ISOSVP_THF__Geometry_Optimization = "wB97X-D/def2-TZVPPD/ISOSVP(THF) Geometry Optimization"
    wB97X_D_def2_TZVPPD_ISOSVP_THF__Frequency_Analysis = "wB97X-D/def2-TZVPPD/ISOSVP(THF) Frequency Analysis"
    wB97X_D_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_THF__Unknown = "wB97X-D/def2-TZVPPD/ISOSVP(THF) Unknown"
    wB97X_D_def2_TZVPPD_ISOSVP_G2__Single_Point = "wB97X-D/def2-TZVPPD/ISOSVP(G2) Single Point"
    wB97X_D_def2_TZVPPD_ISOSVP_G2__Force = "wB97X-D/def2-TZVPPD/ISOSVP(G2) Force"
    wB97X_D_def2_TZVPPD_ISOSVP_G2__Geometry_Optimization = "wB97X-D/def2-TZVPPD/ISOSVP(G2) Geometry Optimization"
    wB97X_D_def2_TZVPPD_ISOSVP_G2__Frequency_Analysis = "wB97X-D/def2-TZVPPD/ISOSVP(G2) Frequency Analysis"
    wB97X_D_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_G2__Unknown = "wB97X-D/def2-TZVPPD/ISOSVP(G2) Unknown"
    wB97X_D_def2_TZVPPD_ISOSVP_DME__Single_Point = "wB97X-D/def2-TZVPPD/ISOSVP(DME) Single Point"
    wB97X_D_def2_TZVPPD_ISOSVP_DME__Force = "wB97X-D/def2-TZVPPD/ISOSVP(DME) Force"
    wB97X_D_def2_TZVPPD_ISOSVP_DME__Geometry_Optimization = "wB97X-D/def2-TZVPPD/ISOSVP(DME) Geometry Optimization"
    wB97X_D_def2_TZVPPD_ISOSVP_DME__Frequency_Analysis = "wB97X-D/def2-TZVPPD/ISOSVP(DME) Frequency Analysis"
    wB97X_D_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_DME__Unknown = "wB97X-D/def2-TZVPPD/ISOSVP(DME) Unknown"
    wB97X_D_def2_TZVPPD_ISOSVP_EC__Single_Point = "wB97X-D/def2-TZVPPD/ISOSVP(EC) Single Point"
    wB97X_D_def2_TZVPPD_ISOSVP_EC__Force = "wB97X-D/def2-TZVPPD/ISOSVP(EC) Force"
    wB97X_D_def2_TZVPPD_ISOSVP_EC__Geometry_Optimization = "wB97X-D/def2-TZVPPD/ISOSVP(EC) Geometry Optimization"
    wB97X_D_def2_TZVPPD_ISOSVP_EC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/ISOSVP(EC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_EC__Unknown = "wB97X-D/def2-TZVPPD/ISOSVP(EC) Unknown"
    wB97X_D_def2_TZVPPD_ISOSVP_EC_EMC__Single_Point = "wB97X-D/def2-TZVPPD/ISOSVP(EC/EMC) Single Point"
    wB97X_D_def2_TZVPPD_ISOSVP_EC_EMC__Force = "wB97X-D/def2-TZVPPD/ISOSVP(EC/EMC) Force"
    wB97X_D_def2_TZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_EC_EMC__Unknown = "wB97X-D/def2-TZVPPD/ISOSVP(EC/EMC) Unknown"
    wB97X_D_def2_TZVPPD_ISOSVP_PC__Single_Point = "wB97X-D/def2-TZVPPD/ISOSVP(PC) Single Point"
    wB97X_D_def2_TZVPPD_ISOSVP_PC__Force = "wB97X-D/def2-TZVPPD/ISOSVP(PC) Force"
    wB97X_D_def2_TZVPPD_ISOSVP_PC__Geometry_Optimization = "wB97X-D/def2-TZVPPD/ISOSVP(PC) Geometry Optimization"
    wB97X_D_def2_TZVPPD_ISOSVP_PC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/ISOSVP(PC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_ISOSVP_PC__Unknown = "wB97X-D/def2-TZVPPD/ISOSVP(PC) Unknown"
    wB97X_D_def2_TZVPPD_SMD_WATER__Single_Point = "wB97X-D/def2-TZVPPD/SMD(WATER) Single Point"
    wB97X_D_def2_TZVPPD_SMD_WATER__Force = "wB97X-D/def2-TZVPPD/SMD(WATER) Force"
    wB97X_D_def2_TZVPPD_SMD_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPPD/SMD(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPPD_SMD_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPPD/SMD(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_WATER__Unknown = "wB97X-D/def2-TZVPPD/SMD(WATER) Unknown"
    wB97X_D_def2_TZVPPD_SMD_THF__Single_Point = "wB97X-D/def2-TZVPPD/SMD(THF) Single Point"
    wB97X_D_def2_TZVPPD_SMD_THF__Force = "wB97X-D/def2-TZVPPD/SMD(THF) Force"
    wB97X_D_def2_TZVPPD_SMD_THF__Geometry_Optimization = "wB97X-D/def2-TZVPPD/SMD(THF) Geometry Optimization"
    wB97X_D_def2_TZVPPD_SMD_THF__Frequency_Analysis = "wB97X-D/def2-TZVPPD/SMD(THF) Frequency Analysis"
    wB97X_D_def2_TZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_THF__Unknown = "wB97X-D/def2-TZVPPD/SMD(THF) Unknown"
    wB97X_D_def2_TZVPPD_SMD_G2__Single_Point = "wB97X-D/def2-TZVPPD/SMD(G2) Single Point"
    wB97X_D_def2_TZVPPD_SMD_G2__Force = "wB97X-D/def2-TZVPPD/SMD(G2) Force"
    wB97X_D_def2_TZVPPD_SMD_G2__Geometry_Optimization = "wB97X-D/def2-TZVPPD/SMD(G2) Geometry Optimization"
    wB97X_D_def2_TZVPPD_SMD_G2__Frequency_Analysis = "wB97X-D/def2-TZVPPD/SMD(G2) Frequency Analysis"
    wB97X_D_def2_TZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_G2__Unknown = "wB97X-D/def2-TZVPPD/SMD(G2) Unknown"
    wB97X_D_def2_TZVPPD_SMD_DME__Single_Point = "wB97X-D/def2-TZVPPD/SMD(DME) Single Point"
    wB97X_D_def2_TZVPPD_SMD_DME__Force = "wB97X-D/def2-TZVPPD/SMD(DME) Force"
    wB97X_D_def2_TZVPPD_SMD_DME__Geometry_Optimization = "wB97X-D/def2-TZVPPD/SMD(DME) Geometry Optimization"
    wB97X_D_def2_TZVPPD_SMD_DME__Frequency_Analysis = "wB97X-D/def2-TZVPPD/SMD(DME) Frequency Analysis"
    wB97X_D_def2_TZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_DME__Unknown = "wB97X-D/def2-TZVPPD/SMD(DME) Unknown"
    wB97X_D_def2_TZVPPD_SMD_EC__Single_Point = "wB97X-D/def2-TZVPPD/SMD(EC) Single Point"
    wB97X_D_def2_TZVPPD_SMD_EC__Force = "wB97X-D/def2-TZVPPD/SMD(EC) Force"
    wB97X_D_def2_TZVPPD_SMD_EC__Geometry_Optimization = "wB97X-D/def2-TZVPPD/SMD(EC) Geometry Optimization"
    wB97X_D_def2_TZVPPD_SMD_EC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/SMD(EC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_EC__Unknown = "wB97X-D/def2-TZVPPD/SMD(EC) Unknown"
    wB97X_D_def2_TZVPPD_SMD_EC_EMC__Single_Point = "wB97X-D/def2-TZVPPD/SMD(EC/EMC) Single Point"
    wB97X_D_def2_TZVPPD_SMD_EC_EMC__Force = "wB97X-D/def2-TZVPPD/SMD(EC/EMC) Force"
    wB97X_D_def2_TZVPPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVPPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVPPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_EC_EMC__Unknown = "wB97X-D/def2-TZVPPD/SMD(EC/EMC) Unknown"
    wB97X_D_def2_TZVPPD_SMD_PC__Single_Point = "wB97X-D/def2-TZVPPD/SMD(PC) Single Point"
    wB97X_D_def2_TZVPPD_SMD_PC__Force = "wB97X-D/def2-TZVPPD/SMD(PC) Force"
    wB97X_D_def2_TZVPPD_SMD_PC__Geometry_Optimization = "wB97X-D/def2-TZVPPD/SMD(PC) Geometry Optimization"
    wB97X_D_def2_TZVPPD_SMD_PC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/SMD(PC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_SMD_PC__Unknown = "wB97X-D/def2-TZVPPD/SMD(PC) Unknown"
    wB97X_D_def2_TZVPPD_CMIRS_WATER__Single_Point = "wB97X-D/def2-TZVPPD/CMIRS(WATER) Single Point"
    wB97X_D_def2_TZVPPD_CMIRS_WATER__Force = "wB97X-D/def2-TZVPPD/CMIRS(WATER) Force"
    wB97X_D_def2_TZVPPD_CMIRS_WATER__Geometry_Optimization = "wB97X-D/def2-TZVPPD/CMIRS(WATER) Geometry Optimization"
    wB97X_D_def2_TZVPPD_CMIRS_WATER__Frequency_Analysis = "wB97X-D/def2-TZVPPD/CMIRS(WATER) Frequency Analysis"
    wB97X_D_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_WATER__Unknown = "wB97X-D/def2-TZVPPD/CMIRS(WATER) Unknown"
    wB97X_D_def2_TZVPPD_CMIRS_THF__Single_Point = "wB97X-D/def2-TZVPPD/CMIRS(THF) Single Point"
    wB97X_D_def2_TZVPPD_CMIRS_THF__Force = "wB97X-D/def2-TZVPPD/CMIRS(THF) Force"
    wB97X_D_def2_TZVPPD_CMIRS_THF__Geometry_Optimization = "wB97X-D/def2-TZVPPD/CMIRS(THF) Geometry Optimization"
    wB97X_D_def2_TZVPPD_CMIRS_THF__Frequency_Analysis = "wB97X-D/def2-TZVPPD/CMIRS(THF) Frequency Analysis"
    wB97X_D_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_THF__Unknown = "wB97X-D/def2-TZVPPD/CMIRS(THF) Unknown"
    wB97X_D_def2_TZVPPD_CMIRS_G2__Single_Point = "wB97X-D/def2-TZVPPD/CMIRS(G2) Single Point"
    wB97X_D_def2_TZVPPD_CMIRS_G2__Force = "wB97X-D/def2-TZVPPD/CMIRS(G2) Force"
    wB97X_D_def2_TZVPPD_CMIRS_G2__Geometry_Optimization = "wB97X-D/def2-TZVPPD/CMIRS(G2) Geometry Optimization"
    wB97X_D_def2_TZVPPD_CMIRS_G2__Frequency_Analysis = "wB97X-D/def2-TZVPPD/CMIRS(G2) Frequency Analysis"
    wB97X_D_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_G2__Unknown = "wB97X-D/def2-TZVPPD/CMIRS(G2) Unknown"
    wB97X_D_def2_TZVPPD_CMIRS_DME__Single_Point = "wB97X-D/def2-TZVPPD/CMIRS(DME) Single Point"
    wB97X_D_def2_TZVPPD_CMIRS_DME__Force = "wB97X-D/def2-TZVPPD/CMIRS(DME) Force"
    wB97X_D_def2_TZVPPD_CMIRS_DME__Geometry_Optimization = "wB97X-D/def2-TZVPPD/CMIRS(DME) Geometry Optimization"
    wB97X_D_def2_TZVPPD_CMIRS_DME__Frequency_Analysis = "wB97X-D/def2-TZVPPD/CMIRS(DME) Frequency Analysis"
    wB97X_D_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_DME__Unknown = "wB97X-D/def2-TZVPPD/CMIRS(DME) Unknown"
    wB97X_D_def2_TZVPPD_CMIRS_EC__Single_Point = "wB97X-D/def2-TZVPPD/CMIRS(EC) Single Point"
    wB97X_D_def2_TZVPPD_CMIRS_EC__Force = "wB97X-D/def2-TZVPPD/CMIRS(EC) Force"
    wB97X_D_def2_TZVPPD_CMIRS_EC__Geometry_Optimization = "wB97X-D/def2-TZVPPD/CMIRS(EC) Geometry Optimization"
    wB97X_D_def2_TZVPPD_CMIRS_EC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/CMIRS(EC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_EC__Unknown = "wB97X-D/def2-TZVPPD/CMIRS(EC) Unknown"
    wB97X_D_def2_TZVPPD_CMIRS_EC_EMC__Single_Point = "wB97X-D/def2-TZVPPD/CMIRS(EC/EMC) Single Point"
    wB97X_D_def2_TZVPPD_CMIRS_EC_EMC__Force = "wB97X-D/def2-TZVPPD/CMIRS(EC/EMC) Force"
    wB97X_D_def2_TZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D/def2-TZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_EC_EMC__Unknown = "wB97X-D/def2-TZVPPD/CMIRS(EC/EMC) Unknown"
    wB97X_D_def2_TZVPPD_CMIRS_PC__Single_Point = "wB97X-D/def2-TZVPPD/CMIRS(PC) Single Point"
    wB97X_D_def2_TZVPPD_CMIRS_PC__Force = "wB97X-D/def2-TZVPPD/CMIRS(PC) Force"
    wB97X_D_def2_TZVPPD_CMIRS_PC__Geometry_Optimization = "wB97X-D/def2-TZVPPD/CMIRS(PC) Geometry Optimization"
    wB97X_D_def2_TZVPPD_CMIRS_PC__Frequency_Analysis = "wB97X-D/def2-TZVPPD/CMIRS(PC) Frequency Analysis"
    wB97X_D_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-TZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_TZVPPD_CMIRS_PC__Unknown = "wB97X-D/def2-TZVPPD/CMIRS(PC) Unknown"
    wB97X_D_def2_QZVPPD_VACUUM_Single_Point = "wB97X-D/def2-QZVPPD/VACUUM Single Point"
    wB97X_D_def2_QZVPPD_VACUUM_Force = "wB97X-D/def2-QZVPPD/VACUUM Force"
    wB97X_D_def2_QZVPPD_VACUUM_Geometry_Optimization = "wB97X-D/def2-QZVPPD/VACUUM Geometry Optimization"
    wB97X_D_def2_QZVPPD_VACUUM_Frequency_Analysis = "wB97X-D/def2-QZVPPD/VACUUM Frequency Analysis"
    wB97X_D_def2_QZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_VACUUM_Unknown = "wB97X-D/def2-QZVPPD/VACUUM Unknown"
    wB97X_D_def2_QZVPPD_PCM_WATER__Single_Point = "wB97X-D/def2-QZVPPD/PCM(WATER) Single Point"
    wB97X_D_def2_QZVPPD_PCM_WATER__Force = "wB97X-D/def2-QZVPPD/PCM(WATER) Force"
    wB97X_D_def2_QZVPPD_PCM_WATER__Geometry_Optimization = "wB97X-D/def2-QZVPPD/PCM(WATER) Geometry Optimization"
    wB97X_D_def2_QZVPPD_PCM_WATER__Frequency_Analysis = "wB97X-D/def2-QZVPPD/PCM(WATER) Frequency Analysis"
    wB97X_D_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_WATER__Unknown = "wB97X-D/def2-QZVPPD/PCM(WATER) Unknown"
    wB97X_D_def2_QZVPPD_PCM_THF__Single_Point = "wB97X-D/def2-QZVPPD/PCM(THF) Single Point"
    wB97X_D_def2_QZVPPD_PCM_THF__Force = "wB97X-D/def2-QZVPPD/PCM(THF) Force"
    wB97X_D_def2_QZVPPD_PCM_THF__Geometry_Optimization = "wB97X-D/def2-QZVPPD/PCM(THF) Geometry Optimization"
    wB97X_D_def2_QZVPPD_PCM_THF__Frequency_Analysis = "wB97X-D/def2-QZVPPD/PCM(THF) Frequency Analysis"
    wB97X_D_def2_QZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_THF__Unknown = "wB97X-D/def2-QZVPPD/PCM(THF) Unknown"
    wB97X_D_def2_QZVPPD_PCM_G2__Single_Point = "wB97X-D/def2-QZVPPD/PCM(G2) Single Point"
    wB97X_D_def2_QZVPPD_PCM_G2__Force = "wB97X-D/def2-QZVPPD/PCM(G2) Force"
    wB97X_D_def2_QZVPPD_PCM_G2__Geometry_Optimization = "wB97X-D/def2-QZVPPD/PCM(G2) Geometry Optimization"
    wB97X_D_def2_QZVPPD_PCM_G2__Frequency_Analysis = "wB97X-D/def2-QZVPPD/PCM(G2) Frequency Analysis"
    wB97X_D_def2_QZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_G2__Unknown = "wB97X-D/def2-QZVPPD/PCM(G2) Unknown"
    wB97X_D_def2_QZVPPD_PCM_DME__Single_Point = "wB97X-D/def2-QZVPPD/PCM(DME) Single Point"
    wB97X_D_def2_QZVPPD_PCM_DME__Force = "wB97X-D/def2-QZVPPD/PCM(DME) Force"
    wB97X_D_def2_QZVPPD_PCM_DME__Geometry_Optimization = "wB97X-D/def2-QZVPPD/PCM(DME) Geometry Optimization"
    wB97X_D_def2_QZVPPD_PCM_DME__Frequency_Analysis = "wB97X-D/def2-QZVPPD/PCM(DME) Frequency Analysis"
    wB97X_D_def2_QZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_DME__Unknown = "wB97X-D/def2-QZVPPD/PCM(DME) Unknown"
    wB97X_D_def2_QZVPPD_PCM_EC__Single_Point = "wB97X-D/def2-QZVPPD/PCM(EC) Single Point"
    wB97X_D_def2_QZVPPD_PCM_EC__Force = "wB97X-D/def2-QZVPPD/PCM(EC) Force"
    wB97X_D_def2_QZVPPD_PCM_EC__Geometry_Optimization = "wB97X-D/def2-QZVPPD/PCM(EC) Geometry Optimization"
    wB97X_D_def2_QZVPPD_PCM_EC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/PCM(EC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_EC__Unknown = "wB97X-D/def2-QZVPPD/PCM(EC) Unknown"
    wB97X_D_def2_QZVPPD_PCM_EC_EMC__Single_Point = "wB97X-D/def2-QZVPPD/PCM(EC/EMC) Single Point"
    wB97X_D_def2_QZVPPD_PCM_EC_EMC__Force = "wB97X-D/def2-QZVPPD/PCM(EC/EMC) Force"
    wB97X_D_def2_QZVPPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-D/def2-QZVPPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_D_def2_QZVPPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_EC_EMC__Unknown = "wB97X-D/def2-QZVPPD/PCM(EC/EMC) Unknown"
    wB97X_D_def2_QZVPPD_PCM_PC__Single_Point = "wB97X-D/def2-QZVPPD/PCM(PC) Single Point"
    wB97X_D_def2_QZVPPD_PCM_PC__Force = "wB97X-D/def2-QZVPPD/PCM(PC) Force"
    wB97X_D_def2_QZVPPD_PCM_PC__Geometry_Optimization = "wB97X-D/def2-QZVPPD/PCM(PC) Geometry Optimization"
    wB97X_D_def2_QZVPPD_PCM_PC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/PCM(PC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_PCM_PC__Unknown = "wB97X-D/def2-QZVPPD/PCM(PC) Unknown"
    wB97X_D_def2_QZVPPD_ISOSVP_WATER__Single_Point = "wB97X-D/def2-QZVPPD/ISOSVP(WATER) Single Point"
    wB97X_D_def2_QZVPPD_ISOSVP_WATER__Force = "wB97X-D/def2-QZVPPD/ISOSVP(WATER) Force"
    wB97X_D_def2_QZVPPD_ISOSVP_WATER__Geometry_Optimization = "wB97X-D/def2-QZVPPD/ISOSVP(WATER) Geometry Optimization"
    wB97X_D_def2_QZVPPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-D/def2-QZVPPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_D_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_WATER__Unknown = "wB97X-D/def2-QZVPPD/ISOSVP(WATER) Unknown"
    wB97X_D_def2_QZVPPD_ISOSVP_THF__Single_Point = "wB97X-D/def2-QZVPPD/ISOSVP(THF) Single Point"
    wB97X_D_def2_QZVPPD_ISOSVP_THF__Force = "wB97X-D/def2-QZVPPD/ISOSVP(THF) Force"
    wB97X_D_def2_QZVPPD_ISOSVP_THF__Geometry_Optimization = "wB97X-D/def2-QZVPPD/ISOSVP(THF) Geometry Optimization"
    wB97X_D_def2_QZVPPD_ISOSVP_THF__Frequency_Analysis = "wB97X-D/def2-QZVPPD/ISOSVP(THF) Frequency Analysis"
    wB97X_D_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_THF__Unknown = "wB97X-D/def2-QZVPPD/ISOSVP(THF) Unknown"
    wB97X_D_def2_QZVPPD_ISOSVP_G2__Single_Point = "wB97X-D/def2-QZVPPD/ISOSVP(G2) Single Point"
    wB97X_D_def2_QZVPPD_ISOSVP_G2__Force = "wB97X-D/def2-QZVPPD/ISOSVP(G2) Force"
    wB97X_D_def2_QZVPPD_ISOSVP_G2__Geometry_Optimization = "wB97X-D/def2-QZVPPD/ISOSVP(G2) Geometry Optimization"
    wB97X_D_def2_QZVPPD_ISOSVP_G2__Frequency_Analysis = "wB97X-D/def2-QZVPPD/ISOSVP(G2) Frequency Analysis"
    wB97X_D_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_G2__Unknown = "wB97X-D/def2-QZVPPD/ISOSVP(G2) Unknown"
    wB97X_D_def2_QZVPPD_ISOSVP_DME__Single_Point = "wB97X-D/def2-QZVPPD/ISOSVP(DME) Single Point"
    wB97X_D_def2_QZVPPD_ISOSVP_DME__Force = "wB97X-D/def2-QZVPPD/ISOSVP(DME) Force"
    wB97X_D_def2_QZVPPD_ISOSVP_DME__Geometry_Optimization = "wB97X-D/def2-QZVPPD/ISOSVP(DME) Geometry Optimization"
    wB97X_D_def2_QZVPPD_ISOSVP_DME__Frequency_Analysis = "wB97X-D/def2-QZVPPD/ISOSVP(DME) Frequency Analysis"
    wB97X_D_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_DME__Unknown = "wB97X-D/def2-QZVPPD/ISOSVP(DME) Unknown"
    wB97X_D_def2_QZVPPD_ISOSVP_EC__Single_Point = "wB97X-D/def2-QZVPPD/ISOSVP(EC) Single Point"
    wB97X_D_def2_QZVPPD_ISOSVP_EC__Force = "wB97X-D/def2-QZVPPD/ISOSVP(EC) Force"
    wB97X_D_def2_QZVPPD_ISOSVP_EC__Geometry_Optimization = "wB97X-D/def2-QZVPPD/ISOSVP(EC) Geometry Optimization"
    wB97X_D_def2_QZVPPD_ISOSVP_EC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/ISOSVP(EC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_EC__Unknown = "wB97X-D/def2-QZVPPD/ISOSVP(EC) Unknown"
    wB97X_D_def2_QZVPPD_ISOSVP_EC_EMC__Single_Point = "wB97X-D/def2-QZVPPD/ISOSVP(EC/EMC) Single Point"
    wB97X_D_def2_QZVPPD_ISOSVP_EC_EMC__Force = "wB97X-D/def2-QZVPPD/ISOSVP(EC/EMC) Force"
    wB97X_D_def2_QZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_EC_EMC__Unknown = "wB97X-D/def2-QZVPPD/ISOSVP(EC/EMC) Unknown"
    wB97X_D_def2_QZVPPD_ISOSVP_PC__Single_Point = "wB97X-D/def2-QZVPPD/ISOSVP(PC) Single Point"
    wB97X_D_def2_QZVPPD_ISOSVP_PC__Force = "wB97X-D/def2-QZVPPD/ISOSVP(PC) Force"
    wB97X_D_def2_QZVPPD_ISOSVP_PC__Geometry_Optimization = "wB97X-D/def2-QZVPPD/ISOSVP(PC) Geometry Optimization"
    wB97X_D_def2_QZVPPD_ISOSVP_PC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/ISOSVP(PC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_ISOSVP_PC__Unknown = "wB97X-D/def2-QZVPPD/ISOSVP(PC) Unknown"
    wB97X_D_def2_QZVPPD_SMD_WATER__Single_Point = "wB97X-D/def2-QZVPPD/SMD(WATER) Single Point"
    wB97X_D_def2_QZVPPD_SMD_WATER__Force = "wB97X-D/def2-QZVPPD/SMD(WATER) Force"
    wB97X_D_def2_QZVPPD_SMD_WATER__Geometry_Optimization = "wB97X-D/def2-QZVPPD/SMD(WATER) Geometry Optimization"
    wB97X_D_def2_QZVPPD_SMD_WATER__Frequency_Analysis = "wB97X-D/def2-QZVPPD/SMD(WATER) Frequency Analysis"
    wB97X_D_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_WATER__Unknown = "wB97X-D/def2-QZVPPD/SMD(WATER) Unknown"
    wB97X_D_def2_QZVPPD_SMD_THF__Single_Point = "wB97X-D/def2-QZVPPD/SMD(THF) Single Point"
    wB97X_D_def2_QZVPPD_SMD_THF__Force = "wB97X-D/def2-QZVPPD/SMD(THF) Force"
    wB97X_D_def2_QZVPPD_SMD_THF__Geometry_Optimization = "wB97X-D/def2-QZVPPD/SMD(THF) Geometry Optimization"
    wB97X_D_def2_QZVPPD_SMD_THF__Frequency_Analysis = "wB97X-D/def2-QZVPPD/SMD(THF) Frequency Analysis"
    wB97X_D_def2_QZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_THF__Unknown = "wB97X-D/def2-QZVPPD/SMD(THF) Unknown"
    wB97X_D_def2_QZVPPD_SMD_G2__Single_Point = "wB97X-D/def2-QZVPPD/SMD(G2) Single Point"
    wB97X_D_def2_QZVPPD_SMD_G2__Force = "wB97X-D/def2-QZVPPD/SMD(G2) Force"
    wB97X_D_def2_QZVPPD_SMD_G2__Geometry_Optimization = "wB97X-D/def2-QZVPPD/SMD(G2) Geometry Optimization"
    wB97X_D_def2_QZVPPD_SMD_G2__Frequency_Analysis = "wB97X-D/def2-QZVPPD/SMD(G2) Frequency Analysis"
    wB97X_D_def2_QZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_G2__Unknown = "wB97X-D/def2-QZVPPD/SMD(G2) Unknown"
    wB97X_D_def2_QZVPPD_SMD_DME__Single_Point = "wB97X-D/def2-QZVPPD/SMD(DME) Single Point"
    wB97X_D_def2_QZVPPD_SMD_DME__Force = "wB97X-D/def2-QZVPPD/SMD(DME) Force"
    wB97X_D_def2_QZVPPD_SMD_DME__Geometry_Optimization = "wB97X-D/def2-QZVPPD/SMD(DME) Geometry Optimization"
    wB97X_D_def2_QZVPPD_SMD_DME__Frequency_Analysis = "wB97X-D/def2-QZVPPD/SMD(DME) Frequency Analysis"
    wB97X_D_def2_QZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_DME__Unknown = "wB97X-D/def2-QZVPPD/SMD(DME) Unknown"
    wB97X_D_def2_QZVPPD_SMD_EC__Single_Point = "wB97X-D/def2-QZVPPD/SMD(EC) Single Point"
    wB97X_D_def2_QZVPPD_SMD_EC__Force = "wB97X-D/def2-QZVPPD/SMD(EC) Force"
    wB97X_D_def2_QZVPPD_SMD_EC__Geometry_Optimization = "wB97X-D/def2-QZVPPD/SMD(EC) Geometry Optimization"
    wB97X_D_def2_QZVPPD_SMD_EC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/SMD(EC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_EC__Unknown = "wB97X-D/def2-QZVPPD/SMD(EC) Unknown"
    wB97X_D_def2_QZVPPD_SMD_EC_EMC__Single_Point = "wB97X-D/def2-QZVPPD/SMD(EC/EMC) Single Point"
    wB97X_D_def2_QZVPPD_SMD_EC_EMC__Force = "wB97X-D/def2-QZVPPD/SMD(EC/EMC) Force"
    wB97X_D_def2_QZVPPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-D/def2-QZVPPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_D_def2_QZVPPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_EC_EMC__Unknown = "wB97X-D/def2-QZVPPD/SMD(EC/EMC) Unknown"
    wB97X_D_def2_QZVPPD_SMD_PC__Single_Point = "wB97X-D/def2-QZVPPD/SMD(PC) Single Point"
    wB97X_D_def2_QZVPPD_SMD_PC__Force = "wB97X-D/def2-QZVPPD/SMD(PC) Force"
    wB97X_D_def2_QZVPPD_SMD_PC__Geometry_Optimization = "wB97X-D/def2-QZVPPD/SMD(PC) Geometry Optimization"
    wB97X_D_def2_QZVPPD_SMD_PC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/SMD(PC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_SMD_PC__Unknown = "wB97X-D/def2-QZVPPD/SMD(PC) Unknown"
    wB97X_D_def2_QZVPPD_CMIRS_WATER__Single_Point = "wB97X-D/def2-QZVPPD/CMIRS(WATER) Single Point"
    wB97X_D_def2_QZVPPD_CMIRS_WATER__Force = "wB97X-D/def2-QZVPPD/CMIRS(WATER) Force"
    wB97X_D_def2_QZVPPD_CMIRS_WATER__Geometry_Optimization = "wB97X-D/def2-QZVPPD/CMIRS(WATER) Geometry Optimization"
    wB97X_D_def2_QZVPPD_CMIRS_WATER__Frequency_Analysis = "wB97X-D/def2-QZVPPD/CMIRS(WATER) Frequency Analysis"
    wB97X_D_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_WATER__Unknown = "wB97X-D/def2-QZVPPD/CMIRS(WATER) Unknown"
    wB97X_D_def2_QZVPPD_CMIRS_THF__Single_Point = "wB97X-D/def2-QZVPPD/CMIRS(THF) Single Point"
    wB97X_D_def2_QZVPPD_CMIRS_THF__Force = "wB97X-D/def2-QZVPPD/CMIRS(THF) Force"
    wB97X_D_def2_QZVPPD_CMIRS_THF__Geometry_Optimization = "wB97X-D/def2-QZVPPD/CMIRS(THF) Geometry Optimization"
    wB97X_D_def2_QZVPPD_CMIRS_THF__Frequency_Analysis = "wB97X-D/def2-QZVPPD/CMIRS(THF) Frequency Analysis"
    wB97X_D_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_THF__Unknown = "wB97X-D/def2-QZVPPD/CMIRS(THF) Unknown"
    wB97X_D_def2_QZVPPD_CMIRS_G2__Single_Point = "wB97X-D/def2-QZVPPD/CMIRS(G2) Single Point"
    wB97X_D_def2_QZVPPD_CMIRS_G2__Force = "wB97X-D/def2-QZVPPD/CMIRS(G2) Force"
    wB97X_D_def2_QZVPPD_CMIRS_G2__Geometry_Optimization = "wB97X-D/def2-QZVPPD/CMIRS(G2) Geometry Optimization"
    wB97X_D_def2_QZVPPD_CMIRS_G2__Frequency_Analysis = "wB97X-D/def2-QZVPPD/CMIRS(G2) Frequency Analysis"
    wB97X_D_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_G2__Unknown = "wB97X-D/def2-QZVPPD/CMIRS(G2) Unknown"
    wB97X_D_def2_QZVPPD_CMIRS_DME__Single_Point = "wB97X-D/def2-QZVPPD/CMIRS(DME) Single Point"
    wB97X_D_def2_QZVPPD_CMIRS_DME__Force = "wB97X-D/def2-QZVPPD/CMIRS(DME) Force"
    wB97X_D_def2_QZVPPD_CMIRS_DME__Geometry_Optimization = "wB97X-D/def2-QZVPPD/CMIRS(DME) Geometry Optimization"
    wB97X_D_def2_QZVPPD_CMIRS_DME__Frequency_Analysis = "wB97X-D/def2-QZVPPD/CMIRS(DME) Frequency Analysis"
    wB97X_D_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_DME__Unknown = "wB97X-D/def2-QZVPPD/CMIRS(DME) Unknown"
    wB97X_D_def2_QZVPPD_CMIRS_EC__Single_Point = "wB97X-D/def2-QZVPPD/CMIRS(EC) Single Point"
    wB97X_D_def2_QZVPPD_CMIRS_EC__Force = "wB97X-D/def2-QZVPPD/CMIRS(EC) Force"
    wB97X_D_def2_QZVPPD_CMIRS_EC__Geometry_Optimization = "wB97X-D/def2-QZVPPD/CMIRS(EC) Geometry Optimization"
    wB97X_D_def2_QZVPPD_CMIRS_EC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/CMIRS(EC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_EC__Unknown = "wB97X-D/def2-QZVPPD/CMIRS(EC) Unknown"
    wB97X_D_def2_QZVPPD_CMIRS_EC_EMC__Single_Point = "wB97X-D/def2-QZVPPD/CMIRS(EC/EMC) Single Point"
    wB97X_D_def2_QZVPPD_CMIRS_EC_EMC__Force = "wB97X-D/def2-QZVPPD/CMIRS(EC/EMC) Force"
    wB97X_D_def2_QZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D/def2-QZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_EC_EMC__Unknown = "wB97X-D/def2-QZVPPD/CMIRS(EC/EMC) Unknown"
    wB97X_D_def2_QZVPPD_CMIRS_PC__Single_Point = "wB97X-D/def2-QZVPPD/CMIRS(PC) Single Point"
    wB97X_D_def2_QZVPPD_CMIRS_PC__Force = "wB97X-D/def2-QZVPPD/CMIRS(PC) Force"
    wB97X_D_def2_QZVPPD_CMIRS_PC__Geometry_Optimization = "wB97X-D/def2-QZVPPD/CMIRS(PC) Geometry Optimization"
    wB97X_D_def2_QZVPPD_CMIRS_PC__Frequency_Analysis = "wB97X-D/def2-QZVPPD/CMIRS(PC) Frequency Analysis"
    wB97X_D_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D/def2-QZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D_def2_QZVPPD_CMIRS_PC__Unknown = "wB97X-D/def2-QZVPPD/CMIRS(PC) Unknown"
    wB97X_D3_6_31g_d_VACUUM_Single_Point = "wB97X-D3/6-31g*/VACUUM Single Point"
    wB97X_D3_6_31g_d_VACUUM_Force = "wB97X-D3/6-31g*/VACUUM Force"
    wB97X_D3_6_31g_d_VACUUM_Geometry_Optimization = "wB97X-D3/6-31g*/VACUUM Geometry Optimization"
    wB97X_D3_6_31g_d_VACUUM_Frequency_Analysis = "wB97X-D3/6-31g*/VACUUM Frequency Analysis"
    wB97X_D3_6_31g_d_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_VACUUM_Unknown = "wB97X-D3/6-31g*/VACUUM Unknown"
    wB97X_D3_6_31g_d_PCM_WATER__Single_Point = "wB97X-D3/6-31g*/PCM(WATER) Single Point"
    wB97X_D3_6_31g_d_PCM_WATER__Force = "wB97X-D3/6-31g*/PCM(WATER) Force"
    wB97X_D3_6_31g_d_PCM_WATER__Geometry_Optimization = "wB97X-D3/6-31g*/PCM(WATER) Geometry Optimization"
    wB97X_D3_6_31g_d_PCM_WATER__Frequency_Analysis = "wB97X-D3/6-31g*/PCM(WATER) Frequency Analysis"
    wB97X_D3_6_31g_d_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_WATER__Unknown = "wB97X-D3/6-31g*/PCM(WATER) Unknown"
    wB97X_D3_6_31g_d_PCM_THF__Single_Point = "wB97X-D3/6-31g*/PCM(THF) Single Point"
    wB97X_D3_6_31g_d_PCM_THF__Force = "wB97X-D3/6-31g*/PCM(THF) Force"
    wB97X_D3_6_31g_d_PCM_THF__Geometry_Optimization = "wB97X-D3/6-31g*/PCM(THF) Geometry Optimization"
    wB97X_D3_6_31g_d_PCM_THF__Frequency_Analysis = "wB97X-D3/6-31g*/PCM(THF) Frequency Analysis"
    wB97X_D3_6_31g_d_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_THF__Unknown = "wB97X-D3/6-31g*/PCM(THF) Unknown"
    wB97X_D3_6_31g_d_PCM_G2__Single_Point = "wB97X-D3/6-31g*/PCM(G2) Single Point"
    wB97X_D3_6_31g_d_PCM_G2__Force = "wB97X-D3/6-31g*/PCM(G2) Force"
    wB97X_D3_6_31g_d_PCM_G2__Geometry_Optimization = "wB97X-D3/6-31g*/PCM(G2) Geometry Optimization"
    wB97X_D3_6_31g_d_PCM_G2__Frequency_Analysis = "wB97X-D3/6-31g*/PCM(G2) Frequency Analysis"
    wB97X_D3_6_31g_d_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_G2__Unknown = "wB97X-D3/6-31g*/PCM(G2) Unknown"
    wB97X_D3_6_31g_d_PCM_DME__Single_Point = "wB97X-D3/6-31g*/PCM(DME) Single Point"
    wB97X_D3_6_31g_d_PCM_DME__Force = "wB97X-D3/6-31g*/PCM(DME) Force"
    wB97X_D3_6_31g_d_PCM_DME__Geometry_Optimization = "wB97X-D3/6-31g*/PCM(DME) Geometry Optimization"
    wB97X_D3_6_31g_d_PCM_DME__Frequency_Analysis = "wB97X-D3/6-31g*/PCM(DME) Frequency Analysis"
    wB97X_D3_6_31g_d_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_DME__Unknown = "wB97X-D3/6-31g*/PCM(DME) Unknown"
    wB97X_D3_6_31g_d_PCM_EC__Single_Point = "wB97X-D3/6-31g*/PCM(EC) Single Point"
    wB97X_D3_6_31g_d_PCM_EC__Force = "wB97X-D3/6-31g*/PCM(EC) Force"
    wB97X_D3_6_31g_d_PCM_EC__Geometry_Optimization = "wB97X-D3/6-31g*/PCM(EC) Geometry Optimization"
    wB97X_D3_6_31g_d_PCM_EC__Frequency_Analysis = "wB97X-D3/6-31g*/PCM(EC) Frequency Analysis"
    wB97X_D3_6_31g_d_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_EC__Unknown = "wB97X-D3/6-31g*/PCM(EC) Unknown"
    wB97X_D3_6_31g_d_PCM_EC_EMC__Single_Point = "wB97X-D3/6-31g*/PCM(EC/EMC) Single Point"
    wB97X_D3_6_31g_d_PCM_EC_EMC__Force = "wB97X-D3/6-31g*/PCM(EC/EMC) Force"
    wB97X_D3_6_31g_d_PCM_EC_EMC__Geometry_Optimization = "wB97X-D3/6-31g*/PCM(EC/EMC) Geometry Optimization"
    wB97X_D3_6_31g_d_PCM_EC_EMC__Frequency_Analysis = "wB97X-D3/6-31g*/PCM(EC/EMC) Frequency Analysis"
    wB97X_D3_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_EC_EMC__Unknown = "wB97X-D3/6-31g*/PCM(EC/EMC) Unknown"
    wB97X_D3_6_31g_d_PCM_PC__Single_Point = "wB97X-D3/6-31g*/PCM(PC) Single Point"
    wB97X_D3_6_31g_d_PCM_PC__Force = "wB97X-D3/6-31g*/PCM(PC) Force"
    wB97X_D3_6_31g_d_PCM_PC__Geometry_Optimization = "wB97X-D3/6-31g*/PCM(PC) Geometry Optimization"
    wB97X_D3_6_31g_d_PCM_PC__Frequency_Analysis = "wB97X-D3/6-31g*/PCM(PC) Frequency Analysis"
    wB97X_D3_6_31g_d_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_PCM_PC__Unknown = "wB97X-D3/6-31g*/PCM(PC) Unknown"
    wB97X_D3_6_31g_d_ISOSVP_WATER__Single_Point = "wB97X-D3/6-31g*/ISOSVP(WATER) Single Point"
    wB97X_D3_6_31g_d_ISOSVP_WATER__Force = "wB97X-D3/6-31g*/ISOSVP(WATER) Force"
    wB97X_D3_6_31g_d_ISOSVP_WATER__Geometry_Optimization = "wB97X-D3/6-31g*/ISOSVP(WATER) Geometry Optimization"
    wB97X_D3_6_31g_d_ISOSVP_WATER__Frequency_Analysis = "wB97X-D3/6-31g*/ISOSVP(WATER) Frequency Analysis"
    wB97X_D3_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_WATER__Unknown = "wB97X-D3/6-31g*/ISOSVP(WATER) Unknown"
    wB97X_D3_6_31g_d_ISOSVP_THF__Single_Point = "wB97X-D3/6-31g*/ISOSVP(THF) Single Point"
    wB97X_D3_6_31g_d_ISOSVP_THF__Force = "wB97X-D3/6-31g*/ISOSVP(THF) Force"
    wB97X_D3_6_31g_d_ISOSVP_THF__Geometry_Optimization = "wB97X-D3/6-31g*/ISOSVP(THF) Geometry Optimization"
    wB97X_D3_6_31g_d_ISOSVP_THF__Frequency_Analysis = "wB97X-D3/6-31g*/ISOSVP(THF) Frequency Analysis"
    wB97X_D3_6_31g_d_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_THF__Unknown = "wB97X-D3/6-31g*/ISOSVP(THF) Unknown"
    wB97X_D3_6_31g_d_ISOSVP_G2__Single_Point = "wB97X-D3/6-31g*/ISOSVP(G2) Single Point"
    wB97X_D3_6_31g_d_ISOSVP_G2__Force = "wB97X-D3/6-31g*/ISOSVP(G2) Force"
    wB97X_D3_6_31g_d_ISOSVP_G2__Geometry_Optimization = "wB97X-D3/6-31g*/ISOSVP(G2) Geometry Optimization"
    wB97X_D3_6_31g_d_ISOSVP_G2__Frequency_Analysis = "wB97X-D3/6-31g*/ISOSVP(G2) Frequency Analysis"
    wB97X_D3_6_31g_d_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_G2__Unknown = "wB97X-D3/6-31g*/ISOSVP(G2) Unknown"
    wB97X_D3_6_31g_d_ISOSVP_DME__Single_Point = "wB97X-D3/6-31g*/ISOSVP(DME) Single Point"
    wB97X_D3_6_31g_d_ISOSVP_DME__Force = "wB97X-D3/6-31g*/ISOSVP(DME) Force"
    wB97X_D3_6_31g_d_ISOSVP_DME__Geometry_Optimization = "wB97X-D3/6-31g*/ISOSVP(DME) Geometry Optimization"
    wB97X_D3_6_31g_d_ISOSVP_DME__Frequency_Analysis = "wB97X-D3/6-31g*/ISOSVP(DME) Frequency Analysis"
    wB97X_D3_6_31g_d_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_DME__Unknown = "wB97X-D3/6-31g*/ISOSVP(DME) Unknown"
    wB97X_D3_6_31g_d_ISOSVP_EC__Single_Point = "wB97X-D3/6-31g*/ISOSVP(EC) Single Point"
    wB97X_D3_6_31g_d_ISOSVP_EC__Force = "wB97X-D3/6-31g*/ISOSVP(EC) Force"
    wB97X_D3_6_31g_d_ISOSVP_EC__Geometry_Optimization = "wB97X-D3/6-31g*/ISOSVP(EC) Geometry Optimization"
    wB97X_D3_6_31g_d_ISOSVP_EC__Frequency_Analysis = "wB97X-D3/6-31g*/ISOSVP(EC) Frequency Analysis"
    wB97X_D3_6_31g_d_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_EC__Unknown = "wB97X-D3/6-31g*/ISOSVP(EC) Unknown"
    wB97X_D3_6_31g_d_ISOSVP_EC_EMC__Single_Point = "wB97X-D3/6-31g*/ISOSVP(EC/EMC) Single Point"
    wB97X_D3_6_31g_d_ISOSVP_EC_EMC__Force = "wB97X-D3/6-31g*/ISOSVP(EC/EMC) Force"
    wB97X_D3_6_31g_d_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-D3/6-31g*/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_D3_6_31g_d_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D3/6-31g*/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D3_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_EC_EMC__Unknown = "wB97X-D3/6-31g*/ISOSVP(EC/EMC) Unknown"
    wB97X_D3_6_31g_d_ISOSVP_PC__Single_Point = "wB97X-D3/6-31g*/ISOSVP(PC) Single Point"
    wB97X_D3_6_31g_d_ISOSVP_PC__Force = "wB97X-D3/6-31g*/ISOSVP(PC) Force"
    wB97X_D3_6_31g_d_ISOSVP_PC__Geometry_Optimization = "wB97X-D3/6-31g*/ISOSVP(PC) Geometry Optimization"
    wB97X_D3_6_31g_d_ISOSVP_PC__Frequency_Analysis = "wB97X-D3/6-31g*/ISOSVP(PC) Frequency Analysis"
    wB97X_D3_6_31g_d_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_ISOSVP_PC__Unknown = "wB97X-D3/6-31g*/ISOSVP(PC) Unknown"
    wB97X_D3_6_31g_d_SMD_WATER__Single_Point = "wB97X-D3/6-31g*/SMD(WATER) Single Point"
    wB97X_D3_6_31g_d_SMD_WATER__Force = "wB97X-D3/6-31g*/SMD(WATER) Force"
    wB97X_D3_6_31g_d_SMD_WATER__Geometry_Optimization = "wB97X-D3/6-31g*/SMD(WATER) Geometry Optimization"
    wB97X_D3_6_31g_d_SMD_WATER__Frequency_Analysis = "wB97X-D3/6-31g*/SMD(WATER) Frequency Analysis"
    wB97X_D3_6_31g_d_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_WATER__Unknown = "wB97X-D3/6-31g*/SMD(WATER) Unknown"
    wB97X_D3_6_31g_d_SMD_THF__Single_Point = "wB97X-D3/6-31g*/SMD(THF) Single Point"
    wB97X_D3_6_31g_d_SMD_THF__Force = "wB97X-D3/6-31g*/SMD(THF) Force"
    wB97X_D3_6_31g_d_SMD_THF__Geometry_Optimization = "wB97X-D3/6-31g*/SMD(THF) Geometry Optimization"
    wB97X_D3_6_31g_d_SMD_THF__Frequency_Analysis = "wB97X-D3/6-31g*/SMD(THF) Frequency Analysis"
    wB97X_D3_6_31g_d_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_THF__Unknown = "wB97X-D3/6-31g*/SMD(THF) Unknown"
    wB97X_D3_6_31g_d_SMD_G2__Single_Point = "wB97X-D3/6-31g*/SMD(G2) Single Point"
    wB97X_D3_6_31g_d_SMD_G2__Force = "wB97X-D3/6-31g*/SMD(G2) Force"
    wB97X_D3_6_31g_d_SMD_G2__Geometry_Optimization = "wB97X-D3/6-31g*/SMD(G2) Geometry Optimization"
    wB97X_D3_6_31g_d_SMD_G2__Frequency_Analysis = "wB97X-D3/6-31g*/SMD(G2) Frequency Analysis"
    wB97X_D3_6_31g_d_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_G2__Unknown = "wB97X-D3/6-31g*/SMD(G2) Unknown"
    wB97X_D3_6_31g_d_SMD_DME__Single_Point = "wB97X-D3/6-31g*/SMD(DME) Single Point"
    wB97X_D3_6_31g_d_SMD_DME__Force = "wB97X-D3/6-31g*/SMD(DME) Force"
    wB97X_D3_6_31g_d_SMD_DME__Geometry_Optimization = "wB97X-D3/6-31g*/SMD(DME) Geometry Optimization"
    wB97X_D3_6_31g_d_SMD_DME__Frequency_Analysis = "wB97X-D3/6-31g*/SMD(DME) Frequency Analysis"
    wB97X_D3_6_31g_d_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_DME__Unknown = "wB97X-D3/6-31g*/SMD(DME) Unknown"
    wB97X_D3_6_31g_d_SMD_EC__Single_Point = "wB97X-D3/6-31g*/SMD(EC) Single Point"
    wB97X_D3_6_31g_d_SMD_EC__Force = "wB97X-D3/6-31g*/SMD(EC) Force"
    wB97X_D3_6_31g_d_SMD_EC__Geometry_Optimization = "wB97X-D3/6-31g*/SMD(EC) Geometry Optimization"
    wB97X_D3_6_31g_d_SMD_EC__Frequency_Analysis = "wB97X-D3/6-31g*/SMD(EC) Frequency Analysis"
    wB97X_D3_6_31g_d_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_EC__Unknown = "wB97X-D3/6-31g*/SMD(EC) Unknown"
    wB97X_D3_6_31g_d_SMD_EC_EMC__Single_Point = "wB97X-D3/6-31g*/SMD(EC/EMC) Single Point"
    wB97X_D3_6_31g_d_SMD_EC_EMC__Force = "wB97X-D3/6-31g*/SMD(EC/EMC) Force"
    wB97X_D3_6_31g_d_SMD_EC_EMC__Geometry_Optimization = "wB97X-D3/6-31g*/SMD(EC/EMC) Geometry Optimization"
    wB97X_D3_6_31g_d_SMD_EC_EMC__Frequency_Analysis = "wB97X-D3/6-31g*/SMD(EC/EMC) Frequency Analysis"
    wB97X_D3_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_EC_EMC__Unknown = "wB97X-D3/6-31g*/SMD(EC/EMC) Unknown"
    wB97X_D3_6_31g_d_SMD_PC__Single_Point = "wB97X-D3/6-31g*/SMD(PC) Single Point"
    wB97X_D3_6_31g_d_SMD_PC__Force = "wB97X-D3/6-31g*/SMD(PC) Force"
    wB97X_D3_6_31g_d_SMD_PC__Geometry_Optimization = "wB97X-D3/6-31g*/SMD(PC) Geometry Optimization"
    wB97X_D3_6_31g_d_SMD_PC__Frequency_Analysis = "wB97X-D3/6-31g*/SMD(PC) Frequency Analysis"
    wB97X_D3_6_31g_d_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_SMD_PC__Unknown = "wB97X-D3/6-31g*/SMD(PC) Unknown"
    wB97X_D3_6_31g_d_CMIRS_WATER__Single_Point = "wB97X-D3/6-31g*/CMIRS(WATER) Single Point"
    wB97X_D3_6_31g_d_CMIRS_WATER__Force = "wB97X-D3/6-31g*/CMIRS(WATER) Force"
    wB97X_D3_6_31g_d_CMIRS_WATER__Geometry_Optimization = "wB97X-D3/6-31g*/CMIRS(WATER) Geometry Optimization"
    wB97X_D3_6_31g_d_CMIRS_WATER__Frequency_Analysis = "wB97X-D3/6-31g*/CMIRS(WATER) Frequency Analysis"
    wB97X_D3_6_31g_d_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_WATER__Unknown = "wB97X-D3/6-31g*/CMIRS(WATER) Unknown"
    wB97X_D3_6_31g_d_CMIRS_THF__Single_Point = "wB97X-D3/6-31g*/CMIRS(THF) Single Point"
    wB97X_D3_6_31g_d_CMIRS_THF__Force = "wB97X-D3/6-31g*/CMIRS(THF) Force"
    wB97X_D3_6_31g_d_CMIRS_THF__Geometry_Optimization = "wB97X-D3/6-31g*/CMIRS(THF) Geometry Optimization"
    wB97X_D3_6_31g_d_CMIRS_THF__Frequency_Analysis = "wB97X-D3/6-31g*/CMIRS(THF) Frequency Analysis"
    wB97X_D3_6_31g_d_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_THF__Unknown = "wB97X-D3/6-31g*/CMIRS(THF) Unknown"
    wB97X_D3_6_31g_d_CMIRS_G2__Single_Point = "wB97X-D3/6-31g*/CMIRS(G2) Single Point"
    wB97X_D3_6_31g_d_CMIRS_G2__Force = "wB97X-D3/6-31g*/CMIRS(G2) Force"
    wB97X_D3_6_31g_d_CMIRS_G2__Geometry_Optimization = "wB97X-D3/6-31g*/CMIRS(G2) Geometry Optimization"
    wB97X_D3_6_31g_d_CMIRS_G2__Frequency_Analysis = "wB97X-D3/6-31g*/CMIRS(G2) Frequency Analysis"
    wB97X_D3_6_31g_d_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_G2__Unknown = "wB97X-D3/6-31g*/CMIRS(G2) Unknown"
    wB97X_D3_6_31g_d_CMIRS_DME__Single_Point = "wB97X-D3/6-31g*/CMIRS(DME) Single Point"
    wB97X_D3_6_31g_d_CMIRS_DME__Force = "wB97X-D3/6-31g*/CMIRS(DME) Force"
    wB97X_D3_6_31g_d_CMIRS_DME__Geometry_Optimization = "wB97X-D3/6-31g*/CMIRS(DME) Geometry Optimization"
    wB97X_D3_6_31g_d_CMIRS_DME__Frequency_Analysis = "wB97X-D3/6-31g*/CMIRS(DME) Frequency Analysis"
    wB97X_D3_6_31g_d_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_DME__Unknown = "wB97X-D3/6-31g*/CMIRS(DME) Unknown"
    wB97X_D3_6_31g_d_CMIRS_EC__Single_Point = "wB97X-D3/6-31g*/CMIRS(EC) Single Point"
    wB97X_D3_6_31g_d_CMIRS_EC__Force = "wB97X-D3/6-31g*/CMIRS(EC) Force"
    wB97X_D3_6_31g_d_CMIRS_EC__Geometry_Optimization = "wB97X-D3/6-31g*/CMIRS(EC) Geometry Optimization"
    wB97X_D3_6_31g_d_CMIRS_EC__Frequency_Analysis = "wB97X-D3/6-31g*/CMIRS(EC) Frequency Analysis"
    wB97X_D3_6_31g_d_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_EC__Unknown = "wB97X-D3/6-31g*/CMIRS(EC) Unknown"
    wB97X_D3_6_31g_d_CMIRS_EC_EMC__Single_Point = "wB97X-D3/6-31g*/CMIRS(EC/EMC) Single Point"
    wB97X_D3_6_31g_d_CMIRS_EC_EMC__Force = "wB97X-D3/6-31g*/CMIRS(EC/EMC) Force"
    wB97X_D3_6_31g_d_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D3/6-31g*/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D3_6_31g_d_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D3/6-31g*/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D3_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_EC_EMC__Unknown = "wB97X-D3/6-31g*/CMIRS(EC/EMC) Unknown"
    wB97X_D3_6_31g_d_CMIRS_PC__Single_Point = "wB97X-D3/6-31g*/CMIRS(PC) Single Point"
    wB97X_D3_6_31g_d_CMIRS_PC__Force = "wB97X-D3/6-31g*/CMIRS(PC) Force"
    wB97X_D3_6_31g_d_CMIRS_PC__Geometry_Optimization = "wB97X-D3/6-31g*/CMIRS(PC) Geometry Optimization"
    wB97X_D3_6_31g_d_CMIRS_PC__Frequency_Analysis = "wB97X-D3/6-31g*/CMIRS(PC) Frequency Analysis"
    wB97X_D3_6_31g_d_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/6-31g*/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_6_31g_d_CMIRS_PC__Unknown = "wB97X-D3/6-31g*/CMIRS(PC) Unknown"
    wB97X_D3_def2_SVPD_VACUUM_Single_Point = "wB97X-D3/def2-SVPD/VACUUM Single Point"
    wB97X_D3_def2_SVPD_VACUUM_Force = "wB97X-D3/def2-SVPD/VACUUM Force"
    wB97X_D3_def2_SVPD_VACUUM_Geometry_Optimization = "wB97X-D3/def2-SVPD/VACUUM Geometry Optimization"
    wB97X_D3_def2_SVPD_VACUUM_Frequency_Analysis = "wB97X-D3/def2-SVPD/VACUUM Frequency Analysis"
    wB97X_D3_def2_SVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_VACUUM_Unknown = "wB97X-D3/def2-SVPD/VACUUM Unknown"
    wB97X_D3_def2_SVPD_PCM_WATER__Single_Point = "wB97X-D3/def2-SVPD/PCM(WATER) Single Point"
    wB97X_D3_def2_SVPD_PCM_WATER__Force = "wB97X-D3/def2-SVPD/PCM(WATER) Force"
    wB97X_D3_def2_SVPD_PCM_WATER__Geometry_Optimization = "wB97X-D3/def2-SVPD/PCM(WATER) Geometry Optimization"
    wB97X_D3_def2_SVPD_PCM_WATER__Frequency_Analysis = "wB97X-D3/def2-SVPD/PCM(WATER) Frequency Analysis"
    wB97X_D3_def2_SVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_WATER__Unknown = "wB97X-D3/def2-SVPD/PCM(WATER) Unknown"
    wB97X_D3_def2_SVPD_PCM_THF__Single_Point = "wB97X-D3/def2-SVPD/PCM(THF) Single Point"
    wB97X_D3_def2_SVPD_PCM_THF__Force = "wB97X-D3/def2-SVPD/PCM(THF) Force"
    wB97X_D3_def2_SVPD_PCM_THF__Geometry_Optimization = "wB97X-D3/def2-SVPD/PCM(THF) Geometry Optimization"
    wB97X_D3_def2_SVPD_PCM_THF__Frequency_Analysis = "wB97X-D3/def2-SVPD/PCM(THF) Frequency Analysis"
    wB97X_D3_def2_SVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_THF__Unknown = "wB97X-D3/def2-SVPD/PCM(THF) Unknown"
    wB97X_D3_def2_SVPD_PCM_G2__Single_Point = "wB97X-D3/def2-SVPD/PCM(G2) Single Point"
    wB97X_D3_def2_SVPD_PCM_G2__Force = "wB97X-D3/def2-SVPD/PCM(G2) Force"
    wB97X_D3_def2_SVPD_PCM_G2__Geometry_Optimization = "wB97X-D3/def2-SVPD/PCM(G2) Geometry Optimization"
    wB97X_D3_def2_SVPD_PCM_G2__Frequency_Analysis = "wB97X-D3/def2-SVPD/PCM(G2) Frequency Analysis"
    wB97X_D3_def2_SVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_G2__Unknown = "wB97X-D3/def2-SVPD/PCM(G2) Unknown"
    wB97X_D3_def2_SVPD_PCM_DME__Single_Point = "wB97X-D3/def2-SVPD/PCM(DME) Single Point"
    wB97X_D3_def2_SVPD_PCM_DME__Force = "wB97X-D3/def2-SVPD/PCM(DME) Force"
    wB97X_D3_def2_SVPD_PCM_DME__Geometry_Optimization = "wB97X-D3/def2-SVPD/PCM(DME) Geometry Optimization"
    wB97X_D3_def2_SVPD_PCM_DME__Frequency_Analysis = "wB97X-D3/def2-SVPD/PCM(DME) Frequency Analysis"
    wB97X_D3_def2_SVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_DME__Unknown = "wB97X-D3/def2-SVPD/PCM(DME) Unknown"
    wB97X_D3_def2_SVPD_PCM_EC__Single_Point = "wB97X-D3/def2-SVPD/PCM(EC) Single Point"
    wB97X_D3_def2_SVPD_PCM_EC__Force = "wB97X-D3/def2-SVPD/PCM(EC) Force"
    wB97X_D3_def2_SVPD_PCM_EC__Geometry_Optimization = "wB97X-D3/def2-SVPD/PCM(EC) Geometry Optimization"
    wB97X_D3_def2_SVPD_PCM_EC__Frequency_Analysis = "wB97X-D3/def2-SVPD/PCM(EC) Frequency Analysis"
    wB97X_D3_def2_SVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_EC__Unknown = "wB97X-D3/def2-SVPD/PCM(EC) Unknown"
    wB97X_D3_def2_SVPD_PCM_EC_EMC__Single_Point = "wB97X-D3/def2-SVPD/PCM(EC/EMC) Single Point"
    wB97X_D3_def2_SVPD_PCM_EC_EMC__Force = "wB97X-D3/def2-SVPD/PCM(EC/EMC) Force"
    wB97X_D3_def2_SVPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-SVPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_SVPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-SVPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_EC_EMC__Unknown = "wB97X-D3/def2-SVPD/PCM(EC/EMC) Unknown"
    wB97X_D3_def2_SVPD_PCM_PC__Single_Point = "wB97X-D3/def2-SVPD/PCM(PC) Single Point"
    wB97X_D3_def2_SVPD_PCM_PC__Force = "wB97X-D3/def2-SVPD/PCM(PC) Force"
    wB97X_D3_def2_SVPD_PCM_PC__Geometry_Optimization = "wB97X-D3/def2-SVPD/PCM(PC) Geometry Optimization"
    wB97X_D3_def2_SVPD_PCM_PC__Frequency_Analysis = "wB97X-D3/def2-SVPD/PCM(PC) Frequency Analysis"
    wB97X_D3_def2_SVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_PCM_PC__Unknown = "wB97X-D3/def2-SVPD/PCM(PC) Unknown"
    wB97X_D3_def2_SVPD_ISOSVP_WATER__Single_Point = "wB97X-D3/def2-SVPD/ISOSVP(WATER) Single Point"
    wB97X_D3_def2_SVPD_ISOSVP_WATER__Force = "wB97X-D3/def2-SVPD/ISOSVP(WATER) Force"
    wB97X_D3_def2_SVPD_ISOSVP_WATER__Geometry_Optimization = "wB97X-D3/def2-SVPD/ISOSVP(WATER) Geometry Optimization"
    wB97X_D3_def2_SVPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-D3/def2-SVPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_D3_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_WATER__Unknown = "wB97X-D3/def2-SVPD/ISOSVP(WATER) Unknown"
    wB97X_D3_def2_SVPD_ISOSVP_THF__Single_Point = "wB97X-D3/def2-SVPD/ISOSVP(THF) Single Point"
    wB97X_D3_def2_SVPD_ISOSVP_THF__Force = "wB97X-D3/def2-SVPD/ISOSVP(THF) Force"
    wB97X_D3_def2_SVPD_ISOSVP_THF__Geometry_Optimization = "wB97X-D3/def2-SVPD/ISOSVP(THF) Geometry Optimization"
    wB97X_D3_def2_SVPD_ISOSVP_THF__Frequency_Analysis = "wB97X-D3/def2-SVPD/ISOSVP(THF) Frequency Analysis"
    wB97X_D3_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_THF__Unknown = "wB97X-D3/def2-SVPD/ISOSVP(THF) Unknown"
    wB97X_D3_def2_SVPD_ISOSVP_G2__Single_Point = "wB97X-D3/def2-SVPD/ISOSVP(G2) Single Point"
    wB97X_D3_def2_SVPD_ISOSVP_G2__Force = "wB97X-D3/def2-SVPD/ISOSVP(G2) Force"
    wB97X_D3_def2_SVPD_ISOSVP_G2__Geometry_Optimization = "wB97X-D3/def2-SVPD/ISOSVP(G2) Geometry Optimization"
    wB97X_D3_def2_SVPD_ISOSVP_G2__Frequency_Analysis = "wB97X-D3/def2-SVPD/ISOSVP(G2) Frequency Analysis"
    wB97X_D3_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_G2__Unknown = "wB97X-D3/def2-SVPD/ISOSVP(G2) Unknown"
    wB97X_D3_def2_SVPD_ISOSVP_DME__Single_Point = "wB97X-D3/def2-SVPD/ISOSVP(DME) Single Point"
    wB97X_D3_def2_SVPD_ISOSVP_DME__Force = "wB97X-D3/def2-SVPD/ISOSVP(DME) Force"
    wB97X_D3_def2_SVPD_ISOSVP_DME__Geometry_Optimization = "wB97X-D3/def2-SVPD/ISOSVP(DME) Geometry Optimization"
    wB97X_D3_def2_SVPD_ISOSVP_DME__Frequency_Analysis = "wB97X-D3/def2-SVPD/ISOSVP(DME) Frequency Analysis"
    wB97X_D3_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_DME__Unknown = "wB97X-D3/def2-SVPD/ISOSVP(DME) Unknown"
    wB97X_D3_def2_SVPD_ISOSVP_EC__Single_Point = "wB97X-D3/def2-SVPD/ISOSVP(EC) Single Point"
    wB97X_D3_def2_SVPD_ISOSVP_EC__Force = "wB97X-D3/def2-SVPD/ISOSVP(EC) Force"
    wB97X_D3_def2_SVPD_ISOSVP_EC__Geometry_Optimization = "wB97X-D3/def2-SVPD/ISOSVP(EC) Geometry Optimization"
    wB97X_D3_def2_SVPD_ISOSVP_EC__Frequency_Analysis = "wB97X-D3/def2-SVPD/ISOSVP(EC) Frequency Analysis"
    wB97X_D3_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_EC__Unknown = "wB97X-D3/def2-SVPD/ISOSVP(EC) Unknown"
    wB97X_D3_def2_SVPD_ISOSVP_EC_EMC__Single_Point = "wB97X-D3/def2-SVPD/ISOSVP(EC/EMC) Single Point"
    wB97X_D3_def2_SVPD_ISOSVP_EC_EMC__Force = "wB97X-D3/def2-SVPD/ISOSVP(EC/EMC) Force"
    wB97X_D3_def2_SVPD_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-SVPD/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_SVPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-SVPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_EC_EMC__Unknown = "wB97X-D3/def2-SVPD/ISOSVP(EC/EMC) Unknown"
    wB97X_D3_def2_SVPD_ISOSVP_PC__Single_Point = "wB97X-D3/def2-SVPD/ISOSVP(PC) Single Point"
    wB97X_D3_def2_SVPD_ISOSVP_PC__Force = "wB97X-D3/def2-SVPD/ISOSVP(PC) Force"
    wB97X_D3_def2_SVPD_ISOSVP_PC__Geometry_Optimization = "wB97X-D3/def2-SVPD/ISOSVP(PC) Geometry Optimization"
    wB97X_D3_def2_SVPD_ISOSVP_PC__Frequency_Analysis = "wB97X-D3/def2-SVPD/ISOSVP(PC) Frequency Analysis"
    wB97X_D3_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_ISOSVP_PC__Unknown = "wB97X-D3/def2-SVPD/ISOSVP(PC) Unknown"
    wB97X_D3_def2_SVPD_SMD_WATER__Single_Point = "wB97X-D3/def2-SVPD/SMD(WATER) Single Point"
    wB97X_D3_def2_SVPD_SMD_WATER__Force = "wB97X-D3/def2-SVPD/SMD(WATER) Force"
    wB97X_D3_def2_SVPD_SMD_WATER__Geometry_Optimization = "wB97X-D3/def2-SVPD/SMD(WATER) Geometry Optimization"
    wB97X_D3_def2_SVPD_SMD_WATER__Frequency_Analysis = "wB97X-D3/def2-SVPD/SMD(WATER) Frequency Analysis"
    wB97X_D3_def2_SVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_WATER__Unknown = "wB97X-D3/def2-SVPD/SMD(WATER) Unknown"
    wB97X_D3_def2_SVPD_SMD_THF__Single_Point = "wB97X-D3/def2-SVPD/SMD(THF) Single Point"
    wB97X_D3_def2_SVPD_SMD_THF__Force = "wB97X-D3/def2-SVPD/SMD(THF) Force"
    wB97X_D3_def2_SVPD_SMD_THF__Geometry_Optimization = "wB97X-D3/def2-SVPD/SMD(THF) Geometry Optimization"
    wB97X_D3_def2_SVPD_SMD_THF__Frequency_Analysis = "wB97X-D3/def2-SVPD/SMD(THF) Frequency Analysis"
    wB97X_D3_def2_SVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_THF__Unknown = "wB97X-D3/def2-SVPD/SMD(THF) Unknown"
    wB97X_D3_def2_SVPD_SMD_G2__Single_Point = "wB97X-D3/def2-SVPD/SMD(G2) Single Point"
    wB97X_D3_def2_SVPD_SMD_G2__Force = "wB97X-D3/def2-SVPD/SMD(G2) Force"
    wB97X_D3_def2_SVPD_SMD_G2__Geometry_Optimization = "wB97X-D3/def2-SVPD/SMD(G2) Geometry Optimization"
    wB97X_D3_def2_SVPD_SMD_G2__Frequency_Analysis = "wB97X-D3/def2-SVPD/SMD(G2) Frequency Analysis"
    wB97X_D3_def2_SVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_G2__Unknown = "wB97X-D3/def2-SVPD/SMD(G2) Unknown"
    wB97X_D3_def2_SVPD_SMD_DME__Single_Point = "wB97X-D3/def2-SVPD/SMD(DME) Single Point"
    wB97X_D3_def2_SVPD_SMD_DME__Force = "wB97X-D3/def2-SVPD/SMD(DME) Force"
    wB97X_D3_def2_SVPD_SMD_DME__Geometry_Optimization = "wB97X-D3/def2-SVPD/SMD(DME) Geometry Optimization"
    wB97X_D3_def2_SVPD_SMD_DME__Frequency_Analysis = "wB97X-D3/def2-SVPD/SMD(DME) Frequency Analysis"
    wB97X_D3_def2_SVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_DME__Unknown = "wB97X-D3/def2-SVPD/SMD(DME) Unknown"
    wB97X_D3_def2_SVPD_SMD_EC__Single_Point = "wB97X-D3/def2-SVPD/SMD(EC) Single Point"
    wB97X_D3_def2_SVPD_SMD_EC__Force = "wB97X-D3/def2-SVPD/SMD(EC) Force"
    wB97X_D3_def2_SVPD_SMD_EC__Geometry_Optimization = "wB97X-D3/def2-SVPD/SMD(EC) Geometry Optimization"
    wB97X_D3_def2_SVPD_SMD_EC__Frequency_Analysis = "wB97X-D3/def2-SVPD/SMD(EC) Frequency Analysis"
    wB97X_D3_def2_SVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_EC__Unknown = "wB97X-D3/def2-SVPD/SMD(EC) Unknown"
    wB97X_D3_def2_SVPD_SMD_EC_EMC__Single_Point = "wB97X-D3/def2-SVPD/SMD(EC/EMC) Single Point"
    wB97X_D3_def2_SVPD_SMD_EC_EMC__Force = "wB97X-D3/def2-SVPD/SMD(EC/EMC) Force"
    wB97X_D3_def2_SVPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-SVPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_SVPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-SVPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_EC_EMC__Unknown = "wB97X-D3/def2-SVPD/SMD(EC/EMC) Unknown"
    wB97X_D3_def2_SVPD_SMD_PC__Single_Point = "wB97X-D3/def2-SVPD/SMD(PC) Single Point"
    wB97X_D3_def2_SVPD_SMD_PC__Force = "wB97X-D3/def2-SVPD/SMD(PC) Force"
    wB97X_D3_def2_SVPD_SMD_PC__Geometry_Optimization = "wB97X-D3/def2-SVPD/SMD(PC) Geometry Optimization"
    wB97X_D3_def2_SVPD_SMD_PC__Frequency_Analysis = "wB97X-D3/def2-SVPD/SMD(PC) Frequency Analysis"
    wB97X_D3_def2_SVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_SMD_PC__Unknown = "wB97X-D3/def2-SVPD/SMD(PC) Unknown"
    wB97X_D3_def2_SVPD_CMIRS_WATER__Single_Point = "wB97X-D3/def2-SVPD/CMIRS(WATER) Single Point"
    wB97X_D3_def2_SVPD_CMIRS_WATER__Force = "wB97X-D3/def2-SVPD/CMIRS(WATER) Force"
    wB97X_D3_def2_SVPD_CMIRS_WATER__Geometry_Optimization = "wB97X-D3/def2-SVPD/CMIRS(WATER) Geometry Optimization"
    wB97X_D3_def2_SVPD_CMIRS_WATER__Frequency_Analysis = "wB97X-D3/def2-SVPD/CMIRS(WATER) Frequency Analysis"
    wB97X_D3_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_WATER__Unknown = "wB97X-D3/def2-SVPD/CMIRS(WATER) Unknown"
    wB97X_D3_def2_SVPD_CMIRS_THF__Single_Point = "wB97X-D3/def2-SVPD/CMIRS(THF) Single Point"
    wB97X_D3_def2_SVPD_CMIRS_THF__Force = "wB97X-D3/def2-SVPD/CMIRS(THF) Force"
    wB97X_D3_def2_SVPD_CMIRS_THF__Geometry_Optimization = "wB97X-D3/def2-SVPD/CMIRS(THF) Geometry Optimization"
    wB97X_D3_def2_SVPD_CMIRS_THF__Frequency_Analysis = "wB97X-D3/def2-SVPD/CMIRS(THF) Frequency Analysis"
    wB97X_D3_def2_SVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_THF__Unknown = "wB97X-D3/def2-SVPD/CMIRS(THF) Unknown"
    wB97X_D3_def2_SVPD_CMIRS_G2__Single_Point = "wB97X-D3/def2-SVPD/CMIRS(G2) Single Point"
    wB97X_D3_def2_SVPD_CMIRS_G2__Force = "wB97X-D3/def2-SVPD/CMIRS(G2) Force"
    wB97X_D3_def2_SVPD_CMIRS_G2__Geometry_Optimization = "wB97X-D3/def2-SVPD/CMIRS(G2) Geometry Optimization"
    wB97X_D3_def2_SVPD_CMIRS_G2__Frequency_Analysis = "wB97X-D3/def2-SVPD/CMIRS(G2) Frequency Analysis"
    wB97X_D3_def2_SVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_G2__Unknown = "wB97X-D3/def2-SVPD/CMIRS(G2) Unknown"
    wB97X_D3_def2_SVPD_CMIRS_DME__Single_Point = "wB97X-D3/def2-SVPD/CMIRS(DME) Single Point"
    wB97X_D3_def2_SVPD_CMIRS_DME__Force = "wB97X-D3/def2-SVPD/CMIRS(DME) Force"
    wB97X_D3_def2_SVPD_CMIRS_DME__Geometry_Optimization = "wB97X-D3/def2-SVPD/CMIRS(DME) Geometry Optimization"
    wB97X_D3_def2_SVPD_CMIRS_DME__Frequency_Analysis = "wB97X-D3/def2-SVPD/CMIRS(DME) Frequency Analysis"
    wB97X_D3_def2_SVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_DME__Unknown = "wB97X-D3/def2-SVPD/CMIRS(DME) Unknown"
    wB97X_D3_def2_SVPD_CMIRS_EC__Single_Point = "wB97X-D3/def2-SVPD/CMIRS(EC) Single Point"
    wB97X_D3_def2_SVPD_CMIRS_EC__Force = "wB97X-D3/def2-SVPD/CMIRS(EC) Force"
    wB97X_D3_def2_SVPD_CMIRS_EC__Geometry_Optimization = "wB97X-D3/def2-SVPD/CMIRS(EC) Geometry Optimization"
    wB97X_D3_def2_SVPD_CMIRS_EC__Frequency_Analysis = "wB97X-D3/def2-SVPD/CMIRS(EC) Frequency Analysis"
    wB97X_D3_def2_SVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_EC__Unknown = "wB97X-D3/def2-SVPD/CMIRS(EC) Unknown"
    wB97X_D3_def2_SVPD_CMIRS_EC_EMC__Single_Point = "wB97X-D3/def2-SVPD/CMIRS(EC/EMC) Single Point"
    wB97X_D3_def2_SVPD_CMIRS_EC_EMC__Force = "wB97X-D3/def2-SVPD/CMIRS(EC/EMC) Force"
    wB97X_D3_def2_SVPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-SVPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_SVPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-SVPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_EC_EMC__Unknown = "wB97X-D3/def2-SVPD/CMIRS(EC/EMC) Unknown"
    wB97X_D3_def2_SVPD_CMIRS_PC__Single_Point = "wB97X-D3/def2-SVPD/CMIRS(PC) Single Point"
    wB97X_D3_def2_SVPD_CMIRS_PC__Force = "wB97X-D3/def2-SVPD/CMIRS(PC) Force"
    wB97X_D3_def2_SVPD_CMIRS_PC__Geometry_Optimization = "wB97X-D3/def2-SVPD/CMIRS(PC) Geometry Optimization"
    wB97X_D3_def2_SVPD_CMIRS_PC__Frequency_Analysis = "wB97X-D3/def2-SVPD/CMIRS(PC) Frequency Analysis"
    wB97X_D3_def2_SVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-SVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_SVPD_CMIRS_PC__Unknown = "wB97X-D3/def2-SVPD/CMIRS(PC) Unknown"
    wB97X_D3_def2_TZVP_VACUUM_Single_Point = "wB97X-D3/def2-TZVP/VACUUM Single Point"
    wB97X_D3_def2_TZVP_VACUUM_Force = "wB97X-D3/def2-TZVP/VACUUM Force"
    wB97X_D3_def2_TZVP_VACUUM_Geometry_Optimization = "wB97X-D3/def2-TZVP/VACUUM Geometry Optimization"
    wB97X_D3_def2_TZVP_VACUUM_Frequency_Analysis = "wB97X-D3/def2-TZVP/VACUUM Frequency Analysis"
    wB97X_D3_def2_TZVP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_VACUUM_Unknown = "wB97X-D3/def2-TZVP/VACUUM Unknown"
    wB97X_D3_def2_TZVP_PCM_WATER__Single_Point = "wB97X-D3/def2-TZVP/PCM(WATER) Single Point"
    wB97X_D3_def2_TZVP_PCM_WATER__Force = "wB97X-D3/def2-TZVP/PCM(WATER) Force"
    wB97X_D3_def2_TZVP_PCM_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVP/PCM(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVP_PCM_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVP/PCM(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_WATER__Unknown = "wB97X-D3/def2-TZVP/PCM(WATER) Unknown"
    wB97X_D3_def2_TZVP_PCM_THF__Single_Point = "wB97X-D3/def2-TZVP/PCM(THF) Single Point"
    wB97X_D3_def2_TZVP_PCM_THF__Force = "wB97X-D3/def2-TZVP/PCM(THF) Force"
    wB97X_D3_def2_TZVP_PCM_THF__Geometry_Optimization = "wB97X-D3/def2-TZVP/PCM(THF) Geometry Optimization"
    wB97X_D3_def2_TZVP_PCM_THF__Frequency_Analysis = "wB97X-D3/def2-TZVP/PCM(THF) Frequency Analysis"
    wB97X_D3_def2_TZVP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_THF__Unknown = "wB97X-D3/def2-TZVP/PCM(THF) Unknown"
    wB97X_D3_def2_TZVP_PCM_G2__Single_Point = "wB97X-D3/def2-TZVP/PCM(G2) Single Point"
    wB97X_D3_def2_TZVP_PCM_G2__Force = "wB97X-D3/def2-TZVP/PCM(G2) Force"
    wB97X_D3_def2_TZVP_PCM_G2__Geometry_Optimization = "wB97X-D3/def2-TZVP/PCM(G2) Geometry Optimization"
    wB97X_D3_def2_TZVP_PCM_G2__Frequency_Analysis = "wB97X-D3/def2-TZVP/PCM(G2) Frequency Analysis"
    wB97X_D3_def2_TZVP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_G2__Unknown = "wB97X-D3/def2-TZVP/PCM(G2) Unknown"
    wB97X_D3_def2_TZVP_PCM_DME__Single_Point = "wB97X-D3/def2-TZVP/PCM(DME) Single Point"
    wB97X_D3_def2_TZVP_PCM_DME__Force = "wB97X-D3/def2-TZVP/PCM(DME) Force"
    wB97X_D3_def2_TZVP_PCM_DME__Geometry_Optimization = "wB97X-D3/def2-TZVP/PCM(DME) Geometry Optimization"
    wB97X_D3_def2_TZVP_PCM_DME__Frequency_Analysis = "wB97X-D3/def2-TZVP/PCM(DME) Frequency Analysis"
    wB97X_D3_def2_TZVP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_DME__Unknown = "wB97X-D3/def2-TZVP/PCM(DME) Unknown"
    wB97X_D3_def2_TZVP_PCM_EC__Single_Point = "wB97X-D3/def2-TZVP/PCM(EC) Single Point"
    wB97X_D3_def2_TZVP_PCM_EC__Force = "wB97X-D3/def2-TZVP/PCM(EC) Force"
    wB97X_D3_def2_TZVP_PCM_EC__Geometry_Optimization = "wB97X-D3/def2-TZVP/PCM(EC) Geometry Optimization"
    wB97X_D3_def2_TZVP_PCM_EC__Frequency_Analysis = "wB97X-D3/def2-TZVP/PCM(EC) Frequency Analysis"
    wB97X_D3_def2_TZVP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_EC__Unknown = "wB97X-D3/def2-TZVP/PCM(EC) Unknown"
    wB97X_D3_def2_TZVP_PCM_EC_EMC__Single_Point = "wB97X-D3/def2-TZVP/PCM(EC/EMC) Single Point"
    wB97X_D3_def2_TZVP_PCM_EC_EMC__Force = "wB97X-D3/def2-TZVP/PCM(EC/EMC) Force"
    wB97X_D3_def2_TZVP_PCM_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVP/PCM(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVP_PCM_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVP/PCM(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_EC_EMC__Unknown = "wB97X-D3/def2-TZVP/PCM(EC/EMC) Unknown"
    wB97X_D3_def2_TZVP_PCM_PC__Single_Point = "wB97X-D3/def2-TZVP/PCM(PC) Single Point"
    wB97X_D3_def2_TZVP_PCM_PC__Force = "wB97X-D3/def2-TZVP/PCM(PC) Force"
    wB97X_D3_def2_TZVP_PCM_PC__Geometry_Optimization = "wB97X-D3/def2-TZVP/PCM(PC) Geometry Optimization"
    wB97X_D3_def2_TZVP_PCM_PC__Frequency_Analysis = "wB97X-D3/def2-TZVP/PCM(PC) Frequency Analysis"
    wB97X_D3_def2_TZVP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_PCM_PC__Unknown = "wB97X-D3/def2-TZVP/PCM(PC) Unknown"
    wB97X_D3_def2_TZVP_ISOSVP_WATER__Single_Point = "wB97X-D3/def2-TZVP/ISOSVP(WATER) Single Point"
    wB97X_D3_def2_TZVP_ISOSVP_WATER__Force = "wB97X-D3/def2-TZVP/ISOSVP(WATER) Force"
    wB97X_D3_def2_TZVP_ISOSVP_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVP/ISOSVP(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVP_ISOSVP_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVP/ISOSVP(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_WATER__Unknown = "wB97X-D3/def2-TZVP/ISOSVP(WATER) Unknown"
    wB97X_D3_def2_TZVP_ISOSVP_THF__Single_Point = "wB97X-D3/def2-TZVP/ISOSVP(THF) Single Point"
    wB97X_D3_def2_TZVP_ISOSVP_THF__Force = "wB97X-D3/def2-TZVP/ISOSVP(THF) Force"
    wB97X_D3_def2_TZVP_ISOSVP_THF__Geometry_Optimization = "wB97X-D3/def2-TZVP/ISOSVP(THF) Geometry Optimization"
    wB97X_D3_def2_TZVP_ISOSVP_THF__Frequency_Analysis = "wB97X-D3/def2-TZVP/ISOSVP(THF) Frequency Analysis"
    wB97X_D3_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_THF__Unknown = "wB97X-D3/def2-TZVP/ISOSVP(THF) Unknown"
    wB97X_D3_def2_TZVP_ISOSVP_G2__Single_Point = "wB97X-D3/def2-TZVP/ISOSVP(G2) Single Point"
    wB97X_D3_def2_TZVP_ISOSVP_G2__Force = "wB97X-D3/def2-TZVP/ISOSVP(G2) Force"
    wB97X_D3_def2_TZVP_ISOSVP_G2__Geometry_Optimization = "wB97X-D3/def2-TZVP/ISOSVP(G2) Geometry Optimization"
    wB97X_D3_def2_TZVP_ISOSVP_G2__Frequency_Analysis = "wB97X-D3/def2-TZVP/ISOSVP(G2) Frequency Analysis"
    wB97X_D3_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_G2__Unknown = "wB97X-D3/def2-TZVP/ISOSVP(G2) Unknown"
    wB97X_D3_def2_TZVP_ISOSVP_DME__Single_Point = "wB97X-D3/def2-TZVP/ISOSVP(DME) Single Point"
    wB97X_D3_def2_TZVP_ISOSVP_DME__Force = "wB97X-D3/def2-TZVP/ISOSVP(DME) Force"
    wB97X_D3_def2_TZVP_ISOSVP_DME__Geometry_Optimization = "wB97X-D3/def2-TZVP/ISOSVP(DME) Geometry Optimization"
    wB97X_D3_def2_TZVP_ISOSVP_DME__Frequency_Analysis = "wB97X-D3/def2-TZVP/ISOSVP(DME) Frequency Analysis"
    wB97X_D3_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_DME__Unknown = "wB97X-D3/def2-TZVP/ISOSVP(DME) Unknown"
    wB97X_D3_def2_TZVP_ISOSVP_EC__Single_Point = "wB97X-D3/def2-TZVP/ISOSVP(EC) Single Point"
    wB97X_D3_def2_TZVP_ISOSVP_EC__Force = "wB97X-D3/def2-TZVP/ISOSVP(EC) Force"
    wB97X_D3_def2_TZVP_ISOSVP_EC__Geometry_Optimization = "wB97X-D3/def2-TZVP/ISOSVP(EC) Geometry Optimization"
    wB97X_D3_def2_TZVP_ISOSVP_EC__Frequency_Analysis = "wB97X-D3/def2-TZVP/ISOSVP(EC) Frequency Analysis"
    wB97X_D3_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_EC__Unknown = "wB97X-D3/def2-TZVP/ISOSVP(EC) Unknown"
    wB97X_D3_def2_TZVP_ISOSVP_EC_EMC__Single_Point = "wB97X-D3/def2-TZVP/ISOSVP(EC/EMC) Single Point"
    wB97X_D3_def2_TZVP_ISOSVP_EC_EMC__Force = "wB97X-D3/def2-TZVP/ISOSVP(EC/EMC) Force"
    wB97X_D3_def2_TZVP_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVP/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVP_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVP/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_EC_EMC__Unknown = "wB97X-D3/def2-TZVP/ISOSVP(EC/EMC) Unknown"
    wB97X_D3_def2_TZVP_ISOSVP_PC__Single_Point = "wB97X-D3/def2-TZVP/ISOSVP(PC) Single Point"
    wB97X_D3_def2_TZVP_ISOSVP_PC__Force = "wB97X-D3/def2-TZVP/ISOSVP(PC) Force"
    wB97X_D3_def2_TZVP_ISOSVP_PC__Geometry_Optimization = "wB97X-D3/def2-TZVP/ISOSVP(PC) Geometry Optimization"
    wB97X_D3_def2_TZVP_ISOSVP_PC__Frequency_Analysis = "wB97X-D3/def2-TZVP/ISOSVP(PC) Frequency Analysis"
    wB97X_D3_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_ISOSVP_PC__Unknown = "wB97X-D3/def2-TZVP/ISOSVP(PC) Unknown"
    wB97X_D3_def2_TZVP_SMD_WATER__Single_Point = "wB97X-D3/def2-TZVP/SMD(WATER) Single Point"
    wB97X_D3_def2_TZVP_SMD_WATER__Force = "wB97X-D3/def2-TZVP/SMD(WATER) Force"
    wB97X_D3_def2_TZVP_SMD_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVP/SMD(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVP_SMD_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVP/SMD(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_WATER__Unknown = "wB97X-D3/def2-TZVP/SMD(WATER) Unknown"
    wB97X_D3_def2_TZVP_SMD_THF__Single_Point = "wB97X-D3/def2-TZVP/SMD(THF) Single Point"
    wB97X_D3_def2_TZVP_SMD_THF__Force = "wB97X-D3/def2-TZVP/SMD(THF) Force"
    wB97X_D3_def2_TZVP_SMD_THF__Geometry_Optimization = "wB97X-D3/def2-TZVP/SMD(THF) Geometry Optimization"
    wB97X_D3_def2_TZVP_SMD_THF__Frequency_Analysis = "wB97X-D3/def2-TZVP/SMD(THF) Frequency Analysis"
    wB97X_D3_def2_TZVP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_THF__Unknown = "wB97X-D3/def2-TZVP/SMD(THF) Unknown"
    wB97X_D3_def2_TZVP_SMD_G2__Single_Point = "wB97X-D3/def2-TZVP/SMD(G2) Single Point"
    wB97X_D3_def2_TZVP_SMD_G2__Force = "wB97X-D3/def2-TZVP/SMD(G2) Force"
    wB97X_D3_def2_TZVP_SMD_G2__Geometry_Optimization = "wB97X-D3/def2-TZVP/SMD(G2) Geometry Optimization"
    wB97X_D3_def2_TZVP_SMD_G2__Frequency_Analysis = "wB97X-D3/def2-TZVP/SMD(G2) Frequency Analysis"
    wB97X_D3_def2_TZVP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_G2__Unknown = "wB97X-D3/def2-TZVP/SMD(G2) Unknown"
    wB97X_D3_def2_TZVP_SMD_DME__Single_Point = "wB97X-D3/def2-TZVP/SMD(DME) Single Point"
    wB97X_D3_def2_TZVP_SMD_DME__Force = "wB97X-D3/def2-TZVP/SMD(DME) Force"
    wB97X_D3_def2_TZVP_SMD_DME__Geometry_Optimization = "wB97X-D3/def2-TZVP/SMD(DME) Geometry Optimization"
    wB97X_D3_def2_TZVP_SMD_DME__Frequency_Analysis = "wB97X-D3/def2-TZVP/SMD(DME) Frequency Analysis"
    wB97X_D3_def2_TZVP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_DME__Unknown = "wB97X-D3/def2-TZVP/SMD(DME) Unknown"
    wB97X_D3_def2_TZVP_SMD_EC__Single_Point = "wB97X-D3/def2-TZVP/SMD(EC) Single Point"
    wB97X_D3_def2_TZVP_SMD_EC__Force = "wB97X-D3/def2-TZVP/SMD(EC) Force"
    wB97X_D3_def2_TZVP_SMD_EC__Geometry_Optimization = "wB97X-D3/def2-TZVP/SMD(EC) Geometry Optimization"
    wB97X_D3_def2_TZVP_SMD_EC__Frequency_Analysis = "wB97X-D3/def2-TZVP/SMD(EC) Frequency Analysis"
    wB97X_D3_def2_TZVP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_EC__Unknown = "wB97X-D3/def2-TZVP/SMD(EC) Unknown"
    wB97X_D3_def2_TZVP_SMD_EC_EMC__Single_Point = "wB97X-D3/def2-TZVP/SMD(EC/EMC) Single Point"
    wB97X_D3_def2_TZVP_SMD_EC_EMC__Force = "wB97X-D3/def2-TZVP/SMD(EC/EMC) Force"
    wB97X_D3_def2_TZVP_SMD_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVP/SMD(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVP_SMD_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVP/SMD(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_EC_EMC__Unknown = "wB97X-D3/def2-TZVP/SMD(EC/EMC) Unknown"
    wB97X_D3_def2_TZVP_SMD_PC__Single_Point = "wB97X-D3/def2-TZVP/SMD(PC) Single Point"
    wB97X_D3_def2_TZVP_SMD_PC__Force = "wB97X-D3/def2-TZVP/SMD(PC) Force"
    wB97X_D3_def2_TZVP_SMD_PC__Geometry_Optimization = "wB97X-D3/def2-TZVP/SMD(PC) Geometry Optimization"
    wB97X_D3_def2_TZVP_SMD_PC__Frequency_Analysis = "wB97X-D3/def2-TZVP/SMD(PC) Frequency Analysis"
    wB97X_D3_def2_TZVP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_SMD_PC__Unknown = "wB97X-D3/def2-TZVP/SMD(PC) Unknown"
    wB97X_D3_def2_TZVP_CMIRS_WATER__Single_Point = "wB97X-D3/def2-TZVP/CMIRS(WATER) Single Point"
    wB97X_D3_def2_TZVP_CMIRS_WATER__Force = "wB97X-D3/def2-TZVP/CMIRS(WATER) Force"
    wB97X_D3_def2_TZVP_CMIRS_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVP/CMIRS(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVP_CMIRS_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVP/CMIRS(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_WATER__Unknown = "wB97X-D3/def2-TZVP/CMIRS(WATER) Unknown"
    wB97X_D3_def2_TZVP_CMIRS_THF__Single_Point = "wB97X-D3/def2-TZVP/CMIRS(THF) Single Point"
    wB97X_D3_def2_TZVP_CMIRS_THF__Force = "wB97X-D3/def2-TZVP/CMIRS(THF) Force"
    wB97X_D3_def2_TZVP_CMIRS_THF__Geometry_Optimization = "wB97X-D3/def2-TZVP/CMIRS(THF) Geometry Optimization"
    wB97X_D3_def2_TZVP_CMIRS_THF__Frequency_Analysis = "wB97X-D3/def2-TZVP/CMIRS(THF) Frequency Analysis"
    wB97X_D3_def2_TZVP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_THF__Unknown = "wB97X-D3/def2-TZVP/CMIRS(THF) Unknown"
    wB97X_D3_def2_TZVP_CMIRS_G2__Single_Point = "wB97X-D3/def2-TZVP/CMIRS(G2) Single Point"
    wB97X_D3_def2_TZVP_CMIRS_G2__Force = "wB97X-D3/def2-TZVP/CMIRS(G2) Force"
    wB97X_D3_def2_TZVP_CMIRS_G2__Geometry_Optimization = "wB97X-D3/def2-TZVP/CMIRS(G2) Geometry Optimization"
    wB97X_D3_def2_TZVP_CMIRS_G2__Frequency_Analysis = "wB97X-D3/def2-TZVP/CMIRS(G2) Frequency Analysis"
    wB97X_D3_def2_TZVP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_G2__Unknown = "wB97X-D3/def2-TZVP/CMIRS(G2) Unknown"
    wB97X_D3_def2_TZVP_CMIRS_DME__Single_Point = "wB97X-D3/def2-TZVP/CMIRS(DME) Single Point"
    wB97X_D3_def2_TZVP_CMIRS_DME__Force = "wB97X-D3/def2-TZVP/CMIRS(DME) Force"
    wB97X_D3_def2_TZVP_CMIRS_DME__Geometry_Optimization = "wB97X-D3/def2-TZVP/CMIRS(DME) Geometry Optimization"
    wB97X_D3_def2_TZVP_CMIRS_DME__Frequency_Analysis = "wB97X-D3/def2-TZVP/CMIRS(DME) Frequency Analysis"
    wB97X_D3_def2_TZVP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_DME__Unknown = "wB97X-D3/def2-TZVP/CMIRS(DME) Unknown"
    wB97X_D3_def2_TZVP_CMIRS_EC__Single_Point = "wB97X-D3/def2-TZVP/CMIRS(EC) Single Point"
    wB97X_D3_def2_TZVP_CMIRS_EC__Force = "wB97X-D3/def2-TZVP/CMIRS(EC) Force"
    wB97X_D3_def2_TZVP_CMIRS_EC__Geometry_Optimization = "wB97X-D3/def2-TZVP/CMIRS(EC) Geometry Optimization"
    wB97X_D3_def2_TZVP_CMIRS_EC__Frequency_Analysis = "wB97X-D3/def2-TZVP/CMIRS(EC) Frequency Analysis"
    wB97X_D3_def2_TZVP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_EC__Unknown = "wB97X-D3/def2-TZVP/CMIRS(EC) Unknown"
    wB97X_D3_def2_TZVP_CMIRS_EC_EMC__Single_Point = "wB97X-D3/def2-TZVP/CMIRS(EC/EMC) Single Point"
    wB97X_D3_def2_TZVP_CMIRS_EC_EMC__Force = "wB97X-D3/def2-TZVP/CMIRS(EC/EMC) Force"
    wB97X_D3_def2_TZVP_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVP/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVP_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVP/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_EC_EMC__Unknown = "wB97X-D3/def2-TZVP/CMIRS(EC/EMC) Unknown"
    wB97X_D3_def2_TZVP_CMIRS_PC__Single_Point = "wB97X-D3/def2-TZVP/CMIRS(PC) Single Point"
    wB97X_D3_def2_TZVP_CMIRS_PC__Force = "wB97X-D3/def2-TZVP/CMIRS(PC) Force"
    wB97X_D3_def2_TZVP_CMIRS_PC__Geometry_Optimization = "wB97X-D3/def2-TZVP/CMIRS(PC) Geometry Optimization"
    wB97X_D3_def2_TZVP_CMIRS_PC__Frequency_Analysis = "wB97X-D3/def2-TZVP/CMIRS(PC) Frequency Analysis"
    wB97X_D3_def2_TZVP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVP_CMIRS_PC__Unknown = "wB97X-D3/def2-TZVP/CMIRS(PC) Unknown"
    wB97X_D3_def2_TZVPD_VACUUM_Single_Point = "wB97X-D3/def2-TZVPD/VACUUM Single Point"
    wB97X_D3_def2_TZVPD_VACUUM_Force = "wB97X-D3/def2-TZVPD/VACUUM Force"
    wB97X_D3_def2_TZVPD_VACUUM_Geometry_Optimization = "wB97X-D3/def2-TZVPD/VACUUM Geometry Optimization"
    wB97X_D3_def2_TZVPD_VACUUM_Frequency_Analysis = "wB97X-D3/def2-TZVPD/VACUUM Frequency Analysis"
    wB97X_D3_def2_TZVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_VACUUM_Unknown = "wB97X-D3/def2-TZVPD/VACUUM Unknown"
    wB97X_D3_def2_TZVPD_PCM_WATER__Single_Point = "wB97X-D3/def2-TZVPD/PCM(WATER) Single Point"
    wB97X_D3_def2_TZVPD_PCM_WATER__Force = "wB97X-D3/def2-TZVPD/PCM(WATER) Force"
    wB97X_D3_def2_TZVPD_PCM_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVPD/PCM(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVPD_PCM_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPD/PCM(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_WATER__Unknown = "wB97X-D3/def2-TZVPD/PCM(WATER) Unknown"
    wB97X_D3_def2_TZVPD_PCM_THF__Single_Point = "wB97X-D3/def2-TZVPD/PCM(THF) Single Point"
    wB97X_D3_def2_TZVPD_PCM_THF__Force = "wB97X-D3/def2-TZVPD/PCM(THF) Force"
    wB97X_D3_def2_TZVPD_PCM_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPD/PCM(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPD_PCM_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPD/PCM(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_THF__Unknown = "wB97X-D3/def2-TZVPD/PCM(THF) Unknown"
    wB97X_D3_def2_TZVPD_PCM_G2__Single_Point = "wB97X-D3/def2-TZVPD/PCM(G2) Single Point"
    wB97X_D3_def2_TZVPD_PCM_G2__Force = "wB97X-D3/def2-TZVPD/PCM(G2) Force"
    wB97X_D3_def2_TZVPD_PCM_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPD/PCM(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPD_PCM_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPD/PCM(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_G2__Unknown = "wB97X-D3/def2-TZVPD/PCM(G2) Unknown"
    wB97X_D3_def2_TZVPD_PCM_DME__Single_Point = "wB97X-D3/def2-TZVPD/PCM(DME) Single Point"
    wB97X_D3_def2_TZVPD_PCM_DME__Force = "wB97X-D3/def2-TZVPD/PCM(DME) Force"
    wB97X_D3_def2_TZVPD_PCM_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPD/PCM(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPD_PCM_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPD/PCM(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_DME__Unknown = "wB97X-D3/def2-TZVPD/PCM(DME) Unknown"
    wB97X_D3_def2_TZVPD_PCM_EC__Single_Point = "wB97X-D3/def2-TZVPD/PCM(EC) Single Point"
    wB97X_D3_def2_TZVPD_PCM_EC__Force = "wB97X-D3/def2-TZVPD/PCM(EC) Force"
    wB97X_D3_def2_TZVPD_PCM_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPD/PCM(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPD_PCM_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/PCM(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_EC__Unknown = "wB97X-D3/def2-TZVPD/PCM(EC) Unknown"
    wB97X_D3_def2_TZVPD_PCM_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPD/PCM(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPD_PCM_EC_EMC__Force = "wB97X-D3/def2-TZVPD/PCM(EC/EMC) Force"
    wB97X_D3_def2_TZVPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_EC_EMC__Unknown = "wB97X-D3/def2-TZVPD/PCM(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPD_PCM_PC__Single_Point = "wB97X-D3/def2-TZVPD/PCM(PC) Single Point"
    wB97X_D3_def2_TZVPD_PCM_PC__Force = "wB97X-D3/def2-TZVPD/PCM(PC) Force"
    wB97X_D3_def2_TZVPD_PCM_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPD/PCM(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPD_PCM_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/PCM(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_PCM_PC__Unknown = "wB97X-D3/def2-TZVPD/PCM(PC) Unknown"
    wB97X_D3_def2_TZVPD_ISOSVP_WATER__Single_Point = "wB97X-D3/def2-TZVPD/ISOSVP(WATER) Single Point"
    wB97X_D3_def2_TZVPD_ISOSVP_WATER__Force = "wB97X-D3/def2-TZVPD/ISOSVP(WATER) Force"
    wB97X_D3_def2_TZVPD_ISOSVP_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVPD/ISOSVP(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_WATER__Unknown = "wB97X-D3/def2-TZVPD/ISOSVP(WATER) Unknown"
    wB97X_D3_def2_TZVPD_ISOSVP_THF__Single_Point = "wB97X-D3/def2-TZVPD/ISOSVP(THF) Single Point"
    wB97X_D3_def2_TZVPD_ISOSVP_THF__Force = "wB97X-D3/def2-TZVPD/ISOSVP(THF) Force"
    wB97X_D3_def2_TZVPD_ISOSVP_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPD/ISOSVP(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPD_ISOSVP_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPD/ISOSVP(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_THF__Unknown = "wB97X-D3/def2-TZVPD/ISOSVP(THF) Unknown"
    wB97X_D3_def2_TZVPD_ISOSVP_G2__Single_Point = "wB97X-D3/def2-TZVPD/ISOSVP(G2) Single Point"
    wB97X_D3_def2_TZVPD_ISOSVP_G2__Force = "wB97X-D3/def2-TZVPD/ISOSVP(G2) Force"
    wB97X_D3_def2_TZVPD_ISOSVP_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPD/ISOSVP(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPD_ISOSVP_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPD/ISOSVP(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_G2__Unknown = "wB97X-D3/def2-TZVPD/ISOSVP(G2) Unknown"
    wB97X_D3_def2_TZVPD_ISOSVP_DME__Single_Point = "wB97X-D3/def2-TZVPD/ISOSVP(DME) Single Point"
    wB97X_D3_def2_TZVPD_ISOSVP_DME__Force = "wB97X-D3/def2-TZVPD/ISOSVP(DME) Force"
    wB97X_D3_def2_TZVPD_ISOSVP_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPD/ISOSVP(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPD_ISOSVP_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPD/ISOSVP(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_DME__Unknown = "wB97X-D3/def2-TZVPD/ISOSVP(DME) Unknown"
    wB97X_D3_def2_TZVPD_ISOSVP_EC__Single_Point = "wB97X-D3/def2-TZVPD/ISOSVP(EC) Single Point"
    wB97X_D3_def2_TZVPD_ISOSVP_EC__Force = "wB97X-D3/def2-TZVPD/ISOSVP(EC) Force"
    wB97X_D3_def2_TZVPD_ISOSVP_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPD/ISOSVP(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPD_ISOSVP_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/ISOSVP(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_EC__Unknown = "wB97X-D3/def2-TZVPD/ISOSVP(EC) Unknown"
    wB97X_D3_def2_TZVPD_ISOSVP_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPD/ISOSVP(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPD_ISOSVP_EC_EMC__Force = "wB97X-D3/def2-TZVPD/ISOSVP(EC/EMC) Force"
    wB97X_D3_def2_TZVPD_ISOSVP_EC_EMC__Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(EC/EMC) Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_EC_EMC__Unknown = "wB97X-D3/def2-TZVPD/ISOSVP(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPD_ISOSVP_PC__Single_Point = "wB97X-D3/def2-TZVPD/ISOSVP(PC) Single Point"
    wB97X_D3_def2_TZVPD_ISOSVP_PC__Force = "wB97X-D3/def2-TZVPD/ISOSVP(PC) Force"
    wB97X_D3_def2_TZVPD_ISOSVP_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPD/ISOSVP(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPD_ISOSVP_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/ISOSVP(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_ISOSVP_PC__Unknown = "wB97X-D3/def2-TZVPD/ISOSVP(PC) Unknown"
    wB97X_D3_def2_TZVPD_SMD_WATER__Single_Point = "wB97X-D3/def2-TZVPD/SMD(WATER) Single Point"
    wB97X_D3_def2_TZVPD_SMD_WATER__Force = "wB97X-D3/def2-TZVPD/SMD(WATER) Force"
    wB97X_D3_def2_TZVPD_SMD_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVPD/SMD(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVPD_SMD_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPD/SMD(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_WATER__Unknown = "wB97X-D3/def2-TZVPD/SMD(WATER) Unknown"
    wB97X_D3_def2_TZVPD_SMD_THF__Single_Point = "wB97X-D3/def2-TZVPD/SMD(THF) Single Point"
    wB97X_D3_def2_TZVPD_SMD_THF__Force = "wB97X-D3/def2-TZVPD/SMD(THF) Force"
    wB97X_D3_def2_TZVPD_SMD_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPD/SMD(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPD_SMD_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPD/SMD(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_THF__Unknown = "wB97X-D3/def2-TZVPD/SMD(THF) Unknown"
    wB97X_D3_def2_TZVPD_SMD_G2__Single_Point = "wB97X-D3/def2-TZVPD/SMD(G2) Single Point"
    wB97X_D3_def2_TZVPD_SMD_G2__Force = "wB97X-D3/def2-TZVPD/SMD(G2) Force"
    wB97X_D3_def2_TZVPD_SMD_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPD/SMD(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPD_SMD_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPD/SMD(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_G2__Unknown = "wB97X-D3/def2-TZVPD/SMD(G2) Unknown"
    wB97X_D3_def2_TZVPD_SMD_DME__Single_Point = "wB97X-D3/def2-TZVPD/SMD(DME) Single Point"
    wB97X_D3_def2_TZVPD_SMD_DME__Force = "wB97X-D3/def2-TZVPD/SMD(DME) Force"
    wB97X_D3_def2_TZVPD_SMD_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPD/SMD(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPD_SMD_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPD/SMD(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_DME__Unknown = "wB97X-D3/def2-TZVPD/SMD(DME) Unknown"
    wB97X_D3_def2_TZVPD_SMD_EC__Single_Point = "wB97X-D3/def2-TZVPD/SMD(EC) Single Point"
    wB97X_D3_def2_TZVPD_SMD_EC__Force = "wB97X-D3/def2-TZVPD/SMD(EC) Force"
    wB97X_D3_def2_TZVPD_SMD_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPD/SMD(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPD_SMD_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/SMD(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_EC__Unknown = "wB97X-D3/def2-TZVPD/SMD(EC) Unknown"
    wB97X_D3_def2_TZVPD_SMD_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPD/SMD(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPD_SMD_EC_EMC__Force = "wB97X-D3/def2-TZVPD/SMD(EC/EMC) Force"
    wB97X_D3_def2_TZVPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_EC_EMC__Unknown = "wB97X-D3/def2-TZVPD/SMD(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPD_SMD_PC__Single_Point = "wB97X-D3/def2-TZVPD/SMD(PC) Single Point"
    wB97X_D3_def2_TZVPD_SMD_PC__Force = "wB97X-D3/def2-TZVPD/SMD(PC) Force"
    wB97X_D3_def2_TZVPD_SMD_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPD/SMD(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPD_SMD_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/SMD(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_SMD_PC__Unknown = "wB97X-D3/def2-TZVPD/SMD(PC) Unknown"
    wB97X_D3_def2_TZVPD_CMIRS_WATER__Single_Point = "wB97X-D3/def2-TZVPD/CMIRS(WATER) Single Point"
    wB97X_D3_def2_TZVPD_CMIRS_WATER__Force = "wB97X-D3/def2-TZVPD/CMIRS(WATER) Force"
    wB97X_D3_def2_TZVPD_CMIRS_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVPD/CMIRS(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVPD_CMIRS_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPD/CMIRS(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_WATER__Unknown = "wB97X-D3/def2-TZVPD/CMIRS(WATER) Unknown"
    wB97X_D3_def2_TZVPD_CMIRS_THF__Single_Point = "wB97X-D3/def2-TZVPD/CMIRS(THF) Single Point"
    wB97X_D3_def2_TZVPD_CMIRS_THF__Force = "wB97X-D3/def2-TZVPD/CMIRS(THF) Force"
    wB97X_D3_def2_TZVPD_CMIRS_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPD/CMIRS(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPD_CMIRS_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPD/CMIRS(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_THF__Unknown = "wB97X-D3/def2-TZVPD/CMIRS(THF) Unknown"
    wB97X_D3_def2_TZVPD_CMIRS_G2__Single_Point = "wB97X-D3/def2-TZVPD/CMIRS(G2) Single Point"
    wB97X_D3_def2_TZVPD_CMIRS_G2__Force = "wB97X-D3/def2-TZVPD/CMIRS(G2) Force"
    wB97X_D3_def2_TZVPD_CMIRS_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPD/CMIRS(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPD_CMIRS_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPD/CMIRS(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_G2__Unknown = "wB97X-D3/def2-TZVPD/CMIRS(G2) Unknown"
    wB97X_D3_def2_TZVPD_CMIRS_DME__Single_Point = "wB97X-D3/def2-TZVPD/CMIRS(DME) Single Point"
    wB97X_D3_def2_TZVPD_CMIRS_DME__Force = "wB97X-D3/def2-TZVPD/CMIRS(DME) Force"
    wB97X_D3_def2_TZVPD_CMIRS_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPD/CMIRS(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPD_CMIRS_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPD/CMIRS(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_DME__Unknown = "wB97X-D3/def2-TZVPD/CMIRS(DME) Unknown"
    wB97X_D3_def2_TZVPD_CMIRS_EC__Single_Point = "wB97X-D3/def2-TZVPD/CMIRS(EC) Single Point"
    wB97X_D3_def2_TZVPD_CMIRS_EC__Force = "wB97X-D3/def2-TZVPD/CMIRS(EC) Force"
    wB97X_D3_def2_TZVPD_CMIRS_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPD/CMIRS(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPD_CMIRS_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/CMIRS(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_EC__Unknown = "wB97X-D3/def2-TZVPD/CMIRS(EC) Unknown"
    wB97X_D3_def2_TZVPD_CMIRS_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPD/CMIRS(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPD_CMIRS_EC_EMC__Force = "wB97X-D3/def2-TZVPD/CMIRS(EC/EMC) Force"
    wB97X_D3_def2_TZVPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_EC_EMC__Unknown = "wB97X-D3/def2-TZVPD/CMIRS(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPD_CMIRS_PC__Single_Point = "wB97X-D3/def2-TZVPD/CMIRS(PC) Single Point"
    wB97X_D3_def2_TZVPD_CMIRS_PC__Force = "wB97X-D3/def2-TZVPD/CMIRS(PC) Force"
    wB97X_D3_def2_TZVPD_CMIRS_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPD/CMIRS(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPD_CMIRS_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPD/CMIRS(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPD_CMIRS_PC__Unknown = "wB97X-D3/def2-TZVPD/CMIRS(PC) Unknown"
    wB97X_D3_def2_TZVPP_VACUUM_Single_Point = "wB97X-D3/def2-TZVPP/VACUUM Single Point"
    wB97X_D3_def2_TZVPP_VACUUM_Force = "wB97X-D3/def2-TZVPP/VACUUM Force"
    wB97X_D3_def2_TZVPP_VACUUM_Geometry_Optimization = "wB97X-D3/def2-TZVPP/VACUUM Geometry Optimization"
    wB97X_D3_def2_TZVPP_VACUUM_Frequency_Analysis = "wB97X-D3/def2-TZVPP/VACUUM Frequency Analysis"
    wB97X_D3_def2_TZVPP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_VACUUM_Unknown = "wB97X-D3/def2-TZVPP/VACUUM Unknown"
    wB97X_D3_def2_TZVPP_PCM_WATER__Single_Point = "wB97X-D3/def2-TZVPP/PCM(WATER) Single Point"
    wB97X_D3_def2_TZVPP_PCM_WATER__Force = "wB97X-D3/def2-TZVPP/PCM(WATER) Force"
    wB97X_D3_def2_TZVPP_PCM_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVPP/PCM(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVPP_PCM_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPP/PCM(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_WATER__Unknown = "wB97X-D3/def2-TZVPP/PCM(WATER) Unknown"
    wB97X_D3_def2_TZVPP_PCM_THF__Single_Point = "wB97X-D3/def2-TZVPP/PCM(THF) Single Point"
    wB97X_D3_def2_TZVPP_PCM_THF__Force = "wB97X-D3/def2-TZVPP/PCM(THF) Force"
    wB97X_D3_def2_TZVPP_PCM_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPP/PCM(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPP_PCM_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPP/PCM(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_THF__Unknown = "wB97X-D3/def2-TZVPP/PCM(THF) Unknown"
    wB97X_D3_def2_TZVPP_PCM_G2__Single_Point = "wB97X-D3/def2-TZVPP/PCM(G2) Single Point"
    wB97X_D3_def2_TZVPP_PCM_G2__Force = "wB97X-D3/def2-TZVPP/PCM(G2) Force"
    wB97X_D3_def2_TZVPP_PCM_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPP/PCM(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPP_PCM_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPP/PCM(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_G2__Unknown = "wB97X-D3/def2-TZVPP/PCM(G2) Unknown"
    wB97X_D3_def2_TZVPP_PCM_DME__Single_Point = "wB97X-D3/def2-TZVPP/PCM(DME) Single Point"
    wB97X_D3_def2_TZVPP_PCM_DME__Force = "wB97X-D3/def2-TZVPP/PCM(DME) Force"
    wB97X_D3_def2_TZVPP_PCM_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPP/PCM(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPP_PCM_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPP/PCM(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_DME__Unknown = "wB97X-D3/def2-TZVPP/PCM(DME) Unknown"
    wB97X_D3_def2_TZVPP_PCM_EC__Single_Point = "wB97X-D3/def2-TZVPP/PCM(EC) Single Point"
    wB97X_D3_def2_TZVPP_PCM_EC__Force = "wB97X-D3/def2-TZVPP/PCM(EC) Force"
    wB97X_D3_def2_TZVPP_PCM_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPP/PCM(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPP_PCM_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/PCM(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_EC__Unknown = "wB97X-D3/def2-TZVPP/PCM(EC) Unknown"
    wB97X_D3_def2_TZVPP_PCM_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPP/PCM(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPP_PCM_EC_EMC__Force = "wB97X-D3/def2-TZVPP/PCM(EC/EMC) Force"
    wB97X_D3_def2_TZVPP_PCM_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVPP/PCM(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVPP_PCM_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/PCM(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_EC_EMC__Unknown = "wB97X-D3/def2-TZVPP/PCM(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPP_PCM_PC__Single_Point = "wB97X-D3/def2-TZVPP/PCM(PC) Single Point"
    wB97X_D3_def2_TZVPP_PCM_PC__Force = "wB97X-D3/def2-TZVPP/PCM(PC) Force"
    wB97X_D3_def2_TZVPP_PCM_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPP/PCM(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPP_PCM_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/PCM(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_PCM_PC__Unknown = "wB97X-D3/def2-TZVPP/PCM(PC) Unknown"
    wB97X_D3_def2_TZVPP_ISOSVP_WATER__Single_Point = "wB97X-D3/def2-TZVPP/ISOSVP(WATER) Single Point"
    wB97X_D3_def2_TZVPP_ISOSVP_WATER__Force = "wB97X-D3/def2-TZVPP/ISOSVP(WATER) Force"
    wB97X_D3_def2_TZVPP_ISOSVP_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVPP/ISOSVP(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVPP_ISOSVP_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPP/ISOSVP(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_WATER__Unknown = "wB97X-D3/def2-TZVPP/ISOSVP(WATER) Unknown"
    wB97X_D3_def2_TZVPP_ISOSVP_THF__Single_Point = "wB97X-D3/def2-TZVPP/ISOSVP(THF) Single Point"
    wB97X_D3_def2_TZVPP_ISOSVP_THF__Force = "wB97X-D3/def2-TZVPP/ISOSVP(THF) Force"
    wB97X_D3_def2_TZVPP_ISOSVP_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPP/ISOSVP(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPP_ISOSVP_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPP/ISOSVP(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_THF__Unknown = "wB97X-D3/def2-TZVPP/ISOSVP(THF) Unknown"
    wB97X_D3_def2_TZVPP_ISOSVP_G2__Single_Point = "wB97X-D3/def2-TZVPP/ISOSVP(G2) Single Point"
    wB97X_D3_def2_TZVPP_ISOSVP_G2__Force = "wB97X-D3/def2-TZVPP/ISOSVP(G2) Force"
    wB97X_D3_def2_TZVPP_ISOSVP_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPP/ISOSVP(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPP_ISOSVP_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPP/ISOSVP(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_G2__Unknown = "wB97X-D3/def2-TZVPP/ISOSVP(G2) Unknown"
    wB97X_D3_def2_TZVPP_ISOSVP_DME__Single_Point = "wB97X-D3/def2-TZVPP/ISOSVP(DME) Single Point"
    wB97X_D3_def2_TZVPP_ISOSVP_DME__Force = "wB97X-D3/def2-TZVPP/ISOSVP(DME) Force"
    wB97X_D3_def2_TZVPP_ISOSVP_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPP/ISOSVP(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPP_ISOSVP_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPP/ISOSVP(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_DME__Unknown = "wB97X-D3/def2-TZVPP/ISOSVP(DME) Unknown"
    wB97X_D3_def2_TZVPP_ISOSVP_EC__Single_Point = "wB97X-D3/def2-TZVPP/ISOSVP(EC) Single Point"
    wB97X_D3_def2_TZVPP_ISOSVP_EC__Force = "wB97X-D3/def2-TZVPP/ISOSVP(EC) Force"
    wB97X_D3_def2_TZVPP_ISOSVP_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPP/ISOSVP(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPP_ISOSVP_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/ISOSVP(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_EC__Unknown = "wB97X-D3/def2-TZVPP/ISOSVP(EC) Unknown"
    wB97X_D3_def2_TZVPP_ISOSVP_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPP/ISOSVP(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPP_ISOSVP_EC_EMC__Force = "wB97X-D3/def2-TZVPP/ISOSVP(EC/EMC) Force"
    wB97X_D3_def2_TZVPP_ISOSVP_EC_EMC__Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(EC/EMC) Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_EC_EMC__Unknown = "wB97X-D3/def2-TZVPP/ISOSVP(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPP_ISOSVP_PC__Single_Point = "wB97X-D3/def2-TZVPP/ISOSVP(PC) Single Point"
    wB97X_D3_def2_TZVPP_ISOSVP_PC__Force = "wB97X-D3/def2-TZVPP/ISOSVP(PC) Force"
    wB97X_D3_def2_TZVPP_ISOSVP_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPP/ISOSVP(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPP_ISOSVP_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/ISOSVP(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_ISOSVP_PC__Unknown = "wB97X-D3/def2-TZVPP/ISOSVP(PC) Unknown"
    wB97X_D3_def2_TZVPP_SMD_WATER__Single_Point = "wB97X-D3/def2-TZVPP/SMD(WATER) Single Point"
    wB97X_D3_def2_TZVPP_SMD_WATER__Force = "wB97X-D3/def2-TZVPP/SMD(WATER) Force"
    wB97X_D3_def2_TZVPP_SMD_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVPP/SMD(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVPP_SMD_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPP/SMD(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_WATER__Unknown = "wB97X-D3/def2-TZVPP/SMD(WATER) Unknown"
    wB97X_D3_def2_TZVPP_SMD_THF__Single_Point = "wB97X-D3/def2-TZVPP/SMD(THF) Single Point"
    wB97X_D3_def2_TZVPP_SMD_THF__Force = "wB97X-D3/def2-TZVPP/SMD(THF) Force"
    wB97X_D3_def2_TZVPP_SMD_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPP/SMD(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPP_SMD_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPP/SMD(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_THF__Unknown = "wB97X-D3/def2-TZVPP/SMD(THF) Unknown"
    wB97X_D3_def2_TZVPP_SMD_G2__Single_Point = "wB97X-D3/def2-TZVPP/SMD(G2) Single Point"
    wB97X_D3_def2_TZVPP_SMD_G2__Force = "wB97X-D3/def2-TZVPP/SMD(G2) Force"
    wB97X_D3_def2_TZVPP_SMD_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPP/SMD(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPP_SMD_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPP/SMD(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_G2__Unknown = "wB97X-D3/def2-TZVPP/SMD(G2) Unknown"
    wB97X_D3_def2_TZVPP_SMD_DME__Single_Point = "wB97X-D3/def2-TZVPP/SMD(DME) Single Point"
    wB97X_D3_def2_TZVPP_SMD_DME__Force = "wB97X-D3/def2-TZVPP/SMD(DME) Force"
    wB97X_D3_def2_TZVPP_SMD_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPP/SMD(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPP_SMD_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPP/SMD(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_DME__Unknown = "wB97X-D3/def2-TZVPP/SMD(DME) Unknown"
    wB97X_D3_def2_TZVPP_SMD_EC__Single_Point = "wB97X-D3/def2-TZVPP/SMD(EC) Single Point"
    wB97X_D3_def2_TZVPP_SMD_EC__Force = "wB97X-D3/def2-TZVPP/SMD(EC) Force"
    wB97X_D3_def2_TZVPP_SMD_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPP/SMD(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPP_SMD_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/SMD(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_EC__Unknown = "wB97X-D3/def2-TZVPP/SMD(EC) Unknown"
    wB97X_D3_def2_TZVPP_SMD_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPP/SMD(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPP_SMD_EC_EMC__Force = "wB97X-D3/def2-TZVPP/SMD(EC/EMC) Force"
    wB97X_D3_def2_TZVPP_SMD_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVPP/SMD(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVPP_SMD_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/SMD(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_EC_EMC__Unknown = "wB97X-D3/def2-TZVPP/SMD(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPP_SMD_PC__Single_Point = "wB97X-D3/def2-TZVPP/SMD(PC) Single Point"
    wB97X_D3_def2_TZVPP_SMD_PC__Force = "wB97X-D3/def2-TZVPP/SMD(PC) Force"
    wB97X_D3_def2_TZVPP_SMD_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPP/SMD(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPP_SMD_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/SMD(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_SMD_PC__Unknown = "wB97X-D3/def2-TZVPP/SMD(PC) Unknown"
    wB97X_D3_def2_TZVPP_CMIRS_WATER__Single_Point = "wB97X-D3/def2-TZVPP/CMIRS(WATER) Single Point"
    wB97X_D3_def2_TZVPP_CMIRS_WATER__Force = "wB97X-D3/def2-TZVPP/CMIRS(WATER) Force"
    wB97X_D3_def2_TZVPP_CMIRS_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVPP/CMIRS(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVPP_CMIRS_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPP/CMIRS(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_WATER__Unknown = "wB97X-D3/def2-TZVPP/CMIRS(WATER) Unknown"
    wB97X_D3_def2_TZVPP_CMIRS_THF__Single_Point = "wB97X-D3/def2-TZVPP/CMIRS(THF) Single Point"
    wB97X_D3_def2_TZVPP_CMIRS_THF__Force = "wB97X-D3/def2-TZVPP/CMIRS(THF) Force"
    wB97X_D3_def2_TZVPP_CMIRS_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPP/CMIRS(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPP_CMIRS_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPP/CMIRS(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_THF__Unknown = "wB97X-D3/def2-TZVPP/CMIRS(THF) Unknown"
    wB97X_D3_def2_TZVPP_CMIRS_G2__Single_Point = "wB97X-D3/def2-TZVPP/CMIRS(G2) Single Point"
    wB97X_D3_def2_TZVPP_CMIRS_G2__Force = "wB97X-D3/def2-TZVPP/CMIRS(G2) Force"
    wB97X_D3_def2_TZVPP_CMIRS_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPP/CMIRS(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPP_CMIRS_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPP/CMIRS(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_G2__Unknown = "wB97X-D3/def2-TZVPP/CMIRS(G2) Unknown"
    wB97X_D3_def2_TZVPP_CMIRS_DME__Single_Point = "wB97X-D3/def2-TZVPP/CMIRS(DME) Single Point"
    wB97X_D3_def2_TZVPP_CMIRS_DME__Force = "wB97X-D3/def2-TZVPP/CMIRS(DME) Force"
    wB97X_D3_def2_TZVPP_CMIRS_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPP/CMIRS(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPP_CMIRS_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPP/CMIRS(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_DME__Unknown = "wB97X-D3/def2-TZVPP/CMIRS(DME) Unknown"
    wB97X_D3_def2_TZVPP_CMIRS_EC__Single_Point = "wB97X-D3/def2-TZVPP/CMIRS(EC) Single Point"
    wB97X_D3_def2_TZVPP_CMIRS_EC__Force = "wB97X-D3/def2-TZVPP/CMIRS(EC) Force"
    wB97X_D3_def2_TZVPP_CMIRS_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPP/CMIRS(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPP_CMIRS_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/CMIRS(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_EC__Unknown = "wB97X-D3/def2-TZVPP/CMIRS(EC) Unknown"
    wB97X_D3_def2_TZVPP_CMIRS_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPP/CMIRS(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPP_CMIRS_EC_EMC__Force = "wB97X-D3/def2-TZVPP/CMIRS(EC/EMC) Force"
    wB97X_D3_def2_TZVPP_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVPP/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVPP_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_EC_EMC__Unknown = "wB97X-D3/def2-TZVPP/CMIRS(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPP_CMIRS_PC__Single_Point = "wB97X-D3/def2-TZVPP/CMIRS(PC) Single Point"
    wB97X_D3_def2_TZVPP_CMIRS_PC__Force = "wB97X-D3/def2-TZVPP/CMIRS(PC) Force"
    wB97X_D3_def2_TZVPP_CMIRS_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPP/CMIRS(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPP_CMIRS_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPP/CMIRS(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPP_CMIRS_PC__Unknown = "wB97X-D3/def2-TZVPP/CMIRS(PC) Unknown"
    wB97X_D3_def2_TZVPPD_VACUUM_Single_Point = "wB97X-D3/def2-TZVPPD/VACUUM Single Point"
    wB97X_D3_def2_TZVPPD_VACUUM_Force = "wB97X-D3/def2-TZVPPD/VACUUM Force"
    wB97X_D3_def2_TZVPPD_VACUUM_Geometry_Optimization = "wB97X-D3/def2-TZVPPD/VACUUM Geometry Optimization"
    wB97X_D3_def2_TZVPPD_VACUUM_Frequency_Analysis = "wB97X-D3/def2-TZVPPD/VACUUM Frequency Analysis"
    wB97X_D3_def2_TZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_VACUUM_Unknown = "wB97X-D3/def2-TZVPPD/VACUUM Unknown"
    wB97X_D3_def2_TZVPPD_PCM_WATER__Single_Point = "wB97X-D3/def2-TZVPPD/PCM(WATER) Single Point"
    wB97X_D3_def2_TZVPPD_PCM_WATER__Force = "wB97X-D3/def2-TZVPPD/PCM(WATER) Force"
    wB97X_D3_def2_TZVPPD_PCM_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/PCM(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_PCM_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/PCM(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_WATER__Unknown = "wB97X-D3/def2-TZVPPD/PCM(WATER) Unknown"
    wB97X_D3_def2_TZVPPD_PCM_THF__Single_Point = "wB97X-D3/def2-TZVPPD/PCM(THF) Single Point"
    wB97X_D3_def2_TZVPPD_PCM_THF__Force = "wB97X-D3/def2-TZVPPD/PCM(THF) Force"
    wB97X_D3_def2_TZVPPD_PCM_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/PCM(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_PCM_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/PCM(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_THF__Unknown = "wB97X-D3/def2-TZVPPD/PCM(THF) Unknown"
    wB97X_D3_def2_TZVPPD_PCM_G2__Single_Point = "wB97X-D3/def2-TZVPPD/PCM(G2) Single Point"
    wB97X_D3_def2_TZVPPD_PCM_G2__Force = "wB97X-D3/def2-TZVPPD/PCM(G2) Force"
    wB97X_D3_def2_TZVPPD_PCM_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/PCM(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_PCM_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/PCM(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_G2__Unknown = "wB97X-D3/def2-TZVPPD/PCM(G2) Unknown"
    wB97X_D3_def2_TZVPPD_PCM_DME__Single_Point = "wB97X-D3/def2-TZVPPD/PCM(DME) Single Point"
    wB97X_D3_def2_TZVPPD_PCM_DME__Force = "wB97X-D3/def2-TZVPPD/PCM(DME) Force"
    wB97X_D3_def2_TZVPPD_PCM_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/PCM(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_PCM_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/PCM(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_DME__Unknown = "wB97X-D3/def2-TZVPPD/PCM(DME) Unknown"
    wB97X_D3_def2_TZVPPD_PCM_EC__Single_Point = "wB97X-D3/def2-TZVPPD/PCM(EC) Single Point"
    wB97X_D3_def2_TZVPPD_PCM_EC__Force = "wB97X-D3/def2-TZVPPD/PCM(EC) Force"
    wB97X_D3_def2_TZVPPD_PCM_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/PCM(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_PCM_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/PCM(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_EC__Unknown = "wB97X-D3/def2-TZVPPD/PCM(EC) Unknown"
    wB97X_D3_def2_TZVPPD_PCM_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPPD/PCM(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPPD_PCM_EC_EMC__Force = "wB97X-D3/def2-TZVPPD/PCM(EC/EMC) Force"
    wB97X_D3_def2_TZVPPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_EC_EMC__Unknown = "wB97X-D3/def2-TZVPPD/PCM(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPPD_PCM_PC__Single_Point = "wB97X-D3/def2-TZVPPD/PCM(PC) Single Point"
    wB97X_D3_def2_TZVPPD_PCM_PC__Force = "wB97X-D3/def2-TZVPPD/PCM(PC) Force"
    wB97X_D3_def2_TZVPPD_PCM_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/PCM(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_PCM_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/PCM(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_PCM_PC__Unknown = "wB97X-D3/def2-TZVPPD/PCM(PC) Unknown"
    wB97X_D3_def2_TZVPPD_ISOSVP_WATER__Single_Point = "wB97X-D3/def2-TZVPPD/ISOSVP(WATER) Single Point"
    wB97X_D3_def2_TZVPPD_ISOSVP_WATER__Force = "wB97X-D3/def2-TZVPPD/ISOSVP(WATER) Force"
    wB97X_D3_def2_TZVPPD_ISOSVP_WATER__Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(WATER) Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_WATER__Unknown = "wB97X-D3/def2-TZVPPD/ISOSVP(WATER) Unknown"
    wB97X_D3_def2_TZVPPD_ISOSVP_THF__Single_Point = "wB97X-D3/def2-TZVPPD/ISOSVP(THF) Single Point"
    wB97X_D3_def2_TZVPPD_ISOSVP_THF__Force = "wB97X-D3/def2-TZVPPD/ISOSVP(THF) Force"
    wB97X_D3_def2_TZVPPD_ISOSVP_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/ISOSVP(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_ISOSVP_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/ISOSVP(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_THF__Unknown = "wB97X-D3/def2-TZVPPD/ISOSVP(THF) Unknown"
    wB97X_D3_def2_TZVPPD_ISOSVP_G2__Single_Point = "wB97X-D3/def2-TZVPPD/ISOSVP(G2) Single Point"
    wB97X_D3_def2_TZVPPD_ISOSVP_G2__Force = "wB97X-D3/def2-TZVPPD/ISOSVP(G2) Force"
    wB97X_D3_def2_TZVPPD_ISOSVP_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/ISOSVP(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_ISOSVP_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/ISOSVP(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_G2__Unknown = "wB97X-D3/def2-TZVPPD/ISOSVP(G2) Unknown"
    wB97X_D3_def2_TZVPPD_ISOSVP_DME__Single_Point = "wB97X-D3/def2-TZVPPD/ISOSVP(DME) Single Point"
    wB97X_D3_def2_TZVPPD_ISOSVP_DME__Force = "wB97X-D3/def2-TZVPPD/ISOSVP(DME) Force"
    wB97X_D3_def2_TZVPPD_ISOSVP_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/ISOSVP(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_ISOSVP_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/ISOSVP(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_DME__Unknown = "wB97X-D3/def2-TZVPPD/ISOSVP(DME) Unknown"
    wB97X_D3_def2_TZVPPD_ISOSVP_EC__Single_Point = "wB97X-D3/def2-TZVPPD/ISOSVP(EC) Single Point"
    wB97X_D3_def2_TZVPPD_ISOSVP_EC__Force = "wB97X-D3/def2-TZVPPD/ISOSVP(EC) Force"
    wB97X_D3_def2_TZVPPD_ISOSVP_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/ISOSVP(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_ISOSVP_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/ISOSVP(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_EC__Unknown = "wB97X-D3/def2-TZVPPD/ISOSVP(EC) Unknown"
    wB97X_D3_def2_TZVPPD_ISOSVP_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPPD/ISOSVP(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPPD_ISOSVP_EC_EMC__Force = "wB97X-D3/def2-TZVPPD/ISOSVP(EC/EMC) Force"
    wB97X_D3_def2_TZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_EC_EMC__Unknown = "wB97X-D3/def2-TZVPPD/ISOSVP(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPPD_ISOSVP_PC__Single_Point = "wB97X-D3/def2-TZVPPD/ISOSVP(PC) Single Point"
    wB97X_D3_def2_TZVPPD_ISOSVP_PC__Force = "wB97X-D3/def2-TZVPPD/ISOSVP(PC) Force"
    wB97X_D3_def2_TZVPPD_ISOSVP_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/ISOSVP(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_ISOSVP_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/ISOSVP(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_ISOSVP_PC__Unknown = "wB97X-D3/def2-TZVPPD/ISOSVP(PC) Unknown"
    wB97X_D3_def2_TZVPPD_SMD_WATER__Single_Point = "wB97X-D3/def2-TZVPPD/SMD(WATER) Single Point"
    wB97X_D3_def2_TZVPPD_SMD_WATER__Force = "wB97X-D3/def2-TZVPPD/SMD(WATER) Force"
    wB97X_D3_def2_TZVPPD_SMD_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/SMD(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_SMD_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/SMD(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_WATER__Unknown = "wB97X-D3/def2-TZVPPD/SMD(WATER) Unknown"
    wB97X_D3_def2_TZVPPD_SMD_THF__Single_Point = "wB97X-D3/def2-TZVPPD/SMD(THF) Single Point"
    wB97X_D3_def2_TZVPPD_SMD_THF__Force = "wB97X-D3/def2-TZVPPD/SMD(THF) Force"
    wB97X_D3_def2_TZVPPD_SMD_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/SMD(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_SMD_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/SMD(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_THF__Unknown = "wB97X-D3/def2-TZVPPD/SMD(THF) Unknown"
    wB97X_D3_def2_TZVPPD_SMD_G2__Single_Point = "wB97X-D3/def2-TZVPPD/SMD(G2) Single Point"
    wB97X_D3_def2_TZVPPD_SMD_G2__Force = "wB97X-D3/def2-TZVPPD/SMD(G2) Force"
    wB97X_D3_def2_TZVPPD_SMD_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/SMD(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_SMD_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/SMD(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_G2__Unknown = "wB97X-D3/def2-TZVPPD/SMD(G2) Unknown"
    wB97X_D3_def2_TZVPPD_SMD_DME__Single_Point = "wB97X-D3/def2-TZVPPD/SMD(DME) Single Point"
    wB97X_D3_def2_TZVPPD_SMD_DME__Force = "wB97X-D3/def2-TZVPPD/SMD(DME) Force"
    wB97X_D3_def2_TZVPPD_SMD_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/SMD(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_SMD_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/SMD(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_DME__Unknown = "wB97X-D3/def2-TZVPPD/SMD(DME) Unknown"
    wB97X_D3_def2_TZVPPD_SMD_EC__Single_Point = "wB97X-D3/def2-TZVPPD/SMD(EC) Single Point"
    wB97X_D3_def2_TZVPPD_SMD_EC__Force = "wB97X-D3/def2-TZVPPD/SMD(EC) Force"
    wB97X_D3_def2_TZVPPD_SMD_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/SMD(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_SMD_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/SMD(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_EC__Unknown = "wB97X-D3/def2-TZVPPD/SMD(EC) Unknown"
    wB97X_D3_def2_TZVPPD_SMD_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPPD/SMD(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPPD_SMD_EC_EMC__Force = "wB97X-D3/def2-TZVPPD/SMD(EC/EMC) Force"
    wB97X_D3_def2_TZVPPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_EC_EMC__Unknown = "wB97X-D3/def2-TZVPPD/SMD(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPPD_SMD_PC__Single_Point = "wB97X-D3/def2-TZVPPD/SMD(PC) Single Point"
    wB97X_D3_def2_TZVPPD_SMD_PC__Force = "wB97X-D3/def2-TZVPPD/SMD(PC) Force"
    wB97X_D3_def2_TZVPPD_SMD_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/SMD(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_SMD_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/SMD(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_SMD_PC__Unknown = "wB97X-D3/def2-TZVPPD/SMD(PC) Unknown"
    wB97X_D3_def2_TZVPPD_CMIRS_WATER__Single_Point = "wB97X-D3/def2-TZVPPD/CMIRS(WATER) Single Point"
    wB97X_D3_def2_TZVPPD_CMIRS_WATER__Force = "wB97X-D3/def2-TZVPPD/CMIRS(WATER) Force"
    wB97X_D3_def2_TZVPPD_CMIRS_WATER__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/CMIRS(WATER) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_CMIRS_WATER__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/CMIRS(WATER) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_WATER__Unknown = "wB97X-D3/def2-TZVPPD/CMIRS(WATER) Unknown"
    wB97X_D3_def2_TZVPPD_CMIRS_THF__Single_Point = "wB97X-D3/def2-TZVPPD/CMIRS(THF) Single Point"
    wB97X_D3_def2_TZVPPD_CMIRS_THF__Force = "wB97X-D3/def2-TZVPPD/CMIRS(THF) Force"
    wB97X_D3_def2_TZVPPD_CMIRS_THF__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/CMIRS(THF) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_CMIRS_THF__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/CMIRS(THF) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_THF__Unknown = "wB97X-D3/def2-TZVPPD/CMIRS(THF) Unknown"
    wB97X_D3_def2_TZVPPD_CMIRS_G2__Single_Point = "wB97X-D3/def2-TZVPPD/CMIRS(G2) Single Point"
    wB97X_D3_def2_TZVPPD_CMIRS_G2__Force = "wB97X-D3/def2-TZVPPD/CMIRS(G2) Force"
    wB97X_D3_def2_TZVPPD_CMIRS_G2__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/CMIRS(G2) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_CMIRS_G2__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/CMIRS(G2) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_G2__Unknown = "wB97X-D3/def2-TZVPPD/CMIRS(G2) Unknown"
    wB97X_D3_def2_TZVPPD_CMIRS_DME__Single_Point = "wB97X-D3/def2-TZVPPD/CMIRS(DME) Single Point"
    wB97X_D3_def2_TZVPPD_CMIRS_DME__Force = "wB97X-D3/def2-TZVPPD/CMIRS(DME) Force"
    wB97X_D3_def2_TZVPPD_CMIRS_DME__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/CMIRS(DME) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_CMIRS_DME__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/CMIRS(DME) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_DME__Unknown = "wB97X-D3/def2-TZVPPD/CMIRS(DME) Unknown"
    wB97X_D3_def2_TZVPPD_CMIRS_EC__Single_Point = "wB97X-D3/def2-TZVPPD/CMIRS(EC) Single Point"
    wB97X_D3_def2_TZVPPD_CMIRS_EC__Force = "wB97X-D3/def2-TZVPPD/CMIRS(EC) Force"
    wB97X_D3_def2_TZVPPD_CMIRS_EC__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/CMIRS(EC) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_CMIRS_EC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/CMIRS(EC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_EC__Unknown = "wB97X-D3/def2-TZVPPD/CMIRS(EC) Unknown"
    wB97X_D3_def2_TZVPPD_CMIRS_EC_EMC__Single_Point = "wB97X-D3/def2-TZVPPD/CMIRS(EC/EMC) Single Point"
    wB97X_D3_def2_TZVPPD_CMIRS_EC_EMC__Force = "wB97X-D3/def2-TZVPPD/CMIRS(EC/EMC) Force"
    wB97X_D3_def2_TZVPPD_CMIRS_EC_EMC__Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_EC_EMC__Unknown = "wB97X-D3/def2-TZVPPD/CMIRS(EC/EMC) Unknown"
    wB97X_D3_def2_TZVPPD_CMIRS_PC__Single_Point = "wB97X-D3/def2-TZVPPD/CMIRS(PC) Single Point"
    wB97X_D3_def2_TZVPPD_CMIRS_PC__Force = "wB97X-D3/def2-TZVPPD/CMIRS(PC) Force"
    wB97X_D3_def2_TZVPPD_CMIRS_PC__Geometry_Optimization = "wB97X-D3/def2-TZVPPD/CMIRS(PC) Geometry Optimization"
    wB97X_D3_def2_TZVPPD_CMIRS_PC__Frequency_Analysis = "wB97X-D3/def2-TZVPPD/CMIRS(PC) Frequency Analysis"
    wB97X_D3_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-TZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_TZVPPD_CMIRS_PC__Unknown = "wB97X-D3/def2-TZVPPD/CMIRS(PC) Unknown"
    wB97X_D3_def2_QZVPPD_VACUUM_Single_Point = "wB97X-D3/def2-QZVPPD/VACUUM Single Point"
    wB97X_D3_def2_QZVPPD_VACUUM_Force = "wB97X-D3/def2-QZVPPD/VACUUM Force"
    wB97X_D3_def2_QZVPPD_VACUUM_Geometry_Optimization = "wB97X-D3/def2-QZVPPD/VACUUM Geometry Optimization"
    wB97X_D3_def2_QZVPPD_VACUUM_Frequency_Analysis = "wB97X-D3/def2-QZVPPD/VACUUM Frequency Analysis"
    wB97X_D3_def2_QZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_VACUUM_Unknown = "wB97X-D3/def2-QZVPPD/VACUUM Unknown"
    wB97X_D3_def2_QZVPPD_PCM_WATER__Single_Point = "wB97X-D3/def2-QZVPPD/PCM(WATER) Single Point"
    wB97X_D3_def2_QZVPPD_PCM_WATER__Force = "wB97X-D3/def2-QZVPPD/PCM(WATER) Force"
    wB97X_D3_def2_QZVPPD_PCM_WATER__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/PCM(WATER) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_PCM_WATER__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/PCM(WATER) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_WATER__Unknown = "wB97X-D3/def2-QZVPPD/PCM(WATER) Unknown"
    wB97X_D3_def2_QZVPPD_PCM_THF__Single_Point = "wB97X-D3/def2-QZVPPD/PCM(THF) Single Point"
    wB97X_D3_def2_QZVPPD_PCM_THF__Force = "wB97X-D3/def2-QZVPPD/PCM(THF) Force"
    wB97X_D3_def2_QZVPPD_PCM_THF__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/PCM(THF) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_PCM_THF__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/PCM(THF) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_THF__Unknown = "wB97X-D3/def2-QZVPPD/PCM(THF) Unknown"
    wB97X_D3_def2_QZVPPD_PCM_G2__Single_Point = "wB97X-D3/def2-QZVPPD/PCM(G2) Single Point"
    wB97X_D3_def2_QZVPPD_PCM_G2__Force = "wB97X-D3/def2-QZVPPD/PCM(G2) Force"
    wB97X_D3_def2_QZVPPD_PCM_G2__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/PCM(G2) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_PCM_G2__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/PCM(G2) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_G2__Unknown = "wB97X-D3/def2-QZVPPD/PCM(G2) Unknown"
    wB97X_D3_def2_QZVPPD_PCM_DME__Single_Point = "wB97X-D3/def2-QZVPPD/PCM(DME) Single Point"
    wB97X_D3_def2_QZVPPD_PCM_DME__Force = "wB97X-D3/def2-QZVPPD/PCM(DME) Force"
    wB97X_D3_def2_QZVPPD_PCM_DME__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/PCM(DME) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_PCM_DME__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/PCM(DME) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_DME__Unknown = "wB97X-D3/def2-QZVPPD/PCM(DME) Unknown"
    wB97X_D3_def2_QZVPPD_PCM_EC__Single_Point = "wB97X-D3/def2-QZVPPD/PCM(EC) Single Point"
    wB97X_D3_def2_QZVPPD_PCM_EC__Force = "wB97X-D3/def2-QZVPPD/PCM(EC) Force"
    wB97X_D3_def2_QZVPPD_PCM_EC__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/PCM(EC) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_PCM_EC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/PCM(EC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_EC__Unknown = "wB97X-D3/def2-QZVPPD/PCM(EC) Unknown"
    wB97X_D3_def2_QZVPPD_PCM_EC_EMC__Single_Point = "wB97X-D3/def2-QZVPPD/PCM(EC/EMC) Single Point"
    wB97X_D3_def2_QZVPPD_PCM_EC_EMC__Force = "wB97X-D3/def2-QZVPPD/PCM(EC/EMC) Force"
    wB97X_D3_def2_QZVPPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_EC_EMC__Unknown = "wB97X-D3/def2-QZVPPD/PCM(EC/EMC) Unknown"
    wB97X_D3_def2_QZVPPD_PCM_PC__Single_Point = "wB97X-D3/def2-QZVPPD/PCM(PC) Single Point"
    wB97X_D3_def2_QZVPPD_PCM_PC__Force = "wB97X-D3/def2-QZVPPD/PCM(PC) Force"
    wB97X_D3_def2_QZVPPD_PCM_PC__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/PCM(PC) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_PCM_PC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/PCM(PC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_PCM_PC__Unknown = "wB97X-D3/def2-QZVPPD/PCM(PC) Unknown"
    wB97X_D3_def2_QZVPPD_ISOSVP_WATER__Single_Point = "wB97X-D3/def2-QZVPPD/ISOSVP(WATER) Single Point"
    wB97X_D3_def2_QZVPPD_ISOSVP_WATER__Force = "wB97X-D3/def2-QZVPPD/ISOSVP(WATER) Force"
    wB97X_D3_def2_QZVPPD_ISOSVP_WATER__Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(WATER) Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_WATER__Unknown = "wB97X-D3/def2-QZVPPD/ISOSVP(WATER) Unknown"
    wB97X_D3_def2_QZVPPD_ISOSVP_THF__Single_Point = "wB97X-D3/def2-QZVPPD/ISOSVP(THF) Single Point"
    wB97X_D3_def2_QZVPPD_ISOSVP_THF__Force = "wB97X-D3/def2-QZVPPD/ISOSVP(THF) Force"
    wB97X_D3_def2_QZVPPD_ISOSVP_THF__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/ISOSVP(THF) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_ISOSVP_THF__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/ISOSVP(THF) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_THF__Unknown = "wB97X-D3/def2-QZVPPD/ISOSVP(THF) Unknown"
    wB97X_D3_def2_QZVPPD_ISOSVP_G2__Single_Point = "wB97X-D3/def2-QZVPPD/ISOSVP(G2) Single Point"
    wB97X_D3_def2_QZVPPD_ISOSVP_G2__Force = "wB97X-D3/def2-QZVPPD/ISOSVP(G2) Force"
    wB97X_D3_def2_QZVPPD_ISOSVP_G2__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/ISOSVP(G2) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_ISOSVP_G2__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/ISOSVP(G2) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_G2__Unknown = "wB97X-D3/def2-QZVPPD/ISOSVP(G2) Unknown"
    wB97X_D3_def2_QZVPPD_ISOSVP_DME__Single_Point = "wB97X-D3/def2-QZVPPD/ISOSVP(DME) Single Point"
    wB97X_D3_def2_QZVPPD_ISOSVP_DME__Force = "wB97X-D3/def2-QZVPPD/ISOSVP(DME) Force"
    wB97X_D3_def2_QZVPPD_ISOSVP_DME__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/ISOSVP(DME) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_ISOSVP_DME__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/ISOSVP(DME) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_DME__Unknown = "wB97X-D3/def2-QZVPPD/ISOSVP(DME) Unknown"
    wB97X_D3_def2_QZVPPD_ISOSVP_EC__Single_Point = "wB97X-D3/def2-QZVPPD/ISOSVP(EC) Single Point"
    wB97X_D3_def2_QZVPPD_ISOSVP_EC__Force = "wB97X-D3/def2-QZVPPD/ISOSVP(EC) Force"
    wB97X_D3_def2_QZVPPD_ISOSVP_EC__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/ISOSVP(EC) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_ISOSVP_EC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/ISOSVP(EC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_EC__Unknown = "wB97X-D3/def2-QZVPPD/ISOSVP(EC) Unknown"
    wB97X_D3_def2_QZVPPD_ISOSVP_EC_EMC__Single_Point = "wB97X-D3/def2-QZVPPD/ISOSVP(EC/EMC) Single Point"
    wB97X_D3_def2_QZVPPD_ISOSVP_EC_EMC__Force = "wB97X-D3/def2-QZVPPD/ISOSVP(EC/EMC) Force"
    wB97X_D3_def2_QZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_EC_EMC__Unknown = "wB97X-D3/def2-QZVPPD/ISOSVP(EC/EMC) Unknown"
    wB97X_D3_def2_QZVPPD_ISOSVP_PC__Single_Point = "wB97X-D3/def2-QZVPPD/ISOSVP(PC) Single Point"
    wB97X_D3_def2_QZVPPD_ISOSVP_PC__Force = "wB97X-D3/def2-QZVPPD/ISOSVP(PC) Force"
    wB97X_D3_def2_QZVPPD_ISOSVP_PC__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/ISOSVP(PC) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_ISOSVP_PC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/ISOSVP(PC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_ISOSVP_PC__Unknown = "wB97X-D3/def2-QZVPPD/ISOSVP(PC) Unknown"
    wB97X_D3_def2_QZVPPD_SMD_WATER__Single_Point = "wB97X-D3/def2-QZVPPD/SMD(WATER) Single Point"
    wB97X_D3_def2_QZVPPD_SMD_WATER__Force = "wB97X-D3/def2-QZVPPD/SMD(WATER) Force"
    wB97X_D3_def2_QZVPPD_SMD_WATER__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/SMD(WATER) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_SMD_WATER__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/SMD(WATER) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_WATER__Unknown = "wB97X-D3/def2-QZVPPD/SMD(WATER) Unknown"
    wB97X_D3_def2_QZVPPD_SMD_THF__Single_Point = "wB97X-D3/def2-QZVPPD/SMD(THF) Single Point"
    wB97X_D3_def2_QZVPPD_SMD_THF__Force = "wB97X-D3/def2-QZVPPD/SMD(THF) Force"
    wB97X_D3_def2_QZVPPD_SMD_THF__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/SMD(THF) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_SMD_THF__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/SMD(THF) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_THF__Unknown = "wB97X-D3/def2-QZVPPD/SMD(THF) Unknown"
    wB97X_D3_def2_QZVPPD_SMD_G2__Single_Point = "wB97X-D3/def2-QZVPPD/SMD(G2) Single Point"
    wB97X_D3_def2_QZVPPD_SMD_G2__Force = "wB97X-D3/def2-QZVPPD/SMD(G2) Force"
    wB97X_D3_def2_QZVPPD_SMD_G2__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/SMD(G2) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_SMD_G2__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/SMD(G2) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_G2__Unknown = "wB97X-D3/def2-QZVPPD/SMD(G2) Unknown"
    wB97X_D3_def2_QZVPPD_SMD_DME__Single_Point = "wB97X-D3/def2-QZVPPD/SMD(DME) Single Point"
    wB97X_D3_def2_QZVPPD_SMD_DME__Force = "wB97X-D3/def2-QZVPPD/SMD(DME) Force"
    wB97X_D3_def2_QZVPPD_SMD_DME__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/SMD(DME) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_SMD_DME__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/SMD(DME) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_DME__Unknown = "wB97X-D3/def2-QZVPPD/SMD(DME) Unknown"
    wB97X_D3_def2_QZVPPD_SMD_EC__Single_Point = "wB97X-D3/def2-QZVPPD/SMD(EC) Single Point"
    wB97X_D3_def2_QZVPPD_SMD_EC__Force = "wB97X-D3/def2-QZVPPD/SMD(EC) Force"
    wB97X_D3_def2_QZVPPD_SMD_EC__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/SMD(EC) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_SMD_EC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/SMD(EC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_EC__Unknown = "wB97X-D3/def2-QZVPPD/SMD(EC) Unknown"
    wB97X_D3_def2_QZVPPD_SMD_EC_EMC__Single_Point = "wB97X-D3/def2-QZVPPD/SMD(EC/EMC) Single Point"
    wB97X_D3_def2_QZVPPD_SMD_EC_EMC__Force = "wB97X-D3/def2-QZVPPD/SMD(EC/EMC) Force"
    wB97X_D3_def2_QZVPPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_EC_EMC__Unknown = "wB97X-D3/def2-QZVPPD/SMD(EC/EMC) Unknown"
    wB97X_D3_def2_QZVPPD_SMD_PC__Single_Point = "wB97X-D3/def2-QZVPPD/SMD(PC) Single Point"
    wB97X_D3_def2_QZVPPD_SMD_PC__Force = "wB97X-D3/def2-QZVPPD/SMD(PC) Force"
    wB97X_D3_def2_QZVPPD_SMD_PC__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/SMD(PC) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_SMD_PC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/SMD(PC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_SMD_PC__Unknown = "wB97X-D3/def2-QZVPPD/SMD(PC) Unknown"
    wB97X_D3_def2_QZVPPD_CMIRS_WATER__Single_Point = "wB97X-D3/def2-QZVPPD/CMIRS(WATER) Single Point"
    wB97X_D3_def2_QZVPPD_CMIRS_WATER__Force = "wB97X-D3/def2-QZVPPD/CMIRS(WATER) Force"
    wB97X_D3_def2_QZVPPD_CMIRS_WATER__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/CMIRS(WATER) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_CMIRS_WATER__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/CMIRS(WATER) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_WATER__Unknown = "wB97X-D3/def2-QZVPPD/CMIRS(WATER) Unknown"
    wB97X_D3_def2_QZVPPD_CMIRS_THF__Single_Point = "wB97X-D3/def2-QZVPPD/CMIRS(THF) Single Point"
    wB97X_D3_def2_QZVPPD_CMIRS_THF__Force = "wB97X-D3/def2-QZVPPD/CMIRS(THF) Force"
    wB97X_D3_def2_QZVPPD_CMIRS_THF__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/CMIRS(THF) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_CMIRS_THF__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/CMIRS(THF) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_THF__Unknown = "wB97X-D3/def2-QZVPPD/CMIRS(THF) Unknown"
    wB97X_D3_def2_QZVPPD_CMIRS_G2__Single_Point = "wB97X-D3/def2-QZVPPD/CMIRS(G2) Single Point"
    wB97X_D3_def2_QZVPPD_CMIRS_G2__Force = "wB97X-D3/def2-QZVPPD/CMIRS(G2) Force"
    wB97X_D3_def2_QZVPPD_CMIRS_G2__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/CMIRS(G2) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_CMIRS_G2__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/CMIRS(G2) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_G2__Unknown = "wB97X-D3/def2-QZVPPD/CMIRS(G2) Unknown"
    wB97X_D3_def2_QZVPPD_CMIRS_DME__Single_Point = "wB97X-D3/def2-QZVPPD/CMIRS(DME) Single Point"
    wB97X_D3_def2_QZVPPD_CMIRS_DME__Force = "wB97X-D3/def2-QZVPPD/CMIRS(DME) Force"
    wB97X_D3_def2_QZVPPD_CMIRS_DME__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/CMIRS(DME) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_CMIRS_DME__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/CMIRS(DME) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_DME__Unknown = "wB97X-D3/def2-QZVPPD/CMIRS(DME) Unknown"
    wB97X_D3_def2_QZVPPD_CMIRS_EC__Single_Point = "wB97X-D3/def2-QZVPPD/CMIRS(EC) Single Point"
    wB97X_D3_def2_QZVPPD_CMIRS_EC__Force = "wB97X-D3/def2-QZVPPD/CMIRS(EC) Force"
    wB97X_D3_def2_QZVPPD_CMIRS_EC__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/CMIRS(EC) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_CMIRS_EC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/CMIRS(EC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_EC__Unknown = "wB97X-D3/def2-QZVPPD/CMIRS(EC) Unknown"
    wB97X_D3_def2_QZVPPD_CMIRS_EC_EMC__Single_Point = "wB97X-D3/def2-QZVPPD/CMIRS(EC/EMC) Single Point"
    wB97X_D3_def2_QZVPPD_CMIRS_EC_EMC__Force = "wB97X-D3/def2-QZVPPD/CMIRS(EC/EMC) Force"
    wB97X_D3_def2_QZVPPD_CMIRS_EC_EMC__Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_EC_EMC__Unknown = "wB97X-D3/def2-QZVPPD/CMIRS(EC/EMC) Unknown"
    wB97X_D3_def2_QZVPPD_CMIRS_PC__Single_Point = "wB97X-D3/def2-QZVPPD/CMIRS(PC) Single Point"
    wB97X_D3_def2_QZVPPD_CMIRS_PC__Force = "wB97X-D3/def2-QZVPPD/CMIRS(PC) Force"
    wB97X_D3_def2_QZVPPD_CMIRS_PC__Geometry_Optimization = "wB97X-D3/def2-QZVPPD/CMIRS(PC) Geometry Optimization"
    wB97X_D3_def2_QZVPPD_CMIRS_PC__Frequency_Analysis = "wB97X-D3/def2-QZVPPD/CMIRS(PC) Frequency Analysis"
    wB97X_D3_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-D3/def2-QZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_D3_def2_QZVPPD_CMIRS_PC__Unknown = "wB97X-D3/def2-QZVPPD/CMIRS(PC) Unknown"
    wB97X_V_6_31g_d_VACUUM_Single_Point = "wB97X-V/6-31g*/VACUUM Single Point"
    wB97X_V_6_31g_d_VACUUM_Force = "wB97X-V/6-31g*/VACUUM Force"
    wB97X_V_6_31g_d_VACUUM_Geometry_Optimization = "wB97X-V/6-31g*/VACUUM Geometry Optimization"
    wB97X_V_6_31g_d_VACUUM_Frequency_Analysis = "wB97X-V/6-31g*/VACUUM Frequency Analysis"
    wB97X_V_6_31g_d_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/VACUUM Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_VACUUM_Unknown = "wB97X-V/6-31g*/VACUUM Unknown"
    wB97X_V_6_31g_d_PCM_WATER__Single_Point = "wB97X-V/6-31g*/PCM(WATER) Single Point"
    wB97X_V_6_31g_d_PCM_WATER__Force = "wB97X-V/6-31g*/PCM(WATER) Force"
    wB97X_V_6_31g_d_PCM_WATER__Geometry_Optimization = "wB97X-V/6-31g*/PCM(WATER) Geometry Optimization"
    wB97X_V_6_31g_d_PCM_WATER__Frequency_Analysis = "wB97X-V/6-31g*/PCM(WATER) Frequency Analysis"
    wB97X_V_6_31g_d_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_WATER__Unknown = "wB97X-V/6-31g*/PCM(WATER) Unknown"
    wB97X_V_6_31g_d_PCM_THF__Single_Point = "wB97X-V/6-31g*/PCM(THF) Single Point"
    wB97X_V_6_31g_d_PCM_THF__Force = "wB97X-V/6-31g*/PCM(THF) Force"
    wB97X_V_6_31g_d_PCM_THF__Geometry_Optimization = "wB97X-V/6-31g*/PCM(THF) Geometry Optimization"
    wB97X_V_6_31g_d_PCM_THF__Frequency_Analysis = "wB97X-V/6-31g*/PCM(THF) Frequency Analysis"
    wB97X_V_6_31g_d_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_THF__Unknown = "wB97X-V/6-31g*/PCM(THF) Unknown"
    wB97X_V_6_31g_d_PCM_G2__Single_Point = "wB97X-V/6-31g*/PCM(G2) Single Point"
    wB97X_V_6_31g_d_PCM_G2__Force = "wB97X-V/6-31g*/PCM(G2) Force"
    wB97X_V_6_31g_d_PCM_G2__Geometry_Optimization = "wB97X-V/6-31g*/PCM(G2) Geometry Optimization"
    wB97X_V_6_31g_d_PCM_G2__Frequency_Analysis = "wB97X-V/6-31g*/PCM(G2) Frequency Analysis"
    wB97X_V_6_31g_d_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_G2__Unknown = "wB97X-V/6-31g*/PCM(G2) Unknown"
    wB97X_V_6_31g_d_PCM_DME__Single_Point = "wB97X-V/6-31g*/PCM(DME) Single Point"
    wB97X_V_6_31g_d_PCM_DME__Force = "wB97X-V/6-31g*/PCM(DME) Force"
    wB97X_V_6_31g_d_PCM_DME__Geometry_Optimization = "wB97X-V/6-31g*/PCM(DME) Geometry Optimization"
    wB97X_V_6_31g_d_PCM_DME__Frequency_Analysis = "wB97X-V/6-31g*/PCM(DME) Frequency Analysis"
    wB97X_V_6_31g_d_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_DME__Unknown = "wB97X-V/6-31g*/PCM(DME) Unknown"
    wB97X_V_6_31g_d_PCM_EC__Single_Point = "wB97X-V/6-31g*/PCM(EC) Single Point"
    wB97X_V_6_31g_d_PCM_EC__Force = "wB97X-V/6-31g*/PCM(EC) Force"
    wB97X_V_6_31g_d_PCM_EC__Geometry_Optimization = "wB97X-V/6-31g*/PCM(EC) Geometry Optimization"
    wB97X_V_6_31g_d_PCM_EC__Frequency_Analysis = "wB97X-V/6-31g*/PCM(EC) Frequency Analysis"
    wB97X_V_6_31g_d_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_EC__Unknown = "wB97X-V/6-31g*/PCM(EC) Unknown"
    wB97X_V_6_31g_d_PCM_EC_EMC__Single_Point = "wB97X-V/6-31g*/PCM(EC/EMC) Single Point"
    wB97X_V_6_31g_d_PCM_EC_EMC__Force = "wB97X-V/6-31g*/PCM(EC/EMC) Force"
    wB97X_V_6_31g_d_PCM_EC_EMC__Geometry_Optimization = "wB97X-V/6-31g*/PCM(EC/EMC) Geometry Optimization"
    wB97X_V_6_31g_d_PCM_EC_EMC__Frequency_Analysis = "wB97X-V/6-31g*/PCM(EC/EMC) Frequency Analysis"
    wB97X_V_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_EC_EMC__Unknown = "wB97X-V/6-31g*/PCM(EC/EMC) Unknown"
    wB97X_V_6_31g_d_PCM_PC__Single_Point = "wB97X-V/6-31g*/PCM(PC) Single Point"
    wB97X_V_6_31g_d_PCM_PC__Force = "wB97X-V/6-31g*/PCM(PC) Force"
    wB97X_V_6_31g_d_PCM_PC__Geometry_Optimization = "wB97X-V/6-31g*/PCM(PC) Geometry Optimization"
    wB97X_V_6_31g_d_PCM_PC__Frequency_Analysis = "wB97X-V/6-31g*/PCM(PC) Frequency Analysis"
    wB97X_V_6_31g_d_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_PCM_PC__Unknown = "wB97X-V/6-31g*/PCM(PC) Unknown"
    wB97X_V_6_31g_d_ISOSVP_WATER__Single_Point = "wB97X-V/6-31g*/ISOSVP(WATER) Single Point"
    wB97X_V_6_31g_d_ISOSVP_WATER__Force = "wB97X-V/6-31g*/ISOSVP(WATER) Force"
    wB97X_V_6_31g_d_ISOSVP_WATER__Geometry_Optimization = "wB97X-V/6-31g*/ISOSVP(WATER) Geometry Optimization"
    wB97X_V_6_31g_d_ISOSVP_WATER__Frequency_Analysis = "wB97X-V/6-31g*/ISOSVP(WATER) Frequency Analysis"
    wB97X_V_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_WATER__Unknown = "wB97X-V/6-31g*/ISOSVP(WATER) Unknown"
    wB97X_V_6_31g_d_ISOSVP_THF__Single_Point = "wB97X-V/6-31g*/ISOSVP(THF) Single Point"
    wB97X_V_6_31g_d_ISOSVP_THF__Force = "wB97X-V/6-31g*/ISOSVP(THF) Force"
    wB97X_V_6_31g_d_ISOSVP_THF__Geometry_Optimization = "wB97X-V/6-31g*/ISOSVP(THF) Geometry Optimization"
    wB97X_V_6_31g_d_ISOSVP_THF__Frequency_Analysis = "wB97X-V/6-31g*/ISOSVP(THF) Frequency Analysis"
    wB97X_V_6_31g_d_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_THF__Unknown = "wB97X-V/6-31g*/ISOSVP(THF) Unknown"
    wB97X_V_6_31g_d_ISOSVP_G2__Single_Point = "wB97X-V/6-31g*/ISOSVP(G2) Single Point"
    wB97X_V_6_31g_d_ISOSVP_G2__Force = "wB97X-V/6-31g*/ISOSVP(G2) Force"
    wB97X_V_6_31g_d_ISOSVP_G2__Geometry_Optimization = "wB97X-V/6-31g*/ISOSVP(G2) Geometry Optimization"
    wB97X_V_6_31g_d_ISOSVP_G2__Frequency_Analysis = "wB97X-V/6-31g*/ISOSVP(G2) Frequency Analysis"
    wB97X_V_6_31g_d_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_G2__Unknown = "wB97X-V/6-31g*/ISOSVP(G2) Unknown"
    wB97X_V_6_31g_d_ISOSVP_DME__Single_Point = "wB97X-V/6-31g*/ISOSVP(DME) Single Point"
    wB97X_V_6_31g_d_ISOSVP_DME__Force = "wB97X-V/6-31g*/ISOSVP(DME) Force"
    wB97X_V_6_31g_d_ISOSVP_DME__Geometry_Optimization = "wB97X-V/6-31g*/ISOSVP(DME) Geometry Optimization"
    wB97X_V_6_31g_d_ISOSVP_DME__Frequency_Analysis = "wB97X-V/6-31g*/ISOSVP(DME) Frequency Analysis"
    wB97X_V_6_31g_d_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_DME__Unknown = "wB97X-V/6-31g*/ISOSVP(DME) Unknown"
    wB97X_V_6_31g_d_ISOSVP_EC__Single_Point = "wB97X-V/6-31g*/ISOSVP(EC) Single Point"
    wB97X_V_6_31g_d_ISOSVP_EC__Force = "wB97X-V/6-31g*/ISOSVP(EC) Force"
    wB97X_V_6_31g_d_ISOSVP_EC__Geometry_Optimization = "wB97X-V/6-31g*/ISOSVP(EC) Geometry Optimization"
    wB97X_V_6_31g_d_ISOSVP_EC__Frequency_Analysis = "wB97X-V/6-31g*/ISOSVP(EC) Frequency Analysis"
    wB97X_V_6_31g_d_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_EC__Unknown = "wB97X-V/6-31g*/ISOSVP(EC) Unknown"
    wB97X_V_6_31g_d_ISOSVP_EC_EMC__Single_Point = "wB97X-V/6-31g*/ISOSVP(EC/EMC) Single Point"
    wB97X_V_6_31g_d_ISOSVP_EC_EMC__Force = "wB97X-V/6-31g*/ISOSVP(EC/EMC) Force"
    wB97X_V_6_31g_d_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-V/6-31g*/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_V_6_31g_d_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-V/6-31g*/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_V_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_EC_EMC__Unknown = "wB97X-V/6-31g*/ISOSVP(EC/EMC) Unknown"
    wB97X_V_6_31g_d_ISOSVP_PC__Single_Point = "wB97X-V/6-31g*/ISOSVP(PC) Single Point"
    wB97X_V_6_31g_d_ISOSVP_PC__Force = "wB97X-V/6-31g*/ISOSVP(PC) Force"
    wB97X_V_6_31g_d_ISOSVP_PC__Geometry_Optimization = "wB97X-V/6-31g*/ISOSVP(PC) Geometry Optimization"
    wB97X_V_6_31g_d_ISOSVP_PC__Frequency_Analysis = "wB97X-V/6-31g*/ISOSVP(PC) Frequency Analysis"
    wB97X_V_6_31g_d_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_ISOSVP_PC__Unknown = "wB97X-V/6-31g*/ISOSVP(PC) Unknown"
    wB97X_V_6_31g_d_SMD_WATER__Single_Point = "wB97X-V/6-31g*/SMD(WATER) Single Point"
    wB97X_V_6_31g_d_SMD_WATER__Force = "wB97X-V/6-31g*/SMD(WATER) Force"
    wB97X_V_6_31g_d_SMD_WATER__Geometry_Optimization = "wB97X-V/6-31g*/SMD(WATER) Geometry Optimization"
    wB97X_V_6_31g_d_SMD_WATER__Frequency_Analysis = "wB97X-V/6-31g*/SMD(WATER) Frequency Analysis"
    wB97X_V_6_31g_d_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_WATER__Unknown = "wB97X-V/6-31g*/SMD(WATER) Unknown"
    wB97X_V_6_31g_d_SMD_THF__Single_Point = "wB97X-V/6-31g*/SMD(THF) Single Point"
    wB97X_V_6_31g_d_SMD_THF__Force = "wB97X-V/6-31g*/SMD(THF) Force"
    wB97X_V_6_31g_d_SMD_THF__Geometry_Optimization = "wB97X-V/6-31g*/SMD(THF) Geometry Optimization"
    wB97X_V_6_31g_d_SMD_THF__Frequency_Analysis = "wB97X-V/6-31g*/SMD(THF) Frequency Analysis"
    wB97X_V_6_31g_d_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_THF__Unknown = "wB97X-V/6-31g*/SMD(THF) Unknown"
    wB97X_V_6_31g_d_SMD_G2__Single_Point = "wB97X-V/6-31g*/SMD(G2) Single Point"
    wB97X_V_6_31g_d_SMD_G2__Force = "wB97X-V/6-31g*/SMD(G2) Force"
    wB97X_V_6_31g_d_SMD_G2__Geometry_Optimization = "wB97X-V/6-31g*/SMD(G2) Geometry Optimization"
    wB97X_V_6_31g_d_SMD_G2__Frequency_Analysis = "wB97X-V/6-31g*/SMD(G2) Frequency Analysis"
    wB97X_V_6_31g_d_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_G2__Unknown = "wB97X-V/6-31g*/SMD(G2) Unknown"
    wB97X_V_6_31g_d_SMD_DME__Single_Point = "wB97X-V/6-31g*/SMD(DME) Single Point"
    wB97X_V_6_31g_d_SMD_DME__Force = "wB97X-V/6-31g*/SMD(DME) Force"
    wB97X_V_6_31g_d_SMD_DME__Geometry_Optimization = "wB97X-V/6-31g*/SMD(DME) Geometry Optimization"
    wB97X_V_6_31g_d_SMD_DME__Frequency_Analysis = "wB97X-V/6-31g*/SMD(DME) Frequency Analysis"
    wB97X_V_6_31g_d_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_DME__Unknown = "wB97X-V/6-31g*/SMD(DME) Unknown"
    wB97X_V_6_31g_d_SMD_EC__Single_Point = "wB97X-V/6-31g*/SMD(EC) Single Point"
    wB97X_V_6_31g_d_SMD_EC__Force = "wB97X-V/6-31g*/SMD(EC) Force"
    wB97X_V_6_31g_d_SMD_EC__Geometry_Optimization = "wB97X-V/6-31g*/SMD(EC) Geometry Optimization"
    wB97X_V_6_31g_d_SMD_EC__Frequency_Analysis = "wB97X-V/6-31g*/SMD(EC) Frequency Analysis"
    wB97X_V_6_31g_d_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_EC__Unknown = "wB97X-V/6-31g*/SMD(EC) Unknown"
    wB97X_V_6_31g_d_SMD_EC_EMC__Single_Point = "wB97X-V/6-31g*/SMD(EC/EMC) Single Point"
    wB97X_V_6_31g_d_SMD_EC_EMC__Force = "wB97X-V/6-31g*/SMD(EC/EMC) Force"
    wB97X_V_6_31g_d_SMD_EC_EMC__Geometry_Optimization = "wB97X-V/6-31g*/SMD(EC/EMC) Geometry Optimization"
    wB97X_V_6_31g_d_SMD_EC_EMC__Frequency_Analysis = "wB97X-V/6-31g*/SMD(EC/EMC) Frequency Analysis"
    wB97X_V_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_EC_EMC__Unknown = "wB97X-V/6-31g*/SMD(EC/EMC) Unknown"
    wB97X_V_6_31g_d_SMD_PC__Single_Point = "wB97X-V/6-31g*/SMD(PC) Single Point"
    wB97X_V_6_31g_d_SMD_PC__Force = "wB97X-V/6-31g*/SMD(PC) Force"
    wB97X_V_6_31g_d_SMD_PC__Geometry_Optimization = "wB97X-V/6-31g*/SMD(PC) Geometry Optimization"
    wB97X_V_6_31g_d_SMD_PC__Frequency_Analysis = "wB97X-V/6-31g*/SMD(PC) Frequency Analysis"
    wB97X_V_6_31g_d_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_SMD_PC__Unknown = "wB97X-V/6-31g*/SMD(PC) Unknown"
    wB97X_V_6_31g_d_CMIRS_WATER__Single_Point = "wB97X-V/6-31g*/CMIRS(WATER) Single Point"
    wB97X_V_6_31g_d_CMIRS_WATER__Force = "wB97X-V/6-31g*/CMIRS(WATER) Force"
    wB97X_V_6_31g_d_CMIRS_WATER__Geometry_Optimization = "wB97X-V/6-31g*/CMIRS(WATER) Geometry Optimization"
    wB97X_V_6_31g_d_CMIRS_WATER__Frequency_Analysis = "wB97X-V/6-31g*/CMIRS(WATER) Frequency Analysis"
    wB97X_V_6_31g_d_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_WATER__Unknown = "wB97X-V/6-31g*/CMIRS(WATER) Unknown"
    wB97X_V_6_31g_d_CMIRS_THF__Single_Point = "wB97X-V/6-31g*/CMIRS(THF) Single Point"
    wB97X_V_6_31g_d_CMIRS_THF__Force = "wB97X-V/6-31g*/CMIRS(THF) Force"
    wB97X_V_6_31g_d_CMIRS_THF__Geometry_Optimization = "wB97X-V/6-31g*/CMIRS(THF) Geometry Optimization"
    wB97X_V_6_31g_d_CMIRS_THF__Frequency_Analysis = "wB97X-V/6-31g*/CMIRS(THF) Frequency Analysis"
    wB97X_V_6_31g_d_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_THF__Unknown = "wB97X-V/6-31g*/CMIRS(THF) Unknown"
    wB97X_V_6_31g_d_CMIRS_G2__Single_Point = "wB97X-V/6-31g*/CMIRS(G2) Single Point"
    wB97X_V_6_31g_d_CMIRS_G2__Force = "wB97X-V/6-31g*/CMIRS(G2) Force"
    wB97X_V_6_31g_d_CMIRS_G2__Geometry_Optimization = "wB97X-V/6-31g*/CMIRS(G2) Geometry Optimization"
    wB97X_V_6_31g_d_CMIRS_G2__Frequency_Analysis = "wB97X-V/6-31g*/CMIRS(G2) Frequency Analysis"
    wB97X_V_6_31g_d_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_G2__Unknown = "wB97X-V/6-31g*/CMIRS(G2) Unknown"
    wB97X_V_6_31g_d_CMIRS_DME__Single_Point = "wB97X-V/6-31g*/CMIRS(DME) Single Point"
    wB97X_V_6_31g_d_CMIRS_DME__Force = "wB97X-V/6-31g*/CMIRS(DME) Force"
    wB97X_V_6_31g_d_CMIRS_DME__Geometry_Optimization = "wB97X-V/6-31g*/CMIRS(DME) Geometry Optimization"
    wB97X_V_6_31g_d_CMIRS_DME__Frequency_Analysis = "wB97X-V/6-31g*/CMIRS(DME) Frequency Analysis"
    wB97X_V_6_31g_d_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_DME__Unknown = "wB97X-V/6-31g*/CMIRS(DME) Unknown"
    wB97X_V_6_31g_d_CMIRS_EC__Single_Point = "wB97X-V/6-31g*/CMIRS(EC) Single Point"
    wB97X_V_6_31g_d_CMIRS_EC__Force = "wB97X-V/6-31g*/CMIRS(EC) Force"
    wB97X_V_6_31g_d_CMIRS_EC__Geometry_Optimization = "wB97X-V/6-31g*/CMIRS(EC) Geometry Optimization"
    wB97X_V_6_31g_d_CMIRS_EC__Frequency_Analysis = "wB97X-V/6-31g*/CMIRS(EC) Frequency Analysis"
    wB97X_V_6_31g_d_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_EC__Unknown = "wB97X-V/6-31g*/CMIRS(EC) Unknown"
    wB97X_V_6_31g_d_CMIRS_EC_EMC__Single_Point = "wB97X-V/6-31g*/CMIRS(EC/EMC) Single Point"
    wB97X_V_6_31g_d_CMIRS_EC_EMC__Force = "wB97X-V/6-31g*/CMIRS(EC/EMC) Force"
    wB97X_V_6_31g_d_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-V/6-31g*/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_V_6_31g_d_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-V/6-31g*/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_V_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_EC_EMC__Unknown = "wB97X-V/6-31g*/CMIRS(EC/EMC) Unknown"
    wB97X_V_6_31g_d_CMIRS_PC__Single_Point = "wB97X-V/6-31g*/CMIRS(PC) Single Point"
    wB97X_V_6_31g_d_CMIRS_PC__Force = "wB97X-V/6-31g*/CMIRS(PC) Force"
    wB97X_V_6_31g_d_CMIRS_PC__Geometry_Optimization = "wB97X-V/6-31g*/CMIRS(PC) Geometry Optimization"
    wB97X_V_6_31g_d_CMIRS_PC__Frequency_Analysis = "wB97X-V/6-31g*/CMIRS(PC) Frequency Analysis"
    wB97X_V_6_31g_d_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/6-31g*/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_6_31g_d_CMIRS_PC__Unknown = "wB97X-V/6-31g*/CMIRS(PC) Unknown"
    wB97X_V_def2_SVPD_VACUUM_Single_Point = "wB97X-V/def2-SVPD/VACUUM Single Point"
    wB97X_V_def2_SVPD_VACUUM_Force = "wB97X-V/def2-SVPD/VACUUM Force"
    wB97X_V_def2_SVPD_VACUUM_Geometry_Optimization = "wB97X-V/def2-SVPD/VACUUM Geometry Optimization"
    wB97X_V_def2_SVPD_VACUUM_Frequency_Analysis = "wB97X-V/def2-SVPD/VACUUM Frequency Analysis"
    wB97X_V_def2_SVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_VACUUM_Unknown = "wB97X-V/def2-SVPD/VACUUM Unknown"
    wB97X_V_def2_SVPD_PCM_WATER__Single_Point = "wB97X-V/def2-SVPD/PCM(WATER) Single Point"
    wB97X_V_def2_SVPD_PCM_WATER__Force = "wB97X-V/def2-SVPD/PCM(WATER) Force"
    wB97X_V_def2_SVPD_PCM_WATER__Geometry_Optimization = "wB97X-V/def2-SVPD/PCM(WATER) Geometry Optimization"
    wB97X_V_def2_SVPD_PCM_WATER__Frequency_Analysis = "wB97X-V/def2-SVPD/PCM(WATER) Frequency Analysis"
    wB97X_V_def2_SVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_WATER__Unknown = "wB97X-V/def2-SVPD/PCM(WATER) Unknown"
    wB97X_V_def2_SVPD_PCM_THF__Single_Point = "wB97X-V/def2-SVPD/PCM(THF) Single Point"
    wB97X_V_def2_SVPD_PCM_THF__Force = "wB97X-V/def2-SVPD/PCM(THF) Force"
    wB97X_V_def2_SVPD_PCM_THF__Geometry_Optimization = "wB97X-V/def2-SVPD/PCM(THF) Geometry Optimization"
    wB97X_V_def2_SVPD_PCM_THF__Frequency_Analysis = "wB97X-V/def2-SVPD/PCM(THF) Frequency Analysis"
    wB97X_V_def2_SVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_THF__Unknown = "wB97X-V/def2-SVPD/PCM(THF) Unknown"
    wB97X_V_def2_SVPD_PCM_G2__Single_Point = "wB97X-V/def2-SVPD/PCM(G2) Single Point"
    wB97X_V_def2_SVPD_PCM_G2__Force = "wB97X-V/def2-SVPD/PCM(G2) Force"
    wB97X_V_def2_SVPD_PCM_G2__Geometry_Optimization = "wB97X-V/def2-SVPD/PCM(G2) Geometry Optimization"
    wB97X_V_def2_SVPD_PCM_G2__Frequency_Analysis = "wB97X-V/def2-SVPD/PCM(G2) Frequency Analysis"
    wB97X_V_def2_SVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_G2__Unknown = "wB97X-V/def2-SVPD/PCM(G2) Unknown"
    wB97X_V_def2_SVPD_PCM_DME__Single_Point = "wB97X-V/def2-SVPD/PCM(DME) Single Point"
    wB97X_V_def2_SVPD_PCM_DME__Force = "wB97X-V/def2-SVPD/PCM(DME) Force"
    wB97X_V_def2_SVPD_PCM_DME__Geometry_Optimization = "wB97X-V/def2-SVPD/PCM(DME) Geometry Optimization"
    wB97X_V_def2_SVPD_PCM_DME__Frequency_Analysis = "wB97X-V/def2-SVPD/PCM(DME) Frequency Analysis"
    wB97X_V_def2_SVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_DME__Unknown = "wB97X-V/def2-SVPD/PCM(DME) Unknown"
    wB97X_V_def2_SVPD_PCM_EC__Single_Point = "wB97X-V/def2-SVPD/PCM(EC) Single Point"
    wB97X_V_def2_SVPD_PCM_EC__Force = "wB97X-V/def2-SVPD/PCM(EC) Force"
    wB97X_V_def2_SVPD_PCM_EC__Geometry_Optimization = "wB97X-V/def2-SVPD/PCM(EC) Geometry Optimization"
    wB97X_V_def2_SVPD_PCM_EC__Frequency_Analysis = "wB97X-V/def2-SVPD/PCM(EC) Frequency Analysis"
    wB97X_V_def2_SVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_EC__Unknown = "wB97X-V/def2-SVPD/PCM(EC) Unknown"
    wB97X_V_def2_SVPD_PCM_EC_EMC__Single_Point = "wB97X-V/def2-SVPD/PCM(EC/EMC) Single Point"
    wB97X_V_def2_SVPD_PCM_EC_EMC__Force = "wB97X-V/def2-SVPD/PCM(EC/EMC) Force"
    wB97X_V_def2_SVPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-V/def2-SVPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_V_def2_SVPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-V/def2-SVPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_V_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_EC_EMC__Unknown = "wB97X-V/def2-SVPD/PCM(EC/EMC) Unknown"
    wB97X_V_def2_SVPD_PCM_PC__Single_Point = "wB97X-V/def2-SVPD/PCM(PC) Single Point"
    wB97X_V_def2_SVPD_PCM_PC__Force = "wB97X-V/def2-SVPD/PCM(PC) Force"
    wB97X_V_def2_SVPD_PCM_PC__Geometry_Optimization = "wB97X-V/def2-SVPD/PCM(PC) Geometry Optimization"
    wB97X_V_def2_SVPD_PCM_PC__Frequency_Analysis = "wB97X-V/def2-SVPD/PCM(PC) Frequency Analysis"
    wB97X_V_def2_SVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_PCM_PC__Unknown = "wB97X-V/def2-SVPD/PCM(PC) Unknown"
    wB97X_V_def2_SVPD_ISOSVP_WATER__Single_Point = "wB97X-V/def2-SVPD/ISOSVP(WATER) Single Point"
    wB97X_V_def2_SVPD_ISOSVP_WATER__Force = "wB97X-V/def2-SVPD/ISOSVP(WATER) Force"
    wB97X_V_def2_SVPD_ISOSVP_WATER__Geometry_Optimization = "wB97X-V/def2-SVPD/ISOSVP(WATER) Geometry Optimization"
    wB97X_V_def2_SVPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-V/def2-SVPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_V_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_WATER__Unknown = "wB97X-V/def2-SVPD/ISOSVP(WATER) Unknown"
    wB97X_V_def2_SVPD_ISOSVP_THF__Single_Point = "wB97X-V/def2-SVPD/ISOSVP(THF) Single Point"
    wB97X_V_def2_SVPD_ISOSVP_THF__Force = "wB97X-V/def2-SVPD/ISOSVP(THF) Force"
    wB97X_V_def2_SVPD_ISOSVP_THF__Geometry_Optimization = "wB97X-V/def2-SVPD/ISOSVP(THF) Geometry Optimization"
    wB97X_V_def2_SVPD_ISOSVP_THF__Frequency_Analysis = "wB97X-V/def2-SVPD/ISOSVP(THF) Frequency Analysis"
    wB97X_V_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_THF__Unknown = "wB97X-V/def2-SVPD/ISOSVP(THF) Unknown"
    wB97X_V_def2_SVPD_ISOSVP_G2__Single_Point = "wB97X-V/def2-SVPD/ISOSVP(G2) Single Point"
    wB97X_V_def2_SVPD_ISOSVP_G2__Force = "wB97X-V/def2-SVPD/ISOSVP(G2) Force"
    wB97X_V_def2_SVPD_ISOSVP_G2__Geometry_Optimization = "wB97X-V/def2-SVPD/ISOSVP(G2) Geometry Optimization"
    wB97X_V_def2_SVPD_ISOSVP_G2__Frequency_Analysis = "wB97X-V/def2-SVPD/ISOSVP(G2) Frequency Analysis"
    wB97X_V_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_G2__Unknown = "wB97X-V/def2-SVPD/ISOSVP(G2) Unknown"
    wB97X_V_def2_SVPD_ISOSVP_DME__Single_Point = "wB97X-V/def2-SVPD/ISOSVP(DME) Single Point"
    wB97X_V_def2_SVPD_ISOSVP_DME__Force = "wB97X-V/def2-SVPD/ISOSVP(DME) Force"
    wB97X_V_def2_SVPD_ISOSVP_DME__Geometry_Optimization = "wB97X-V/def2-SVPD/ISOSVP(DME) Geometry Optimization"
    wB97X_V_def2_SVPD_ISOSVP_DME__Frequency_Analysis = "wB97X-V/def2-SVPD/ISOSVP(DME) Frequency Analysis"
    wB97X_V_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_DME__Unknown = "wB97X-V/def2-SVPD/ISOSVP(DME) Unknown"
    wB97X_V_def2_SVPD_ISOSVP_EC__Single_Point = "wB97X-V/def2-SVPD/ISOSVP(EC) Single Point"
    wB97X_V_def2_SVPD_ISOSVP_EC__Force = "wB97X-V/def2-SVPD/ISOSVP(EC) Force"
    wB97X_V_def2_SVPD_ISOSVP_EC__Geometry_Optimization = "wB97X-V/def2-SVPD/ISOSVP(EC) Geometry Optimization"
    wB97X_V_def2_SVPD_ISOSVP_EC__Frequency_Analysis = "wB97X-V/def2-SVPD/ISOSVP(EC) Frequency Analysis"
    wB97X_V_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_EC__Unknown = "wB97X-V/def2-SVPD/ISOSVP(EC) Unknown"
    wB97X_V_def2_SVPD_ISOSVP_EC_EMC__Single_Point = "wB97X-V/def2-SVPD/ISOSVP(EC/EMC) Single Point"
    wB97X_V_def2_SVPD_ISOSVP_EC_EMC__Force = "wB97X-V/def2-SVPD/ISOSVP(EC/EMC) Force"
    wB97X_V_def2_SVPD_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-V/def2-SVPD/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_V_def2_SVPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-V/def2-SVPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_V_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_EC_EMC__Unknown = "wB97X-V/def2-SVPD/ISOSVP(EC/EMC) Unknown"
    wB97X_V_def2_SVPD_ISOSVP_PC__Single_Point = "wB97X-V/def2-SVPD/ISOSVP(PC) Single Point"
    wB97X_V_def2_SVPD_ISOSVP_PC__Force = "wB97X-V/def2-SVPD/ISOSVP(PC) Force"
    wB97X_V_def2_SVPD_ISOSVP_PC__Geometry_Optimization = "wB97X-V/def2-SVPD/ISOSVP(PC) Geometry Optimization"
    wB97X_V_def2_SVPD_ISOSVP_PC__Frequency_Analysis = "wB97X-V/def2-SVPD/ISOSVP(PC) Frequency Analysis"
    wB97X_V_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_ISOSVP_PC__Unknown = "wB97X-V/def2-SVPD/ISOSVP(PC) Unknown"
    wB97X_V_def2_SVPD_SMD_WATER__Single_Point = "wB97X-V/def2-SVPD/SMD(WATER) Single Point"
    wB97X_V_def2_SVPD_SMD_WATER__Force = "wB97X-V/def2-SVPD/SMD(WATER) Force"
    wB97X_V_def2_SVPD_SMD_WATER__Geometry_Optimization = "wB97X-V/def2-SVPD/SMD(WATER) Geometry Optimization"
    wB97X_V_def2_SVPD_SMD_WATER__Frequency_Analysis = "wB97X-V/def2-SVPD/SMD(WATER) Frequency Analysis"
    wB97X_V_def2_SVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_WATER__Unknown = "wB97X-V/def2-SVPD/SMD(WATER) Unknown"
    wB97X_V_def2_SVPD_SMD_THF__Single_Point = "wB97X-V/def2-SVPD/SMD(THF) Single Point"
    wB97X_V_def2_SVPD_SMD_THF__Force = "wB97X-V/def2-SVPD/SMD(THF) Force"
    wB97X_V_def2_SVPD_SMD_THF__Geometry_Optimization = "wB97X-V/def2-SVPD/SMD(THF) Geometry Optimization"
    wB97X_V_def2_SVPD_SMD_THF__Frequency_Analysis = "wB97X-V/def2-SVPD/SMD(THF) Frequency Analysis"
    wB97X_V_def2_SVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_THF__Unknown = "wB97X-V/def2-SVPD/SMD(THF) Unknown"
    wB97X_V_def2_SVPD_SMD_G2__Single_Point = "wB97X-V/def2-SVPD/SMD(G2) Single Point"
    wB97X_V_def2_SVPD_SMD_G2__Force = "wB97X-V/def2-SVPD/SMD(G2) Force"
    wB97X_V_def2_SVPD_SMD_G2__Geometry_Optimization = "wB97X-V/def2-SVPD/SMD(G2) Geometry Optimization"
    wB97X_V_def2_SVPD_SMD_G2__Frequency_Analysis = "wB97X-V/def2-SVPD/SMD(G2) Frequency Analysis"
    wB97X_V_def2_SVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_G2__Unknown = "wB97X-V/def2-SVPD/SMD(G2) Unknown"
    wB97X_V_def2_SVPD_SMD_DME__Single_Point = "wB97X-V/def2-SVPD/SMD(DME) Single Point"
    wB97X_V_def2_SVPD_SMD_DME__Force = "wB97X-V/def2-SVPD/SMD(DME) Force"
    wB97X_V_def2_SVPD_SMD_DME__Geometry_Optimization = "wB97X-V/def2-SVPD/SMD(DME) Geometry Optimization"
    wB97X_V_def2_SVPD_SMD_DME__Frequency_Analysis = "wB97X-V/def2-SVPD/SMD(DME) Frequency Analysis"
    wB97X_V_def2_SVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_DME__Unknown = "wB97X-V/def2-SVPD/SMD(DME) Unknown"
    wB97X_V_def2_SVPD_SMD_EC__Single_Point = "wB97X-V/def2-SVPD/SMD(EC) Single Point"
    wB97X_V_def2_SVPD_SMD_EC__Force = "wB97X-V/def2-SVPD/SMD(EC) Force"
    wB97X_V_def2_SVPD_SMD_EC__Geometry_Optimization = "wB97X-V/def2-SVPD/SMD(EC) Geometry Optimization"
    wB97X_V_def2_SVPD_SMD_EC__Frequency_Analysis = "wB97X-V/def2-SVPD/SMD(EC) Frequency Analysis"
    wB97X_V_def2_SVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_EC__Unknown = "wB97X-V/def2-SVPD/SMD(EC) Unknown"
    wB97X_V_def2_SVPD_SMD_EC_EMC__Single_Point = "wB97X-V/def2-SVPD/SMD(EC/EMC) Single Point"
    wB97X_V_def2_SVPD_SMD_EC_EMC__Force = "wB97X-V/def2-SVPD/SMD(EC/EMC) Force"
    wB97X_V_def2_SVPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-V/def2-SVPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_V_def2_SVPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-V/def2-SVPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_V_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_EC_EMC__Unknown = "wB97X-V/def2-SVPD/SMD(EC/EMC) Unknown"
    wB97X_V_def2_SVPD_SMD_PC__Single_Point = "wB97X-V/def2-SVPD/SMD(PC) Single Point"
    wB97X_V_def2_SVPD_SMD_PC__Force = "wB97X-V/def2-SVPD/SMD(PC) Force"
    wB97X_V_def2_SVPD_SMD_PC__Geometry_Optimization = "wB97X-V/def2-SVPD/SMD(PC) Geometry Optimization"
    wB97X_V_def2_SVPD_SMD_PC__Frequency_Analysis = "wB97X-V/def2-SVPD/SMD(PC) Frequency Analysis"
    wB97X_V_def2_SVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_SMD_PC__Unknown = "wB97X-V/def2-SVPD/SMD(PC) Unknown"
    wB97X_V_def2_SVPD_CMIRS_WATER__Single_Point = "wB97X-V/def2-SVPD/CMIRS(WATER) Single Point"
    wB97X_V_def2_SVPD_CMIRS_WATER__Force = "wB97X-V/def2-SVPD/CMIRS(WATER) Force"
    wB97X_V_def2_SVPD_CMIRS_WATER__Geometry_Optimization = "wB97X-V/def2-SVPD/CMIRS(WATER) Geometry Optimization"
    wB97X_V_def2_SVPD_CMIRS_WATER__Frequency_Analysis = "wB97X-V/def2-SVPD/CMIRS(WATER) Frequency Analysis"
    wB97X_V_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_WATER__Unknown = "wB97X-V/def2-SVPD/CMIRS(WATER) Unknown"
    wB97X_V_def2_SVPD_CMIRS_THF__Single_Point = "wB97X-V/def2-SVPD/CMIRS(THF) Single Point"
    wB97X_V_def2_SVPD_CMIRS_THF__Force = "wB97X-V/def2-SVPD/CMIRS(THF) Force"
    wB97X_V_def2_SVPD_CMIRS_THF__Geometry_Optimization = "wB97X-V/def2-SVPD/CMIRS(THF) Geometry Optimization"
    wB97X_V_def2_SVPD_CMIRS_THF__Frequency_Analysis = "wB97X-V/def2-SVPD/CMIRS(THF) Frequency Analysis"
    wB97X_V_def2_SVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_THF__Unknown = "wB97X-V/def2-SVPD/CMIRS(THF) Unknown"
    wB97X_V_def2_SVPD_CMIRS_G2__Single_Point = "wB97X-V/def2-SVPD/CMIRS(G2) Single Point"
    wB97X_V_def2_SVPD_CMIRS_G2__Force = "wB97X-V/def2-SVPD/CMIRS(G2) Force"
    wB97X_V_def2_SVPD_CMIRS_G2__Geometry_Optimization = "wB97X-V/def2-SVPD/CMIRS(G2) Geometry Optimization"
    wB97X_V_def2_SVPD_CMIRS_G2__Frequency_Analysis = "wB97X-V/def2-SVPD/CMIRS(G2) Frequency Analysis"
    wB97X_V_def2_SVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_G2__Unknown = "wB97X-V/def2-SVPD/CMIRS(G2) Unknown"
    wB97X_V_def2_SVPD_CMIRS_DME__Single_Point = "wB97X-V/def2-SVPD/CMIRS(DME) Single Point"
    wB97X_V_def2_SVPD_CMIRS_DME__Force = "wB97X-V/def2-SVPD/CMIRS(DME) Force"
    wB97X_V_def2_SVPD_CMIRS_DME__Geometry_Optimization = "wB97X-V/def2-SVPD/CMIRS(DME) Geometry Optimization"
    wB97X_V_def2_SVPD_CMIRS_DME__Frequency_Analysis = "wB97X-V/def2-SVPD/CMIRS(DME) Frequency Analysis"
    wB97X_V_def2_SVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_DME__Unknown = "wB97X-V/def2-SVPD/CMIRS(DME) Unknown"
    wB97X_V_def2_SVPD_CMIRS_EC__Single_Point = "wB97X-V/def2-SVPD/CMIRS(EC) Single Point"
    wB97X_V_def2_SVPD_CMIRS_EC__Force = "wB97X-V/def2-SVPD/CMIRS(EC) Force"
    wB97X_V_def2_SVPD_CMIRS_EC__Geometry_Optimization = "wB97X-V/def2-SVPD/CMIRS(EC) Geometry Optimization"
    wB97X_V_def2_SVPD_CMIRS_EC__Frequency_Analysis = "wB97X-V/def2-SVPD/CMIRS(EC) Frequency Analysis"
    wB97X_V_def2_SVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_EC__Unknown = "wB97X-V/def2-SVPD/CMIRS(EC) Unknown"
    wB97X_V_def2_SVPD_CMIRS_EC_EMC__Single_Point = "wB97X-V/def2-SVPD/CMIRS(EC/EMC) Single Point"
    wB97X_V_def2_SVPD_CMIRS_EC_EMC__Force = "wB97X-V/def2-SVPD/CMIRS(EC/EMC) Force"
    wB97X_V_def2_SVPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-V/def2-SVPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_V_def2_SVPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-V/def2-SVPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_V_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_EC_EMC__Unknown = "wB97X-V/def2-SVPD/CMIRS(EC/EMC) Unknown"
    wB97X_V_def2_SVPD_CMIRS_PC__Single_Point = "wB97X-V/def2-SVPD/CMIRS(PC) Single Point"
    wB97X_V_def2_SVPD_CMIRS_PC__Force = "wB97X-V/def2-SVPD/CMIRS(PC) Force"
    wB97X_V_def2_SVPD_CMIRS_PC__Geometry_Optimization = "wB97X-V/def2-SVPD/CMIRS(PC) Geometry Optimization"
    wB97X_V_def2_SVPD_CMIRS_PC__Frequency_Analysis = "wB97X-V/def2-SVPD/CMIRS(PC) Frequency Analysis"
    wB97X_V_def2_SVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-SVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_SVPD_CMIRS_PC__Unknown = "wB97X-V/def2-SVPD/CMIRS(PC) Unknown"
    wB97X_V_def2_TZVP_VACUUM_Single_Point = "wB97X-V/def2-TZVP/VACUUM Single Point"
    wB97X_V_def2_TZVP_VACUUM_Force = "wB97X-V/def2-TZVP/VACUUM Force"
    wB97X_V_def2_TZVP_VACUUM_Geometry_Optimization = "wB97X-V/def2-TZVP/VACUUM Geometry Optimization"
    wB97X_V_def2_TZVP_VACUUM_Frequency_Analysis = "wB97X-V/def2-TZVP/VACUUM Frequency Analysis"
    wB97X_V_def2_TZVP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/VACUUM Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_VACUUM_Unknown = "wB97X-V/def2-TZVP/VACUUM Unknown"
    wB97X_V_def2_TZVP_PCM_WATER__Single_Point = "wB97X-V/def2-TZVP/PCM(WATER) Single Point"
    wB97X_V_def2_TZVP_PCM_WATER__Force = "wB97X-V/def2-TZVP/PCM(WATER) Force"
    wB97X_V_def2_TZVP_PCM_WATER__Geometry_Optimization = "wB97X-V/def2-TZVP/PCM(WATER) Geometry Optimization"
    wB97X_V_def2_TZVP_PCM_WATER__Frequency_Analysis = "wB97X-V/def2-TZVP/PCM(WATER) Frequency Analysis"
    wB97X_V_def2_TZVP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_WATER__Unknown = "wB97X-V/def2-TZVP/PCM(WATER) Unknown"
    wB97X_V_def2_TZVP_PCM_THF__Single_Point = "wB97X-V/def2-TZVP/PCM(THF) Single Point"
    wB97X_V_def2_TZVP_PCM_THF__Force = "wB97X-V/def2-TZVP/PCM(THF) Force"
    wB97X_V_def2_TZVP_PCM_THF__Geometry_Optimization = "wB97X-V/def2-TZVP/PCM(THF) Geometry Optimization"
    wB97X_V_def2_TZVP_PCM_THF__Frequency_Analysis = "wB97X-V/def2-TZVP/PCM(THF) Frequency Analysis"
    wB97X_V_def2_TZVP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_THF__Unknown = "wB97X-V/def2-TZVP/PCM(THF) Unknown"
    wB97X_V_def2_TZVP_PCM_G2__Single_Point = "wB97X-V/def2-TZVP/PCM(G2) Single Point"
    wB97X_V_def2_TZVP_PCM_G2__Force = "wB97X-V/def2-TZVP/PCM(G2) Force"
    wB97X_V_def2_TZVP_PCM_G2__Geometry_Optimization = "wB97X-V/def2-TZVP/PCM(G2) Geometry Optimization"
    wB97X_V_def2_TZVP_PCM_G2__Frequency_Analysis = "wB97X-V/def2-TZVP/PCM(G2) Frequency Analysis"
    wB97X_V_def2_TZVP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_G2__Unknown = "wB97X-V/def2-TZVP/PCM(G2) Unknown"
    wB97X_V_def2_TZVP_PCM_DME__Single_Point = "wB97X-V/def2-TZVP/PCM(DME) Single Point"
    wB97X_V_def2_TZVP_PCM_DME__Force = "wB97X-V/def2-TZVP/PCM(DME) Force"
    wB97X_V_def2_TZVP_PCM_DME__Geometry_Optimization = "wB97X-V/def2-TZVP/PCM(DME) Geometry Optimization"
    wB97X_V_def2_TZVP_PCM_DME__Frequency_Analysis = "wB97X-V/def2-TZVP/PCM(DME) Frequency Analysis"
    wB97X_V_def2_TZVP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_DME__Unknown = "wB97X-V/def2-TZVP/PCM(DME) Unknown"
    wB97X_V_def2_TZVP_PCM_EC__Single_Point = "wB97X-V/def2-TZVP/PCM(EC) Single Point"
    wB97X_V_def2_TZVP_PCM_EC__Force = "wB97X-V/def2-TZVP/PCM(EC) Force"
    wB97X_V_def2_TZVP_PCM_EC__Geometry_Optimization = "wB97X-V/def2-TZVP/PCM(EC) Geometry Optimization"
    wB97X_V_def2_TZVP_PCM_EC__Frequency_Analysis = "wB97X-V/def2-TZVP/PCM(EC) Frequency Analysis"
    wB97X_V_def2_TZVP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_EC__Unknown = "wB97X-V/def2-TZVP/PCM(EC) Unknown"
    wB97X_V_def2_TZVP_PCM_EC_EMC__Single_Point = "wB97X-V/def2-TZVP/PCM(EC/EMC) Single Point"
    wB97X_V_def2_TZVP_PCM_EC_EMC__Force = "wB97X-V/def2-TZVP/PCM(EC/EMC) Force"
    wB97X_V_def2_TZVP_PCM_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVP/PCM(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVP_PCM_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVP/PCM(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_EC_EMC__Unknown = "wB97X-V/def2-TZVP/PCM(EC/EMC) Unknown"
    wB97X_V_def2_TZVP_PCM_PC__Single_Point = "wB97X-V/def2-TZVP/PCM(PC) Single Point"
    wB97X_V_def2_TZVP_PCM_PC__Force = "wB97X-V/def2-TZVP/PCM(PC) Force"
    wB97X_V_def2_TZVP_PCM_PC__Geometry_Optimization = "wB97X-V/def2-TZVP/PCM(PC) Geometry Optimization"
    wB97X_V_def2_TZVP_PCM_PC__Frequency_Analysis = "wB97X-V/def2-TZVP/PCM(PC) Frequency Analysis"
    wB97X_V_def2_TZVP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_PCM_PC__Unknown = "wB97X-V/def2-TZVP/PCM(PC) Unknown"
    wB97X_V_def2_TZVP_ISOSVP_WATER__Single_Point = "wB97X-V/def2-TZVP/ISOSVP(WATER) Single Point"
    wB97X_V_def2_TZVP_ISOSVP_WATER__Force = "wB97X-V/def2-TZVP/ISOSVP(WATER) Force"
    wB97X_V_def2_TZVP_ISOSVP_WATER__Geometry_Optimization = "wB97X-V/def2-TZVP/ISOSVP(WATER) Geometry Optimization"
    wB97X_V_def2_TZVP_ISOSVP_WATER__Frequency_Analysis = "wB97X-V/def2-TZVP/ISOSVP(WATER) Frequency Analysis"
    wB97X_V_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_WATER__Unknown = "wB97X-V/def2-TZVP/ISOSVP(WATER) Unknown"
    wB97X_V_def2_TZVP_ISOSVP_THF__Single_Point = "wB97X-V/def2-TZVP/ISOSVP(THF) Single Point"
    wB97X_V_def2_TZVP_ISOSVP_THF__Force = "wB97X-V/def2-TZVP/ISOSVP(THF) Force"
    wB97X_V_def2_TZVP_ISOSVP_THF__Geometry_Optimization = "wB97X-V/def2-TZVP/ISOSVP(THF) Geometry Optimization"
    wB97X_V_def2_TZVP_ISOSVP_THF__Frequency_Analysis = "wB97X-V/def2-TZVP/ISOSVP(THF) Frequency Analysis"
    wB97X_V_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_THF__Unknown = "wB97X-V/def2-TZVP/ISOSVP(THF) Unknown"
    wB97X_V_def2_TZVP_ISOSVP_G2__Single_Point = "wB97X-V/def2-TZVP/ISOSVP(G2) Single Point"
    wB97X_V_def2_TZVP_ISOSVP_G2__Force = "wB97X-V/def2-TZVP/ISOSVP(G2) Force"
    wB97X_V_def2_TZVP_ISOSVP_G2__Geometry_Optimization = "wB97X-V/def2-TZVP/ISOSVP(G2) Geometry Optimization"
    wB97X_V_def2_TZVP_ISOSVP_G2__Frequency_Analysis = "wB97X-V/def2-TZVP/ISOSVP(G2) Frequency Analysis"
    wB97X_V_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_G2__Unknown = "wB97X-V/def2-TZVP/ISOSVP(G2) Unknown"
    wB97X_V_def2_TZVP_ISOSVP_DME__Single_Point = "wB97X-V/def2-TZVP/ISOSVP(DME) Single Point"
    wB97X_V_def2_TZVP_ISOSVP_DME__Force = "wB97X-V/def2-TZVP/ISOSVP(DME) Force"
    wB97X_V_def2_TZVP_ISOSVP_DME__Geometry_Optimization = "wB97X-V/def2-TZVP/ISOSVP(DME) Geometry Optimization"
    wB97X_V_def2_TZVP_ISOSVP_DME__Frequency_Analysis = "wB97X-V/def2-TZVP/ISOSVP(DME) Frequency Analysis"
    wB97X_V_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_DME__Unknown = "wB97X-V/def2-TZVP/ISOSVP(DME) Unknown"
    wB97X_V_def2_TZVP_ISOSVP_EC__Single_Point = "wB97X-V/def2-TZVP/ISOSVP(EC) Single Point"
    wB97X_V_def2_TZVP_ISOSVP_EC__Force = "wB97X-V/def2-TZVP/ISOSVP(EC) Force"
    wB97X_V_def2_TZVP_ISOSVP_EC__Geometry_Optimization = "wB97X-V/def2-TZVP/ISOSVP(EC) Geometry Optimization"
    wB97X_V_def2_TZVP_ISOSVP_EC__Frequency_Analysis = "wB97X-V/def2-TZVP/ISOSVP(EC) Frequency Analysis"
    wB97X_V_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_EC__Unknown = "wB97X-V/def2-TZVP/ISOSVP(EC) Unknown"
    wB97X_V_def2_TZVP_ISOSVP_EC_EMC__Single_Point = "wB97X-V/def2-TZVP/ISOSVP(EC/EMC) Single Point"
    wB97X_V_def2_TZVP_ISOSVP_EC_EMC__Force = "wB97X-V/def2-TZVP/ISOSVP(EC/EMC) Force"
    wB97X_V_def2_TZVP_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVP/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVP_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVP/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_EC_EMC__Unknown = "wB97X-V/def2-TZVP/ISOSVP(EC/EMC) Unknown"
    wB97X_V_def2_TZVP_ISOSVP_PC__Single_Point = "wB97X-V/def2-TZVP/ISOSVP(PC) Single Point"
    wB97X_V_def2_TZVP_ISOSVP_PC__Force = "wB97X-V/def2-TZVP/ISOSVP(PC) Force"
    wB97X_V_def2_TZVP_ISOSVP_PC__Geometry_Optimization = "wB97X-V/def2-TZVP/ISOSVP(PC) Geometry Optimization"
    wB97X_V_def2_TZVP_ISOSVP_PC__Frequency_Analysis = "wB97X-V/def2-TZVP/ISOSVP(PC) Frequency Analysis"
    wB97X_V_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_ISOSVP_PC__Unknown = "wB97X-V/def2-TZVP/ISOSVP(PC) Unknown"
    wB97X_V_def2_TZVP_SMD_WATER__Single_Point = "wB97X-V/def2-TZVP/SMD(WATER) Single Point"
    wB97X_V_def2_TZVP_SMD_WATER__Force = "wB97X-V/def2-TZVP/SMD(WATER) Force"
    wB97X_V_def2_TZVP_SMD_WATER__Geometry_Optimization = "wB97X-V/def2-TZVP/SMD(WATER) Geometry Optimization"
    wB97X_V_def2_TZVP_SMD_WATER__Frequency_Analysis = "wB97X-V/def2-TZVP/SMD(WATER) Frequency Analysis"
    wB97X_V_def2_TZVP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_WATER__Unknown = "wB97X-V/def2-TZVP/SMD(WATER) Unknown"
    wB97X_V_def2_TZVP_SMD_THF__Single_Point = "wB97X-V/def2-TZVP/SMD(THF) Single Point"
    wB97X_V_def2_TZVP_SMD_THF__Force = "wB97X-V/def2-TZVP/SMD(THF) Force"
    wB97X_V_def2_TZVP_SMD_THF__Geometry_Optimization = "wB97X-V/def2-TZVP/SMD(THF) Geometry Optimization"
    wB97X_V_def2_TZVP_SMD_THF__Frequency_Analysis = "wB97X-V/def2-TZVP/SMD(THF) Frequency Analysis"
    wB97X_V_def2_TZVP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_THF__Unknown = "wB97X-V/def2-TZVP/SMD(THF) Unknown"
    wB97X_V_def2_TZVP_SMD_G2__Single_Point = "wB97X-V/def2-TZVP/SMD(G2) Single Point"
    wB97X_V_def2_TZVP_SMD_G2__Force = "wB97X-V/def2-TZVP/SMD(G2) Force"
    wB97X_V_def2_TZVP_SMD_G2__Geometry_Optimization = "wB97X-V/def2-TZVP/SMD(G2) Geometry Optimization"
    wB97X_V_def2_TZVP_SMD_G2__Frequency_Analysis = "wB97X-V/def2-TZVP/SMD(G2) Frequency Analysis"
    wB97X_V_def2_TZVP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_G2__Unknown = "wB97X-V/def2-TZVP/SMD(G2) Unknown"
    wB97X_V_def2_TZVP_SMD_DME__Single_Point = "wB97X-V/def2-TZVP/SMD(DME) Single Point"
    wB97X_V_def2_TZVP_SMD_DME__Force = "wB97X-V/def2-TZVP/SMD(DME) Force"
    wB97X_V_def2_TZVP_SMD_DME__Geometry_Optimization = "wB97X-V/def2-TZVP/SMD(DME) Geometry Optimization"
    wB97X_V_def2_TZVP_SMD_DME__Frequency_Analysis = "wB97X-V/def2-TZVP/SMD(DME) Frequency Analysis"
    wB97X_V_def2_TZVP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_DME__Unknown = "wB97X-V/def2-TZVP/SMD(DME) Unknown"
    wB97X_V_def2_TZVP_SMD_EC__Single_Point = "wB97X-V/def2-TZVP/SMD(EC) Single Point"
    wB97X_V_def2_TZVP_SMD_EC__Force = "wB97X-V/def2-TZVP/SMD(EC) Force"
    wB97X_V_def2_TZVP_SMD_EC__Geometry_Optimization = "wB97X-V/def2-TZVP/SMD(EC) Geometry Optimization"
    wB97X_V_def2_TZVP_SMD_EC__Frequency_Analysis = "wB97X-V/def2-TZVP/SMD(EC) Frequency Analysis"
    wB97X_V_def2_TZVP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_EC__Unknown = "wB97X-V/def2-TZVP/SMD(EC) Unknown"
    wB97X_V_def2_TZVP_SMD_EC_EMC__Single_Point = "wB97X-V/def2-TZVP/SMD(EC/EMC) Single Point"
    wB97X_V_def2_TZVP_SMD_EC_EMC__Force = "wB97X-V/def2-TZVP/SMD(EC/EMC) Force"
    wB97X_V_def2_TZVP_SMD_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVP/SMD(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVP_SMD_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVP/SMD(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_EC_EMC__Unknown = "wB97X-V/def2-TZVP/SMD(EC/EMC) Unknown"
    wB97X_V_def2_TZVP_SMD_PC__Single_Point = "wB97X-V/def2-TZVP/SMD(PC) Single Point"
    wB97X_V_def2_TZVP_SMD_PC__Force = "wB97X-V/def2-TZVP/SMD(PC) Force"
    wB97X_V_def2_TZVP_SMD_PC__Geometry_Optimization = "wB97X-V/def2-TZVP/SMD(PC) Geometry Optimization"
    wB97X_V_def2_TZVP_SMD_PC__Frequency_Analysis = "wB97X-V/def2-TZVP/SMD(PC) Frequency Analysis"
    wB97X_V_def2_TZVP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_SMD_PC__Unknown = "wB97X-V/def2-TZVP/SMD(PC) Unknown"
    wB97X_V_def2_TZVP_CMIRS_WATER__Single_Point = "wB97X-V/def2-TZVP/CMIRS(WATER) Single Point"
    wB97X_V_def2_TZVP_CMIRS_WATER__Force = "wB97X-V/def2-TZVP/CMIRS(WATER) Force"
    wB97X_V_def2_TZVP_CMIRS_WATER__Geometry_Optimization = "wB97X-V/def2-TZVP/CMIRS(WATER) Geometry Optimization"
    wB97X_V_def2_TZVP_CMIRS_WATER__Frequency_Analysis = "wB97X-V/def2-TZVP/CMIRS(WATER) Frequency Analysis"
    wB97X_V_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_WATER__Unknown = "wB97X-V/def2-TZVP/CMIRS(WATER) Unknown"
    wB97X_V_def2_TZVP_CMIRS_THF__Single_Point = "wB97X-V/def2-TZVP/CMIRS(THF) Single Point"
    wB97X_V_def2_TZVP_CMIRS_THF__Force = "wB97X-V/def2-TZVP/CMIRS(THF) Force"
    wB97X_V_def2_TZVP_CMIRS_THF__Geometry_Optimization = "wB97X-V/def2-TZVP/CMIRS(THF) Geometry Optimization"
    wB97X_V_def2_TZVP_CMIRS_THF__Frequency_Analysis = "wB97X-V/def2-TZVP/CMIRS(THF) Frequency Analysis"
    wB97X_V_def2_TZVP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_THF__Unknown = "wB97X-V/def2-TZVP/CMIRS(THF) Unknown"
    wB97X_V_def2_TZVP_CMIRS_G2__Single_Point = "wB97X-V/def2-TZVP/CMIRS(G2) Single Point"
    wB97X_V_def2_TZVP_CMIRS_G2__Force = "wB97X-V/def2-TZVP/CMIRS(G2) Force"
    wB97X_V_def2_TZVP_CMIRS_G2__Geometry_Optimization = "wB97X-V/def2-TZVP/CMIRS(G2) Geometry Optimization"
    wB97X_V_def2_TZVP_CMIRS_G2__Frequency_Analysis = "wB97X-V/def2-TZVP/CMIRS(G2) Frequency Analysis"
    wB97X_V_def2_TZVP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_G2__Unknown = "wB97X-V/def2-TZVP/CMIRS(G2) Unknown"
    wB97X_V_def2_TZVP_CMIRS_DME__Single_Point = "wB97X-V/def2-TZVP/CMIRS(DME) Single Point"
    wB97X_V_def2_TZVP_CMIRS_DME__Force = "wB97X-V/def2-TZVP/CMIRS(DME) Force"
    wB97X_V_def2_TZVP_CMIRS_DME__Geometry_Optimization = "wB97X-V/def2-TZVP/CMIRS(DME) Geometry Optimization"
    wB97X_V_def2_TZVP_CMIRS_DME__Frequency_Analysis = "wB97X-V/def2-TZVP/CMIRS(DME) Frequency Analysis"
    wB97X_V_def2_TZVP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_DME__Unknown = "wB97X-V/def2-TZVP/CMIRS(DME) Unknown"
    wB97X_V_def2_TZVP_CMIRS_EC__Single_Point = "wB97X-V/def2-TZVP/CMIRS(EC) Single Point"
    wB97X_V_def2_TZVP_CMIRS_EC__Force = "wB97X-V/def2-TZVP/CMIRS(EC) Force"
    wB97X_V_def2_TZVP_CMIRS_EC__Geometry_Optimization = "wB97X-V/def2-TZVP/CMIRS(EC) Geometry Optimization"
    wB97X_V_def2_TZVP_CMIRS_EC__Frequency_Analysis = "wB97X-V/def2-TZVP/CMIRS(EC) Frequency Analysis"
    wB97X_V_def2_TZVP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_EC__Unknown = "wB97X-V/def2-TZVP/CMIRS(EC) Unknown"
    wB97X_V_def2_TZVP_CMIRS_EC_EMC__Single_Point = "wB97X-V/def2-TZVP/CMIRS(EC/EMC) Single Point"
    wB97X_V_def2_TZVP_CMIRS_EC_EMC__Force = "wB97X-V/def2-TZVP/CMIRS(EC/EMC) Force"
    wB97X_V_def2_TZVP_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVP/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVP_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVP/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_EC_EMC__Unknown = "wB97X-V/def2-TZVP/CMIRS(EC/EMC) Unknown"
    wB97X_V_def2_TZVP_CMIRS_PC__Single_Point = "wB97X-V/def2-TZVP/CMIRS(PC) Single Point"
    wB97X_V_def2_TZVP_CMIRS_PC__Force = "wB97X-V/def2-TZVP/CMIRS(PC) Force"
    wB97X_V_def2_TZVP_CMIRS_PC__Geometry_Optimization = "wB97X-V/def2-TZVP/CMIRS(PC) Geometry Optimization"
    wB97X_V_def2_TZVP_CMIRS_PC__Frequency_Analysis = "wB97X-V/def2-TZVP/CMIRS(PC) Frequency Analysis"
    wB97X_V_def2_TZVP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVP_CMIRS_PC__Unknown = "wB97X-V/def2-TZVP/CMIRS(PC) Unknown"
    wB97X_V_def2_TZVPD_VACUUM_Single_Point = "wB97X-V/def2-TZVPD/VACUUM Single Point"
    wB97X_V_def2_TZVPD_VACUUM_Force = "wB97X-V/def2-TZVPD/VACUUM Force"
    wB97X_V_def2_TZVPD_VACUUM_Geometry_Optimization = "wB97X-V/def2-TZVPD/VACUUM Geometry Optimization"
    wB97X_V_def2_TZVPD_VACUUM_Frequency_Analysis = "wB97X-V/def2-TZVPD/VACUUM Frequency Analysis"
    wB97X_V_def2_TZVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_VACUUM_Unknown = "wB97X-V/def2-TZVPD/VACUUM Unknown"
    wB97X_V_def2_TZVPD_PCM_WATER__Single_Point = "wB97X-V/def2-TZVPD/PCM(WATER) Single Point"
    wB97X_V_def2_TZVPD_PCM_WATER__Force = "wB97X-V/def2-TZVPD/PCM(WATER) Force"
    wB97X_V_def2_TZVPD_PCM_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPD/PCM(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPD_PCM_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPD/PCM(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_WATER__Unknown = "wB97X-V/def2-TZVPD/PCM(WATER) Unknown"
    wB97X_V_def2_TZVPD_PCM_THF__Single_Point = "wB97X-V/def2-TZVPD/PCM(THF) Single Point"
    wB97X_V_def2_TZVPD_PCM_THF__Force = "wB97X-V/def2-TZVPD/PCM(THF) Force"
    wB97X_V_def2_TZVPD_PCM_THF__Geometry_Optimization = "wB97X-V/def2-TZVPD/PCM(THF) Geometry Optimization"
    wB97X_V_def2_TZVPD_PCM_THF__Frequency_Analysis = "wB97X-V/def2-TZVPD/PCM(THF) Frequency Analysis"
    wB97X_V_def2_TZVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_THF__Unknown = "wB97X-V/def2-TZVPD/PCM(THF) Unknown"
    wB97X_V_def2_TZVPD_PCM_G2__Single_Point = "wB97X-V/def2-TZVPD/PCM(G2) Single Point"
    wB97X_V_def2_TZVPD_PCM_G2__Force = "wB97X-V/def2-TZVPD/PCM(G2) Force"
    wB97X_V_def2_TZVPD_PCM_G2__Geometry_Optimization = "wB97X-V/def2-TZVPD/PCM(G2) Geometry Optimization"
    wB97X_V_def2_TZVPD_PCM_G2__Frequency_Analysis = "wB97X-V/def2-TZVPD/PCM(G2) Frequency Analysis"
    wB97X_V_def2_TZVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_G2__Unknown = "wB97X-V/def2-TZVPD/PCM(G2) Unknown"
    wB97X_V_def2_TZVPD_PCM_DME__Single_Point = "wB97X-V/def2-TZVPD/PCM(DME) Single Point"
    wB97X_V_def2_TZVPD_PCM_DME__Force = "wB97X-V/def2-TZVPD/PCM(DME) Force"
    wB97X_V_def2_TZVPD_PCM_DME__Geometry_Optimization = "wB97X-V/def2-TZVPD/PCM(DME) Geometry Optimization"
    wB97X_V_def2_TZVPD_PCM_DME__Frequency_Analysis = "wB97X-V/def2-TZVPD/PCM(DME) Frequency Analysis"
    wB97X_V_def2_TZVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_DME__Unknown = "wB97X-V/def2-TZVPD/PCM(DME) Unknown"
    wB97X_V_def2_TZVPD_PCM_EC__Single_Point = "wB97X-V/def2-TZVPD/PCM(EC) Single Point"
    wB97X_V_def2_TZVPD_PCM_EC__Force = "wB97X-V/def2-TZVPD/PCM(EC) Force"
    wB97X_V_def2_TZVPD_PCM_EC__Geometry_Optimization = "wB97X-V/def2-TZVPD/PCM(EC) Geometry Optimization"
    wB97X_V_def2_TZVPD_PCM_EC__Frequency_Analysis = "wB97X-V/def2-TZVPD/PCM(EC) Frequency Analysis"
    wB97X_V_def2_TZVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_EC__Unknown = "wB97X-V/def2-TZVPD/PCM(EC) Unknown"
    wB97X_V_def2_TZVPD_PCM_EC_EMC__Single_Point = "wB97X-V/def2-TZVPD/PCM(EC/EMC) Single Point"
    wB97X_V_def2_TZVPD_PCM_EC_EMC__Force = "wB97X-V/def2-TZVPD/PCM(EC/EMC) Force"
    wB97X_V_def2_TZVPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_EC_EMC__Unknown = "wB97X-V/def2-TZVPD/PCM(EC/EMC) Unknown"
    wB97X_V_def2_TZVPD_PCM_PC__Single_Point = "wB97X-V/def2-TZVPD/PCM(PC) Single Point"
    wB97X_V_def2_TZVPD_PCM_PC__Force = "wB97X-V/def2-TZVPD/PCM(PC) Force"
    wB97X_V_def2_TZVPD_PCM_PC__Geometry_Optimization = "wB97X-V/def2-TZVPD/PCM(PC) Geometry Optimization"
    wB97X_V_def2_TZVPD_PCM_PC__Frequency_Analysis = "wB97X-V/def2-TZVPD/PCM(PC) Frequency Analysis"
    wB97X_V_def2_TZVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_PCM_PC__Unknown = "wB97X-V/def2-TZVPD/PCM(PC) Unknown"
    wB97X_V_def2_TZVPD_ISOSVP_WATER__Single_Point = "wB97X-V/def2-TZVPD/ISOSVP(WATER) Single Point"
    wB97X_V_def2_TZVPD_ISOSVP_WATER__Force = "wB97X-V/def2-TZVPD/ISOSVP(WATER) Force"
    wB97X_V_def2_TZVPD_ISOSVP_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPD/ISOSVP(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_WATER__Unknown = "wB97X-V/def2-TZVPD/ISOSVP(WATER) Unknown"
    wB97X_V_def2_TZVPD_ISOSVP_THF__Single_Point = "wB97X-V/def2-TZVPD/ISOSVP(THF) Single Point"
    wB97X_V_def2_TZVPD_ISOSVP_THF__Force = "wB97X-V/def2-TZVPD/ISOSVP(THF) Force"
    wB97X_V_def2_TZVPD_ISOSVP_THF__Geometry_Optimization = "wB97X-V/def2-TZVPD/ISOSVP(THF) Geometry Optimization"
    wB97X_V_def2_TZVPD_ISOSVP_THF__Frequency_Analysis = "wB97X-V/def2-TZVPD/ISOSVP(THF) Frequency Analysis"
    wB97X_V_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_THF__Unknown = "wB97X-V/def2-TZVPD/ISOSVP(THF) Unknown"
    wB97X_V_def2_TZVPD_ISOSVP_G2__Single_Point = "wB97X-V/def2-TZVPD/ISOSVP(G2) Single Point"
    wB97X_V_def2_TZVPD_ISOSVP_G2__Force = "wB97X-V/def2-TZVPD/ISOSVP(G2) Force"
    wB97X_V_def2_TZVPD_ISOSVP_G2__Geometry_Optimization = "wB97X-V/def2-TZVPD/ISOSVP(G2) Geometry Optimization"
    wB97X_V_def2_TZVPD_ISOSVP_G2__Frequency_Analysis = "wB97X-V/def2-TZVPD/ISOSVP(G2) Frequency Analysis"
    wB97X_V_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_G2__Unknown = "wB97X-V/def2-TZVPD/ISOSVP(G2) Unknown"
    wB97X_V_def2_TZVPD_ISOSVP_DME__Single_Point = "wB97X-V/def2-TZVPD/ISOSVP(DME) Single Point"
    wB97X_V_def2_TZVPD_ISOSVP_DME__Force = "wB97X-V/def2-TZVPD/ISOSVP(DME) Force"
    wB97X_V_def2_TZVPD_ISOSVP_DME__Geometry_Optimization = "wB97X-V/def2-TZVPD/ISOSVP(DME) Geometry Optimization"
    wB97X_V_def2_TZVPD_ISOSVP_DME__Frequency_Analysis = "wB97X-V/def2-TZVPD/ISOSVP(DME) Frequency Analysis"
    wB97X_V_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_DME__Unknown = "wB97X-V/def2-TZVPD/ISOSVP(DME) Unknown"
    wB97X_V_def2_TZVPD_ISOSVP_EC__Single_Point = "wB97X-V/def2-TZVPD/ISOSVP(EC) Single Point"
    wB97X_V_def2_TZVPD_ISOSVP_EC__Force = "wB97X-V/def2-TZVPD/ISOSVP(EC) Force"
    wB97X_V_def2_TZVPD_ISOSVP_EC__Geometry_Optimization = "wB97X-V/def2-TZVPD/ISOSVP(EC) Geometry Optimization"
    wB97X_V_def2_TZVPD_ISOSVP_EC__Frequency_Analysis = "wB97X-V/def2-TZVPD/ISOSVP(EC) Frequency Analysis"
    wB97X_V_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_EC__Unknown = "wB97X-V/def2-TZVPD/ISOSVP(EC) Unknown"
    wB97X_V_def2_TZVPD_ISOSVP_EC_EMC__Single_Point = "wB97X-V/def2-TZVPD/ISOSVP(EC/EMC) Single Point"
    wB97X_V_def2_TZVPD_ISOSVP_EC_EMC__Force = "wB97X-V/def2-TZVPD/ISOSVP(EC/EMC) Force"
    wB97X_V_def2_TZVPD_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVPD/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_EC_EMC__Unknown = "wB97X-V/def2-TZVPD/ISOSVP(EC/EMC) Unknown"
    wB97X_V_def2_TZVPD_ISOSVP_PC__Single_Point = "wB97X-V/def2-TZVPD/ISOSVP(PC) Single Point"
    wB97X_V_def2_TZVPD_ISOSVP_PC__Force = "wB97X-V/def2-TZVPD/ISOSVP(PC) Force"
    wB97X_V_def2_TZVPD_ISOSVP_PC__Geometry_Optimization = "wB97X-V/def2-TZVPD/ISOSVP(PC) Geometry Optimization"
    wB97X_V_def2_TZVPD_ISOSVP_PC__Frequency_Analysis = "wB97X-V/def2-TZVPD/ISOSVP(PC) Frequency Analysis"
    wB97X_V_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_ISOSVP_PC__Unknown = "wB97X-V/def2-TZVPD/ISOSVP(PC) Unknown"
    wB97X_V_def2_TZVPD_SMD_WATER__Single_Point = "wB97X-V/def2-TZVPD/SMD(WATER) Single Point"
    wB97X_V_def2_TZVPD_SMD_WATER__Force = "wB97X-V/def2-TZVPD/SMD(WATER) Force"
    wB97X_V_def2_TZVPD_SMD_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPD/SMD(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPD_SMD_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPD/SMD(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_WATER__Unknown = "wB97X-V/def2-TZVPD/SMD(WATER) Unknown"
    wB97X_V_def2_TZVPD_SMD_THF__Single_Point = "wB97X-V/def2-TZVPD/SMD(THF) Single Point"
    wB97X_V_def2_TZVPD_SMD_THF__Force = "wB97X-V/def2-TZVPD/SMD(THF) Force"
    wB97X_V_def2_TZVPD_SMD_THF__Geometry_Optimization = "wB97X-V/def2-TZVPD/SMD(THF) Geometry Optimization"
    wB97X_V_def2_TZVPD_SMD_THF__Frequency_Analysis = "wB97X-V/def2-TZVPD/SMD(THF) Frequency Analysis"
    wB97X_V_def2_TZVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_THF__Unknown = "wB97X-V/def2-TZVPD/SMD(THF) Unknown"
    wB97X_V_def2_TZVPD_SMD_G2__Single_Point = "wB97X-V/def2-TZVPD/SMD(G2) Single Point"
    wB97X_V_def2_TZVPD_SMD_G2__Force = "wB97X-V/def2-TZVPD/SMD(G2) Force"
    wB97X_V_def2_TZVPD_SMD_G2__Geometry_Optimization = "wB97X-V/def2-TZVPD/SMD(G2) Geometry Optimization"
    wB97X_V_def2_TZVPD_SMD_G2__Frequency_Analysis = "wB97X-V/def2-TZVPD/SMD(G2) Frequency Analysis"
    wB97X_V_def2_TZVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_G2__Unknown = "wB97X-V/def2-TZVPD/SMD(G2) Unknown"
    wB97X_V_def2_TZVPD_SMD_DME__Single_Point = "wB97X-V/def2-TZVPD/SMD(DME) Single Point"
    wB97X_V_def2_TZVPD_SMD_DME__Force = "wB97X-V/def2-TZVPD/SMD(DME) Force"
    wB97X_V_def2_TZVPD_SMD_DME__Geometry_Optimization = "wB97X-V/def2-TZVPD/SMD(DME) Geometry Optimization"
    wB97X_V_def2_TZVPD_SMD_DME__Frequency_Analysis = "wB97X-V/def2-TZVPD/SMD(DME) Frequency Analysis"
    wB97X_V_def2_TZVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_DME__Unknown = "wB97X-V/def2-TZVPD/SMD(DME) Unknown"
    wB97X_V_def2_TZVPD_SMD_EC__Single_Point = "wB97X-V/def2-TZVPD/SMD(EC) Single Point"
    wB97X_V_def2_TZVPD_SMD_EC__Force = "wB97X-V/def2-TZVPD/SMD(EC) Force"
    wB97X_V_def2_TZVPD_SMD_EC__Geometry_Optimization = "wB97X-V/def2-TZVPD/SMD(EC) Geometry Optimization"
    wB97X_V_def2_TZVPD_SMD_EC__Frequency_Analysis = "wB97X-V/def2-TZVPD/SMD(EC) Frequency Analysis"
    wB97X_V_def2_TZVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_EC__Unknown = "wB97X-V/def2-TZVPD/SMD(EC) Unknown"
    wB97X_V_def2_TZVPD_SMD_EC_EMC__Single_Point = "wB97X-V/def2-TZVPD/SMD(EC/EMC) Single Point"
    wB97X_V_def2_TZVPD_SMD_EC_EMC__Force = "wB97X-V/def2-TZVPD/SMD(EC/EMC) Force"
    wB97X_V_def2_TZVPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_EC_EMC__Unknown = "wB97X-V/def2-TZVPD/SMD(EC/EMC) Unknown"
    wB97X_V_def2_TZVPD_SMD_PC__Single_Point = "wB97X-V/def2-TZVPD/SMD(PC) Single Point"
    wB97X_V_def2_TZVPD_SMD_PC__Force = "wB97X-V/def2-TZVPD/SMD(PC) Force"
    wB97X_V_def2_TZVPD_SMD_PC__Geometry_Optimization = "wB97X-V/def2-TZVPD/SMD(PC) Geometry Optimization"
    wB97X_V_def2_TZVPD_SMD_PC__Frequency_Analysis = "wB97X-V/def2-TZVPD/SMD(PC) Frequency Analysis"
    wB97X_V_def2_TZVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_SMD_PC__Unknown = "wB97X-V/def2-TZVPD/SMD(PC) Unknown"
    wB97X_V_def2_TZVPD_CMIRS_WATER__Single_Point = "wB97X-V/def2-TZVPD/CMIRS(WATER) Single Point"
    wB97X_V_def2_TZVPD_CMIRS_WATER__Force = "wB97X-V/def2-TZVPD/CMIRS(WATER) Force"
    wB97X_V_def2_TZVPD_CMIRS_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPD/CMIRS(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPD_CMIRS_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPD/CMIRS(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_WATER__Unknown = "wB97X-V/def2-TZVPD/CMIRS(WATER) Unknown"
    wB97X_V_def2_TZVPD_CMIRS_THF__Single_Point = "wB97X-V/def2-TZVPD/CMIRS(THF) Single Point"
    wB97X_V_def2_TZVPD_CMIRS_THF__Force = "wB97X-V/def2-TZVPD/CMIRS(THF) Force"
    wB97X_V_def2_TZVPD_CMIRS_THF__Geometry_Optimization = "wB97X-V/def2-TZVPD/CMIRS(THF) Geometry Optimization"
    wB97X_V_def2_TZVPD_CMIRS_THF__Frequency_Analysis = "wB97X-V/def2-TZVPD/CMIRS(THF) Frequency Analysis"
    wB97X_V_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_THF__Unknown = "wB97X-V/def2-TZVPD/CMIRS(THF) Unknown"
    wB97X_V_def2_TZVPD_CMIRS_G2__Single_Point = "wB97X-V/def2-TZVPD/CMIRS(G2) Single Point"
    wB97X_V_def2_TZVPD_CMIRS_G2__Force = "wB97X-V/def2-TZVPD/CMIRS(G2) Force"
    wB97X_V_def2_TZVPD_CMIRS_G2__Geometry_Optimization = "wB97X-V/def2-TZVPD/CMIRS(G2) Geometry Optimization"
    wB97X_V_def2_TZVPD_CMIRS_G2__Frequency_Analysis = "wB97X-V/def2-TZVPD/CMIRS(G2) Frequency Analysis"
    wB97X_V_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_G2__Unknown = "wB97X-V/def2-TZVPD/CMIRS(G2) Unknown"
    wB97X_V_def2_TZVPD_CMIRS_DME__Single_Point = "wB97X-V/def2-TZVPD/CMIRS(DME) Single Point"
    wB97X_V_def2_TZVPD_CMIRS_DME__Force = "wB97X-V/def2-TZVPD/CMIRS(DME) Force"
    wB97X_V_def2_TZVPD_CMIRS_DME__Geometry_Optimization = "wB97X-V/def2-TZVPD/CMIRS(DME) Geometry Optimization"
    wB97X_V_def2_TZVPD_CMIRS_DME__Frequency_Analysis = "wB97X-V/def2-TZVPD/CMIRS(DME) Frequency Analysis"
    wB97X_V_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_DME__Unknown = "wB97X-V/def2-TZVPD/CMIRS(DME) Unknown"
    wB97X_V_def2_TZVPD_CMIRS_EC__Single_Point = "wB97X-V/def2-TZVPD/CMIRS(EC) Single Point"
    wB97X_V_def2_TZVPD_CMIRS_EC__Force = "wB97X-V/def2-TZVPD/CMIRS(EC) Force"
    wB97X_V_def2_TZVPD_CMIRS_EC__Geometry_Optimization = "wB97X-V/def2-TZVPD/CMIRS(EC) Geometry Optimization"
    wB97X_V_def2_TZVPD_CMIRS_EC__Frequency_Analysis = "wB97X-V/def2-TZVPD/CMIRS(EC) Frequency Analysis"
    wB97X_V_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_EC__Unknown = "wB97X-V/def2-TZVPD/CMIRS(EC) Unknown"
    wB97X_V_def2_TZVPD_CMIRS_EC_EMC__Single_Point = "wB97X-V/def2-TZVPD/CMIRS(EC/EMC) Single Point"
    wB97X_V_def2_TZVPD_CMIRS_EC_EMC__Force = "wB97X-V/def2-TZVPD/CMIRS(EC/EMC) Force"
    wB97X_V_def2_TZVPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_EC_EMC__Unknown = "wB97X-V/def2-TZVPD/CMIRS(EC/EMC) Unknown"
    wB97X_V_def2_TZVPD_CMIRS_PC__Single_Point = "wB97X-V/def2-TZVPD/CMIRS(PC) Single Point"
    wB97X_V_def2_TZVPD_CMIRS_PC__Force = "wB97X-V/def2-TZVPD/CMIRS(PC) Force"
    wB97X_V_def2_TZVPD_CMIRS_PC__Geometry_Optimization = "wB97X-V/def2-TZVPD/CMIRS(PC) Geometry Optimization"
    wB97X_V_def2_TZVPD_CMIRS_PC__Frequency_Analysis = "wB97X-V/def2-TZVPD/CMIRS(PC) Frequency Analysis"
    wB97X_V_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPD_CMIRS_PC__Unknown = "wB97X-V/def2-TZVPD/CMIRS(PC) Unknown"
    wB97X_V_def2_TZVPP_VACUUM_Single_Point = "wB97X-V/def2-TZVPP/VACUUM Single Point"
    wB97X_V_def2_TZVPP_VACUUM_Force = "wB97X-V/def2-TZVPP/VACUUM Force"
    wB97X_V_def2_TZVPP_VACUUM_Geometry_Optimization = "wB97X-V/def2-TZVPP/VACUUM Geometry Optimization"
    wB97X_V_def2_TZVPP_VACUUM_Frequency_Analysis = "wB97X-V/def2-TZVPP/VACUUM Frequency Analysis"
    wB97X_V_def2_TZVPP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/VACUUM Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_VACUUM_Unknown = "wB97X-V/def2-TZVPP/VACUUM Unknown"
    wB97X_V_def2_TZVPP_PCM_WATER__Single_Point = "wB97X-V/def2-TZVPP/PCM(WATER) Single Point"
    wB97X_V_def2_TZVPP_PCM_WATER__Force = "wB97X-V/def2-TZVPP/PCM(WATER) Force"
    wB97X_V_def2_TZVPP_PCM_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPP/PCM(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPP_PCM_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPP/PCM(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_WATER__Unknown = "wB97X-V/def2-TZVPP/PCM(WATER) Unknown"
    wB97X_V_def2_TZVPP_PCM_THF__Single_Point = "wB97X-V/def2-TZVPP/PCM(THF) Single Point"
    wB97X_V_def2_TZVPP_PCM_THF__Force = "wB97X-V/def2-TZVPP/PCM(THF) Force"
    wB97X_V_def2_TZVPP_PCM_THF__Geometry_Optimization = "wB97X-V/def2-TZVPP/PCM(THF) Geometry Optimization"
    wB97X_V_def2_TZVPP_PCM_THF__Frequency_Analysis = "wB97X-V/def2-TZVPP/PCM(THF) Frequency Analysis"
    wB97X_V_def2_TZVPP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_THF__Unknown = "wB97X-V/def2-TZVPP/PCM(THF) Unknown"
    wB97X_V_def2_TZVPP_PCM_G2__Single_Point = "wB97X-V/def2-TZVPP/PCM(G2) Single Point"
    wB97X_V_def2_TZVPP_PCM_G2__Force = "wB97X-V/def2-TZVPP/PCM(G2) Force"
    wB97X_V_def2_TZVPP_PCM_G2__Geometry_Optimization = "wB97X-V/def2-TZVPP/PCM(G2) Geometry Optimization"
    wB97X_V_def2_TZVPP_PCM_G2__Frequency_Analysis = "wB97X-V/def2-TZVPP/PCM(G2) Frequency Analysis"
    wB97X_V_def2_TZVPP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_G2__Unknown = "wB97X-V/def2-TZVPP/PCM(G2) Unknown"
    wB97X_V_def2_TZVPP_PCM_DME__Single_Point = "wB97X-V/def2-TZVPP/PCM(DME) Single Point"
    wB97X_V_def2_TZVPP_PCM_DME__Force = "wB97X-V/def2-TZVPP/PCM(DME) Force"
    wB97X_V_def2_TZVPP_PCM_DME__Geometry_Optimization = "wB97X-V/def2-TZVPP/PCM(DME) Geometry Optimization"
    wB97X_V_def2_TZVPP_PCM_DME__Frequency_Analysis = "wB97X-V/def2-TZVPP/PCM(DME) Frequency Analysis"
    wB97X_V_def2_TZVPP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_DME__Unknown = "wB97X-V/def2-TZVPP/PCM(DME) Unknown"
    wB97X_V_def2_TZVPP_PCM_EC__Single_Point = "wB97X-V/def2-TZVPP/PCM(EC) Single Point"
    wB97X_V_def2_TZVPP_PCM_EC__Force = "wB97X-V/def2-TZVPP/PCM(EC) Force"
    wB97X_V_def2_TZVPP_PCM_EC__Geometry_Optimization = "wB97X-V/def2-TZVPP/PCM(EC) Geometry Optimization"
    wB97X_V_def2_TZVPP_PCM_EC__Frequency_Analysis = "wB97X-V/def2-TZVPP/PCM(EC) Frequency Analysis"
    wB97X_V_def2_TZVPP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_EC__Unknown = "wB97X-V/def2-TZVPP/PCM(EC) Unknown"
    wB97X_V_def2_TZVPP_PCM_EC_EMC__Single_Point = "wB97X-V/def2-TZVPP/PCM(EC/EMC) Single Point"
    wB97X_V_def2_TZVPP_PCM_EC_EMC__Force = "wB97X-V/def2-TZVPP/PCM(EC/EMC) Force"
    wB97X_V_def2_TZVPP_PCM_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVPP/PCM(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVPP_PCM_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPP/PCM(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_EC_EMC__Unknown = "wB97X-V/def2-TZVPP/PCM(EC/EMC) Unknown"
    wB97X_V_def2_TZVPP_PCM_PC__Single_Point = "wB97X-V/def2-TZVPP/PCM(PC) Single Point"
    wB97X_V_def2_TZVPP_PCM_PC__Force = "wB97X-V/def2-TZVPP/PCM(PC) Force"
    wB97X_V_def2_TZVPP_PCM_PC__Geometry_Optimization = "wB97X-V/def2-TZVPP/PCM(PC) Geometry Optimization"
    wB97X_V_def2_TZVPP_PCM_PC__Frequency_Analysis = "wB97X-V/def2-TZVPP/PCM(PC) Frequency Analysis"
    wB97X_V_def2_TZVPP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_PCM_PC__Unknown = "wB97X-V/def2-TZVPP/PCM(PC) Unknown"
    wB97X_V_def2_TZVPP_ISOSVP_WATER__Single_Point = "wB97X-V/def2-TZVPP/ISOSVP(WATER) Single Point"
    wB97X_V_def2_TZVPP_ISOSVP_WATER__Force = "wB97X-V/def2-TZVPP/ISOSVP(WATER) Force"
    wB97X_V_def2_TZVPP_ISOSVP_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPP/ISOSVP(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPP_ISOSVP_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPP/ISOSVP(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_WATER__Unknown = "wB97X-V/def2-TZVPP/ISOSVP(WATER) Unknown"
    wB97X_V_def2_TZVPP_ISOSVP_THF__Single_Point = "wB97X-V/def2-TZVPP/ISOSVP(THF) Single Point"
    wB97X_V_def2_TZVPP_ISOSVP_THF__Force = "wB97X-V/def2-TZVPP/ISOSVP(THF) Force"
    wB97X_V_def2_TZVPP_ISOSVP_THF__Geometry_Optimization = "wB97X-V/def2-TZVPP/ISOSVP(THF) Geometry Optimization"
    wB97X_V_def2_TZVPP_ISOSVP_THF__Frequency_Analysis = "wB97X-V/def2-TZVPP/ISOSVP(THF) Frequency Analysis"
    wB97X_V_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_THF__Unknown = "wB97X-V/def2-TZVPP/ISOSVP(THF) Unknown"
    wB97X_V_def2_TZVPP_ISOSVP_G2__Single_Point = "wB97X-V/def2-TZVPP/ISOSVP(G2) Single Point"
    wB97X_V_def2_TZVPP_ISOSVP_G2__Force = "wB97X-V/def2-TZVPP/ISOSVP(G2) Force"
    wB97X_V_def2_TZVPP_ISOSVP_G2__Geometry_Optimization = "wB97X-V/def2-TZVPP/ISOSVP(G2) Geometry Optimization"
    wB97X_V_def2_TZVPP_ISOSVP_G2__Frequency_Analysis = "wB97X-V/def2-TZVPP/ISOSVP(G2) Frequency Analysis"
    wB97X_V_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_G2__Unknown = "wB97X-V/def2-TZVPP/ISOSVP(G2) Unknown"
    wB97X_V_def2_TZVPP_ISOSVP_DME__Single_Point = "wB97X-V/def2-TZVPP/ISOSVP(DME) Single Point"
    wB97X_V_def2_TZVPP_ISOSVP_DME__Force = "wB97X-V/def2-TZVPP/ISOSVP(DME) Force"
    wB97X_V_def2_TZVPP_ISOSVP_DME__Geometry_Optimization = "wB97X-V/def2-TZVPP/ISOSVP(DME) Geometry Optimization"
    wB97X_V_def2_TZVPP_ISOSVP_DME__Frequency_Analysis = "wB97X-V/def2-TZVPP/ISOSVP(DME) Frequency Analysis"
    wB97X_V_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_DME__Unknown = "wB97X-V/def2-TZVPP/ISOSVP(DME) Unknown"
    wB97X_V_def2_TZVPP_ISOSVP_EC__Single_Point = "wB97X-V/def2-TZVPP/ISOSVP(EC) Single Point"
    wB97X_V_def2_TZVPP_ISOSVP_EC__Force = "wB97X-V/def2-TZVPP/ISOSVP(EC) Force"
    wB97X_V_def2_TZVPP_ISOSVP_EC__Geometry_Optimization = "wB97X-V/def2-TZVPP/ISOSVP(EC) Geometry Optimization"
    wB97X_V_def2_TZVPP_ISOSVP_EC__Frequency_Analysis = "wB97X-V/def2-TZVPP/ISOSVP(EC) Frequency Analysis"
    wB97X_V_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_EC__Unknown = "wB97X-V/def2-TZVPP/ISOSVP(EC) Unknown"
    wB97X_V_def2_TZVPP_ISOSVP_EC_EMC__Single_Point = "wB97X-V/def2-TZVPP/ISOSVP(EC/EMC) Single Point"
    wB97X_V_def2_TZVPP_ISOSVP_EC_EMC__Force = "wB97X-V/def2-TZVPP/ISOSVP(EC/EMC) Force"
    wB97X_V_def2_TZVPP_ISOSVP_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVPP/ISOSVP(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPP/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_EC_EMC__Unknown = "wB97X-V/def2-TZVPP/ISOSVP(EC/EMC) Unknown"
    wB97X_V_def2_TZVPP_ISOSVP_PC__Single_Point = "wB97X-V/def2-TZVPP/ISOSVP(PC) Single Point"
    wB97X_V_def2_TZVPP_ISOSVP_PC__Force = "wB97X-V/def2-TZVPP/ISOSVP(PC) Force"
    wB97X_V_def2_TZVPP_ISOSVP_PC__Geometry_Optimization = "wB97X-V/def2-TZVPP/ISOSVP(PC) Geometry Optimization"
    wB97X_V_def2_TZVPP_ISOSVP_PC__Frequency_Analysis = "wB97X-V/def2-TZVPP/ISOSVP(PC) Frequency Analysis"
    wB97X_V_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_ISOSVP_PC__Unknown = "wB97X-V/def2-TZVPP/ISOSVP(PC) Unknown"
    wB97X_V_def2_TZVPP_SMD_WATER__Single_Point = "wB97X-V/def2-TZVPP/SMD(WATER) Single Point"
    wB97X_V_def2_TZVPP_SMD_WATER__Force = "wB97X-V/def2-TZVPP/SMD(WATER) Force"
    wB97X_V_def2_TZVPP_SMD_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPP/SMD(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPP_SMD_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPP/SMD(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_WATER__Unknown = "wB97X-V/def2-TZVPP/SMD(WATER) Unknown"
    wB97X_V_def2_TZVPP_SMD_THF__Single_Point = "wB97X-V/def2-TZVPP/SMD(THF) Single Point"
    wB97X_V_def2_TZVPP_SMD_THF__Force = "wB97X-V/def2-TZVPP/SMD(THF) Force"
    wB97X_V_def2_TZVPP_SMD_THF__Geometry_Optimization = "wB97X-V/def2-TZVPP/SMD(THF) Geometry Optimization"
    wB97X_V_def2_TZVPP_SMD_THF__Frequency_Analysis = "wB97X-V/def2-TZVPP/SMD(THF) Frequency Analysis"
    wB97X_V_def2_TZVPP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_THF__Unknown = "wB97X-V/def2-TZVPP/SMD(THF) Unknown"
    wB97X_V_def2_TZVPP_SMD_G2__Single_Point = "wB97X-V/def2-TZVPP/SMD(G2) Single Point"
    wB97X_V_def2_TZVPP_SMD_G2__Force = "wB97X-V/def2-TZVPP/SMD(G2) Force"
    wB97X_V_def2_TZVPP_SMD_G2__Geometry_Optimization = "wB97X-V/def2-TZVPP/SMD(G2) Geometry Optimization"
    wB97X_V_def2_TZVPP_SMD_G2__Frequency_Analysis = "wB97X-V/def2-TZVPP/SMD(G2) Frequency Analysis"
    wB97X_V_def2_TZVPP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_G2__Unknown = "wB97X-V/def2-TZVPP/SMD(G2) Unknown"
    wB97X_V_def2_TZVPP_SMD_DME__Single_Point = "wB97X-V/def2-TZVPP/SMD(DME) Single Point"
    wB97X_V_def2_TZVPP_SMD_DME__Force = "wB97X-V/def2-TZVPP/SMD(DME) Force"
    wB97X_V_def2_TZVPP_SMD_DME__Geometry_Optimization = "wB97X-V/def2-TZVPP/SMD(DME) Geometry Optimization"
    wB97X_V_def2_TZVPP_SMD_DME__Frequency_Analysis = "wB97X-V/def2-TZVPP/SMD(DME) Frequency Analysis"
    wB97X_V_def2_TZVPP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_DME__Unknown = "wB97X-V/def2-TZVPP/SMD(DME) Unknown"
    wB97X_V_def2_TZVPP_SMD_EC__Single_Point = "wB97X-V/def2-TZVPP/SMD(EC) Single Point"
    wB97X_V_def2_TZVPP_SMD_EC__Force = "wB97X-V/def2-TZVPP/SMD(EC) Force"
    wB97X_V_def2_TZVPP_SMD_EC__Geometry_Optimization = "wB97X-V/def2-TZVPP/SMD(EC) Geometry Optimization"
    wB97X_V_def2_TZVPP_SMD_EC__Frequency_Analysis = "wB97X-V/def2-TZVPP/SMD(EC) Frequency Analysis"
    wB97X_V_def2_TZVPP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_EC__Unknown = "wB97X-V/def2-TZVPP/SMD(EC) Unknown"
    wB97X_V_def2_TZVPP_SMD_EC_EMC__Single_Point = "wB97X-V/def2-TZVPP/SMD(EC/EMC) Single Point"
    wB97X_V_def2_TZVPP_SMD_EC_EMC__Force = "wB97X-V/def2-TZVPP/SMD(EC/EMC) Force"
    wB97X_V_def2_TZVPP_SMD_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVPP/SMD(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVPP_SMD_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPP/SMD(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_EC_EMC__Unknown = "wB97X-V/def2-TZVPP/SMD(EC/EMC) Unknown"
    wB97X_V_def2_TZVPP_SMD_PC__Single_Point = "wB97X-V/def2-TZVPP/SMD(PC) Single Point"
    wB97X_V_def2_TZVPP_SMD_PC__Force = "wB97X-V/def2-TZVPP/SMD(PC) Force"
    wB97X_V_def2_TZVPP_SMD_PC__Geometry_Optimization = "wB97X-V/def2-TZVPP/SMD(PC) Geometry Optimization"
    wB97X_V_def2_TZVPP_SMD_PC__Frequency_Analysis = "wB97X-V/def2-TZVPP/SMD(PC) Frequency Analysis"
    wB97X_V_def2_TZVPP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_SMD_PC__Unknown = "wB97X-V/def2-TZVPP/SMD(PC) Unknown"
    wB97X_V_def2_TZVPP_CMIRS_WATER__Single_Point = "wB97X-V/def2-TZVPP/CMIRS(WATER) Single Point"
    wB97X_V_def2_TZVPP_CMIRS_WATER__Force = "wB97X-V/def2-TZVPP/CMIRS(WATER) Force"
    wB97X_V_def2_TZVPP_CMIRS_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPP/CMIRS(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPP_CMIRS_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPP/CMIRS(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_WATER__Unknown = "wB97X-V/def2-TZVPP/CMIRS(WATER) Unknown"
    wB97X_V_def2_TZVPP_CMIRS_THF__Single_Point = "wB97X-V/def2-TZVPP/CMIRS(THF) Single Point"
    wB97X_V_def2_TZVPP_CMIRS_THF__Force = "wB97X-V/def2-TZVPP/CMIRS(THF) Force"
    wB97X_V_def2_TZVPP_CMIRS_THF__Geometry_Optimization = "wB97X-V/def2-TZVPP/CMIRS(THF) Geometry Optimization"
    wB97X_V_def2_TZVPP_CMIRS_THF__Frequency_Analysis = "wB97X-V/def2-TZVPP/CMIRS(THF) Frequency Analysis"
    wB97X_V_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_THF__Unknown = "wB97X-V/def2-TZVPP/CMIRS(THF) Unknown"
    wB97X_V_def2_TZVPP_CMIRS_G2__Single_Point = "wB97X-V/def2-TZVPP/CMIRS(G2) Single Point"
    wB97X_V_def2_TZVPP_CMIRS_G2__Force = "wB97X-V/def2-TZVPP/CMIRS(G2) Force"
    wB97X_V_def2_TZVPP_CMIRS_G2__Geometry_Optimization = "wB97X-V/def2-TZVPP/CMIRS(G2) Geometry Optimization"
    wB97X_V_def2_TZVPP_CMIRS_G2__Frequency_Analysis = "wB97X-V/def2-TZVPP/CMIRS(G2) Frequency Analysis"
    wB97X_V_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_G2__Unknown = "wB97X-V/def2-TZVPP/CMIRS(G2) Unknown"
    wB97X_V_def2_TZVPP_CMIRS_DME__Single_Point = "wB97X-V/def2-TZVPP/CMIRS(DME) Single Point"
    wB97X_V_def2_TZVPP_CMIRS_DME__Force = "wB97X-V/def2-TZVPP/CMIRS(DME) Force"
    wB97X_V_def2_TZVPP_CMIRS_DME__Geometry_Optimization = "wB97X-V/def2-TZVPP/CMIRS(DME) Geometry Optimization"
    wB97X_V_def2_TZVPP_CMIRS_DME__Frequency_Analysis = "wB97X-V/def2-TZVPP/CMIRS(DME) Frequency Analysis"
    wB97X_V_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_DME__Unknown = "wB97X-V/def2-TZVPP/CMIRS(DME) Unknown"
    wB97X_V_def2_TZVPP_CMIRS_EC__Single_Point = "wB97X-V/def2-TZVPP/CMIRS(EC) Single Point"
    wB97X_V_def2_TZVPP_CMIRS_EC__Force = "wB97X-V/def2-TZVPP/CMIRS(EC) Force"
    wB97X_V_def2_TZVPP_CMIRS_EC__Geometry_Optimization = "wB97X-V/def2-TZVPP/CMIRS(EC) Geometry Optimization"
    wB97X_V_def2_TZVPP_CMIRS_EC__Frequency_Analysis = "wB97X-V/def2-TZVPP/CMIRS(EC) Frequency Analysis"
    wB97X_V_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_EC__Unknown = "wB97X-V/def2-TZVPP/CMIRS(EC) Unknown"
    wB97X_V_def2_TZVPP_CMIRS_EC_EMC__Single_Point = "wB97X-V/def2-TZVPP/CMIRS(EC/EMC) Single Point"
    wB97X_V_def2_TZVPP_CMIRS_EC_EMC__Force = "wB97X-V/def2-TZVPP/CMIRS(EC/EMC) Force"
    wB97X_V_def2_TZVPP_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVPP/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVPP_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPP/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_EC_EMC__Unknown = "wB97X-V/def2-TZVPP/CMIRS(EC/EMC) Unknown"
    wB97X_V_def2_TZVPP_CMIRS_PC__Single_Point = "wB97X-V/def2-TZVPP/CMIRS(PC) Single Point"
    wB97X_V_def2_TZVPP_CMIRS_PC__Force = "wB97X-V/def2-TZVPP/CMIRS(PC) Force"
    wB97X_V_def2_TZVPP_CMIRS_PC__Geometry_Optimization = "wB97X-V/def2-TZVPP/CMIRS(PC) Geometry Optimization"
    wB97X_V_def2_TZVPP_CMIRS_PC__Frequency_Analysis = "wB97X-V/def2-TZVPP/CMIRS(PC) Frequency Analysis"
    wB97X_V_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPP_CMIRS_PC__Unknown = "wB97X-V/def2-TZVPP/CMIRS(PC) Unknown"
    wB97X_V_def2_TZVPPD_VACUUM_Single_Point = "wB97X-V/def2-TZVPPD/VACUUM Single Point"
    wB97X_V_def2_TZVPPD_VACUUM_Force = "wB97X-V/def2-TZVPPD/VACUUM Force"
    wB97X_V_def2_TZVPPD_VACUUM_Geometry_Optimization = "wB97X-V/def2-TZVPPD/VACUUM Geometry Optimization"
    wB97X_V_def2_TZVPPD_VACUUM_Frequency_Analysis = "wB97X-V/def2-TZVPPD/VACUUM Frequency Analysis"
    wB97X_V_def2_TZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_VACUUM_Unknown = "wB97X-V/def2-TZVPPD/VACUUM Unknown"
    wB97X_V_def2_TZVPPD_PCM_WATER__Single_Point = "wB97X-V/def2-TZVPPD/PCM(WATER) Single Point"
    wB97X_V_def2_TZVPPD_PCM_WATER__Force = "wB97X-V/def2-TZVPPD/PCM(WATER) Force"
    wB97X_V_def2_TZVPPD_PCM_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPPD/PCM(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPPD_PCM_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPPD/PCM(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_WATER__Unknown = "wB97X-V/def2-TZVPPD/PCM(WATER) Unknown"
    wB97X_V_def2_TZVPPD_PCM_THF__Single_Point = "wB97X-V/def2-TZVPPD/PCM(THF) Single Point"
    wB97X_V_def2_TZVPPD_PCM_THF__Force = "wB97X-V/def2-TZVPPD/PCM(THF) Force"
    wB97X_V_def2_TZVPPD_PCM_THF__Geometry_Optimization = "wB97X-V/def2-TZVPPD/PCM(THF) Geometry Optimization"
    wB97X_V_def2_TZVPPD_PCM_THF__Frequency_Analysis = "wB97X-V/def2-TZVPPD/PCM(THF) Frequency Analysis"
    wB97X_V_def2_TZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_THF__Unknown = "wB97X-V/def2-TZVPPD/PCM(THF) Unknown"
    wB97X_V_def2_TZVPPD_PCM_G2__Single_Point = "wB97X-V/def2-TZVPPD/PCM(G2) Single Point"
    wB97X_V_def2_TZVPPD_PCM_G2__Force = "wB97X-V/def2-TZVPPD/PCM(G2) Force"
    wB97X_V_def2_TZVPPD_PCM_G2__Geometry_Optimization = "wB97X-V/def2-TZVPPD/PCM(G2) Geometry Optimization"
    wB97X_V_def2_TZVPPD_PCM_G2__Frequency_Analysis = "wB97X-V/def2-TZVPPD/PCM(G2) Frequency Analysis"
    wB97X_V_def2_TZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_G2__Unknown = "wB97X-V/def2-TZVPPD/PCM(G2) Unknown"
    wB97X_V_def2_TZVPPD_PCM_DME__Single_Point = "wB97X-V/def2-TZVPPD/PCM(DME) Single Point"
    wB97X_V_def2_TZVPPD_PCM_DME__Force = "wB97X-V/def2-TZVPPD/PCM(DME) Force"
    wB97X_V_def2_TZVPPD_PCM_DME__Geometry_Optimization = "wB97X-V/def2-TZVPPD/PCM(DME) Geometry Optimization"
    wB97X_V_def2_TZVPPD_PCM_DME__Frequency_Analysis = "wB97X-V/def2-TZVPPD/PCM(DME) Frequency Analysis"
    wB97X_V_def2_TZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_DME__Unknown = "wB97X-V/def2-TZVPPD/PCM(DME) Unknown"
    wB97X_V_def2_TZVPPD_PCM_EC__Single_Point = "wB97X-V/def2-TZVPPD/PCM(EC) Single Point"
    wB97X_V_def2_TZVPPD_PCM_EC__Force = "wB97X-V/def2-TZVPPD/PCM(EC) Force"
    wB97X_V_def2_TZVPPD_PCM_EC__Geometry_Optimization = "wB97X-V/def2-TZVPPD/PCM(EC) Geometry Optimization"
    wB97X_V_def2_TZVPPD_PCM_EC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/PCM(EC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_EC__Unknown = "wB97X-V/def2-TZVPPD/PCM(EC) Unknown"
    wB97X_V_def2_TZVPPD_PCM_EC_EMC__Single_Point = "wB97X-V/def2-TZVPPD/PCM(EC/EMC) Single Point"
    wB97X_V_def2_TZVPPD_PCM_EC_EMC__Force = "wB97X-V/def2-TZVPPD/PCM(EC/EMC) Force"
    wB97X_V_def2_TZVPPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVPPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVPPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_EC_EMC__Unknown = "wB97X-V/def2-TZVPPD/PCM(EC/EMC) Unknown"
    wB97X_V_def2_TZVPPD_PCM_PC__Single_Point = "wB97X-V/def2-TZVPPD/PCM(PC) Single Point"
    wB97X_V_def2_TZVPPD_PCM_PC__Force = "wB97X-V/def2-TZVPPD/PCM(PC) Force"
    wB97X_V_def2_TZVPPD_PCM_PC__Geometry_Optimization = "wB97X-V/def2-TZVPPD/PCM(PC) Geometry Optimization"
    wB97X_V_def2_TZVPPD_PCM_PC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/PCM(PC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_PCM_PC__Unknown = "wB97X-V/def2-TZVPPD/PCM(PC) Unknown"
    wB97X_V_def2_TZVPPD_ISOSVP_WATER__Single_Point = "wB97X-V/def2-TZVPPD/ISOSVP(WATER) Single Point"
    wB97X_V_def2_TZVPPD_ISOSVP_WATER__Force = "wB97X-V/def2-TZVPPD/ISOSVP(WATER) Force"
    wB97X_V_def2_TZVPPD_ISOSVP_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPPD/ISOSVP(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_WATER__Unknown = "wB97X-V/def2-TZVPPD/ISOSVP(WATER) Unknown"
    wB97X_V_def2_TZVPPD_ISOSVP_THF__Single_Point = "wB97X-V/def2-TZVPPD/ISOSVP(THF) Single Point"
    wB97X_V_def2_TZVPPD_ISOSVP_THF__Force = "wB97X-V/def2-TZVPPD/ISOSVP(THF) Force"
    wB97X_V_def2_TZVPPD_ISOSVP_THF__Geometry_Optimization = "wB97X-V/def2-TZVPPD/ISOSVP(THF) Geometry Optimization"
    wB97X_V_def2_TZVPPD_ISOSVP_THF__Frequency_Analysis = "wB97X-V/def2-TZVPPD/ISOSVP(THF) Frequency Analysis"
    wB97X_V_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_THF__Unknown = "wB97X-V/def2-TZVPPD/ISOSVP(THF) Unknown"
    wB97X_V_def2_TZVPPD_ISOSVP_G2__Single_Point = "wB97X-V/def2-TZVPPD/ISOSVP(G2) Single Point"
    wB97X_V_def2_TZVPPD_ISOSVP_G2__Force = "wB97X-V/def2-TZVPPD/ISOSVP(G2) Force"
    wB97X_V_def2_TZVPPD_ISOSVP_G2__Geometry_Optimization = "wB97X-V/def2-TZVPPD/ISOSVP(G2) Geometry Optimization"
    wB97X_V_def2_TZVPPD_ISOSVP_G2__Frequency_Analysis = "wB97X-V/def2-TZVPPD/ISOSVP(G2) Frequency Analysis"
    wB97X_V_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_G2__Unknown = "wB97X-V/def2-TZVPPD/ISOSVP(G2) Unknown"
    wB97X_V_def2_TZVPPD_ISOSVP_DME__Single_Point = "wB97X-V/def2-TZVPPD/ISOSVP(DME) Single Point"
    wB97X_V_def2_TZVPPD_ISOSVP_DME__Force = "wB97X-V/def2-TZVPPD/ISOSVP(DME) Force"
    wB97X_V_def2_TZVPPD_ISOSVP_DME__Geometry_Optimization = "wB97X-V/def2-TZVPPD/ISOSVP(DME) Geometry Optimization"
    wB97X_V_def2_TZVPPD_ISOSVP_DME__Frequency_Analysis = "wB97X-V/def2-TZVPPD/ISOSVP(DME) Frequency Analysis"
    wB97X_V_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_DME__Unknown = "wB97X-V/def2-TZVPPD/ISOSVP(DME) Unknown"
    wB97X_V_def2_TZVPPD_ISOSVP_EC__Single_Point = "wB97X-V/def2-TZVPPD/ISOSVP(EC) Single Point"
    wB97X_V_def2_TZVPPD_ISOSVP_EC__Force = "wB97X-V/def2-TZVPPD/ISOSVP(EC) Force"
    wB97X_V_def2_TZVPPD_ISOSVP_EC__Geometry_Optimization = "wB97X-V/def2-TZVPPD/ISOSVP(EC) Geometry Optimization"
    wB97X_V_def2_TZVPPD_ISOSVP_EC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/ISOSVP(EC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_EC__Unknown = "wB97X-V/def2-TZVPPD/ISOSVP(EC) Unknown"
    wB97X_V_def2_TZVPPD_ISOSVP_EC_EMC__Single_Point = "wB97X-V/def2-TZVPPD/ISOSVP(EC/EMC) Single Point"
    wB97X_V_def2_TZVPPD_ISOSVP_EC_EMC__Force = "wB97X-V/def2-TZVPPD/ISOSVP(EC/EMC) Force"
    wB97X_V_def2_TZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_EC_EMC__Unknown = "wB97X-V/def2-TZVPPD/ISOSVP(EC/EMC) Unknown"
    wB97X_V_def2_TZVPPD_ISOSVP_PC__Single_Point = "wB97X-V/def2-TZVPPD/ISOSVP(PC) Single Point"
    wB97X_V_def2_TZVPPD_ISOSVP_PC__Force = "wB97X-V/def2-TZVPPD/ISOSVP(PC) Force"
    wB97X_V_def2_TZVPPD_ISOSVP_PC__Geometry_Optimization = "wB97X-V/def2-TZVPPD/ISOSVP(PC) Geometry Optimization"
    wB97X_V_def2_TZVPPD_ISOSVP_PC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/ISOSVP(PC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_ISOSVP_PC__Unknown = "wB97X-V/def2-TZVPPD/ISOSVP(PC) Unknown"
    wB97X_V_def2_TZVPPD_SMD_WATER__Single_Point = "wB97X-V/def2-TZVPPD/SMD(WATER) Single Point"
    wB97X_V_def2_TZVPPD_SMD_WATER__Force = "wB97X-V/def2-TZVPPD/SMD(WATER) Force"
    wB97X_V_def2_TZVPPD_SMD_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPPD/SMD(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPPD_SMD_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPPD/SMD(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_WATER__Unknown = "wB97X-V/def2-TZVPPD/SMD(WATER) Unknown"
    wB97X_V_def2_TZVPPD_SMD_THF__Single_Point = "wB97X-V/def2-TZVPPD/SMD(THF) Single Point"
    wB97X_V_def2_TZVPPD_SMD_THF__Force = "wB97X-V/def2-TZVPPD/SMD(THF) Force"
    wB97X_V_def2_TZVPPD_SMD_THF__Geometry_Optimization = "wB97X-V/def2-TZVPPD/SMD(THF) Geometry Optimization"
    wB97X_V_def2_TZVPPD_SMD_THF__Frequency_Analysis = "wB97X-V/def2-TZVPPD/SMD(THF) Frequency Analysis"
    wB97X_V_def2_TZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_THF__Unknown = "wB97X-V/def2-TZVPPD/SMD(THF) Unknown"
    wB97X_V_def2_TZVPPD_SMD_G2__Single_Point = "wB97X-V/def2-TZVPPD/SMD(G2) Single Point"
    wB97X_V_def2_TZVPPD_SMD_G2__Force = "wB97X-V/def2-TZVPPD/SMD(G2) Force"
    wB97X_V_def2_TZVPPD_SMD_G2__Geometry_Optimization = "wB97X-V/def2-TZVPPD/SMD(G2) Geometry Optimization"
    wB97X_V_def2_TZVPPD_SMD_G2__Frequency_Analysis = "wB97X-V/def2-TZVPPD/SMD(G2) Frequency Analysis"
    wB97X_V_def2_TZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_G2__Unknown = "wB97X-V/def2-TZVPPD/SMD(G2) Unknown"
    wB97X_V_def2_TZVPPD_SMD_DME__Single_Point = "wB97X-V/def2-TZVPPD/SMD(DME) Single Point"
    wB97X_V_def2_TZVPPD_SMD_DME__Force = "wB97X-V/def2-TZVPPD/SMD(DME) Force"
    wB97X_V_def2_TZVPPD_SMD_DME__Geometry_Optimization = "wB97X-V/def2-TZVPPD/SMD(DME) Geometry Optimization"
    wB97X_V_def2_TZVPPD_SMD_DME__Frequency_Analysis = "wB97X-V/def2-TZVPPD/SMD(DME) Frequency Analysis"
    wB97X_V_def2_TZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_DME__Unknown = "wB97X-V/def2-TZVPPD/SMD(DME) Unknown"
    wB97X_V_def2_TZVPPD_SMD_EC__Single_Point = "wB97X-V/def2-TZVPPD/SMD(EC) Single Point"
    wB97X_V_def2_TZVPPD_SMD_EC__Force = "wB97X-V/def2-TZVPPD/SMD(EC) Force"
    wB97X_V_def2_TZVPPD_SMD_EC__Geometry_Optimization = "wB97X-V/def2-TZVPPD/SMD(EC) Geometry Optimization"
    wB97X_V_def2_TZVPPD_SMD_EC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/SMD(EC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_EC__Unknown = "wB97X-V/def2-TZVPPD/SMD(EC) Unknown"
    wB97X_V_def2_TZVPPD_SMD_EC_EMC__Single_Point = "wB97X-V/def2-TZVPPD/SMD(EC/EMC) Single Point"
    wB97X_V_def2_TZVPPD_SMD_EC_EMC__Force = "wB97X-V/def2-TZVPPD/SMD(EC/EMC) Force"
    wB97X_V_def2_TZVPPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVPPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVPPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_EC_EMC__Unknown = "wB97X-V/def2-TZVPPD/SMD(EC/EMC) Unknown"
    wB97X_V_def2_TZVPPD_SMD_PC__Single_Point = "wB97X-V/def2-TZVPPD/SMD(PC) Single Point"
    wB97X_V_def2_TZVPPD_SMD_PC__Force = "wB97X-V/def2-TZVPPD/SMD(PC) Force"
    wB97X_V_def2_TZVPPD_SMD_PC__Geometry_Optimization = "wB97X-V/def2-TZVPPD/SMD(PC) Geometry Optimization"
    wB97X_V_def2_TZVPPD_SMD_PC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/SMD(PC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_SMD_PC__Unknown = "wB97X-V/def2-TZVPPD/SMD(PC) Unknown"
    wB97X_V_def2_TZVPPD_CMIRS_WATER__Single_Point = "wB97X-V/def2-TZVPPD/CMIRS(WATER) Single Point"
    wB97X_V_def2_TZVPPD_CMIRS_WATER__Force = "wB97X-V/def2-TZVPPD/CMIRS(WATER) Force"
    wB97X_V_def2_TZVPPD_CMIRS_WATER__Geometry_Optimization = "wB97X-V/def2-TZVPPD/CMIRS(WATER) Geometry Optimization"
    wB97X_V_def2_TZVPPD_CMIRS_WATER__Frequency_Analysis = "wB97X-V/def2-TZVPPD/CMIRS(WATER) Frequency Analysis"
    wB97X_V_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_WATER__Unknown = "wB97X-V/def2-TZVPPD/CMIRS(WATER) Unknown"
    wB97X_V_def2_TZVPPD_CMIRS_THF__Single_Point = "wB97X-V/def2-TZVPPD/CMIRS(THF) Single Point"
    wB97X_V_def2_TZVPPD_CMIRS_THF__Force = "wB97X-V/def2-TZVPPD/CMIRS(THF) Force"
    wB97X_V_def2_TZVPPD_CMIRS_THF__Geometry_Optimization = "wB97X-V/def2-TZVPPD/CMIRS(THF) Geometry Optimization"
    wB97X_V_def2_TZVPPD_CMIRS_THF__Frequency_Analysis = "wB97X-V/def2-TZVPPD/CMIRS(THF) Frequency Analysis"
    wB97X_V_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_THF__Unknown = "wB97X-V/def2-TZVPPD/CMIRS(THF) Unknown"
    wB97X_V_def2_TZVPPD_CMIRS_G2__Single_Point = "wB97X-V/def2-TZVPPD/CMIRS(G2) Single Point"
    wB97X_V_def2_TZVPPD_CMIRS_G2__Force = "wB97X-V/def2-TZVPPD/CMIRS(G2) Force"
    wB97X_V_def2_TZVPPD_CMIRS_G2__Geometry_Optimization = "wB97X-V/def2-TZVPPD/CMIRS(G2) Geometry Optimization"
    wB97X_V_def2_TZVPPD_CMIRS_G2__Frequency_Analysis = "wB97X-V/def2-TZVPPD/CMIRS(G2) Frequency Analysis"
    wB97X_V_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_G2__Unknown = "wB97X-V/def2-TZVPPD/CMIRS(G2) Unknown"
    wB97X_V_def2_TZVPPD_CMIRS_DME__Single_Point = "wB97X-V/def2-TZVPPD/CMIRS(DME) Single Point"
    wB97X_V_def2_TZVPPD_CMIRS_DME__Force = "wB97X-V/def2-TZVPPD/CMIRS(DME) Force"
    wB97X_V_def2_TZVPPD_CMIRS_DME__Geometry_Optimization = "wB97X-V/def2-TZVPPD/CMIRS(DME) Geometry Optimization"
    wB97X_V_def2_TZVPPD_CMIRS_DME__Frequency_Analysis = "wB97X-V/def2-TZVPPD/CMIRS(DME) Frequency Analysis"
    wB97X_V_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_DME__Unknown = "wB97X-V/def2-TZVPPD/CMIRS(DME) Unknown"
    wB97X_V_def2_TZVPPD_CMIRS_EC__Single_Point = "wB97X-V/def2-TZVPPD/CMIRS(EC) Single Point"
    wB97X_V_def2_TZVPPD_CMIRS_EC__Force = "wB97X-V/def2-TZVPPD/CMIRS(EC) Force"
    wB97X_V_def2_TZVPPD_CMIRS_EC__Geometry_Optimization = "wB97X-V/def2-TZVPPD/CMIRS(EC) Geometry Optimization"
    wB97X_V_def2_TZVPPD_CMIRS_EC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/CMIRS(EC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_EC__Unknown = "wB97X-V/def2-TZVPPD/CMIRS(EC) Unknown"
    wB97X_V_def2_TZVPPD_CMIRS_EC_EMC__Single_Point = "wB97X-V/def2-TZVPPD/CMIRS(EC/EMC) Single Point"
    wB97X_V_def2_TZVPPD_CMIRS_EC_EMC__Force = "wB97X-V/def2-TZVPPD/CMIRS(EC/EMC) Force"
    wB97X_V_def2_TZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-V/def2-TZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_V_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_EC_EMC__Unknown = "wB97X-V/def2-TZVPPD/CMIRS(EC/EMC) Unknown"
    wB97X_V_def2_TZVPPD_CMIRS_PC__Single_Point = "wB97X-V/def2-TZVPPD/CMIRS(PC) Single Point"
    wB97X_V_def2_TZVPPD_CMIRS_PC__Force = "wB97X-V/def2-TZVPPD/CMIRS(PC) Force"
    wB97X_V_def2_TZVPPD_CMIRS_PC__Geometry_Optimization = "wB97X-V/def2-TZVPPD/CMIRS(PC) Geometry Optimization"
    wB97X_V_def2_TZVPPD_CMIRS_PC__Frequency_Analysis = "wB97X-V/def2-TZVPPD/CMIRS(PC) Frequency Analysis"
    wB97X_V_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-TZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_TZVPPD_CMIRS_PC__Unknown = "wB97X-V/def2-TZVPPD/CMIRS(PC) Unknown"
    wB97X_V_def2_QZVPPD_VACUUM_Single_Point = "wB97X-V/def2-QZVPPD/VACUUM Single Point"
    wB97X_V_def2_QZVPPD_VACUUM_Force = "wB97X-V/def2-QZVPPD/VACUUM Force"
    wB97X_V_def2_QZVPPD_VACUUM_Geometry_Optimization = "wB97X-V/def2-QZVPPD/VACUUM Geometry Optimization"
    wB97X_V_def2_QZVPPD_VACUUM_Frequency_Analysis = "wB97X-V/def2-QZVPPD/VACUUM Frequency Analysis"
    wB97X_V_def2_QZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/VACUUM Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_VACUUM_Unknown = "wB97X-V/def2-QZVPPD/VACUUM Unknown"
    wB97X_V_def2_QZVPPD_PCM_WATER__Single_Point = "wB97X-V/def2-QZVPPD/PCM(WATER) Single Point"
    wB97X_V_def2_QZVPPD_PCM_WATER__Force = "wB97X-V/def2-QZVPPD/PCM(WATER) Force"
    wB97X_V_def2_QZVPPD_PCM_WATER__Geometry_Optimization = "wB97X-V/def2-QZVPPD/PCM(WATER) Geometry Optimization"
    wB97X_V_def2_QZVPPD_PCM_WATER__Frequency_Analysis = "wB97X-V/def2-QZVPPD/PCM(WATER) Frequency Analysis"
    wB97X_V_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_WATER__Unknown = "wB97X-V/def2-QZVPPD/PCM(WATER) Unknown"
    wB97X_V_def2_QZVPPD_PCM_THF__Single_Point = "wB97X-V/def2-QZVPPD/PCM(THF) Single Point"
    wB97X_V_def2_QZVPPD_PCM_THF__Force = "wB97X-V/def2-QZVPPD/PCM(THF) Force"
    wB97X_V_def2_QZVPPD_PCM_THF__Geometry_Optimization = "wB97X-V/def2-QZVPPD/PCM(THF) Geometry Optimization"
    wB97X_V_def2_QZVPPD_PCM_THF__Frequency_Analysis = "wB97X-V/def2-QZVPPD/PCM(THF) Frequency Analysis"
    wB97X_V_def2_QZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_THF__Unknown = "wB97X-V/def2-QZVPPD/PCM(THF) Unknown"
    wB97X_V_def2_QZVPPD_PCM_G2__Single_Point = "wB97X-V/def2-QZVPPD/PCM(G2) Single Point"
    wB97X_V_def2_QZVPPD_PCM_G2__Force = "wB97X-V/def2-QZVPPD/PCM(G2) Force"
    wB97X_V_def2_QZVPPD_PCM_G2__Geometry_Optimization = "wB97X-V/def2-QZVPPD/PCM(G2) Geometry Optimization"
    wB97X_V_def2_QZVPPD_PCM_G2__Frequency_Analysis = "wB97X-V/def2-QZVPPD/PCM(G2) Frequency Analysis"
    wB97X_V_def2_QZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_G2__Unknown = "wB97X-V/def2-QZVPPD/PCM(G2) Unknown"
    wB97X_V_def2_QZVPPD_PCM_DME__Single_Point = "wB97X-V/def2-QZVPPD/PCM(DME) Single Point"
    wB97X_V_def2_QZVPPD_PCM_DME__Force = "wB97X-V/def2-QZVPPD/PCM(DME) Force"
    wB97X_V_def2_QZVPPD_PCM_DME__Geometry_Optimization = "wB97X-V/def2-QZVPPD/PCM(DME) Geometry Optimization"
    wB97X_V_def2_QZVPPD_PCM_DME__Frequency_Analysis = "wB97X-V/def2-QZVPPD/PCM(DME) Frequency Analysis"
    wB97X_V_def2_QZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_DME__Unknown = "wB97X-V/def2-QZVPPD/PCM(DME) Unknown"
    wB97X_V_def2_QZVPPD_PCM_EC__Single_Point = "wB97X-V/def2-QZVPPD/PCM(EC) Single Point"
    wB97X_V_def2_QZVPPD_PCM_EC__Force = "wB97X-V/def2-QZVPPD/PCM(EC) Force"
    wB97X_V_def2_QZVPPD_PCM_EC__Geometry_Optimization = "wB97X-V/def2-QZVPPD/PCM(EC) Geometry Optimization"
    wB97X_V_def2_QZVPPD_PCM_EC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/PCM(EC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_EC__Unknown = "wB97X-V/def2-QZVPPD/PCM(EC) Unknown"
    wB97X_V_def2_QZVPPD_PCM_EC_EMC__Single_Point = "wB97X-V/def2-QZVPPD/PCM(EC/EMC) Single Point"
    wB97X_V_def2_QZVPPD_PCM_EC_EMC__Force = "wB97X-V/def2-QZVPPD/PCM(EC/EMC) Force"
    wB97X_V_def2_QZVPPD_PCM_EC_EMC__Geometry_Optimization = "wB97X-V/def2-QZVPPD/PCM(EC/EMC) Geometry Optimization"
    wB97X_V_def2_QZVPPD_PCM_EC_EMC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/PCM(EC/EMC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_EC_EMC__Unknown = "wB97X-V/def2-QZVPPD/PCM(EC/EMC) Unknown"
    wB97X_V_def2_QZVPPD_PCM_PC__Single_Point = "wB97X-V/def2-QZVPPD/PCM(PC) Single Point"
    wB97X_V_def2_QZVPPD_PCM_PC__Force = "wB97X-V/def2-QZVPPD/PCM(PC) Force"
    wB97X_V_def2_QZVPPD_PCM_PC__Geometry_Optimization = "wB97X-V/def2-QZVPPD/PCM(PC) Geometry Optimization"
    wB97X_V_def2_QZVPPD_PCM_PC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/PCM(PC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_PCM_PC__Unknown = "wB97X-V/def2-QZVPPD/PCM(PC) Unknown"
    wB97X_V_def2_QZVPPD_ISOSVP_WATER__Single_Point = "wB97X-V/def2-QZVPPD/ISOSVP(WATER) Single Point"
    wB97X_V_def2_QZVPPD_ISOSVP_WATER__Force = "wB97X-V/def2-QZVPPD/ISOSVP(WATER) Force"
    wB97X_V_def2_QZVPPD_ISOSVP_WATER__Geometry_Optimization = "wB97X-V/def2-QZVPPD/ISOSVP(WATER) Geometry Optimization"
    wB97X_V_def2_QZVPPD_ISOSVP_WATER__Frequency_Analysis = "wB97X-V/def2-QZVPPD/ISOSVP(WATER) Frequency Analysis"
    wB97X_V_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_WATER__Unknown = "wB97X-V/def2-QZVPPD/ISOSVP(WATER) Unknown"
    wB97X_V_def2_QZVPPD_ISOSVP_THF__Single_Point = "wB97X-V/def2-QZVPPD/ISOSVP(THF) Single Point"
    wB97X_V_def2_QZVPPD_ISOSVP_THF__Force = "wB97X-V/def2-QZVPPD/ISOSVP(THF) Force"
    wB97X_V_def2_QZVPPD_ISOSVP_THF__Geometry_Optimization = "wB97X-V/def2-QZVPPD/ISOSVP(THF) Geometry Optimization"
    wB97X_V_def2_QZVPPD_ISOSVP_THF__Frequency_Analysis = "wB97X-V/def2-QZVPPD/ISOSVP(THF) Frequency Analysis"
    wB97X_V_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_THF__Unknown = "wB97X-V/def2-QZVPPD/ISOSVP(THF) Unknown"
    wB97X_V_def2_QZVPPD_ISOSVP_G2__Single_Point = "wB97X-V/def2-QZVPPD/ISOSVP(G2) Single Point"
    wB97X_V_def2_QZVPPD_ISOSVP_G2__Force = "wB97X-V/def2-QZVPPD/ISOSVP(G2) Force"
    wB97X_V_def2_QZVPPD_ISOSVP_G2__Geometry_Optimization = "wB97X-V/def2-QZVPPD/ISOSVP(G2) Geometry Optimization"
    wB97X_V_def2_QZVPPD_ISOSVP_G2__Frequency_Analysis = "wB97X-V/def2-QZVPPD/ISOSVP(G2) Frequency Analysis"
    wB97X_V_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_G2__Unknown = "wB97X-V/def2-QZVPPD/ISOSVP(G2) Unknown"
    wB97X_V_def2_QZVPPD_ISOSVP_DME__Single_Point = "wB97X-V/def2-QZVPPD/ISOSVP(DME) Single Point"
    wB97X_V_def2_QZVPPD_ISOSVP_DME__Force = "wB97X-V/def2-QZVPPD/ISOSVP(DME) Force"
    wB97X_V_def2_QZVPPD_ISOSVP_DME__Geometry_Optimization = "wB97X-V/def2-QZVPPD/ISOSVP(DME) Geometry Optimization"
    wB97X_V_def2_QZVPPD_ISOSVP_DME__Frequency_Analysis = "wB97X-V/def2-QZVPPD/ISOSVP(DME) Frequency Analysis"
    wB97X_V_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_DME__Unknown = "wB97X-V/def2-QZVPPD/ISOSVP(DME) Unknown"
    wB97X_V_def2_QZVPPD_ISOSVP_EC__Single_Point = "wB97X-V/def2-QZVPPD/ISOSVP(EC) Single Point"
    wB97X_V_def2_QZVPPD_ISOSVP_EC__Force = "wB97X-V/def2-QZVPPD/ISOSVP(EC) Force"
    wB97X_V_def2_QZVPPD_ISOSVP_EC__Geometry_Optimization = "wB97X-V/def2-QZVPPD/ISOSVP(EC) Geometry Optimization"
    wB97X_V_def2_QZVPPD_ISOSVP_EC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/ISOSVP(EC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_EC__Unknown = "wB97X-V/def2-QZVPPD/ISOSVP(EC) Unknown"
    wB97X_V_def2_QZVPPD_ISOSVP_EC_EMC__Single_Point = "wB97X-V/def2-QZVPPD/ISOSVP(EC/EMC) Single Point"
    wB97X_V_def2_QZVPPD_ISOSVP_EC_EMC__Force = "wB97X-V/def2-QZVPPD/ISOSVP(EC/EMC) Force"
    wB97X_V_def2_QZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_EC_EMC__Unknown = "wB97X-V/def2-QZVPPD/ISOSVP(EC/EMC) Unknown"
    wB97X_V_def2_QZVPPD_ISOSVP_PC__Single_Point = "wB97X-V/def2-QZVPPD/ISOSVP(PC) Single Point"
    wB97X_V_def2_QZVPPD_ISOSVP_PC__Force = "wB97X-V/def2-QZVPPD/ISOSVP(PC) Force"
    wB97X_V_def2_QZVPPD_ISOSVP_PC__Geometry_Optimization = "wB97X-V/def2-QZVPPD/ISOSVP(PC) Geometry Optimization"
    wB97X_V_def2_QZVPPD_ISOSVP_PC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/ISOSVP(PC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_ISOSVP_PC__Unknown = "wB97X-V/def2-QZVPPD/ISOSVP(PC) Unknown"
    wB97X_V_def2_QZVPPD_SMD_WATER__Single_Point = "wB97X-V/def2-QZVPPD/SMD(WATER) Single Point"
    wB97X_V_def2_QZVPPD_SMD_WATER__Force = "wB97X-V/def2-QZVPPD/SMD(WATER) Force"
    wB97X_V_def2_QZVPPD_SMD_WATER__Geometry_Optimization = "wB97X-V/def2-QZVPPD/SMD(WATER) Geometry Optimization"
    wB97X_V_def2_QZVPPD_SMD_WATER__Frequency_Analysis = "wB97X-V/def2-QZVPPD/SMD(WATER) Frequency Analysis"
    wB97X_V_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_WATER__Unknown = "wB97X-V/def2-QZVPPD/SMD(WATER) Unknown"
    wB97X_V_def2_QZVPPD_SMD_THF__Single_Point = "wB97X-V/def2-QZVPPD/SMD(THF) Single Point"
    wB97X_V_def2_QZVPPD_SMD_THF__Force = "wB97X-V/def2-QZVPPD/SMD(THF) Force"
    wB97X_V_def2_QZVPPD_SMD_THF__Geometry_Optimization = "wB97X-V/def2-QZVPPD/SMD(THF) Geometry Optimization"
    wB97X_V_def2_QZVPPD_SMD_THF__Frequency_Analysis = "wB97X-V/def2-QZVPPD/SMD(THF) Frequency Analysis"
    wB97X_V_def2_QZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_THF__Unknown = "wB97X-V/def2-QZVPPD/SMD(THF) Unknown"
    wB97X_V_def2_QZVPPD_SMD_G2__Single_Point = "wB97X-V/def2-QZVPPD/SMD(G2) Single Point"
    wB97X_V_def2_QZVPPD_SMD_G2__Force = "wB97X-V/def2-QZVPPD/SMD(G2) Force"
    wB97X_V_def2_QZVPPD_SMD_G2__Geometry_Optimization = "wB97X-V/def2-QZVPPD/SMD(G2) Geometry Optimization"
    wB97X_V_def2_QZVPPD_SMD_G2__Frequency_Analysis = "wB97X-V/def2-QZVPPD/SMD(G2) Frequency Analysis"
    wB97X_V_def2_QZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_G2__Unknown = "wB97X-V/def2-QZVPPD/SMD(G2) Unknown"
    wB97X_V_def2_QZVPPD_SMD_DME__Single_Point = "wB97X-V/def2-QZVPPD/SMD(DME) Single Point"
    wB97X_V_def2_QZVPPD_SMD_DME__Force = "wB97X-V/def2-QZVPPD/SMD(DME) Force"
    wB97X_V_def2_QZVPPD_SMD_DME__Geometry_Optimization = "wB97X-V/def2-QZVPPD/SMD(DME) Geometry Optimization"
    wB97X_V_def2_QZVPPD_SMD_DME__Frequency_Analysis = "wB97X-V/def2-QZVPPD/SMD(DME) Frequency Analysis"
    wB97X_V_def2_QZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_DME__Unknown = "wB97X-V/def2-QZVPPD/SMD(DME) Unknown"
    wB97X_V_def2_QZVPPD_SMD_EC__Single_Point = "wB97X-V/def2-QZVPPD/SMD(EC) Single Point"
    wB97X_V_def2_QZVPPD_SMD_EC__Force = "wB97X-V/def2-QZVPPD/SMD(EC) Force"
    wB97X_V_def2_QZVPPD_SMD_EC__Geometry_Optimization = "wB97X-V/def2-QZVPPD/SMD(EC) Geometry Optimization"
    wB97X_V_def2_QZVPPD_SMD_EC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/SMD(EC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_EC__Unknown = "wB97X-V/def2-QZVPPD/SMD(EC) Unknown"
    wB97X_V_def2_QZVPPD_SMD_EC_EMC__Single_Point = "wB97X-V/def2-QZVPPD/SMD(EC/EMC) Single Point"
    wB97X_V_def2_QZVPPD_SMD_EC_EMC__Force = "wB97X-V/def2-QZVPPD/SMD(EC/EMC) Force"
    wB97X_V_def2_QZVPPD_SMD_EC_EMC__Geometry_Optimization = "wB97X-V/def2-QZVPPD/SMD(EC/EMC) Geometry Optimization"
    wB97X_V_def2_QZVPPD_SMD_EC_EMC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/SMD(EC/EMC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_EC_EMC__Unknown = "wB97X-V/def2-QZVPPD/SMD(EC/EMC) Unknown"
    wB97X_V_def2_QZVPPD_SMD_PC__Single_Point = "wB97X-V/def2-QZVPPD/SMD(PC) Single Point"
    wB97X_V_def2_QZVPPD_SMD_PC__Force = "wB97X-V/def2-QZVPPD/SMD(PC) Force"
    wB97X_V_def2_QZVPPD_SMD_PC__Geometry_Optimization = "wB97X-V/def2-QZVPPD/SMD(PC) Geometry Optimization"
    wB97X_V_def2_QZVPPD_SMD_PC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/SMD(PC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_SMD_PC__Unknown = "wB97X-V/def2-QZVPPD/SMD(PC) Unknown"
    wB97X_V_def2_QZVPPD_CMIRS_WATER__Single_Point = "wB97X-V/def2-QZVPPD/CMIRS(WATER) Single Point"
    wB97X_V_def2_QZVPPD_CMIRS_WATER__Force = "wB97X-V/def2-QZVPPD/CMIRS(WATER) Force"
    wB97X_V_def2_QZVPPD_CMIRS_WATER__Geometry_Optimization = "wB97X-V/def2-QZVPPD/CMIRS(WATER) Geometry Optimization"
    wB97X_V_def2_QZVPPD_CMIRS_WATER__Frequency_Analysis = "wB97X-V/def2-QZVPPD/CMIRS(WATER) Frequency Analysis"
    wB97X_V_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_WATER__Unknown = "wB97X-V/def2-QZVPPD/CMIRS(WATER) Unknown"
    wB97X_V_def2_QZVPPD_CMIRS_THF__Single_Point = "wB97X-V/def2-QZVPPD/CMIRS(THF) Single Point"
    wB97X_V_def2_QZVPPD_CMIRS_THF__Force = "wB97X-V/def2-QZVPPD/CMIRS(THF) Force"
    wB97X_V_def2_QZVPPD_CMIRS_THF__Geometry_Optimization = "wB97X-V/def2-QZVPPD/CMIRS(THF) Geometry Optimization"
    wB97X_V_def2_QZVPPD_CMIRS_THF__Frequency_Analysis = "wB97X-V/def2-QZVPPD/CMIRS(THF) Frequency Analysis"
    wB97X_V_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_THF__Unknown = "wB97X-V/def2-QZVPPD/CMIRS(THF) Unknown"
    wB97X_V_def2_QZVPPD_CMIRS_G2__Single_Point = "wB97X-V/def2-QZVPPD/CMIRS(G2) Single Point"
    wB97X_V_def2_QZVPPD_CMIRS_G2__Force = "wB97X-V/def2-QZVPPD/CMIRS(G2) Force"
    wB97X_V_def2_QZVPPD_CMIRS_G2__Geometry_Optimization = "wB97X-V/def2-QZVPPD/CMIRS(G2) Geometry Optimization"
    wB97X_V_def2_QZVPPD_CMIRS_G2__Frequency_Analysis = "wB97X-V/def2-QZVPPD/CMIRS(G2) Frequency Analysis"
    wB97X_V_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_G2__Unknown = "wB97X-V/def2-QZVPPD/CMIRS(G2) Unknown"
    wB97X_V_def2_QZVPPD_CMIRS_DME__Single_Point = "wB97X-V/def2-QZVPPD/CMIRS(DME) Single Point"
    wB97X_V_def2_QZVPPD_CMIRS_DME__Force = "wB97X-V/def2-QZVPPD/CMIRS(DME) Force"
    wB97X_V_def2_QZVPPD_CMIRS_DME__Geometry_Optimization = "wB97X-V/def2-QZVPPD/CMIRS(DME) Geometry Optimization"
    wB97X_V_def2_QZVPPD_CMIRS_DME__Frequency_Analysis = "wB97X-V/def2-QZVPPD/CMIRS(DME) Frequency Analysis"
    wB97X_V_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_DME__Unknown = "wB97X-V/def2-QZVPPD/CMIRS(DME) Unknown"
    wB97X_V_def2_QZVPPD_CMIRS_EC__Single_Point = "wB97X-V/def2-QZVPPD/CMIRS(EC) Single Point"
    wB97X_V_def2_QZVPPD_CMIRS_EC__Force = "wB97X-V/def2-QZVPPD/CMIRS(EC) Force"
    wB97X_V_def2_QZVPPD_CMIRS_EC__Geometry_Optimization = "wB97X-V/def2-QZVPPD/CMIRS(EC) Geometry Optimization"
    wB97X_V_def2_QZVPPD_CMIRS_EC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/CMIRS(EC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_EC__Unknown = "wB97X-V/def2-QZVPPD/CMIRS(EC) Unknown"
    wB97X_V_def2_QZVPPD_CMIRS_EC_EMC__Single_Point = "wB97X-V/def2-QZVPPD/CMIRS(EC/EMC) Single Point"
    wB97X_V_def2_QZVPPD_CMIRS_EC_EMC__Force = "wB97X-V/def2-QZVPPD/CMIRS(EC/EMC) Force"
    wB97X_V_def2_QZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97X-V/def2-QZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97X_V_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_EC_EMC__Unknown = "wB97X-V/def2-QZVPPD/CMIRS(EC/EMC) Unknown"
    wB97X_V_def2_QZVPPD_CMIRS_PC__Single_Point = "wB97X-V/def2-QZVPPD/CMIRS(PC) Single Point"
    wB97X_V_def2_QZVPPD_CMIRS_PC__Force = "wB97X-V/def2-QZVPPD/CMIRS(PC) Force"
    wB97X_V_def2_QZVPPD_CMIRS_PC__Geometry_Optimization = "wB97X-V/def2-QZVPPD/CMIRS(PC) Geometry Optimization"
    wB97X_V_def2_QZVPPD_CMIRS_PC__Frequency_Analysis = "wB97X-V/def2-QZVPPD/CMIRS(PC) Frequency Analysis"
    wB97X_V_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97X-V/def2-QZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97X_V_def2_QZVPPD_CMIRS_PC__Unknown = "wB97X-V/def2-QZVPPD/CMIRS(PC) Unknown"
    wB97M_V_6_31g_d_VACUUM_Single_Point = "wB97M-V/6-31g*/VACUUM Single Point"
    wB97M_V_6_31g_d_VACUUM_Force = "wB97M-V/6-31g*/VACUUM Force"
    wB97M_V_6_31g_d_VACUUM_Geometry_Optimization = "wB97M-V/6-31g*/VACUUM Geometry Optimization"
    wB97M_V_6_31g_d_VACUUM_Frequency_Analysis = "wB97M-V/6-31g*/VACUUM Frequency Analysis"
    wB97M_V_6_31g_d_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/VACUUM Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_VACUUM_Unknown = "wB97M-V/6-31g*/VACUUM Unknown"
    wB97M_V_6_31g_d_PCM_WATER__Single_Point = "wB97M-V/6-31g*/PCM(WATER) Single Point"
    wB97M_V_6_31g_d_PCM_WATER__Force = "wB97M-V/6-31g*/PCM(WATER) Force"
    wB97M_V_6_31g_d_PCM_WATER__Geometry_Optimization = "wB97M-V/6-31g*/PCM(WATER) Geometry Optimization"
    wB97M_V_6_31g_d_PCM_WATER__Frequency_Analysis = "wB97M-V/6-31g*/PCM(WATER) Frequency Analysis"
    wB97M_V_6_31g_d_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_WATER__Unknown = "wB97M-V/6-31g*/PCM(WATER) Unknown"
    wB97M_V_6_31g_d_PCM_THF__Single_Point = "wB97M-V/6-31g*/PCM(THF) Single Point"
    wB97M_V_6_31g_d_PCM_THF__Force = "wB97M-V/6-31g*/PCM(THF) Force"
    wB97M_V_6_31g_d_PCM_THF__Geometry_Optimization = "wB97M-V/6-31g*/PCM(THF) Geometry Optimization"
    wB97M_V_6_31g_d_PCM_THF__Frequency_Analysis = "wB97M-V/6-31g*/PCM(THF) Frequency Analysis"
    wB97M_V_6_31g_d_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(THF) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_THF__Unknown = "wB97M-V/6-31g*/PCM(THF) Unknown"
    wB97M_V_6_31g_d_PCM_G2__Single_Point = "wB97M-V/6-31g*/PCM(G2) Single Point"
    wB97M_V_6_31g_d_PCM_G2__Force = "wB97M-V/6-31g*/PCM(G2) Force"
    wB97M_V_6_31g_d_PCM_G2__Geometry_Optimization = "wB97M-V/6-31g*/PCM(G2) Geometry Optimization"
    wB97M_V_6_31g_d_PCM_G2__Frequency_Analysis = "wB97M-V/6-31g*/PCM(G2) Frequency Analysis"
    wB97M_V_6_31g_d_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(G2) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_G2__Unknown = "wB97M-V/6-31g*/PCM(G2) Unknown"
    wB97M_V_6_31g_d_PCM_DME__Single_Point = "wB97M-V/6-31g*/PCM(DME) Single Point"
    wB97M_V_6_31g_d_PCM_DME__Force = "wB97M-V/6-31g*/PCM(DME) Force"
    wB97M_V_6_31g_d_PCM_DME__Geometry_Optimization = "wB97M-V/6-31g*/PCM(DME) Geometry Optimization"
    wB97M_V_6_31g_d_PCM_DME__Frequency_Analysis = "wB97M-V/6-31g*/PCM(DME) Frequency Analysis"
    wB97M_V_6_31g_d_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(DME) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_DME__Unknown = "wB97M-V/6-31g*/PCM(DME) Unknown"
    wB97M_V_6_31g_d_PCM_EC__Single_Point = "wB97M-V/6-31g*/PCM(EC) Single Point"
    wB97M_V_6_31g_d_PCM_EC__Force = "wB97M-V/6-31g*/PCM(EC) Force"
    wB97M_V_6_31g_d_PCM_EC__Geometry_Optimization = "wB97M-V/6-31g*/PCM(EC) Geometry Optimization"
    wB97M_V_6_31g_d_PCM_EC__Frequency_Analysis = "wB97M-V/6-31g*/PCM(EC) Frequency Analysis"
    wB97M_V_6_31g_d_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(EC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_EC__Unknown = "wB97M-V/6-31g*/PCM(EC) Unknown"
    wB97M_V_6_31g_d_PCM_EC_EMC__Single_Point = "wB97M-V/6-31g*/PCM(EC/EMC) Single Point"
    wB97M_V_6_31g_d_PCM_EC_EMC__Force = "wB97M-V/6-31g*/PCM(EC/EMC) Force"
    wB97M_V_6_31g_d_PCM_EC_EMC__Geometry_Optimization = "wB97M-V/6-31g*/PCM(EC/EMC) Geometry Optimization"
    wB97M_V_6_31g_d_PCM_EC_EMC__Frequency_Analysis = "wB97M-V/6-31g*/PCM(EC/EMC) Frequency Analysis"
    wB97M_V_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_EC_EMC__Unknown = "wB97M-V/6-31g*/PCM(EC/EMC) Unknown"
    wB97M_V_6_31g_d_PCM_PC__Single_Point = "wB97M-V/6-31g*/PCM(PC) Single Point"
    wB97M_V_6_31g_d_PCM_PC__Force = "wB97M-V/6-31g*/PCM(PC) Force"
    wB97M_V_6_31g_d_PCM_PC__Geometry_Optimization = "wB97M-V/6-31g*/PCM(PC) Geometry Optimization"
    wB97M_V_6_31g_d_PCM_PC__Frequency_Analysis = "wB97M-V/6-31g*/PCM(PC) Frequency Analysis"
    wB97M_V_6_31g_d_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(PC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_PCM_PC__Unknown = "wB97M-V/6-31g*/PCM(PC) Unknown"
    wB97M_V_6_31g_d_ISOSVP_WATER__Single_Point = "wB97M-V/6-31g*/ISOSVP(WATER) Single Point"
    wB97M_V_6_31g_d_ISOSVP_WATER__Force = "wB97M-V/6-31g*/ISOSVP(WATER) Force"
    wB97M_V_6_31g_d_ISOSVP_WATER__Geometry_Optimization = "wB97M-V/6-31g*/ISOSVP(WATER) Geometry Optimization"
    wB97M_V_6_31g_d_ISOSVP_WATER__Frequency_Analysis = "wB97M-V/6-31g*/ISOSVP(WATER) Frequency Analysis"
    wB97M_V_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_WATER__Unknown = "wB97M-V/6-31g*/ISOSVP(WATER) Unknown"
    wB97M_V_6_31g_d_ISOSVP_THF__Single_Point = "wB97M-V/6-31g*/ISOSVP(THF) Single Point"
    wB97M_V_6_31g_d_ISOSVP_THF__Force = "wB97M-V/6-31g*/ISOSVP(THF) Force"
    wB97M_V_6_31g_d_ISOSVP_THF__Geometry_Optimization = "wB97M-V/6-31g*/ISOSVP(THF) Geometry Optimization"
    wB97M_V_6_31g_d_ISOSVP_THF__Frequency_Analysis = "wB97M-V/6-31g*/ISOSVP(THF) Frequency Analysis"
    wB97M_V_6_31g_d_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_THF__Unknown = "wB97M-V/6-31g*/ISOSVP(THF) Unknown"
    wB97M_V_6_31g_d_ISOSVP_G2__Single_Point = "wB97M-V/6-31g*/ISOSVP(G2) Single Point"
    wB97M_V_6_31g_d_ISOSVP_G2__Force = "wB97M-V/6-31g*/ISOSVP(G2) Force"
    wB97M_V_6_31g_d_ISOSVP_G2__Geometry_Optimization = "wB97M-V/6-31g*/ISOSVP(G2) Geometry Optimization"
    wB97M_V_6_31g_d_ISOSVP_G2__Frequency_Analysis = "wB97M-V/6-31g*/ISOSVP(G2) Frequency Analysis"
    wB97M_V_6_31g_d_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_G2__Unknown = "wB97M-V/6-31g*/ISOSVP(G2) Unknown"
    wB97M_V_6_31g_d_ISOSVP_DME__Single_Point = "wB97M-V/6-31g*/ISOSVP(DME) Single Point"
    wB97M_V_6_31g_d_ISOSVP_DME__Force = "wB97M-V/6-31g*/ISOSVP(DME) Force"
    wB97M_V_6_31g_d_ISOSVP_DME__Geometry_Optimization = "wB97M-V/6-31g*/ISOSVP(DME) Geometry Optimization"
    wB97M_V_6_31g_d_ISOSVP_DME__Frequency_Analysis = "wB97M-V/6-31g*/ISOSVP(DME) Frequency Analysis"
    wB97M_V_6_31g_d_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_DME__Unknown = "wB97M-V/6-31g*/ISOSVP(DME) Unknown"
    wB97M_V_6_31g_d_ISOSVP_EC__Single_Point = "wB97M-V/6-31g*/ISOSVP(EC) Single Point"
    wB97M_V_6_31g_d_ISOSVP_EC__Force = "wB97M-V/6-31g*/ISOSVP(EC) Force"
    wB97M_V_6_31g_d_ISOSVP_EC__Geometry_Optimization = "wB97M-V/6-31g*/ISOSVP(EC) Geometry Optimization"
    wB97M_V_6_31g_d_ISOSVP_EC__Frequency_Analysis = "wB97M-V/6-31g*/ISOSVP(EC) Frequency Analysis"
    wB97M_V_6_31g_d_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_EC__Unknown = "wB97M-V/6-31g*/ISOSVP(EC) Unknown"
    wB97M_V_6_31g_d_ISOSVP_EC_EMC__Single_Point = "wB97M-V/6-31g*/ISOSVP(EC/EMC) Single Point"
    wB97M_V_6_31g_d_ISOSVP_EC_EMC__Force = "wB97M-V/6-31g*/ISOSVP(EC/EMC) Force"
    wB97M_V_6_31g_d_ISOSVP_EC_EMC__Geometry_Optimization = "wB97M-V/6-31g*/ISOSVP(EC/EMC) Geometry Optimization"
    wB97M_V_6_31g_d_ISOSVP_EC_EMC__Frequency_Analysis = "wB97M-V/6-31g*/ISOSVP(EC/EMC) Frequency Analysis"
    wB97M_V_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_EC_EMC__Unknown = "wB97M-V/6-31g*/ISOSVP(EC/EMC) Unknown"
    wB97M_V_6_31g_d_ISOSVP_PC__Single_Point = "wB97M-V/6-31g*/ISOSVP(PC) Single Point"
    wB97M_V_6_31g_d_ISOSVP_PC__Force = "wB97M-V/6-31g*/ISOSVP(PC) Force"
    wB97M_V_6_31g_d_ISOSVP_PC__Geometry_Optimization = "wB97M-V/6-31g*/ISOSVP(PC) Geometry Optimization"
    wB97M_V_6_31g_d_ISOSVP_PC__Frequency_Analysis = "wB97M-V/6-31g*/ISOSVP(PC) Frequency Analysis"
    wB97M_V_6_31g_d_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_ISOSVP_PC__Unknown = "wB97M-V/6-31g*/ISOSVP(PC) Unknown"
    wB97M_V_6_31g_d_SMD_WATER__Single_Point = "wB97M-V/6-31g*/SMD(WATER) Single Point"
    wB97M_V_6_31g_d_SMD_WATER__Force = "wB97M-V/6-31g*/SMD(WATER) Force"
    wB97M_V_6_31g_d_SMD_WATER__Geometry_Optimization = "wB97M-V/6-31g*/SMD(WATER) Geometry Optimization"
    wB97M_V_6_31g_d_SMD_WATER__Frequency_Analysis = "wB97M-V/6-31g*/SMD(WATER) Frequency Analysis"
    wB97M_V_6_31g_d_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_WATER__Unknown = "wB97M-V/6-31g*/SMD(WATER) Unknown"
    wB97M_V_6_31g_d_SMD_THF__Single_Point = "wB97M-V/6-31g*/SMD(THF) Single Point"
    wB97M_V_6_31g_d_SMD_THF__Force = "wB97M-V/6-31g*/SMD(THF) Force"
    wB97M_V_6_31g_d_SMD_THF__Geometry_Optimization = "wB97M-V/6-31g*/SMD(THF) Geometry Optimization"
    wB97M_V_6_31g_d_SMD_THF__Frequency_Analysis = "wB97M-V/6-31g*/SMD(THF) Frequency Analysis"
    wB97M_V_6_31g_d_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(THF) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_THF__Unknown = "wB97M-V/6-31g*/SMD(THF) Unknown"
    wB97M_V_6_31g_d_SMD_G2__Single_Point = "wB97M-V/6-31g*/SMD(G2) Single Point"
    wB97M_V_6_31g_d_SMD_G2__Force = "wB97M-V/6-31g*/SMD(G2) Force"
    wB97M_V_6_31g_d_SMD_G2__Geometry_Optimization = "wB97M-V/6-31g*/SMD(G2) Geometry Optimization"
    wB97M_V_6_31g_d_SMD_G2__Frequency_Analysis = "wB97M-V/6-31g*/SMD(G2) Frequency Analysis"
    wB97M_V_6_31g_d_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(G2) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_G2__Unknown = "wB97M-V/6-31g*/SMD(G2) Unknown"
    wB97M_V_6_31g_d_SMD_DME__Single_Point = "wB97M-V/6-31g*/SMD(DME) Single Point"
    wB97M_V_6_31g_d_SMD_DME__Force = "wB97M-V/6-31g*/SMD(DME) Force"
    wB97M_V_6_31g_d_SMD_DME__Geometry_Optimization = "wB97M-V/6-31g*/SMD(DME) Geometry Optimization"
    wB97M_V_6_31g_d_SMD_DME__Frequency_Analysis = "wB97M-V/6-31g*/SMD(DME) Frequency Analysis"
    wB97M_V_6_31g_d_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(DME) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_DME__Unknown = "wB97M-V/6-31g*/SMD(DME) Unknown"
    wB97M_V_6_31g_d_SMD_EC__Single_Point = "wB97M-V/6-31g*/SMD(EC) Single Point"
    wB97M_V_6_31g_d_SMD_EC__Force = "wB97M-V/6-31g*/SMD(EC) Force"
    wB97M_V_6_31g_d_SMD_EC__Geometry_Optimization = "wB97M-V/6-31g*/SMD(EC) Geometry Optimization"
    wB97M_V_6_31g_d_SMD_EC__Frequency_Analysis = "wB97M-V/6-31g*/SMD(EC) Frequency Analysis"
    wB97M_V_6_31g_d_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(EC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_EC__Unknown = "wB97M-V/6-31g*/SMD(EC) Unknown"
    wB97M_V_6_31g_d_SMD_EC_EMC__Single_Point = "wB97M-V/6-31g*/SMD(EC/EMC) Single Point"
    wB97M_V_6_31g_d_SMD_EC_EMC__Force = "wB97M-V/6-31g*/SMD(EC/EMC) Force"
    wB97M_V_6_31g_d_SMD_EC_EMC__Geometry_Optimization = "wB97M-V/6-31g*/SMD(EC/EMC) Geometry Optimization"
    wB97M_V_6_31g_d_SMD_EC_EMC__Frequency_Analysis = "wB97M-V/6-31g*/SMD(EC/EMC) Frequency Analysis"
    wB97M_V_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_EC_EMC__Unknown = "wB97M-V/6-31g*/SMD(EC/EMC) Unknown"
    wB97M_V_6_31g_d_SMD_PC__Single_Point = "wB97M-V/6-31g*/SMD(PC) Single Point"
    wB97M_V_6_31g_d_SMD_PC__Force = "wB97M-V/6-31g*/SMD(PC) Force"
    wB97M_V_6_31g_d_SMD_PC__Geometry_Optimization = "wB97M-V/6-31g*/SMD(PC) Geometry Optimization"
    wB97M_V_6_31g_d_SMD_PC__Frequency_Analysis = "wB97M-V/6-31g*/SMD(PC) Frequency Analysis"
    wB97M_V_6_31g_d_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(PC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_SMD_PC__Unknown = "wB97M-V/6-31g*/SMD(PC) Unknown"
    wB97M_V_6_31g_d_CMIRS_WATER__Single_Point = "wB97M-V/6-31g*/CMIRS(WATER) Single Point"
    wB97M_V_6_31g_d_CMIRS_WATER__Force = "wB97M-V/6-31g*/CMIRS(WATER) Force"
    wB97M_V_6_31g_d_CMIRS_WATER__Geometry_Optimization = "wB97M-V/6-31g*/CMIRS(WATER) Geometry Optimization"
    wB97M_V_6_31g_d_CMIRS_WATER__Frequency_Analysis = "wB97M-V/6-31g*/CMIRS(WATER) Frequency Analysis"
    wB97M_V_6_31g_d_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_WATER__Unknown = "wB97M-V/6-31g*/CMIRS(WATER) Unknown"
    wB97M_V_6_31g_d_CMIRS_THF__Single_Point = "wB97M-V/6-31g*/CMIRS(THF) Single Point"
    wB97M_V_6_31g_d_CMIRS_THF__Force = "wB97M-V/6-31g*/CMIRS(THF) Force"
    wB97M_V_6_31g_d_CMIRS_THF__Geometry_Optimization = "wB97M-V/6-31g*/CMIRS(THF) Geometry Optimization"
    wB97M_V_6_31g_d_CMIRS_THF__Frequency_Analysis = "wB97M-V/6-31g*/CMIRS(THF) Frequency Analysis"
    wB97M_V_6_31g_d_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_THF__Unknown = "wB97M-V/6-31g*/CMIRS(THF) Unknown"
    wB97M_V_6_31g_d_CMIRS_G2__Single_Point = "wB97M-V/6-31g*/CMIRS(G2) Single Point"
    wB97M_V_6_31g_d_CMIRS_G2__Force = "wB97M-V/6-31g*/CMIRS(G2) Force"
    wB97M_V_6_31g_d_CMIRS_G2__Geometry_Optimization = "wB97M-V/6-31g*/CMIRS(G2) Geometry Optimization"
    wB97M_V_6_31g_d_CMIRS_G2__Frequency_Analysis = "wB97M-V/6-31g*/CMIRS(G2) Frequency Analysis"
    wB97M_V_6_31g_d_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_G2__Unknown = "wB97M-V/6-31g*/CMIRS(G2) Unknown"
    wB97M_V_6_31g_d_CMIRS_DME__Single_Point = "wB97M-V/6-31g*/CMIRS(DME) Single Point"
    wB97M_V_6_31g_d_CMIRS_DME__Force = "wB97M-V/6-31g*/CMIRS(DME) Force"
    wB97M_V_6_31g_d_CMIRS_DME__Geometry_Optimization = "wB97M-V/6-31g*/CMIRS(DME) Geometry Optimization"
    wB97M_V_6_31g_d_CMIRS_DME__Frequency_Analysis = "wB97M-V/6-31g*/CMIRS(DME) Frequency Analysis"
    wB97M_V_6_31g_d_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_DME__Unknown = "wB97M-V/6-31g*/CMIRS(DME) Unknown"
    wB97M_V_6_31g_d_CMIRS_EC__Single_Point = "wB97M-V/6-31g*/CMIRS(EC) Single Point"
    wB97M_V_6_31g_d_CMIRS_EC__Force = "wB97M-V/6-31g*/CMIRS(EC) Force"
    wB97M_V_6_31g_d_CMIRS_EC__Geometry_Optimization = "wB97M-V/6-31g*/CMIRS(EC) Geometry Optimization"
    wB97M_V_6_31g_d_CMIRS_EC__Frequency_Analysis = "wB97M-V/6-31g*/CMIRS(EC) Frequency Analysis"
    wB97M_V_6_31g_d_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_EC__Unknown = "wB97M-V/6-31g*/CMIRS(EC) Unknown"
    wB97M_V_6_31g_d_CMIRS_EC_EMC__Single_Point = "wB97M-V/6-31g*/CMIRS(EC/EMC) Single Point"
    wB97M_V_6_31g_d_CMIRS_EC_EMC__Force = "wB97M-V/6-31g*/CMIRS(EC/EMC) Force"
    wB97M_V_6_31g_d_CMIRS_EC_EMC__Geometry_Optimization = "wB97M-V/6-31g*/CMIRS(EC/EMC) Geometry Optimization"
    wB97M_V_6_31g_d_CMIRS_EC_EMC__Frequency_Analysis = "wB97M-V/6-31g*/CMIRS(EC/EMC) Frequency Analysis"
    wB97M_V_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_EC_EMC__Unknown = "wB97M-V/6-31g*/CMIRS(EC/EMC) Unknown"
    wB97M_V_6_31g_d_CMIRS_PC__Single_Point = "wB97M-V/6-31g*/CMIRS(PC) Single Point"
    wB97M_V_6_31g_d_CMIRS_PC__Force = "wB97M-V/6-31g*/CMIRS(PC) Force"
    wB97M_V_6_31g_d_CMIRS_PC__Geometry_Optimization = "wB97M-V/6-31g*/CMIRS(PC) Geometry Optimization"
    wB97M_V_6_31g_d_CMIRS_PC__Frequency_Analysis = "wB97M-V/6-31g*/CMIRS(PC) Frequency Analysis"
    wB97M_V_6_31g_d_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/6-31g*/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_6_31g_d_CMIRS_PC__Unknown = "wB97M-V/6-31g*/CMIRS(PC) Unknown"
    wB97M_V_def2_SVPD_VACUUM_Single_Point = "wB97M-V/def2-SVPD/VACUUM Single Point"
    wB97M_V_def2_SVPD_VACUUM_Force = "wB97M-V/def2-SVPD/VACUUM Force"
    wB97M_V_def2_SVPD_VACUUM_Geometry_Optimization = "wB97M-V/def2-SVPD/VACUUM Geometry Optimization"
    wB97M_V_def2_SVPD_VACUUM_Frequency_Analysis = "wB97M-V/def2-SVPD/VACUUM Frequency Analysis"
    wB97M_V_def2_SVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/VACUUM Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_VACUUM_Unknown = "wB97M-V/def2-SVPD/VACUUM Unknown"
    wB97M_V_def2_SVPD_PCM_WATER__Single_Point = "wB97M-V/def2-SVPD/PCM(WATER) Single Point"
    wB97M_V_def2_SVPD_PCM_WATER__Force = "wB97M-V/def2-SVPD/PCM(WATER) Force"
    wB97M_V_def2_SVPD_PCM_WATER__Geometry_Optimization = "wB97M-V/def2-SVPD/PCM(WATER) Geometry Optimization"
    wB97M_V_def2_SVPD_PCM_WATER__Frequency_Analysis = "wB97M-V/def2-SVPD/PCM(WATER) Frequency Analysis"
    wB97M_V_def2_SVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_WATER__Unknown = "wB97M-V/def2-SVPD/PCM(WATER) Unknown"
    wB97M_V_def2_SVPD_PCM_THF__Single_Point = "wB97M-V/def2-SVPD/PCM(THF) Single Point"
    wB97M_V_def2_SVPD_PCM_THF__Force = "wB97M-V/def2-SVPD/PCM(THF) Force"
    wB97M_V_def2_SVPD_PCM_THF__Geometry_Optimization = "wB97M-V/def2-SVPD/PCM(THF) Geometry Optimization"
    wB97M_V_def2_SVPD_PCM_THF__Frequency_Analysis = "wB97M-V/def2-SVPD/PCM(THF) Frequency Analysis"
    wB97M_V_def2_SVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_THF__Unknown = "wB97M-V/def2-SVPD/PCM(THF) Unknown"
    wB97M_V_def2_SVPD_PCM_G2__Single_Point = "wB97M-V/def2-SVPD/PCM(G2) Single Point"
    wB97M_V_def2_SVPD_PCM_G2__Force = "wB97M-V/def2-SVPD/PCM(G2) Force"
    wB97M_V_def2_SVPD_PCM_G2__Geometry_Optimization = "wB97M-V/def2-SVPD/PCM(G2) Geometry Optimization"
    wB97M_V_def2_SVPD_PCM_G2__Frequency_Analysis = "wB97M-V/def2-SVPD/PCM(G2) Frequency Analysis"
    wB97M_V_def2_SVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_G2__Unknown = "wB97M-V/def2-SVPD/PCM(G2) Unknown"
    wB97M_V_def2_SVPD_PCM_DME__Single_Point = "wB97M-V/def2-SVPD/PCM(DME) Single Point"
    wB97M_V_def2_SVPD_PCM_DME__Force = "wB97M-V/def2-SVPD/PCM(DME) Force"
    wB97M_V_def2_SVPD_PCM_DME__Geometry_Optimization = "wB97M-V/def2-SVPD/PCM(DME) Geometry Optimization"
    wB97M_V_def2_SVPD_PCM_DME__Frequency_Analysis = "wB97M-V/def2-SVPD/PCM(DME) Frequency Analysis"
    wB97M_V_def2_SVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_DME__Unknown = "wB97M-V/def2-SVPD/PCM(DME) Unknown"
    wB97M_V_def2_SVPD_PCM_EC__Single_Point = "wB97M-V/def2-SVPD/PCM(EC) Single Point"
    wB97M_V_def2_SVPD_PCM_EC__Force = "wB97M-V/def2-SVPD/PCM(EC) Force"
    wB97M_V_def2_SVPD_PCM_EC__Geometry_Optimization = "wB97M-V/def2-SVPD/PCM(EC) Geometry Optimization"
    wB97M_V_def2_SVPD_PCM_EC__Frequency_Analysis = "wB97M-V/def2-SVPD/PCM(EC) Frequency Analysis"
    wB97M_V_def2_SVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_EC__Unknown = "wB97M-V/def2-SVPD/PCM(EC) Unknown"
    wB97M_V_def2_SVPD_PCM_EC_EMC__Single_Point = "wB97M-V/def2-SVPD/PCM(EC/EMC) Single Point"
    wB97M_V_def2_SVPD_PCM_EC_EMC__Force = "wB97M-V/def2-SVPD/PCM(EC/EMC) Force"
    wB97M_V_def2_SVPD_PCM_EC_EMC__Geometry_Optimization = "wB97M-V/def2-SVPD/PCM(EC/EMC) Geometry Optimization"
    wB97M_V_def2_SVPD_PCM_EC_EMC__Frequency_Analysis = "wB97M-V/def2-SVPD/PCM(EC/EMC) Frequency Analysis"
    wB97M_V_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_EC_EMC__Unknown = "wB97M-V/def2-SVPD/PCM(EC/EMC) Unknown"
    wB97M_V_def2_SVPD_PCM_PC__Single_Point = "wB97M-V/def2-SVPD/PCM(PC) Single Point"
    wB97M_V_def2_SVPD_PCM_PC__Force = "wB97M-V/def2-SVPD/PCM(PC) Force"
    wB97M_V_def2_SVPD_PCM_PC__Geometry_Optimization = "wB97M-V/def2-SVPD/PCM(PC) Geometry Optimization"
    wB97M_V_def2_SVPD_PCM_PC__Frequency_Analysis = "wB97M-V/def2-SVPD/PCM(PC) Frequency Analysis"
    wB97M_V_def2_SVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_PCM_PC__Unknown = "wB97M-V/def2-SVPD/PCM(PC) Unknown"
    wB97M_V_def2_SVPD_ISOSVP_WATER__Single_Point = "wB97M-V/def2-SVPD/ISOSVP(WATER) Single Point"
    wB97M_V_def2_SVPD_ISOSVP_WATER__Force = "wB97M-V/def2-SVPD/ISOSVP(WATER) Force"
    wB97M_V_def2_SVPD_ISOSVP_WATER__Geometry_Optimization = "wB97M-V/def2-SVPD/ISOSVP(WATER) Geometry Optimization"
    wB97M_V_def2_SVPD_ISOSVP_WATER__Frequency_Analysis = "wB97M-V/def2-SVPD/ISOSVP(WATER) Frequency Analysis"
    wB97M_V_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_WATER__Unknown = "wB97M-V/def2-SVPD/ISOSVP(WATER) Unknown"
    wB97M_V_def2_SVPD_ISOSVP_THF__Single_Point = "wB97M-V/def2-SVPD/ISOSVP(THF) Single Point"
    wB97M_V_def2_SVPD_ISOSVP_THF__Force = "wB97M-V/def2-SVPD/ISOSVP(THF) Force"
    wB97M_V_def2_SVPD_ISOSVP_THF__Geometry_Optimization = "wB97M-V/def2-SVPD/ISOSVP(THF) Geometry Optimization"
    wB97M_V_def2_SVPD_ISOSVP_THF__Frequency_Analysis = "wB97M-V/def2-SVPD/ISOSVP(THF) Frequency Analysis"
    wB97M_V_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_THF__Unknown = "wB97M-V/def2-SVPD/ISOSVP(THF) Unknown"
    wB97M_V_def2_SVPD_ISOSVP_G2__Single_Point = "wB97M-V/def2-SVPD/ISOSVP(G2) Single Point"
    wB97M_V_def2_SVPD_ISOSVP_G2__Force = "wB97M-V/def2-SVPD/ISOSVP(G2) Force"
    wB97M_V_def2_SVPD_ISOSVP_G2__Geometry_Optimization = "wB97M-V/def2-SVPD/ISOSVP(G2) Geometry Optimization"
    wB97M_V_def2_SVPD_ISOSVP_G2__Frequency_Analysis = "wB97M-V/def2-SVPD/ISOSVP(G2) Frequency Analysis"
    wB97M_V_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_G2__Unknown = "wB97M-V/def2-SVPD/ISOSVP(G2) Unknown"
    wB97M_V_def2_SVPD_ISOSVP_DME__Single_Point = "wB97M-V/def2-SVPD/ISOSVP(DME) Single Point"
    wB97M_V_def2_SVPD_ISOSVP_DME__Force = "wB97M-V/def2-SVPD/ISOSVP(DME) Force"
    wB97M_V_def2_SVPD_ISOSVP_DME__Geometry_Optimization = "wB97M-V/def2-SVPD/ISOSVP(DME) Geometry Optimization"
    wB97M_V_def2_SVPD_ISOSVP_DME__Frequency_Analysis = "wB97M-V/def2-SVPD/ISOSVP(DME) Frequency Analysis"
    wB97M_V_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_DME__Unknown = "wB97M-V/def2-SVPD/ISOSVP(DME) Unknown"
    wB97M_V_def2_SVPD_ISOSVP_EC__Single_Point = "wB97M-V/def2-SVPD/ISOSVP(EC) Single Point"
    wB97M_V_def2_SVPD_ISOSVP_EC__Force = "wB97M-V/def2-SVPD/ISOSVP(EC) Force"
    wB97M_V_def2_SVPD_ISOSVP_EC__Geometry_Optimization = "wB97M-V/def2-SVPD/ISOSVP(EC) Geometry Optimization"
    wB97M_V_def2_SVPD_ISOSVP_EC__Frequency_Analysis = "wB97M-V/def2-SVPD/ISOSVP(EC) Frequency Analysis"
    wB97M_V_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_EC__Unknown = "wB97M-V/def2-SVPD/ISOSVP(EC) Unknown"
    wB97M_V_def2_SVPD_ISOSVP_EC_EMC__Single_Point = "wB97M-V/def2-SVPD/ISOSVP(EC/EMC) Single Point"
    wB97M_V_def2_SVPD_ISOSVP_EC_EMC__Force = "wB97M-V/def2-SVPD/ISOSVP(EC/EMC) Force"
    wB97M_V_def2_SVPD_ISOSVP_EC_EMC__Geometry_Optimization = "wB97M-V/def2-SVPD/ISOSVP(EC/EMC) Geometry Optimization"
    wB97M_V_def2_SVPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97M-V/def2-SVPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97M_V_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_EC_EMC__Unknown = "wB97M-V/def2-SVPD/ISOSVP(EC/EMC) Unknown"
    wB97M_V_def2_SVPD_ISOSVP_PC__Single_Point = "wB97M-V/def2-SVPD/ISOSVP(PC) Single Point"
    wB97M_V_def2_SVPD_ISOSVP_PC__Force = "wB97M-V/def2-SVPD/ISOSVP(PC) Force"
    wB97M_V_def2_SVPD_ISOSVP_PC__Geometry_Optimization = "wB97M-V/def2-SVPD/ISOSVP(PC) Geometry Optimization"
    wB97M_V_def2_SVPD_ISOSVP_PC__Frequency_Analysis = "wB97M-V/def2-SVPD/ISOSVP(PC) Frequency Analysis"
    wB97M_V_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_ISOSVP_PC__Unknown = "wB97M-V/def2-SVPD/ISOSVP(PC) Unknown"
    wB97M_V_def2_SVPD_SMD_WATER__Single_Point = "wB97M-V/def2-SVPD/SMD(WATER) Single Point"
    wB97M_V_def2_SVPD_SMD_WATER__Force = "wB97M-V/def2-SVPD/SMD(WATER) Force"
    wB97M_V_def2_SVPD_SMD_WATER__Geometry_Optimization = "wB97M-V/def2-SVPD/SMD(WATER) Geometry Optimization"
    wB97M_V_def2_SVPD_SMD_WATER__Frequency_Analysis = "wB97M-V/def2-SVPD/SMD(WATER) Frequency Analysis"
    wB97M_V_def2_SVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_WATER__Unknown = "wB97M-V/def2-SVPD/SMD(WATER) Unknown"
    wB97M_V_def2_SVPD_SMD_THF__Single_Point = "wB97M-V/def2-SVPD/SMD(THF) Single Point"
    wB97M_V_def2_SVPD_SMD_THF__Force = "wB97M-V/def2-SVPD/SMD(THF) Force"
    wB97M_V_def2_SVPD_SMD_THF__Geometry_Optimization = "wB97M-V/def2-SVPD/SMD(THF) Geometry Optimization"
    wB97M_V_def2_SVPD_SMD_THF__Frequency_Analysis = "wB97M-V/def2-SVPD/SMD(THF) Frequency Analysis"
    wB97M_V_def2_SVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_THF__Unknown = "wB97M-V/def2-SVPD/SMD(THF) Unknown"
    wB97M_V_def2_SVPD_SMD_G2__Single_Point = "wB97M-V/def2-SVPD/SMD(G2) Single Point"
    wB97M_V_def2_SVPD_SMD_G2__Force = "wB97M-V/def2-SVPD/SMD(G2) Force"
    wB97M_V_def2_SVPD_SMD_G2__Geometry_Optimization = "wB97M-V/def2-SVPD/SMD(G2) Geometry Optimization"
    wB97M_V_def2_SVPD_SMD_G2__Frequency_Analysis = "wB97M-V/def2-SVPD/SMD(G2) Frequency Analysis"
    wB97M_V_def2_SVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_G2__Unknown = "wB97M-V/def2-SVPD/SMD(G2) Unknown"
    wB97M_V_def2_SVPD_SMD_DME__Single_Point = "wB97M-V/def2-SVPD/SMD(DME) Single Point"
    wB97M_V_def2_SVPD_SMD_DME__Force = "wB97M-V/def2-SVPD/SMD(DME) Force"
    wB97M_V_def2_SVPD_SMD_DME__Geometry_Optimization = "wB97M-V/def2-SVPD/SMD(DME) Geometry Optimization"
    wB97M_V_def2_SVPD_SMD_DME__Frequency_Analysis = "wB97M-V/def2-SVPD/SMD(DME) Frequency Analysis"
    wB97M_V_def2_SVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_DME__Unknown = "wB97M-V/def2-SVPD/SMD(DME) Unknown"
    wB97M_V_def2_SVPD_SMD_EC__Single_Point = "wB97M-V/def2-SVPD/SMD(EC) Single Point"
    wB97M_V_def2_SVPD_SMD_EC__Force = "wB97M-V/def2-SVPD/SMD(EC) Force"
    wB97M_V_def2_SVPD_SMD_EC__Geometry_Optimization = "wB97M-V/def2-SVPD/SMD(EC) Geometry Optimization"
    wB97M_V_def2_SVPD_SMD_EC__Frequency_Analysis = "wB97M-V/def2-SVPD/SMD(EC) Frequency Analysis"
    wB97M_V_def2_SVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_EC__Unknown = "wB97M-V/def2-SVPD/SMD(EC) Unknown"
    wB97M_V_def2_SVPD_SMD_EC_EMC__Single_Point = "wB97M-V/def2-SVPD/SMD(EC/EMC) Single Point"
    wB97M_V_def2_SVPD_SMD_EC_EMC__Force = "wB97M-V/def2-SVPD/SMD(EC/EMC) Force"
    wB97M_V_def2_SVPD_SMD_EC_EMC__Geometry_Optimization = "wB97M-V/def2-SVPD/SMD(EC/EMC) Geometry Optimization"
    wB97M_V_def2_SVPD_SMD_EC_EMC__Frequency_Analysis = "wB97M-V/def2-SVPD/SMD(EC/EMC) Frequency Analysis"
    wB97M_V_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_EC_EMC__Unknown = "wB97M-V/def2-SVPD/SMD(EC/EMC) Unknown"
    wB97M_V_def2_SVPD_SMD_PC__Single_Point = "wB97M-V/def2-SVPD/SMD(PC) Single Point"
    wB97M_V_def2_SVPD_SMD_PC__Force = "wB97M-V/def2-SVPD/SMD(PC) Force"
    wB97M_V_def2_SVPD_SMD_PC__Geometry_Optimization = "wB97M-V/def2-SVPD/SMD(PC) Geometry Optimization"
    wB97M_V_def2_SVPD_SMD_PC__Frequency_Analysis = "wB97M-V/def2-SVPD/SMD(PC) Frequency Analysis"
    wB97M_V_def2_SVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_SMD_PC__Unknown = "wB97M-V/def2-SVPD/SMD(PC) Unknown"
    wB97M_V_def2_SVPD_CMIRS_WATER__Single_Point = "wB97M-V/def2-SVPD/CMIRS(WATER) Single Point"
    wB97M_V_def2_SVPD_CMIRS_WATER__Force = "wB97M-V/def2-SVPD/CMIRS(WATER) Force"
    wB97M_V_def2_SVPD_CMIRS_WATER__Geometry_Optimization = "wB97M-V/def2-SVPD/CMIRS(WATER) Geometry Optimization"
    wB97M_V_def2_SVPD_CMIRS_WATER__Frequency_Analysis = "wB97M-V/def2-SVPD/CMIRS(WATER) Frequency Analysis"
    wB97M_V_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_WATER__Unknown = "wB97M-V/def2-SVPD/CMIRS(WATER) Unknown"
    wB97M_V_def2_SVPD_CMIRS_THF__Single_Point = "wB97M-V/def2-SVPD/CMIRS(THF) Single Point"
    wB97M_V_def2_SVPD_CMIRS_THF__Force = "wB97M-V/def2-SVPD/CMIRS(THF) Force"
    wB97M_V_def2_SVPD_CMIRS_THF__Geometry_Optimization = "wB97M-V/def2-SVPD/CMIRS(THF) Geometry Optimization"
    wB97M_V_def2_SVPD_CMIRS_THF__Frequency_Analysis = "wB97M-V/def2-SVPD/CMIRS(THF) Frequency Analysis"
    wB97M_V_def2_SVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_THF__Unknown = "wB97M-V/def2-SVPD/CMIRS(THF) Unknown"
    wB97M_V_def2_SVPD_CMIRS_G2__Single_Point = "wB97M-V/def2-SVPD/CMIRS(G2) Single Point"
    wB97M_V_def2_SVPD_CMIRS_G2__Force = "wB97M-V/def2-SVPD/CMIRS(G2) Force"
    wB97M_V_def2_SVPD_CMIRS_G2__Geometry_Optimization = "wB97M-V/def2-SVPD/CMIRS(G2) Geometry Optimization"
    wB97M_V_def2_SVPD_CMIRS_G2__Frequency_Analysis = "wB97M-V/def2-SVPD/CMIRS(G2) Frequency Analysis"
    wB97M_V_def2_SVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_G2__Unknown = "wB97M-V/def2-SVPD/CMIRS(G2) Unknown"
    wB97M_V_def2_SVPD_CMIRS_DME__Single_Point = "wB97M-V/def2-SVPD/CMIRS(DME) Single Point"
    wB97M_V_def2_SVPD_CMIRS_DME__Force = "wB97M-V/def2-SVPD/CMIRS(DME) Force"
    wB97M_V_def2_SVPD_CMIRS_DME__Geometry_Optimization = "wB97M-V/def2-SVPD/CMIRS(DME) Geometry Optimization"
    wB97M_V_def2_SVPD_CMIRS_DME__Frequency_Analysis = "wB97M-V/def2-SVPD/CMIRS(DME) Frequency Analysis"
    wB97M_V_def2_SVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_DME__Unknown = "wB97M-V/def2-SVPD/CMIRS(DME) Unknown"
    wB97M_V_def2_SVPD_CMIRS_EC__Single_Point = "wB97M-V/def2-SVPD/CMIRS(EC) Single Point"
    wB97M_V_def2_SVPD_CMIRS_EC__Force = "wB97M-V/def2-SVPD/CMIRS(EC) Force"
    wB97M_V_def2_SVPD_CMIRS_EC__Geometry_Optimization = "wB97M-V/def2-SVPD/CMIRS(EC) Geometry Optimization"
    wB97M_V_def2_SVPD_CMIRS_EC__Frequency_Analysis = "wB97M-V/def2-SVPD/CMIRS(EC) Frequency Analysis"
    wB97M_V_def2_SVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_EC__Unknown = "wB97M-V/def2-SVPD/CMIRS(EC) Unknown"
    wB97M_V_def2_SVPD_CMIRS_EC_EMC__Single_Point = "wB97M-V/def2-SVPD/CMIRS(EC/EMC) Single Point"
    wB97M_V_def2_SVPD_CMIRS_EC_EMC__Force = "wB97M-V/def2-SVPD/CMIRS(EC/EMC) Force"
    wB97M_V_def2_SVPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97M-V/def2-SVPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97M_V_def2_SVPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97M-V/def2-SVPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97M_V_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_EC_EMC__Unknown = "wB97M-V/def2-SVPD/CMIRS(EC/EMC) Unknown"
    wB97M_V_def2_SVPD_CMIRS_PC__Single_Point = "wB97M-V/def2-SVPD/CMIRS(PC) Single Point"
    wB97M_V_def2_SVPD_CMIRS_PC__Force = "wB97M-V/def2-SVPD/CMIRS(PC) Force"
    wB97M_V_def2_SVPD_CMIRS_PC__Geometry_Optimization = "wB97M-V/def2-SVPD/CMIRS(PC) Geometry Optimization"
    wB97M_V_def2_SVPD_CMIRS_PC__Frequency_Analysis = "wB97M-V/def2-SVPD/CMIRS(PC) Frequency Analysis"
    wB97M_V_def2_SVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-SVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_SVPD_CMIRS_PC__Unknown = "wB97M-V/def2-SVPD/CMIRS(PC) Unknown"
    wB97M_V_def2_TZVP_VACUUM_Single_Point = "wB97M-V/def2-TZVP/VACUUM Single Point"
    wB97M_V_def2_TZVP_VACUUM_Force = "wB97M-V/def2-TZVP/VACUUM Force"
    wB97M_V_def2_TZVP_VACUUM_Geometry_Optimization = "wB97M-V/def2-TZVP/VACUUM Geometry Optimization"
    wB97M_V_def2_TZVP_VACUUM_Frequency_Analysis = "wB97M-V/def2-TZVP/VACUUM Frequency Analysis"
    wB97M_V_def2_TZVP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/VACUUM Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_VACUUM_Unknown = "wB97M-V/def2-TZVP/VACUUM Unknown"
    wB97M_V_def2_TZVP_PCM_WATER__Single_Point = "wB97M-V/def2-TZVP/PCM(WATER) Single Point"
    wB97M_V_def2_TZVP_PCM_WATER__Force = "wB97M-V/def2-TZVP/PCM(WATER) Force"
    wB97M_V_def2_TZVP_PCM_WATER__Geometry_Optimization = "wB97M-V/def2-TZVP/PCM(WATER) Geometry Optimization"
    wB97M_V_def2_TZVP_PCM_WATER__Frequency_Analysis = "wB97M-V/def2-TZVP/PCM(WATER) Frequency Analysis"
    wB97M_V_def2_TZVP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_WATER__Unknown = "wB97M-V/def2-TZVP/PCM(WATER) Unknown"
    wB97M_V_def2_TZVP_PCM_THF__Single_Point = "wB97M-V/def2-TZVP/PCM(THF) Single Point"
    wB97M_V_def2_TZVP_PCM_THF__Force = "wB97M-V/def2-TZVP/PCM(THF) Force"
    wB97M_V_def2_TZVP_PCM_THF__Geometry_Optimization = "wB97M-V/def2-TZVP/PCM(THF) Geometry Optimization"
    wB97M_V_def2_TZVP_PCM_THF__Frequency_Analysis = "wB97M-V/def2-TZVP/PCM(THF) Frequency Analysis"
    wB97M_V_def2_TZVP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_THF__Unknown = "wB97M-V/def2-TZVP/PCM(THF) Unknown"
    wB97M_V_def2_TZVP_PCM_G2__Single_Point = "wB97M-V/def2-TZVP/PCM(G2) Single Point"
    wB97M_V_def2_TZVP_PCM_G2__Force = "wB97M-V/def2-TZVP/PCM(G2) Force"
    wB97M_V_def2_TZVP_PCM_G2__Geometry_Optimization = "wB97M-V/def2-TZVP/PCM(G2) Geometry Optimization"
    wB97M_V_def2_TZVP_PCM_G2__Frequency_Analysis = "wB97M-V/def2-TZVP/PCM(G2) Frequency Analysis"
    wB97M_V_def2_TZVP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_G2__Unknown = "wB97M-V/def2-TZVP/PCM(G2) Unknown"
    wB97M_V_def2_TZVP_PCM_DME__Single_Point = "wB97M-V/def2-TZVP/PCM(DME) Single Point"
    wB97M_V_def2_TZVP_PCM_DME__Force = "wB97M-V/def2-TZVP/PCM(DME) Force"
    wB97M_V_def2_TZVP_PCM_DME__Geometry_Optimization = "wB97M-V/def2-TZVP/PCM(DME) Geometry Optimization"
    wB97M_V_def2_TZVP_PCM_DME__Frequency_Analysis = "wB97M-V/def2-TZVP/PCM(DME) Frequency Analysis"
    wB97M_V_def2_TZVP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_DME__Unknown = "wB97M-V/def2-TZVP/PCM(DME) Unknown"
    wB97M_V_def2_TZVP_PCM_EC__Single_Point = "wB97M-V/def2-TZVP/PCM(EC) Single Point"
    wB97M_V_def2_TZVP_PCM_EC__Force = "wB97M-V/def2-TZVP/PCM(EC) Force"
    wB97M_V_def2_TZVP_PCM_EC__Geometry_Optimization = "wB97M-V/def2-TZVP/PCM(EC) Geometry Optimization"
    wB97M_V_def2_TZVP_PCM_EC__Frequency_Analysis = "wB97M-V/def2-TZVP/PCM(EC) Frequency Analysis"
    wB97M_V_def2_TZVP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_EC__Unknown = "wB97M-V/def2-TZVP/PCM(EC) Unknown"
    wB97M_V_def2_TZVP_PCM_EC_EMC__Single_Point = "wB97M-V/def2-TZVP/PCM(EC/EMC) Single Point"
    wB97M_V_def2_TZVP_PCM_EC_EMC__Force = "wB97M-V/def2-TZVP/PCM(EC/EMC) Force"
    wB97M_V_def2_TZVP_PCM_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVP/PCM(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVP_PCM_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVP/PCM(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_EC_EMC__Unknown = "wB97M-V/def2-TZVP/PCM(EC/EMC) Unknown"
    wB97M_V_def2_TZVP_PCM_PC__Single_Point = "wB97M-V/def2-TZVP/PCM(PC) Single Point"
    wB97M_V_def2_TZVP_PCM_PC__Force = "wB97M-V/def2-TZVP/PCM(PC) Force"
    wB97M_V_def2_TZVP_PCM_PC__Geometry_Optimization = "wB97M-V/def2-TZVP/PCM(PC) Geometry Optimization"
    wB97M_V_def2_TZVP_PCM_PC__Frequency_Analysis = "wB97M-V/def2-TZVP/PCM(PC) Frequency Analysis"
    wB97M_V_def2_TZVP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_PCM_PC__Unknown = "wB97M-V/def2-TZVP/PCM(PC) Unknown"
    wB97M_V_def2_TZVP_ISOSVP_WATER__Single_Point = "wB97M-V/def2-TZVP/ISOSVP(WATER) Single Point"
    wB97M_V_def2_TZVP_ISOSVP_WATER__Force = "wB97M-V/def2-TZVP/ISOSVP(WATER) Force"
    wB97M_V_def2_TZVP_ISOSVP_WATER__Geometry_Optimization = "wB97M-V/def2-TZVP/ISOSVP(WATER) Geometry Optimization"
    wB97M_V_def2_TZVP_ISOSVP_WATER__Frequency_Analysis = "wB97M-V/def2-TZVP/ISOSVP(WATER) Frequency Analysis"
    wB97M_V_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_WATER__Unknown = "wB97M-V/def2-TZVP/ISOSVP(WATER) Unknown"
    wB97M_V_def2_TZVP_ISOSVP_THF__Single_Point = "wB97M-V/def2-TZVP/ISOSVP(THF) Single Point"
    wB97M_V_def2_TZVP_ISOSVP_THF__Force = "wB97M-V/def2-TZVP/ISOSVP(THF) Force"
    wB97M_V_def2_TZVP_ISOSVP_THF__Geometry_Optimization = "wB97M-V/def2-TZVP/ISOSVP(THF) Geometry Optimization"
    wB97M_V_def2_TZVP_ISOSVP_THF__Frequency_Analysis = "wB97M-V/def2-TZVP/ISOSVP(THF) Frequency Analysis"
    wB97M_V_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_THF__Unknown = "wB97M-V/def2-TZVP/ISOSVP(THF) Unknown"
    wB97M_V_def2_TZVP_ISOSVP_G2__Single_Point = "wB97M-V/def2-TZVP/ISOSVP(G2) Single Point"
    wB97M_V_def2_TZVP_ISOSVP_G2__Force = "wB97M-V/def2-TZVP/ISOSVP(G2) Force"
    wB97M_V_def2_TZVP_ISOSVP_G2__Geometry_Optimization = "wB97M-V/def2-TZVP/ISOSVP(G2) Geometry Optimization"
    wB97M_V_def2_TZVP_ISOSVP_G2__Frequency_Analysis = "wB97M-V/def2-TZVP/ISOSVP(G2) Frequency Analysis"
    wB97M_V_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_G2__Unknown = "wB97M-V/def2-TZVP/ISOSVP(G2) Unknown"
    wB97M_V_def2_TZVP_ISOSVP_DME__Single_Point = "wB97M-V/def2-TZVP/ISOSVP(DME) Single Point"
    wB97M_V_def2_TZVP_ISOSVP_DME__Force = "wB97M-V/def2-TZVP/ISOSVP(DME) Force"
    wB97M_V_def2_TZVP_ISOSVP_DME__Geometry_Optimization = "wB97M-V/def2-TZVP/ISOSVP(DME) Geometry Optimization"
    wB97M_V_def2_TZVP_ISOSVP_DME__Frequency_Analysis = "wB97M-V/def2-TZVP/ISOSVP(DME) Frequency Analysis"
    wB97M_V_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_DME__Unknown = "wB97M-V/def2-TZVP/ISOSVP(DME) Unknown"
    wB97M_V_def2_TZVP_ISOSVP_EC__Single_Point = "wB97M-V/def2-TZVP/ISOSVP(EC) Single Point"
    wB97M_V_def2_TZVP_ISOSVP_EC__Force = "wB97M-V/def2-TZVP/ISOSVP(EC) Force"
    wB97M_V_def2_TZVP_ISOSVP_EC__Geometry_Optimization = "wB97M-V/def2-TZVP/ISOSVP(EC) Geometry Optimization"
    wB97M_V_def2_TZVP_ISOSVP_EC__Frequency_Analysis = "wB97M-V/def2-TZVP/ISOSVP(EC) Frequency Analysis"
    wB97M_V_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_EC__Unknown = "wB97M-V/def2-TZVP/ISOSVP(EC) Unknown"
    wB97M_V_def2_TZVP_ISOSVP_EC_EMC__Single_Point = "wB97M-V/def2-TZVP/ISOSVP(EC/EMC) Single Point"
    wB97M_V_def2_TZVP_ISOSVP_EC_EMC__Force = "wB97M-V/def2-TZVP/ISOSVP(EC/EMC) Force"
    wB97M_V_def2_TZVP_ISOSVP_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVP/ISOSVP(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVP_ISOSVP_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVP/ISOSVP(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_EC_EMC__Unknown = "wB97M-V/def2-TZVP/ISOSVP(EC/EMC) Unknown"
    wB97M_V_def2_TZVP_ISOSVP_PC__Single_Point = "wB97M-V/def2-TZVP/ISOSVP(PC) Single Point"
    wB97M_V_def2_TZVP_ISOSVP_PC__Force = "wB97M-V/def2-TZVP/ISOSVP(PC) Force"
    wB97M_V_def2_TZVP_ISOSVP_PC__Geometry_Optimization = "wB97M-V/def2-TZVP/ISOSVP(PC) Geometry Optimization"
    wB97M_V_def2_TZVP_ISOSVP_PC__Frequency_Analysis = "wB97M-V/def2-TZVP/ISOSVP(PC) Frequency Analysis"
    wB97M_V_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_ISOSVP_PC__Unknown = "wB97M-V/def2-TZVP/ISOSVP(PC) Unknown"
    wB97M_V_def2_TZVP_SMD_WATER__Single_Point = "wB97M-V/def2-TZVP/SMD(WATER) Single Point"
    wB97M_V_def2_TZVP_SMD_WATER__Force = "wB97M-V/def2-TZVP/SMD(WATER) Force"
    wB97M_V_def2_TZVP_SMD_WATER__Geometry_Optimization = "wB97M-V/def2-TZVP/SMD(WATER) Geometry Optimization"
    wB97M_V_def2_TZVP_SMD_WATER__Frequency_Analysis = "wB97M-V/def2-TZVP/SMD(WATER) Frequency Analysis"
    wB97M_V_def2_TZVP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_WATER__Unknown = "wB97M-V/def2-TZVP/SMD(WATER) Unknown"
    wB97M_V_def2_TZVP_SMD_THF__Single_Point = "wB97M-V/def2-TZVP/SMD(THF) Single Point"
    wB97M_V_def2_TZVP_SMD_THF__Force = "wB97M-V/def2-TZVP/SMD(THF) Force"
    wB97M_V_def2_TZVP_SMD_THF__Geometry_Optimization = "wB97M-V/def2-TZVP/SMD(THF) Geometry Optimization"
    wB97M_V_def2_TZVP_SMD_THF__Frequency_Analysis = "wB97M-V/def2-TZVP/SMD(THF) Frequency Analysis"
    wB97M_V_def2_TZVP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_THF__Unknown = "wB97M-V/def2-TZVP/SMD(THF) Unknown"
    wB97M_V_def2_TZVP_SMD_G2__Single_Point = "wB97M-V/def2-TZVP/SMD(G2) Single Point"
    wB97M_V_def2_TZVP_SMD_G2__Force = "wB97M-V/def2-TZVP/SMD(G2) Force"
    wB97M_V_def2_TZVP_SMD_G2__Geometry_Optimization = "wB97M-V/def2-TZVP/SMD(G2) Geometry Optimization"
    wB97M_V_def2_TZVP_SMD_G2__Frequency_Analysis = "wB97M-V/def2-TZVP/SMD(G2) Frequency Analysis"
    wB97M_V_def2_TZVP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_G2__Unknown = "wB97M-V/def2-TZVP/SMD(G2) Unknown"
    wB97M_V_def2_TZVP_SMD_DME__Single_Point = "wB97M-V/def2-TZVP/SMD(DME) Single Point"
    wB97M_V_def2_TZVP_SMD_DME__Force = "wB97M-V/def2-TZVP/SMD(DME) Force"
    wB97M_V_def2_TZVP_SMD_DME__Geometry_Optimization = "wB97M-V/def2-TZVP/SMD(DME) Geometry Optimization"
    wB97M_V_def2_TZVP_SMD_DME__Frequency_Analysis = "wB97M-V/def2-TZVP/SMD(DME) Frequency Analysis"
    wB97M_V_def2_TZVP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_DME__Unknown = "wB97M-V/def2-TZVP/SMD(DME) Unknown"
    wB97M_V_def2_TZVP_SMD_EC__Single_Point = "wB97M-V/def2-TZVP/SMD(EC) Single Point"
    wB97M_V_def2_TZVP_SMD_EC__Force = "wB97M-V/def2-TZVP/SMD(EC) Force"
    wB97M_V_def2_TZVP_SMD_EC__Geometry_Optimization = "wB97M-V/def2-TZVP/SMD(EC) Geometry Optimization"
    wB97M_V_def2_TZVP_SMD_EC__Frequency_Analysis = "wB97M-V/def2-TZVP/SMD(EC) Frequency Analysis"
    wB97M_V_def2_TZVP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_EC__Unknown = "wB97M-V/def2-TZVP/SMD(EC) Unknown"
    wB97M_V_def2_TZVP_SMD_EC_EMC__Single_Point = "wB97M-V/def2-TZVP/SMD(EC/EMC) Single Point"
    wB97M_V_def2_TZVP_SMD_EC_EMC__Force = "wB97M-V/def2-TZVP/SMD(EC/EMC) Force"
    wB97M_V_def2_TZVP_SMD_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVP/SMD(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVP_SMD_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVP/SMD(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_EC_EMC__Unknown = "wB97M-V/def2-TZVP/SMD(EC/EMC) Unknown"
    wB97M_V_def2_TZVP_SMD_PC__Single_Point = "wB97M-V/def2-TZVP/SMD(PC) Single Point"
    wB97M_V_def2_TZVP_SMD_PC__Force = "wB97M-V/def2-TZVP/SMD(PC) Force"
    wB97M_V_def2_TZVP_SMD_PC__Geometry_Optimization = "wB97M-V/def2-TZVP/SMD(PC) Geometry Optimization"
    wB97M_V_def2_TZVP_SMD_PC__Frequency_Analysis = "wB97M-V/def2-TZVP/SMD(PC) Frequency Analysis"
    wB97M_V_def2_TZVP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_SMD_PC__Unknown = "wB97M-V/def2-TZVP/SMD(PC) Unknown"
    wB97M_V_def2_TZVP_CMIRS_WATER__Single_Point = "wB97M-V/def2-TZVP/CMIRS(WATER) Single Point"
    wB97M_V_def2_TZVP_CMIRS_WATER__Force = "wB97M-V/def2-TZVP/CMIRS(WATER) Force"
    wB97M_V_def2_TZVP_CMIRS_WATER__Geometry_Optimization = "wB97M-V/def2-TZVP/CMIRS(WATER) Geometry Optimization"
    wB97M_V_def2_TZVP_CMIRS_WATER__Frequency_Analysis = "wB97M-V/def2-TZVP/CMIRS(WATER) Frequency Analysis"
    wB97M_V_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_WATER__Unknown = "wB97M-V/def2-TZVP/CMIRS(WATER) Unknown"
    wB97M_V_def2_TZVP_CMIRS_THF__Single_Point = "wB97M-V/def2-TZVP/CMIRS(THF) Single Point"
    wB97M_V_def2_TZVP_CMIRS_THF__Force = "wB97M-V/def2-TZVP/CMIRS(THF) Force"
    wB97M_V_def2_TZVP_CMIRS_THF__Geometry_Optimization = "wB97M-V/def2-TZVP/CMIRS(THF) Geometry Optimization"
    wB97M_V_def2_TZVP_CMIRS_THF__Frequency_Analysis = "wB97M-V/def2-TZVP/CMIRS(THF) Frequency Analysis"
    wB97M_V_def2_TZVP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_THF__Unknown = "wB97M-V/def2-TZVP/CMIRS(THF) Unknown"
    wB97M_V_def2_TZVP_CMIRS_G2__Single_Point = "wB97M-V/def2-TZVP/CMIRS(G2) Single Point"
    wB97M_V_def2_TZVP_CMIRS_G2__Force = "wB97M-V/def2-TZVP/CMIRS(G2) Force"
    wB97M_V_def2_TZVP_CMIRS_G2__Geometry_Optimization = "wB97M-V/def2-TZVP/CMIRS(G2) Geometry Optimization"
    wB97M_V_def2_TZVP_CMIRS_G2__Frequency_Analysis = "wB97M-V/def2-TZVP/CMIRS(G2) Frequency Analysis"
    wB97M_V_def2_TZVP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_G2__Unknown = "wB97M-V/def2-TZVP/CMIRS(G2) Unknown"
    wB97M_V_def2_TZVP_CMIRS_DME__Single_Point = "wB97M-V/def2-TZVP/CMIRS(DME) Single Point"
    wB97M_V_def2_TZVP_CMIRS_DME__Force = "wB97M-V/def2-TZVP/CMIRS(DME) Force"
    wB97M_V_def2_TZVP_CMIRS_DME__Geometry_Optimization = "wB97M-V/def2-TZVP/CMIRS(DME) Geometry Optimization"
    wB97M_V_def2_TZVP_CMIRS_DME__Frequency_Analysis = "wB97M-V/def2-TZVP/CMIRS(DME) Frequency Analysis"
    wB97M_V_def2_TZVP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_DME__Unknown = "wB97M-V/def2-TZVP/CMIRS(DME) Unknown"
    wB97M_V_def2_TZVP_CMIRS_EC__Single_Point = "wB97M-V/def2-TZVP/CMIRS(EC) Single Point"
    wB97M_V_def2_TZVP_CMIRS_EC__Force = "wB97M-V/def2-TZVP/CMIRS(EC) Force"
    wB97M_V_def2_TZVP_CMIRS_EC__Geometry_Optimization = "wB97M-V/def2-TZVP/CMIRS(EC) Geometry Optimization"
    wB97M_V_def2_TZVP_CMIRS_EC__Frequency_Analysis = "wB97M-V/def2-TZVP/CMIRS(EC) Frequency Analysis"
    wB97M_V_def2_TZVP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_EC__Unknown = "wB97M-V/def2-TZVP/CMIRS(EC) Unknown"
    wB97M_V_def2_TZVP_CMIRS_EC_EMC__Single_Point = "wB97M-V/def2-TZVP/CMIRS(EC/EMC) Single Point"
    wB97M_V_def2_TZVP_CMIRS_EC_EMC__Force = "wB97M-V/def2-TZVP/CMIRS(EC/EMC) Force"
    wB97M_V_def2_TZVP_CMIRS_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVP/CMIRS(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVP_CMIRS_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVP/CMIRS(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_EC_EMC__Unknown = "wB97M-V/def2-TZVP/CMIRS(EC/EMC) Unknown"
    wB97M_V_def2_TZVP_CMIRS_PC__Single_Point = "wB97M-V/def2-TZVP/CMIRS(PC) Single Point"
    wB97M_V_def2_TZVP_CMIRS_PC__Force = "wB97M-V/def2-TZVP/CMIRS(PC) Force"
    wB97M_V_def2_TZVP_CMIRS_PC__Geometry_Optimization = "wB97M-V/def2-TZVP/CMIRS(PC) Geometry Optimization"
    wB97M_V_def2_TZVP_CMIRS_PC__Frequency_Analysis = "wB97M-V/def2-TZVP/CMIRS(PC) Frequency Analysis"
    wB97M_V_def2_TZVP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVP_CMIRS_PC__Unknown = "wB97M-V/def2-TZVP/CMIRS(PC) Unknown"
    wB97M_V_def2_TZVPD_VACUUM_Single_Point = "wB97M-V/def2-TZVPD/VACUUM Single Point"
    wB97M_V_def2_TZVPD_VACUUM_Force = "wB97M-V/def2-TZVPD/VACUUM Force"
    wB97M_V_def2_TZVPD_VACUUM_Geometry_Optimization = "wB97M-V/def2-TZVPD/VACUUM Geometry Optimization"
    wB97M_V_def2_TZVPD_VACUUM_Frequency_Analysis = "wB97M-V/def2-TZVPD/VACUUM Frequency Analysis"
    wB97M_V_def2_TZVPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/VACUUM Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_VACUUM_Unknown = "wB97M-V/def2-TZVPD/VACUUM Unknown"
    wB97M_V_def2_TZVPD_PCM_WATER__Single_Point = "wB97M-V/def2-TZVPD/PCM(WATER) Single Point"
    wB97M_V_def2_TZVPD_PCM_WATER__Force = "wB97M-V/def2-TZVPD/PCM(WATER) Force"
    wB97M_V_def2_TZVPD_PCM_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPD/PCM(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPD_PCM_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPD/PCM(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_WATER__Unknown = "wB97M-V/def2-TZVPD/PCM(WATER) Unknown"
    wB97M_V_def2_TZVPD_PCM_THF__Single_Point = "wB97M-V/def2-TZVPD/PCM(THF) Single Point"
    wB97M_V_def2_TZVPD_PCM_THF__Force = "wB97M-V/def2-TZVPD/PCM(THF) Force"
    wB97M_V_def2_TZVPD_PCM_THF__Geometry_Optimization = "wB97M-V/def2-TZVPD/PCM(THF) Geometry Optimization"
    wB97M_V_def2_TZVPD_PCM_THF__Frequency_Analysis = "wB97M-V/def2-TZVPD/PCM(THF) Frequency Analysis"
    wB97M_V_def2_TZVPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_THF__Unknown = "wB97M-V/def2-TZVPD/PCM(THF) Unknown"
    wB97M_V_def2_TZVPD_PCM_G2__Single_Point = "wB97M-V/def2-TZVPD/PCM(G2) Single Point"
    wB97M_V_def2_TZVPD_PCM_G2__Force = "wB97M-V/def2-TZVPD/PCM(G2) Force"
    wB97M_V_def2_TZVPD_PCM_G2__Geometry_Optimization = "wB97M-V/def2-TZVPD/PCM(G2) Geometry Optimization"
    wB97M_V_def2_TZVPD_PCM_G2__Frequency_Analysis = "wB97M-V/def2-TZVPD/PCM(G2) Frequency Analysis"
    wB97M_V_def2_TZVPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_G2__Unknown = "wB97M-V/def2-TZVPD/PCM(G2) Unknown"
    wB97M_V_def2_TZVPD_PCM_DME__Single_Point = "wB97M-V/def2-TZVPD/PCM(DME) Single Point"
    wB97M_V_def2_TZVPD_PCM_DME__Force = "wB97M-V/def2-TZVPD/PCM(DME) Force"
    wB97M_V_def2_TZVPD_PCM_DME__Geometry_Optimization = "wB97M-V/def2-TZVPD/PCM(DME) Geometry Optimization"
    wB97M_V_def2_TZVPD_PCM_DME__Frequency_Analysis = "wB97M-V/def2-TZVPD/PCM(DME) Frequency Analysis"
    wB97M_V_def2_TZVPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_DME__Unknown = "wB97M-V/def2-TZVPD/PCM(DME) Unknown"
    wB97M_V_def2_TZVPD_PCM_EC__Single_Point = "wB97M-V/def2-TZVPD/PCM(EC) Single Point"
    wB97M_V_def2_TZVPD_PCM_EC__Force = "wB97M-V/def2-TZVPD/PCM(EC) Force"
    wB97M_V_def2_TZVPD_PCM_EC__Geometry_Optimization = "wB97M-V/def2-TZVPD/PCM(EC) Geometry Optimization"
    wB97M_V_def2_TZVPD_PCM_EC__Frequency_Analysis = "wB97M-V/def2-TZVPD/PCM(EC) Frequency Analysis"
    wB97M_V_def2_TZVPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_EC__Unknown = "wB97M-V/def2-TZVPD/PCM(EC) Unknown"
    wB97M_V_def2_TZVPD_PCM_EC_EMC__Single_Point = "wB97M-V/def2-TZVPD/PCM(EC/EMC) Single Point"
    wB97M_V_def2_TZVPD_PCM_EC_EMC__Force = "wB97M-V/def2-TZVPD/PCM(EC/EMC) Force"
    wB97M_V_def2_TZVPD_PCM_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVPD/PCM(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVPD_PCM_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPD/PCM(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_EC_EMC__Unknown = "wB97M-V/def2-TZVPD/PCM(EC/EMC) Unknown"
    wB97M_V_def2_TZVPD_PCM_PC__Single_Point = "wB97M-V/def2-TZVPD/PCM(PC) Single Point"
    wB97M_V_def2_TZVPD_PCM_PC__Force = "wB97M-V/def2-TZVPD/PCM(PC) Force"
    wB97M_V_def2_TZVPD_PCM_PC__Geometry_Optimization = "wB97M-V/def2-TZVPD/PCM(PC) Geometry Optimization"
    wB97M_V_def2_TZVPD_PCM_PC__Frequency_Analysis = "wB97M-V/def2-TZVPD/PCM(PC) Frequency Analysis"
    wB97M_V_def2_TZVPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_PCM_PC__Unknown = "wB97M-V/def2-TZVPD/PCM(PC) Unknown"
    wB97M_V_def2_TZVPD_ISOSVP_WATER__Single_Point = "wB97M-V/def2-TZVPD/ISOSVP(WATER) Single Point"
    wB97M_V_def2_TZVPD_ISOSVP_WATER__Force = "wB97M-V/def2-TZVPD/ISOSVP(WATER) Force"
    wB97M_V_def2_TZVPD_ISOSVP_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPD/ISOSVP(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPD_ISOSVP_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPD/ISOSVP(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_WATER__Unknown = "wB97M-V/def2-TZVPD/ISOSVP(WATER) Unknown"
    wB97M_V_def2_TZVPD_ISOSVP_THF__Single_Point = "wB97M-V/def2-TZVPD/ISOSVP(THF) Single Point"
    wB97M_V_def2_TZVPD_ISOSVP_THF__Force = "wB97M-V/def2-TZVPD/ISOSVP(THF) Force"
    wB97M_V_def2_TZVPD_ISOSVP_THF__Geometry_Optimization = "wB97M-V/def2-TZVPD/ISOSVP(THF) Geometry Optimization"
    wB97M_V_def2_TZVPD_ISOSVP_THF__Frequency_Analysis = "wB97M-V/def2-TZVPD/ISOSVP(THF) Frequency Analysis"
    wB97M_V_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_THF__Unknown = "wB97M-V/def2-TZVPD/ISOSVP(THF) Unknown"
    wB97M_V_def2_TZVPD_ISOSVP_G2__Single_Point = "wB97M-V/def2-TZVPD/ISOSVP(G2) Single Point"
    wB97M_V_def2_TZVPD_ISOSVP_G2__Force = "wB97M-V/def2-TZVPD/ISOSVP(G2) Force"
    wB97M_V_def2_TZVPD_ISOSVP_G2__Geometry_Optimization = "wB97M-V/def2-TZVPD/ISOSVP(G2) Geometry Optimization"
    wB97M_V_def2_TZVPD_ISOSVP_G2__Frequency_Analysis = "wB97M-V/def2-TZVPD/ISOSVP(G2) Frequency Analysis"
    wB97M_V_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_G2__Unknown = "wB97M-V/def2-TZVPD/ISOSVP(G2) Unknown"
    wB97M_V_def2_TZVPD_ISOSVP_DME__Single_Point = "wB97M-V/def2-TZVPD/ISOSVP(DME) Single Point"
    wB97M_V_def2_TZVPD_ISOSVP_DME__Force = "wB97M-V/def2-TZVPD/ISOSVP(DME) Force"
    wB97M_V_def2_TZVPD_ISOSVP_DME__Geometry_Optimization = "wB97M-V/def2-TZVPD/ISOSVP(DME) Geometry Optimization"
    wB97M_V_def2_TZVPD_ISOSVP_DME__Frequency_Analysis = "wB97M-V/def2-TZVPD/ISOSVP(DME) Frequency Analysis"
    wB97M_V_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_DME__Unknown = "wB97M-V/def2-TZVPD/ISOSVP(DME) Unknown"
    wB97M_V_def2_TZVPD_ISOSVP_EC__Single_Point = "wB97M-V/def2-TZVPD/ISOSVP(EC) Single Point"
    wB97M_V_def2_TZVPD_ISOSVP_EC__Force = "wB97M-V/def2-TZVPD/ISOSVP(EC) Force"
    wB97M_V_def2_TZVPD_ISOSVP_EC__Geometry_Optimization = "wB97M-V/def2-TZVPD/ISOSVP(EC) Geometry Optimization"
    wB97M_V_def2_TZVPD_ISOSVP_EC__Frequency_Analysis = "wB97M-V/def2-TZVPD/ISOSVP(EC) Frequency Analysis"
    wB97M_V_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_EC__Unknown = "wB97M-V/def2-TZVPD/ISOSVP(EC) Unknown"
    wB97M_V_def2_TZVPD_ISOSVP_EC_EMC__Single_Point = "wB97M-V/def2-TZVPD/ISOSVP(EC/EMC) Single Point"
    wB97M_V_def2_TZVPD_ISOSVP_EC_EMC__Force = "wB97M-V/def2-TZVPD/ISOSVP(EC/EMC) Force"
    wB97M_V_def2_TZVPD_ISOSVP_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVPD/ISOSVP(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_EC_EMC__Unknown = "wB97M-V/def2-TZVPD/ISOSVP(EC/EMC) Unknown"
    wB97M_V_def2_TZVPD_ISOSVP_PC__Single_Point = "wB97M-V/def2-TZVPD/ISOSVP(PC) Single Point"
    wB97M_V_def2_TZVPD_ISOSVP_PC__Force = "wB97M-V/def2-TZVPD/ISOSVP(PC) Force"
    wB97M_V_def2_TZVPD_ISOSVP_PC__Geometry_Optimization = "wB97M-V/def2-TZVPD/ISOSVP(PC) Geometry Optimization"
    wB97M_V_def2_TZVPD_ISOSVP_PC__Frequency_Analysis = "wB97M-V/def2-TZVPD/ISOSVP(PC) Frequency Analysis"
    wB97M_V_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_ISOSVP_PC__Unknown = "wB97M-V/def2-TZVPD/ISOSVP(PC) Unknown"
    wB97M_V_def2_TZVPD_SMD_WATER__Single_Point = "wB97M-V/def2-TZVPD/SMD(WATER) Single Point"
    wB97M_V_def2_TZVPD_SMD_WATER__Force = "wB97M-V/def2-TZVPD/SMD(WATER) Force"
    wB97M_V_def2_TZVPD_SMD_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPD/SMD(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPD_SMD_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPD/SMD(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_WATER__Unknown = "wB97M-V/def2-TZVPD/SMD(WATER) Unknown"
    wB97M_V_def2_TZVPD_SMD_THF__Single_Point = "wB97M-V/def2-TZVPD/SMD(THF) Single Point"
    wB97M_V_def2_TZVPD_SMD_THF__Force = "wB97M-V/def2-TZVPD/SMD(THF) Force"
    wB97M_V_def2_TZVPD_SMD_THF__Geometry_Optimization = "wB97M-V/def2-TZVPD/SMD(THF) Geometry Optimization"
    wB97M_V_def2_TZVPD_SMD_THF__Frequency_Analysis = "wB97M-V/def2-TZVPD/SMD(THF) Frequency Analysis"
    wB97M_V_def2_TZVPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_THF__Unknown = "wB97M-V/def2-TZVPD/SMD(THF) Unknown"
    wB97M_V_def2_TZVPD_SMD_G2__Single_Point = "wB97M-V/def2-TZVPD/SMD(G2) Single Point"
    wB97M_V_def2_TZVPD_SMD_G2__Force = "wB97M-V/def2-TZVPD/SMD(G2) Force"
    wB97M_V_def2_TZVPD_SMD_G2__Geometry_Optimization = "wB97M-V/def2-TZVPD/SMD(G2) Geometry Optimization"
    wB97M_V_def2_TZVPD_SMD_G2__Frequency_Analysis = "wB97M-V/def2-TZVPD/SMD(G2) Frequency Analysis"
    wB97M_V_def2_TZVPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_G2__Unknown = "wB97M-V/def2-TZVPD/SMD(G2) Unknown"
    wB97M_V_def2_TZVPD_SMD_DME__Single_Point = "wB97M-V/def2-TZVPD/SMD(DME) Single Point"
    wB97M_V_def2_TZVPD_SMD_DME__Force = "wB97M-V/def2-TZVPD/SMD(DME) Force"
    wB97M_V_def2_TZVPD_SMD_DME__Geometry_Optimization = "wB97M-V/def2-TZVPD/SMD(DME) Geometry Optimization"
    wB97M_V_def2_TZVPD_SMD_DME__Frequency_Analysis = "wB97M-V/def2-TZVPD/SMD(DME) Frequency Analysis"
    wB97M_V_def2_TZVPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_DME__Unknown = "wB97M-V/def2-TZVPD/SMD(DME) Unknown"
    wB97M_V_def2_TZVPD_SMD_EC__Single_Point = "wB97M-V/def2-TZVPD/SMD(EC) Single Point"
    wB97M_V_def2_TZVPD_SMD_EC__Force = "wB97M-V/def2-TZVPD/SMD(EC) Force"
    wB97M_V_def2_TZVPD_SMD_EC__Geometry_Optimization = "wB97M-V/def2-TZVPD/SMD(EC) Geometry Optimization"
    wB97M_V_def2_TZVPD_SMD_EC__Frequency_Analysis = "wB97M-V/def2-TZVPD/SMD(EC) Frequency Analysis"
    wB97M_V_def2_TZVPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_EC__Unknown = "wB97M-V/def2-TZVPD/SMD(EC) Unknown"
    wB97M_V_def2_TZVPD_SMD_EC_EMC__Single_Point = "wB97M-V/def2-TZVPD/SMD(EC/EMC) Single Point"
    wB97M_V_def2_TZVPD_SMD_EC_EMC__Force = "wB97M-V/def2-TZVPD/SMD(EC/EMC) Force"
    wB97M_V_def2_TZVPD_SMD_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVPD/SMD(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVPD_SMD_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPD/SMD(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_EC_EMC__Unknown = "wB97M-V/def2-TZVPD/SMD(EC/EMC) Unknown"
    wB97M_V_def2_TZVPD_SMD_PC__Single_Point = "wB97M-V/def2-TZVPD/SMD(PC) Single Point"
    wB97M_V_def2_TZVPD_SMD_PC__Force = "wB97M-V/def2-TZVPD/SMD(PC) Force"
    wB97M_V_def2_TZVPD_SMD_PC__Geometry_Optimization = "wB97M-V/def2-TZVPD/SMD(PC) Geometry Optimization"
    wB97M_V_def2_TZVPD_SMD_PC__Frequency_Analysis = "wB97M-V/def2-TZVPD/SMD(PC) Frequency Analysis"
    wB97M_V_def2_TZVPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_SMD_PC__Unknown = "wB97M-V/def2-TZVPD/SMD(PC) Unknown"
    wB97M_V_def2_TZVPD_CMIRS_WATER__Single_Point = "wB97M-V/def2-TZVPD/CMIRS(WATER) Single Point"
    wB97M_V_def2_TZVPD_CMIRS_WATER__Force = "wB97M-V/def2-TZVPD/CMIRS(WATER) Force"
    wB97M_V_def2_TZVPD_CMIRS_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPD/CMIRS(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPD_CMIRS_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPD/CMIRS(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_WATER__Unknown = "wB97M-V/def2-TZVPD/CMIRS(WATER) Unknown"
    wB97M_V_def2_TZVPD_CMIRS_THF__Single_Point = "wB97M-V/def2-TZVPD/CMIRS(THF) Single Point"
    wB97M_V_def2_TZVPD_CMIRS_THF__Force = "wB97M-V/def2-TZVPD/CMIRS(THF) Force"
    wB97M_V_def2_TZVPD_CMIRS_THF__Geometry_Optimization = "wB97M-V/def2-TZVPD/CMIRS(THF) Geometry Optimization"
    wB97M_V_def2_TZVPD_CMIRS_THF__Frequency_Analysis = "wB97M-V/def2-TZVPD/CMIRS(THF) Frequency Analysis"
    wB97M_V_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_THF__Unknown = "wB97M-V/def2-TZVPD/CMIRS(THF) Unknown"
    wB97M_V_def2_TZVPD_CMIRS_G2__Single_Point = "wB97M-V/def2-TZVPD/CMIRS(G2) Single Point"
    wB97M_V_def2_TZVPD_CMIRS_G2__Force = "wB97M-V/def2-TZVPD/CMIRS(G2) Force"
    wB97M_V_def2_TZVPD_CMIRS_G2__Geometry_Optimization = "wB97M-V/def2-TZVPD/CMIRS(G2) Geometry Optimization"
    wB97M_V_def2_TZVPD_CMIRS_G2__Frequency_Analysis = "wB97M-V/def2-TZVPD/CMIRS(G2) Frequency Analysis"
    wB97M_V_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_G2__Unknown = "wB97M-V/def2-TZVPD/CMIRS(G2) Unknown"
    wB97M_V_def2_TZVPD_CMIRS_DME__Single_Point = "wB97M-V/def2-TZVPD/CMIRS(DME) Single Point"
    wB97M_V_def2_TZVPD_CMIRS_DME__Force = "wB97M-V/def2-TZVPD/CMIRS(DME) Force"
    wB97M_V_def2_TZVPD_CMIRS_DME__Geometry_Optimization = "wB97M-V/def2-TZVPD/CMIRS(DME) Geometry Optimization"
    wB97M_V_def2_TZVPD_CMIRS_DME__Frequency_Analysis = "wB97M-V/def2-TZVPD/CMIRS(DME) Frequency Analysis"
    wB97M_V_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_DME__Unknown = "wB97M-V/def2-TZVPD/CMIRS(DME) Unknown"
    wB97M_V_def2_TZVPD_CMIRS_EC__Single_Point = "wB97M-V/def2-TZVPD/CMIRS(EC) Single Point"
    wB97M_V_def2_TZVPD_CMIRS_EC__Force = "wB97M-V/def2-TZVPD/CMIRS(EC) Force"
    wB97M_V_def2_TZVPD_CMIRS_EC__Geometry_Optimization = "wB97M-V/def2-TZVPD/CMIRS(EC) Geometry Optimization"
    wB97M_V_def2_TZVPD_CMIRS_EC__Frequency_Analysis = "wB97M-V/def2-TZVPD/CMIRS(EC) Frequency Analysis"
    wB97M_V_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_EC__Unknown = "wB97M-V/def2-TZVPD/CMIRS(EC) Unknown"
    wB97M_V_def2_TZVPD_CMIRS_EC_EMC__Single_Point = "wB97M-V/def2-TZVPD/CMIRS(EC/EMC) Single Point"
    wB97M_V_def2_TZVPD_CMIRS_EC_EMC__Force = "wB97M-V/def2-TZVPD/CMIRS(EC/EMC) Force"
    wB97M_V_def2_TZVPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_EC_EMC__Unknown = "wB97M-V/def2-TZVPD/CMIRS(EC/EMC) Unknown"
    wB97M_V_def2_TZVPD_CMIRS_PC__Single_Point = "wB97M-V/def2-TZVPD/CMIRS(PC) Single Point"
    wB97M_V_def2_TZVPD_CMIRS_PC__Force = "wB97M-V/def2-TZVPD/CMIRS(PC) Force"
    wB97M_V_def2_TZVPD_CMIRS_PC__Geometry_Optimization = "wB97M-V/def2-TZVPD/CMIRS(PC) Geometry Optimization"
    wB97M_V_def2_TZVPD_CMIRS_PC__Frequency_Analysis = "wB97M-V/def2-TZVPD/CMIRS(PC) Frequency Analysis"
    wB97M_V_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPD_CMIRS_PC__Unknown = "wB97M-V/def2-TZVPD/CMIRS(PC) Unknown"
    wB97M_V_def2_TZVPP_VACUUM_Single_Point = "wB97M-V/def2-TZVPP/VACUUM Single Point"
    wB97M_V_def2_TZVPP_VACUUM_Force = "wB97M-V/def2-TZVPP/VACUUM Force"
    wB97M_V_def2_TZVPP_VACUUM_Geometry_Optimization = "wB97M-V/def2-TZVPP/VACUUM Geometry Optimization"
    wB97M_V_def2_TZVPP_VACUUM_Frequency_Analysis = "wB97M-V/def2-TZVPP/VACUUM Frequency Analysis"
    wB97M_V_def2_TZVPP_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/VACUUM Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_VACUUM_Unknown = "wB97M-V/def2-TZVPP/VACUUM Unknown"
    wB97M_V_def2_TZVPP_PCM_WATER__Single_Point = "wB97M-V/def2-TZVPP/PCM(WATER) Single Point"
    wB97M_V_def2_TZVPP_PCM_WATER__Force = "wB97M-V/def2-TZVPP/PCM(WATER) Force"
    wB97M_V_def2_TZVPP_PCM_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPP/PCM(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPP_PCM_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPP/PCM(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPP_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_WATER__Unknown = "wB97M-V/def2-TZVPP/PCM(WATER) Unknown"
    wB97M_V_def2_TZVPP_PCM_THF__Single_Point = "wB97M-V/def2-TZVPP/PCM(THF) Single Point"
    wB97M_V_def2_TZVPP_PCM_THF__Force = "wB97M-V/def2-TZVPP/PCM(THF) Force"
    wB97M_V_def2_TZVPP_PCM_THF__Geometry_Optimization = "wB97M-V/def2-TZVPP/PCM(THF) Geometry Optimization"
    wB97M_V_def2_TZVPP_PCM_THF__Frequency_Analysis = "wB97M-V/def2-TZVPP/PCM(THF) Frequency Analysis"
    wB97M_V_def2_TZVPP_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_THF__Unknown = "wB97M-V/def2-TZVPP/PCM(THF) Unknown"
    wB97M_V_def2_TZVPP_PCM_G2__Single_Point = "wB97M-V/def2-TZVPP/PCM(G2) Single Point"
    wB97M_V_def2_TZVPP_PCM_G2__Force = "wB97M-V/def2-TZVPP/PCM(G2) Force"
    wB97M_V_def2_TZVPP_PCM_G2__Geometry_Optimization = "wB97M-V/def2-TZVPP/PCM(G2) Geometry Optimization"
    wB97M_V_def2_TZVPP_PCM_G2__Frequency_Analysis = "wB97M-V/def2-TZVPP/PCM(G2) Frequency Analysis"
    wB97M_V_def2_TZVPP_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_G2__Unknown = "wB97M-V/def2-TZVPP/PCM(G2) Unknown"
    wB97M_V_def2_TZVPP_PCM_DME__Single_Point = "wB97M-V/def2-TZVPP/PCM(DME) Single Point"
    wB97M_V_def2_TZVPP_PCM_DME__Force = "wB97M-V/def2-TZVPP/PCM(DME) Force"
    wB97M_V_def2_TZVPP_PCM_DME__Geometry_Optimization = "wB97M-V/def2-TZVPP/PCM(DME) Geometry Optimization"
    wB97M_V_def2_TZVPP_PCM_DME__Frequency_Analysis = "wB97M-V/def2-TZVPP/PCM(DME) Frequency Analysis"
    wB97M_V_def2_TZVPP_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_DME__Unknown = "wB97M-V/def2-TZVPP/PCM(DME) Unknown"
    wB97M_V_def2_TZVPP_PCM_EC__Single_Point = "wB97M-V/def2-TZVPP/PCM(EC) Single Point"
    wB97M_V_def2_TZVPP_PCM_EC__Force = "wB97M-V/def2-TZVPP/PCM(EC) Force"
    wB97M_V_def2_TZVPP_PCM_EC__Geometry_Optimization = "wB97M-V/def2-TZVPP/PCM(EC) Geometry Optimization"
    wB97M_V_def2_TZVPP_PCM_EC__Frequency_Analysis = "wB97M-V/def2-TZVPP/PCM(EC) Frequency Analysis"
    wB97M_V_def2_TZVPP_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_EC__Unknown = "wB97M-V/def2-TZVPP/PCM(EC) Unknown"
    wB97M_V_def2_TZVPP_PCM_EC_EMC__Single_Point = "wB97M-V/def2-TZVPP/PCM(EC/EMC) Single Point"
    wB97M_V_def2_TZVPP_PCM_EC_EMC__Force = "wB97M-V/def2-TZVPP/PCM(EC/EMC) Force"
    wB97M_V_def2_TZVPP_PCM_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVPP/PCM(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVPP_PCM_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPP/PCM(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_EC_EMC__Unknown = "wB97M-V/def2-TZVPP/PCM(EC/EMC) Unknown"
    wB97M_V_def2_TZVPP_PCM_PC__Single_Point = "wB97M-V/def2-TZVPP/PCM(PC) Single Point"
    wB97M_V_def2_TZVPP_PCM_PC__Force = "wB97M-V/def2-TZVPP/PCM(PC) Force"
    wB97M_V_def2_TZVPP_PCM_PC__Geometry_Optimization = "wB97M-V/def2-TZVPP/PCM(PC) Geometry Optimization"
    wB97M_V_def2_TZVPP_PCM_PC__Frequency_Analysis = "wB97M-V/def2-TZVPP/PCM(PC) Frequency Analysis"
    wB97M_V_def2_TZVPP_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_PCM_PC__Unknown = "wB97M-V/def2-TZVPP/PCM(PC) Unknown"
    wB97M_V_def2_TZVPP_ISOSVP_WATER__Single_Point = "wB97M-V/def2-TZVPP/ISOSVP(WATER) Single Point"
    wB97M_V_def2_TZVPP_ISOSVP_WATER__Force = "wB97M-V/def2-TZVPP/ISOSVP(WATER) Force"
    wB97M_V_def2_TZVPP_ISOSVP_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPP/ISOSVP(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPP_ISOSVP_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPP/ISOSVP(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_WATER__Unknown = "wB97M-V/def2-TZVPP/ISOSVP(WATER) Unknown"
    wB97M_V_def2_TZVPP_ISOSVP_THF__Single_Point = "wB97M-V/def2-TZVPP/ISOSVP(THF) Single Point"
    wB97M_V_def2_TZVPP_ISOSVP_THF__Force = "wB97M-V/def2-TZVPP/ISOSVP(THF) Force"
    wB97M_V_def2_TZVPP_ISOSVP_THF__Geometry_Optimization = "wB97M-V/def2-TZVPP/ISOSVP(THF) Geometry Optimization"
    wB97M_V_def2_TZVPP_ISOSVP_THF__Frequency_Analysis = "wB97M-V/def2-TZVPP/ISOSVP(THF) Frequency Analysis"
    wB97M_V_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_THF__Unknown = "wB97M-V/def2-TZVPP/ISOSVP(THF) Unknown"
    wB97M_V_def2_TZVPP_ISOSVP_G2__Single_Point = "wB97M-V/def2-TZVPP/ISOSVP(G2) Single Point"
    wB97M_V_def2_TZVPP_ISOSVP_G2__Force = "wB97M-V/def2-TZVPP/ISOSVP(G2) Force"
    wB97M_V_def2_TZVPP_ISOSVP_G2__Geometry_Optimization = "wB97M-V/def2-TZVPP/ISOSVP(G2) Geometry Optimization"
    wB97M_V_def2_TZVPP_ISOSVP_G2__Frequency_Analysis = "wB97M-V/def2-TZVPP/ISOSVP(G2) Frequency Analysis"
    wB97M_V_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_G2__Unknown = "wB97M-V/def2-TZVPP/ISOSVP(G2) Unknown"
    wB97M_V_def2_TZVPP_ISOSVP_DME__Single_Point = "wB97M-V/def2-TZVPP/ISOSVP(DME) Single Point"
    wB97M_V_def2_TZVPP_ISOSVP_DME__Force = "wB97M-V/def2-TZVPP/ISOSVP(DME) Force"
    wB97M_V_def2_TZVPP_ISOSVP_DME__Geometry_Optimization = "wB97M-V/def2-TZVPP/ISOSVP(DME) Geometry Optimization"
    wB97M_V_def2_TZVPP_ISOSVP_DME__Frequency_Analysis = "wB97M-V/def2-TZVPP/ISOSVP(DME) Frequency Analysis"
    wB97M_V_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_DME__Unknown = "wB97M-V/def2-TZVPP/ISOSVP(DME) Unknown"
    wB97M_V_def2_TZVPP_ISOSVP_EC__Single_Point = "wB97M-V/def2-TZVPP/ISOSVP(EC) Single Point"
    wB97M_V_def2_TZVPP_ISOSVP_EC__Force = "wB97M-V/def2-TZVPP/ISOSVP(EC) Force"
    wB97M_V_def2_TZVPP_ISOSVP_EC__Geometry_Optimization = "wB97M-V/def2-TZVPP/ISOSVP(EC) Geometry Optimization"
    wB97M_V_def2_TZVPP_ISOSVP_EC__Frequency_Analysis = "wB97M-V/def2-TZVPP/ISOSVP(EC) Frequency Analysis"
    wB97M_V_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_EC__Unknown = "wB97M-V/def2-TZVPP/ISOSVP(EC) Unknown"
    wB97M_V_def2_TZVPP_ISOSVP_EC_EMC__Single_Point = "wB97M-V/def2-TZVPP/ISOSVP(EC/EMC) Single Point"
    wB97M_V_def2_TZVPP_ISOSVP_EC_EMC__Force = "wB97M-V/def2-TZVPP/ISOSVP(EC/EMC) Force"
    wB97M_V_def2_TZVPP_ISOSVP_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVPP/ISOSVP(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPP/ISOSVP(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_EC_EMC__Unknown = "wB97M-V/def2-TZVPP/ISOSVP(EC/EMC) Unknown"
    wB97M_V_def2_TZVPP_ISOSVP_PC__Single_Point = "wB97M-V/def2-TZVPP/ISOSVP(PC) Single Point"
    wB97M_V_def2_TZVPP_ISOSVP_PC__Force = "wB97M-V/def2-TZVPP/ISOSVP(PC) Force"
    wB97M_V_def2_TZVPP_ISOSVP_PC__Geometry_Optimization = "wB97M-V/def2-TZVPP/ISOSVP(PC) Geometry Optimization"
    wB97M_V_def2_TZVPP_ISOSVP_PC__Frequency_Analysis = "wB97M-V/def2-TZVPP/ISOSVP(PC) Frequency Analysis"
    wB97M_V_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_ISOSVP_PC__Unknown = "wB97M-V/def2-TZVPP/ISOSVP(PC) Unknown"
    wB97M_V_def2_TZVPP_SMD_WATER__Single_Point = "wB97M-V/def2-TZVPP/SMD(WATER) Single Point"
    wB97M_V_def2_TZVPP_SMD_WATER__Force = "wB97M-V/def2-TZVPP/SMD(WATER) Force"
    wB97M_V_def2_TZVPP_SMD_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPP/SMD(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPP_SMD_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPP/SMD(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPP_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_WATER__Unknown = "wB97M-V/def2-TZVPP/SMD(WATER) Unknown"
    wB97M_V_def2_TZVPP_SMD_THF__Single_Point = "wB97M-V/def2-TZVPP/SMD(THF) Single Point"
    wB97M_V_def2_TZVPP_SMD_THF__Force = "wB97M-V/def2-TZVPP/SMD(THF) Force"
    wB97M_V_def2_TZVPP_SMD_THF__Geometry_Optimization = "wB97M-V/def2-TZVPP/SMD(THF) Geometry Optimization"
    wB97M_V_def2_TZVPP_SMD_THF__Frequency_Analysis = "wB97M-V/def2-TZVPP/SMD(THF) Frequency Analysis"
    wB97M_V_def2_TZVPP_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_THF__Unknown = "wB97M-V/def2-TZVPP/SMD(THF) Unknown"
    wB97M_V_def2_TZVPP_SMD_G2__Single_Point = "wB97M-V/def2-TZVPP/SMD(G2) Single Point"
    wB97M_V_def2_TZVPP_SMD_G2__Force = "wB97M-V/def2-TZVPP/SMD(G2) Force"
    wB97M_V_def2_TZVPP_SMD_G2__Geometry_Optimization = "wB97M-V/def2-TZVPP/SMD(G2) Geometry Optimization"
    wB97M_V_def2_TZVPP_SMD_G2__Frequency_Analysis = "wB97M-V/def2-TZVPP/SMD(G2) Frequency Analysis"
    wB97M_V_def2_TZVPP_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_G2__Unknown = "wB97M-V/def2-TZVPP/SMD(G2) Unknown"
    wB97M_V_def2_TZVPP_SMD_DME__Single_Point = "wB97M-V/def2-TZVPP/SMD(DME) Single Point"
    wB97M_V_def2_TZVPP_SMD_DME__Force = "wB97M-V/def2-TZVPP/SMD(DME) Force"
    wB97M_V_def2_TZVPP_SMD_DME__Geometry_Optimization = "wB97M-V/def2-TZVPP/SMD(DME) Geometry Optimization"
    wB97M_V_def2_TZVPP_SMD_DME__Frequency_Analysis = "wB97M-V/def2-TZVPP/SMD(DME) Frequency Analysis"
    wB97M_V_def2_TZVPP_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_DME__Unknown = "wB97M-V/def2-TZVPP/SMD(DME) Unknown"
    wB97M_V_def2_TZVPP_SMD_EC__Single_Point = "wB97M-V/def2-TZVPP/SMD(EC) Single Point"
    wB97M_V_def2_TZVPP_SMD_EC__Force = "wB97M-V/def2-TZVPP/SMD(EC) Force"
    wB97M_V_def2_TZVPP_SMD_EC__Geometry_Optimization = "wB97M-V/def2-TZVPP/SMD(EC) Geometry Optimization"
    wB97M_V_def2_TZVPP_SMD_EC__Frequency_Analysis = "wB97M-V/def2-TZVPP/SMD(EC) Frequency Analysis"
    wB97M_V_def2_TZVPP_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_EC__Unknown = "wB97M-V/def2-TZVPP/SMD(EC) Unknown"
    wB97M_V_def2_TZVPP_SMD_EC_EMC__Single_Point = "wB97M-V/def2-TZVPP/SMD(EC/EMC) Single Point"
    wB97M_V_def2_TZVPP_SMD_EC_EMC__Force = "wB97M-V/def2-TZVPP/SMD(EC/EMC) Force"
    wB97M_V_def2_TZVPP_SMD_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVPP/SMD(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVPP_SMD_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPP/SMD(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_EC_EMC__Unknown = "wB97M-V/def2-TZVPP/SMD(EC/EMC) Unknown"
    wB97M_V_def2_TZVPP_SMD_PC__Single_Point = "wB97M-V/def2-TZVPP/SMD(PC) Single Point"
    wB97M_V_def2_TZVPP_SMD_PC__Force = "wB97M-V/def2-TZVPP/SMD(PC) Force"
    wB97M_V_def2_TZVPP_SMD_PC__Geometry_Optimization = "wB97M-V/def2-TZVPP/SMD(PC) Geometry Optimization"
    wB97M_V_def2_TZVPP_SMD_PC__Frequency_Analysis = "wB97M-V/def2-TZVPP/SMD(PC) Frequency Analysis"
    wB97M_V_def2_TZVPP_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_SMD_PC__Unknown = "wB97M-V/def2-TZVPP/SMD(PC) Unknown"
    wB97M_V_def2_TZVPP_CMIRS_WATER__Single_Point = "wB97M-V/def2-TZVPP/CMIRS(WATER) Single Point"
    wB97M_V_def2_TZVPP_CMIRS_WATER__Force = "wB97M-V/def2-TZVPP/CMIRS(WATER) Force"
    wB97M_V_def2_TZVPP_CMIRS_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPP/CMIRS(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPP_CMIRS_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPP/CMIRS(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_WATER__Unknown = "wB97M-V/def2-TZVPP/CMIRS(WATER) Unknown"
    wB97M_V_def2_TZVPP_CMIRS_THF__Single_Point = "wB97M-V/def2-TZVPP/CMIRS(THF) Single Point"
    wB97M_V_def2_TZVPP_CMIRS_THF__Force = "wB97M-V/def2-TZVPP/CMIRS(THF) Force"
    wB97M_V_def2_TZVPP_CMIRS_THF__Geometry_Optimization = "wB97M-V/def2-TZVPP/CMIRS(THF) Geometry Optimization"
    wB97M_V_def2_TZVPP_CMIRS_THF__Frequency_Analysis = "wB97M-V/def2-TZVPP/CMIRS(THF) Frequency Analysis"
    wB97M_V_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_THF__Unknown = "wB97M-V/def2-TZVPP/CMIRS(THF) Unknown"
    wB97M_V_def2_TZVPP_CMIRS_G2__Single_Point = "wB97M-V/def2-TZVPP/CMIRS(G2) Single Point"
    wB97M_V_def2_TZVPP_CMIRS_G2__Force = "wB97M-V/def2-TZVPP/CMIRS(G2) Force"
    wB97M_V_def2_TZVPP_CMIRS_G2__Geometry_Optimization = "wB97M-V/def2-TZVPP/CMIRS(G2) Geometry Optimization"
    wB97M_V_def2_TZVPP_CMIRS_G2__Frequency_Analysis = "wB97M-V/def2-TZVPP/CMIRS(G2) Frequency Analysis"
    wB97M_V_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_G2__Unknown = "wB97M-V/def2-TZVPP/CMIRS(G2) Unknown"
    wB97M_V_def2_TZVPP_CMIRS_DME__Single_Point = "wB97M-V/def2-TZVPP/CMIRS(DME) Single Point"
    wB97M_V_def2_TZVPP_CMIRS_DME__Force = "wB97M-V/def2-TZVPP/CMIRS(DME) Force"
    wB97M_V_def2_TZVPP_CMIRS_DME__Geometry_Optimization = "wB97M-V/def2-TZVPP/CMIRS(DME) Geometry Optimization"
    wB97M_V_def2_TZVPP_CMIRS_DME__Frequency_Analysis = "wB97M-V/def2-TZVPP/CMIRS(DME) Frequency Analysis"
    wB97M_V_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_DME__Unknown = "wB97M-V/def2-TZVPP/CMIRS(DME) Unknown"
    wB97M_V_def2_TZVPP_CMIRS_EC__Single_Point = "wB97M-V/def2-TZVPP/CMIRS(EC) Single Point"
    wB97M_V_def2_TZVPP_CMIRS_EC__Force = "wB97M-V/def2-TZVPP/CMIRS(EC) Force"
    wB97M_V_def2_TZVPP_CMIRS_EC__Geometry_Optimization = "wB97M-V/def2-TZVPP/CMIRS(EC) Geometry Optimization"
    wB97M_V_def2_TZVPP_CMIRS_EC__Frequency_Analysis = "wB97M-V/def2-TZVPP/CMIRS(EC) Frequency Analysis"
    wB97M_V_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_EC__Unknown = "wB97M-V/def2-TZVPP/CMIRS(EC) Unknown"
    wB97M_V_def2_TZVPP_CMIRS_EC_EMC__Single_Point = "wB97M-V/def2-TZVPP/CMIRS(EC/EMC) Single Point"
    wB97M_V_def2_TZVPP_CMIRS_EC_EMC__Force = "wB97M-V/def2-TZVPP/CMIRS(EC/EMC) Force"
    wB97M_V_def2_TZVPP_CMIRS_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVPP/CMIRS(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVPP_CMIRS_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPP/CMIRS(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_EC_EMC__Unknown = "wB97M-V/def2-TZVPP/CMIRS(EC/EMC) Unknown"
    wB97M_V_def2_TZVPP_CMIRS_PC__Single_Point = "wB97M-V/def2-TZVPP/CMIRS(PC) Single Point"
    wB97M_V_def2_TZVPP_CMIRS_PC__Force = "wB97M-V/def2-TZVPP/CMIRS(PC) Force"
    wB97M_V_def2_TZVPP_CMIRS_PC__Geometry_Optimization = "wB97M-V/def2-TZVPP/CMIRS(PC) Geometry Optimization"
    wB97M_V_def2_TZVPP_CMIRS_PC__Frequency_Analysis = "wB97M-V/def2-TZVPP/CMIRS(PC) Frequency Analysis"
    wB97M_V_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPP/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPP_CMIRS_PC__Unknown = "wB97M-V/def2-TZVPP/CMIRS(PC) Unknown"
    wB97M_V_def2_TZVPPD_VACUUM_Single_Point = "wB97M-V/def2-TZVPPD/VACUUM Single Point"
    wB97M_V_def2_TZVPPD_VACUUM_Force = "wB97M-V/def2-TZVPPD/VACUUM Force"
    wB97M_V_def2_TZVPPD_VACUUM_Geometry_Optimization = "wB97M-V/def2-TZVPPD/VACUUM Geometry Optimization"
    wB97M_V_def2_TZVPPD_VACUUM_Frequency_Analysis = "wB97M-V/def2-TZVPPD/VACUUM Frequency Analysis"
    wB97M_V_def2_TZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/VACUUM Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_VACUUM_Unknown = "wB97M-V/def2-TZVPPD/VACUUM Unknown"
    wB97M_V_def2_TZVPPD_PCM_WATER__Single_Point = "wB97M-V/def2-TZVPPD/PCM(WATER) Single Point"
    wB97M_V_def2_TZVPPD_PCM_WATER__Force = "wB97M-V/def2-TZVPPD/PCM(WATER) Force"
    wB97M_V_def2_TZVPPD_PCM_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPPD/PCM(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPPD_PCM_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPPD/PCM(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_WATER__Unknown = "wB97M-V/def2-TZVPPD/PCM(WATER) Unknown"
    wB97M_V_def2_TZVPPD_PCM_THF__Single_Point = "wB97M-V/def2-TZVPPD/PCM(THF) Single Point"
    wB97M_V_def2_TZVPPD_PCM_THF__Force = "wB97M-V/def2-TZVPPD/PCM(THF) Force"
    wB97M_V_def2_TZVPPD_PCM_THF__Geometry_Optimization = "wB97M-V/def2-TZVPPD/PCM(THF) Geometry Optimization"
    wB97M_V_def2_TZVPPD_PCM_THF__Frequency_Analysis = "wB97M-V/def2-TZVPPD/PCM(THF) Frequency Analysis"
    wB97M_V_def2_TZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_THF__Unknown = "wB97M-V/def2-TZVPPD/PCM(THF) Unknown"
    wB97M_V_def2_TZVPPD_PCM_G2__Single_Point = "wB97M-V/def2-TZVPPD/PCM(G2) Single Point"
    wB97M_V_def2_TZVPPD_PCM_G2__Force = "wB97M-V/def2-TZVPPD/PCM(G2) Force"
    wB97M_V_def2_TZVPPD_PCM_G2__Geometry_Optimization = "wB97M-V/def2-TZVPPD/PCM(G2) Geometry Optimization"
    wB97M_V_def2_TZVPPD_PCM_G2__Frequency_Analysis = "wB97M-V/def2-TZVPPD/PCM(G2) Frequency Analysis"
    wB97M_V_def2_TZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_G2__Unknown = "wB97M-V/def2-TZVPPD/PCM(G2) Unknown"
    wB97M_V_def2_TZVPPD_PCM_DME__Single_Point = "wB97M-V/def2-TZVPPD/PCM(DME) Single Point"
    wB97M_V_def2_TZVPPD_PCM_DME__Force = "wB97M-V/def2-TZVPPD/PCM(DME) Force"
    wB97M_V_def2_TZVPPD_PCM_DME__Geometry_Optimization = "wB97M-V/def2-TZVPPD/PCM(DME) Geometry Optimization"
    wB97M_V_def2_TZVPPD_PCM_DME__Frequency_Analysis = "wB97M-V/def2-TZVPPD/PCM(DME) Frequency Analysis"
    wB97M_V_def2_TZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_DME__Unknown = "wB97M-V/def2-TZVPPD/PCM(DME) Unknown"
    wB97M_V_def2_TZVPPD_PCM_EC__Single_Point = "wB97M-V/def2-TZVPPD/PCM(EC) Single Point"
    wB97M_V_def2_TZVPPD_PCM_EC__Force = "wB97M-V/def2-TZVPPD/PCM(EC) Force"
    wB97M_V_def2_TZVPPD_PCM_EC__Geometry_Optimization = "wB97M-V/def2-TZVPPD/PCM(EC) Geometry Optimization"
    wB97M_V_def2_TZVPPD_PCM_EC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/PCM(EC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_EC__Unknown = "wB97M-V/def2-TZVPPD/PCM(EC) Unknown"
    wB97M_V_def2_TZVPPD_PCM_EC_EMC__Single_Point = "wB97M-V/def2-TZVPPD/PCM(EC/EMC) Single Point"
    wB97M_V_def2_TZVPPD_PCM_EC_EMC__Force = "wB97M-V/def2-TZVPPD/PCM(EC/EMC) Force"
    wB97M_V_def2_TZVPPD_PCM_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVPPD/PCM(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVPPD_PCM_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/PCM(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_EC_EMC__Unknown = "wB97M-V/def2-TZVPPD/PCM(EC/EMC) Unknown"
    wB97M_V_def2_TZVPPD_PCM_PC__Single_Point = "wB97M-V/def2-TZVPPD/PCM(PC) Single Point"
    wB97M_V_def2_TZVPPD_PCM_PC__Force = "wB97M-V/def2-TZVPPD/PCM(PC) Force"
    wB97M_V_def2_TZVPPD_PCM_PC__Geometry_Optimization = "wB97M-V/def2-TZVPPD/PCM(PC) Geometry Optimization"
    wB97M_V_def2_TZVPPD_PCM_PC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/PCM(PC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_PCM_PC__Unknown = "wB97M-V/def2-TZVPPD/PCM(PC) Unknown"
    wB97M_V_def2_TZVPPD_ISOSVP_WATER__Single_Point = "wB97M-V/def2-TZVPPD/ISOSVP(WATER) Single Point"
    wB97M_V_def2_TZVPPD_ISOSVP_WATER__Force = "wB97M-V/def2-TZVPPD/ISOSVP(WATER) Force"
    wB97M_V_def2_TZVPPD_ISOSVP_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPPD/ISOSVP(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPPD_ISOSVP_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPPD/ISOSVP(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_WATER__Unknown = "wB97M-V/def2-TZVPPD/ISOSVP(WATER) Unknown"
    wB97M_V_def2_TZVPPD_ISOSVP_THF__Single_Point = "wB97M-V/def2-TZVPPD/ISOSVP(THF) Single Point"
    wB97M_V_def2_TZVPPD_ISOSVP_THF__Force = "wB97M-V/def2-TZVPPD/ISOSVP(THF) Force"
    wB97M_V_def2_TZVPPD_ISOSVP_THF__Geometry_Optimization = "wB97M-V/def2-TZVPPD/ISOSVP(THF) Geometry Optimization"
    wB97M_V_def2_TZVPPD_ISOSVP_THF__Frequency_Analysis = "wB97M-V/def2-TZVPPD/ISOSVP(THF) Frequency Analysis"
    wB97M_V_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_THF__Unknown = "wB97M-V/def2-TZVPPD/ISOSVP(THF) Unknown"
    wB97M_V_def2_TZVPPD_ISOSVP_G2__Single_Point = "wB97M-V/def2-TZVPPD/ISOSVP(G2) Single Point"
    wB97M_V_def2_TZVPPD_ISOSVP_G2__Force = "wB97M-V/def2-TZVPPD/ISOSVP(G2) Force"
    wB97M_V_def2_TZVPPD_ISOSVP_G2__Geometry_Optimization = "wB97M-V/def2-TZVPPD/ISOSVP(G2) Geometry Optimization"
    wB97M_V_def2_TZVPPD_ISOSVP_G2__Frequency_Analysis = "wB97M-V/def2-TZVPPD/ISOSVP(G2) Frequency Analysis"
    wB97M_V_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_G2__Unknown = "wB97M-V/def2-TZVPPD/ISOSVP(G2) Unknown"
    wB97M_V_def2_TZVPPD_ISOSVP_DME__Single_Point = "wB97M-V/def2-TZVPPD/ISOSVP(DME) Single Point"
    wB97M_V_def2_TZVPPD_ISOSVP_DME__Force = "wB97M-V/def2-TZVPPD/ISOSVP(DME) Force"
    wB97M_V_def2_TZVPPD_ISOSVP_DME__Geometry_Optimization = "wB97M-V/def2-TZVPPD/ISOSVP(DME) Geometry Optimization"
    wB97M_V_def2_TZVPPD_ISOSVP_DME__Frequency_Analysis = "wB97M-V/def2-TZVPPD/ISOSVP(DME) Frequency Analysis"
    wB97M_V_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_DME__Unknown = "wB97M-V/def2-TZVPPD/ISOSVP(DME) Unknown"
    wB97M_V_def2_TZVPPD_ISOSVP_EC__Single_Point = "wB97M-V/def2-TZVPPD/ISOSVP(EC) Single Point"
    wB97M_V_def2_TZVPPD_ISOSVP_EC__Force = "wB97M-V/def2-TZVPPD/ISOSVP(EC) Force"
    wB97M_V_def2_TZVPPD_ISOSVP_EC__Geometry_Optimization = "wB97M-V/def2-TZVPPD/ISOSVP(EC) Geometry Optimization"
    wB97M_V_def2_TZVPPD_ISOSVP_EC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/ISOSVP(EC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_EC__Unknown = "wB97M-V/def2-TZVPPD/ISOSVP(EC) Unknown"
    wB97M_V_def2_TZVPPD_ISOSVP_EC_EMC__Single_Point = "wB97M-V/def2-TZVPPD/ISOSVP(EC/EMC) Single Point"
    wB97M_V_def2_TZVPPD_ISOSVP_EC_EMC__Force = "wB97M-V/def2-TZVPPD/ISOSVP(EC/EMC) Force"
    wB97M_V_def2_TZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_EC_EMC__Unknown = "wB97M-V/def2-TZVPPD/ISOSVP(EC/EMC) Unknown"
    wB97M_V_def2_TZVPPD_ISOSVP_PC__Single_Point = "wB97M-V/def2-TZVPPD/ISOSVP(PC) Single Point"
    wB97M_V_def2_TZVPPD_ISOSVP_PC__Force = "wB97M-V/def2-TZVPPD/ISOSVP(PC) Force"
    wB97M_V_def2_TZVPPD_ISOSVP_PC__Geometry_Optimization = "wB97M-V/def2-TZVPPD/ISOSVP(PC) Geometry Optimization"
    wB97M_V_def2_TZVPPD_ISOSVP_PC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/ISOSVP(PC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_ISOSVP_PC__Unknown = "wB97M-V/def2-TZVPPD/ISOSVP(PC) Unknown"
    wB97M_V_def2_TZVPPD_SMD_WATER__Single_Point = "wB97M-V/def2-TZVPPD/SMD(WATER) Single Point"
    wB97M_V_def2_TZVPPD_SMD_WATER__Force = "wB97M-V/def2-TZVPPD/SMD(WATER) Force"
    wB97M_V_def2_TZVPPD_SMD_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPPD/SMD(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPPD_SMD_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPPD/SMD(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_WATER__Unknown = "wB97M-V/def2-TZVPPD/SMD(WATER) Unknown"
    wB97M_V_def2_TZVPPD_SMD_THF__Single_Point = "wB97M-V/def2-TZVPPD/SMD(THF) Single Point"
    wB97M_V_def2_TZVPPD_SMD_THF__Force = "wB97M-V/def2-TZVPPD/SMD(THF) Force"
    wB97M_V_def2_TZVPPD_SMD_THF__Geometry_Optimization = "wB97M-V/def2-TZVPPD/SMD(THF) Geometry Optimization"
    wB97M_V_def2_TZVPPD_SMD_THF__Frequency_Analysis = "wB97M-V/def2-TZVPPD/SMD(THF) Frequency Analysis"
    wB97M_V_def2_TZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_THF__Unknown = "wB97M-V/def2-TZVPPD/SMD(THF) Unknown"
    wB97M_V_def2_TZVPPD_SMD_G2__Single_Point = "wB97M-V/def2-TZVPPD/SMD(G2) Single Point"
    wB97M_V_def2_TZVPPD_SMD_G2__Force = "wB97M-V/def2-TZVPPD/SMD(G2) Force"
    wB97M_V_def2_TZVPPD_SMD_G2__Geometry_Optimization = "wB97M-V/def2-TZVPPD/SMD(G2) Geometry Optimization"
    wB97M_V_def2_TZVPPD_SMD_G2__Frequency_Analysis = "wB97M-V/def2-TZVPPD/SMD(G2) Frequency Analysis"
    wB97M_V_def2_TZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_G2__Unknown = "wB97M-V/def2-TZVPPD/SMD(G2) Unknown"
    wB97M_V_def2_TZVPPD_SMD_DME__Single_Point = "wB97M-V/def2-TZVPPD/SMD(DME) Single Point"
    wB97M_V_def2_TZVPPD_SMD_DME__Force = "wB97M-V/def2-TZVPPD/SMD(DME) Force"
    wB97M_V_def2_TZVPPD_SMD_DME__Geometry_Optimization = "wB97M-V/def2-TZVPPD/SMD(DME) Geometry Optimization"
    wB97M_V_def2_TZVPPD_SMD_DME__Frequency_Analysis = "wB97M-V/def2-TZVPPD/SMD(DME) Frequency Analysis"
    wB97M_V_def2_TZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_DME__Unknown = "wB97M-V/def2-TZVPPD/SMD(DME) Unknown"
    wB97M_V_def2_TZVPPD_SMD_EC__Single_Point = "wB97M-V/def2-TZVPPD/SMD(EC) Single Point"
    wB97M_V_def2_TZVPPD_SMD_EC__Force = "wB97M-V/def2-TZVPPD/SMD(EC) Force"
    wB97M_V_def2_TZVPPD_SMD_EC__Geometry_Optimization = "wB97M-V/def2-TZVPPD/SMD(EC) Geometry Optimization"
    wB97M_V_def2_TZVPPD_SMD_EC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/SMD(EC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_EC__Unknown = "wB97M-V/def2-TZVPPD/SMD(EC) Unknown"
    wB97M_V_def2_TZVPPD_SMD_EC_EMC__Single_Point = "wB97M-V/def2-TZVPPD/SMD(EC/EMC) Single Point"
    wB97M_V_def2_TZVPPD_SMD_EC_EMC__Force = "wB97M-V/def2-TZVPPD/SMD(EC/EMC) Force"
    wB97M_V_def2_TZVPPD_SMD_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVPPD/SMD(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVPPD_SMD_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/SMD(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_EC_EMC__Unknown = "wB97M-V/def2-TZVPPD/SMD(EC/EMC) Unknown"
    wB97M_V_def2_TZVPPD_SMD_PC__Single_Point = "wB97M-V/def2-TZVPPD/SMD(PC) Single Point"
    wB97M_V_def2_TZVPPD_SMD_PC__Force = "wB97M-V/def2-TZVPPD/SMD(PC) Force"
    wB97M_V_def2_TZVPPD_SMD_PC__Geometry_Optimization = "wB97M-V/def2-TZVPPD/SMD(PC) Geometry Optimization"
    wB97M_V_def2_TZVPPD_SMD_PC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/SMD(PC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_SMD_PC__Unknown = "wB97M-V/def2-TZVPPD/SMD(PC) Unknown"
    wB97M_V_def2_TZVPPD_CMIRS_WATER__Single_Point = "wB97M-V/def2-TZVPPD/CMIRS(WATER) Single Point"
    wB97M_V_def2_TZVPPD_CMIRS_WATER__Force = "wB97M-V/def2-TZVPPD/CMIRS(WATER) Force"
    wB97M_V_def2_TZVPPD_CMIRS_WATER__Geometry_Optimization = "wB97M-V/def2-TZVPPD/CMIRS(WATER) Geometry Optimization"
    wB97M_V_def2_TZVPPD_CMIRS_WATER__Frequency_Analysis = "wB97M-V/def2-TZVPPD/CMIRS(WATER) Frequency Analysis"
    wB97M_V_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_WATER__Unknown = "wB97M-V/def2-TZVPPD/CMIRS(WATER) Unknown"
    wB97M_V_def2_TZVPPD_CMIRS_THF__Single_Point = "wB97M-V/def2-TZVPPD/CMIRS(THF) Single Point"
    wB97M_V_def2_TZVPPD_CMIRS_THF__Force = "wB97M-V/def2-TZVPPD/CMIRS(THF) Force"
    wB97M_V_def2_TZVPPD_CMIRS_THF__Geometry_Optimization = "wB97M-V/def2-TZVPPD/CMIRS(THF) Geometry Optimization"
    wB97M_V_def2_TZVPPD_CMIRS_THF__Frequency_Analysis = "wB97M-V/def2-TZVPPD/CMIRS(THF) Frequency Analysis"
    wB97M_V_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_THF__Unknown = "wB97M-V/def2-TZVPPD/CMIRS(THF) Unknown"
    wB97M_V_def2_TZVPPD_CMIRS_G2__Single_Point = "wB97M-V/def2-TZVPPD/CMIRS(G2) Single Point"
    wB97M_V_def2_TZVPPD_CMIRS_G2__Force = "wB97M-V/def2-TZVPPD/CMIRS(G2) Force"
    wB97M_V_def2_TZVPPD_CMIRS_G2__Geometry_Optimization = "wB97M-V/def2-TZVPPD/CMIRS(G2) Geometry Optimization"
    wB97M_V_def2_TZVPPD_CMIRS_G2__Frequency_Analysis = "wB97M-V/def2-TZVPPD/CMIRS(G2) Frequency Analysis"
    wB97M_V_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_G2__Unknown = "wB97M-V/def2-TZVPPD/CMIRS(G2) Unknown"
    wB97M_V_def2_TZVPPD_CMIRS_DME__Single_Point = "wB97M-V/def2-TZVPPD/CMIRS(DME) Single Point"
    wB97M_V_def2_TZVPPD_CMIRS_DME__Force = "wB97M-V/def2-TZVPPD/CMIRS(DME) Force"
    wB97M_V_def2_TZVPPD_CMIRS_DME__Geometry_Optimization = "wB97M-V/def2-TZVPPD/CMIRS(DME) Geometry Optimization"
    wB97M_V_def2_TZVPPD_CMIRS_DME__Frequency_Analysis = "wB97M-V/def2-TZVPPD/CMIRS(DME) Frequency Analysis"
    wB97M_V_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_DME__Unknown = "wB97M-V/def2-TZVPPD/CMIRS(DME) Unknown"
    wB97M_V_def2_TZVPPD_CMIRS_EC__Single_Point = "wB97M-V/def2-TZVPPD/CMIRS(EC) Single Point"
    wB97M_V_def2_TZVPPD_CMIRS_EC__Force = "wB97M-V/def2-TZVPPD/CMIRS(EC) Force"
    wB97M_V_def2_TZVPPD_CMIRS_EC__Geometry_Optimization = "wB97M-V/def2-TZVPPD/CMIRS(EC) Geometry Optimization"
    wB97M_V_def2_TZVPPD_CMIRS_EC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/CMIRS(EC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_EC__Unknown = "wB97M-V/def2-TZVPPD/CMIRS(EC) Unknown"
    wB97M_V_def2_TZVPPD_CMIRS_EC_EMC__Single_Point = "wB97M-V/def2-TZVPPD/CMIRS(EC/EMC) Single Point"
    wB97M_V_def2_TZVPPD_CMIRS_EC_EMC__Force = "wB97M-V/def2-TZVPPD/CMIRS(EC/EMC) Force"
    wB97M_V_def2_TZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97M-V/def2-TZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97M_V_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_EC_EMC__Unknown = "wB97M-V/def2-TZVPPD/CMIRS(EC/EMC) Unknown"
    wB97M_V_def2_TZVPPD_CMIRS_PC__Single_Point = "wB97M-V/def2-TZVPPD/CMIRS(PC) Single Point"
    wB97M_V_def2_TZVPPD_CMIRS_PC__Force = "wB97M-V/def2-TZVPPD/CMIRS(PC) Force"
    wB97M_V_def2_TZVPPD_CMIRS_PC__Geometry_Optimization = "wB97M-V/def2-TZVPPD/CMIRS(PC) Geometry Optimization"
    wB97M_V_def2_TZVPPD_CMIRS_PC__Frequency_Analysis = "wB97M-V/def2-TZVPPD/CMIRS(PC) Frequency Analysis"
    wB97M_V_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-TZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_TZVPPD_CMIRS_PC__Unknown = "wB97M-V/def2-TZVPPD/CMIRS(PC) Unknown"
    wB97M_V_def2_QZVPPD_VACUUM_Single_Point = "wB97M-V/def2-QZVPPD/VACUUM Single Point"
    wB97M_V_def2_QZVPPD_VACUUM_Force = "wB97M-V/def2-QZVPPD/VACUUM Force"
    wB97M_V_def2_QZVPPD_VACUUM_Geometry_Optimization = "wB97M-V/def2-QZVPPD/VACUUM Geometry Optimization"
    wB97M_V_def2_QZVPPD_VACUUM_Frequency_Analysis = "wB97M-V/def2-QZVPPD/VACUUM Frequency Analysis"
    wB97M_V_def2_QZVPPD_VACUUM_Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/VACUUM Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_VACUUM_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/VACUUM Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_VACUUM_Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/VACUUM Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_VACUUM_Unknown = "wB97M-V/def2-QZVPPD/VACUUM Unknown"
    wB97M_V_def2_QZVPPD_PCM_WATER__Single_Point = "wB97M-V/def2-QZVPPD/PCM(WATER) Single Point"
    wB97M_V_def2_QZVPPD_PCM_WATER__Force = "wB97M-V/def2-QZVPPD/PCM(WATER) Force"
    wB97M_V_def2_QZVPPD_PCM_WATER__Geometry_Optimization = "wB97M-V/def2-QZVPPD/PCM(WATER) Geometry Optimization"
    wB97M_V_def2_QZVPPD_PCM_WATER__Frequency_Analysis = "wB97M-V/def2-QZVPPD/PCM(WATER) Frequency Analysis"
    wB97M_V_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_WATER__Unknown = "wB97M-V/def2-QZVPPD/PCM(WATER) Unknown"
    wB97M_V_def2_QZVPPD_PCM_THF__Single_Point = "wB97M-V/def2-QZVPPD/PCM(THF) Single Point"
    wB97M_V_def2_QZVPPD_PCM_THF__Force = "wB97M-V/def2-QZVPPD/PCM(THF) Force"
    wB97M_V_def2_QZVPPD_PCM_THF__Geometry_Optimization = "wB97M-V/def2-QZVPPD/PCM(THF) Geometry Optimization"
    wB97M_V_def2_QZVPPD_PCM_THF__Frequency_Analysis = "wB97M-V/def2-QZVPPD/PCM(THF) Frequency Analysis"
    wB97M_V_def2_QZVPPD_PCM_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_THF__Unknown = "wB97M-V/def2-QZVPPD/PCM(THF) Unknown"
    wB97M_V_def2_QZVPPD_PCM_G2__Single_Point = "wB97M-V/def2-QZVPPD/PCM(G2) Single Point"
    wB97M_V_def2_QZVPPD_PCM_G2__Force = "wB97M-V/def2-QZVPPD/PCM(G2) Force"
    wB97M_V_def2_QZVPPD_PCM_G2__Geometry_Optimization = "wB97M-V/def2-QZVPPD/PCM(G2) Geometry Optimization"
    wB97M_V_def2_QZVPPD_PCM_G2__Frequency_Analysis = "wB97M-V/def2-QZVPPD/PCM(G2) Frequency Analysis"
    wB97M_V_def2_QZVPPD_PCM_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_G2__Unknown = "wB97M-V/def2-QZVPPD/PCM(G2) Unknown"
    wB97M_V_def2_QZVPPD_PCM_DME__Single_Point = "wB97M-V/def2-QZVPPD/PCM(DME) Single Point"
    wB97M_V_def2_QZVPPD_PCM_DME__Force = "wB97M-V/def2-QZVPPD/PCM(DME) Force"
    wB97M_V_def2_QZVPPD_PCM_DME__Geometry_Optimization = "wB97M-V/def2-QZVPPD/PCM(DME) Geometry Optimization"
    wB97M_V_def2_QZVPPD_PCM_DME__Frequency_Analysis = "wB97M-V/def2-QZVPPD/PCM(DME) Frequency Analysis"
    wB97M_V_def2_QZVPPD_PCM_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_DME__Unknown = "wB97M-V/def2-QZVPPD/PCM(DME) Unknown"
    wB97M_V_def2_QZVPPD_PCM_EC__Single_Point = "wB97M-V/def2-QZVPPD/PCM(EC) Single Point"
    wB97M_V_def2_QZVPPD_PCM_EC__Force = "wB97M-V/def2-QZVPPD/PCM(EC) Force"
    wB97M_V_def2_QZVPPD_PCM_EC__Geometry_Optimization = "wB97M-V/def2-QZVPPD/PCM(EC) Geometry Optimization"
    wB97M_V_def2_QZVPPD_PCM_EC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/PCM(EC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_PCM_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_EC__Unknown = "wB97M-V/def2-QZVPPD/PCM(EC) Unknown"
    wB97M_V_def2_QZVPPD_PCM_EC_EMC__Single_Point = "wB97M-V/def2-QZVPPD/PCM(EC/EMC) Single Point"
    wB97M_V_def2_QZVPPD_PCM_EC_EMC__Force = "wB97M-V/def2-QZVPPD/PCM(EC/EMC) Force"
    wB97M_V_def2_QZVPPD_PCM_EC_EMC__Geometry_Optimization = "wB97M-V/def2-QZVPPD/PCM(EC/EMC) Geometry Optimization"
    wB97M_V_def2_QZVPPD_PCM_EC_EMC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/PCM(EC/EMC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_EC_EMC__Unknown = "wB97M-V/def2-QZVPPD/PCM(EC/EMC) Unknown"
    wB97M_V_def2_QZVPPD_PCM_PC__Single_Point = "wB97M-V/def2-QZVPPD/PCM(PC) Single Point"
    wB97M_V_def2_QZVPPD_PCM_PC__Force = "wB97M-V/def2-QZVPPD/PCM(PC) Force"
    wB97M_V_def2_QZVPPD_PCM_PC__Geometry_Optimization = "wB97M-V/def2-QZVPPD/PCM(PC) Geometry Optimization"
    wB97M_V_def2_QZVPPD_PCM_PC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/PCM(PC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_PCM_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/PCM(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_PCM_PC__Unknown = "wB97M-V/def2-QZVPPD/PCM(PC) Unknown"
    wB97M_V_def2_QZVPPD_ISOSVP_WATER__Single_Point = "wB97M-V/def2-QZVPPD/ISOSVP(WATER) Single Point"
    wB97M_V_def2_QZVPPD_ISOSVP_WATER__Force = "wB97M-V/def2-QZVPPD/ISOSVP(WATER) Force"
    wB97M_V_def2_QZVPPD_ISOSVP_WATER__Geometry_Optimization = "wB97M-V/def2-QZVPPD/ISOSVP(WATER) Geometry Optimization"
    wB97M_V_def2_QZVPPD_ISOSVP_WATER__Frequency_Analysis = "wB97M-V/def2-QZVPPD/ISOSVP(WATER) Frequency Analysis"
    wB97M_V_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_WATER__Unknown = "wB97M-V/def2-QZVPPD/ISOSVP(WATER) Unknown"
    wB97M_V_def2_QZVPPD_ISOSVP_THF__Single_Point = "wB97M-V/def2-QZVPPD/ISOSVP(THF) Single Point"
    wB97M_V_def2_QZVPPD_ISOSVP_THF__Force = "wB97M-V/def2-QZVPPD/ISOSVP(THF) Force"
    wB97M_V_def2_QZVPPD_ISOSVP_THF__Geometry_Optimization = "wB97M-V/def2-QZVPPD/ISOSVP(THF) Geometry Optimization"
    wB97M_V_def2_QZVPPD_ISOSVP_THF__Frequency_Analysis = "wB97M-V/def2-QZVPPD/ISOSVP(THF) Frequency Analysis"
    wB97M_V_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_THF__Unknown = "wB97M-V/def2-QZVPPD/ISOSVP(THF) Unknown"
    wB97M_V_def2_QZVPPD_ISOSVP_G2__Single_Point = "wB97M-V/def2-QZVPPD/ISOSVP(G2) Single Point"
    wB97M_V_def2_QZVPPD_ISOSVP_G2__Force = "wB97M-V/def2-QZVPPD/ISOSVP(G2) Force"
    wB97M_V_def2_QZVPPD_ISOSVP_G2__Geometry_Optimization = "wB97M-V/def2-QZVPPD/ISOSVP(G2) Geometry Optimization"
    wB97M_V_def2_QZVPPD_ISOSVP_G2__Frequency_Analysis = "wB97M-V/def2-QZVPPD/ISOSVP(G2) Frequency Analysis"
    wB97M_V_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_G2__Unknown = "wB97M-V/def2-QZVPPD/ISOSVP(G2) Unknown"
    wB97M_V_def2_QZVPPD_ISOSVP_DME__Single_Point = "wB97M-V/def2-QZVPPD/ISOSVP(DME) Single Point"
    wB97M_V_def2_QZVPPD_ISOSVP_DME__Force = "wB97M-V/def2-QZVPPD/ISOSVP(DME) Force"
    wB97M_V_def2_QZVPPD_ISOSVP_DME__Geometry_Optimization = "wB97M-V/def2-QZVPPD/ISOSVP(DME) Geometry Optimization"
    wB97M_V_def2_QZVPPD_ISOSVP_DME__Frequency_Analysis = "wB97M-V/def2-QZVPPD/ISOSVP(DME) Frequency Analysis"
    wB97M_V_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_DME__Unknown = "wB97M-V/def2-QZVPPD/ISOSVP(DME) Unknown"
    wB97M_V_def2_QZVPPD_ISOSVP_EC__Single_Point = "wB97M-V/def2-QZVPPD/ISOSVP(EC) Single Point"
    wB97M_V_def2_QZVPPD_ISOSVP_EC__Force = "wB97M-V/def2-QZVPPD/ISOSVP(EC) Force"
    wB97M_V_def2_QZVPPD_ISOSVP_EC__Geometry_Optimization = "wB97M-V/def2-QZVPPD/ISOSVP(EC) Geometry Optimization"
    wB97M_V_def2_QZVPPD_ISOSVP_EC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/ISOSVP(EC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_EC__Unknown = "wB97M-V/def2-QZVPPD/ISOSVP(EC) Unknown"
    wB97M_V_def2_QZVPPD_ISOSVP_EC_EMC__Single_Point = "wB97M-V/def2-QZVPPD/ISOSVP(EC/EMC) Single Point"
    wB97M_V_def2_QZVPPD_ISOSVP_EC_EMC__Force = "wB97M-V/def2-QZVPPD/ISOSVP(EC/EMC) Force"
    wB97M_V_def2_QZVPPD_ISOSVP_EC_EMC__Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(EC/EMC) Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_EC_EMC__Unknown = "wB97M-V/def2-QZVPPD/ISOSVP(EC/EMC) Unknown"
    wB97M_V_def2_QZVPPD_ISOSVP_PC__Single_Point = "wB97M-V/def2-QZVPPD/ISOSVP(PC) Single Point"
    wB97M_V_def2_QZVPPD_ISOSVP_PC__Force = "wB97M-V/def2-QZVPPD/ISOSVP(PC) Force"
    wB97M_V_def2_QZVPPD_ISOSVP_PC__Geometry_Optimization = "wB97M-V/def2-QZVPPD/ISOSVP(PC) Geometry Optimization"
    wB97M_V_def2_QZVPPD_ISOSVP_PC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/ISOSVP(PC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/ISOSVP(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_ISOSVP_PC__Unknown = "wB97M-V/def2-QZVPPD/ISOSVP(PC) Unknown"
    wB97M_V_def2_QZVPPD_SMD_WATER__Single_Point = "wB97M-V/def2-QZVPPD/SMD(WATER) Single Point"
    wB97M_V_def2_QZVPPD_SMD_WATER__Force = "wB97M-V/def2-QZVPPD/SMD(WATER) Force"
    wB97M_V_def2_QZVPPD_SMD_WATER__Geometry_Optimization = "wB97M-V/def2-QZVPPD/SMD(WATER) Geometry Optimization"
    wB97M_V_def2_QZVPPD_SMD_WATER__Frequency_Analysis = "wB97M-V/def2-QZVPPD/SMD(WATER) Frequency Analysis"
    wB97M_V_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_WATER__Unknown = "wB97M-V/def2-QZVPPD/SMD(WATER) Unknown"
    wB97M_V_def2_QZVPPD_SMD_THF__Single_Point = "wB97M-V/def2-QZVPPD/SMD(THF) Single Point"
    wB97M_V_def2_QZVPPD_SMD_THF__Force = "wB97M-V/def2-QZVPPD/SMD(THF) Force"
    wB97M_V_def2_QZVPPD_SMD_THF__Geometry_Optimization = "wB97M-V/def2-QZVPPD/SMD(THF) Geometry Optimization"
    wB97M_V_def2_QZVPPD_SMD_THF__Frequency_Analysis = "wB97M-V/def2-QZVPPD/SMD(THF) Frequency Analysis"
    wB97M_V_def2_QZVPPD_SMD_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_THF__Unknown = "wB97M-V/def2-QZVPPD/SMD(THF) Unknown"
    wB97M_V_def2_QZVPPD_SMD_G2__Single_Point = "wB97M-V/def2-QZVPPD/SMD(G2) Single Point"
    wB97M_V_def2_QZVPPD_SMD_G2__Force = "wB97M-V/def2-QZVPPD/SMD(G2) Force"
    wB97M_V_def2_QZVPPD_SMD_G2__Geometry_Optimization = "wB97M-V/def2-QZVPPD/SMD(G2) Geometry Optimization"
    wB97M_V_def2_QZVPPD_SMD_G2__Frequency_Analysis = "wB97M-V/def2-QZVPPD/SMD(G2) Frequency Analysis"
    wB97M_V_def2_QZVPPD_SMD_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_G2__Unknown = "wB97M-V/def2-QZVPPD/SMD(G2) Unknown"
    wB97M_V_def2_QZVPPD_SMD_DME__Single_Point = "wB97M-V/def2-QZVPPD/SMD(DME) Single Point"
    wB97M_V_def2_QZVPPD_SMD_DME__Force = "wB97M-V/def2-QZVPPD/SMD(DME) Force"
    wB97M_V_def2_QZVPPD_SMD_DME__Geometry_Optimization = "wB97M-V/def2-QZVPPD/SMD(DME) Geometry Optimization"
    wB97M_V_def2_QZVPPD_SMD_DME__Frequency_Analysis = "wB97M-V/def2-QZVPPD/SMD(DME) Frequency Analysis"
    wB97M_V_def2_QZVPPD_SMD_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_DME__Unknown = "wB97M-V/def2-QZVPPD/SMD(DME) Unknown"
    wB97M_V_def2_QZVPPD_SMD_EC__Single_Point = "wB97M-V/def2-QZVPPD/SMD(EC) Single Point"
    wB97M_V_def2_QZVPPD_SMD_EC__Force = "wB97M-V/def2-QZVPPD/SMD(EC) Force"
    wB97M_V_def2_QZVPPD_SMD_EC__Geometry_Optimization = "wB97M-V/def2-QZVPPD/SMD(EC) Geometry Optimization"
    wB97M_V_def2_QZVPPD_SMD_EC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/SMD(EC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_SMD_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_EC__Unknown = "wB97M-V/def2-QZVPPD/SMD(EC) Unknown"
    wB97M_V_def2_QZVPPD_SMD_EC_EMC__Single_Point = "wB97M-V/def2-QZVPPD/SMD(EC/EMC) Single Point"
    wB97M_V_def2_QZVPPD_SMD_EC_EMC__Force = "wB97M-V/def2-QZVPPD/SMD(EC/EMC) Force"
    wB97M_V_def2_QZVPPD_SMD_EC_EMC__Geometry_Optimization = "wB97M-V/def2-QZVPPD/SMD(EC/EMC) Geometry Optimization"
    wB97M_V_def2_QZVPPD_SMD_EC_EMC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/SMD(EC/EMC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_EC_EMC__Unknown = "wB97M-V/def2-QZVPPD/SMD(EC/EMC) Unknown"
    wB97M_V_def2_QZVPPD_SMD_PC__Single_Point = "wB97M-V/def2-QZVPPD/SMD(PC) Single Point"
    wB97M_V_def2_QZVPPD_SMD_PC__Force = "wB97M-V/def2-QZVPPD/SMD(PC) Force"
    wB97M_V_def2_QZVPPD_SMD_PC__Geometry_Optimization = "wB97M-V/def2-QZVPPD/SMD(PC) Geometry Optimization"
    wB97M_V_def2_QZVPPD_SMD_PC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/SMD(PC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_SMD_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/SMD(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_SMD_PC__Unknown = "wB97M-V/def2-QZVPPD/SMD(PC) Unknown"
    wB97M_V_def2_QZVPPD_CMIRS_WATER__Single_Point = "wB97M-V/def2-QZVPPD/CMIRS(WATER) Single Point"
    wB97M_V_def2_QZVPPD_CMIRS_WATER__Force = "wB97M-V/def2-QZVPPD/CMIRS(WATER) Force"
    wB97M_V_def2_QZVPPD_CMIRS_WATER__Geometry_Optimization = "wB97M-V/def2-QZVPPD/CMIRS(WATER) Geometry Optimization"
    wB97M_V_def2_QZVPPD_CMIRS_WATER__Frequency_Analysis = "wB97M-V/def2-QZVPPD/CMIRS(WATER) Frequency Analysis"
    wB97M_V_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_WATER__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(WATER) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_WATER__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(WATER) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_WATER__Unknown = "wB97M-V/def2-QZVPPD/CMIRS(WATER) Unknown"
    wB97M_V_def2_QZVPPD_CMIRS_THF__Single_Point = "wB97M-V/def2-QZVPPD/CMIRS(THF) Single Point"
    wB97M_V_def2_QZVPPD_CMIRS_THF__Force = "wB97M-V/def2-QZVPPD/CMIRS(THF) Force"
    wB97M_V_def2_QZVPPD_CMIRS_THF__Geometry_Optimization = "wB97M-V/def2-QZVPPD/CMIRS(THF) Geometry Optimization"
    wB97M_V_def2_QZVPPD_CMIRS_THF__Frequency_Analysis = "wB97M-V/def2-QZVPPD/CMIRS(THF) Frequency Analysis"
    wB97M_V_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(THF) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_THF__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(THF) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_THF__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(THF) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_THF__Unknown = "wB97M-V/def2-QZVPPD/CMIRS(THF) Unknown"
    wB97M_V_def2_QZVPPD_CMIRS_G2__Single_Point = "wB97M-V/def2-QZVPPD/CMIRS(G2) Single Point"
    wB97M_V_def2_QZVPPD_CMIRS_G2__Force = "wB97M-V/def2-QZVPPD/CMIRS(G2) Force"
    wB97M_V_def2_QZVPPD_CMIRS_G2__Geometry_Optimization = "wB97M-V/def2-QZVPPD/CMIRS(G2) Geometry Optimization"
    wB97M_V_def2_QZVPPD_CMIRS_G2__Frequency_Analysis = "wB97M-V/def2-QZVPPD/CMIRS(G2) Frequency Analysis"
    wB97M_V_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(G2) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_G2__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(G2) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_G2__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(G2) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_G2__Unknown = "wB97M-V/def2-QZVPPD/CMIRS(G2) Unknown"
    wB97M_V_def2_QZVPPD_CMIRS_DME__Single_Point = "wB97M-V/def2-QZVPPD/CMIRS(DME) Single Point"
    wB97M_V_def2_QZVPPD_CMIRS_DME__Force = "wB97M-V/def2-QZVPPD/CMIRS(DME) Force"
    wB97M_V_def2_QZVPPD_CMIRS_DME__Geometry_Optimization = "wB97M-V/def2-QZVPPD/CMIRS(DME) Geometry Optimization"
    wB97M_V_def2_QZVPPD_CMIRS_DME__Frequency_Analysis = "wB97M-V/def2-QZVPPD/CMIRS(DME) Frequency Analysis"
    wB97M_V_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(DME) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_DME__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(DME) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_DME__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(DME) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_DME__Unknown = "wB97M-V/def2-QZVPPD/CMIRS(DME) Unknown"
    wB97M_V_def2_QZVPPD_CMIRS_EC__Single_Point = "wB97M-V/def2-QZVPPD/CMIRS(EC) Single Point"
    wB97M_V_def2_QZVPPD_CMIRS_EC__Force = "wB97M-V/def2-QZVPPD/CMIRS(EC) Force"
    wB97M_V_def2_QZVPPD_CMIRS_EC__Geometry_Optimization = "wB97M-V/def2-QZVPPD/CMIRS(EC) Geometry Optimization"
    wB97M_V_def2_QZVPPD_CMIRS_EC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/CMIRS(EC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(EC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_EC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(EC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_EC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(EC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_EC__Unknown = "wB97M-V/def2-QZVPPD/CMIRS(EC) Unknown"
    wB97M_V_def2_QZVPPD_CMIRS_EC_EMC__Single_Point = "wB97M-V/def2-QZVPPD/CMIRS(EC/EMC) Single Point"
    wB97M_V_def2_QZVPPD_CMIRS_EC_EMC__Force = "wB97M-V/def2-QZVPPD/CMIRS(EC/EMC) Force"
    wB97M_V_def2_QZVPPD_CMIRS_EC_EMC__Geometry_Optimization = "wB97M-V/def2-QZVPPD/CMIRS(EC/EMC) Geometry Optimization"
    wB97M_V_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/CMIRS(EC/EMC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_EC_EMC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(EC/EMC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_EC_EMC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(EC/EMC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_EC_EMC__Unknown = "wB97M-V/def2-QZVPPD/CMIRS(EC/EMC) Unknown"
    wB97M_V_def2_QZVPPD_CMIRS_PC__Single_Point = "wB97M-V/def2-QZVPPD/CMIRS(PC) Single Point"
    wB97M_V_def2_QZVPPD_CMIRS_PC__Force = "wB97M-V/def2-QZVPPD/CMIRS(PC) Force"
    wB97M_V_def2_QZVPPD_CMIRS_PC__Geometry_Optimization = "wB97M-V/def2-QZVPPD/CMIRS(PC) Geometry Optimization"
    wB97M_V_def2_QZVPPD_CMIRS_PC__Frequency_Analysis = "wB97M-V/def2-QZVPPD/CMIRS(PC) Frequency Analysis"
    wB97M_V_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(PC) Frequency Flattening Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_PC__Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(PC) Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_PC__Frequency_Flattening_Transition_State_Geometry_Optimization = (
        "wB97M-V/def2-QZVPPD/CMIRS(PC) Frequency Flattening Transition State Geometry Optimization"
    )
    wB97M_V_def2_QZVPPD_CMIRS_PC__Unknown = "wB97M-V/def2-QZVPPD/CMIRS(PC) Unknown"
